Package: writexl
Type: Package
Title: Export Data Frames to Excel 'xlsx' Format
Version: 1.5.4
Authors@R: c(
    person("Jeroen", "Ooms", ,"jeroenooms@gmail.com", role = c("aut", "cre"),
      comment = c(ORCID = "0000-0002-4035-0289")),
    person("John McNamara", role = "cph",
      comment = "Author of libxlsxwriter (see AUTHORS and COPYRIGHT files for details)"))
Description: Zero-dependency data frame to xlsx exporter based on 'libxlsxwriter'
    <https://libxlsxwriter.github.io>. Fast and no Java or Excel required.
License: BSD_2_clause + file LICENSE
Encoding: UTF-8
URL: https://ropensci.r-universe.dev/writexl
        https://docs.ropensci.org/writexl/
BugReports: https://github.com/ropensci/writexl/issues
RoxygenNote: 7.0.2
Suggests: spelling, readxl, nycflights13, testthat, bit64
Language: en-US
SystemRequirements: zlib
NeedsCompilation: yes
Packaged: 2025-04-15 13:34:41 UTC; jeroen
Author: Jeroen Ooms [aut, cre] (<https://orcid.org/0000-0002-4035-0289>),
  John McNamara [cph] (Author of libxlsxwriter (see AUTHORS and COPYRIGHT
    files for details))
Maintainer: Jeroen Ooms <jeroenooms@gmail.com>
Repository: CRAN
Date/Publication: 2025-04-15 14:40:03 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 09:21:15 UTC; unix
Archs: writexl.so.dSYM

Package: writexl
Type: Package
Title: Export Data Frames to Excel 'xlsx' Format
Version: 2.0.0
Authors@R: c(
    person("Jeroen", "Ooms", role = "aut", comment = c(ORCID = "0000-0002-4035-0289")),
    person("Bill", "Denney", ,"wdenney@humanpredictions.com", role = c("aut", "cre"),
      comment = c(ORCID="0000-0002-5759-428X")),
    person("John McNamara", role = "cph",
      comment = "Author of libxlsxwriter (see AUTHORS and COPYRIGHT files for details)"))
Description: Zero-dependency data frame to xlsx exporter based on 'libxlsxwriter'
    <https://libxlsxwriter.github.io>. Fast and no Java or Excel required.
License: BSD_2_clause + file LICENSE
Copyright: Jeroen Ooms. For copyright notices of bundled third-party
        code (libxlsxwriter and dependencies), see COPYRIGHT.
Encoding: UTF-8
URL: https://ropensci.r-universe.dev/writexl
        https://docs.ropensci.org/writexl/
BugReports: https://github.com/ropensci/writexl/issues
Suggests: grDevices, graphics, spelling, readxl, nycflights13, testthat
        (>= 3.2.0), xml2, bit64, knitr, rmarkdown
VignetteBuilder: knitr
Language: en-US
SystemRequirements: zlib
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-05 12:30:06 UTC; bill
Author: Jeroen Ooms [aut] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  Bill Denney [aut, cre] (ORCID: <https://orcid.org/0000-0002-5759-428X>),
  John McNamara [cph] (Author of libxlsxwriter (see AUTHORS and COPYRIGHT
    files for details))
Maintainer: Bill Denney <wdenney@humanpredictions.com>
Repository: CRAN
Date/Publication: 2026-08-05 14:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("+",xl_format)
S3method("[",xl_cell_general)
S3method("[<-",xl_cell_general)
S3method("[[",xl_cell_general)
S3method("[[<-",xl_cell_general)
S3method("length<-",xl_cell_general)
S3method(as.character,xl_cell_general)
S3method(as.character,xl_rich_string)
S3method(as.data.frame,xl_cell_general)
S3method(c,xl_cell_general)
S3method(format,xl_rich_string)
S3method(length,xl_cell_general)
S3method(print,xl_cell_general)
S3method(print,xl_chart)
S3method(print,xl_chart_axis)
S3method(print,xl_chart_error_bars)
S3method(print,xl_chart_label)
S3method(print,xl_chart_labels)
S3method(print,xl_chart_legend)
S3method(print,xl_chart_marker)
S3method(print,xl_chart_series)
S3method(print,xl_chart_table)
S3method(print,xl_chart_trendline)
S3method(print,xl_chartsheet)
S3method(print,xl_comment)
S3method(print,xl_conditional)
S3method(print,xl_filter)
S3method(print,xl_format)
S3method(print,xl_image)
S3method(print,xl_merge)
S3method(print,xl_outline)
S3method(print,xl_page_setup)
S3method(print,xl_properties)
S3method(print,xl_rich_run)
S3method(print,xl_rich_string)
S3method(print,xl_sheet)
S3method(print,xl_sheet_view)
S3method(print,xl_table)
S3method(print,xl_table_column)
S3method(print,xl_validation)
S3method(print,xl_workbook)
S3method(rep,xl_cell_general)
export(is_xl_comment)
export(is_xl_format)
export(is_xl_rich_string)
export(lxw_version)
export(write_xlsx)
export(xl_align)
export(xl_border)
export(xl_cell_general)
export(xl_chart)
export(xl_chart_axis)
export(xl_chart_error_bars)
export(xl_chart_label)
export(xl_chart_labels)
export(xl_chart_legend)
export(xl_chart_marker)
export(xl_chart_series)
export(xl_chart_table)
export(xl_chart_trendline)
export(xl_chartsheet)
export(xl_col_spec)
export(xl_color)
export(xl_comment)
export(xl_cond_bar)
export(xl_cond_cell)
export(xl_cond_icons)
export(xl_cond_scale)
export(xl_fill)
export(xl_filter)
export(xl_filter_keep)
export(xl_font)
export(xl_format)
export(xl_formula)
export(xl_hyperlink)
export(xl_hyperlink_cell)
export(xl_image)
export(xl_merge)
export(xl_num_format)
export(xl_outline)
export(xl_page_setup)
export(xl_properties)
export(xl_protection)
export(xl_rich_run)
export(xl_rich_string)
export(xl_row_spec)
export(xl_sheet)
export(xl_sheet_view)
export(xl_table)
export(xl_table_column)
export(xl_validation)
export(xl_workbook)
useDynLib(writexl,C_lxw_version)
useDynLib(writexl,C_set_tempdir)
useDynLib(writexl,C_write_data_frame_list)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)

## -----------------------------------------------------------------------------
sales <- data.frame(
  quarter = c("Q1", "Q2", "Q3", "Q4"),
  revenue = c(120000, 145000, 133000, 180000),
  cost    = c(90000, 110000, 105000, 130000),
  region  = c("North", "South", "North", "South"),
  stringsAsFactors = FALSE)

## -----------------------------------------------------------------------------
money <- xl_num_format("$#,##0") + xl_font(color = "darkgreen")

styled <- xl_sheet(
  sales,
  cols = list(xl_col_spec("quarter", width = 10),
              xl_col_spec(c("revenue", "cost"), width = 14, format = money)),
  rows = xl_row_spec(1, height = 18),
  conditional = list(
    xl_cond_cell(list(cols = "revenue"), "cell", ">=", 1.4e5,
                 format = xl_fill(background = "lightgreen")),
    xl_cond_bar(list(cols = "cost"), color = "steelblue")),
  freeze = "A2", tab_color = "steelblue")

## -----------------------------------------------------------------------------
cells <- data.frame(item = c("Widget", "Gadget", "Gizmo"),
                    stringsAsFactors = FALSE)
cells$price   <- xl_cell_general(value = c(1234.5, 67.25, 890), format = money)
cells$margin  <- xl_formula(sprintf("=B%d*0.3", 2:4))
cells$link    <- xl_hyperlink(rep("https://ropensci.org", 3), "rOpenSci")
cells$note    <- xl_cell_general(
  value   = c("ok", NA, "check"),
  comment = list(xl_comment("Reviewed", author = "QA"), NULL, NULL),
  format  = list(NULL, xl_fill(background = "#EEEEEE"), NULL))
# a rich string is one cell's worth of text, so it goes in a list alongside
# the other rows' values
cells$rich    <- xl_cell_general(value = list(
  xl_rich_string("Plain ", xl_rich_run("bold", xl_font(bold = TRUE)),
                 " and ", xl_rich_run("red", xl_font(color = "red"))),
  "plain text", NA))

content <- xl_sheet(
  cells,
  merge = xl_merge("A5:C5", "A merged heading",
                   format = xl_align(horizontal = "center") +
                            xl_font(bold = TRUE)),
  validation = xl_validation(list(cols = "item"), type = "list",
                             list = c("Widget", "Gadget", "Gizmo"),
                             input_title = "Pick one"),
  ignore_errors = list(number_stored_as_text = "A2:A4"))

## -----------------------------------------------------------------------------
filtered <- xl_sheet(sales, filter = xl_filter("region", "==", "North"))

tabled <- xl_sheet(sales[, c("quarter", "revenue")],
                   table = xl_table(name = "Revenue", style = "medium 9",
                                    total_row = TRUE,
                                    columns = xl_table_column("revenue",
                                                              total = "sum")))

## -----------------------------------------------------------------------------
printed <- xl_sheet(
  sales,
  page = xl_page_setup(orientation = "landscape", paper = "A4",
                       fit_to = c(1, 0), center_horizontally = TRUE,
                       header = "&LwritexlShowcase&RPage &P of &N",
                       repeat_rows = 1),
  view = xl_sheet_view(hide_zero = TRUE),
  outline = xl_outline(symbols_below = FALSE),
  protect = list(password = "secret", sort = TRUE))

## -----------------------------------------------------------------------------
charted <- xl_sheet(
  sales,
  chart = list(
    xl_chart("column",
             list(xl_chart_series(values = list(cols = "revenue"),
                                  categories = list(cols = "quarter"),
                                  labels = xl_chart_labels(num_format = "$#,##0",
                                                           position = "outside_end")),
                  xl_chart_series(values = list(cols = "cost"),
                                  categories = list(cols = "quarter"))),
             title = "Revenue and cost",
             title_format = xl_font(size = 13, bold = TRUE),
             x_axis = xl_chart_axis(title = "Quarter"),
             y_axis = xl_chart_axis(title = "Dollars", min = 0,
                                    num_format = "$#,##0"),
             legend = xl_chart_legend(position = "bottom"),
             series_gap = 60, at = "G2"),
    xl_chart("pie",
             xl_chart_series(values = list(cols = "revenue"),
                             categories = list(cols = "quarter"),
                             labels = xl_chart_labels(show_percentage = TRUE),
                             points = list(xl_fill(background = "#C00000"),
                                           NULL, NULL, NULL)),
             title = "Share of revenue", at = "G20"),
    xl_chart("scatter_straight_markers",
             xl_chart_series(values = list(cols = "revenue"),
                             categories = list(cols = "cost"),
                             marker = xl_chart_marker("circle", size = 7),
                             trendline = xl_chart_trendline("linear",
                                                            equation = TRUE),
                             y_error_bars = xl_chart_error_bars("percentage", 5)),
             title = "Revenue against cost", at = "P2")))

overview <- xl_chartsheet(
  xl_chart("line",
           xl_chart_series(values = list(sheet = "Charts", cols = "revenue"),
                           categories = list(sheet = "Charts",
                                             cols = "quarter")),
           title = "Revenue, full page",
           y_axis = xl_chart_axis(title = "Revenue", min = 0)),
  tab_color = "red")

## -----------------------------------------------------------------------------
have_png <- isTRUE(capabilities("png")) &&
  requireNamespace("grDevices", quietly = TRUE)

## ----eval = have_png----------------------------------------------------------
card <- as.raster(matrix(c("#4472C4", "#ED7D31", "#A5A5A5", "#FFC000"),
                         nrow = 2))
pictured <- xl_sheet(sales[, 1:2], image = xl_image(card, at = "D2",
                                                    scale = 30,
                                                    description = "A test card"))

## ----eval = !have_png, echo = FALSE-------------------------------------------
# pictured <- xl_sheet(sales[, 1:2])

## -----------------------------------------------------------------------------
wb <- xl_workbook(
  list(Formatted = styled,
       Content   = content,
       Filtered  = filtered,
       Table     = tabled,
       Printed   = printed,
       Charts    = charted,
       Picture   = pictured,
       Overview  = overview),
  properties = xl_properties(
    title    = "writexl showcase",
    author   = "writexl",
    subject  = "Every feature in one workbook",
    custom   = list(Generated = Sys.Date(), Reviewed = FALSE),
    header_format = xl_font(bold = TRUE, color = "white") +
                    xl_fill(background = "#1F3864")))

showcase <- write_xlsx(wb, tempfile(fileext = ".xlsx"))

## ----eval = have_png----------------------------------------------------------
embedded <- write_xlsx(
  list(Embedded = xl_sheet(sales[, 1:2],
                           image = xl_image(card, at = "D2", embed = TRUE))),
  tempfile(fileext = ".xlsx"))

## -----------------------------------------------------------------------------
sheets <- c("Formatted", "Content", "Filtered", "Table", "Printed",
            "Charts", "Picture", "Overview")
data.frame(sheet = sheets,
           shows = c("formats, conditional rules, frozen pane",
                     "formulas, links, comments, rich text, merge, validation",
                     "an autofilter with its rows hidden",
                     "a worksheet table with a total row",
                     "page setup, outline symbols, protection",
                     "a column, a pie and a scatter chart",
                     "a picture anchored to a cell",
                     "a chartsheet: one chart, no cells"))

## -----------------------------------------------------------------------------
file.size(showcase)

# =============================================================================
# Images
# =============================================================================
#
# An image is either *inserted* -- floating over the sheet, anchored to a cell,
# which is what Excel has always done -- or *embedded* in a cell, which is the
# Excel 365 "place in cell" feature.  Older Excel shows #VALUE! where an
# embedded image sits, so the two are not interchangeable and `embed` is an
# explicit choice rather than a default.
#
# An image may arrive in four shapes: a file path, a raw vector of encoded
# bytes, or an in-memory picture as either a `raster` (or anything as.raster()
# accepts -- a colour matrix, a numeric matrix, an RGB/RGBA array) or a
# `nativeRaster`.  The last two are what rasterImage() draws, so anything that
# can be plotted can be written.
#
# xlsx stores encoded images, so an in-memory picture has to become PNG bytes
# first.  Base R has no PNG encoder outside its graphics devices, so the
# conversion renders the raster through grDevices::png() at exactly its own
# pixel dimensions.  That needs capabilities("png"), which is not guaranteed on
# a headless build -- when it is missing the error says how to do the
# conversion by hand rather than failing obscurely.
#
# Paths and raw vectors map to libxlsxwriter's file and _buffer variants
# respectively; a converted raster becomes a raw vector.
#
# Format is detected from the leading bytes rather than the file extension: a
# ".png" that is really a JPEG would otherwise reach libxlsxwriter and be
# misfiled in the zip.  Excel supports PNG, JPEG, BMP and GIF; libxlsxwriter
# declines SVG deliberately, since Excel converts it to PNG on the way in.
# -----------------------------------------------------------------------------

.LXW_OBJECT_POSITION <- c(
  "default" = 0L, move_and_size = 1L, move_dont_size = 2L,
  dont_move_dont_size = 3L, move_and_size_after = 4L
)

# The formats Excel accepts, by their leading bytes.
.IMAGE_MAGIC <- list(
  png  = as.raw(c(0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A)),
  jpeg = as.raw(c(0xFF, 0xD8, 0xFF)),
  gif  = as.raw(c(0x47, 0x49, 0x46, 0x38)),
  bmp  = as.raw(c(0x42, 0x4D))
)

# Identify an image from its first bytes, or NULL if it is none of them.
.image_format <- function(bytes) {
  for (nm in names(.IMAGE_MAGIC)) {
    m <- .IMAGE_MAGIC[[nm]]
    if (length(bytes) >= length(m) &&
        identical(bytes[seq_along(m)], m))
      return(nm)
  }
  NULL
}

# The first bytes of an image given as a path or a raw vector.
.image_head <- function(x, n = 8L) {
  if (is.raw(x)) return(utils::head(x, n))
  con <- file(x, "rb")
  on.exit(close(con))
  readBin(con, "raw", n)
}

# Is this an in-memory picture rather than a path or encoded bytes?
.is_raster_like <- function(x) {
  inherits(x, "nativeRaster") || inherits(x, "raster") ||
    is.matrix(x) || (is.array(x) && length(dim(x)) == 3L)
}

# Render an in-memory picture to PNG bytes, at exactly its own pixel size.
#
# grDevices and graphics are Suggests rather than Imports: writexl needs them
# only to encode an in-memory image, and only this one path uses them, so the
# package keeps its zero-dependency footing and the requirement is checked
# where it arises.
.raster_to_png <- function(x, arg) {
  missing_ns <- !vapply(c("grDevices", "graphics"), requireNamespace,
                        logical(1), quietly = TRUE)
  # nocov start  (needs a build with no PNG device; a test skips there instead)
  if (any(missing_ns) || !isTRUE(capabilities("png")))
    stop(sprintf(paste0("`%s` is an in-memory image, which writexl encodes ",
                        "with R's PNG device -- unavailable here (%s).\n",
                        "  Encode it yourself and pass the result instead:\n",
                        "    xl_image(png::writePNG(img))            # raw ",
                        "vector\n",
                        "    png::writePNG(img, \"plot.png\")           # or a ",
                        "file\n",
                        "    xl_image(\"plot.png\")"),
                 arg,
                 if (any(missing_ns))
                   paste("missing package(s):",
                         paste(names(missing_ns)[missing_ns], collapse = ", "))
                 else "capabilities(\"png\") is FALSE"),
         call. = FALSE)
  # nocov end
  d <- dim(x)
  if (length(d) < 2L || anyNA(d[1:2]) || any(d[1:2] < 1L))
    stop(sprintf("`%s` has no pixels to write", arg), call. = FALSE)
  height <- d[1L]
  width  <- d[2L]

  path <- tempfile(fileext = ".png")
  # png() warns that a small pixel size is "unlikely", meaning it suspects the
  # caller wanted inches.  Here the size is the image's own, so the warning is
  # noise -- muted by message, not blanket-suppressed, so anything else the
  # device has to say still comes through.
  withCallingHandlers(
    grDevices::png(path, width = width, height = height, units = "px",
                   bg = "transparent"),
    warning = function(w) {
      if (grepl("unlikely values in pixels", conditionMessage(w), fixed = TRUE))
        invokeRestart("muffleWarning")
    })
  # dev.off() must run even if drawing fails, or the device is left open
  on.exit({
    # only fires when drawing threw: the happy path has closed the device by
    # the time on.exit() runs
    if (grDevices::dev.cur() > 1L) grDevices::dev.off()  # nocov
    unlink(path)
  }, add = TRUE)
  graphics::par(mar = c(0, 0, 0, 0), xaxs = "i", yaxs = "i")
  graphics::plot.new()
  graphics::plot.window(c(0, 1), c(0, 1))
  graphics::rasterImage(x, 0, 0, 1, 1, interpolate = FALSE)
  grDevices::dev.off()
  readBin(path, "raw", file.size(path))
}

# Check that an image is one Excel can read, and say what it is otherwise.
.check_image <- function(x, arg = "image") {
  if (is.raw(x)) {
    if (!length(x))
      stop(sprintf("`%s` is an empty raw vector", arg), call. = FALSE)
  } else {
    if (!is.character(x) || length(x) != 1L || is.na(x) || !nzchar(x))
      stop(sprintf(paste0("`%s` must be a file path, a raw vector of encoded ",
                          "bytes, or an image rasterImage() can draw"), arg),
           call. = FALSE)
    if (!file.exists(x))
      stop(sprintf("`%s`: file does not exist: %s", arg, x), call. = FALSE)
    if (dir.exists(x))
      stop(sprintf("`%s` is a directory, not an image: %s", arg, x),
           call. = FALSE)
  }
  fmt <- .image_format(.image_head(x))
  if (is.null(fmt)) {
    what <- if (is.raw(x)) "the raw vector" else sprintf("\"%s\"", x)
    hint <- if (!is.raw(x) && grepl("\\.svgz?$", x, ignore.case = TRUE))
      paste0(" SVG is not supported: Excel converts it to PNG on the way in, ",
             "so save the plot as a PNG instead.")
    else ""
      stop(sprintf(paste0("`%s`: %s is not a PNG, JPEG, GIF or BMP -- those ",
                          "are the formats Excel reads.%s"), arg, what, hint),
           call. = FALSE)
  }
  fmt
}

#' Insert an image into a worksheet
#'
#' @description
#' `xl_image()` places an image on a sheet, either floating over the cells and
#' anchored to one of them (the default) or, with `embed = TRUE`, inside a cell.
#' Pass one or a list of them as `xl_sheet(image = )`.
#'
#' The image may be a file path or a raw vector, which is convenient when a plot
#' has just been written by a graphics device and never touched the disk.  PNG,
#' JPEG, GIF and BMP are supported --- the formats Excel reads --- and the
#' format is detected from the file's own bytes rather than its extension.
#'
#' @section Floating versus embedded:
#' An inserted image floats above the grid: it has a position but occupies no
#' cell, and `position` decides whether it moves and resizes as rows and columns
#' change.  An embedded image (`embed = TRUE`) lives *in* a cell and sizes with
#' it, which is the Excel 365 "place in cell" behaviour.  Excel versions without
#' that feature show `#VALUE!` in place of an embedded image, so it is worth
#' choosing deliberately.
#'
#' @section Caveats inherited from Excel:
#' An image's scaling can shift if it crosses a row whose height changed --- for
#' a taller font or wrapped text --- so set the height explicitly with
#' [xl_row_spec()] for rows an image spans.  BMP is supported only for backward
#' compatibility and must be 24-bit true colour; prefer PNG.  SVG is refused,
#' because Excel stores it converted to PNG anyway.
#'
#' @param image The image, in any of four shapes: a path to a PNG, JPEG, GIF or
#'   BMP; a raw vector holding one of those encoded; a `raster` (or anything
#'   [grDevices::as.raster()] accepts, such as a colour matrix or an RGB/RGBA
#'   array); or a `nativeRaster`.  The last two are what
#'   [graphics::rasterImage()] draws, so anything you can plot can be written.
#' @param at The cell the image is anchored to, such as `"B2"`, or a
#'   `list(rows = , cols = )` spec selecting a single cell.
#' @param scale Scale factor: one number for both axes, or `c(x, y)`.
#' @param offset Offset from the anchor cell's top-left corner in pixels, as
#'   `c(x, y)`.
#' @param position How the image behaves when rows and columns change size:
#'   `"move_and_size"` (the default), `"move_dont_size"`,
#'   `"dont_move_dont_size"`, `"move_and_size_after"`, or `"default"` for
#'   Excel's own default.  Ignored when `embed = TRUE`.
#' @param description Alt text, for screen readers.  Excel defaults it to the
#'   file name; `""` writes none.
#' @param decorative Mark the image as decorative, so screen readers skip it.
#'   Excel does not write a description for a decorative image.
#' @param url An optional hyperlink the image links to.
#' @param tip An optional mouseover tip for `url`.
#' @param embed Place the image inside the cell rather than floating above it.
#'   Requires a version of Excel that supports images in cells.
#' @param format An [xl_format] for the cell holding an embedded image.  Only
#'   meaningful with `embed = TRUE`.
#' @return An `xl_image` object.
#' @family images and charts
#' @seealso [xl_sheet]
#' @export
#' @examples
#' logo <- system.file("help", "figures", "logo.png", package = "writexl")
#' if (nzchar(logo)) {
#'   xl_image(logo, at = "C2", scale = 0.5)
#'   xl_image(logo, at = "C2", url = "https://example.com", tip = "Home")
#' }
xl_image <- function(image, at = "A1", scale = 1, offset = NULL,
                     position = "move_and_size", description = NULL,
                     decorative = FALSE, url = NULL, tip = NULL,
                     embed = FALSE, format = NULL) {
  if (missing(image) || is.null(image))
    stop("`image` must be a file path, a raw vector, or an image ",
         "rasterImage() can draw", call. = FALSE)
  # an in-memory picture is encoded now, so a later failure cannot be blamed on
  # a file that was fine when it was named
  if (.is_raster_like(image)) image <- .raster_to_png(image, "image")
  fmt <- .check_image(image, "image")

  str1 <- function(x, arg) {
    if (is.null(x)) return(NULL)
    if (!is.character(x) || length(x) != 1L || is.na(x))
      stop(sprintf("`%s` must be a single non-NA string", arg), call. = FALSE)
    x
  }
  pair <- function(x, arg, what) {
    if (is.null(x)) return(NULL)
    if (!is.numeric(x) || !length(x) %in% c(1L, 2L) || anyNA(x))
      stop(sprintf("`%s` must be one number or two, %s", arg, what),
           call. = FALSE)
    if (length(x) == 1L) x <- c(x, x)
    as.numeric(x)
  }

  sc <- pair(scale, "scale", "as c(x, y)")
  if (!is.null(sc) && any(sc <= 0))
    stop("`scale` must be positive", call. = FALSE)
  off <- pair(offset, "offset", "of pixels as c(x, y)")

  pos <- .val_enum(position, names(.LXW_OBJECT_POSITION), "position")
  dec <- .val_flag(decorative, "decorative")
  emb <- .val_flag(embed, "embed")
  if (!is.null(format) && !is_xl_format(format))
    stop("`format` must be an xl_format object", call. = FALSE)
  # cell_format reaches libxlsxwriter only through worksheet_embed_image_opt(),
  # so accepting it for a floating image would be accepting a no-op
  if (!is.null(format) && !isTRUE(emb))
    stop("`format` applies to the cell an embedded image sits in, so it needs ",
         "`embed = TRUE`", call. = FALSE)
  if (isTRUE(emb) && !is.null(url))
    stop("`url` is not available for an embedded image; Excel puts the image ",
         "in the cell, leaving nowhere for the link", call. = FALSE)

  structure(.drop_null(list(
    image = image, image_format = fmt, at = at, scale = sc, offset = off,
    position = pos, description = str1(description, "description"),
    decorative = dec, url = str1(url, "url"), tip = str1(tip, "tip"),
    embed = emb, format = format
  )), class = "xl_image")
}

#' @export
print.xl_image <- function(x, ...) {
  p <- unclass(x)
  src <- if (is.raw(p[["image"]]))
    sprintf("<%s, %d bytes>", p[["image_format"]], length(p[["image"]]))
  else basename(p[["image"]])
  cat(sprintf("<xl_image: %s at %s%s>\n", src,
              if (is.character(p[["at"]])) p[["at"]] else "a cell",
              if (isTRUE(p[["embed"]])) ", embedded" else ""))
  invisible(x)
}

# Resolve a sheet's images to the C payloads.  Applied after the row loop
# beside the merges and tables: an embedded image writes a cell, and a floating
# one anchors to a row whose height the row loop may still be setting.
.resolve_images <- function(el, df, reg, header_offset, props) {
  if (!inherits(el, "xl_sheet")) return(list())
  ims <- .image_list(el$image)
  lapply(seq_along(ims), function(i) {
    p <- unclass(ims[[i]])
    at <- .xl_resolve_range(p[["at"]], arg = sprintf("image[[%d]] at", i),
                            df = df, header_offset = header_offset,
                            allow_cell = TRUE)
    if (at[1L] != at[3L] || at[2L] != at[4L])
      stop(sprintf(paste0("`image[[%d]] at` must name a single cell, not a ",
                          "range: an image is anchored to one cell"), i),
           call. = FALSE)
    ent <- list(row = at[1L], col = at[2L],
                embed = as.integer(isTRUE(p[["embed"]])))
    if (is.raw(p[["image"]])) ent$buffer <- p[["image"]]
    else                      ent$filename <- p[["image"]]
    if (!is.null(p[["scale"]])) {
      ent$x_scale <- p[["scale"]][1L]
      ent$y_scale <- p[["scale"]][2L]
    }
    if (!is.null(p[["offset"]])) {
      ent$x_offset <- as.integer(p[["offset"]][1L])
      ent$y_offset <- as.integer(p[["offset"]][2L])
    }
    if (!is.null(p[["position"]]))
      ent$object_position <- unname(.LXW_OBJECT_POSITION[[p[["position"]]]])
    if (!is.null(p[["description"]])) ent$description <- p[["description"]]
    if (isTRUE(p[["decorative"]]))    ent$decorative <- 1L
    if (!is.null(p[["url"]]))         ent$url <- p[["url"]]
    if (!is.null(p[["tip"]]))         ent$tip <- p[["tip"]]
    if (!is.null(p[["format"]]))
      ent$cell_format_id <- .register_format(reg,
        merge_xl_format(props$default_format, p[["format"]]))
    ent
  })
}

# Resolve xl_sheet(background_image =) to the C payload.  It takes the same
# shapes as an image but no options: Excel tiles it to fill the sheet, and it
# is a display-only backdrop that never prints.
.resolve_background <- function(el, arg = "background_image") {
  if (!inherits(el, "xl_sheet") || is.null(el$background_image)) return(NULL)
  x <- el$background_image
  if (.is_raster_like(x)) x <- .raster_to_png(x, arg)
  .check_image(x, arg)
  if (is.raw(x)) list(buffer = x) else list(filename = x)
}

# --- Guard against libxlsxwriter's drawing-id desync ---------------------------
#
# _prepare_drawings() in workbook.c increments drawing_id for any sheet holding
# images, embedded images, charts, header/footer VML or a background image, and
# uses it to build the sheet's "../drawings/drawingN.xml" relationship target.
# The packager numbers the drawing *files* separately, counting only sheets that
# actually have a drawing (_get_drawing_count() in packager.c).
#
# A sheet whose only image is embedded, or lives in a header or footer, is
# therefore counted without producing a drawing file, and every floating image
# on a LATER sheet gets a target one too high.  Excel finds the missing part and
# repairs the file, dropping the image.
#
# writexl cannot renumber the relationship without patching vendored code, so
# the combination is refused.  Ordering the sheets so the floating images come
# first is a real workaround, which is why the message says so.
.check_drawing_order <- function(sheets, sheet_names) {
  has_floating <- function(s)
    any(vapply(s$images, function(i) !isTRUE(as.logical(i$embed)), logical(1)))
  has_embedded <- function(s)
    any(vapply(s$images, function(i) isTRUE(as.logical(i$embed)), logical(1)))
  has_hf_image <- function(s)
    any(grepl("_image_(left|center|right)$", names(s$page)))
  has_background <- function(s) !is.null(s$background)
  # A chart produces a drawing of its own, so it never *causes* the desync --
  # but it is subject to it exactly as a floating image is, since its
  # relationship target is numbered from the same counter.  Established with a
  # pure-C reprex against libxlsxwriter 1.2.4 before charts existed here: a
  # header-image sheet followed by a chart sheet points at a missing drawing
  # part, while the reverse order is fine.  No test could have caught this
  # omission, because nothing else notices a chart is a victim.
  has_drawing <- function(s) has_floating(s) || length(s$charts) > 0L

  # 1. An embedded image alongside any other kind.
  #
  # _prepare_drawings() gives an embedded image the shared image_ref_id and
  # passes it to worksheet_set_error_cell(), which writes it as the cell's `vm`
  # attribute.  But `vm` indexes the <valueMetadata> blocks, of which there is
  # one per *embedded* image.  With only embedded images in the workbook the two
  # counters move together and the file is right; any other image shifts the
  # ref id and `vm` then points past the end of the metadata, which Excel
  # repairs away.  The richValueRel target goes to the wrong image for the same
  # reason.
  emb <- vapply(sheets, has_embedded, logical(1))
  other <- vapply(sheets, function(s) has_floating(s) || has_hf_image(s),
                  logical(1))
  if (any(emb) && any(other))
    stop(sprintf(paste0("sheet \"%s\" has an embedded image and sheet \"%s\" ",
                        "has another image.\n  libxlsxwriter numbers the ",
                        "embedded image's metadata from a counter shared with ",
                        "every other image, so mixing them writes a cell ",
                        "reference that points past the end of the metadata ",
                        "and Excel repairs the sheet.\n  Use `embed = TRUE` ",
                        "only in a workbook whose images are all embedded, or ",
                        "make this one a floating image."),
                 sheet_names[which(emb)[1L]], sheet_names[which(other)[1L]]),
         call. = FALSE)

  # 2. A sheet counted for a drawing id that produces no drawing file, before a
  # sheet that does produce one -- a floating image or a chart.
  offender <- NULL
  for (i in seq_along(sheets)) {
    s <- sheets[[i]]
    if (!is.null(offender) && has_drawing(s))
      stop(sprintf(paste0("sheet \"%s\" has %s, but sheet \"%s\" earlier in ",
                          "the workbook has %s.\n  libxlsxwriter numbers the ",
                          "drawing relationship wrongly in that order, and ",
                          "Excel repairs the file and drops it.\n  Put the ",
                          "sheet(s) with images or charts before \"%s\"."),
                   sheet_names[i],
                   if (has_floating(s)) "a floating image" else "a chart",
                   offender$name, offender$what,
                   offender$name), call. = FALSE)
    if (is.null(offender) && !has_drawing(s)) {
      what <- if (has_hf_image(s)) "a header or footer image"
              else if (has_background(s)) "a background image"
              else NULL
      if (!is.null(what)) offender <- list(name = sheet_names[i], what = what)
    }
  }
  invisible(NULL)
}

# Normalise one image or a list of them to a checked list.
.image_list <- function(image, arg = "image") {
  if (is.null(image)) return(list())
  is_one <- inherits(image, "xl_image")
  ims <- if (is_one) list(image) else image
  if (!is.list(ims))
    stop(sprintf("`%s` must be an xl_image object or a list of them", arg),
         call. = FALSE)
  for (i in seq_along(ims))
    if (!inherits(ims[[i]], "xl_image"))
      stop(sprintf("`%s[[%d]]` must be an xl_image object", arg, i),
           call. = FALSE)
  ims
}

# Images: the constructor, its validation, and what reaches the file.
# PNG_1x1 and png_file() are in helper-xlsx.R -- other files need them too.

# ── Format detection ──────────────────────────────────────────────────────────

test_that("every format Excel reads is recognised by its leading bytes", {
  # enumerated rather than sampled: a format added to .IMAGE_MAGIC without a
  # case here fails the count check below
  expect_equal(.image_format(PNG_1x1), "png")
  expect_equal(.image_format(as.raw(c(0xFF, 0xD8, 0xFF, 0xE0))), "jpeg")
  expect_equal(.image_format(charToRaw("GIF89a")), "gif")
  expect_equal(.image_format(charToRaw("GIF87a")), "gif")
  expect_equal(.image_format(charToRaw("BM  ")), "bmp")
  expect_equal(length(.IMAGE_MAGIC), 4L)
})

test_that("anything else is not an image", {
  expect_null(.image_format(charToRaw("nope")))
  expect_null(.image_format(charToRaw("<?xml version=\"1.0\"?>")))
  expect_null(.image_format(raw(0)))
  # a truncated header must not match on a prefix
  expect_null(.image_format(as.raw(c(0x89, 0x50))))
})

test_that("the format comes from the bytes, not the extension", {
  # a .png that is really a JPEG would otherwise be misfiled in the zip
  p <- tempfile(fileext = ".png")
  writeBin(as.raw(c(0xFF, 0xD8, 0xFF, 0xE0, 0x00, 0x10)), p)
  expect_equal(.check_image(p), "jpeg")
  # and a real image with no extension at all is still accepted
  q <- tempfile()
  writeBin(PNG_1x1, q)
  expect_equal(.check_image(q), "png")
})

test_that("a non-image file is refused, naming it", {
  p <- tempfile(fileext = ".png")
  writeLines("not an image", p)
  expect_error(xl_image(p), "is not a PNG, JPEG, GIF or BMP")
})

test_that("SVG is refused with a reason", {
  # libxlsxwriter declines SVG because Excel stores it converted to PNG, so the
  # error should say that rather than just listing the formats
  s <- tempfile(fileext = ".svg")
  writeLines("<svg xmlns=\"http://www.w3.org/2000/svg\"/>", s)
  expect_error(xl_image(s), "SVG is not supported")
  expect_error(xl_image(s), "save the plot as a PNG")
})

# ── The image itself ──────────────────────────────────────────────────────────

test_that("an image may be a path or a raw vector", {
  expect_s3_class(xl_image(png_file()), "xl_image")
  expect_s3_class(xl_image(PNG_1x1), "xl_image")
  expect_equal(unclass(xl_image(PNG_1x1))$image_format, "png")
  # the argument is `image`, not `file`: it takes more than a filename
  expect_equal(names(formals(xl_image))[1L], "image")
})

test_that("a missing or unreadable source is refused", {
  expect_error(xl_image(), "must be a file path, a raw vector")
  expect_error(xl_image("no_such_file.png"), "file does not exist")
  expect_error(xl_image(tempdir()), "is a directory, not an image")
  expect_error(xl_image(raw(0)), "empty raw vector")
  expect_error(xl_image(42), "must be a file path, a raw vector")
})

# ── Options ───────────────────────────────────────────────────────────────────

test_that("scale and offset take one number or two", {
  expect_equal(unclass(xl_image(png_file(), scale = 0.5))$scale, c(0.5, 0.5))
  expect_equal(unclass(xl_image(png_file(), scale = c(2, 3)))$scale, c(2, 3))
  expect_equal(unclass(xl_image(png_file(), offset = c(4, 5)))$offset, c(4, 5))
  expect_null(unclass(xl_image(png_file()))$offset)

  expect_error(xl_image(png_file(), scale = -1), "`scale` must be positive")
  expect_error(xl_image(png_file(), scale = 0), "`scale` must be positive")
  expect_error(xl_image(png_file(), scale = c(1, 2, 3)), "one number or two")
  expect_error(xl_image(png_file(), scale = "big"), "one number or two")
  expect_error(xl_image(png_file(), offset = c(1, NA)), "one number or two")
})

test_that("every object position is accepted", {
  for (p in names(.LXW_OBJECT_POSITION))
    expect_equal(unclass(xl_image(png_file(), position = p))$position, p)
  expect_equal(length(.LXW_OBJECT_POSITION), 5L)
  expect_error(xl_image(png_file(), position = "wander"), "`position` must be one of")
})

test_that("a format without embed is refused rather than ignored", {
  # cell_format reaches libxlsxwriter only through worksheet_embed_image_opt(),
  # so accepting it for a floating image would be accepting a silent no-op
  expect_error(xl_image(png_file(), format = xl_font(bold = TRUE)),
               "needs `embed = TRUE`")
  expect_s3_class(xl_image(png_file(), embed = TRUE,
                           format = xl_font(bold = TRUE)), "xl_image")
  expect_error(xl_image(png_file(), embed = TRUE, format = "bold"),
               "must be an xl_format")
})

test_that("a url on an embedded image is refused", {
  # the image occupies the cell, so there is nowhere for the link to live
  expect_error(xl_image(png_file(), embed = TRUE, url = "https://example.com"),
               "not available for an embedded image")
  expect_s3_class(xl_image(png_file(), url = "https://example.com"), "xl_image")
})

test_that("text and flag arguments are validated", {
  expect_error(xl_image(png_file(), description = 1), "single non-NA string")
  expect_error(xl_image(png_file(), url = c("a", "b")), "single non-NA string")
  expect_error(xl_image(png_file(), tip = NA), "single non-NA string")
  expect_error(xl_image(png_file(), decorative = "yes"), "`decorative` must be")
  expect_error(xl_image(png_file(), embed = "yes"), "`embed` must be")
  # NA means "unset" throughout writexl, so it is accepted and drops out
  expect_null(unclass(xl_image(png_file(), decorative = NA))$decorative)
  expect_null(unclass(xl_image(png_file(), embed = NA))$embed)
})

test_that("images normalise from one or a list", {
  expect_length(.image_list(xl_image(png_file())), 1L)
  expect_length(.image_list(list(xl_image(png_file()), xl_image(PNG_1x1))), 2L)
  expect_length(.image_list(NULL), 0L)
  expect_error(.image_list("logo.png"), "must be an xl_image object")
  expect_error(.image_list(list(xl_image(PNG_1x1), "x")),
               "`image\\[\\[2\\]\\]` must be an xl_image object")
})

test_that("the print method runs", {
  expect_output(print(xl_image(png_file())), "xl_image")
  expect_output(print(xl_image(PNG_1x1)), "67 bytes")
  expect_output(print(xl_image(png_file(), embed = TRUE)), "embedded")
})

# ── In-memory pictures ────────────────────────────────────────────────────────

# A PNG's real pixel size, read from its IHDR chunk -- no decoder needed.
png_dims <- function(bytes) {
  be <- function(i) sum(as.integer(bytes[i]) * c(256^3, 256^2, 256, 1))
  c(width = be(17:20), height = be(21:24))
}

test_that("a raster is encoded at its own pixel size", {
  skip_if_not(isTRUE(capabilities("png")))
  # dim(raster) is rows x cols, and a PNG is width x height -- so a 2x3 raster
  # must come out 3 wide and 2 tall, not the other way round
  r <- as.raster(matrix(c("red", "green", "blue", "black", "white", "grey"),
                        nrow = 2, byrow = TRUE))
  expect_equal(dim(r), c(2L, 3L))
  b <- unclass(xl_image(r))$image
  expect_true(is.raw(b))
  expect_equal(.image_format(b), "png")
  expect_equal(unname(png_dims(b)), c(3, 2))
})

test_that("everything as.raster() accepts is encoded", {
  skip_if_not(isTRUE(capabilities("png")))
  # colour matrix
  m <- matrix(c("#FF0000", "#00FF00", "#0000FF", "#FFFFFF"), nrow = 2)
  expect_equal(unname(png_dims(unclass(xl_image(m))$image)), c(2, 2))
  # numeric matrix (greyscale)
  g <- matrix(c(0, 0.5, 1, 0.25, 0.75, 0.1), nrow = 2)
  expect_equal(unname(png_dims(unclass(xl_image(g))$image)), c(3, 2))
  # RGB array
  a <- array(0.5, dim = c(4, 5, 3))
  expect_equal(unname(png_dims(unclass(xl_image(a))$image)), c(5, 4))
})

test_that("a nativeRaster is encoded", {
  skip_if_not(isTRUE(capabilities("png")))
  nr <- structure(as.integer(c(-1, -16777216, -65536, -16711936)),
                  dim = c(2L, 2L), class = "nativeRaster")
  expect_equal(unname(png_dims(unclass(xl_image(nr))$image)), c(2, 2))
})

test_that("an in-memory image is encoded once, at construction", {
  skip_if_not(isTRUE(capabilities("png")))
  # the object carries bytes, not the raster, so a later change to the raster
  # cannot alter what gets written
  r <- as.raster(matrix("red", 1, 1))
  im <- xl_image(r)
  expect_true(is.raw(unclass(im)$image))
  expect_equal(unclass(im)$image_format, "png")
})

test_that("a degenerate raster is refused", {
  skip_if_not(isTRUE(capabilities("png")))
  expect_error(xl_image(matrix(character(0), nrow = 0, ncol = 0)),
               "no pixels to write")
})

test_that("encoding a raster leaves no device open", {
  skip_if_not(isTRUE(capabilities("png")))
  before <- grDevices::dev.cur()
  xl_image(as.raster(matrix("red", 2, 2)))
  expect_equal(grDevices::dev.cur(), before)
})

# ── What reaches the file ─────────────────────────────────────────────────────

img_parts <- function(..., df = data.frame(a = 1:4)) {
  p <- write_tmp(list(S = xl_sheet(df, ...)))
  d <- tempfile(); dir.create(d); utils::unzip(p, exdir = d)
  part <- function(f) {
    q <- file.path(d, f)
    if (file.exists(q)) paste(readLines(q, warn = FALSE), collapse = "") else ""
  }
  list(files = list.files(d, recursive = TRUE),
       drawing = part("xl/drawings/drawing1.xml"),
       sheet = part("xl/worksheets/sheet1.xml"),
       rels = part("xl/drawings/_rels/drawing1.xml.rels"))
}

test_that("a floating image reaches the drawing and the media", {
  r <- img_parts(image = xl_image(png_file(), at = "C2"))
  expect_true("xl/media/image1.png" %in% r$files)
  expect_true("xl/drawings/drawing1.xml" %in% r$files)
  expect_match(r$sheet, "<drawing r:id=", fixed = TRUE)
  # anchored to C2, which is 0-based col 2, row 1
  expect_match(r$drawing, "<xdr:col>2</xdr:col>", fixed = TRUE)
  expect_match(r$drawing, "<xdr:row>1</xdr:row>", fixed = TRUE)
})

test_that("a path and a raw vector differ only in the default alt text", {
  a <- img_parts(image = xl_image(png_file(), at = "A1"))
  b <- img_parts(image = xl_image(PNG_1x1, at = "A1"))
  expect_equal(a$files, b$files)
  # Excel defaults an image's alt text to its file name, so the path form
  # carries a descr the raw form cannot have -- the rest must be identical
  expect_match(a$drawing, "descr=", fixed = TRUE)
  expect_no_match(b$drawing, "descr=", fixed = TRUE)
  strip <- function(x) sub(' descr="[^"]*"', "", x)
  expect_equal(strip(a$drawing), strip(b$drawing))
})

test_that("offsets and the object position are written", {
  r <- img_parts(image = xl_image(png_file(), at = "A1", offset = c(40, 20)))
  # Pixels become EMU at 9525 each.  The absolute offset is the stable thing to
  # assert: an offset larger than the row's height is normalised into the next
  # row, so a 20px y-offset on a 20px row comes out as row 1 with rowOff 0 --
  # the position is right, the spelling is not what a naive test expects.
  expect_match(r$drawing, sprintf('<a:off x="%d" y="%d"/>', 40 * 9525,
                                  20 * 9525), fixed = TRUE)
  expect_match(r$drawing, sprintf("<xdr:colOff>%d</xdr:colOff>", 40 * 9525),
               fixed = TRUE)
  # the position controls the anchor kind
  expect_match(img_parts(image = xl_image(png_file(), at = "A1",
                                          position = "dont_move_dont_size"))$drawing,
               "absoluteAnchor|editAs=\"absolute\"")
  expect_match(img_parts(image = xl_image(png_file(), at = "A1",
                                          position = "move_dont_size"))$drawing,
               "oneCellAnchor|editAs=\"oneCell\"")
})

test_that("a description is written, and decorative replaces it", {
  expect_match(img_parts(image = xl_image(png_file(), at = "A1",
                                          description = "A logo"))$drawing,
               "A logo", fixed = TRUE)
  # Excel does not write a description for a decorative image
  d <- img_parts(image = xl_image(png_file(), at = "A1",
                                  description = "A logo",
                                  decorative = TRUE))$drawing
  expect_no_match(d, "descr=\"A logo\"", fixed = TRUE)
})

test_that("a url becomes an external relationship", {
  r <- img_parts(image = xl_image(png_file(), at = "A1",
                                  url = "https://example.com", tip = "Home"))
  expect_match(r$rels, "https://example.com", fixed = TRUE)
  expect_match(r$rels, "TargetMode=\"External\"", fixed = TRUE)
})

test_that("several images share one drawing", {
  r <- img_parts(image = list(xl_image(png_file(), at = "A1"),
                              xl_image(png_file(), at = "A5"),
                              xl_image(PNG_1x1, at = "A9")))
  # the same picture three times is deduplicated to one media file
  expect_equal(sum(grepl("^xl/media/", r$files)), 1L)
  expect_equal(sum(grepl("^xl/drawings/drawing", r$files)), 1L)
  expect_equal(length(gregexpr("<xdr:pic>", r$drawing)[[1L]]), 3L)
})

test_that("an anchor must name a single cell", {
  expect_error(write_tmp(list(S = xl_sheet(data.frame(a = 1:4),
                                           image = xl_image(PNG_1x1, at = "A1:B2")))),
               "must name a single cell")
})

test_that("an embedded image writes a cell, not a drawing", {
  r <- img_parts(image = xl_image(PNG_1x1, at = "C2", embed = TRUE))
  expect_true("xl/media/image1.png" %in% r$files)
  expect_false("xl/drawings/drawing1.xml" %in% r$files)
  expect_true("xl/richData/richValueRel.xml" %in% r$files)
  # the cell carries the value-metadata index, and it must be in range
  expect_match(r$sheet, "vm=\"1\"", fixed = TRUE)
})

# ── Background ────────────────────────────────────────────────────────────────

test_that("a background image is stored and referenced", {
  r <- img_parts(background_image = png_file())
  expect_true("xl/media/image1.png" %in% r$files)
  expect_match(r$sheet, "<picture r:id=", fixed = TRUE)
  # a background is a backdrop, not a drawing
  expect_false("xl/drawings/drawing1.xml" %in% r$files)
})

test_that("a background may be a raw vector or an in-memory image", {
  expect_true("xl/media/image1.png" %in% img_parts(background_image = PNG_1x1)$files)
  skip_if_not(isTRUE(capabilities("png")))
  expect_true("xl/media/image1.png" %in%
                img_parts(background_image = as.raster(matrix("red", 2, 2)))$files)
})

test_that("a bad background is refused like any other image", {
  bad <- tempfile(fileext = ".png")
  writeLines("not an image", bad)
  expect_error(write_tmp(list(S = xl_sheet(data.frame(a = 1), background_image = bad))),
               "is not a PNG, JPEG, GIF or BMP")
})

# ── Guards against libxlsxwriter's counter bugs ───────────────────────────────

test_that("an embedded image may not share a workbook with any other image", {
  # _prepare_drawings() gives the embedded image the shared image ref id and
  # writes it as the cell's vm, which indexes the valueMetadata blocks instead
  df <- data.frame(a = 1:4)
  expect_error(write_tmp(list(
    A = xl_sheet(df, image = xl_image(PNG_1x1, at = "C2")),
    B = xl_sheet(df, image = xl_image(PNG_1x1, at = "C2", embed = TRUE)))),
    "points past the end of the metadata")
  # embedded images on their own are fine, in any number
  expect_silent(write_tmp(list(
    A = xl_sheet(df, image = xl_image(PNG_1x1, at = "C2", embed = TRUE)),
    B = xl_sheet(df, image = xl_image(PNG_1x1, at = "C4", embed = TRUE)))))
})

test_that("a drawing-less image sheet may not precede a floating image", {
  # header/footer and background images each consume a drawing id without
  # writing a drawing part, so a later floating image points at a missing file
  df <- data.frame(a = 1:4)
  logo <- png_file()
  hdr <- xl_sheet(df, page = xl_page_setup(header = "&L&G",
                                           header_image = list(left = logo)))
  expect_error(write_tmp(list(A = hdr,
                              B = xl_sheet(df, image = xl_image(logo, "C2")))),
               "earlier in the workbook has a header or footer image")
  expect_error(write_tmp(list(A = xl_sheet(df, background_image = logo),
                              B = xl_sheet(df, image = xl_image(logo, "C2")))),
               "earlier in the workbook has a background image")
  # the other order is fine, and is what the message recommends
  expect_silent(write_tmp(list(A = xl_sheet(df, image = xl_image(logo, "C2")),
                               B = hdr)))
  expect_silent(write_tmp(list(A = xl_sheet(df, image = xl_image(logo, "C2")),
                               B = xl_sheet(df, background_image = logo))))
})

test_that("only an embedded image costs row streaming", {
  df <- data.frame(a = 1:4)
  plan_float <- list(list(images = list(list(embed = 0L))))
  plan_embed <- list(list(images = list(list(embed = 1L))))
  # request = TRUE isolates the question: does the feature itself force it
  # off?  Without it a frame this small would be off for size alone.
  expect_equal(.resolve_constant_memory(list(D = df), list(), plan_float,
                                        request = TRUE)$on, 1L)
  # the request cannot be honoured, and says so rather than failing quietly
  expect_warning(
    cm <- .resolve_constant_memory(list(D = df), list(), plan_embed,
                                   request = TRUE),
    "cannot be honoured")
  expect_equal(cm$on, 0L)
  expect_match(cm$reasons, "embedded image")
})

test_that("an embedded image takes a cell format", {
  r <- img_parts(image = xl_image(png_file(), "B2", embed = TRUE,
                                  format = xl_align(horizontal = "center")))
  expect_match(r$sheet, "vm=", fixed = TRUE)
  expect_true("xl/richData/rdrichvalue.xml" %in% r$files ||
                any(grepl("richValue", r$files)))
})

test_that("an in-memory image needs R's PNG device", {
  skip_if(isTRUE(capabilities("png")), "this build has a PNG device")
  expect_error(xl_image(as.raster(matrix("red", 2, 2)), "A1"),
               "unavailable here")
})

# Cross-function API consistency.
#
# An argument that appears in several exported functions should mean the same
# thing and appear in the same relative order, so that reading one signature
# teaches you the others.  This is a mechanical gate rather than a convention in
# prose, because a convention in prose is exactly what got violated: `format`
# was briefly moved after `author` in xl_comment(), breaking the idiom that
# `format` follows the content arguments everywhere else.

# Argument names and positions for every exported function.
api_arg_positions <- function() {
  ns <- asNamespace("writexl")
  out <- list()
  for (f in sort(getNamespaceExports("writexl"))) {
    obj <- get(f, envir = ns)
    if (!is.function(obj)) next
    a <- names(formals(obj))
    a <- a[a != "..."]
    if (length(a)) out[[f]] <- a
  }
  out
}

# Same name, unrelated concept: two functions and the one argument they are
# allowed to disagree about.  Naming the argument rather than exempting the
# whole pair keeps every other argument the two share under the check.
FALSE_FRIENDS <- list(
  # `name` in xl_hyperlink() is the deprecated alias for `value`, kept last on
  # purpose; elsewhere it is a label
  list(c("xl_hyperlink", "xl_chart_series"), "name"),
  list(c("xl_hyperlink", "xl_chart_axis"), "name"),
  list(c("xl_hyperlink", "xl_chart_trendline"), "name"),
  # `position` in xl_image() is how the picture behaves when rows and columns
  # resize; in xl_chart_axis() it is whether the categories sit on the tick
  # marks or between them
  list(c("xl_image", "xl_chart_axis"), "position"),
  # `chart` on a worksheet is one or more charts floating over the cells, and
  # optional; on a chartsheet it is the sheet's whole content, exactly one of
  # them, and the only required argument -- so it comes first there and last
  # among the overlays here
  list(c("xl_sheet", "xl_chartsheet"), "chart")
)

# The argument two functions may disagree about, or NULL.
false_friend_arg <- function(a, b) {
  for (p in FALSE_FRIENDS)
    if (setequal(p[[1L]], c(a, b))) return(p[[2L]])
  NULL
}

test_that("shared arguments keep the same relative order across functions", {
  api <- api_arg_positions()
  fns <- names(api)
  conflicts <- character(0)
  for (i in seq_along(fns)) for (j in seq_along(fns)) {
    if (j <= i) next
    fa <- fns[i]; fb <- fns[j]
    common <- setdiff(intersect(api[[fa]], api[[fb]]),
                      false_friend_arg(fa, fb))
    if (length(common) < 2L) next
    oa <- order(match(common, api[[fa]]))
    ob <- order(match(common, api[[fb]]))
    if (!identical(oa, ob))
      conflicts <- c(conflicts, sprintf(
        "%s(%s) vs %s(%s)", fa, paste(common[oa], collapse = ", "),
        fb, paste(common[ob], collapse = ", ")))
  }
  expect_equal(conflicts, character(0))
})

test_that("`format` follows the content arguments, everywhere it appears", {
  # The idiom across the package: identify the content, then style it, then
  # everything else.  The number is how many content arguments precede it.
  expected <- c(xl_num_format = 1L,   # `format` *is* the content here
                xl_formula = 2L, xl_rich_run = 2L, xl_comment = 2L,
                xl_merge = 3L, xl_hyperlink = 3L, xl_hyperlink_cell = 3L,
                xl_cell_general = 4L)
  api <- api_arg_positions()
  got <- vapply(names(expected), function(f) {
    p <- match("format", api[[f]])
    if (is.na(p)) NA_integer_ else as.integer(p)
  }, integer(1))
  expect_equal(got, expected)
})

test_that("what a cell shows is always called `value`", {
  # xl_comment(), xl_merge() and xl_rich_run() each called it `text` at one
  # point, which made three functions read differently for one idea.  `value`
  # is the word everywhere something is shown, whatever its type.
  api <- api_arg_positions()
  expect_equal(names(Filter(function(a) "text" %in% a, api)), character(0))
})

test_that("sheet selection has exactly one spelling", {
  # Which tab is selected is an xl_sheet_view() setting, and only that: a
  # worksheet and a chartsheet take the same argument and write the same
  # element.  A second route -- xl_sheet(selected =), say -- would have to be
  # kept in step with the visibility rules in sheet_visibility.R, which read
  # the view and nothing else.
  api <- api_arg_positions()
  expect_equal(names(Filter(function(a) "selected" %in% a, api)),
               "xl_sheet_view")
  expect_equal(names(Filter(function(a) "active" %in% a, api)),
               "xl_sheet_view")
})

test_that("data comes first in the functions that take data", {
  api <- api_arg_positions()
  expect_equal(api$write_xlsx[1L], "x")
  expect_equal(api$xl_sheet[1L], "data")
  expect_equal(api$xl_workbook[1L], "sheets")
  # and the range/target identifies what is being acted on
  expect_equal(api$xl_merge[1L], "range")
  expect_equal(api$xl_col_spec[1L], "cols")
  expect_equal(api$xl_row_spec[1L], "rows")
})

test_that("the object-taking sheet arguments are named for their constructor", {
  # xl_sheet(page = xl_page_setup(), view = xl_sheet_view(), merge = xl_merge())
  # -- the argument name should make the constructor obvious
  api <- api_arg_positions()
  for (a in c("page", "view", "merge"))
    expect_true(a %in% api$xl_sheet, label = a)
})

# =============================================================================
# Merged cells
# =============================================================================
#
# worksheet_merge_range() does more than record a merge: it writes the top-left
# cell's text itself and blanks the rest of the range with the same format.  A
# merge therefore carries its own content, which is why xl_merge() takes `value`
# rather than reusing whatever the data frame put there.
#
# That also settles when merges are applied.  They run *after* the sheet's rows
# have been written, so a merge over cells the data frame filled discards all
# but the top-left value -- which is exactly what merging does in Excel.
# Applying them before the rows would instead have the row loop overwrite the
# merge's own blanks.
#
# Writing to already-written rows is not possible while libxlsxwriter is
# streaming rows to disk, so any merge turns constant memory off; see
# .resolve_constant_memory().
# -----------------------------------------------------------------------------

#' Merge a range of cells
#'
#' @description
#' `xl_merge()` merges a rectangle of cells into one, as Excel's "Merge and
#' Centre" does.  Pass one or a list of them as `xl_sheet(merge = )`.
#'
#' A merged range holds a single value, so `xl_merge()` carries its own `value`
#' rather than taking it from the data frame.  Merging over cells the data frame
#' filled keeps only the merged value, exactly as merging in Excel discards
#' everything but the top-left value.
#'
#' @param range The cells to merge: an Excel range string such as `"A1:C1"`, or
#'   a `list(rows = , cols = )` spec.  It must cover more than one cell ---
#'   Excel has no single-cell merge.
#' @param value The value shown in the merged cell: a string, or anything with
#'   an [as.character()] method such as an [xl_rich_string()].  `NULL` leaves
#'   the cell empty.
#' @param format An optional [xl_format] applied to the whole merged range.
#'   Merged cells usually want `xl_align(horizontal = "center")`.
#' @return An `xl_merge` object.
#' @family worksheet features
#' @seealso [xl_sheet], [xl_format]
#' @export
#' @examples
#' xl_merge("A1:C1", "Quarterly results",
#'          format = xl_align(horizontal = "center") + xl_font(bold = TRUE))
#'
#' df <- data.frame(a = 1:3, b = 4:6)
#' sheet <- xl_sheet(df, merge = xl_merge("A5:B5", "Total",
#'                                        format = xl_font(bold = TRUE)))
#' tmp <- write_xlsx(list(Data = sheet))
xl_merge <- function(range, value = NULL, format = NULL) {
  if (missing(range) || is.null(range))
    stop("`range` must name the cells to merge", call. = FALSE)
  value <- .as_display_string(value, "value")
  if (!is.null(format) && !is_xl_format(format))
    stop("`format` must be an xl_format object", call. = FALSE)
  structure(list(range = range, value = value, format = format),
            class = "xl_merge")
}

#' @export
print.xl_merge <- function(x, ...) {
  cat(sprintf("<xl_merge: %s%s>\n",
              if (is.character(x$range)) x$range else "<spec>",
              if (is.null(x$value)) ""
              else sprintf(" %s", encodeString(x$value, quote = '"'))))
  invisible(x)
}

# Resolve a sheet's merges to the payloads C applies, registering each format.
# Normalise one merge or a list of them to a checked list.
.merge_list <- function(merge, arg = "merge") {
  if (is.null(merge)) return(list())
  ms <- if (inherits(merge, "xl_merge")) list(merge) else merge
  if (!is.list(ms))
    stop(sprintf("`%s` must be an xl_merge object or a list of them", arg),
         call. = FALSE)
  for (i in seq_along(ms))
    if (!inherits(ms[[i]], "xl_merge"))
      stop(sprintf("`%s[[%d]]` must be an xl_merge object", arg, i),
           call. = FALSE)
  ms
}

.resolve_merges <- function(el, df, reg, header_offset, props) {
  if (!inherits(el, "xl_sheet")) return(list())
  ms <- .merge_list(el$merge)
  lapply(seq_along(ms), function(i) {
    m <- ms[[i]]
    arg <- sprintf("merge[[%d]] range", i)
    q <- .xl_resolve_range(m$range, arg = arg, df = df,
                           header_offset = header_offset, allow_cell = FALSE)
    # libxlsxwriter refuses a single-cell merge, and so does Excel; catch it
    # here so the message names the merge rather than coming back as a bare
    # parameter-validation error
    if (q[1L] == q[3L] && q[2L] == q[4L])
      stop(sprintf("`%s` covers a single cell; a merge needs more than one",
                   arg), call. = FALSE)
    list(kind = "merge", range = as.integer(q),
         value = if (is.null(m$value)) NA_character_ else m$value,
         format_id = .register_format(reg,
                                      merge_xl_format(props$default_format,
                                                      m$format)))
  })
}

# =============================================================================
# xl_format -> chart line / fill / pattern / font
# =============================================================================
#
# libxlsxwriter styles a chart through four small structs, and offers a
# _set_line / _set_fill / _set_pattern / _set_font function for each of fourteen
# targets -- series, data labels, markers, error bars, trendlines, the axis, its
# name and number fonts, major and minor gridlines, the legend, the title, the
# plot area, the chart area and the data table.  That is ~29 C functions
# expressing one idea: apply a format to a named part of the chart.
#
# So writexl exposes no chart-specific format type.  An ordinary xl_format is
# translated here, and the caller passes it to whichever chart argument it
# belongs to.
#
# The translation is lossy in known ways, and every loss is an error rather
# than a silent drop:
#
#   * A chart shape has no number format, no alignment and no cell protection,
#     so those groups are refused.  (An axis and data labels *do* take a number
#     format, but as their own argument -- it is not part of a shape's style.)
#   * lxw_chart_line has colour, dash and transparency but no *width*, so the
#     border styles that encode a width -- medium, thick, double, hair, and the
#     medium-* variants -- have no representation.  Refused, rather than
#     flattened to a plain line.
#   * A chart has one line, not four sides, so a border with differing sides is
#     refused rather than one side being picked.
# -----------------------------------------------------------------------------

# A colour on its way into a chart struct.  Every chart colour is tested for
# truth -- `if (line->color)` in _chart_write_a_ln(), and the same for fills and
# fonts -- so plain black, which xl_color() gives as 0, reads as "no colour set"
# and the part is drawn with Excel's default instead.  libxlsxwriter's own enum
# spells black LXW_COLOR_BLACK = 0x1000000 for exactly this reason, and
# LXW_COLOR_MASK strips the sentinel bit again on the way out.
.chart_color <- function(x) {
  if (is.null(x)) return(NULL)
  i <- as.integer(x)
  if (identical(i, 0L)) 0x1000000L else i
}

# Cell border style -> chart dash type.  Only exact equivalents are listed;
# .chart_dash() refuses the rest rather than approximating them.
.CHART_DASH <- c(
  thin       = 0L,   # LXW_CHART_LINE_DASH_SOLID
  dashed     = 3L,   # DASH
  "dash-dot" = 4L,   # DASH_DOT
  dotted     = 8L    # DOT
)

# Border styles a chart line cannot express, and why.
.CHART_DASH_UNSUPPORTED <- c(
  medium = "width", thick = "width", double = "width", hair = "width",
  "medium-dashed" = "width", "medium-dash-dot" = "width",
  "medium-dash-dot-dot" = "width", "dash-dot-dot" = "pattern",
  "slant-dash-dot" = "pattern"
)

.chart_dash <- function(style, arg) {
  if (is.null(style) || identical(style, "none")) return(NULL)
  if (style %in% names(.CHART_DASH)) return(unname(.CHART_DASH[[style]]))
  why <- .CHART_DASH_UNSUPPORTED[[style]]
  reason <- if (identical(why, "width"))
    paste0("a chart line has a colour, a dash pattern and a transparency, but ",
           "no width")
  else "Excel has no chart dash pattern matching it"
  stop(sprintf(paste0("`%s`: border style \"%s\" cannot be drawn on a chart -- ",
                      "%s.\n  Chart lines accept: %s (or \"none\" for no ",
                      "line)."),
               arg, style, reason,
               paste(sprintf("\"%s\"", names(.CHART_DASH)), collapse = ", ")),
       call. = FALSE)
}

# Cell fill pattern -> chart pattern.  A cell's patterns are ECMA-376's
# ST_PatternValue and a chart's are ST_PresetPatternVal: two vocabularies, so
# only the names that mean the same thing in both are listed, and
# .chart_pattern() refuses the rest rather than approximating them.  The grey
# screens map by the percentage their names state; 12.5% and 6.25% have no
# chart member, and the chart's grid and trellis sets have no dark/light pair.
.CHART_PATTERN <- c(
  "light-gray"       = 4L,   # LXW_CHART_PATTERN_PERCENT_25
  "medium-gray"      = 7L,   # PERCENT_50
  "dark-gray"        = 10L,  # PERCENT_75
  "light-down"       = 13L,  # LIGHT_DOWNWARD_DIAGONAL
  "light-up"         = 14L,  # LIGHT_UPWARD_DIAGONAL
  "dark-down"        = 15L,  # DARK_DOWNWARD_DIAGONAL
  "dark-up"          = 16L,  # DARK_UPWARD_DIAGONAL
  "light-vertical"   = 19L,  # LIGHT_VERTICAL
  "light-horizontal" = 20L,  # LIGHT_HORIZONTAL
  "dark-vertical"    = 23L,  # DARK_VERTICAL
  "dark-horizontal"  = 24L   # DARK_HORIZONTAL
)

.chart_pattern <- function(pattern, arg) {
  if (pattern %in% names(.CHART_PATTERN))
    return(unname(.CHART_PATTERN[[pattern]]))
  stop(sprintf(paste0("`%s`: fill pattern \"%s\" cannot be drawn on a chart -- ",
                      "Excel's chart patterns are a different set from its ",
                      "cell patterns, and none of them means the same thing.",
                      "\n  Chart patterns accept: %s."),
               arg, pattern,
               paste(sprintf("\"%s\"", names(.CHART_PATTERN)), collapse = ", ")),
       call. = FALSE)
}

# The groups a chart shape cannot express at all.  Alignment is the partial
# case: a chart shape has no horizontal or vertical placement for its text, but
# lxw_chart_font does carry a rotation, so xl_align(rotation =) turns a header
# cell and an axis label by the same angle.
.CHART_FORMAT_REFUSED <- c(
  num_format = paste0("a chart shape has no number format; set it on the axis ",
                      "or the data labels instead"),
  protection = "a chart shape cannot be locked or hidden"
)

# Not every chart part takes every piece.  A series is a shape and has no
# text, so it takes a line, a fill and a pattern but no font; a title is text
# and takes only a font.  A group the target cannot use has nowhere to go, so
# it is an error rather than a silent drop.
.CHART_PART_CANNOT <- c(
  font    = paste0("a font, but %s is a shape and has no text -- set the font ",
                   "on the chart title"),
  line    = "a border, but %s takes no line",
  fill    = "a fill, but %s takes no fill",
  pattern = "a pattern fill, but %s takes no fill"
)

.CHART_PART_GROUP <- c(font = "xl_font()", line = "xl_border()",
                       fill = "xl_fill()", pattern = "xl_fill()")

# Translate an xl_format into the pieces a chart understands.  Returns a list
# with any of `line`, `fill`, `pattern` and `font`, each ready for C.  `accept`
# names the pieces this particular chart part can use, and `part` names it for
# the error message.
.chart_format_payload <- function(fmt, arg,
                                  accept = c("line", "fill", "pattern", "font"),
                                  part = "a chart series") {
  if (is.null(fmt)) return(NULL)
  if (!is_xl_format(fmt))
    stop(sprintf("`%s` must be an xl_format object", arg), call. = FALSE)
  f <- unclass(fmt)

  for (g in names(.CHART_FORMAT_REFUSED))
    if (!is.null(f[[g]]))
      stop(sprintf("`%s` sets %s, which a chart cannot use: %s.", arg,
                   switch(g, num_format = "a number format",
                          protection = "protection"),
                   .CHART_FORMAT_REFUSED[[g]]), call. = FALSE)

  rotation <- NULL
  if (!is.null(f$align)) {
    placement <- setdiff(names(f$align), "rotation")
    if (length(placement))
      stop(sprintf(paste0("`%s` sets %s, which a chart cannot use: a chart ",
                          "shape places its own text. Of the alignment group ",
                          "it takes only `rotation`."),
                   arg, paste(sprintf("`%s`", placement), collapse = ", ")),
           call. = FALSE)
    # 0 has to travel as 360: lxw_chart_font treats a rotation of 0 as unset,
    # which is why libxlsxwriter spells "explicitly upright" 360
    rotation <- if (identical(as.integer(f$align$rotation), 0L)) 360L
                else as.integer(f$align$rotation)
  }

  out <- list()

  if (!is.null(f$font) || !is.null(rotation)) {
    fo <- f$font
    ent <- list()
    if (!is.null(fo$name))  ent$name <- fo$name
    if (!is.null(fo$size))  ent$size <- as.numeric(fo$size)
    if (!is.null(fo$color)) ent$color <- .chart_color(fo$color)
    if (isTRUE(fo$bold))    ent$bold <- 1L
    if (isTRUE(fo$italic))  ent$italic <- 1L
    # lxw_chart_font.underline is a flag, not Excel's five-way choice, so any
    # underline at all becomes underlined; "none" means not underlined
    if (!is.null(fo$underline) && !identical(fo$underline, "none"))
      ent$underline <- 1L
    if (!is.null(rotation)) ent$rotation <- rotation
    if (length(ent)) out$font <- ent
  }

  if (!is.null(f$fill)) {
    fi <- f$fill
    # a real pattern needs two colours and becomes lxw_chart_pattern; a plain
    # or solid fill becomes lxw_chart_fill
    if (!is.null(fi$pattern) && !identical(fi$pattern, "solid") &&
        !identical(fi$pattern, "none")) {
      if (is.null(fi$foreground) || is.null(fi$background))
        stop(sprintf(paste0("`%s` sets a pattern, which a chart draws from a ",
                            "foreground and a background colour -- give both."),
                     arg), call. = FALSE)
      out$pattern <- list(fg_color = .chart_color(fi$foreground),
                          bg_color = .chart_color(fi$background),
                          type = .chart_pattern(fi$pattern, arg))
    } else {
      ent <- list()
      if (!is.null(fi$background)) ent$color <- .chart_color(fi$background)
      if (identical(fi$pattern, "none")) ent$none <- 1L
      if (!is.null(fi$transparency))
        ent$transparency <- as.integer(fi$transparency)
      if (length(ent)) out$fill <- ent
    }
  }

  if (!is.null(f$border)) {
    b <- f$border
    sides <- c(b$left, b$right, b$top, b$bottom)
    cols  <- c(b$left_color, b$right_color, b$top_color, b$bottom_color)
    if (length(unique(sides)) > 1L || length(unique(cols)) > 1L)
      stop(sprintf(paste0("`%s` gives different borders on different sides; a ",
                          "chart has one line. Use xl_border(all = ) with a ",
                          "single style and colour."), arg), call. = FALSE)
    style <- if (length(sides)) sides[[1L]] else NULL
    ent <- list()
    if (identical(style, "none")) ent$none <- 1L
    else {
      d <- .chart_dash(style, arg)
      if (!is.null(d)) ent$dash_type <- d
    }
    if (length(cols)) ent$color <- .chart_color(cols[[1L]])
    if (!is.null(b$transparency))
      ent$transparency <- as.integer(b$transparency)
    if (length(ent)) out$line <- ent
  }

  extra <- setdiff(names(out), accept)
  if (length(extra))
    stop(sprintf("`%s` sets %s. Use %s instead.", arg,
                 sprintf(.CHART_PART_CANNOT[[extra[[1L]]]], part),
                 paste(unique(.CHART_PART_GROUP[accept]), collapse = " or ")),
         call. = FALSE)

  if (!length(out))
    stop(sprintf("`%s` is an empty format; give %s for the chart to use.", arg,
                 paste(unique(.CHART_PART_GROUP[accept]), collapse = " or ")),
         call. = FALSE)
  out
}

# =============================================================================
# Rich strings: one cell, several differently formatted runs
# =============================================================================
#
# A rich string is a single cell whose text is split into runs, each with its
# own font.  libxlsxwriter renders a run's format with lxw_styles_write_rich_font(),
# which emits only the <font> element -- so a run carries font properties and
# nothing else.  As with xl_comment(), the styling input is an ordinary
# xl_format and the unsupported groups are warned about rather than silently
# dropped (see .check_rich_string_format()).
#
# Excel itself imposes the shape: a run must have text, and a "rich" string
# needs at least two runs.  A single-format string is a plain character value.
# -----------------------------------------------------------------------------

#' One run of a rich (multi-format) string
#'
#' @description
#' `xl_rich_run()` is one fragment of an [xl_rich_string()]: a piece of text
#' plus the font it is drawn in.
#'
#' @param value A single non-`NA`, non-empty string: the run's text.
#' @param format An optional [xl_format].  Only its **font** properties apply
#'   (see [xl_font()]); a run has no fill, border, alignment or number format,
#'   and supplying one warns.  `NULL` draws the run in the cell's own font.
#' @return An `xl_rich_run` object.
#' @family cell content
#' @seealso [xl_rich_string], [xl_format]
#' @export
#' @examples
#' xl_rich_run("bold", xl_font(bold = TRUE))
#' xl_rich_run("plain")
xl_rich_run <- function(value, format = NULL) {
  value <- .as_display_string(value, "value", null_ok = FALSE)
  if (!nzchar(value))
    stop("`value` must not be empty: Excel has no representation for an empty ",
         "run of a rich string", call. = FALSE)
  if (!is.null(format)) {
    if (!is_xl_format(format))
      stop("`format` must be an xl_format object", call. = FALSE)
    .check_rich_string_format(format)
  }
  structure(list(value = value, format = format), class = "xl_rich_run")
}

#' A cell whose text has several formats
#'
#' @description
#' `xl_rich_string()` builds the value of a single cell out of differently
#' formatted runs, so that one cell can read "This is **bold** text".  Pass it
#' as the `value` of [xl_cell_general()].
#'
#' Excel requires at least two runs: a string with one format is an ordinary
#' character value, so pass it as one.
#'
#' @param ... Runs, in order.  A bare string is taken as an unformatted run; an
#'   [xl_rich_run()] carries its own font.  Lists of either are flattened, so
#'   runs can be assembled programmatically.
#' @return An `xl_rich_string` object: a list of runs.
#' @family cell content
#' @seealso [xl_rich_run], [xl_cell_general], [xl_font]
#' @export
#' @examples
#' xl_rich_string("This is ", xl_rich_run("bold", xl_font(bold = TRUE)), " text")
#'
#' # in a cell, with a cell-wide format alongside the per-run fonts
#' xl_cell_general(
#'   value  = xl_rich_string("2 H", xl_rich_run("2", xl_font(script = "sub")), "O"),
#'   format = xl_align(horizontal = "center")
#' )
xl_rich_string <- function(...) {
  runs <- .flatten_rich_runs(list(...))
  if (length(runs) < 2L)
    stop("a rich string needs at least 2 runs (got ", length(runs),
         "); a string with a single format is an ordinary character value",
         call. = FALSE)
  structure(runs, class = "xl_rich_string")
}

#' @rdname xl_rich_string
#' @param x An object to test.
#' @export
is_xl_rich_string <- function(x) inherits(x, "xl_rich_string")

# Turn the `...` of xl_rich_string() into a flat list of xl_rich_run objects.
# Bare strings become unformatted runs; nested lists are flattened so runs may
# be built up with lapply() and spliced in.
.flatten_rich_runs <- function(x) {
  out <- list()
  for (el in x) {
    if (inherits(el, "xl_rich_run")) {
      out <- c(out, list(el))
    } else if (is.character(el)) {
      if (!length(el))
        stop("a character run must contain at least one string", call. = FALSE)
      out <- c(out, lapply(el, xl_rich_run))
    } else if (is.list(el)) {
      out <- c(out, .flatten_rich_runs(el))
    } else {
      stop("each run must be a string or an xl_rich_run object, not ",
           paste(class(el), collapse = "/"), call. = FALSE)
    }
  }
  out
}

# Warn about xl_format properties a rich-string run cannot render.  A run is
# drawn by lxw_styles_write_rich_font(), which emits a <font> element only, so
# the whole font group applies and no other group does.
.check_rich_string_format <- function(format) {
  f <- unclass(format)
  bad <- character(0)
  for (g in c("fill", "border", "align", "num_format", "protection"))
    if (!is.null(f[[g]])) bad <- c(bad, g)
  if (isTRUE(f$quote_prefix)) bad <- c(bad, "quote_prefix")
  if (length(bad))
    warning("xl_rich_run(): a rich string run is drawn with a font only, so ",
            "these format properties will be ignored: ",
            paste(bad, collapse = ", "), call. = FALSE)
  invisible(bad)
}

#' @export
print.xl_rich_run <- function(x, ...) {
  cat(sprintf("<xl_rich_run: %s%s>\n", encodeString(x$value, quote = '"'),
              if (is.null(x$format)) "" else " (formatted)"))
  invisible(x)
}

#' @export
print.xl_rich_string <- function(x, ...) {
  runs <- unclass(x)
  cat(sprintf("<xl_rich_string: %d runs>\n", length(runs)))
  for (r in runs)
    cat(sprintf("  %s%s\n", encodeString(r$value, quote = '"'),
                if (is.null(r$format)) "" else "  <formatted>"))
  invisible(x)
}

#' @export
format.xl_rich_string <- function(x, ...) {
  paste(vapply(unclass(x), function(r) r$value, character(1)), collapse = "")
}

#' @description
#' `as.character()` returns the cell's text with the per-run fonts dropped.
#' @rdname xl_rich_string
#' @export
as.character.xl_rich_string <- function(x, ...) format(x)

# Flatten a rich string to the parallel text / format-id vectors C consumes,
# registering each run's format in the workbook registry.  Run formats cascade
# over the workbook default_format like every other format path.
.rich_string_c_payload <- function(rs, reg, props) {
  runs <- unclass(rs)
  list(
    rich_text = vapply(runs, function(r) r$value, character(1)),
    rich_format_id = vapply(runs, function(r)
      .register_format(reg, merge_xl_format(props$default_format, r$format)),
      integer(1))
  )
}

# Test helpers: write an xlsx and read back parts of its XML for inspection.

# Unzip a written workbook and return the parsed/raw contents of one part.
xlsx_part <- function(path, part = "xl/styles.xml", raw = FALSE) {
  d <- tempfile("xlsx_")
  dir.create(d)
  utils::unzip(path, exdir = d)
  f <- file.path(d, part)
  if (!file.exists(f)) return(if (raw) "" else NULL)
  if (raw) paste(readLines(f, warn = FALSE), collapse = "\n") else xml2::read_xml(f)
}

# Convenience: write a data frame (or workbook) and return styles.xml as a
# single string for token matching.
styles_string <- function(x, ...) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(x, tmp, ...)
  xlsx_part(tmp, "xl/styles.xml", raw = TRUE)
}

# Write and return the path (for readback / multi-part inspection).
write_tmp <- function(x, ...) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(x, tmp, ...)
  tmp
}

# A minimal valid 1x1 PNG.  Hardcoded rather than shipped as a fixture or drawn
# with png(), which needs a working graphics device that CRAN machines may not
# have.
PNG_1x1 <- as.raw(c(
  0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D,
  0x49, 0x48, 0x44, 0x52, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01,
  0x08, 0x06, 0x00, 0x00, 0x00, 0x1F, 0x15, 0xC4, 0x89, 0x00, 0x00, 0x00,
  0x0A, 0x49, 0x44, 0x41, 0x54, 0x78, 0x9C, 0x63, 0x00, 0x01, 0x00, 0x00,
  0x05, 0x00, 0x01, 0x0D, 0x0A, 0x2D, 0xB4, 0x00, 0x00, 0x00, 0x00, 0x49,
  0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82))

png_file <- function() {
  p <- tempfile(fileext = ".png")
  writeBin(PNG_1x1, p)
  p
}

# Data validation.  libxlsxwriter's 17 validation types are collapsed to five
# kinds plus list/custom/any, with the formula and datetime variants inferred
# from the R type of the limit.

dv <- function(..., df = data.frame(qty = 1:3)) {
  x <- xlsx_part(write_tmp(list(D = xl_sheet(df, validation = list(...)))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  m <- regmatches(x, regexpr("<dataValidations.*</dataValidations>", x))
  if (!length(m)) NA_character_ else m
}

# ── Numeric and length kinds ──────────────────────────────────────────────────

test_that("an integer bound writes a whole-number validation", {
  x <- dv(xl_validation("A2:A4", type = "integer", min = 1, max = 10))
  expect_match(x, 'type="whole"', fixed = TRUE)
  expect_match(x, "<formula1>1</formula1>", fixed = TRUE)
  expect_match(x, "<formula2>10</formula2>", fixed = TRUE)
  expect_match(x, 'sqref="A2:A4"', fixed = TRUE)
})

test_that("min and max together imply between", {
  # criteria may be omitted when both bounds are given
  a <- dv(xl_validation("A2:A4", type = "integer", min = 1, max = 10))
  b <- dv(xl_validation("A2:A4", type = "integer", criteria = "between",
                        min = 1, max = 10))
  expect_equal(a, b)
})

test_that("the single-limit criteria are written with an operator", {
  x <- dv(xl_validation("A2:A4", type = "integer", criteria = ">", value = 5))
  expect_match(x, 'operator="greaterThan"', fixed = TRUE)
  expect_match(x, "<formula1>5</formula1>", fixed = TRUE)
  expect_match(dv(xl_validation("A2:A4", type = "decimal", criteria = "<=",
                                value = 2.5)),
               'operator="lessThanOrEqual"', fixed = TRUE)
})

test_that("decimal and length kinds are distinct types", {
  expect_match(dv(xl_validation("A2:A4", type = "decimal", criteria = ">",
                                value = 0)),
               'type="decimal"', fixed = TRUE)
  expect_match(dv(xl_validation("A2:A4", type = "length", criteria = "<=",
                                value = 10)),
               'type="textLength"', fixed = TRUE)
})

test_that("a formula limit selects the formula variant", {
  # the type is inferred from the limit being a "=" string
  x <- dv(xl_validation("A2:A4", type = "integer", criteria = ">",
                        value = "=$B$1"))
  expect_match(x, 'type="whole"', fixed = TRUE)
  expect_match(x, "<formula1>$B$1</formula1>", fixed = TRUE)
})

# ── Dates and times ───────────────────────────────────────────────────────────

test_that("a Date bound writes a date validation with a serial number", {
  x <- dv(xl_validation("A2:A4", type = "date", criteria = ">=",
                        value = as.Date("2024-01-01")))
  expect_match(x, 'type="date"', fixed = TRUE)
  # 2024-01-01 is Excel serial 45292
  expect_match(x, "<formula1>45292</formula1>", fixed = TRUE)
})

test_that("a POSIXct bound carries the time of day", {
  x <- dv(xl_validation("A2:A4", type = "date", criteria = ">=",
                        value = as.POSIXct("2024-01-01 12:00:00", tz = "UTC")))
  expect_match(x, 'type="date"', fixed = TRUE)
  expect_match(x, "45292.5", fixed = TRUE)   # midday
})

test_that("a date range uses both bounds", {
  x <- dv(xl_validation("A2:A4", type = "date",
                        min = as.Date("2024-01-01"),
                        max = as.Date("2024-12-31")))
  expect_match(x, "<formula1>45292</formula1>", fixed = TRUE)
  expect_match(x, "<formula2>45657</formula2>", fixed = TRUE)
})

# ── Dropdown lists ────────────────────────────────────────────────────────────

test_that("a list becomes a quoted CSV formula", {
  x <- dv(xl_validation("A2:A4", list = c("open", "high", "close")))
  expect_match(x, 'type="list"', fixed = TRUE)
  expect_match(x, '<formula1>"open,high,close"</formula1>', fixed = TRUE)
})

test_that("a list given as a range formula is used as-is", {
  x <- dv(xl_validation("A2:A4", list = "=$H$1:$H$5"))
  expect_match(x, 'type="list"', fixed = TRUE)
  expect_match(x, "<formula1>$H$1:$H$5</formula1>", fixed = TRUE)
})

test_that("a single-choice list is allowed", {
  expect_match(dv(xl_validation("A2:A4", list = "only")),
               '<formula1>"only"</formula1>', fixed = TRUE)
})

test_that("the dropdown arrow can be turned off", {
  expect_match(dv(xl_validation("A2:A4", list = c("a", "b"), dropdown = FALSE)),
               'showDropDown="1"', fixed = TRUE)   # Excel stores it inverted
})

test_that("a list whose joined length exceeds 255 characters is refused", {
  # Excel's limit is on the comma-joined string, not the number of choices
  long <- rep(strrep("x", 20), 20)          # 20 * 21 - 1 = 419 characters
  expect_error(xl_validation("A2:A4", list = long), "joined by commas")
  expect_error(xl_validation("A2:A4", list = long), "255")
  # and just under the limit is fine
  ok <- rep(strrep("x", 20), 12)            # 251 characters
  expect_s3_class(xl_validation("A2:A4", list = ok), "xl_validation")
})

# ── custom and any ────────────────────────────────────────────────────────────

test_that("a custom formula validation is written", {
  x <- dv(xl_validation("A2:A4", type = "custom", value = "=A2>A1"))
  expect_match(x, 'type="custom"', fixed = TRUE)
  expect_match(x, "<formula1>A2&gt;A1</formula1>", fixed = TRUE)
})

test_that("type = any carries only the messages", {
  x <- dv(xl_validation("A2:A4", type = "any", input_title = "Hint",
                        input_message = "Anything goes"))
  expect_match(x, 'promptTitle="Hint"', fixed = TRUE)
  expect_match(x, 'prompt="Anything goes"', fixed = TRUE)
})

# ── Messages and toggles ──────────────────────────────────────────────────────

test_that("input and error text reach the file", {
  x <- dv(xl_validation("A2:A4", type = "integer", criteria = ">", value = 0,
                        input_title = "Quantity", input_message = "A number",
                        error_title = "Bad", error_message = "Must exceed 0"))
  expect_match(x, 'promptTitle="Quantity"', fixed = TRUE)
  expect_match(x, 'prompt="A number"', fixed = TRUE)
  expect_match(x, 'errorTitle="Bad"', fixed = TRUE)
  expect_match(x, 'error="Must exceed 0"', fixed = TRUE)
})

test_that("error_type selects the alert style", {
  base <- xl_validation("A2:A4", type = "integer", criteria = ">", value = 0)
  expect_false(grepl("errorStyle", dv(base), fixed = TRUE))   # stop is default
  expect_match(dv(xl_validation("A2:A4", type = "integer", criteria = ">",
                                value = 0, error_type = "warning")),
               'errorStyle="warning"', fixed = TRUE)
  expect_match(dv(xl_validation("A2:A4", type = "integer", criteria = ">",
                                value = 0, error_type = "information")),
               'errorStyle="information"', fixed = TRUE)
})

test_that("the on-by-default toggles are only written when turned off", {
  on <- dv(xl_validation("A2:A4", type = "integer", criteria = ">", value = 0))
  expect_match(on, 'allowBlank="1"', fixed = TRUE)
  off <- dv(xl_validation("A2:A4", type = "integer", criteria = ">", value = 0,
                          ignore_blank = FALSE, show_input = FALSE,
                          show_error = FALSE))
  expect_false(grepl('allowBlank="1"', off, fixed = TRUE))
  expect_false(grepl('showInputMessage="1"', off, fixed = TRUE))
})

# ── Validation of the arguments themselves ────────────────────────────────────

test_that("criteria is required for the comparison kinds", {
  for (k in c("integer", "decimal", "date", "time", "length"))
    expect_error(xl_validation("A2:A4", type = k, value = 1),
                 "`criteria` is required", label = k)
})

test_that("criteria is refused for the kinds that carry their own meaning", {
  expect_error(xl_validation("A2:A4", list = c("a", "b"), criteria = ">"),
               "does not apply")
  expect_error(xl_validation("A2:A4", type = "custom", value = "=A1",
                             criteria = ">"), "does not apply")
  expect_error(xl_validation("A2:A4", type = "any", criteria = ">"),
               "does not apply")
})

test_that("between needs both bounds and the others need a value", {
  expect_error(xl_validation("A2:A4", type = "integer", criteria = "between",
                             min = 1), "needs both `min` and `max`")
  expect_error(xl_validation("A2:A4", type = "integer",
                             criteria = "not between", max = 9),
               "needs both `min` and `max`")
  expect_error(xl_validation("A2:A4", type = "integer", criteria = ">"),
               "needs a `value`")
})

test_that("custom needs its formula and list needs its choices", {
  expect_error(xl_validation("A2:A4", type = "custom"), "needs a `value`")
  expect_error(xl_validation("A2:A4", list = character(0)),
               "at least one choice")
  expect_error(xl_validation("A2:A4", list = c("a", NA)), "must not be NA")
  expect_error(xl_validation("A2:A4", list = 1:3), "character vector")
})

test_that("titles and messages are capped at Excel's limits", {
  expect_error(xl_validation("A2:A4", type = "any",
                             input_title = strrep("x", 33)),
               "at most 32 characters")
  expect_error(xl_validation("A2:A4", type = "any",
                             error_title = strrep("x", 33)),
               "at most 32 characters")
  expect_error(xl_validation("A2:A4", type = "any",
                             input_message = strrep("x", 256)),
               "at most 255 characters")
  # exactly at the limit is fine
  expect_s3_class(xl_validation("A2:A4", type = "any",
                                input_title = strrep("x", 32)),
                  "xl_validation")
})

test_that("the limits count characters, not bytes", {
  s <- strrep("é", 32)          # 64 bytes, 32 characters
  expect_gt(nchar(s, type = "bytes"), 32L)
  expect_s3_class(xl_validation("A2:A4", type = "any", input_title = s),
                  "xl_validation")
})

test_that("bad types and ranges are rejected", {
  expect_error(xl_validation("A2:A4", type = "colour"), "type")
  expect_error(xl_validation(), "must name the cells")
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1),
                                           validation = "A1:B2"))),
               "must be an xl_validation object")
})

# ── Wiring ────────────────────────────────────────────────────────────────────

test_that("a single validation need not be wrapped in a list", {
  x <- xlsx_part(write_tmp(list(D = xl_sheet(data.frame(qty = 1:3),
    validation = xl_validation("A2:A4", type = "integer", criteria = ">",
                               value = 0)))),
    "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(x, 'type="whole"', fixed = TRUE)
})

test_that("several validations coexist", {
  x <- dv(xl_validation("A2:A4", type = "integer", min = 1, max = 10),
          xl_validation("B2:B4", list = c("a", "b")))
  expect_match(x, 'count="2"', fixed = TRUE)
})

test_that("a validation accepts a data-frame-relative spec", {
  df <- data.frame(a = 1:3, b = 1:3)
  x <- dv(xl_validation(list(rows = 1:3, cols = "a"), type = "integer",
                        criteria = ">", value = 0), df = df)
  # data rows 1:3 are sheet rows 2:4 once the header is counted
  expect_match(x, 'sqref="A2:A4"', fixed = TRUE)
})

test_that("validation does not disturb the cells or need streaming off", {
  skip_if_not_installed("readxl")
  df <- data.frame(qty = 1:3)
  p <- write_tmp(list(D = xl_sheet(df,
    validation = xl_validation("A2:A4", type = "integer", min = 1, max = 10))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
  # validations are written at packaging time, so row streaming stays on when
  # it is asked for -- a frame this small would not stream on size alone
  expect_true(is.null(xlsx_part(
    write_tmp(list(D = xl_sheet(df, validation = xl_validation(
      "A2:A4", type = "integer", min = 1, max = 10))),
      constant_memory = TRUE),
    "xl/sharedStrings.xml")))
})

test_that("the print method runs", {
  expect_output(print(xl_validation("A2:A4", list = c("a", "b"))),
                "xl_validation")
})

test_that("a validation with no type is the `any` kind", {
  # no `type` and no `list` leaves the "any" kind, which validates nothing but
  # still shows the input message
  v <- unclass(xl_validation("A1:A5", input_message = "type something"))
  expect_equal(v$kind, "any")
  p <- write_tmp(list(D = xl_sheet(data.frame(a = 1:3), validation =
    xl_validation("A2:A4", input_message = "type something"))))
  expect_match(xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE),
               "dataValidation", fixed = TRUE)
})

test_that("a character bound is passed through as a formula", {
  p <- write_tmp(list(D = xl_sheet(data.frame(a = 1:3), validation =
    xl_validation("A2:A4", type = "integer", criteria = "between",
                  min = "=MIN($A$2:$A$4)", max = "=MAX($A$2:$A$4)"))))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(x, "MIN($A$2:$A$4)", fixed = TRUE)
  expect_match(x, "MAX($A$2:$A$4)", fixed = TRUE)
})

test_that("a validation list must hold xl_validation objects", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                           validation = list("nope")))),
               "must be an xl_validation object")
})

test_that("a type of NA is the `any` kind", {
  expect_equal(unclass(xl_validation("A1:A5", type = NA))$kind, "any")
})

# An xl_cell_general is a vector, not a list.
#
# The distinction is not cosmetic.  `[<-.data.frame` branches on
# is.list(value): a list is read as a *list of columns*, so a list-backed cell
# vector could not be assigned with `df[, j] <- cells` at all -- it scattered
# one record per column and dropped the class, which is how writexl 2.0.0 broke
# a reverse dependency that had worked since 1.5.4.
#
# The carrier is an integer index; the per-cell records ride in an attribute,
# the shape factor uses.  These tests pin the vector behaviour rather than the
# representation, except where the representation *is* the behaviour.

test_that("a cell vector is not a list", {
  x <- xl_cell_general(value = 1:3)
  expect_false(is.list(x))
  expect_type(unclass(x), "integer")
  expect_length(x, 3L)
  # the records are the content, and there is one per cell
  expect_length(.cell_records(x), 3L)
})

test_that("mixed types survive: the carrier constrains nothing", {
  x <- xl_cell_general(value = list(1.5, "text", TRUE, as.Date("2024-01-01")))
  expect_equal(vapply(.cell_records(x), function(r) class(r$value)[1L],
                      character(1)),
               c("numeric", "character", "logical", "Date"))
  # and they reach the file as their own types.  The Date arrives as its Excel
  # serial rather than a formatted date: a mixed cell column carries no column
  # number format, which is behaviour this rewrite inherited unchanged --
  # verified identical against the pre-rewrite tree, and pinned here so that
  # fixing it is a deliberate act rather than an accident.
  p <- write_tmp(data.frame(a = 1:4, b = x))
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$b, c("1.5", "text", "TRUE", "45292"))
})

# ── Assignment into a data frame ──────────────────────────────────────────────

test_that("`df[, j] <- cells` keeps the class, which is the whole point", {
  d <- data.frame(a = 1:3)
  d[, 2] <- xl_cell_general(value = 4:6)
  expect_s3_class(d[[2]], "xl_cell_general")
  expect_length(.cell_records(d[[2]]), 3L)
})

test_that("the pattern that broke a reverse dependency works", {
  # BioMonTools seeds a one-row frame with a formula cell, then fills rows in
  # one at a time.  Under a list-backed cell vector the seed warned, dropped
  # the class, and left a bare list column that write_xlsx() refused.
  NOTES <- data.frame(matrix(ncol = 3))
  NOTES[, 2] <- xl_formula('=""')
  expect_s3_class(NOTES[[2]], "xl_cell_general")

  NOTES[1, 1] <- "Title"
  # writing a formula string as a value is what this pattern does; the warning
  # is the subject of its own test below
  expect_warning(NOTES[3, 2] <- "=LEFT(A1,1)", "not a formula")
  expect_s3_class(NOTES[[2]], "xl_cell_general")
  # the carrier and the records stay in step: a stale record list is how this
  # corrupts silently rather than loudly
  expect_length(.cell_records(NOTES[[2]]), length(NOTES[[2]]))
  expect_equal(nrow(NOTES), length(NOTES[[2]]))
  expect_equal(.cell_records(NOTES[[2]])[[3L]]$value, "=LEFT(A1,1)")

  expect_silent(write_tmp(list(NOTES = NOTES)))
})

test_that("`df[i, j] <- value` reaches the record, not the carrier", {
  d <- data.frame(a = 1:3)
  d[, 2] <- xl_cell_general(value = c(10, 20, 30))
  d[2, 2] <- 99
  expect_s3_class(d[[2]], "xl_cell_general")
  expect_equal(.cell_records(d[[2]])[[2L]]$value, 99)
  # unchanged cells keep theirs
  expect_equal(.cell_records(d[[2]])[[1L]]$value, 10)
  # and an xl_cell_general on the right replaces the cell outright
  d[3, 2] <- xl_cell_general(formula = "=A1+1")
  expect_equal(.cell_records(d[[2]])[[3L]]$formula, "=A1+1")
})

test_that("the other ways a cell column gets built still work", {
  x <- xl_cell_general(value = 1:3)
  d <- data.frame(a = 1:3)
  d$b <- x
  expect_s3_class(d$b, "xl_cell_general")
  expect_s3_class(data.frame(a = 1:3, b = x)$b, "xl_cell_general")
  d2 <- data.frame(a = 1:3)
  d2[["c"]] <- x
  expect_s3_class(d2$c, "xl_cell_general")
})

# ── Vector semantics ──────────────────────────────────────────────────────────

test_that("subsetting, concatenation and repetition act on cells", {
  x <- xl_cell_general(value = c(1, 2, 3))
  expect_length(x[2:3], 2L)
  expect_equal(x[2:3][[1L]]$value, 2)
  expect_length(c(x, x), 6L)
  expect_length(rep(x, 2L), 6L)
  expect_length(rep(x, length.out = 5L), 5L)
  # `[[` is one cell's record, never the carrier's integer
  expect_type(x[[1L]], "list")
  expect_equal(x[[1L]]$value, 1)
})

test_that("a cell vector round-trips through a write unchanged by the rewrite", {
  # the representation changed; what reaches the file must not have
  df <- data.frame(a = 1:3)
  df$b <- xl_cell_general(value = list(1, "two", NA),
                          formula = c(NA, NA, "=A2+1"),
                          format = xl_font(bold = TRUE))
  p <- write_tmp(list(D = df))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(x, "A2+1", fixed = TRUE)
  expect_match(xlsx_part(p, "xl/styles.xml", raw = TRUE), "<b/>", fixed = TRUE)
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$b[1L], "1")
  expect_equal(got$b[2L], "two")
})

test_that("assigning a bare string writes a string, not a formula", {
  # writexl's model is per cell: a formula is xl_formula(), anything else is a
  # value.  writexl 1.5.4 carried the "this is a formula column" idea on the
  # column's class, so a string assigned into one became a formula; there is no
  # column-level kind here, and `d[i, j] <- "=A1"` writes the six characters.
  d <- data.frame(a = 1:2)
  d[, 2] <- xl_formula(c("=A1", "=A2"))
  expect_warning(d[2, 2] <- "=SUM(A1:A2)", "not a formula")
  expect_equal(.cell_records(d[[2]])[[2L]]$value, "=SUM(A1:A2)")
  expect_true(is.na(.cell_records(d[[2]])[[2L]]$formula))
  # the formula spelling is explicit, and still available
  d[2, 2] <- xl_formula("=SUM(A1:A2)")
  expect_equal(.cell_records(d[[2]])[[2L]]$formula, "=SUM(A1:A2)")
})

test_that("a frame extended past a cell column pads it with empty cells", {
  # R extends rows behind our back, through xpdrows.data.frame(); the records
  # are padded on read so the new cells are blank rather than absent
  d <- data.frame(matrix(ncol = 2))
  d[, 1] <- xl_cell_general(value = 1)
  d[4, 2] <- "x"
  expect_equal(nrow(d), 4L)
  expect_length(d[[1]], 4L)
  expect_length(.cell_records(d[[1]]), 4L)
  expect_true(is.na(.cell_records(d[[1]])[[3L]]$value))
  expect_silent(write_tmp(list(D = d)))
})

test_that("`[[<-` replaces one cell's record", {
  x <- xl_cell_general(value = c(1, 2, 3))
  x[[2L]] <- .cell_records(xl_cell_general(formula = "=A1"))[[1L]]
  expect_s3_class(x, "xl_cell_general")
  expect_equal(x[[2L]]$formula, "=A1")
  expect_equal(x[[1L]]$value, 1)
  expect_length(x, 3L)
})

test_that("`length<-` grows and shrinks a cell vector", {
  x <- xl_cell_general(value = c(1, 2, 3))
  length(x) <- 5L
  expect_length(x, 5L)
  expect_length(.cell_records(x), 5L)
  # the new cells are empty rather than absent
  expect_true(is.na(x[[5L]]$value))
  length(x) <- 2L
  expect_length(x, 2L)
  expect_equal(x[[2L]]$value, 2)
})

test_that("writing a formula string as a value says so", {
  # a cell column has no column-wide "these are all formulas", so this writes
  # the characters.  That is the rule, but it is a quiet way to lose every
  # formula on a sheet, so it warns -- and points at the spelling that works.
  d <- data.frame(a = 1:2)
  d[, 2] <- xl_cell_general(value = c(1, 2))
  expect_warning(d[2, 2] <- "=SUM(A1)", "Use xl_formula")
  expect_warning(d[2, 2] <- "=SUM(A1)", "df[[j]] <- xl_formula", fixed = TRUE)

  # and stays quiet everywhere the intent is already clear
  expect_silent(d[2, 2] <- "plain text")
  expect_silent(d[2, 2] <- 99)
  expect_silent(d[2, 2] <- NA)
  expect_silent(d[2, 2] <- xl_formula("=SUM(A1)"))
  expect_silent(d[2, 2] <- xl_cell_general(formula = "=SUM(A1)"))
  # constructing a text cell that looks like a formula is deliberate and
  # documented, so it is not the assignment's business to second-guess it
  expect_silent(xl_cell_general(value = "=SUM(A1)"))
})

# Page setup: how a worksheet prints.  None of this touches the cells, so the
# assertions are all against the print-related parts of the worksheet XML.

page_xml <- function(page, df = data.frame(x = 1:3)) {
  xlsx_part(write_tmp(list(D = xl_sheet(df, page = page))),
            "xl/worksheets/sheet1.xml", raw = TRUE)
}

# ── Construction and validation ───────────────────────────────────────────────

test_that("an empty page setup is inert", {
  expect_s3_class(xl_page_setup(), "xl_page_setup")
  expect_length(unclass(xl_page_setup()), 0L)
  # a sheet with an empty setup writes the same print XML as one with none
  plain <- xlsx_part(write_tmp(list(D = xl_sheet(data.frame(x = 1:3)))),
                     "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_equal(page_xml(xl_page_setup()), plain)
})

test_that("orientation is validated and applied", {
  expect_error(xl_page_setup(orientation = "sideways"), "orientation")
  expect_match(page_xml(xl_page_setup(orientation = "landscape")),
               'orientation="landscape"', fixed = TRUE)
  expect_match(page_xml(xl_page_setup(orientation = "portrait")),
               'orientation="portrait"', fixed = TRUE)
})

test_that("paper accepts names case-insensitively and raw integers", {
  expect_match(page_xml(xl_page_setup(paper = "A4")), 'paperSize="9"',
               fixed = TRUE)
  expect_match(page_xml(xl_page_setup(paper = "a4")), 'paperSize="9"',
               fixed = TRUE)
  expect_match(page_xml(xl_page_setup(paper = "letter")), 'paperSize="1"',
               fixed = TRUE)
  # an unnamed specialist size is still reachable by index (37 = Monarch)
  expect_match(page_xml(xl_page_setup(paper = 37L)), 'paperSize="37"',
               fixed = TRUE)
  expect_error(xl_page_setup(paper = "A9"), "must be one of")
  expect_error(xl_page_setup(paper = 999L), "paper")
})

test_that("margins accept one value, four values, or a named subset", {
  expect_match(page_xml(xl_page_setup(margins = 1)),
               'left="1" right="1" top="1" bottom="1"', fixed = TRUE)
  expect_match(page_xml(xl_page_setup(margins = c(0.5, 0.6, 0.7, 0.8))),
               'left="0.5" right="0.6" top="0.7" bottom="0.8"', fixed = TRUE)
  # a named subset leaves the other sides at Excel's defaults
  m <- page_xml(xl_page_setup(margins = c(left = 1.25)))
  expect_match(m, 'left="1.25"', fixed = TRUE)
  expect_match(m, 'right="0.7"', fixed = TRUE)
})

test_that("margins are validated", {
  expect_error(xl_page_setup(margins = "wide"), "numeric")
  expect_error(xl_page_setup(margins = c(1, 2)), "one value, four values")
  expect_error(xl_page_setup(margins = c(middle = 1)), "unknown `margins`")
  expect_error(xl_page_setup(margins = -1), "non-negative")
  # zero is a legitimate margin
  expect_match(page_xml(xl_page_setup(margins = 0)), 'left="0"', fixed = TRUE)
})

test_that("scale is bounded to Excel's range", {
  expect_error(xl_page_setup(scale = 5), "scale")
  expect_error(xl_page_setup(scale = 500), "scale")
  expect_match(page_xml(xl_page_setup(scale = 150)), 'scale="150"', fixed = TRUE)
})

test_that("fit_to sets the page counts and the fitToPage flag", {
  # 0 means "as many pages as needed in that direction"
  x <- page_xml(xl_page_setup(fit_to = c(width = 1, height = 0)))
  expect_match(x, 'fitToHeight="0"', fixed = TRUE)
  expect_match(x, '<pageSetUpPr fitToPage="1"/>', fixed = TRUE)

  y <- page_xml(xl_page_setup(fit_to = c(2, 3)))
  expect_match(y, 'fitToWidth="2"', fixed = TRUE)
  expect_match(y, 'fitToHeight="3"', fixed = TRUE)

  # a single value is the width, height defaulting to 1 page
  expect_match(page_xml(xl_page_setup(fit_to = 1)),
               '<pageSetUpPr fitToPage="1"/>', fixed = TRUE)
})

test_that("fit_to is validated", {
  expect_error(xl_page_setup(fit_to = "one page"), "non-negative numbers")
  expect_error(xl_page_setup(fit_to = c(1, 2, 3)), "non-negative numbers")
  expect_error(xl_page_setup(fit_to = -1), "non-negative numbers")
  expect_error(xl_page_setup(fit_to = c(depth = 1)), "unknown `fit_to`")
})

# ── The flag options ──────────────────────────────────────────────────────────

test_that("centring is applied on each axis independently", {
  expect_match(page_xml(xl_page_setup(center_horizontally = TRUE)),
               'horizontalCentered="1"', fixed = TRUE)
  expect_match(page_xml(xl_page_setup(center_vertically = TRUE)),
               'verticalCentered="1"', fixed = TRUE)
  both <- page_xml(xl_page_setup(center_horizontally = TRUE,
                                 center_vertically = TRUE))
  expect_match(both, 'horizontalCentered="1"', fixed = TRUE)
  expect_match(both, 'verticalCentered="1"', fixed = TRUE)
  # FALSE is not TRUE: nothing is written
  expect_false(grepl("horizontalCentered",
                     page_xml(xl_page_setup(center_horizontally = FALSE)),
                     fixed = TRUE))
})

test_that("the print options are applied", {
  expect_match(page_xml(xl_page_setup(black_and_white = TRUE)),
               'blackAndWhite="1"', fixed = TRUE)
  expect_match(page_xml(xl_page_setup(row_col_headers = TRUE)),
               'headings="1"', fixed = TRUE)
  expect_match(page_xml(xl_page_setup(across = TRUE)),
               'pageOrder="overThenDown"', fixed = TRUE)
})

test_that("first_page sets the starting page number", {
  x <- page_xml(xl_page_setup(first_page = 3))
  expect_match(x, 'firstPageNumber="3"', fixed = TRUE)
  # Excel needs the companion flag or it ignores the number
  expect_match(x, 'useFirstPageNumber="1"', fixed = TRUE)
})

test_that("page_view opens the sheet in page-layout view", {
  expect_match(page_xml(xl_page_setup(page_view = TRUE)),
               'view="pageLayout"', fixed = TRUE)
})

# ── Wiring ────────────────────────────────────────────────────────────────────

test_that("settings combine without interfering", {
  x <- page_xml(xl_page_setup(orientation = "landscape", paper = "legal",
                              margins = 0.25, scale = 80,
                              center_horizontally = TRUE, first_page = 2))
  expect_match(x, 'orientation="landscape"', fixed = TRUE)
  expect_match(x, 'paperSize="5"', fixed = TRUE)
  expect_match(x, 'left="0.25"', fixed = TRUE)
  expect_match(x, 'scale="80"', fixed = TRUE)
  expect_match(x, 'horizontalCentered="1"', fixed = TRUE)
  expect_match(x, 'firstPageNumber="2"', fixed = TRUE)
})

test_that("page setup does not disturb the cells", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = 1:3, b = c("x", "y", "z"), stringsAsFactors = FALSE)
  p <- write_tmp(list(D = xl_sheet(df, page = xl_page_setup(
    orientation = "landscape", margins = 1, fit_to = c(1, 0)))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
})

test_that("`page` must be an xl_page_setup", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(x = 1), page = "landscape"))),
               "must be an xl_page_setup")
})

test_that("the print method runs", {
  expect_output(print(xl_page_setup(orientation = "landscape", margins = 1)),
                "xl_page_setup")
  expect_output(print(xl_page_setup()), "0 settings")
})

# ── Headers and footers ───────────────────────────────────────────────────────

test_that("header and footer text is written with Excel's codes intact", {
  x <- page_xml(xl_page_setup(header = "&LQ1 report&RPage &P of &N",
                              footer = "&Cdraft"))
  # the & codes survive, XML-escaped
  expect_match(x, "<oddHeader>&amp;LQ1 report&amp;RPage &amp;P of &amp;N</oddHeader>",
               fixed = TRUE)
  expect_match(x, "<oddFooter>&amp;Cdraft</oddFooter>", fixed = TRUE)
})

test_that("a header or footer works without a margin", {
  # no margin means the plain setter rather than the _opt form
  x <- page_xml(xl_page_setup(header = "&Cplain"))
  expect_match(x, "<oddHeader>&amp;Cplain</oddHeader>", fixed = TRUE)
  expect_match(x, 'header="0.3"', fixed = TRUE)   # Excel's default retained
})

test_that("header and footer margins reach pageMargins", {
  x <- page_xml(xl_page_setup(header = "&Ca", footer = "&Cb",
                              header_margin = 0.5, footer_margin = 0.6))
  expect_match(x, 'header="0.5"', fixed = TRUE)
  expect_match(x, 'footer="0.6"', fixed = TRUE)
})

test_that("a header/footer margin must be greater than zero", {
  # libxlsxwriter applies the margin only when it is > 0, so a zero would be
  # silently ignored rather than doing what it says
  expect_error(xl_page_setup(header_margin = 0), "greater than 0")
  expect_error(xl_page_setup(footer_margin = 0), "greater than 0")
  expect_error(xl_page_setup(header_margin = -1), "header_margin")
})

test_that("header and footer length is capped at 255 characters", {
  ok <- strrep("x", 255)
  expect_s3_class(xl_page_setup(header = ok), "xl_page_setup")
  expect_error(xl_page_setup(header = strrep("x", 256)),
               "at most 255 characters")
  expect_error(xl_page_setup(footer = strrep("x", 256)),
               "at most 255 characters")
})

test_that("the length limit counts characters, not bytes", {
  # libxlsxwriter measures with lxw_utf8_strlen(); a 200-character string of
  # 3-byte characters is well over 255 bytes but under the limit
  s <- strrep("\u00e9", 200)
  expect_gt(nchar(s, type = "bytes"), 255L)
  expect_s3_class(xl_page_setup(header = s), "xl_page_setup")
})

# ── Header and footer images ──────────────────────────────────────────────────

# A minimal valid 1x1 PNG, as in test-xl_image.R.
HF_PNG <- as.raw(c(
  0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D,
  0x49, 0x48, 0x44, 0x52, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01,
  0x08, 0x06, 0x00, 0x00, 0x00, 0x1F, 0x15, 0xC4, 0x89, 0x00, 0x00, 0x00,
  0x0A, 0x49, 0x44, 0x41, 0x54, 0x78, 0x9C, 0x63, 0x00, 0x01, 0x00, 0x00,
  0x05, 0x00, 0x01, 0x0D, 0x0A, 0x2D, 0xB4, 0x00, 0x00, 0x00, 0x00, 0x49,
  0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82))

hf_logo <- function() {
  p <- tempfile(fileext = ".png")
  writeBin(HF_PNG, p)
  p
}

# Write a sheet and list every part of the resulting zip.
hf_parts <- function(page) {
  p <- write_tmp(list(S = xl_sheet(data.frame(a = 1:3), page = page)))
  d <- tempfile()
  dir.create(d)
  utils::unzip(p, exdir = d)
  list(files = list.files(d, recursive = TRUE),
       sheet = xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE))
}

test_that("both spellings of the image placeholder are counted", {
  expect_equal(.hf_placeholders("&L&G"), 1L)
  expect_equal(.hf_placeholders("&G&C&G"), 2L)
  expect_equal(.hf_placeholders("&L&[Picture]"), 1L)
  expect_equal(.hf_placeholders("&G&[Picture]"), 2L)
  expect_equal(.hf_placeholders("plain"), 0L)
  expect_equal(.hf_placeholders(NULL), 0L)
})

test_that("a header image reaches the file with everything Excel needs", {
  # the image alone is not enough: Excel renders a header image through a VML
  # drawing, and without the legacyDrawingHF reference nothing appears
  r <- hf_parts(xl_page_setup(header = "&L&G",
                              header_image = list(left = hf_logo())))
  expect_true("xl/media/image1.png" %in% r$files)
  expect_true("xl/drawings/vmlDrawing1.vml" %in% r$files)
  expect_true("xl/drawings/_rels/vmlDrawing1.vml.rels" %in% r$files)
  expect_match(r$sheet, "<legacyDrawingHF r:id=", fixed = TRUE)
})

test_that("a footer image works, and may be given in memory", {
  r <- hf_parts(xl_page_setup(footer = "&R&G",
                              footer_image = list(right = HF_PNG)))
  expect_true("xl/media/image1.png" %in% r$files)
  expect_match(r$sheet, "<legacyDrawingHF r:id=", fixed = TRUE)
})

test_that("all three positions can carry an image", {
  a <- hf_logo()
  b <- tempfile(fileext = ".png"); writeBin(c(HF_PNG, as.raw(0)), b)
  cimg <- tempfile(fileext = ".png"); writeBin(c(HF_PNG, as.raw(c(0, 0))), cimg)
  r <- hf_parts(xl_page_setup(
    header = "&L&G&C&G&R&G",
    header_image = list(left = a, center = b, right = cimg)))
  # three distinct images, so three media files -- identical ones would be
  # deduplicated by libxlsxwriter and prove nothing
  expect_true(all(sprintf("xl/media/image%d.png", 1:3) %in% r$files))
})

test_that("placeholders and images must agree in number", {
  # libxlsxwriter enforces this too, but its error does not say which side is
  # short, and that is the only thing the caller needs to know
  expect_error(xl_page_setup(header = "&LPlain",
                             header_image = list(left = hf_logo())),
               "gives 1 image\\(s\\) but `header` has 0 image placeholder")
  expect_error(xl_page_setup(header = "&L&G"),
               "has 1 image placeholder.*but `header_image` gives 0")
  expect_error(xl_page_setup(footer = "&C&G"), "gives 0 image")
  # the message says how to fix the commoner direction
  expect_error(xl_page_setup(header = "&LPlain",
                             header_image = list(left = hf_logo())),
               "add &G where each image should sit")
})

test_that("header image positions are validated", {
  expect_error(xl_page_setup(header = "&L&G",
                             header_image = list(middle = hf_logo())),
               "unknown position\\(s\\): middle")
  expect_error(xl_page_setup(header = "&L&G", header_image = hf_logo()),
               "must be named by position")
  expect_error(xl_page_setup(header = "&L&G",
                             header_image = list(left = "nope.png")),
               "file does not exist")
  # the image itself goes through the same checks as xl_image()
  bad <- tempfile(fileext = ".png"); writeLines("not an image", bad)
  expect_error(xl_page_setup(header = "&L&G", header_image = list(left = bad)),
               "is not a PNG, JPEG, GIF or BMP")
})

test_that("a header without an image is unaffected", {
  r <- hf_parts(xl_page_setup(header = "&CPlain"))
  expect_match(r$sheet, "<oddHeader>&amp;CPlain</oddHeader>", fixed = TRUE)
  expect_false(any(grepl("vmlDrawing", r$files)))
  # a literal ampersand or an unrelated code is still fine
  expect_s3_class(xl_page_setup(header = "R&&D&Cmid"), "xl_page_setup")
})

test_that("header and footer do not disturb the cells", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = 1:2, stringsAsFactors = FALSE)
  p <- write_tmp(list(D = xl_sheet(df, page = xl_page_setup(
    header = "&Ctitle", footer = "&Rpage &P", header_margin = 0.4))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
})

# ── Print area and repeat rows/columns ───────────────────────────────────────
#
# These write defined names into workbook.xml, not into the worksheet part.

defined_names <- function(page, df = data.frame(x = 1:30)) {
  wb <- xlsx_part(write_tmp(list(D = xl_sheet(df, page = page))),
                  "xl/workbook.xml", raw = TRUE)
  m <- regmatches(wb, regexpr("<definedNames>.*</definedNames>", wb))
  if (!length(m)) NA_character_ else m
}

test_that("print_area becomes a Print_Area defined name", {
  expect_match(defined_names(xl_page_setup(print_area = "A1:B20")),
               "_xlnm.Print_Area", fixed = TRUE)
  expect_match(defined_names(xl_page_setup(print_area = "A1:B20")),
               "D!$A$1:$B$20", fixed = TRUE)
})

test_that("print_area accepts a data-frame-relative spec", {
  df <- data.frame(a = 1:5, b = 1:5, cc = 1:5)
  x <- defined_names(xl_page_setup(print_area = list(rows = 1:3, cols = c("a", "b"))),
                     df = df)
  # rows 1:3 are data rows, so sheet rows 2:4 with the header
  expect_match(x, "D!$A$2:$B$4", fixed = TRUE)
})

test_that("repeat_rows and repeat_cols become Print_Titles", {
  # a count repeats that many sheet rows, starting at the header
  expect_match(defined_names(xl_page_setup(repeat_rows = 1)),
               "D!$1:$1", fixed = TRUE)
  expect_match(defined_names(xl_page_setup(repeat_rows = 2)),
               "D!$1:$2", fixed = TRUE)
  expect_match(defined_names(xl_page_setup(repeat_cols = 1)),
               "D!$A:$A", fixed = TRUE)
  # a range string works too
  expect_match(defined_names(xl_page_setup(repeat_rows = "1:2")),
               "D!$1:$2", fixed = TRUE)
  expect_match(defined_names(xl_page_setup(repeat_cols = "A:B")),
               "D!$A:$B", fixed = TRUE)
  # both together
  expect_match(defined_names(xl_page_setup(repeat_rows = 1, repeat_cols = "A:A")),
               "_xlnm.Print_Titles", fixed = TRUE)
})

test_that("repeat arguments reject the wrong kind of range", {
  expect_error(xl_page_setup(repeat_rows = "A:B"), "row range")
  expect_error(xl_page_setup(repeat_cols = "1:2"), "column range")
  expect_error(xl_page_setup(repeat_rows = 0), "repeat_rows")
  expect_error(xl_page_setup(repeat_rows = list(1)), "count or range string")
})

test_that("a print area outside the data still resolves", {
  # a print area is a sheet range, not restricted to the written cells
  expect_match(defined_names(xl_page_setup(print_area = "A1:Z100")),
               "D!$A$1:$Z$100", fixed = TRUE)
})

# ── Page breaks ───────────────────────────────────────────────────────────────

breaks_xml <- function(page, df = data.frame(x = 1:30)) page_xml(page, df)

test_that("page breaks are written 0-based, one before the named row", {
  # h_breaks = 21 means the new page starts at row 21, which is 0-based 20
  x <- breaks_xml(xl_page_setup(h_breaks = c(11, 21)))
  expect_match(x, '<brk id="10"', fixed = TRUE)
  expect_match(x, '<brk id="20"', fixed = TRUE)
  expect_match(x, 'manualBreakCount="2"', fixed = TRUE)
})

test_that("vertical breaks accept numbers and column letters", {
  expect_match(breaks_xml(xl_page_setup(v_breaks = 2)), "<colBreaks",
               fixed = TRUE)
  expect_match(breaks_xml(xl_page_setup(v_breaks = 2)), '<brk id="1"', fixed = TRUE)
  # column "C" is the third column, so a break before it is 0-based 2
  expect_match(breaks_xml(xl_page_setup(v_breaks = "C")), '<brk id="2"',
               fixed = TRUE)
})

test_that("breaks are sorted and deduplicated", {
  x <- breaks_xml(xl_page_setup(h_breaks = c(21, 11, 21)))
  expect_match(x, 'manualBreakCount="2"', fixed = TRUE)
  # sorted: 10 must appear before 20
  expect_lt(regexpr('<brk id="10"', x, fixed = TRUE),
            regexpr('<brk id="20"', x, fixed = TRUE))
})

test_that("a break at position 1 is rejected", {
  # 0-based 0 terminates libxlsxwriter's array, so this would silently drop
  # every break rather than being merely useless
  expect_error(xl_page_setup(h_breaks = 1), "2 or greater")
  expect_error(xl_page_setup(v_breaks = 1), "2 or greater")
  expect_error(xl_page_setup(h_breaks = c(5, 1)), "2 or greater")
})

test_that("the break count is capped at Excel's limit", {
  expect_s3_class(xl_page_setup(h_breaks = 2:1024), "xl_page_setup")   # 1023
  expect_error(xl_page_setup(h_breaks = 2:1025), "at most 1023")
})

test_that("breaks are validated", {
  expect_error(xl_page_setup(h_breaks = "row 5"), "numeric vector")
  expect_error(xl_page_setup(h_breaks = c(5, NA)), "numeric vector")
})

test_that("combining breaks with fit_to warns", {
  # Excel ignores manual breaks when fit-to-page is on
  expect_warning(xl_page_setup(h_breaks = 10, fit_to = c(1, 0)),
                 "ignores manual page breaks")
  expect_silent(xl_page_setup(h_breaks = 10))
})

test_that("print area and breaks together leave the cells alone", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = 1:10, stringsAsFactors = FALSE)
  p <- write_tmp(list(D = xl_sheet(df, page = xl_page_setup(
    print_area = "A1:A11", repeat_rows = 1, h_breaks = 6))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
})

test_that("a header or footer image position may be given only once", {
  logo <- png_file()
  expect_error(xl_page_setup(header = "&L&G",
                             header_image = list(left = logo, left = logo)),
               "each position may be given only once")
  expect_error(xl_page_setup(header = "&L&G",
                             header_image = list(middle = logo)),
               "unknown position")
})

test_that("a header image may be an in-memory picture", {
  skip_if_not(isTRUE(capabilities("png")))
  p <- write_tmp(list(D = xl_sheet(data.frame(a = 1:3), page = xl_page_setup(
    header = "&L&G",
    header_image = list(left = as.raster(matrix("red", 2, 2)))))))
  d <- tempfile(); dir.create(d); utils::unzip(p, exdir = d)
  expect_true(any(grepl("^xl/media/", list.files(d, recursive = TRUE))))
})

# Tests for xl_sheet() autofilter and worksheet protection.

sheet1_xml <- function(x) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(x, tmp)
  xlsx_part(tmp, "xl/worksheets/sheet1.xml", raw = TRUE)
}

test_that("autofilter = TRUE covers the whole used range", {
  df <- data.frame(name = c("a", "b", "c"), value = 1:3)
  w <- sheet1_xml(list(D = xl_sheet(df, autofilter = TRUE)))
  expect_match(w, '<autoFilter ref="A1:B4"')   # header + 3 data rows, 2 cols
})

test_that("autofilter accepts an explicit range", {
  df <- data.frame(a = 1:3, b = 4:6, c = 7:9)
  w <- sheet1_xml(list(D = xl_sheet(df, autofilter = "A1:C4")))
  expect_match(w, '<autoFilter ref="A1:C4"')
})

test_that("autofilter is absent by default and validated", {
  w <- sheet1_xml(data.frame(a = 1:2))
  expect_false(grepl("autoFilter", w))
  expect_error(xl_sheet(data.frame(a = 1), autofilter = 1), "TRUE/FALSE or an Excel range")
  expect_error(xl_sheet(data.frame(a = 1), autofilter = c(TRUE, FALSE)), "TRUE/FALSE")
})

test_that(".parse_range parses and rejects", {
  expect_equal(.parse_range("A1:D51"), c(0L, 0L, 50L, 3L))
  expect_equal(.parse_range("B2:B2"), c(1L, 1L, 1L, 1L))
  expect_error(.parse_range("A1"), "A1:D51")
  expect_error(.parse_range("A1:9"), "A1:D51")
})

test_that("protect = TRUE adds default sheet protection", {
  w <- sheet1_xml(list(D = xl_sheet(data.frame(a = 1:2), protect = TRUE)))
  expect_match(w, "<sheetProtection")
  expect_match(w, 'sheet="1"')
})

test_that("protect = password sets a hashed password", {
  w <- sheet1_xml(list(D = xl_sheet(data.frame(a = 1:2), protect = "secret")))
  expect_match(w, "<sheetProtection")
  expect_match(w, "password=")
})

test_that("protect = list gives fine-grained control", {
  w <- sheet1_xml(list(D = xl_sheet(data.frame(a = 1:2),
    protect = list(password = "pw", format_cells = TRUE,
                   insert_rows = TRUE, sort = TRUE))))
  # a libxlsxwriter option set TRUE -> the action is allowed (attr = "0")
  expect_match(w, 'formatCells="0"')
  expect_match(w, 'insertRows="0"')
  expect_match(w, 'sort="0"')
  expect_match(w, "password=")
})

test_that("protect list without options uses defaults", {
  w <- sheet1_xml(list(D = xl_sheet(data.frame(a = 1:2),
                                    protect = list(password = "pw"))))
  expect_match(w, "<sheetProtection")
  expect_match(w, "password=")
})

test_that("protection is absent by default and validated", {
  w <- sheet1_xml(data.frame(a = 1:2))
  expect_false(grepl("sheetProtection", w))
  expect_error(xl_sheet(data.frame(a = 1), protect = list(1, 2)), "fully named")
  expect_error(xl_sheet(data.frame(a = 1), protect = list(bogus = TRUE)), "unknown")
  expect_error(xl_sheet(data.frame(a = 1), protect = 42), "TRUE/FALSE")
})

test_that("protect = NULL means no protection", {
  expect_silent(xl_sheet(data.frame(a = 1), protect = NULL))
  w <- sheet1_xml(list(D = xl_sheet(data.frame(a = 1:2), protect = NULL)))
  expect_false(grepl("sheetProtection", w))
})

test_that("cell locking interacts with sheet protection", {
  # a cell explicitly unlocked, on a protected sheet
  df <- data.frame(a = 1)
  df$b <- xl_cell_general(value = 2, format = xl_protection(locked = FALSE))
  w <- sheet1_xml(list(D = xl_sheet(df, protect = TRUE)))
  expect_match(w, "<sheetProtection")     # sheet protected
  s <- styles_string(list(D = xl_sheet(df, protect = TRUE)))
  expect_match(s, 'locked="0"')           # the cell is unlocked in the style
})

test_that("protection formatting on an unprotected sheet warns", {
  df <- data.frame(a = 1)
  df$b <- xl_cell_general(value = 2, format = xl_protection(locked = FALSE))
  expect_warning(write_tmp(df), "not protected")
  dfh <- data.frame(a = 1)
  dfh$b <- xl_cell_general(value = 2, format = xl_protection(hidden = TRUE))
  expect_warning(write_tmp(dfh), "not protected")
  # column-level protection warns and names the sheet
  sc <- xl_sheet(data.frame(x = 1:2),
                 cols = xl_col_spec("x", format = xl_protection(locked = FALSE)))
  expect_warning(write_tmp(list(MySheet = sc)), "MySheet")
})

test_that("no warning when protected or when there is no protection formatting", {
  df <- data.frame(a = 1)
  df$b <- xl_cell_general(value = 2, format = xl_protection(locked = FALSE))
  expect_no_warning(write_tmp(list(D = xl_sheet(df, protect = TRUE))))
  expect_no_warning(write_tmp(data.frame(a = 1:3)))
  # a cell left at the default locked = TRUE must NOT warn
  df2 <- data.frame(a = 1)
  df2$b <- xl_cell_general(value = 2, format = xl_protection(locked = TRUE))
  expect_no_warning(write_tmp(df2))
})

roundtrip <- function(df){
  readxl::read_xlsx(writexl::write_xlsx(df))
}

test_that("Types roundtrip properly", {
  kremlin <- "http://\u043F\u0440\u0435\u0437\u0438\u0434\u0435\u043D\u0442.\u0440\u0444"
  num <- c(NA_real_, pi, 1.2345e80)
  int <- c(NA_integer_, 0L, -100L)
  str <- c(NA_character_, "foo", kremlin) #note empty strings don't work yet
  # The zone is set explicitly: writexl drops a single shared time zone and
  # writes local wall-clock time, so pinning this column to UTC (offset 0) keeps
  # this test about type round-tripping. Time zones are covered separately below.
  time <- structure(Sys.time() + 1:3, tzone = "UTC")
  bigint <- bit64::as.integer64(.Machine$integer.max) ^ c(0, 1, 1.5)
  input <- data.frame(num = num, int = int, bigint = bigint, str = str,
                      time = time, stringsAsFactors = FALSE)
  expect_warning(output <- roundtrip(input), "int64")
  output$bigint <- bit64::as.integer64(output$bigint)
  attr(output$time, "tzone") <- attr(time, "tzone")
  expect_equal(input, as.data.frame(output))
})

# Extract the numeric <v> values of a written worksheet, keyed by cell
# reference (e.g. c(A2 = 1, B2 = 1)).  Reading the raw serials straight out of
# the XML keeps these assertions independent of readxl, so a writexl bug cannot
# be masked by a compensating reader bug.
sheet_serials <- function(path, part = "xl/worksheets/sheet1.xml") {
  xml <- xlsx_part(path, part, raw = TRUE)
  cells <- regmatches(xml, gregexpr("<c r=\"[A-Z]+[0-9]+\"[^>]*><v>[^<]*</v></c>", xml))[[1]]
  refs <- sub("^<c r=\"([A-Z]+[0-9]+)\".*$", "\\1", cells)
  vals <- as.numeric(sub("^.*<v>([^<]*)</v>.*$", "\\1", cells))
  stats::setNames(vals, refs)
}

# Excel's 1900 date system pretends 1900 was a leap year, so serial 60 is the
# nonexistent 1900-02-29 and every date from 1900-03-01 onwards is shifted by
# one.  These serials are Excel's documented ground truth.
date_boundary_cases <- data.frame(
  date = as.Date(c("1900-01-01", "1900-02-28", "1900-03-01", "1970-01-01", "2020-03-01")),
  serial = c(1, 59, 61, 25569, 43891)
)

test_that("Date and POSIXct columns write Excel's documented serial numbers", {
  d <- date_boundary_cases$date
  expected <- date_boundary_cases$serial
  df <- data.frame(dt = d, pt = as.POSIXct(paste(d, "00:00:00"), tz = "UTC"))
  serials <- sheet_serials(write_tmp(df))

  # Row 1 is the header; data starts at row 2.  Column A = Date, B = POSIXct.
  rows <- seq_along(d) + 1L
  expect_equal(unname(serials[paste0("A", rows)]), expected)
  expect_equal(unname(serials[paste0("B", rows)]), expected)

  # The two writers must agree cell for cell, not merely both look plausible.
  expect_equal(unname(serials[paste0("A", rows)]), unname(serials[paste0("B", rows)]))

  # Serial 60 is Excel's phantom 1900-02-29 and must never be written.
  expect_false(any(serials[paste0(c("A", "B"), rep(rows, each = 2))] == 60))
})

test_that("Integer-backed Date columns write the same serials as double-backed", {
  d <- date_boundary_cases$date
  int_date <- structure(as.integer(unclass(d)), class = "Date")
  expect_identical(typeof(int_date), "integer")

  serials <- sheet_serials(write_tmp(data.frame(dt = int_date)))
  expect_equal(unname(serials[paste0("A", seq_along(d) + 1L)]), date_boundary_cases$serial)
})

test_that("NA is preserved in Date columns of either backing type", {
  d <- c(as.Date("1900-02-28"), NA, as.Date("2020-03-01"))
  int_date <- structure(as.integer(unclass(d)), class = "Date")

  for (col in list(d, int_date)) {
    path <- write_tmp(data.frame(dt = col))
    serials <- sheet_serials(path)
    # The NA cell is written as an empty cell, so no <v> exists for A3.
    expect_equal(unname(serials[c("A2", "A4")]), c(59, 43891))
    expect_false("A3" %in% names(serials))
    expect_equal(as.Date(readxl::read_xlsx(path)[[1]]), d)
  }
})

test_that("Dates around the 1900 leap-day boundary roundtrip through readxl", {
  d <- date_boundary_cases$date
  out <- readxl::read_xlsx(write_tmp(data.frame(dt = d)))
  expect_equal(as.Date(out$dt), d)

  # Every day of the 1900 boundary window, not just the hand-picked ones.
  window <- seq(as.Date("1900-01-01"), as.Date("1900-03-05"), by = "day")
  out2 <- readxl::read_xlsx(write_tmp(data.frame(dt = window)))
  expect_equal(as.Date(out2$dt), window)
  expect_false(anyNA(out2$dt))
})

test_that("Writing formulas", {
  df <- data.frame(
    name = c("UCLA", "Berkeley"),
    founded = c(1919, 1868),
    website = xl_hyperlink(c("http://www.ucla.edu", "http://www.berkeley.edu"), "website")
  )

  # repeats a formula for entire column
  df$age <- xl_formula('=(YEAR(TODAY()) - INDIRECT("B" & ROW()))')

  # currently readxl does not support formulas so inspect manually
  expect_true(file.exists(write_xlsx(df)))
})

test_that("unsupported column types raise an error naming the column", {
  df <- data.frame(a = 1)
  df$bad <- complex(real = 1, imaginary = 1)
  expect_error(write_xlsx(df), "cannot write column")
  expect_error(write_xlsx(df), "'bad' (column 2)", fixed = TRUE)
  expect_error(write_xlsx(df), "complex")

  # a bare list column and a raw vector are equally unrepresentable; the
  # underlying type is reported when the class alone would be unhelpful
  df2 <- data.frame(a = 1); df2$bad <- I(list(1))
  expect_error(write_xlsx(df2), "(list)", fixed = TRUE)
  df3 <- data.frame(a = 1); df3$bad <- as.raw(1)
  expect_error(write_xlsx(df3), "raw")

  # every unsupported column is reported, not just the first
  df4 <- data.frame(a = 1)
  df4$b <- complex(real = 1, imaginary = 1)
  df4$c <- as.raw(2)
  expect_error(write_xlsx(df4), "'b'")
  expect_error(write_xlsx(df4), "'c'")
})

test_that("classed numeric columns such as difftime are still written", {
  # difftime is a double underneath, so it is written from its numeric value
  df <- data.frame(a = 1)
  df$elapsed <- as.difftime(c(1.5), units = "days")
  expect_silent(path <- write_xlsx(df))
  expect_equal(readxl::read_xlsx(path)$elapsed, 1.5)
})

test_that("supported column types are still accepted", {
  df <- data.frame(int = 1:2, dbl = c(1.5, 2.5), chr = c("x", "y"),
                   lgl = c(TRUE, FALSE), date = as.Date("2020-01-01") + 0:1,
                   stringsAsFactors = FALSE)
  df$time <- as.POSIXct(c("2020-01-01 10:00", "2020-01-02 10:00"), tz = "UTC")
  df$cell <- xl_cell_general(value = 1:2)
  expect_true(file.exists(write_xlsx(df)))
  # classes that normalize_df coerces must not trip the check
  expect_true(file.exists(write_xlsx(data.frame(f = factor(c("a", "b"))))))
  expect_true(file.exists(write_xlsx(data.frame(p = as.POSIXlt("2020-01-01")))))
})

# ── POSIXct time zones (Excel stores naive datetimes) ─────────────────────────

# The Excel serial for a naive wall clock, independent of writexl's own maths.
expected_serial <- function(wall_clock) {
  25569 + as.numeric(as.POSIXct(wall_clock, tz = "UTC")) / 86400
}
datetime_format_of <- function(path) {
  xml <- xlsx_part(path, "xl/styles.xml", raw = TRUE)
  if (grepl("HH:mm:ss UTC", xml, fixed = TRUE)) "utc-labelled" else "unlabelled"
}

test_that("a single time zone is dropped, writing local wall-clock time", {
  # the reproducer from the issue: Perth 01:00 must not become 17:00 UTC
  df <- data.frame(dt = as.POSIXct("2025-12-01 01:00:00", tz = "Australia/Perth"))
  path <- write_tmp(df)
  expect_equal(sheet_serials(path)[["A2"]],
               expected_serial("2025-12-01 01:00:00"))
  # and the default format must no longer claim the value is UTC
  expect_equal(datetime_format_of(path), "unlabelled")
})

test_that("dropping a time zone is daylight-saving correct", {
  # both instants read 12:00 locally despite different UTC offsets
  df <- data.frame(dt = as.POSIXct(c("2025-01-15 12:00:00", "2025-07-15 12:00:00"),
                                   tz = "America/New_York"))
  s <- sheet_serials(write_tmp(df))
  expect_equal(s[["A2"]], expected_serial("2025-01-15 12:00:00"))
  expect_equal(s[["A3"]], expected_serial("2025-07-15 12:00:00"))
})

test_that("an all-UTC workbook is unchanged in value", {
  df <- data.frame(dt = as.POSIXct("2020-06-01 12:00:00", tz = "UTC"))
  expect_equal(sheet_serials(write_tmp(df))[["A2"]],
               expected_serial("2020-06-01 12:00:00"))
})

test_that("differing time zones warn and are converted to UTC", {
  df <- data.frame(a = as.POSIXct("2025-12-01 01:00:00", tz = "Australia/Perth"))
  df$b <- as.POSIXct("2025-12-01 01:00:00", tz = "UTC")
  expect_warning(path <- write_tmp(df), "different time zones")
  expect_warning(write_tmp(df), "converted to UTC")
  s <- sheet_serials(path)
  expect_equal(s[["A2"]], expected_serial("2025-11-30 17:00:00"))  # Perth -> UTC
  expect_equal(s[["B2"]], expected_serial("2025-12-01 01:00:00"))
  # here the UTC label is accurate, so it is kept
  expect_equal(datetime_format_of(path), "utc-labelled")
})

test_that("the time zone survey includes POSIXct inside cell objects", {
  df <- data.frame(a = as.POSIXct("2025-12-01 01:00:00", tz = "Australia/Perth"))
  df$b <- xl_cell_general(value = as.POSIXct("2025-12-01 01:00:00", tz = "UTC"))
  expect_warning(write_tmp(df), "different time zones")
  # a cell object sharing the column's zone must NOT warn, and is shifted too
  df2 <- data.frame(a = as.POSIXct("2025-12-01 01:00:00", tz = "Australia/Perth"))
  df2$b <- xl_cell_general(value = as.POSIXct("2025-12-01 03:00:00", tz = "Australia/Perth"))
  expect_no_warning(path <- write_tmp(df2))
  s <- sheet_serials(path)
  expect_equal(s[["A2"]], expected_serial("2025-12-01 01:00:00"))
  expect_equal(s[["B2"]], expected_serial("2025-12-01 03:00:00"))
})

test_that("time zones are surveyed across every sheet, not per sheet", {
  s1 <- data.frame(a = as.POSIXct("2025-12-01 01:00:00", tz = "Australia/Perth"))
  s2 <- data.frame(a = as.POSIXct("2025-12-01 01:00:00", tz = "UTC"))
  expect_warning(write_tmp(list(one = s1, two = s2)), "different time zones")
})

test_that("a supplied datetime_format is never overridden", {
  df <- data.frame(dt = as.POSIXct("2025-12-01 01:00:00", tz = "Australia/Perth"))
  wb <- xl_workbook(df, properties = xl_properties(
    datetime_format = xl_num_format("yyyy/mm/dd hh:mm")))
  expect_match(xlsx_part(write_tmp(wb), "xl/styles.xml", raw = TRUE),
               "yyyy/mm/dd hh:mm", fixed = TRUE)
})

test_that("workbooks without POSIXct are untouched", {
  expect_no_warning(write_tmp(data.frame(x = 1:3, d = as.Date("2020-01-01") + 0:2)))
})

test_that("NA datetimes survive the time zone pass", {
  df <- data.frame(dt = as.POSIXct(c("2025-12-01 01:00:00", NA), tz = "Australia/Perth"))
  path <- write_tmp(df)
  s <- sheet_serials(path)
  expect_equal(s[["A2"]], expected_serial("2025-12-01 01:00:00"))
  expect_false("A3" %in% names(s))     # NA stays an empty cell
})

test_that("dropping a time zone preserves length and NA positions", {
  # A platform that cannot report a UTC offset must fall back to leaving the
  # value alone. Without a length check a zero-length offset would recycle to
  # nothing and silently empty the whole column, so pin the invariant.
  x <- as.POSIXct(c("2025-01-01 10:00:00", NA, "2025-07-01 10:00:00"),
                  tz = "America/New_York")
  y <- .drop_tzone(x)
  expect_length(y, length(x))
  expect_equal(is.na(y), is.na(x))
  expect_length(.drop_tzone(as.POSIXct(character(0), tz = "UTC")), 0L)
})

# ── Column types normalised on the way in ────────────────────────────────────

test_that("POSIXlt and integer-storage POSIXct become ordinary datetimes", {
  lt <- as.POSIXlt("2024-03-01 12:00:00", tz = "UTC")
  int <- as.POSIXct("2024-03-01 12:00:00", tz = "UTC")
  storage.mode(int) <- "integer"
  df <- data.frame(a = 1)
  df$lt <- lt
  df$int <- int
  p <- write_tmp(df)
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_s3_class(got$lt, "POSIXct")
  expect_equal(format(got$lt, "%Y-%m-%d %H:%M", tz = "UTC"), "2024-03-01 12:00")
  expect_equal(format(got$int, "%Y-%m-%d %H:%M", tz = "UTC"), "2024-03-01 12:00")
})

test_that("a datetime with no time zone falls back to the session's", {
  # attr(x, "tzone") unset means "local", which is what the writer has to use
  naive <- as.POSIXct("2024-03-01 12:00:00")
  attr(naive, "tzone") <- NULL
  df <- data.frame(a = 1)
  df$t <- naive
  p <- write_tmp(df)
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(format(got$t, "%Y-%m-%d %H:%M", tz = "UTC"), "2024-03-01 12:00")
})

test_that("a non-character column type is refused by name", {
  expect_error(xl_formula(1:3), "must be a character vector")
  expect_error(xl_hyperlink(1:3), "must be a character vector")
})

test_that("the row limit is refused before anything is written", {
  # one column of doubles: big enough to cross the limit, small enough to build
  big <- data.frame(a = numeric(1024L^2 + 1L))
  expect_error(write_tmp(big), "does not support tables with 1M+ rows",
               fixed = TRUE)
})

test_that("-Inf is written as text, like Inf", {
  p <- write_tmp(data.frame(a = c(-Inf, Inf, 1)))
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$a, c("-Inf", "Inf", "1"))
})

# Sheet names are repaired to Excel's rules rather than erroring; see
# R/sheet_names.R for the rules and the ordering.

# The sheet names actually written to a workbook.
written_names <- function(x) {
  p <- suppressWarnings(write_tmp(x))
  wb <- xlsx_part(p, "xl/workbook.xml", raw = TRUE)
  m <- regmatches(wb, gregexpr('<sheet name="[^"]*"', wb))[[1L]]
  sub('"$', "", sub('^<sheet name="', "", m))
}

one <- function(...) {
  l <- list(...)
  lapply(l, function(i) data.frame(a = 1))
}

# ── Length ────────────────────────────────────────────────────────────────────

test_that("an over-long name is truncated to exactly 31 characters", {
  nm <- strrep("x", 40)
  s <- setNames(list(data.frame(a = 1)), nm)
  expect_warning(write_tmp(s), "Truncating")
  # the old code truncated to 29 while claiming 31
  expect_equal(written_names(s), strrep("x", 31))
  expect_equal(nchar(written_names(s)), 31L)
})

test_that("a name of exactly 31 characters is left alone", {
  nm <- strrep("x", 31)
  s <- setNames(list(data.frame(a = 1)), nm)
  expect_silent(write_tmp(s))
  expect_equal(written_names(s), nm)
})

test_that("a name of 32 characters loses exactly one character", {
  nm <- strrep("x", 32)
  s <- setNames(list(data.frame(a = 1)), nm)
  expect_equal(written_names(s), strrep("x", 31))
})

# ── Invalid characters ────────────────────────────────────────────────────────

test_that("characters Excel forbids are replaced", {
  s <- setNames(list(data.frame(a = 1)), "2024/Q1")
  expect_warning(write_tmp(s), "Replacing characters")
  expect_equal(written_names(s), "2024-Q1")
})

test_that("every forbidden character is handled", {
  for (ch in c("[", "]", ":", "*", "?", "/", "\\")) {
    s <- setNames(list(data.frame(a = 1)), paste0("a", ch, "b"))
    expect_equal(written_names(s), "a-b", label = ch)
  }
})

test_that("a name of nothing but forbidden characters still writes", {
  s <- setNames(list(data.frame(a = 1)), "///")
  expect_equal(written_names(s), "---")
})

test_that("leading and trailing apostrophes are stripped", {
  expect_equal(written_names(setNames(list(data.frame(a = 1)), "'q1'")), "q1")
  expect_equal(written_names(setNames(list(data.frame(a = 1)), "'q1")), "q1")
  expect_equal(written_names(setNames(list(data.frame(a = 1)), "q1'")), "q1")
  # an interior apostrophe is legal and must survive
  expect_equal(written_names(setNames(list(data.frame(a = 1)), "q1's")), "q1's")
})

# ── Duplicates ────────────────────────────────────────────────────────────────

test_that("duplicate names are made unique", {
  s <- setNames(one(1, 2), c("S", "S"))
  expect_warning(write_tmp(s), "Deduplicating")
  expect_equal(written_names(s), c("S", "S_1"))
})

test_that("three-way duplicates are all made unique", {
  s <- setNames(one(1, 2, 3), c("S", "S", "S"))
  expect_equal(written_names(s), c("S", "S_1", "S_2"))
})

test_that("dedup keeps names within the 31-character maximum", {
  # the suffix must shorten the base, not push the name over the limit
  nm <- strrep("x", 31)
  s <- setNames(one(1, 2), c(nm, nm))
  got <- written_names(s)
  expect_equal(nchar(got), c(31L, 31L))
  expect_equal(got[2L], paste0(strrep("x", 29), "_1"))
  expect_false(anyDuplicated(got) > 0L)
})

test_that("collisions created by sanitizing are caught", {
  # "a/b" and "a-b" both sanitize to "a-b"; only a dedup pass that runs
  # *after* sanitizing sees that collision
  s <- setNames(one(1, 2), c("a/b", "a-b"))
  got <- written_names(s)
  expect_equal(got, c("a-b", "a-b_1"))
  expect_false(anyDuplicated(got) > 0L)
})

test_that("collisions created by truncating are caught", {
  a <- paste0(strrep("x", 31), "AAA")
  b <- paste0(strrep("x", 31), "BBB")
  got <- written_names(setNames(one(1, 2), c(a, b)))
  expect_equal(nchar(got), c(31L, 31L))
  expect_false(anyDuplicated(got) > 0L)
})

# ── Empty and absent names ────────────────────────────────────────────────────

test_that("unnamed input is left for libxlsxwriter to name", {
  # a bare data frame has no sheet name, and must keep getting the default
  expect_equal(written_names(data.frame(a = 1)), "Sheet1")
  expect_equal(written_names(one(1, 2)), c("Sheet1", "Sheet2"))
})

test_that("an empty or NA name is left for libxlsxwriter to name", {
  expect_equal(written_names(setNames(list(data.frame(a = 1)), "")), "Sheet1")
  expect_equal(written_names(setNames(list(data.frame(a = 1)), NA)), "Sheet1")
  # mixed: the named one keeps its name
  expect_equal(written_names(setNames(one(1, 2), c("", "Data"))),
               c("Sheet1", "Data"))
})

# ── Warning content ───────────────────────────────────────────────────────────

test_that("a rename warning names the original and the replacement", {
  s <- setNames(list(data.frame(a = 1)), "2024/Q1")
  expect_warning(write_tmp(s), '"2024/Q1" -> "2024-Q1"', fixed = TRUE)
})

test_that("many renames are summarised rather than listed in full", {
  nms <- paste0("a/", seq_len(8))
  s <- setNames(one(1, 2, 3, 4, 5, 6, 7, 8), nms)
  expect_warning(write_tmp(s), "and 3 more", fixed = TRUE)
})

# ── The C-side backstop ───────────────────────────────────────────────────────

test_that("a name the R side failed to repair is rejected in C", {
  # drive the C validation directly by bypassing .resolve_sheet_names(); this
  # is the guard against a future R-side regression, so it must actually fire
  local_mocked_bindings(.resolve_sheet_names = function(nms, n) rep("a/b", n))
  expect_error(write_tmp(setNames(list(data.frame(a = 1)), "ok")),
               "cannot contain invalid characters")
})

# What gets written where a value has none.
#
# writexl has always left such a cell blank.  `na` substitutes something else,
# at three scopes -- a cell, a column, the workbook -- with the innermost one
# that is set winning.  The default, `na = NA`, must keep the blank, and the
# first test proves that byte for byte rather than by inspection: this is the
# whole compatibility story for a package whose reverse dependencies all write
# data frames with NAs in them.

na_df <- function() data.frame(
  num  = c(1.5, NA, NaN),
  int  = c(1L, NA, 3L),
  chr  = c("a", NA, "c"),
  lgl  = c(TRUE, NA, FALSE),
  date = as.Date(c("2024-01-01", NA, "2024-03-01")),
  time = as.POSIXct(c("2024-01-01 10:00:00", NA, "2024-03-01 12:00:00"),
                    tz = "UTC"),
  stringsAsFactors = FALSE
)

sheet_xml_of <- function(...) {
  p <- tempfile(fileext = ".xlsx")
  write_xlsx(..., path = p)
  xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
}

# A string cell holds an index into sharedStrings.xml, so counting occurrences
# of a substitute means counting them there -- once per distinct string, which
# is why these tests count cells that point at it rather than the string.
shared_has <- function(p, s)
  grepl(s, xlsx_part(p, "xl/sharedStrings.xml", raw = TRUE), fixed = TRUE)

# ── The default changes nothing ───────────────────────────────────────────────

test_that("na = NA writes exactly what writing no `na` at all writes", {
  df <- na_df()
  # every scope explicitly unset, which is what the defaults already are
  explicit <- sheet_xml_of(list(D = xl_sheet(
    df, cols = xl_col_spec("num", na = NA))), na = NA)
  implicit <- sheet_xml_of(list(D = df))
  expect_equal(explicit, implicit)
  # and a blank cell is still no cell: the row holds only the non-NA columns
  expect_false(grepl("<c r=\"A3\"", implicit, fixed = TRUE))
})

test_that("the whole file is unchanged by the default, not just one sheet", {
  # xl_properties() gained an argument; the properties payload must not have
  # grown a key that reaches the file when nothing asked for one.  `created` is
  # pinned so the only thing that could differ is the change under test rather
  # than the timestamp libxlsxwriter stamps on each run.
  df <- na_df()
  when <- as.POSIXct("2024-01-01 00:00:00", tz = "UTC")
  bytes <- function(...) {
    p <- write_tmp(xl_workbook(list(D = df),
                               properties = xl_properties(created = when, ...)))
    readBin(p, "raw", file.size(p))
  }
  expect_equal(bytes(na = NA), bytes())
})

# ── Every column type ─────────────────────────────────────────────────────────

test_that("a workbook-wide na fills in for every column type, NaN included", {
  p <- write_tmp(na_df(), na = "none")
  got <- as.data.frame(readxl::read_xlsx(p))
  # the substitute makes each column mixed, so readxl reports character
  for (col in names(got))
    expect_equal(got[[col]][2L], "none", info = col)
  # NaN is as unwritable as NA, so it is filled in too
  expect_equal(got$num[3L], "none")
  # and the values that do exist are untouched
  expect_equal(got$chr[1L], "a")
  expect_equal(got$int[3L], "3")
})

test_that("na keeps its own type rather than becoming a string", {
  p <- write_tmp(data.frame(x = c(1.5, NA)), na = 0)
  got <- as.data.frame(readxl::read_xlsx(p))
  # still a numeric column, because the substitute is a number
  expect_type(got$x, "double")
  expect_equal(got$x, c(1.5, 0))

  # a logical substitute stays a boolean cell
  x <- sheet_xml_of(list(D = data.frame(a = c("k", NA), stringsAsFactors = FALSE)),
                    na = TRUE)
  expect_match(x, 't="b"', fixed = TRUE)
})

test_that("Inf is unaffected: it has a representation already", {
  p <- write_tmp(data.frame(x = c(Inf, -Inf, NA)), na = "none")
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$x, c("Inf", "-Inf", "none"))
})

# ── Scopes ────────────────────────────────────────────────────────────────────

test_that("a column's na overrides the workbook's", {
  df <- data.frame(a = c(1, NA), b = c(1, NA), c = c(1, NA))
  p <- write_tmp(xl_workbook(
    list(D = xl_sheet(df, cols = list(xl_col_spec("a", na = -999),
                                      xl_col_spec("b", na = "gone")))),
    properties = xl_properties(na = "WB")))
  got <- as.data.frame(readxl::read_xlsx(p))
  # a number keeps the column numeric; a string makes it mixed
  expect_equal(got$a[2L], -999)     # column wins
  expect_equal(got$b[2L], "gone")   # column wins
  expect_equal(got$c[2L], "WB")     # no column setting, so the workbook's
})

test_that("a cell's na overrides the column's and the workbook's", {
  df <- data.frame(a = 1:3)
  df$mix <- xl_cell_general(value = list(1, NA, NA), na = c(NA, "CELL", NA))
  p <- write_tmp(xl_workbook(
    list(D = xl_sheet(df, cols = xl_col_spec("mix", na = "COL"))),
    properties = xl_properties(na = "WB")))
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$mix, c("1", "CELL", "COL"))
})

test_that("a cell reaches the workbook's na with no column between", {
  df <- data.frame(a = 1:2)
  df$mix <- xl_cell_general(value = list(1, NA))
  p <- write_tmp(xl_workbook(list(D = xl_sheet(df)),
                             properties = xl_properties(na = "WB")))
  expect_equal(as.data.frame(readxl::read_xlsx(p))$mix, c("1", "WB"))
})

# ── What na must not touch ────────────────────────────────────────────────────

test_that("a cell that has content keeps it", {
  df <- data.frame(a = 1:3)
  # a formula and a hyperlink are content, not missing values
  df$f <- xl_cell_general(formula = c("=A2+1", NA, NA))
  df$h <- xl_cell_general(hyperlink = c("https://example.com", NA, NA))
  p <- write_tmp(xl_workbook(list(D = xl_sheet(df)),
                             properties = xl_properties(na = "none")))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  # the formula cell still holds its formula
  expect_match(x, "A2+1", fixed = TRUE)
  expect_true(shared_has(p, "none"))
  # the four cells that have neither formula nor value take the substitute,
  # and the two that have content do not
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$f[2:3], c("none", "none"))
  expect_equal(got$h[2:3], c("none", "none"))
  expect_false(identical(got$f[1L], "none"))
})

test_that("the header row is never substituted", {
  # a header is always written, so it has no missing value to stand in for --
  # this pins that `na` never reaches it
  df <- data.frame(none = c(1, NA))
  p <- write_tmp(list(D = df), na = "none")
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(names(got), "none")
  expect_equal(got$none, c("1", "none"))
})

test_that("a formatted blank keeps its format when substituted", {
  df <- data.frame(a = 1:2)
  df$b <- xl_cell_general(value = list(1, NA), format = xl_font(bold = TRUE),
                          na = "none")
  p <- write_tmp(list(D = xl_sheet(df)))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_true(shared_has(p, "none"))
  # the substituted cell carries the same style as the value cell above it,
  # which is the format the blank it replaces would have kept
  styles <- regmatches(x, gregexpr('<c r="B[23]" s="[0-9]+"', x))[[1L]]
  expect_length(styles, 2L)
  expect_equal(sub('.*s="', "", styles[1L]), sub('.*s="', "", styles[2L]))
  expect_match(xlsx_part(p, "xl/styles.xml", raw = TRUE), "<b/>", fixed = TRUE)
})

# ── Validation ────────────────────────────────────────────────────────────────

test_that("na must be a single value of some atomic type", {
  expect_error(xl_properties(na = c("a", "b")), "must be a single value")
  expect_error(xl_properties(na = list(1)), "must be a single value")
  expect_error(xl_col_spec("a", na = mean), "must be a single value")
  expect_error(xl_cell_general(value = 1, na = list(1, 2)),
               "must be length 1 or 1")
  expect_error(xl_cell_general(value = 1:3, na = c("a", "b")),
               "must be length 1 or 3")
  # a factor is taken as its label
  expect_equal(unclass(xl_properties(na = factor("gone")))$na, "gone")
  # NaN cannot stand in for a missing value either, so it reads as unset
  expect_null(unclass(xl_properties(na = NaN))$na)
  expect_null(unclass(xl_properties(na = NA))$na)
})

test_that("a per-cell na may leave some cells to inherit", {
  # NULL and NA both mean "nothing set here", so a list can mix cells that
  # override with cells that fall through to the column or the workbook
  df <- data.frame(a = 1:3)
  df$mix <- xl_cell_general(value = list(NA, NA, NA),
                            na = list("CELL", NULL, NA))
  p <- write_tmp(xl_workbook(list(D = xl_sheet(df)),
                             properties = xl_properties(na = "WB")))
  expect_equal(as.data.frame(readxl::read_xlsx(p))$mix,
               c("CELL", "WB", "WB"))
})

test_that("write_xlsx's na shorthand refuses to be silently ignored", {
  wb <- xl_workbook(list(D = data.frame(a = c(1, NA))))
  expect_error(write_tmp(wb, na = "none"), "already an xl_workbook")
  # without the shorthand the workbook's own setting applies
  expect_silent(write_tmp(wb))
})

# =============================================================================
# Chart axes
# =============================================================================
#
# libxlsxwriter offers 32 chart_axis_*() functions.  Most take one value, and
# several of them are only meaningful on one *kind* of axis -- the header
# documents that per function ("Axis types: value axes only") and then does not
# enforce it: setting a minimum on a category axis is accepted, written, and
# ignored by Excel.
#
# There are two kinds of axis in reach.  libxlsxwriter also mentions date axes,
# but nothing in its API turns an axis into one, so every axis writexl can
# produce is either a category axis or a value axis:
#
#   * a scatter chart plots numbers against numbers, so both of its axes are
#     value axes;
#   * every other type plots values against labels, so x is the category axis
#     and y the value axis.  This holds for bar charts too, which Excel draws
#     with the categories running up the side -- the drawing is rotated, the
#     axes are not renamed.
#
# .check_axis_option() refuses an option the axis cannot use, naming the axis
# that can.
# -----------------------------------------------------------------------------

.LXW_AXIS_POSITION <- c(on_tick = 1L, between = 2L)

.LXW_AXIS_LABEL_POSITION <- c(next_to = 0L, high = 1L, low = 2L, none = 3L)

.LXW_AXIS_LABEL_ALIGN <- c(center = 0L, left = 1L, right = 2L)

.LXW_AXIS_TICK_MARK <- c(default = 0L, none = 1L, inside = 2L, outside = 3L,
                         crossing = 4L)

# The names are Excel's, not the enum's: "ten_thousands" rather than
# TEN_THOUSANDS, and the order is the enum's own.
.LXW_AXIS_UNITS <- c(
  none = 0L, hundreds = 1L, thousands = 2L, ten_thousands = 3L,
  hundred_thousands = 4L, millions = 5L, ten_millions = 6L,
  hundred_millions = 7L, billions = 8L, trillions = 9L
)

# Options only one kind of axis can use, and which kind that is.  Taken from
# the "Axis types" line of each function's documentation in chart.h; everything
# not listed applies to both.
.AXIS_OPTION_KIND <- c(
  position      = "category",   # chart_axis_set_position()
  label_align   = "category",   # chart_axis_set_label_align()
  interval_unit = "category",   # chart_axis_set_interval_unit()
  interval_tick = "category",   # chart_axis_set_interval_tick()
  min           = "value",      # chart_axis_set_min()
  max           = "value",      # chart_axis_set_max()
  log_base      = "value",      # chart_axis_set_log_base()
  major_unit    = "value",      # chart_axis_set_major_unit()
  minor_unit    = "value",      # chart_axis_set_minor_unit()
  display_units = "value",      # chart_axis_set_display_units()
  display_units_visible = "value"
)

# Which kind an axis is, given the chart type it belongs to.
.axis_kind <- function(type, which) {
  if (identical(.CHART_FAMILY(type), "scatter")) "value"
  else if (identical(which, "x")) "category" else "value"
}

# `opt` is one of .AXIS_OPTION_KIND's names: .check_axis() intersects with them
# before calling, and a character vector's [[ ]] errors on anything else.
.check_axis_option <- function(opt, type, which) {
  kind <- .AXIS_OPTION_KIND[[opt]]
  have <- .axis_kind(type, which)
  if (identical(kind, have)) return(invisible(NULL))
  other <- if (identical(which, "x")) "y_axis" else "x_axis"
  stop(sprintf(paste0("`%s_axis$%s` applies to a %s axis, and the %s axis of a ",
                      "\"%s\" chart is a %s axis; Excel ignores it there.\n  ",
                      "%s"),
               which, opt, kind, which, type, have,
               if (identical(.CHART_FAMILY(type), "scatter"))
                 "Both axes of a scatter chart are value axes."
               else sprintf("Use `%s`, or a chart type whose %s axis is a %s axis.",
                            other, which, kind)),
       call. = FALSE)
}

#' An axis of a chart
#'
#' @description
#' `xl_chart_axis()` describes one axis, and is given to [xl_chart()] as
#' `x_axis` or `y_axis`.
#'
#' Several options apply to one kind of axis only, and Excel discards the rest
#' without a word, so writexl refuses them instead.  A scatter chart plots
#' numbers against numbers, so both of its axes are **value** axes; every other
#' type has a **category** x axis and a value y axis.  (A bar chart is drawn
#' with its categories up the side, but the axes keep their names.)  Pie and
#' doughnut charts have no axes at all.
#'
#' * value axes only --- `min`, `max`, `log_base`, `major_unit`, `minor_unit`,
#'   `display_units`, `display_units_visible`;
#' * category axes only --- `position`, `label_align`, `interval_unit`,
#'   `interval_tick`.
#'
#' @param title The axis title: a string, or a range spec holding one --- see
#'   [xl_chart_series()] for the spellings, including
#'   `list(header = "revenue")`.
#' @param title_format An [xl_format()] styling the axis title.  A title is
#'   text, so only the [xl_font()] group applies.
#' @param title_layout Where to put the axis title by hand, as `c(x, y)`
#'   fractions of the chart, each above 0 and at most 1.  Excel places it for
#'   you otherwise.
#' @param label_format An [xl_format()] styling the tick labels --- the
#'   [xl_font()] group only.
#' @param num_format A number format for the tick labels, as an Excel format
#'   string (`"#,##0"`) or an [xl_num_format()].
#' @param line_format An [xl_format()] styling the axis line itself:
#'   [xl_border()] for the line, [xl_fill()] for the fill behind it.  An axis
#'   has four parts that can be styled, so none of them is just `format`.
#' @param visible `FALSE` hides the axis.
#' @param reverse Draw the axis in the opposite direction.
#' @param min,max The axis bounds.  Value axes only.
#' @param log_base Use a logarithmic scale with this base, 2 or more.  Value
#'   axes only.
#' @param major_unit,minor_unit The spacing between major and minor tick marks.
#'   Value axes only.
#' @param display_units Scale the labels by `"thousands"`, `"millions"`,
#'   `"billions"` and so on; see Details for the full set.  Value axes only.
#' @param display_units_visible Whether the caption naming the units --- the
#'   small rotated "Millions" beside the axis --- is drawn.  Setting
#'   `display_units` turns it **on**, as Excel does, so this is really for
#'   `FALSE`: rescaled labels with no caption.  Value axes only.
#' @param interval_unit Label one category in every `n`.  Category axes only.
#' @param interval_tick Put a tick mark on one category in every `n`.
#'   Category axes only.
#' @param position Whether the data sits `"on_tick"` or `"between"` the tick
#'   marks.  Category axes only.
#' @param label_position Where the tick labels go: `"next_to"`, `"high"`,
#'   `"low"`, or `"none"` for no labels.
#' @param label_align Tick-label alignment: `"center"`, `"left"` or `"right"`.
#'   Category axes only.
#' @param major_tick,minor_tick The tick marks: `"default"`, `"none"`,
#'   `"inside"`, `"outside"` or `"crossing"`.
#' @param crossing Where the other axis crosses this one: a number, or `"min"`
#'   or `"max"` for either end.
#' @param major_gridlines,minor_gridlines Show the gridlines.  A chart's major
#'   y gridlines are on by default and everything else is off.
#' @param major_gridlines_format,minor_gridlines_format An [xl_format()]
#'   styling the gridlines --- [xl_border()] only, since a gridline is a line.
#' @return An `xl_chart_axis` object.
#' @details
#' `display_units` is one of `"none"`, `"hundreds"`, `"thousands"`,
#' `"ten_thousands"`, `"hundred_thousands"`, `"millions"`, `"ten_millions"`,
#' `"hundred_millions"`, `"billions"` or `"trillions"`.
#' @family images and charts
#' @seealso [xl_chart], [xl_chart_series]
#' @export
#' @examples
#' xl_chart_axis(title = "Quarter")
#' xl_chart_axis(title = "Revenue", min = 0, num_format = "$#,##0",
#'               major_gridlines = FALSE)
xl_chart_axis <- function(title = NULL, title_format = NULL, title_layout = NULL,
                          label_format = NULL, num_format = NULL,
                          line_format = NULL, visible = NA, reverse = NA,
                          min = NA, max = NA, log_base = NA,
                          major_unit = NA, minor_unit = NA,
                          display_units = NULL, display_units_visible = NA,
                          interval_unit = NA, interval_tick = NA,
                          position = NULL, label_position = NULL,
                          label_align = NULL,
                          major_tick = NULL, minor_tick = NULL,
                          crossing = NULL,
                          major_gridlines = NA, minor_gridlines = NA,
                          major_gridlines_format = NULL,
                          minor_gridlines_format = NULL) {
  nm <- if (is.null(title)) NULL
        else if (is.character(title) && length(title) == 1L && !is.na(title))
          list(text = title)
        else .chart_range(title, "title")
  if (!is.null(title_format) && is.null(nm))
    stop(paste0("`title_format` styles the axis title, so give a `title` too. ",
                "An axis with no title has nothing to style."), call. = FALSE)

  # chart_axis_set_num_format() takes the format string itself rather than an
  # xl_format; the constructor is accepted too so either spelling works.
  nf <- num_format
  if (is_xl_format(nf)) {
    grp <- unclass(nf)
    nf <- grp$num_format$format
    if (is.null(nf))
      stop("`num_format` is a format with no number format in it",
           call. = FALSE)
  }
  if (!is.null(nf) && (!is.character(nf) || length(nf) != 1L || is.na(nf)))
    stop("`num_format` must be an Excel format string, e.g. \"#,##0.0\"",
         call. = FALSE)

  cr <- crossing
  if (!is.null(cr)) {
    if (is.character(cr)) {
      if (!identical(cr, "min") && !identical(cr, "max"))
        stop("`crossing` must be a number, \"min\" or \"max\"", call. = FALSE)
    } else if (!is.numeric(cr) || length(cr) != 1L || is.na(cr))
      stop("`crossing` must be a number, \"min\" or \"max\"", call. = FALSE)
  }

  at <- title_layout
  if (!is.null(at)) {
    if (!is.numeric(at) || length(at) != 2L || anyNA(at) ||
        any(at <= 0) || any(at > 1))
      stop(paste0("`title_layout` must be c(x, y), each above 0 and at most 1 ",
                  "-- they are fractions of the chart's width and height"),
           call. = FALSE)
    at <- as.numeric(at)
  }

  lb <- .val_int(log_base, "log_base", min = 2, max = 1000)

  structure(.drop_null(list(
    title = nm,
    title_format = .chart_format_payload(title_format, "title_format",
                                        accept = "font",
                                        part = "an axis title"),
    title_layout = at,
    label_format = .chart_format_payload(label_format, "label_format",
                                         accept = "font",
                                         part = "an axis tick label"),
    num_format = nf,
    line_format = .chart_format_payload(line_format, "line_format",
                                        accept = c("line", "fill", "pattern"),
                                        part = "an axis"),
    visible = .val_flag(visible, "visible"),
    reverse = .val_flag(reverse, "reverse"),
    min = .val_num(min, "min"), max = .val_num(max, "max"),
    log_base = lb,
    major_unit = .val_num(major_unit, "major_unit", min = 0),
    minor_unit = .val_num(minor_unit, "minor_unit", min = 0),
    display_units = .val_enum(display_units, names(.LXW_AXIS_UNITS),
                              "display_units"),
    display_units_visible = .val_flag(display_units_visible,
                                      "display_units_visible"),
    interval_unit = .val_int(interval_unit, "interval_unit", min = 1),
    interval_tick = .val_int(interval_tick, "interval_tick", min = 1),
    position = .val_enum(position, names(.LXW_AXIS_POSITION), "position"),
    label_position = .val_enum(label_position,
                               names(.LXW_AXIS_LABEL_POSITION),
                               "label_position"),
    label_align = .val_enum(label_align, names(.LXW_AXIS_LABEL_ALIGN),
                            "label_align"),
    major_tick = .val_enum(major_tick, names(.LXW_AXIS_TICK_MARK),
                           "major_tick"),
    minor_tick = .val_enum(minor_tick, names(.LXW_AXIS_TICK_MARK),
                           "minor_tick"),
    crossing = cr,
    major_gridlines = .val_flag(major_gridlines, "major_gridlines"),
    minor_gridlines = .val_flag(minor_gridlines, "minor_gridlines"),
    major_gridlines_format = .chart_format_payload(
      major_gridlines_format, "major_gridlines_format", accept = "line",
      part = "a gridline"),
    minor_gridlines_format = .chart_format_payload(
      minor_gridlines_format, "minor_gridlines_format", accept = "line",
      part = "a gridline")
  )), class = "xl_chart_axis")
}

#' @export
print.xl_chart_axis <- function(x, ...) {
  p <- unclass(x)
  cat("<xl_chart_axis>",
      if (!is.null(p[["name"]][["text"]]))
        paste0(" \"", p[["name"]][["text"]], "\"") else "",
      "\n", sep = "")
  set <- setdiff(names(p), "name")
  if (length(set)) cat("  set:", paste(set, collapse = ", "), "\n")
  invisible(x)
}

# --- Resolution --------------------------------------------------------------

# Everything about an axis that the chart's type decides.  Checked here so it
# is refused where the caller wrote it rather than at write time.
.check_axis <- function(ax, which, type, arg) {
  if (is.null(ax)) return(invisible(NULL))
  if (!inherits(ax, "xl_chart_axis"))
    stop(sprintf("`%s` must be an xl_chart_axis() object", arg), call. = FALSE)
  # pie and doughnut charts have no axes to put this on
  .check_chart_feature(type, "axes", arg)
  for (opt in intersect(names(unclass(ax)), names(.AXIS_OPTION_KIND)))
    .check_axis_option(opt, type, which)
  invisible(NULL)
}

# One axis -> the payload C reads.  Anything type-dependent has already been
# checked by .check_axis().
.axis_payload <- function(ax, arg, sheets, own, header_offset) {
  if (is.null(ax)) return(NULL)
  p <- unclass(ax)

  ent <- list()
  nm <- p[["title"]]
  if (!is.null(nm)) {
    if (!is.null(nm[["text"]])) ent$name <- nm[["text"]]
    else {
      r <- .resolve_chart_range(nm, sprintf("%s title", arg), sheets, own,
                                header_offset)
      ent$name_sheet <- r$sheet
      ent$name_range <- r$range
    }
  }
  # the payload itself, not its $font: chart_font_of() looks the font up by
  # key, the same way it reads a series format
  ent$name_font  <- p[["title_format"]]
  ent$num_font   <- p[["label_format"]]
  ent$num_format <- p[["num_format"]]
  if (!is.null(p[["title_layout"]])) {
    ent$name_x <- p[["title_layout"]][[1L]]
    ent$name_y <- p[["title_layout"]][[2L]]
  }
  ent$line    <- p[["line_format"]][["line"]]
  ent$fill    <- p[["line_format"]][["fill"]]
  ent$pattern <- p[["line_format"]][["pattern"]]

  if (identical(p[["visible"]], FALSE)) ent$off <- 1L
  if (isTRUE(p[["reverse"]]))           ent$reverse <- 1L
  ent$min <- p[["min"]]
  ent$max <- p[["max"]]
  ent$log_base <- p[["log_base"]]
  ent$major_unit <- p[["major_unit"]]
  ent$minor_unit <- p[["minor_unit"]]
  ent$interval_unit <- p[["interval_unit"]]
  ent$interval_tick <- p[["interval_tick"]]
  if (!is.null(p[["display_units"]]))
    ent$display_units <- unname(.LXW_AXIS_UNITS[[p[["display_units"]]]])
  if (!is.null(p[["display_units_visible"]]))
    ent$display_units_visible <- as.integer(p[["display_units_visible"]])
  if (!is.null(p[["position"]]))
    ent$position <- unname(.LXW_AXIS_POSITION[[p[["position"]]]])
  if (!is.null(p[["label_position"]]))
    ent$label_position <-
      unname(.LXW_AXIS_LABEL_POSITION[[p[["label_position"]]]])
  if (!is.null(p[["label_align"]]))
    ent$label_align <- unname(.LXW_AXIS_LABEL_ALIGN[[p[["label_align"]]]])
  if (!is.null(p[["major_tick"]]))
    ent$major_tick <- unname(.LXW_AXIS_TICK_MARK[[p[["major_tick"]]]])
  if (!is.null(p[["minor_tick"]]))
    ent$minor_tick <- unname(.LXW_AXIS_TICK_MARK[[p[["minor_tick"]]]])

  cr <- p[["crossing"]]
  if (!is.null(cr)) {
    if (identical(cr, "min")) ent$crossing_min <- 1L
    else if (identical(cr, "max")) ent$crossing_max <- 1L
    else ent$crossing <- as.numeric(cr)
  }

  if (!is.null(p[["major_gridlines"]]))
    ent$major_gridlines <- as.integer(p[["major_gridlines"]])
  if (!is.null(p[["minor_gridlines"]]))
    ent$minor_gridlines <- as.integer(p[["minor_gridlines"]])
  # a gridline styled but never switched on would be invisible
  for (g in c("major", "minor")) {
    ln <- p[[paste0(g, "_gridlines_format")]][["line"]]
    if (is.null(ln)) next
    ent[[paste0(g, "_gridlines_line")]] <- ln
    if (is.null(ent[[paste0(g, "_gridlines")]]))
      ent[[paste0(g, "_gridlines")]] <- 1L
  }

  .drop_null(ent)
}

# Chart axes.  Two things are being pinned here: that each of libxlsxwriter's
# 32 chart_axis_*() functions is reachable and reaches the file, and that the
# options only one *kind* of axis can use are refused on the other kind rather
# than written and ignored by Excel.

axis_sales <- data.frame(quarter = c("Q1", "Q2", "Q3", "Q4"),
                         revenue = c(10, 20, 30, 40),
                         stringsAsFactors = FALSE)

axis_ser <- function(...)
  xl_chart_series(values = list(cols = "revenue"),
                  categories = list(cols = "quarter"), ...)

# The chart XML for one chart over axis_sales.
axis_xml <- function(...) {
  t <- tempfile(fileext = ".xlsx")
  write_xlsx(list(Data = xl_sheet(axis_sales,
                                  chart = xl_chart("column", axis_ser(), ...))),
             t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  f <- list.files(file.path(d, "xl/charts"), pattern = "^chart",
                  full.names = TRUE)
  paste(readLines(f[1L], warn = FALSE), collapse = "")
}

# ── Which kind of axis is which ───────────────────────────────────────────────

test_that("a scatter chart has two value axes and everything else one", {
  # this is the whole basis of the option check, so it is pinned directly
  for (ty in names(.LXW_CHART_TYPE)) {
    if (identical(.CHART_FAMILY(ty), "pie")) next    # no axes at all
    expect_equal(.axis_kind(ty, "y"), "value", info = ty)
    expect_equal(.axis_kind(ty, "x"),
                 if (identical(.CHART_FAMILY(ty), "scatter")) "value"
                 else "category", info = ty)
  }
  # a bar chart draws its categories up the side, but they are still the x axis
  expect_equal(.axis_kind("bar", "x"), "category")
  expect_equal(.axis_kind("bar_stacked", "x"), "category")
})

test_that("a value-only option is refused on a category axis, and vice versa", {
  for (opt in names(.AXIS_OPTION_KIND)) {
    kind <- .AXIS_OPTION_KIND[[opt]]
    # a column chart has a category x and a value y, so one of the two is wrong
    wrong <- if (identical(kind, "value")) "x_axis" else "y_axis"
    right <- if (identical(kind, "value")) "y_axis" else "x_axis"
    val <- switch(opt, position = "on_tick", label_align = "left",
                  display_units = "thousands",
                  display_units_visible = TRUE, log_base = 10, 2)
    ax <- do.call(xl_chart_axis, stats::setNames(list(val), opt))
    args <- list("column", xl_chart_series(values = "B1:B5"))
    args[[wrong]] <- ax
    expect_error(do.call(xl_chart, args),
                 sprintf("$%s` applies to a %s axis", opt, kind),
                 fixed = TRUE, label = opt)
    args <- list("column", xl_chart_series(values = "B1:B5"))
    args[[right]] <- ax
    expect_true(inherits(do.call(xl_chart, args), "xl_chart"), info = opt)
  }
})

test_that("both axes of a scatter take the value-axis options", {
  se <- xl_chart_series(values = "B1:B5", categories = "A1:A5")
  expect_s3_class(xl_chart("scatter", se, x_axis = xl_chart_axis(min = 0),
                           y_axis = xl_chart_axis(log_base = 10)), "xl_chart")
  # and neither takes the category-axis ones, with a message that says why
  expect_error(xl_chart("scatter", se, x_axis = xl_chart_axis(interval_unit = 2)),
               "Both axes of a scatter chart are value axes", fixed = TRUE)
})

test_that("pie and doughnut charts have no axes to describe", {
  for (ty in c("pie", "doughnut"))
    expect_error(xl_chart(ty, xl_chart_series(values = "A1:A5"),
                          y_axis = xl_chart_axis(title = "n")),
                 "does not apply to a", label = ty)
})

test_that("the option-kind table matches the documented axis types", {
  # the table was read out of chart.h's "Axis types" lines; if a name here
  # stops being an xl_chart_axis() argument the table has drifted
  expect_true(all(names(.AXIS_OPTION_KIND) %in% names(formals(xl_chart_axis))))
  expect_setequal(unique(unname(.AXIS_OPTION_KIND)), c("value", "category"))
})

# ── The constructor ───────────────────────────────────────────────────────────

test_that("the enums cover libxlsxwriter's, in its order", {
  expect_equal(.LXW_AXIS_POSITION, c(on_tick = 1L, between = 2L))
  expect_equal(unname(.LXW_AXIS_LABEL_POSITION), 0:3)
  expect_equal(unname(.LXW_AXIS_LABEL_ALIGN), 0:2)
  expect_equal(unname(.LXW_AXIS_TICK_MARK), 0:4)
  expect_equal(unname(.LXW_AXIS_UNITS), 0:9)
  expect_true("crossing" %in% names(.LXW_AXIS_TICK_MARK))
})

test_that("axis arguments are validated", {
  expect_error(xl_chart_axis(position = "middle"), "`position` must be one of")
  expect_error(xl_chart_axis(major_tick = "big"), "`major_tick` must be one of")
  expect_error(xl_chart_axis(display_units = "squillions"),
               "`display_units` must be one of")
  expect_error(xl_chart_axis(log_base = 1), "must be between 2 and 1000")
  expect_error(xl_chart_axis(interval_unit = 0), "must be between 1 and NA")
  expect_error(xl_chart_axis(crossing = "middle"),
               "must be a number, \"min\" or \"max\"")
  expect_error(xl_chart_axis(crossing = c(1, 2)),
               "must be a number, \"min\" or \"max\"")
  expect_error(xl_chart_axis(num_format = 42), "must be an Excel format string")
  expect_error(xl_chart_axis(title_layout = c(0, 0.5)),
               "each above 0 and at most 1")
  expect_error(xl_chart_axis(title_layout = 0.5), "must be c\\(x, y\\)")
  expect_error(xl_chart("column", xl_chart_series(values = "A1"),
                        x_axis = "bottom"),
               "must be an xl_chart_axis")
})

test_that("a number format may be given either way", {
  expect_equal(unclass(xl_chart_axis(num_format = "#,##0"))$num_format, "#,##0")
  expect_equal(unclass(xl_chart_axis(num_format = xl_num_format("0.0%")))$num_format,
               "0.0%")
  # an xl_format that carries no number format is a mistake, not an empty one
  expect_error(xl_chart_axis(num_format = xl_font(bold = TRUE)),
               "no number format in it")
})

test_that("each part of an axis takes only the format groups it can use", {
  expect_error(xl_chart_axis(title = "n", title_format = xl_fill(background = "red")),
               "takes no fill")
  expect_error(xl_chart_axis(label_format = xl_border(all = "thin")),
               "takes no line")
  expect_error(xl_chart_axis(major_gridlines_format = xl_fill(background = "red")),
               "takes no fill")
  # the axis line itself is a shape, so it takes no font
  expect_error(xl_chart_axis(line_format = xl_font(bold = TRUE)),
               "a shape and has no text")
})

test_that("a title format needs a title to style", {
  expect_error(xl_chart_axis(title_format = xl_font(bold = TRUE)),
               "give a `title` too", fixed = TRUE)
  expect_s3_class(xl_chart_axis(title = "n", title_format = xl_font(bold = TRUE)),
                  "xl_chart_axis")
})

test_that("styling a gridline turns it on", {
  # a styled gridline that stays hidden is not something anyone means
  p <- .axis_payload(xl_chart_axis(major_gridlines_format = xl_border(all = "thin")),
                     "y_axis", list(Data = axis_sales), "Data", 1L)
  expect_equal(p$major_gridlines, 1L)
  expect_false(is.null(p$major_gridlines_line))
  # unless it is switched off explicitly, which stays off
  q <- .axis_payload(xl_chart_axis(major_gridlines = FALSE,
                                   major_gridlines_format = xl_border(all = "thin")),
                     "y_axis", list(Data = axis_sales), "Data", 1L)
  expect_equal(q$major_gridlines, 0L)
})

# ── What reaches the file ─────────────────────────────────────────────────────

test_that("every axis option reaches the chart XML", {
  x <- axis_xml(
    x_axis = xl_chart_axis(title = "Quarter", position = "on_tick",
                           label_align = "left", interval_unit = 2,
                           interval_tick = 3, major_tick = "inside",
                           minor_tick = "crossing", label_position = "low",
                           title_format = xl_font(size = 12, bold = TRUE),
                           label_format = xl_font(italic = TRUE),
                           line_format = xl_border(all = "dashed",
                                                   color = "red")),
    y_axis = xl_chart_axis(title = list(header = "revenue"), min = 5, max = 45,
                           major_unit = 10, minor_unit = 2.5,
                           num_format = "$#,##0.0", reverse = TRUE,
                           crossing = "max", display_units = "thousands",
                           display_units_visible = TRUE,
                           major_gridlines = TRUE, minor_gridlines = TRUE,
                           minor_gridlines_format = xl_border(all = "dotted",
                                                              color = "blue"),
                           title_layout = c(0.02, 0.4)))
  want <- c(
    "Quarter",                              # name
    "<c:f>Data!$B$1</c:f>",                 # name from a header cell
    '<c:tickLblPos val="low"/>',            # label_position
    '<c:lblAlgn val="l"/>',                 # label_align
    '<c:tickLblSkip val="2"/>',             # interval_unit
    '<c:tickMarkSkip val="3"/>',            # interval_tick
    '<c:majorTickMark val="in"/>',          # major_tick
    '<c:minorTickMark val="cross"/>',       # minor_tick
    '<a:prstDash val="dash"/>',             # the axis line
    '<c:min val="5"/>', '<c:max val="45"/>',
    '<c:majorUnit val="10"/>', '<c:minorUnit val="2.5"/>',
    'formatCode="$#,##0.0"',                # num_format
    '<c:orientation val="maxMin"/>',        # reverse
    '<c:crosses val="max"/>',               # crossing
    '<c:builtInUnit val="thousands"/>',     # display_units
    "<c:dispUnitsLbl>",                     # display_units_visible
    "<c:majorGridlines/>", "<c:minorGridlines>",
    '<a:prstDash val="dot"/>',              # the minor gridline's own line
    "0000FF",                               # and its colour
    "<c:manualLayout>",                     # title_layout
    'sz="1200"', 'i="1"'                    # the two fonts
  )
  for (w in want) expect_true(grepl(w, x, fixed = TRUE), info = w)
  # position = "on_tick" is written as the absence of crossBetween="between"
  expect_match(x, '<c:crossBetween val="midCat"/>', fixed = TRUE)
})

test_that("an axis can be hidden, and its gridlines and labels turned off", {
  x <- axis_xml(x_axis = xl_chart_axis(visible = FALSE),
                y_axis = xl_chart_axis(major_gridlines = FALSE,
                                       label_position = "none"))
  expect_match(x, '<c:delete val="1"/>', fixed = TRUE)
  expect_false(grepl("<c:majorGridlines", x, fixed = TRUE))
  expect_match(x, '<c:tickLblPos val="none"/>', fixed = TRUE)
})

test_that("a log scale reaches both axes of a scatter chart", {
  t <- tempfile(fileext = ".xlsx")
  write_xlsx(list(Data = xl_sheet(axis_sales, chart = xl_chart("scatter",
    xl_chart_series(values = list(cols = "revenue"),
                    categories = list(cols = "revenue")),
    x_axis = xl_chart_axis(log_base = 10),
    y_axis = xl_chart_axis(log_base = 2)))), t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  x <- paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                                  full.names = TRUE)[1L], warn = FALSE),
             collapse = "")
  expect_match(x, '<c:logBase val="10"/>', fixed = TRUE)
  expect_match(x, '<c:logBase val="2"/>', fixed = TRUE)
})

test_that("an axis name may come from another sheet", {
  t <- tempfile(fileext = ".xlsx")
  write_xlsx(list(
    Chart = xl_sheet(data.frame(z = 1),
                     chart = xl_chart("column",
                       xl_chart_series(values = list(sheet = "Data",
                                                     cols = "revenue")),
                       y_axis = xl_chart_axis(title = list(sheet = "Data",
                                                          header = "revenue")))),
    Data = axis_sales), t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  x <- paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                                  full.names = TRUE)[1L], warn = FALSE),
             collapse = "")
  expect_match(x, "<c:f>Data!$B$1</c:f>", fixed = TRUE)
})

# ── Coverage of the C API ─────────────────────────────────────────────────────

test_that("every chart_axis_*() function libxlsxwriter offers is called", {
  # A mechanical gate rather than a promise in prose: a function added
  # upstream, or one dropped from apply_axis(), fails here.  Skipped where the
  # sources are not alongside the tests, as in an installed package.
  hdr <- NULL
  for (p in c("../../src/libxlsxwriter/include/xlsxwriter/chart.h",
              "../../../src/libxlsxwriter/include/xlsxwriter/chart.h"))
    if (file.exists(p)) hdr <- p
  skip_if(is.null(hdr), "libxlsxwriter sources not alongside the tests")
  src <- NULL
  for (p in c("../../src/write_xlsx.c", "../../../src/write_xlsx.c"))
    if (file.exists(p)) src <- p
  skip_if(is.null(src), "write_xlsx.c not alongside the tests")

  decl <- grep("^(void|lxw_chart_axis \\*|uint8_t) ?chart_axis",
               readLines(hdr), value = TRUE)
  fns <- unique(regmatches(decl, regexpr("chart_axis_[a-z_]+", decl)))
  # chart_axis_get() fetches an axis by enum; writexl reaches the axes through
  # the chart struct's own x_axis/y_axis members instead.
  fns <- setdiff(fns, "chart_axis_get")
  expect_equal(length(fns), 32L)

  called <- paste(readLines(src), collapse = "\n")
  for (f in fns)
    expect_true(grepl(paste0(f, "("), called, fixed = TRUE), info = f)
})

test_that("the display-units caption is on unless it is turned off", {
  # chart_axis_set_display_units() sets display_units_visible itself, so the
  # caption comes with the rescaling and `TRUE` adds nothing.  Pinned because
  # the argument reads as opt-in, and because a change upstream would silently
  # alter every such chart.
  on <- axis_xml(y_axis = xl_chart_axis(display_units = "thousands"))
  expect_match(on, "<c:dispUnitsLbl>", fixed = TRUE)
  same <- axis_xml(y_axis = xl_chart_axis(display_units = "thousands",
                                          display_units_visible = TRUE))
  expect_match(same, "<c:dispUnitsLbl>", fixed = TRUE)
  off <- axis_xml(y_axis = xl_chart_axis(display_units = "thousands",
                                         display_units_visible = FALSE))
  expect_match(off, '<c:builtInUnit val="thousands"/>', fixed = TRUE)
  expect_false(grepl("<c:dispUnitsLbl>", off, fixed = TRUE))
})

test_that("only the options in the kind table are checked against the axis", {
  # .check_axis() intersects with .AXIS_OPTION_KIND before checking, so an
  # option outside the table -- `title`, `visible`, the gridlines -- reaches
  # every axis of every chart type untouched
  untyped <- setdiff(names(formals(xl_chart_axis)), names(.AXIS_OPTION_KIND))
  expect_true(length(untyped) > 0L)
  for (ty in c("column", "scatter", "line"))
    expect_s3_class(xl_chart(ty, xl_chart_series(values = "A1:A5",
                                                 categories = "B1:B5"),
                             x_axis = xl_chart_axis(title = "t",
                                                    visible = FALSE),
                             y_axis = xl_chart_axis(title = "t",
                                                    visible = FALSE)),
                    "xl_chart")
  # and every option that IS in the table belongs to exactly one kind
  expect_true(all(.AXIS_OPTION_KIND %in% c("category", "value")))
})

# =============================================================================
# Data validation: restricting what may be typed into a range
# =============================================================================
#
# libxlsxwriter exposes 17 LXW_VALIDATION_TYPE_* values, but they are really 5
# kinds crossed with 3 ways of supplying the limit:
#
#            plain value        formula              serial number
#   integer  INTEGER            INTEGER_FORMULA      --
#   decimal  DECIMAL            DECIMAL_FORMULA      --
#   date     DATE               DATE_FORMULA         DATE_NUMBER
#   time     TIME               TIME_FORMULA         TIME_NUMBER
#   length   LENGTH             LENGTH_FORMULA       --
#
# plus LIST, LIST_FORMULA, ANY and CUSTOM_FORMULA.
#
# writexl exposes the five kinds and picks the variant from the R type of the
# limit: a character starting with "=" is a formula, a Date/POSIXct is a
# datetime.  The enum never reaches the user.
#
# libxlsxwriter checks a lot of this itself, but reports it as a warning on
# stderr plus a bare LXW_ERROR_PARAMETER_VALIDATION.  Every rule is therefore
# re-checked here, where the message can name the argument.
# -----------------------------------------------------------------------------

# Enum values, in the header's declaration order.
.LXW_VALIDATION <- c(
  none = 0L, integer = 1L, integer_formula = 2L, decimal = 3L,
  decimal_formula = 4L, list = 5L, list_formula = 6L, date = 7L,
  date_formula = 8L, date_number = 9L, time = 10L, time_formula = 11L,
  time_number = 12L, length = 13L, length_formula = 14L,
  custom_formula = 15L, any = 16L
)

.LXW_VALIDATION_CRITERIA <- c(
  none = 0L, between = 1L, "not between" = 2L, "==" = 3L, "!=" = 4L,
  ">" = 5L, "<" = 6L, ">=" = 7L, "<=" = 8L
)

.LXW_VALIDATION_ERROR_TYPE <- c(stop = 0L, warning = 1L, information = 2L)

# lxw_validation_boolean is three-valued, not a plain flag: 0 leaves
# libxlsxwriter's own default (which is "on" for all four toggles), so a FALSE
# has to be sent as OFF rather than as zero.
.LXW_VALIDATION_BOOL <- c("default" = 0L, off = 1L, on = 2L)
.validation_bool <- function(x) if (isTRUE(x)) 2L else 1L

# The kinds a user names, and whether each has a formula / serial variant.
.VALIDATION_KINDS <- c("integer", "decimal", "date", "time", "length",
                       "list", "custom", "any")

# Excel's own limits, enforced in libxlsxwriter as
# LXW_VALIDATION_MAX_TITLE_LENGTH / MAX_STRING_LENGTH.
.VALIDATION_TITLE_MAX  <- 32L
.VALIDATION_STRING_MAX <- 255L

# Which criteria need two bounds.
.is_between <- function(criteria) criteria %in% c("between", "not between")

# Is this limit an Excel formula rather than a literal?
.is_formula_limit <- function(x) {
  is.character(x) && length(x) == 1L && !is.na(x) && startsWith(x, "=")
}

.is_datetime_limit <- function(x) inherits(x, "Date") || inherits(x, "POSIXct")

# Check a title or message against Excel's length limit, in characters (which
# is what lxw_utf8_strlen() counts).
.check_validation_text <- function(x, arg, max) {
  if (.is_unset(x)) return(NULL)
  s <- .val_str(x, arg)
  if (!is.null(s) && nchar(s, type = "chars") > max)
    stop(sprintf("`%s` must be at most %d characters (got %d)", arg, max,
                 nchar(s, type = "chars")), call. = FALSE)
  s
}

# Resolve the libxlsxwriter type from the kind plus the R type of the limits.
.validation_type <- function(kind, limits) {
  if (kind == "any")    return("any")
  if (kind == "custom") return("custom_formula")
  if (kind == "list")
    return(if (length(limits) && .is_formula_limit(limits[[1L]]))
             "list_formula" else "list")
  supplied <- limits[!vapply(limits, is.null, logical(1))]
  if (length(supplied) && all(vapply(supplied, .is_formula_limit, logical(1))))
    return(paste0(kind, "_formula"))
  # A Date/POSIXct bound on a date/time validation is written as a serial
  # number, which is what the _NUMBER variants are for.
  if (kind %in% c("date", "time") &&
      length(supplied) && all(vapply(supplied, .is_datetime_limit, logical(1))))
    return(kind)
  kind
}

#' Restrict what can be typed into a range
#'
#' @description
#' `xl_validation()` adds Excel data validation to a range: a dropdown list, a
#' numeric or date bound, a text-length limit, or a custom formula.  Pass one or
#' a list of them as `xl_sheet(validation = )`.
#'
#' Excel's 17 internal validation types are collapsed into the five `type`
#' kinds below.  Whether a limit is a literal, a cell formula or a date is
#' inferred from what you pass: a string starting with `"="` is a formula, and a
#' `Date` or `POSIXct` is a date/time bound.
#'
#' @param range The cells to validate: an Excel range string such as
#'   `"B2:B100"`, a single cell, or a `list(rows = , cols = )` spec.
#' @param type The kind of value allowed: `"integer"`, `"decimal"`, `"date"`,
#'   `"time"`, `"length"` (of the text entered), `"custom"` (any formula that
#'   must evaluate `TRUE`), or `"any"` (no restriction, useful when you only
#'   want the input message).  Ignored when `list` is given.
#' @param list A dropdown of allowed values: a character vector of choices, or a
#'   single `"=..."` formula naming a range that holds them.  The choices are
#'   stored joined by commas, and Excel limits that joined string to 255
#'   characters.
#' @param criteria How `value` (or `min`/`max`) limits the entry: `"between"`,
#'   `"not between"`, `"=="`, `"!="`, `">"`, `"<"`, `">="` or `"<="`.  Required
#'   for the numeric, date, time and length kinds; must not be given for
#'   `list`, `custom` or `any`, which carry their own meaning.  Supplying `min`
#'   and `max` implies `"between"`.
#' @param value The single limit the criteria applies to.  A number, a `Date` or
#'   `POSIXct`, or a `"=..."` formula.
#' @param min,max The two limits for `"between"` / `"not between"`.  Supplying
#'   both and omitting `criteria` implies `"between"`.
#' @param input_title,input_message Text shown in a tooltip when the cell is
#'   selected.  Titles are limited to 32 characters and messages to 255.
#' @param error_title,error_message Text shown when an invalid entry is made.
#'   Same limits.
#' @param error_type What Excel does on an invalid entry: `"stop"` (refuse it),
#'   `"warning"` or `"information"` (both allow it through).
#' @param ignore_blank Logical; allow an empty cell.  `TRUE` by default, as in
#'   Excel.
#' @param show_input,show_error Logical; whether the input tooltip and the error
#'   alert are shown at all.  Both `TRUE` by default.
#' @param dropdown Logical; show the in-cell dropdown arrow for a `list`
#'   validation.  `TRUE` by default.
#' @return An `xl_validation` object.
#' @family worksheet features
#' @seealso [xl_sheet]
#' @export
#' @examples
#' # a dropdown
#' xl_validation("C2:C100", list = c("open", "high", "close"))
#'
#' # a numeric bound, with the message Excel shows on a bad entry
#' xl_validation("B2:B100", type = "integer", min = 1, max = 10,
#'               error_message = "Enter a whole number from 1 to 10")
#'
#' # a date bound
#' xl_validation("D2:D100", type = "date", criteria = ">=",
#'               value = as.Date("2024-01-01"))
#'
#' df <- data.frame(qty = 1:3)
#' tmp <- write_xlsx(list(Data = xl_sheet(df,
#'   validation = xl_validation("A2:A4", type = "integer", min = 0, max = 99))))
xl_validation <- function(range, type = "any", criteria = NA, value = NULL,
                          min = NULL, max = NULL, list = NULL,
                          input_title = NA, input_message = NA,
                          error_title = NA, error_message = NA,
                          error_type = NA, ignore_blank = NA,
                          show_input = NA, show_error = NA, dropdown = NA) {
  if (missing(range) || is.null(range))
    stop("`range` must name the cells to validate", call. = FALSE)
  kind <- .val_enum(type, .VALIDATION_KINDS, "type")
  if (is.null(kind)) kind <- "any"
  if (!is.null(list)) kind <- "list"

  crit <- .val_enum(criteria, names(.LXW_VALIDATION_CRITERIA)[-1L], "criteria")
  # Supplying both bounds is unambiguous, so let it stand in for the criteria.
  if (is.null(crit) && !is.null(min) && !is.null(max)) crit <- "between"

  .check_validation_args(kind, crit, list, value, min, max)

  structure(
    .drop_null(base::list(
      range = range, kind = kind, criteria = crit, list = list,
      value = value, min = min, max = max,
      input_title   = .check_validation_text(input_title, "input_title",
                                             .VALIDATION_TITLE_MAX),
      input_message = .check_validation_text(input_message, "input_message",
                                             .VALIDATION_STRING_MAX),
      error_title   = .check_validation_text(error_title, "error_title",
                                             .VALIDATION_TITLE_MAX),
      error_message = .check_validation_text(error_message, "error_message",
                                             .VALIDATION_STRING_MAX),
      error_type    = .val_enum(error_type, names(.LXW_VALIDATION_ERROR_TYPE),
                                "error_type"),
      ignore_blank  = .val_flag(ignore_blank, "ignore_blank"),
      show_input    = .val_flag(show_input, "show_input"),
      show_error    = .val_flag(show_error, "show_error"),
      dropdown      = .val_flag(dropdown, "dropdown")
    )),
    class = "xl_validation"
  )
}

# The rules libxlsxwriter enforces with a stderr warning and a bare error code.
.check_validation_args <- function(kind, crit, lst, value, min, max) {
  needs_criteria <- kind %in% c("integer", "decimal", "date", "time", "length")

  if (needs_criteria && is.null(crit))
    stop(sprintf(paste0("`criteria` is required for type = \"%s\"; give one of ",
                        "\"between\", \"not between\", \"==\", \"!=\", \">\", ",
                        "\"<\", \">=\", \"<=\", or supply both `min` and `max`"),
                 kind), call. = FALSE)
  if (!needs_criteria && !is.null(crit))
    stop(sprintf(paste0("`criteria` does not apply to type = \"%s\": that kind ",
                        "carries its own meaning"), kind), call. = FALSE)

  if (kind == "list") {
    if (is.null(lst) || !length(lst))
      stop("`list` must give at least one choice", call. = FALSE)
    if (!is.character(lst))
      stop("`list` must be a character vector of choices, or a single \"=\" ",
           "formula naming the range that holds them", call. = FALSE)
    if (anyNA(lst))
      stop("`list` choices must not be NA", call. = FALSE)
    if (!.is_formula_limit(lst[[1L]])) {
      joined <- nchar(paste(lst, collapse = ","), type = "chars")
      if (joined > .VALIDATION_STRING_MAX)
        stop(sprintf(paste0("the `list` choices joined by commas come to %d ",
                            "characters; Excel allows %d. Put them in a range ",
                            "and pass a \"=\" formula instead"),
                     joined, .VALIDATION_STRING_MAX), call. = FALSE)
    }
    return(invisible(NULL))
  }

  if (kind == "custom" && is.null(value))
    stop("type = \"custom\" needs a `value` holding the formula to evaluate",
         call. = FALSE)
  if (kind == "any") return(invisible(NULL))

  if (.is_between(crit %||% "")) {
    if (is.null(min) || is.null(max))
      stop(sprintf("`criteria = \"%s\"` needs both `min` and `max`", crit),
           call. = FALSE)
  } else if (needs_criteria && is.null(value)) {
    stop(sprintf("`criteria = \"%s\"` needs a `value` to compare against", crit),
         call. = FALSE)
  }
  invisible(NULL)
}

#' @export
print.xl_validation <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_validation: %s on %s>\n", p$kind,
              if (is.character(p$range)) p$range else "<spec>"))
  invisible(x)
}

# One limit, flattened into the (number, formula, datetime) triple the
# lxw_data_validation struct carries for each of value / minimum / maximum.
.validation_limit <- function(x, prefix) {
  if (is.null(x)) return(base::list())
  if (.is_datetime_limit(x))
    return(stats::setNames(base::list(.datetime_fields(x)),
                           paste0(prefix, "_datetime")))
  # a string is a formula whether or not it starts with "=" -- Excel stores the
  # bound as a formula either way, and libxlsxwriter adds the "=" itself
  if (is.character(x))
    return(stats::setNames(base::list(x), paste0(prefix, "_formula")))
  stats::setNames(base::list(as.numeric(x)), paste0(prefix, "_number"))
}

# Resolve a sheet's validations to the overlay payloads C applies.
.resolve_validations <- function(el, df, header_offset) {
  if (!inherits(el, "xl_sheet") || is.null(el$validation)) return(list())
  vs <- if (inherits(el$validation, "xl_validation")) list(el$validation)
        else el$validation
  if (!is.list(vs))
    stop("`validation` must be an xl_validation object or a list of them",
         call. = FALSE)
  lapply(seq_along(vs), function(i) {
    v <- vs[[i]]
    if (!inherits(v, "xl_validation"))
      stop(sprintf("`validation[[%d]]` must be an xl_validation object", i),
           call. = FALSE)
    p <- unclass(v)
    arg <- sprintf("validation[[%d]] range", i)
    q <- .xl_resolve_range(p$range, arg = arg, df = df,
                           header_offset = header_offset, allow_cell = TRUE)
    limits <- if (identical(p$kind, "list")) list(p$list[[1L]])
              else list(p$value, p$min, p$max)
    type <- .validation_type(p$kind, limits)

    out <- list(kind = "validation", range = as.integer(q),
                validate = unname(.LXW_VALIDATION[[type]]))
    if (!is.null(p$criteria))
      out$criteria <- unname(.LXW_VALIDATION_CRITERIA[[p$criteria]])
    if (!is.null(p$error_type))
      out$error_type <- unname(.LXW_VALIDATION_ERROR_TYPE[[p$error_type]])
    # A dropdown's choices travel as a character vector; C builds the
    # NULL-terminated array libxlsxwriter turns into a CSV formula.
    if (identical(p$kind, "list")) {
      if (.is_formula_limit(p$list[[1L]])) out$value_formula <- p$list[[1L]]
      else                                 out$value_list <- as.character(p$list)
    }
    if (identical(p$kind, "custom")) out$value_formula <- as.character(p$value)
    else {
      out <- c(out, .validation_limit(p$value, "value"),
               .validation_limit(p$min, "minimum"),
               .validation_limit(p$max, "maximum"))
    }
    for (k in c("input_title", "input_message", "error_title", "error_message"))
      if (!is.null(p[[k]])) out[[k]] <- p[[k]]
    # These three are on by default in libxlsxwriter, so only a FALSE is worth
    # sending; the C side distinguishes absent from 0.
    for (k in c("ignore_blank", "show_input", "show_error", "dropdown"))
      if (!is.null(p[[k]])) out[[k]] <- .validation_bool(p[[k]])
    out
  })
}

# =============================================================================
# Range addressing: one resolver for every range-scoped worksheet feature
# =============================================================================
#
# `.xl_resolve_range()` is the single place where a user-supplied range is
# turned into the 0-based `c(first_row, first_col, last_row, last_col)` quad
# that C consumes.  Every range-scoped feature goes through it -- the
# autofilter, merges, data validation, conditional formats, images, charts and
# tables -- so there is one set of accepted spellings and one set of error
# messages.
#
# Two spellings are accepted:
#
#   * an A1-style string:
#       "A1:D51"        a rectangle
#       "A2"            a single cell (a zero-span rectangle)
#       "$A$1:$D$51"    absolute references ($ is ignored)
#       "B:D"           whole columns, spanning the sheet's used rows
#       "2:10"          whole rows, spanning the sheet's used columns
#   * a data-frame-relative spec, `list(rows = , cols = )`, where `cols` names
#     or indexes columns of the sheet's data frame and `rows` gives 1-based
#     *data* row indices (row 1 is the first data row, ignoring the header).
#
# Either spelling may name a sheet -- "Data!A1:B5", "'My Sheet'!A1:B5", or a
# `sheet` element in the list form -- but almost nothing wants one.  An
# autofilter, a merge, a validation, a conditional format, a table, an image
# anchor and a print area all apply to the sheet they are on; a sheet name
# there is a misunderstanding, not a refinement.  Only a chart series can
# genuinely point at another sheet.
#
# So the sheet is *parsed* centrally and *allowed* selectively: callers opt in
# with allow_sheet = TRUE, and everything else reports the mistake instead of
# ignoring it.  Parsing it in one place keeps the quoting rules (and their
# error messages) from being reinvented per feature.
#
# The header offset is applied here, so C never has to know whether a header
# row was written.
# -----------------------------------------------------------------------------

# The xlsx grid limits (mirroring LXW_ROW_MAX / LXW_COL_MAX in libxlsxwriter).
.XLSX_MAX_ROWS <- 1048576L
.XLSX_MAX_COLS <- 16384L

# A1 notation, with the `$` of an absolute reference optional and ignored.
.RE_CELL_REF <- "^[$]?([A-Za-z]{1,3})[$]?([0-9]+)$"
.RE_COL_REF  <- "^[$]?([A-Za-z]{1,3})$"
.RE_ROW_REF  <- "^[$]?([0-9]+)$"

# Convert column letters ("A", "AB") to a 0-based column index.
.col_letters_to_index <- function(letters_str) {
  chars <- utf8ToInt(toupper(letters_str)) - utf8ToInt("A") + 1L
  idx <- 0L
  for (v in chars) idx <- idx * 26L + v
  idx - 1L
}

# Validate a 1-based sheet row number and return it 0-based.
.check_row_number <- function(n, arg) {
  n <- suppressWarnings(as.numeric(n))
  if (is.na(n) || n < 1 || n > .XLSX_MAX_ROWS)
    stop(sprintf("`%s` row must be between 1 and %d", arg, .XLSX_MAX_ROWS),
         call. = FALSE)
  as.integer(n) - 1L
}

# Validate an already 0-based column index.
.check_col_index <- function(i, arg) {
  if (is.na(i) || i < 0L || i >= .XLSX_MAX_COLS)
    stop(sprintf("`%s` column is outside the xlsx grid (maximum column is XFD)",
                 arg), call. = FALSE)
  as.integer(i)
}

# Parse one A1-style cell reference ("A2", "$A$2") to 0-based c(row, col).
.parse_cell_ref <- function(s, arg) {
  m <- regmatches(s, regexec(.RE_CELL_REF, s))[[1]]
  if (length(m) != 3L)
    stop(sprintf('`%s` must be a cell reference like "A2"', arg), call. = FALSE)
  c(.check_row_number(m[3L], arg),
    .check_col_index(.col_letters_to_index(m[2L]), arg))
}

# The rows a whole-column reference ("B:D") spans: the sheet's used rows when a
# data frame is available, otherwise the whole grid.
.sheet_row_extent <- function(df, header_offset, arg) {
  if (is.null(df)) return(c(0L, .XLSX_MAX_ROWS - 1L))
  last <- as.integer(header_offset) + nrow(df) - 1L
  if (last < 0L)
    stop(sprintf("`%s` covers no rows: the sheet is empty", arg), call. = FALSE)
  c(0L, as.integer(last))
}

# The columns a whole-row reference ("2:10") spans.
.sheet_col_extent <- function(df, arg) {
  if (is.null(df)) return(c(0L, .XLSX_MAX_COLS - 1L))
  if (length(df) < 1L)
    stop(sprintf("`%s` covers no columns: the sheet has no columns", arg),
         call. = FALSE)
  c(0L, as.integer(length(df) - 1L))
}

# Reject an inverted or off-grid quad.  A zero-span range (a single cell) is
# legitimate and passes.
.check_range_span <- function(v, arg) {
  # No current caller can reach this: every path in has already been through
  # .check_row_number() / .check_col_index(), which reject anything below the
  # top-left cell.  It stands so that a new producer of a quad cannot leak a
  # negative index into C.
  # nocov start
  if (v[1L] < 0L || v[2L] < 0L)
    stop(sprintf("`%s` range starts before the top-left cell", arg),
         call. = FALSE)
  # nocov end
  if (v[3L] < v[1L] || v[4L] < v[2L])
    stop(sprintf(paste0("`%s` range is inverted: the last cell must not come ",
                        "before the first"), arg), call. = FALSE)
  as.integer(v)
}

# Resolve an A1-style range string.
.resolve_range_string <- function(rng, arg, df, header_offset, allow_cell) {
  malformed <- function()
    stop(sprintf('`%s` range must look like "A1:D51" (got "%s")', arg, rng),
         call. = FALSE)
  has_colon <- grepl(":", rng, fixed = TRUE)
  parts <- strsplit(rng, ":", fixed = TRUE)[[1L]]
  if (!has_colon) {
    # A bare cell is a zero-span range, but only where one makes sense.
    if (!allow_cell || !grepl(.RE_CELL_REF, parts[1L])) malformed()
    rc <- .parse_cell_ref(parts[1L], arg)
    return(c(rc[1L], rc[2L], rc[1L], rc[2L]))
  }
  if (length(parts) != 2L || !all(nzchar(parts))) malformed()
  a <- parts[1L]
  b <- parts[2L]
  if (grepl(.RE_CELL_REF, a) && grepl(.RE_CELL_REF, b)) {
    ra <- .parse_cell_ref(a, arg)
    rb <- .parse_cell_ref(b, arg)
    return(.check_range_span(c(ra[1L], ra[2L], rb[1L], rb[2L]), arg))
  }
  if (grepl(.RE_COL_REF, a) && grepl(.RE_COL_REF, b)) {
    ext <- .sheet_row_extent(df, header_offset, arg)
    return(.check_range_span(c(
      ext[1L], .check_col_index(.col_letters_to_index(sub("^[$]", "", a)), arg),
      ext[2L], .check_col_index(.col_letters_to_index(sub("^[$]", "", b)), arg)
    ), arg))
  }
  if (grepl(.RE_ROW_REF, a) && grepl(.RE_ROW_REF, b)) {
    ext <- .sheet_col_extent(df, arg)
    return(.check_range_span(c(
      .check_row_number(sub("^[$]", "", a), arg), ext[1L],
      .check_row_number(sub("^[$]", "", b), arg), ext[2L]
    ), arg))
  }
  malformed()
}

# The elements a data-frame-relative range spec may carry.
.RANGE_SPEC_FIELDS <- c("rows", "cols")

# Turn a 1-based index vector into the c(first, last) it spans, refusing a
# non-contiguous selection: a range is a rectangle, and quietly filling the
# holes would select cells the caller did not ask for.
.contiguous_span <- function(idx, arg, what) {
  u <- sort(unique(as.integer(idx)))
  if (length(u) > 1L && any(diff(u) != 1L))
    stop(sprintf("`%s` %s must select a contiguous block", arg, what),
         call. = FALSE)
  c(u[1L], u[length(u)])
}

# Resolve list(rows = , cols = ) against the sheet's data frame.
.resolve_range_spec <- function(spec, arg, df, header_offset) {
  nms <- names(spec)
  if (!length(spec) || is.null(nms) || !all(nzchar(nms)))
    stop(sprintf("`%s` list must be fully named, e.g. list(rows = 1:3, cols = \"a\")",
                 arg), call. = FALSE)
  if (anyDuplicated(nms))
    stop(sprintf("`%s` list has duplicated element(s): %s", arg,
                 paste(unique(nms[duplicated(nms)]), collapse = ", ")),
         call. = FALSE)
  unknown <- setdiff(nms, .RANGE_SPEC_FIELDS)
  if (length(unknown))
    stop(sprintf("unknown `%s` element(s): %s", arg,
                 paste(unknown, collapse = ", ")), call. = FALSE)
  if (is.null(df))
    stop(sprintf("`%s` given as list(rows = , cols = ) needs the sheet's data frame",
                 arg), call. = FALSE)

  if (is.null(spec$cols)) {
    if (length(df) < 1L)
      stop(sprintf("`%s` covers no columns: the sheet has no columns", arg),
           call. = FALSE)
    cols <- c(1L, length(df))
  } else {
    cols <- .contiguous_span(.resolve_col_index(spec$cols, names(df), arg),
                             arg, "cols")
  }

  if (is.null(spec$rows)) {
    if (nrow(df) < 1L)
      stop(sprintf("`%s` covers no rows: the sheet has no data rows", arg),
           call. = FALSE)
    rows <- c(1L, nrow(df))
  } else {
    r <- spec$rows
    if (!is.numeric(r) || !length(r) || anyNA(r))
      stop(sprintf("`%s` rows must be a numeric vector of 1-based data-row indices",
                   arg), call. = FALSE)
    if (any(r < 1))
      stop(sprintf("`%s` rows must be positive (1-based data-row indices)", arg),
           call. = FALSE)
    rows <- .contiguous_span(r, arg, "rows")
  }

  first_row <- .check_row_number(rows[1L] + header_offset, arg)
  last_row  <- .check_row_number(rows[2L] + header_offset, arg)
  .check_range_span(c(first_row, cols[1L] - 1L, last_row, cols[2L] - 1L), arg)
}

# Split a sheet qualifier off a range string.  Excel quotes a sheet name that
# contains spaces or punctuation, and doubles an apostrophe inside it:
#   Data!A1:B5      'My Sheet'!A1:B5      'It''s'!A1
.split_sheet_ref <- function(x) {
  if (startsWith(x, "'")) {
    m <- regmatches(x, regexec("^'((?:[^']|'')*)'!(.*)$", x))[[1L]]
    if (length(m) == 3L)
      return(list(sheet = gsub("''", "'", m[2L], fixed = TRUE), rest = m[3L]))
    return(list(sheet = NULL, rest = x))
  }
  i <- regexpr("!", x, fixed = TRUE)
  if (i > 0L)
    return(list(sheet = substr(x, 1L, i - 1L), rest = substring(x, i + 1L)))
  list(sheet = NULL, rest = x)
}

# Reject a sheet qualifier where one cannot mean anything.
.reject_sheet <- function(sheet, arg) {
  if (is.null(sheet)) return(invisible(NULL))
  stop(sprintf(paste0("`%s` may not name a sheet (got \"%s!\"): it applies to ",
                      "the sheet it is on. Drop the sheet name."),
               arg, sheet), call. = FALSE)
}

# Resolve any accepted range spelling to 0-based
# c(first_row, first_col, last_row, last_col).
#
# `arg` names the calling argument in every error message.  `df` and
# `header_offset` supply the sheet context needed by the data-frame-relative
# and whole-column/whole-row forms.  `allow_cell = FALSE` rejects a bare cell
# reference for arguments documented as taking a rectangle.
# `allow_sheet = TRUE` accepts a sheet qualifier and returns it as the "sheet"
# attribute of the quad; otherwise a sheet name is an error.  See the note at
# the top of this file for why it is opt-in.
.xl_resolve_range <- function(x, arg = "range", df = NULL, header_offset = 1L,
                              allow_cell = TRUE, allow_sheet = FALSE) {
  sheet <- NULL
  if (is.list(x) && !is.data.frame(x)) {
    if (!is.null(x[["sheet"]])) {
      sheet <- x[["sheet"]]
      if (!is.character(sheet) || length(sheet) != 1L || is.na(sheet))
        stop(sprintf("`%s$sheet` must be a single sheet name", arg),
             call. = FALSE)
      x <- x[setdiff(names(x), "sheet")]
    }
    if (!allow_sheet) .reject_sheet(sheet, arg)
    out <- .resolve_range_spec(x, arg, df, header_offset)
  } else if (is.character(x) && length(x) == 1L && !is.na(x)) {
    sp <- .split_sheet_ref(x)
    sheet <- sp$sheet
    if (!allow_sheet) .reject_sheet(sheet, arg)
    out <- .resolve_range_string(sp$rest, arg, df, header_offset, allow_cell)
  } else {
    stop(sprintf(paste0('`%s` must be an Excel range string like "A1:D51" or a ',
                        "list(rows = , cols = ) spec"), arg), call. = FALSE)
  }
  if (allow_sheet && !is.null(sheet)) attr(out, "sheet") <- sheet
  out
}

# Parse an Excel range ("A1:D51") into 0-based c(first_row, first_col,
# last_row, last_col).  A thin wrapper over the shared resolver that keeps the
# rectangle-only contract of the arguments documented as taking a *range*.
.parse_range <- function(rng, arg = "autofilter", df = NULL,
                         header_offset = 1L) {
  .xl_resolve_range(rng, arg = arg, df = df, header_offset = header_offset,
                    allow_cell = FALSE)
}

# Parse a freeze specification into c(freeze_row, freeze_col) (counts), or
# c(-1, -1) for none.
#
# Note the different reading of `list(row =, col =)` here: for `freeze` these
# are *counts* of rows/columns to freeze, not indices, so only the cell
# reference form shares the range machinery.
.parse_freeze <- function(freeze) {
  if (is.null(freeze) || (length(freeze) == 1L && is.na(freeze)))
    return(c(-1L, -1L))
  if (is.list(freeze)) {
    r <- if (!is.null(freeze$row)) as.integer(freeze$row) else 0L
    cc <- if (!is.null(freeze$col)) as.integer(freeze$col) else 0L
    return(c(r, cc))
  }
  if (is.character(freeze) && length(freeze) == 1L)
    return(.parse_cell_ref(freeze, "freeze"))
  stop('`freeze` must be a cell reference ("A2") or list(row =, col =)',
       call. = FALSE)
}

# Tests for the shared range resolver in R/xl_range.R.

df3 <- data.frame(a = 1:3, b = 4:6, c = 7:9)

test_that("column letters convert to 0-based indices", {
  expect_equal(.col_letters_to_index("A"), 0L)
  expect_equal(.col_letters_to_index("b"), 1L)
  expect_equal(.col_letters_to_index("Z"), 25L)
  expect_equal(.col_letters_to_index("AA"), 26L)
  expect_equal(.col_letters_to_index("XFD"), 16383L)   # the last xlsx column
})

test_that("cell references parse, including absolute form", {
  expect_equal(.parse_cell_ref("A1", "x"), c(0L, 0L))
  expect_equal(.parse_cell_ref("a2", "x"), c(1L, 0L))
  expect_equal(.parse_cell_ref("D51", "x"), c(50L, 3L))
  expect_equal(.parse_cell_ref("$D$51", "x"), c(50L, 3L))
  expect_equal(.parse_cell_ref("$D51", "x"), c(50L, 3L))
  expect_error(.parse_cell_ref("nonsense", "x"), "cell reference")
  expect_error(.parse_cell_ref("1A", "x"), "cell reference")
  expect_error(.parse_cell_ref("", "x"), "cell reference")
})

test_that("grid limits are enforced on cell references", {
  expect_error(.parse_cell_ref("A0", "x"), "row must be between 1")
  expect_error(.parse_cell_ref("A1048577", "x"), "row must be between 1")
  expect_equal(.parse_cell_ref("XFD1048576", "x"), c(1048575L, 16383L))
  expect_error(.parse_cell_ref("XFE1", "x"), "outside the xlsx grid")
})

test_that("A1 rectangles resolve, absolute or not", {
  expect_equal(.xl_resolve_range("A1:D51", "x"), c(0L, 0L, 50L, 3L))
  expect_equal(.xl_resolve_range("$A$1:$D$51", "x"), c(0L, 0L, 50L, 3L))
  expect_equal(.xl_resolve_range("B2:B2", "x"), c(1L, 1L, 1L, 1L))  # zero span
})

test_that("a bare cell is a zero-span range where one is allowed", {
  expect_equal(.xl_resolve_range("A2", "x"), c(1L, 0L, 1L, 0L))
  expect_equal(.xl_resolve_range("$C$4", "x"), c(3L, 2L, 3L, 2L))
  # ... and is rejected where the argument is documented as taking a rectangle
  expect_error(.xl_resolve_range("A2", "x", allow_cell = FALSE), "A1:D51")
})

test_that("whole-column references span the sheet's used rows", {
  expect_equal(.xl_resolve_range("B:D", "x", df3, 1L), c(0L, 1L, 3L, 3L))
  expect_equal(.xl_resolve_range("B:D", "x", df3, 0L), c(0L, 1L, 2L, 3L))
  expect_equal(.xl_resolve_range("$B:$B", "x", df3, 1L), c(0L, 1L, 3L, 1L))
  # with no data frame in hand, the whole grid
  expect_equal(.xl_resolve_range("B:D", "x"), c(0L, 1L, 1048575L, 3L))
})

test_that("whole-row references span the sheet's used columns", {
  expect_equal(.xl_resolve_range("2:10", "x", df3, 1L), c(1L, 0L, 9L, 2L))
  expect_equal(.xl_resolve_range("$2:$2", "x", df3, 1L), c(1L, 0L, 1L, 2L))
  expect_equal(.xl_resolve_range("2:10", "x"), c(1L, 0L, 9L, 16383L))
})

test_that("malformed range strings are rejected and name the argument", {
  for (bad in c("A1:", ":", ":B2", "A1::B2", "A1:9", "9:A1", "A1:B", "",
                "not a range", "A1:B2:C3")) {
    expect_error(.xl_resolve_range(bad, "merge"), "`merge` range must look like",
                 fixed = TRUE)
  }
})

test_that("non-string, non-list ranges are rejected", {
  expect_error(.xl_resolve_range(42, "merge"), "`merge` must be an Excel range",
               fixed = TRUE)
  expect_error(.xl_resolve_range(NA_character_, "merge"), "must be an Excel range")
  expect_error(.xl_resolve_range(c("A1:B2", "C1:C2"), "merge"),
               "must be an Excel range")
})

test_that("inverted ranges are an error, zero-span ones are not", {
  expect_error(.xl_resolve_range("D1:A1", "merge"), "inverted")
  expect_error(.xl_resolve_range("A5:A2", "merge"), "inverted")
  expect_error(.xl_resolve_range("D:A", "merge", df3, 1L), "inverted")
  expect_error(.xl_resolve_range("10:2", "merge", df3, 1L), "inverted")
  expect_equal(.xl_resolve_range("A1:A1", "merge"), c(0L, 0L, 0L, 0L))
})

test_that("data-frame-relative specs resolve rows and columns", {
  expect_equal(.xl_resolve_range(list(rows = 1:2, cols = c("a", "b")), "m", df3, 1L),
               c(1L, 0L, 2L, 1L))
  expect_equal(.xl_resolve_range(list(rows = 1:2, cols = c("a", "b")), "m", df3, 0L),
               c(0L, 0L, 1L, 1L))
  expect_equal(.xl_resolve_range(list(rows = 2, cols = 2), "m", df3, 1L),
               c(2L, 1L, 2L, 1L))
  # numeric column positions and unsorted input both work
  expect_equal(.xl_resolve_range(list(rows = c(3, 2), cols = c(3, 2)), "m", df3, 1L),
               c(2L, 1L, 3L, 2L))
})

test_that("an omitted rows/cols element means the whole data block", {
  expect_equal(.xl_resolve_range(list(cols = "b"), "m", df3, 1L), c(1L, 1L, 3L, 1L))
  expect_equal(.xl_resolve_range(list(cols = "b"), "m", df3, 0L), c(0L, 1L, 2L, 1L))
  expect_equal(.xl_resolve_range(list(rows = 1), "m", df3, 1L), c(1L, 0L, 1L, 2L))
})

test_that("range specs are validated", {
  expect_error(.xl_resolve_range(list(), "m", df3, 1L), "must be fully named")
  expect_error(.xl_resolve_range(list(1:2), "m", df3, 1L), "must be fully named")
  expect_error(.xl_resolve_range(list(rows = 1, rows = 2), "m", df3, 1L),
               "duplicated element")
  expect_error(.xl_resolve_range(list(row = 1), "m", df3, 1L),
               "unknown `m` element", fixed = TRUE)
  expect_error(.xl_resolve_range(list(cols = "a"), "m"), "needs the sheet's data frame")
  expect_error(.xl_resolve_range(list(cols = "z"), "m", df3, 1L), "unknown column")
  expect_error(.xl_resolve_range(list(cols = 9), "m", df3, 1L), "out of range")
  expect_error(.xl_resolve_range(list(rows = 0), "m", df3, 1L), "must be positive")
  expect_error(.xl_resolve_range(list(rows = "a"), "m", df3, 1L), "must be a numeric")
  expect_error(.xl_resolve_range(list(rows = numeric(0)), "m", df3, 1L),
               "must be a numeric")
  expect_error(.xl_resolve_range(list(rows = c(1, NA)), "m", df3, 1L),
               "must be a numeric")
})

test_that("a range must be a contiguous rectangle", {
  expect_error(.xl_resolve_range(list(cols = c("a", "c")), "m", df3, 1L),
               "cols must select a contiguous block")
  expect_error(.xl_resolve_range(list(rows = c(1, 3)), "m", df3, 1L),
               "rows must select a contiguous block")
})

test_that("degenerate sheets are reported, not silently accepted", {
  empty_rows <- data.frame(a = integer(0), b = integer(0))
  no_cols <- data.frame(a = 1:2)[, character(0), drop = FALSE]
  expect_error(.xl_resolve_range(list(cols = "a"), "m", empty_rows, 1L),
               "covers no rows")
  expect_error(.xl_resolve_range(list(rows = 1), "m", no_cols, 1L),
               "covers no columns")
  expect_error(.xl_resolve_range("1:2", "m", no_cols, 1L), "covers no columns")
  # with a header row a 0-row sheet still has one usable row
  expect_equal(.xl_resolve_range("A:A", "m", empty_rows, 1L), c(0L, 0L, 0L, 0L))
  expect_error(.xl_resolve_range("A:A", "m", empty_rows, 0L), "covers no rows")
})

test_that("rows beyond the xlsx grid are rejected", {
  expect_error(.xl_resolve_range(list(rows = 1048576), "m", df3, 1L),
               "row must be between 1")
})

test_that(".parse_range keeps its rectangle-only contract", {
  expect_equal(.parse_range("A1:D51"), c(0L, 0L, 50L, 3L))
  expect_error(.parse_range("A1"), "`autofilter` range must look like",
               fixed = TRUE)
  expect_error(.parse_range("A1:B2", "merge"), NA)
  expect_error(.parse_range("A1", "merge"), "`merge` range must look like",
               fixed = TRUE)
})

test_that("freeze specs reuse the cell-reference parser but keep count semantics", {
  expect_equal(.parse_freeze("$B$3"), c(2L, 1L))
  # list(row =, col =) here means counts of frozen rows/columns, not indices
  expect_equal(.parse_freeze(list(row = 2, col = 1)), c(2L, 1L))
  expect_equal(.parse_freeze(list(row = 3)), c(3L, 0L))
  expect_equal(.parse_freeze(list(col = 3)), c(0L, 3L))
  expect_equal(.parse_freeze(NA_character_), c(-1L, -1L))
})

# ── Sheet-qualified ranges ───────────────────────────────────────────────────

test_that("a quoted sheet name is split off, apostrophes and all", {
  # Excel quotes a sheet name holding spaces or punctuation and doubles an
  # apostrophe inside it
  expect_equal(.split_sheet_ref("Data!A1:B5"),
               list(sheet = "Data", rest = "A1:B5"))
  expect_equal(.split_sheet_ref("'My Sheet'!A1:B5"),
               list(sheet = "My Sheet", rest = "A1:B5"))
  expect_equal(.split_sheet_ref("'It''s'!A1"),
               list(sheet = "It's", rest = "A1"))
  expect_equal(.split_sheet_ref("A1:B5"), list(sheet = NULL, rest = "A1:B5"))
  # an opening quote with no closing one is not a sheet reference
  expect_equal(.split_sheet_ref("'unterminated"),
               list(sheet = NULL, rest = "'unterminated"))
})

test_that("a sheet name is refused where a range applies to one sheet only", {
  df <- data.frame(a = 1:3, b = 4:6)
  expect_error(.xl_resolve_range("Other!A1:B2", "range", df),
               "may not name a sheet")
  expect_error(.xl_resolve_range(list(sheet = "Other", cols = "a"), "range", df),
               "may not name a sheet")
  # and carried as an attribute where it is allowed
  got <- .xl_resolve_range(list(sheet = "Other", cols = "a"), "range", df,
                           allow_sheet = TRUE)
  expect_equal(attr(got, "sheet"), "Other")
  got <- .xl_resolve_range("'My Sheet'!A1:B2", "range", df, allow_sheet = TRUE)
  expect_equal(attr(got, "sheet"), "My Sheet")
})

test_that("a sheet name in a spec must be a single string", {
  df <- data.frame(a = 1:3)
  for (bad in list(1, c("a", "b"), NA_character_))
    expect_error(.xl_resolve_range(list(sheet = bad, cols = "a"), "range", df,
                                   allow_sheet = TRUE),
                 "must be a single sheet name")
})

# =============================================================================
# xl_comment: Excel cell comments (a.k.a. notes)
# =============================================================================
#
# A comment is a per-cell overlay, independent of the cell's value/formula/
# hyperlink.  Excel comment boxes support only a background color and a font
# name/size/family -- no bold, italic, text color, borders, number formats or
# alignment.  So a comment reuses `xl_format` as its *styling input* (via the
# `format` argument), extracts the four supported properties, and warns about
# anything else through `.check_comment_format()`.
# -----------------------------------------------------------------------------

#' Test whether an object is an `xl_comment`
#' @param x An object.
#' @return `TRUE` if `x` inherits from `"xl_comment"`.
#' @family cell content
#' @export
is_xl_comment <- function(x) inherits(x, "xl_comment")

#' Create a cell comment
#'
#' @description
#' `xl_comment()` builds a comment (note) that can be attached to a cell via
#' [xl_cell_general()]'s `comment` argument.  The simplest form is just text
#' (`xl_cell_general(value = x, comment = "see note")`); `xl_comment()` adds
#' options such as the author, initial visibility, box size and position, and
#' styling.
#'
#' Excel comment boxes support only a **background color** and a **font
#' name/size/family**.  These are supplied by reusing the formatting engine via
#' the `format` argument (e.g. `xl_font(name = "Arial", size = 10) +
#' xl_fill(background = "lightyellow")`); any other format property (bold,
#' borders, number formats, ...) is not supported by comments and triggers a
#' warning.
#'
#' @param value A single string: the comment's text.  Anything with an
#'   [as.character()] method is accepted, so a cell built for a sheet can be
#'   reused here.
#' @param format An optional [xl_format]; only its fill background color and
#'   font name/size/family are used (see [xl_fill()], [xl_font()]).  Other
#'   properties are unsupported and warned about.
#' @param author Comment author (shown in Excel's status bar).  Defaults to the
#'   sheet/workbook `comment_author` when unset (see [xl_sheet()],
#'   [xl_properties()]).
#' @param visible Initial visibility: `NA` follows the sheet default (comments
#'   are hidden unless `show_comments` is set), `TRUE` shows this comment,
#'   `FALSE` hides it.
#' @param width_pixels,height_pixels Comment box size in pixels (defaults
#'   128 x 74).
#' @param x_scale,y_scale Box scale factors.
#' @param start_row,start_col Zero-based anchor cell of the box (by default a
#'   comment sits one row up and one column right of its cell).
#' @param x_offset,y_offset Pixel offset of the box from its anchor.
#' @return An `xl_comment` object.
#' @family cell content
#' @seealso [xl_cell_general], [xl_format]
#' @export
#' @examples
#' # plain text
#' xl_comment("Double-check this figure")
#'
#' # with author and styling reused from the format engine
#' xl_comment("Estimate", author = "Finance",
#'            format = xl_font(name = "Arial", size = 10) +
#'                     xl_fill(background = "lightyellow"))
xl_comment <- function(value, format = NULL, author = NA, visible = NA,
                       width_pixels = NA, height_pixels = NA, x_scale = NA,
                       y_scale = NA, start_row = NA, start_col = NA,
                       x_offset = NA, y_offset = NA) {
  value <- .as_display_string(value, "value", null_ok = FALSE)

  # styling: validate now, but keep the xl_format object itself so it stays
  # easy to inspect and modify (e.g. cm$format <- cm$format + xl_font(size = 9)).
  # It is flattened to the fields libxlsxwriter wants only on the way to C, in
  # .comment_c_payload().
  if (!is.null(format)) {
    if (!is_xl_format(format))
      stop("`format` must be an xl_format object", call. = FALSE)
    .check_comment_format(format)
  }

  vis <- .val_flag(visible, "visible")
  visible_code <- if (is.null(vis)) NULL else if (isTRUE(vis)) 2L else 1L

  payload <- .drop_null(list(
    value       = value,
    format      = format,
    author      = .val_str(author, "author"),
    visible     = visible_code,
    width       = .val_int(width_pixels, "width_pixels", min = 0),
    height      = .val_int(height_pixels, "height_pixels", min = 0),
    x_scale     = .val_num(x_scale, "x_scale", min = 0),
    y_scale     = .val_num(y_scale, "y_scale", min = 0),
    start_row   = .val_int(start_row, "start_row", min = 0),
    start_col   = .val_int(start_col, "start_col", min = 0),
    x_offset    = .val_int(x_offset, "x_offset"),
    y_offset    = .val_int(y_offset, "y_offset")
  ))
  structure(payload, class = "xl_comment")
}

# Flatten a comment payload for C: replace the `format` object with the four
# fields Excel comments actually support.  Called on the final pass to C so the
# xl_format stays editable up to that point.
.comment_c_payload <- function(cm) {
  if (is.null(cm)) return(NULL)
  fmt <- cm$format
  cm$format <- NULL
  if (!is.null(fmt)) {
    f <- unclass(fmt)
    cm$color       <- f$fill$background
    cm$font_name   <- f$font$name
    cm$font_size   <- f$font$size
    cm$font_family <- f$font$family
  }
  .drop_null(cm)
}

#' @export
print.xl_comment <- function(x, ...) {
  p <- unclass(x)
  cat("<xl_comment>\n")
  cat("  value:", p$value, "\n")
  for (o in setdiff(names(p), c("value", "format")))
    cat(sprintf("  %s: %s\n", o, paste(format(p[[o]]), collapse = ",")))
  if (!is.null(p$format)) print(p$format)
  invisible(x)
}

# Warn about xl_format properties Excel comments cannot render.  Supported:
# fill$background, font$name/size/family.  fill$pattern is ignored only when
# "solid" (the artifact xl_fill() adds for a background); a non-solid pattern is
# a real, unsupported request and is warned.  Everything else is warned.
.check_comment_format <- function(format) {
  f <- unclass(format)
  bad <- character(0)
  if (!is.null(f$font)) {
    extra <- setdiff(names(f$font), c("name", "size", "family"))
    if (length(extra)) bad <- c(bad, paste0("font$", extra))
  }
  if (!is.null(f$fill)) {
    extra <- setdiff(names(f$fill), c("background", "pattern"))
    if (length(extra)) bad <- c(bad, paste0("fill$", extra))
    if (!is.null(f$fill$pattern) && !identical(f$fill$pattern, "solid"))
      bad <- c(bad, "fill$pattern")
  }
  for (g in c("border", "align", "num_format", "protection"))
    if (!is.null(f[[g]])) bad <- c(bad, g)
  if (isTRUE(f$quote_prefix)) bad <- c(bad, "quote_prefix")
  if (isTRUE(f$hyperlink))    bad <- c(bad, "hyperlink")
  if (length(bad))
    warning("xl_comment(): these format properties are not supported by Excel ",
            "comments and will be ignored: ", paste(bad, collapse = ", "),
            call. = FALSE)
  invisible(bad)
}

# Normalise one per-cell comment element to a C-ready payload (or NULL).
# Accepts NA (no comment), a single string (default-option comment), or an
# xl_comment.
.comment_payload <- function(x) {
  if (is.null(x)) return(NULL)
  if (is.character(x) && length(x) == 1L && is.na(x)) return(NULL)
  if (length(x) == 1L && !is.list(x) && is.na(x)) return(NULL)
  if (is_xl_comment(x)) return(unclass(x))
  if (is.character(x) && length(x) == 1L) return(unclass(xl_comment(x)))
  stop("each comment must be NA, a single string, or an xl_comment object",
       call. = FALSE)
}

# =============================================================================
# Sheet names: fix them up so Excel will open the file
# =============================================================================
#
# Excel's rules, mirrored from workbook_validate_sheet_name() in
# src/libxlsxwriter/src/workbook.c: a name must be non-empty, at most 31
# characters, must not contain any of [ ] : * ? / \, and must not start or end
# with an apostrophe.  Duplicates are also rejected.
#
# writexl fixes names up and warns rather than erroring, which is how it has
# always treated over-long and duplicate names -- a sheet name is often derived
# from data, and failing the whole export over one stray "/" would be unkind.
# The warnings name the original and the replacement so a rename is never
# silent.
#
# The stages run in this order, because each one has to see the previous one's
# output: sanitize -> truncate -> deduplicate.  Deduplicating last matters:
# "a/b" and "a-b" both sanitize to "a-b", and only a final dedup pass catches
# the collision that sanitizing created.
#
# An empty name is left alone rather than repaired.  writexl uses "" to mean
# "let libxlsxwriter name this sheet", and C passes NULL for it.
# -----------------------------------------------------------------------------

.SHEETNAME_MAX <- 31L

# The characters Excel forbids in a sheet name.
.SHEETNAME_BAD_CHARS <- c("[", "]", ":", "*", "?", "/", "\\")

# Report a set of renames as one warning, listing at most `max_show` of them.
.warn_renamed <- function(what, orig, new, changed, max_show = 5L) {
  shown <- utils::head(changed, max_show)
  detail <- paste(sprintf('"%s" -> "%s"', orig[shown], new[shown]),
                  collapse = ", ")
  if (length(changed) > max_show)
    detail <- paste0(detail, sprintf(", and %d more", length(changed) - max_show))
  warning(what, ": ", detail, call. = FALSE)
}

# Replace the characters Excel forbids, and strip apostrophes from the ends.
.sanitize_sheet_names <- function(x) {
  out <- x
  for (ch in .SHEETNAME_BAD_CHARS)
    out <- gsub(ch, "-", out, fixed = TRUE)
  out <- sub("^'+", "", out)
  out <- sub("'+$", "", out)
  changed <- which(out != x & nzchar(x))
  if (length(changed))
    .warn_renamed("Replacing characters Excel does not allow in sheet name(s)",
                  x, out, changed)
  out
}

# Cut to Excel's 31-character maximum.
.truncate_sheet_names <- function(x) {
  out <- x
  long <- nzchar(x) & nchar(x, type = "chars") > .SHEETNAME_MAX
  out[long] <- substr(x[long], 1L, .SHEETNAME_MAX)
  if (any(long))
    .warn_renamed("Truncating sheet name(s) to 31 characters", x, out,
                  which(long))
  out
}

# Make names unique, keeping every result within the 31-character maximum by
# shortening the base rather than letting the suffix push it over.
.dedupe_sheet_names <- function(x) {
  out <- x
  seen <- character(0)
  for (i in seq_along(x)) {
    nm <- x[i]
    # empty means "libxlsxwriter names this one"; two empties do not collide
    if (!nzchar(nm)) next
    cand <- nm
    k <- 0L
    while (cand %in% seen) {
      k <- k + 1L
      sfx <- paste0("_", k)
      cand <- paste0(substr(nm, 1L, .SHEETNAME_MAX - nchar(sfx, type = "chars")),
                     sfx)
    }
    seen <- c(seen, cand)
    out[i] <- cand
  }
  changed <- which(out != x)
  if (length(changed))
    .warn_renamed("Deduplicating sheet names", x, out, changed)
  out
}

# Resolve the user's sheet names to ones Excel will accept.
.resolve_sheet_names <- function(nms, n) {
  if (is.null(nms)) return(rep("", n))
  nms <- as.character(nms)
  nms[is.na(nms)] <- ""
  .dedupe_sheet_names(.truncate_sheet_names(.sanitize_sheet_names(nms)))
}

# =============================================================================
# Sheet view: the tab strip and what the user sees when the sheet opens
# =============================================================================
#
# Grouped into an object for the same reason xl_page_setup() is: these are a
# cluster of rarely-used settings that would otherwise widen xl_sheet()'s
# signature past the point of being readable.
#
# `freeze` deliberately stays on xl_sheet() rather than moving here.  It is the
# single most common worksheet option -- keeping the header row visible -- and
# making the common case wordier to tidy the rare ones would be a poor trade.
# `split`, its rarely-used divider variant, does live here; the two remain
# mutually exclusive and the check spans both.
# -----------------------------------------------------------------------------

#' How a worksheet appears when it opens
#'
#' @description
#' `xl_sheet_view()` collects a worksheet's tab state and opening view: which
#' tab is active, selected or hidden, where the sheet is scrolled and selected,
#' and a few display options.  Pass it as `xl_sheet(view = )`.
#'
#' None of these affect the cell data.
#'
#' @param active Logical; make this the tab Excel opens on.  At most one sheet
#'   in a workbook may be active.
#' @param selected Logical; include this tab in the selected group.  The active
#'   sheet is always selected.
#' @param visible Logical; `FALSE` hides the sheet's tab.  A hidden sheet cannot
#'   be active or selected, the first sheet cannot be hidden unless another is
#'   made active, and at least one sheet must stay visible or Excel will not
#'   open the file.  All four rules are checked before writing.
#' @param first_tab Logical; make this the leftmost visible tab in the tab
#'   strip.  Independent of which sheet is active.
#' @param selection The cell or range selected when the sheet opens, as an Excel
#'   reference (`"B2"`, `"B2:D10"`) or a `list(rows = , cols = )` spec.
#'
#'   Excel also uses the order of a selection's corners to mark which cell in it
#'   is active; writexl does not expose that, because ranges are normalised by
#'   the shared range parser, which rejects an inverted range.
#' @param top_left The cell scrolled to the top-left of the window when the
#'   sheet opens, as an Excel reference such as `"A5"`.
#' @param hide_zero Logical; display zero values as blank cells.
#' @param right_to_left Logical; order the columns right to left, for a sheet in
#'   a right-to-left language.
#' @param split Split the sheet into scrollable panes with a visible, movable
#'   divider, given as the cell reference the split sits above and to the left
#'   of --- `"B3"` splits above row 3 and left of column B.  Mutually exclusive
#'   with `xl_sheet(freeze = )`, which does the same thing without the divider.
#'
#'   libxlsxwriter positions a split by distance, in row-height and column-width
#'   units, not by row and column number.  writexl converts the cell reference
#'   using the sheet's actual row heights and column widths, so the split lands
#'   where you asked even after resizing.  Pass
#'   `list(vertical = , horizontal = )` to give those units directly.
#'
#'   Note that libxlsxwriter derives the pane's scroll anchor back from that
#'   distance assuming default row heights, so on a sheet with resized rows or
#'   columns the divider is placed correctly but the anchor cell may be a row or
#'   two out.
#' @return An `xl_sheet_view` object.
#' @family worksheet layout
#' @seealso [xl_sheet], [xl_page_setup]
#' @export
#' @examples
#' xl_sheet_view(active = TRUE, selection = "B2")
#' xl_sheet_view(visible = FALSE)
#'
#' df <- data.frame(x = 1:3)
#' tmp <- write_xlsx(list(
#'   Summary = xl_sheet(df, view = xl_sheet_view(active = TRUE)),
#'   Working = xl_sheet(df, view = xl_sheet_view(visible = FALSE))
#' ))
xl_sheet_view <- function(active = NA, selected = NA, visible = NA,
                          first_tab = NA, selection = NULL, top_left = NULL,
                          hide_zero = NA, right_to_left = NA, split = NULL) {
  structure(
    list(active        = .val_flag(active, "active"),
         selected      = .val_flag(selected, "selected"),
         visible       = .val_flag(visible, "visible"),
         first_tab     = .val_flag(first_tab, "first_tab"),
         selection     = selection,
         top_left      = top_left,
         hide_zero     = .val_flag(hide_zero, "hide_zero"),
         right_to_left = .val_flag(right_to_left, "right_to_left"),
         split         = split),
    class = "xl_sheet_view"
  )
}

#' @export
print.xl_sheet_view <- function(x, ...) {
  p <- .drop_null(unclass(x))
  cat(sprintf("<xl_sheet_view: %d setting%s>\n", length(p),
              if (length(p) == 1L) "" else "s"))
  for (k in names(p))
    cat(sprintf("  %s: %s\n", k, paste(format(unlist(p[[k]])), collapse = ", ")))
  invisible(x)
}

# The view settings of a sheet element, with absent ones as NULL.  Accepts a
# sheet whose `view` is unset so callers need not special-case it.
.sheet_view_of <- function(el) {
  if (!inherits(el, "xl_sheet") || is.null(el$view)) return(NULL)
  if (!inherits(el$view, "xl_sheet_view"))
    stop("`view` must be an xl_sheet_view object", call. = FALSE)
  unclass(el$view)
}

# =============================================================================
# The chart itself: legend, data table, plot and chart areas, and the options
# that belong to one family of chart
# =============================================================================
#
# Two constructors -- the legend and the data table -- and a set of xl_chart()
# arguments for the rest, which are single values rather than objects with
# parts of their own.
#
# Six apply to one family of chart only: the doughnut hole, pie rotation,
# up-down bars, high-low lines, drop lines, and the bar gap and overlap.
# .CHART_FEATURE_FAMILIES in xl_chart.R holds which, since Excel drops the
# rest without a word.
# -----------------------------------------------------------------------------

.LXW_CHART_BLANKS <- c(gap = 0L, zero = 1L, connected = 2L)

# --- The legend ---------------------------------------------------------------

#' A chart's legend
#'
#' @description
#' `xl_chart_legend()` moves, styles or removes the legend, and can leave
#' individual series out of it.
#'
#' @param position Where the legend sits: `"right"` (Excel's default),
#'   `"left"`, `"top"`, `"bottom"`, `"top_right"`, the `"overlay_*"` variants
#'   that let the legend sit over the plot, or `"none"` to remove it.
#' @param format An [xl_format()] styling the legend text --- the [xl_font()]
#'   group only, since libxlsxwriter gives a legend a font and nothing else.
#' @param layout Where to put the legend by hand, as `c(x, y)` or
#'   `c(x, y, width, height)` --- fractions of the chart, each above 0 and at
#'   most 1.  Excel places it for you otherwise.  `at` is a cell everywhere
#'   else in writexl, so a chart's own fractions are a `layout`.
#' @param delete_series Series to leave out of the legend, by position: `2`
#'   drops the second series' entry while still plotting it.  This is how a
#'   trendline or a helper series is kept out of the key.
#' @return An `xl_chart_legend` object.
#' @family images and charts
#' @seealso [xl_chart]
#' @export
#' @examples
#' xl_chart_legend(position = "bottom")
#' xl_chart_legend(position = "none")
#' xl_chart_legend(delete_series = 2)
xl_chart_legend <- function(position = NULL, format = NULL,
                            layout = NULL, delete_series = NULL) {
  ds <- delete_series
  if (!is.null(ds)) {
    if (!is.numeric(ds) || !length(ds) || anyNA(ds) || any(ds < 1))
      stop(paste0("`delete_series` must be the positions of the series to ",
                  "leave out of the legend, counting from 1"), call. = FALSE)
    ds <- as.integer(ds) - 1L                  # C counts from 0
  }
  structure(.drop_null(list(
    position = .val_enum(position, names(.LXW_CHART_LEGEND), "position"),
    format = .chart_format_payload(format, "format", accept = "font",
                                   part = "a chart legend"),
    layout = .chart_layout(layout, "layout"),
    delete_series = ds
  )), class = "xl_chart_legend")
}

# --- The data table -----------------------------------------------------------

#' The table of values under a chart
#'
#' @description
#' `xl_chart_table()` prints the plotted numbers in a grid beneath the chart,
#' which is Excel's "Data Table" chart element.  It is given to [xl_chart()] as
#' `data_table`.
#'
#' Naming none of the grid options leaves Excel's own: horizontal, vertical and
#' outline borders drawn, and no legend keys.
#'
#' @param show_keys Print each series' legend swatch in the table.
#' @param horizontal_border,vertical_border,outline_border Which of the grid's
#'   borders to draw.
#' @param format An [xl_format()] styling the table's text --- the [xl_font()]
#'   group only.
#' @return An `xl_chart_table` object.
#' @family images and charts
#' @seealso [xl_chart]
#' @export
#' @examples
#' xl_chart_table()
#' xl_chart_table(show_keys = TRUE, vertical_border = FALSE)
xl_chart_table <- function(show_keys = NA, horizontal_border = NA,
                           vertical_border = NA, outline_border = NA,
                           format = NULL) {
  structure(.drop_null(list(
    show_keys = .val_flag(show_keys, "show_keys"),
    horizontal = .val_flag(horizontal_border, "horizontal_border"),
    vertical = .val_flag(vertical_border, "vertical_border"),
    outline = .val_flag(outline_border, "outline_border"),
    format = .chart_format_payload(format, "format", accept = "font",
                                   part = "a chart data table")
  )), class = "xl_chart_table")
}

# --- Shared helpers -----------------------------------------------------------

# lxw_chart_layout: fractions of the chart's own width and height.  Two numbers
# place a thing, four also size it.
.chart_layout <- function(x, arg, size_ok = TRUE) {
  if (is.null(x)) return(NULL)
  n <- if (size_ok) c(2L, 4L) else 2L
  if (!is.numeric(x) || !length(x) %in% n || anyNA(x) ||
      any(x <= 0) || any(x > 1))
    stop(sprintf(paste0("`%s` must be %s, each above 0 and at most 1 -- they ",
                        "are fractions of the chart's width and height"),
                 arg, if (size_ok) "c(x, y) or c(x, y, width, height)"
                      else "c(x, y)"), call. = FALSE)
  as.numeric(x)
}

# TRUE, or an xl_format giving the line these are drawn with.
.chart_line_switch <- function(x, arg) {
  if (.is_unset(x)) return(NULL)
  if (isTRUE(x)) return(list(on = 1L))
  if (identical(x, FALSE)) return(NULL)
  if (is_xl_format(x)) {
    f <- .chart_format_payload(x, arg, accept = "line", part = arg)
    return(.drop_null(list(on = 1L, line = f[["line"]])))
  }
  stop(sprintf(paste0("`%s` must be TRUE, or an xl_format() giving the line to ",
                      "draw them with"), arg), call. = FALSE)
}

# TRUE, or list(up = , down = ) with a format for each bar.
.chart_up_down <- function(x, arg) {
  if (.is_unset(x)) return(NULL)
  if (isTRUE(x)) return(list(on = 1L))
  if (identical(x, FALSE)) return(NULL)
  if (!is.list(x) || !length(x) ||
      length(setdiff(names(x), c("up", "down"))) || is.null(names(x)))
    stop(sprintf(paste0("`%s` must be TRUE, or list(up = , down = ) with an ",
                        "xl_format() for either bar"), arg), call. = FALSE)
  out <- list(on = 1L)
  for (side in c("up", "down")) {
    f <- x[[side]]
    if (is.null(f)) next
    p <- .chart_format_payload(f, sprintf("%s$%s", arg, side),
                               accept = c("line", "fill", "pattern"),
                               part = "an up-down bar")
    out[[paste0(side, "_line")]] <- p[["line"]]
    out[[paste0(side, "_fill")]] <- p[["fill"]]
  }
  .drop_null(out)
}

# --- Payloads -----------------------------------------------------------------

.legend_payload <- function(l) {
  if (is.null(l)) return(NULL)
  p <- unclass(l)
  ent <- list()
  if (!is.null(p[["position"]]))
    ent$position <- unname(.LXW_CHART_LEGEND[[p[["position"]]]])
  ent$font <- p[["format"]]
  if (!is.null(p[["layout"]])) ent$layout <- p[["layout"]]
  ent$delete_series <- p[["delete_series"]]
  .drop_null(ent)
}

.chart_table_payload <- function(t) {
  if (is.null(t)) return(NULL)
  p <- unclass(t)
  ent <- list(on = 1L)
  grid <- c("horizontal", "vertical", "outline", "show_keys")
  if (any(vapply(grid, function(k) !is.null(p[[k]]), logical(1))))
    # Excel's own defaults, for whichever were not named
    ent$grid <- as.integer(c(
      isTRUE(p[["horizontal"]] %||% TRUE), isTRUE(p[["vertical"]] %||% TRUE),
      isTRUE(p[["outline"]] %||% TRUE), isTRUE(p[["show_keys"]] %||% FALSE)))
  ent$font <- p[["format"]]
  .drop_null(ent)
}

# The title carries its layout and overlay alongside its text and font, so
# that .chart_list() has one place to look.
.chart_title_extras <- function(ttl, layout, overlay) {
  lay <- .chart_layout(layout, "title_layout", size_ok = FALSE)
  ov <- .val_flag(overlay, "title_overlay")
  if (is.null(lay) && is.null(ov)) return(ttl)
  if (isTRUE(ttl[["off"]]))
    stop(paste0("`title_layout` and `title_overlay` place the title, but ",
                "`title = FALSE` removes it"), call. = FALSE)
  if (is.null(ttl))
    stop(paste0("`title_layout` and `title_overlay` place the title, so give ",
                "a `title` too. The title Excel generates for a single-series ",
                "chart is not in the file, so there is nothing to place."),
         call. = FALSE)
  c(ttl, .drop_null(list(layout = lay,
                         overlay = if (is.null(ov)) NULL else as.integer(ov))))
}

#' @export
print.xl_chart_legend <- function(x, ...) .print_part(x, "xl_chart_legend")

#' @export
print.xl_chart_table <- function(x, ...) .print_part(x, "xl_chart_table")

# The chart itself: the legend, the data table, the plot and chart areas, and
# the options that belong to one family of chart.

ex_sales <- data.frame(quarter = c("Q1", "Q2", "Q3", "Q4"),
                       revenue = c(10, 25, 18, 32),
                       cost = c(8, 12, 25, 20),
                       stringsAsFactors = FALSE)

ex_two <- function() list(
  xl_chart_series(values = list(cols = "revenue"),
                  categories = list(cols = "quarter")),
  xl_chart_series(values = list(cols = "cost"),
                  categories = list(cols = "quarter")))

ex_xml <- function(..., type = "column", series = ex_two()) {
  t <- tempfile(fileext = ".xlsx")
  write_xlsx(list(Data = xl_sheet(ex_sales,
                                  chart = xl_chart(type, series, ...))), t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                             full.names = TRUE)[1L], warn = FALSE),
        collapse = "")
}

# ── The legend ────────────────────────────────────────────────────────────────

test_that("the legend positions are libxlsxwriter's, with none at zero", {
  # NONE is 0 and the rest run 1..8.  Anything above the range
  # chart_legend_set_position() accepts is refused there with a warning, and
  # the legend stays where it is.
  expect_equal(.LXW_CHART_LEGEND[["none"]], 0L)
  expect_equal(unname(.LXW_CHART_LEGEND), 0:8)
  expect_equal(names(.LXW_CHART_LEGEND)[[1L]], "none")
})

test_that("a legend can be moved, styled, placed and removed", {
  expect_match(ex_xml(legend = xl_chart_legend("bottom")),
               '<c:legendPos val="b"/>', fixed = TRUE)
  expect_match(ex_xml(legend = xl_chart_legend("top_right")),
               '<c:legendPos val="tr"/>', fixed = TRUE)
  expect_match(ex_xml(legend = xl_chart_legend(format = xl_font(size = 8,
                                                                italic = TRUE))),
               'i="1"', fixed = TRUE)
  expect_match(ex_xml(legend = xl_chart_legend(layout = c(0.1, 0.1, 0.2, 0.2))),
               "<c:manualLayout>", fixed = TRUE)
  # and "none" removes it outright, with no warning from the C layer
  none <- expect_silent(ex_xml(legend = xl_chart_legend("none")))
  expect_false(grepl("<c:legend>", none, fixed = TRUE))
})

test_that("a series can be left out of the legend while still being plotted", {
  x <- ex_xml(legend = xl_chart_legend(delete_series = 2))
  expect_match(x, "<c:legendEntry>", fixed = TRUE)
  expect_match(x, '<c:idx val="1"/>', fixed = TRUE)   # 0-based on the way out
  # both series are still drawn
  expect_length(regmatches(x, gregexpr("<c:ser>", x))[[1L]], 2L)
})

test_that("legend arguments are validated", {
  expect_error(xl_chart_legend("middle"), "`position` must be one of")
  expect_error(xl_chart_legend(delete_series = 0), "counting from 1")
  expect_error(xl_chart_legend(delete_series = "first"), "counting from 1")
  expect_error(xl_chart_legend(format = xl_fill(background = "red")),
               "takes no fill")
  expect_error(xl_chart_legend(layout = c(0.5, 0.5, 0.5)),
               "c\\(x, y\\) or c\\(x, y, width, height\\)")
  expect_error(xl_chart("column", ex_two(), legend = "bottom"),
               "must be an xl_chart_legend")
})

# ── The data table ────────────────────────────────────────────────────────────

test_that("a data table reaches the file with its grid options", {
  expect_match(ex_xml(data_table = xl_chart_table()), "<c:dTable>",
               fixed = TRUE)
  x <- ex_xml(data_table = xl_chart_table(show_keys = TRUE, vertical_border = FALSE))
  expect_match(x, '<c:showKeys val="1"/>', fixed = TRUE)
  expect_match(x, '<c:showHorzBorder val="1"/>', fixed = TRUE)
  # a border switched off is written as an absence, not as val="0"
  expect_false(grepl("showVertBorder", x, fixed = TRUE))
})

test_that("a data table's font is the only thing it can be styled with", {
  expect_match(ex_xml(data_table = xl_chart_table(format = xl_font(size = 7))),
               'sz="700"', fixed = TRUE)
  expect_error(xl_chart_table(format = xl_border(all = "thin")),
               "takes no line")
  expect_error(xl_chart("column", ex_two(), data_table = TRUE),
               "must be an xl_chart_table")
})

# ── The plot and chart areas ──────────────────────────────────────────────────

test_that("the plot area and the chart area take a line, fill and pattern", {
  x <- ex_xml(plot_area_format = xl_fill(background = "#F2F2F2") +
                                 xl_border(all = "thin", color = "gray"),
              chart_area_format = xl_border(all = "dashed", color = "navy"))
  expect_match(x, "F2F2F2", fixed = TRUE)
  expect_match(x, '<a:prstDash val="dash"/>', fixed = TRUE)
  expect_match(ex_xml(chart_area_format = xl_fill(pattern = "light-up",
                                                  foreground = "#2E75B6",
                                                  background = "white")),
               "<a:pattFill", fixed = TRUE)
  # an area is a shape, so no font
  expect_error(xl_chart("column", ex_two(),
                        plot_area_format = xl_font(bold = TRUE)),
               "a shape and has no text")
})

test_that("the plot area and the title can be placed by hand", {
  expect_match(ex_xml(plot_area_layout = c(0.1, 0.1, 0.8, 0.7)),
               "<c:manualLayout>", fixed = TRUE)
  x <- ex_xml(title = "T", title_layout = c(0.3, 0.02),
              title_overlay = TRUE)
  expect_match(x, "<c:manualLayout>", fixed = TRUE)
  expect_match(x, '<c:overlay val="1"/>', fixed = TRUE)
})

test_that("placing a title needs a title to place", {
  expect_error(xl_chart("column", ex_two(), title_layout = c(0.3, 0.02)),
               "give a `title` too", fixed = TRUE)
  expect_error(xl_chart("column", ex_two(), title = FALSE,
                        title_overlay = TRUE),
               "`title = FALSE` removes it", fixed = TRUE)
  expect_error(xl_chart("column", ex_two(), title = "T",
                        title_layout = c(0, 0.5)),
               "each above 0 and at most 1")
})

# ── The options that belong to one family ─────────────────────────────────────

test_that("each family-only option is refused on the other families", {
  cases <- list(
    hole_size = list(val = 40, ok = "doughnut", no = "pie"),
    rotation = list(val = 90, ok = "pie", no = "column"),
    drop_lines = list(val = TRUE, ok = "line", no = "column"),
    high_low_lines = list(val = TRUE, ok = "line", no = "column"),
    up_down_bars = list(val = TRUE, ok = "line", no = "column"),
    series_gap = list(val = 50, ok = "column", no = "line"),
    series_overlap = list(val = -20, ok = "column", no = "line"))
  for (nm in names(cases)) {
    k <- cases[[nm]]
    args <- list(k$no, xl_chart_series(values = "A1:A5"))
    args[[nm]] <- k$val
    expect_error(do.call(xl_chart, args),
                 sprintf('`%s` does not apply to a "%s" chart', nm, k$no),
                 fixed = TRUE, label = nm)
    args <- list(k$ok, xl_chart_series(values = "A1:A5"))
    args[[nm]] <- k$val
    expect_true(inherits(do.call(xl_chart, args), "xl_chart"), info = nm)
  }
  # a hole belongs to a doughnut and not to a pie, which shares its family
  expect_error(xl_chart("pie", xl_chart_series(values = "A1"), hole_size = 40),
               "It applies to: doughnut", fixed = TRUE)
})

test_that("the family-only options reach the file", {
  d <- ex_xml(type = "doughnut", series = ex_two()[[1L]],
              hole_size = 40, rotation = 90)
  expect_match(d, '<c:holeSize val="40"/>', fixed = TRUE)
  expect_match(d, '<c:firstSliceAng val="90"/>', fixed = TRUE)
  b <- ex_xml(series_gap = 50, series_overlap = -20)
  expect_match(b, '<c:gapWidth val="50"/>', fixed = TRUE)
  expect_match(b, '<c:overlap val="-20"/>', fixed = TRUE)
})

test_that("drop, high-low and up-down lines take a format or a plain TRUE", {
  x <- ex_xml(type = "line",
              drop_lines = xl_border(all = "dotted", color = "gray"),
              high_low_lines = TRUE,
              up_down_bars = list(up = xl_fill(background = "green"),
                                  down = xl_fill(background = "red")))
  expect_match(x, "<c:dropLines>", fixed = TRUE)
  expect_match(x, "<c:hiLowLines", fixed = TRUE)
  expect_match(x, "<c:upDownBars>", fixed = TRUE)
  expect_match(x, "00FF00", fixed = TRUE)      # the up bar
  expect_match(x, "FF0000", fixed = TRUE)      # the down bar
  expect_match(ex_xml(type = "line", up_down_bars = TRUE), "<c:upDownBars>",
               fixed = TRUE)
})

test_that("the switch arguments say what they accept", {
  expect_error(xl_chart("line", ex_two(), drop_lines = "yes"),
               "must be TRUE, or an xl_format")
  expect_error(xl_chart("line", ex_two(), up_down_bars = list(sideways = 1)),
               "must be TRUE, or list\\(up = , down = \\)")
  expect_error(xl_chart("line", ex_two(),
                        drop_lines = xl_fill(background = "red")),
               "takes no fill")
  # FALSE is the same as leaving them out
  expect_null(unclass(xl_chart("line", ex_two(),
                               drop_lines = FALSE))$drop_lines)
})

test_that("the numeric options are held to Excel's ranges", {
  expect_error(xl_chart("doughnut", ex_two()[[1L]], hole_size = 5),
               "must be between 10 and 90")
  expect_error(xl_chart("pie", ex_two()[[1L]], rotation = 400),
               "must be between 0 and 360")
  expect_error(xl_chart("column", ex_two(), series_gap = 600),
               "must be between 0 and 500")
  expect_error(xl_chart("column", ex_two(), series_overlap = -200),
               "must be between -100 and 100")
})

# ── Blanks and hidden data ────────────────────────────────────────────────────

test_that("what an empty cell does to the plot can be chosen", {
  expect_equal(unname(.LXW_CHART_BLANKS), 0:2)
  expect_match(ex_xml(show_blanks = "zero"), '<c:dispBlanksAs val="zero"/>',
               fixed = TRUE)
  expect_match(ex_xml(show_blanks = "connected"),
               '<c:dispBlanksAs val="span"/>', fixed = TRUE)
  expect_error(xl_chart("column", ex_two(), show_blanks = "skip"),
               "`show_blanks` must be one of")
})

test_that("hidden data is plotted by leaving plotVisOnly out", {
  # libxlsxwriter expresses this as an absence rather than as val="0", so the
  # test pins the absence
  expect_match(ex_xml(), "<c:plotVisOnly", fixed = TRUE)
  expect_false(grepl("plotVisOnly", ex_xml(show_hidden_data = TRUE),
                     fixed = TRUE))
})

# ── Coverage of the C API ─────────────────────────────────────────────────────

test_that("every chart-level chart_*() function libxlsxwriter offers is called", {
  hdr <- NULL
  for (p in c("../../src/libxlsxwriter/include/xlsxwriter/chart.h",
              "../../../src/libxlsxwriter/include/xlsxwriter/chart.h"))
    if (file.exists(p)) hdr <- p
  skip_if(is.null(hdr), "libxlsxwriter sources not alongside the tests")
  src <- NULL
  for (p in c("../../src/write_xlsx.c", "../../../src/write_xlsx.c"))
    if (file.exists(p)) src <- p
  skip_if(is.null(src), "write_xlsx.c not alongside the tests")

  decl <- grep("^(void|lxw_error|uint8_t) chart_", readLines(hdr), value = TRUE)
  fns <- unique(regmatches(decl, regexpr("chart_[a-z_0-9]+", decl)))
  # the axes and the series have gates of their own
  fns <- fns[!grepl("^chart_(series|axis)_", fns)]
  # the title and the built-in style are in here too
  expect_equal(length(fns), 31L)

  called <- paste(readLines(src), collapse = "\n")
  for (f in fns)
    expect_true(grepl(paste0(f, "("), called, fixed = TRUE), info = f)
})

test_that("up-down bars accept the shapes they document, and refuse the rest", {
  expect_null(.chart_up_down(NA, "up_down_bars"))
  expect_null(.chart_up_down(FALSE, "up_down_bars"))
  expect_equal(.chart_up_down(TRUE, "up_down_bars"), list(on = 1L))
  # one side alone leaves the other at Excel's default
  one <- .chart_up_down(list(up = xl_fill(background = "green")),
                        "up_down_bars")
  expect_equal(one$on, 1L)
  expect_null(one$down_fill)
  expect_error(.chart_up_down(list(sideways = TRUE), "up_down_bars"),
               "must be TRUE, or list(up = , down = )", fixed = TRUE)
  expect_error(.chart_up_down("yes", "up_down_bars"), "must be TRUE, or list")
})

# =============================================================================
# Sheet visibility: which tab opens, which are selected, which are hidden
# =============================================================================
#
# These are the only worksheet settings whose rules span the whole workbook, so
# unlike everything else in the sheet plan they cannot be validated one sheet at
# a time.  Excel's rules, as documented for worksheet_hide():
#
#   * a hidden sheet cannot be activated or selected -- the calls are mutually
#     exclusive;
#   * the first sheet is active by default, so it cannot be hidden unless some
#     other sheet is made active;
#   * at most one sheet may be active;
#   * at least one sheet must stay visible, or Excel refuses to open the file.
#
# libxlsxwriter enforces none of them: worksheet_hide(), worksheet_activate()
# and worksheet_select() all return void and simply set a flag.  A bad
# combination therefore produces a broken workbook with no diagnostic, so the
# checks live here and name the sheet at fault.
# -----------------------------------------------------------------------------

# One sheet's visibility settings, with absent ones as NA.
.sheet_view_flags <- function(el) {
  vw <- if (inherits(el, "xl_chartsheet")) unclass(el)$view
        else .sheet_view_of(el)
  if (is.null(vw))
    return(list(active = NA, selected = NA, visible = NA, first_tab = NA))
  list(active    = vw$active,
       selected  = vw$selected,
       visible   = vw$visible,
       first_tab = vw$first_tab)
}

# A readable name for a sheet in an error message.
.sheet_label <- function(nms, i) {
  if (!is.null(nms) && length(nms) >= i && !is.na(nms[i]) && nzchar(nms[i]))
    sprintf('"%s"', nms[i])
  else
    sprintf("%d", i)
}

# Validate the workbook-wide visibility rules.  Returns nothing; it exists to
# fail loudly before anything is written.
.resolve_sheet_visibility <- function(elems, nms = names(elems)) {
  n <- length(elems)
  if (!n) return(invisible(NULL))
  f <- lapply(elems, .sheet_view_flags)
  is_set <- function(k, want) vapply(f, function(x) identical(x[[k]], want),
                                     logical(1))

  hidden   <- is_set("visible", FALSE)
  active   <- is_set("active", TRUE)
  selected <- is_set("selected", TRUE)

  # every sheet hidden -> Excel will not open the file
  if (all(hidden))
    stop("every sheet is hidden; a workbook needs at least one visible sheet",
         call. = FALSE)

  # a hidden sheet cannot also be the active or a selected one
  bad <- which(hidden & (active | selected))
  if (length(bad))
    stop(sprintf(paste0("sheet %s is hidden but also marked active/selected; ",
                        "Excel cannot show a hidden sheet"),
                 .sheet_label(nms, bad[1L])), call. = FALSE)

  # only one active sheet
  if (sum(active) > 1L)
    stop(sprintf("sheets %s are both marked active; only one sheet may be active",
                 paste(vapply(which(active), function(i) .sheet_label(nms, i),
                              character(1)), collapse = " and ")),
         call. = FALSE)

  # the first sheet is active by default, so hiding it needs another active one
  if (hidden[1L] && !any(active))
    stop(sprintf(paste0("sheet %s is the first sheet and cannot be hidden ",
                        "unless another sheet is given active = TRUE, since ",
                        "Excel opens on the first sheet by default"),
                 .sheet_label(nms, 1L)), call. = FALSE)

  invisible(NULL)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
sales <- data.frame(quarter = c("Q1", "Q2", "Q3", "Q4"),
                    revenue = c(120, 145, 133, 180),
                    cost    = c(90, 110, 105, 130))

## -----------------------------------------------------------------------------
sheet <- xl_sheet(sales,
                  chart = xl_chart("column",
                                   xl_chart_series(values = list(cols = "revenue")),
                                   title = "Revenue by quarter", at = "E2"))
path <- write_xlsx(list(Sales = sheet), tempfile(fileext = ".xlsx"))

## -----------------------------------------------------------------------------
xl_chart_series(values = list(sheet = "Sales", cols = "revenue", rows = 1:2))

## -----------------------------------------------------------------------------
xl_chart("line",
         xl_chart_series(values = list(cols = "revenue"),
                         format = xl_border(all = "dashed", color = "#4472C4")),
         title = "Revenue", title_format = xl_font(size = 14, bold = TRUE))

## -----------------------------------------------------------------------------
xl_chart("column", xl_chart_series(values = list(cols = "revenue")),
         x_axis = xl_chart_axis(title = "Quarter", major_tick = "none"),
         y_axis = xl_chart_axis(title = "Revenue", min = 0,
                                num_format = "$#,##0",
                                major_gridlines_format =
                                  xl_border(all = "dotted", color = "gray")))

## -----------------------------------------------------------------------------
xl_chart_series(
  values       = list(cols = "revenue"),
  marker       = xl_chart_marker(type = "circle", size = 7),
  labels       = xl_chart_labels(num_format = "$#,##0", position = "above"),
  trendline    = xl_chart_trendline("linear", equation = TRUE),
  y_error_bars = xl_chart_error_bars("percentage", 5))

## -----------------------------------------------------------------------------
xl_chart("pie", xl_chart_series(values = list(cols = "revenue"),
                                points = list(xl_fill(background = "red"),
                                              NULL, NULL, NULL)))

## -----------------------------------------------------------------------------
xl_chart("column", xl_chart_series(values = list(cols = "revenue")),
         legend     = xl_chart_legend(position = "bottom"),
         data_table = xl_chart_table(show_keys = TRUE),
         plot_area_format  = xl_fill(background = "#F8F8F8"),
         chart_area_format = xl_border(all = "thin", color = "gray"))

## -----------------------------------------------------------------------------
chart <- xl_chart("column",
                  xl_chart_series(values = list(sheet = "Data",
                                                cols = "revenue"),
                                  categories = list(sheet = "Data",
                                                    cols = "quarter")),
                  title = "Revenue")
path <- write_xlsx(list(Data = sales, Overview = xl_chartsheet(chart)),
                   tempfile(fileext = ".xlsx"))

## ----eval = isTRUE(capabilities("png"))---------------------------------------
card <- as.raster(matrix(c("red", "blue", "green", "gold"), nrow = 2))
xl_sheet(sales, image = xl_image(card, at = "G2", scale = 20))

## ----eval = FALSE-------------------------------------------------------------
# xl_page_setup(header = "&L&G", header_image = list(left = logo))
# xl_sheet(sales, background_image = logo)

# =============================================================================
# Conditional formatting
# =============================================================================
#
# lxw_conditional_format carries about thirty fields, but they fall into four
# clusters that share almost nothing: the simple rules (type, criteria, value,
# format), colour scales (min/mid/max x value, rule type, colour), data bars
# (a dozen bar_* fields) and icon sets (style, reverse, icons only).  Rather
# than one constructor with thirty arguments, there is one per cluster; each
# returns an xl_conditional carrying its `cluster`, and one overlay kind in C
# applies them all.
#
# The trap here is that the 34 criteria are partitioned by type -- text criteria
# only mean anything for a text rule, the time-period ones only for a time
# rule, and so on -- and libxlsxwriter does not check the pairing.  A mismatch
# produces a rule Excel silently ignores, so .CONDITIONAL_CRITERIA_FOR maps each
# type to what it accepts and the mismatch is an error naming both.
# -----------------------------------------------------------------------------

# Enum values, in the header's declaration order.
.LXW_COND_TYPE <- c(
  none = 0L, cell = 1L, text = 2L, time_period = 3L, average = 4L,
  duplicate = 5L, unique = 6L, top = 7L, bottom = 8L, blanks = 9L,
  no_blanks = 10L, errors = 11L, no_errors = 12L, formula = 13L,
  # Note these three carry no TYPE_ in their libxlsxwriter names, which makes
  # them easy to miss when reading the header -- and getting the offset wrong
  # would silently write the wrong kind of rule.
  "2_color_scale" = 14L, "3_color_scale" = 15L, data_bar = 16L,
  icon_sets = 17L
)

.LXW_COND_CRITERIA <- c(
  none = 0L,
  "==" = 1L, "!=" = 2L, ">" = 3L, "<" = 4L, ">=" = 5L, "<=" = 6L,
  between = 7L, "not between" = 8L,
  contains = 9L, "not contains" = 10L, "begins with" = 11L, "ends with" = 12L,
  yesterday = 13L, today = 14L, tomorrow = 15L, "last 7 days" = 16L,
  "last week" = 17L, "this week" = 18L, "next week" = 19L,
  "last month" = 20L, "this month" = 21L, "next month" = 22L,
  "above" = 23L, "below" = 24L, "above or equal" = 25L, "below or equal" = 26L,
  "1 std dev above" = 27L, "1 std dev below" = 28L,
  "2 std dev above" = 29L, "2 std dev below" = 30L,
  "3 std dev above" = 31L, "3 std dev below" = 32L,
  "percent" = 33L
)

.LXW_COND_RULE_TYPE <- c(
  none = 0L, minimum = 1L, number = 2L, percent = 3L, percentile = 4L,
  formula = 5L, maximum = 6L, auto_min = 7L, auto_max = 8L
)

.LXW_COND_ICONS <- c(
  "3_arrows"                  = 0L,
  "3_arrows_gray"             = 1L,
  "3_flags"                   = 2L,
  "3_traffic_lights"          = 3L,
  "3_traffic_lights_rimmed"   = 4L,
  "3_signs"                   = 5L,
  "3_symbols_circled"         = 6L,
  "3_symbols"                 = 7L,
  "4_arrows"                  = 8L,
  "4_arrows_gray"             = 9L,
  "4_red_to_black"            = 10L,
  "4_ratings"                 = 11L,
  "4_traffic_lights"          = 12L,
  "5_arrows"                  = 13L,
  "5_arrows_gray"             = 14L,
  "5_ratings"                 = 15L,
  "5_quarters"                = 16L
)

# Which criteria each rule type accepts.  libxlsxwriter checks none of this, and
# a mismatched pair produces a rule Excel quietly ignores.
.COND_COMPARISONS <- c("==", "!=", ">", "<", ">=", "<=", "between",
                       "not between")
.COND_TEXT        <- c("contains", "not contains", "begins with", "ends with")
.COND_TIME        <- c("yesterday", "today", "tomorrow", "last 7 days",
                       "last week", "this week", "next week", "last month",
                       "this month", "next month")
.COND_AVERAGE     <- c("above", "below", "above or equal", "below or equal",
                       "1 std dev above", "1 std dev below",
                       "2 std dev above", "2 std dev below",
                       "3 std dev above", "3 std dev below")

.CONDITIONAL_CRITERIA_FOR <- list(
  cell        = .COND_COMPARISONS,
  text        = .COND_TEXT,
  time_period = .COND_TIME,
  average     = .COND_AVERAGE,
  top         = "percent",          # optional: absent means "top N"
  bottom      = "percent",
  duplicate   = character(0),
  unique      = character(0),
  blanks      = character(0),
  no_blanks   = character(0),
  errors      = character(0),
  no_errors   = character(0),
  formula     = character(0)
)

# The criteria that only make sense with two bounds.
.cond_is_between <- function(x) !is.null(x) && x %in% c("between", "not between")

# Infer the rule type from the criteria when the caller did not name one: the
# two are largely redundant, and a comparison plainly means a cell rule.
.cond_infer_type <- function(type, criteria) {
  if (!is.null(type)) return(type)
  if (is.null(criteria)) return("cell")
  if (criteria %in% .COND_TEXT)    return("text")
  if (criteria %in% .COND_TIME)    return("time_period")
  if (criteria %in% .COND_AVERAGE) return("average")
  "cell"
}

.check_conditional_criteria <- function(type, criteria) {
  allowed <- .CONDITIONAL_CRITERIA_FOR[[type]]
  if (is.null(allowed)) return(invisible(NULL))
  if (is.null(criteria)) {
    # only the comparison and text kinds insist on one
    if (type %in% c("cell", "text", "time_period", "average"))
      stop(sprintf(paste0("`criteria` is required for type = \"%s\"; one of: %s"),
                   type, paste(allowed, collapse = ", ")), call. = FALSE)
    return(invisible(NULL))
  }
  if (!length(allowed))
    stop(sprintf(paste0("`criteria` does not apply to type = \"%s\": that rule ",
                        "carries its own meaning"), type), call. = FALSE)
  if (!criteria %in% allowed)
    stop(sprintf(paste0("`criteria = \"%s\"` is not valid for type = \"%s\". ",
                        "Excel would ignore the rule. Valid here: %s"),
                 criteria, type, paste(allowed, collapse = ", ")),
         call. = FALSE)
  invisible(NULL)
}

# Shared tail of every constructor.
.new_conditional <- function(cluster, range, fields, stop_if_true, multi_range) {
  # every constructor checks this first, so it can fail before its own
  # validation runs; this is the backstop for one added later
  if (missing(range) || is.null(range))                                # nocov
    stop("`range` must name the cells to format", call. = FALSE)       # nocov
  structure(
    c(list(cluster = cluster, range = range),
      fields,
      .drop_null(list(stop_if_true = .val_flag(stop_if_true, "stop_if_true"),
                      multi_range  = .val_str(multi_range, "multi_range")))),
    class = "xl_conditional"
  )
}

#' Format cells according to their contents
#'
#' @description
#' Excel's conditional formatting, in four flavours:
#'
#' * `xl_cond_cell()` --- a rule with a format: comparisons, text matches, time
#'   periods, above/below average, top/bottom N, duplicates, blanks, errors, or
#'   an arbitrary formula.
#' * `xl_cond_scale()` --- a two- or three-colour scale across the range.
#' * `xl_cond_bar()` --- in-cell data bars.
#' * `xl_cond_icons()` --- one of Excel's built-in icon sets.
#'
#' Pass one or a list of them as `xl_sheet(conditional = )`.
#'
#' @param range The cells the rule applies to: an Excel range string such as
#'   `"B2:B100"`, a single cell, or a `list(rows = , cols = )` spec.
#' @param criteria How the rule decides, which depends on `type`:
#'   `"=="`, `"!="`, `">"`, `"<"`, `">="`, `"<="`, `"between"`, `"not between"`
#'   for `"cell"`; `"contains"`, `"not contains"`, `"begins with"`,
#'   `"ends with"` for `"text"`; `"yesterday"`, `"today"`, `"tomorrow"`,
#'   `"last 7 days"`, `"last week"`, `"this week"`, `"next week"`,
#'   `"last month"`, `"this month"`, `"next month"` for `"time_period"`;
#'   `"above"`, `"below"`, `"above or equal"`, `"below or equal"` and the
#'   `"N std dev above"` / `"below"` variants for `"average"`; and `"percent"`
#'   for `"top"` / `"bottom"`.
#'
#'   Pairing a criteria with the wrong `type` is an error --- Excel would accept
#'   the file and silently ignore the rule.
#' @param value What the criteria compares against: a number, a string (for the
#'   text criteria), or an `"=..."` formula.  For `"top"` / `"bottom"` it is the
#'   N.  Use `min` and `max` for `"between"` / `"not between"`.
#' @param min,max The two bounds for `"between"` / `"not between"`.
#' @param type The kind of rule, when it cannot be inferred from `criteria`:
#'   `"cell"`, `"text"`, `"time_period"`, `"average"`, `"top"`, `"bottom"`,
#'   `"duplicate"`, `"unique"`, `"blanks"`, `"no_blanks"`, `"errors"`,
#'   `"no_errors"` or `"formula"`.
#' @param format The [xl_format] applied to cells that match.
#' @param stop_if_true Logical; if this rule matches, skip the later rules on
#'   the same cells.
#' @param multi_range A further set of ranges the rule also covers, as an Excel
#'   multi-range string such as `"B3:K6 B9:K12"`.
#' @return An `xl_conditional` object.
#' @family worksheet features
#' @seealso [xl_sheet], [xl_format]
#' @name xl_conditional
#' @examples
#' xl_cond_cell("B2:B100", criteria = ">", value = 100, format = xl_fill(background = "red"))
#' xl_cond_cell("C2:C100", type = "text", criteria = "contains", value = "urgent",
#'              format = xl_font(bold = TRUE))
#' xl_cond_cell("D2:D100", type = "duplicate",
#'              format = xl_fill(background = "yellow"))
NULL

#' @rdname xl_conditional
#' @export
xl_cond_cell <- function(range, type = NA, criteria = NA, value = NULL,
                         min = NULL, max = NULL, format = NULL,
                         stop_if_true = NA, multi_range = NA) {
  # the range is the subject of the call, so check it before anything else
  if (missing(range) || is.null(range))
    stop("`range` must name the cells to format", call. = FALSE)
  crit <- .val_enum(criteria, names(.LXW_COND_CRITERIA)[-1L], "criteria")
  kind <- .val_enum(type, names(.CONDITIONAL_CRITERIA_FOR), "type")
  kind <- .cond_infer_type(kind, crit)
  .check_conditional_criteria(kind, crit)

  if (.cond_is_between(crit) && (is.null(min) || is.null(max)))
    stop(sprintf("`criteria = \"%s\"` needs both `min` and `max`", crit),
         call. = FALSE)
  if (!is.null(crit) && !.cond_is_between(crit) && is.null(value) &&
      kind %in% c("cell", "text"))
    stop(sprintf("`criteria = \"%s\"` needs a `value` to compare against", crit),
         call. = FALSE)
  if (kind == "formula" && is.null(value))
    stop("type = \"formula\" needs a `value` holding the formula", call. = FALSE)
  if (!is.null(format) && !is_xl_format(format))
    stop("`format` must be an xl_format object", call. = FALSE)

  .new_conditional("cell", range,
                   .drop_null(list(type = kind, criteria = crit, value = value,
                                   min = min, max = max, format = format)),
                   stop_if_true, multi_range)
}

#' @export
print.xl_conditional <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_conditional: %s on %s>\n", p$cluster,
              if (is.character(p$range)) p$range else "<spec>"))
  invisible(x)
}

# One limit, flattened into the (number, string) pair the struct carries.  A
# string is either an Excel formula or the literal text a text-rule matches;
# libxlsxwriter takes both through value_string.
.cond_value <- function(x, prefix) {
  if (is.null(x)) return(list())
  if (is.character(x))
    return(stats::setNames(list(as.character(x)), paste0(prefix, "_string")))
  stats::setNames(list(as.numeric(x)), prefix)
}

# Resolve a sheet's conditional formats to the overlay payloads C applies.
.resolve_conditionals <- function(el, df, reg, header_offset, props) {
  if (!inherits(el, "xl_sheet") || is.null(el$conditional)) return(list())
  cs <- if (inherits(el$conditional, "xl_conditional")) list(el$conditional)
        else el$conditional
  if (!is.list(cs))
    stop("`conditional` must be an xl_conditional object or a list of them",
         call. = FALSE)
  lapply(seq_along(cs), function(i) {
    cf <- cs[[i]]
    if (!inherits(cf, "xl_conditional"))
      stop(sprintf("`conditional[[%d]]` must be an xl_conditional object", i),
           call. = FALSE)
    p <- unclass(cf)
    q <- .xl_resolve_range(p$range, arg = sprintf("conditional[[%d]] range", i),
                           df = df, header_offset = header_offset,
                           allow_cell = TRUE)
    out <- list(kind = "conditional", range = as.integer(q))

    if (identical(p$cluster, "cell")) {
      out$type <- unname(.LXW_COND_TYPE[[p$type]])
      if (!is.null(p$criteria))
        out$criteria <- unname(.LXW_COND_CRITERIA[[p$criteria]])
      out <- c(out, .cond_value(p$value, "value"),
               .cond_value(p$min, "min_value"),
               .cond_value(p$max, "max_value"))
      # the format is an ordinary xl_format, so it goes through the same
      # registry as every other format -- libxlsxwriter emits it as a
      # differential format (<dxf>) rather than a cell style
      out$format_id <- .register_format(reg,
                                        merge_xl_format(props$default_format,
                                                        p$format))
    } else {
      # scale, bar and icon clusters: the constructor already produced the
      # flat field names the struct uses, so they pass straight through with
      # only the type name resolved to its enum value
      flat <- p[setdiff(names(p), c("cluster", "range", "type",
                                    "stop_if_true", "multi_range"))]
      out$type <- unname(.LXW_COND_TYPE[[p$type]])
      out <- c(out, flat)
    }
    for (k in c("stop_if_true"))
      if (!is.null(p[[k]])) out[[k]] <- as.integer(isTRUE(p[[k]]))
    if (!is.null(p$multi_range)) out$multi_range <- p$multi_range
    out
  })
}

.LXW_BAR_DIRECTION <- c(context = 0L, "left to right" = 1L, "right to left" = 2L)
.LXW_BAR_AXIS      <- c(automatic = 0L, midpoint = 1L, none = 2L)

# The min/mid/max triple a colour scale or data bar uses.  Each point carries a
# rule type ("what kind of number is this"), a value and, for scales, a colour.
.cond_point <- function(value, rule_type, color, prefix, arg) {
  out <- list()
  rt <- .val_enum(rule_type, names(.LXW_COND_RULE_TYPE)[-1L],
                  paste0(arg, "_rule_type"))
  if (!is.null(rt)) out[[paste0(prefix, "_rule_type")]] <- unname(.LXW_COND_RULE_TYPE[[rt]])
  if (!is.null(value)) {
    if (is.character(value)) out[[paste0(prefix, "_value_string")]] <- value
    else                     out[[paste0(prefix, "_value")]] <- as.numeric(value)
  }
  if (!is.null(color) && !(length(color) == 1L && is.na(color)))
    out[[paste0(prefix, "_color")]] <- xl_color(color)
  out
}

#' @rdname xl_conditional
#' @param colors Two or three colours for the scale, from lowest to highest.
#'   Two gives a two-colour scale, three a three-colour scale.
#' @param values Optional values marking where each colour sits, in the same
#'   order as `colors`.  Defaults to the range's minimum, midpoint and maximum.
#' @param rule_types How each entry of `values` is interpreted: `"minimum"`,
#'   `"maximum"`, `"number"`, `"percent"`, `"percentile"`, `"formula"`,
#'   `"auto_min"` or `"auto_max"`.
#' @export
#' @examples
#' xl_cond_scale("C2:C100", colors = c("red", "yellow", "green"))
#' xl_cond_scale("C2:C100", colors = c("white", "steelblue"))
xl_cond_scale <- function(range, colors = c("red", "yellow", "green"),
                          values = NULL, rule_types = NULL,
                          stop_if_true = NA, multi_range = NA) {
  if (missing(range) || is.null(range))
    stop("`range` must name the cells to format", call. = FALSE)
  n <- length(colors)
  if (!n %in% c(2L, 3L))
    stop("`colors` must give 2 or 3 colours (Excel has no other colour scale)",
         call. = FALSE)
  if (!is.null(values) && length(values) != n)
    stop(sprintf("`values` must have one entry per colour (%d)", n),
         call. = FALSE)
  if (!is.null(rule_types) && length(rule_types) != n)
    stop(sprintf("`rule_types` must have one entry per colour (%d)", n),
         call. = FALSE)
  # a two-colour scale uses min and max; a three-colour one adds mid
  prefixes <- if (n == 2L) c("min", "max") else c("min", "mid", "max")
  fields <- list(type = "2_color_scale")
  if (n == 3L) fields$type <- "3_color_scale"
  for (i in seq_len(n))
    fields <- c(fields, .cond_point(values[[i]], rule_types[[i]],
                                    colors[[i]], prefixes[i], prefixes[i]))
  .new_conditional("scale", range, fields, stop_if_true, multi_range)
}

#' @rdname xl_conditional
#' @param color The bar's fill colour.
#' @param solid Logical; a solid bar rather than Excel's default gradient.
#' @param negative_color,border_color,negative_border_color,axis_color Colours
#'   for the negative portion, the bar border, the negative border, and the
#'   axis line.
#' @param no_border Logical; draw the bar without a border.
#' @param direction `"context"` (follow the sheet), `"left to right"` or
#'   `"right to left"`.
#' @param axis Where the zero axis sits: `"automatic"`, `"midpoint"` or
#'   `"none"`.
#' @param bar_only Logical; show the bar without the cell's value.
#' @export
#' @examples
#' xl_cond_bar("D2:D100", color = "steelblue")
#' xl_cond_bar("D2:D100", color = "green", solid = TRUE, bar_only = TRUE)
xl_cond_bar <- function(range, color = NA, values = NULL, rule_types = NULL,
                        solid = NA, negative_color = NA, border_color = NA,
                        negative_border_color = NA, no_border = NA,
                        direction = NA, axis = NA, axis_color = NA,
                        bar_only = NA, stop_if_true = NA, multi_range = NA) {
  if (missing(range) || is.null(range))
    stop("`range` must name the cells to format", call. = FALSE)
  if (!is.null(values) && length(values) != 2L)
    stop("`values` must give the two ends of the bar", call. = FALSE)
  if (!is.null(rule_types) && length(rule_types) != 2L)
    stop("`rule_types` must give the two ends of the bar", call. = FALSE)
  fields <- list(type = "data_bar")
  for (i in 1:2) {
    pre <- c("min", "max")[i]
    fields <- c(fields, .cond_point(values[[i]], rule_types[[i]],
                                    NA, pre, pre))
  }
  cols <- .drop_null(list(
    bar_color                 = if (!.is_unset(color)) xl_color(color),
    bar_negative_color        = if (!.is_unset(negative_color)) xl_color(negative_color),
    bar_border_color          = if (!.is_unset(border_color)) xl_color(border_color),
    bar_negative_border_color = if (!.is_unset(negative_border_color))
                                  xl_color(negative_border_color),
    bar_axis_color            = if (!.is_unset(axis_color)) xl_color(axis_color)
  ))
  flags <- .drop_null(list(
    bar_solid     = .val_flag(solid, "solid"),
    bar_no_border = .val_flag(no_border, "no_border"),
    bar_only      = .val_flag(bar_only, "bar_only")
  ))
  d <- .val_enum(direction, names(.LXW_BAR_DIRECTION), "direction")
  a <- .val_enum(axis, names(.LXW_BAR_AXIS), "axis")
  if (!is.null(d)) fields$bar_direction     <- unname(.LXW_BAR_DIRECTION[[d]])
  if (!is.null(a)) fields$bar_axis_position <- unname(.LXW_BAR_AXIS[[a]])
  .new_conditional("bar", range, c(fields, cols, flags), stop_if_true,
                   multi_range)
}

#' @rdname xl_conditional
#' @param style Which built-in icon set: one of `"3_arrows"`, `"3_arrows_gray"`,
#'   `"3_flags"`, `"3_traffic_lights"`, `"3_traffic_lights_rimmed"`,
#'   `"3_signs"`, `"3_symbols_circled"`, `"3_symbols"`, `"4_arrows"`,
#'   `"4_arrows_gray"`, `"4_red_to_black"`, `"4_ratings"`,
#'   `"4_traffic_lights"`, `"5_arrows"`, `"5_arrows_gray"`, `"5_ratings"` or
#'   `"5_quarters"`.  These are Excel's own icons, not images, so nothing is
#'   embedded in the file.
#' @param reverse Logical; reverse the order the icons are assigned in.
#' @param icons_only Logical; show the icon without the cell's value.
#' @export
#' @examples
#' xl_cond_icons("E2:E100", style = "3_traffic_lights")
#' xl_cond_icons("E2:E100", style = "5_ratings", icons_only = TRUE)
xl_cond_icons <- function(range, style = "3_traffic_lights", reverse = NA,
                          icons_only = NA, stop_if_true = NA,
                          multi_range = NA) {
  if (missing(range) || is.null(range))
    stop("`range` must name the cells to format", call. = FALSE)
  st <- .val_enum(style, names(.LXW_COND_ICONS), "style")
  if (is.null(st)) st <- "3_traffic_lights"
  fields <- .drop_null(list(
    type          = "icon_sets",
    icon_style    = unname(.LXW_COND_ICONS[[st]]),
    reverse_icons = .val_flag(reverse, "reverse"),
    icons_only    = .val_flag(icons_only, "icons_only")
  ))
  .new_conditional("icons", range, fields, stop_if_true, multi_range)
}

# Conditional formatting.  The rules land in <conditionalFormatting> in the
# worksheet part, and their formats in <dxfs> in styles.xml -- differential
# formats, not the ordinary cell styles.

cf <- function(..., df = data.frame(v = c(50, 150, 250))) {
  x <- xlsx_part(write_tmp(list(D = xl_sheet(df, conditional = list(...)))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  m <- regmatches(x, regexpr("<conditionalFormatting.*</conditionalFormatting>", x))
  if (!length(m)) NA_character_ else m
}

cf_styles <- function(...) {
  p <- write_tmp(list(D = xl_sheet(data.frame(v = c(50, 150, 250)),
                                   conditional = list(...))))
  s <- xlsx_part(p, "xl/styles.xml", raw = TRUE)
  m <- regmatches(s, regexpr("<dxfs.*?</dxfs>", s))
  if (!length(m)) NA_character_ else m
}

# ── Simple cell rules ─────────────────────────────────────────────────────────

test_that("a comparison rule writes cellIs with an operator", {
  x <- cf(xl_cond_cell("A2:A4", criteria = ">", value = 100,
                       format = xl_fill(background = "red")))
  expect_match(x, 'type="cellIs"', fixed = TRUE)
  expect_match(x, 'operator="greaterThan"', fixed = TRUE)
  expect_match(x, "<formula>100</formula>", fixed = TRUE)
  expect_match(x, 'sqref="A2:A4"', fixed = TRUE)
})

test_that("each comparison criteria maps to its Excel operator", {
  ops <- c("==" = "equal", "!=" = "notEqual", ">" = "greaterThan",
           "<" = "lessThan", ">=" = "greaterThanOrEqual",
           "<=" = "lessThanOrEqual")
  for (k in names(ops))
    expect_match(cf(xl_cond_cell("A2:A4", criteria = k, value = 100)),
                 sprintf('operator="%s"', ops[[k]]), fixed = TRUE, label = k)
})

test_that("between uses both bounds", {
  x <- cf(xl_cond_cell("A2:A4", criteria = "between", min = 10, max = 20))
  expect_match(x, 'operator="between"', fixed = TRUE)
  expect_match(x, "<formula>10</formula>", fixed = TRUE)
  expect_match(x, "<formula>20</formula>", fixed = TRUE)
})

test_that("the type is inferred from the criteria", {
  # a text criteria implies a text rule without naming the type
  expect_match(cf(xl_cond_cell("A2:A4", criteria = "contains", value = "urgent")),
               'type="containsText"', fixed = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", criteria = "today")),
               'type="timePeriod"', fixed = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", criteria = "above")),
               'type="aboveAverage"', fixed = TRUE)
})

test_that("the rules that need no criteria are written", {
  for (ty in c("duplicate", "unique", "blanks", "no_blanks", "errors",
               "no_errors"))
    expect_match(cf(xl_cond_cell("A2:A4", type = ty)), "<cfRule", fixed = TRUE,
                 label = ty)
  expect_match(cf(xl_cond_cell("A2:A4", type = "duplicate")),
               'type="duplicateValues"', fixed = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", type = "unique")),
               'type="uniqueValues"', fixed = TRUE)
})

test_that("a formula rule carries its formula", {
  x <- cf(xl_cond_cell("A2:A4", type = "formula", value = "=$A2>100"))
  expect_match(x, 'type="expression"', fixed = TRUE)
  expect_match(x, "$A2&gt;100", fixed = TRUE)
})

test_that("top and bottom rules are written, with and without percent", {
  expect_match(cf(xl_cond_cell("A2:A4", type = "top", value = 10)),
               'type="top10"', fixed = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", criteria = "percent", value = 10, type = "top")),
               'percent="1"', fixed = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", type = "bottom", value = 5)),
               'bottom="1"', fixed = TRUE)
})

# ── Formats land in dxfs ──────────────────────────────────────────────────────

test_that("a rule's format becomes a differential format, not a cell style", {
  # this is the part worth pinning: conditional formats use <dxfs>, so the
  # shared format registry has to reach somewhere other than cellXfs
  d <- cf_styles(xl_cond_cell("A2:A4", criteria = ">", value = 100,
                              format = xl_fill(background = "red")))
  expect_match(d, "<dxf>", fixed = TRUE)
  expect_match(d, "FFFF0000", ignore.case = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", criteria = ">", value = 100,
                               format = xl_fill(background = "red"))),
               'dxfId="0"', fixed = TRUE)
})

test_that("several rules get distinct dxf ids and ascending priorities", {
  x <- cf(xl_cond_cell("A2:A4", criteria = ">", value = 100, format = xl_fill(background = "red")),
          xl_cond_cell("A2:A4", type = "duplicate",
                       format = xl_font(bold = TRUE)))
  expect_match(x, 'dxfId="0"', fixed = TRUE)
  expect_match(x, 'dxfId="1"', fixed = TRUE)
  expect_match(x, 'priority="1"', fixed = TRUE)
  expect_match(x, 'priority="2"', fixed = TRUE)
  expect_match(cf_styles(
    xl_cond_cell("A2:A4", criteria = ">", value = 100, format = xl_fill(background = "red")),
    xl_cond_cell("A2:A4", type = "duplicate", format = xl_font(bold = TRUE))),
    'count="2"', fixed = TRUE)
})

test_that("a rule with no format is still written", {
  expect_match(cf(xl_cond_cell("A2:A4", criteria = ">", value = 100)), "<cfRule", fixed = TRUE)
})

# ── The type / criteria pairing ───────────────────────────────────────────────

test_that("a criteria illegal for its type is refused, naming both", {
  # libxlsxwriter accepts these and Excel then ignores the rule, so the check
  # has to be ours
  expect_error(xl_cond_cell("A2:A4", criteria = "contains", value = "x", type = "cell"),
               'not valid for type = "cell"')
  expect_error(xl_cond_cell("A2:A4", criteria = ">", value = 1, type = "text"),
               'not valid for type = "text"')
  expect_error(xl_cond_cell("A2:A4", criteria = "today", type = "average"),
               'not valid for type = "average"')
  # the message lists what would be valid *for that type*
  expect_error(xl_cond_cell("A2:A4", criteria = "contains", value = "x", type = "cell"),
               "not between", fixed = TRUE)
  expect_error(xl_cond_cell("A2:A4", criteria = ">", value = 1, type = "text"),
               "begins with", fixed = TRUE)
})

test_that("criteria is refused for the rules that carry their own meaning", {
  for (ty in c("duplicate", "unique", "blanks", "errors", "formula"))
    expect_error(xl_cond_cell("A2:A4", criteria = ">", value = 1, type = ty),
                 "does not apply", label = ty)
})

test_that("criteria is required where it decides the rule", {
  for (ty in c("cell", "text", "time_period", "average"))
    expect_error(xl_cond_cell("A2:A4", type = ty, value = 1),
                 "`criteria` is required", label = ty)
})

test_that("between needs both bounds and comparisons need a value", {
  expect_error(xl_cond_cell("A2:A4", criteria = "between", min = 1),
               "needs both `min` and `max`")
  expect_error(xl_cond_cell("A2:A4", criteria = ">"), "needs a `value`")
  expect_error(xl_cond_cell("A2:A4", type = "formula"), "needs a `value`")
})

test_that("bad enum values and ranges are rejected", {
  expect_error(xl_cond_cell("A2:A4", criteria = "≈", value = 1), "criteria")
  expect_error(xl_cond_cell("A2:A4", criteria = ">", value = 1, type = "rainbow"), "type")
  expect_error(xl_cond_cell(), "must name the cells")
  expect_error(xl_cond_cell("A2:A4", criteria = ">", value = 1, format = "red"),
               "must be an xl_format")
})

# ── Wiring ────────────────────────────────────────────────────────────────────

test_that("a single rule need not be wrapped in a list", {
  x <- xlsx_part(write_tmp(list(D = xl_sheet(data.frame(v = c(1, 2, 3)),
    conditional = xl_cond_cell("A2:A4", criteria = ">", value = 1)))),
    "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(x, "<cfRule", fixed = TRUE)
})

test_that("non-conditional objects are rejected", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(v = 1),
                                           conditional = "A1:B2"))),
               "must be an xl_conditional object")
})

test_that("stop_if_true and multi_range reach the file", {
  expect_match(cf(xl_cond_cell("A2:A4", criteria = ">", value = 100, stop_if_true = TRUE)),
               'stopIfTrue="1"', fixed = TRUE)
  expect_match(cf(xl_cond_cell("A2:A4", criteria = ">", value = 100, multi_range = "A2:A4 C2:C4")),
               'sqref="A2:A4 C2:C4"', fixed = TRUE)
})

test_that("a rule accepts a data-frame-relative spec", {
  df <- data.frame(v = c(1, 2, 3))
  x <- cf(xl_cond_cell(list(rows = 1:3, cols = "v"), criteria = ">", value = 1), df = df)
  expect_match(x, 'sqref="A2:A4"', fixed = TRUE)
})

test_that("conditional formatting leaves the cells alone and keeps streaming", {
  skip_if_not_installed("readxl")
  df <- data.frame(v = c(50, 150, 250))
  p <- write_tmp(list(D = xl_sheet(df,
    conditional = xl_cond_cell("A2:A4", criteria = ">", value = 100,
                               format = xl_fill(background = "red")))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
  # written at packaging time, so row streaming stays on when asked for
  expect_true(is.null(xlsx_part(
    write_tmp(list(D = xl_sheet(df, conditional = xl_cond_cell(
      "A2:A4", criteria = ">", value = 100,
      format = xl_fill(background = "red")))), constant_memory = TRUE),
    "xl/sharedStrings.xml")))
})

test_that("the print method runs", {
  expect_output(print(xl_cond_cell("A2:A4", criteria = ">", value = 1)), "xl_conditional")
})

# ── Colour scales ─────────────────────────────────────────────────────────────

test_that("a three-colour scale writes three stops and three colours", {
  x <- cf(xl_cond_scale("A2:A4", colors = c("red", "yellow", "green")))
  expect_match(x, 'type="colorScale"', fixed = TRUE)
  expect_match(x, "FFFF0000", ignore.case = TRUE)   # red
  expect_match(x, "FFFFFF00", ignore.case = TRUE)   # yellow
  expect_match(x, "FF00FF00", ignore.case = TRUE)   # green
  expect_equal(lengths(regmatches(x, gregexpr("<cfvo ", x))), 3L)
})

test_that("a two-colour scale writes two stops", {
  x <- cf(xl_cond_scale("A2:A4", colors = c("white", "blue")))
  expect_match(x, 'type="colorScale"', fixed = TRUE)
  expect_equal(lengths(regmatches(x, gregexpr("<cfvo ", x))), 2L)
})

test_that("scale stops accept values and rule types", {
  x <- cf(xl_cond_scale("A2:A4", colors = c("red", "green"),
                        values = c(10, 90),
                        rule_types = c("number", "number")))
  expect_match(x, 'type="num"', fixed = TRUE)
  expect_match(x, 'val="10"', fixed = TRUE)
  expect_match(x, 'val="90"', fixed = TRUE)
})

test_that("a scale needs two or three colours", {
  expect_error(xl_cond_scale("A2:A4", colors = "red"), "2 or 3 colours")
  expect_error(xl_cond_scale("A2:A4", colors = c("a", "b", "c", "d")),
               "2 or 3 colours")
  expect_error(xl_cond_scale("A2:A4", colors = c("red", "green"),
                             values = 1), "one entry per colour")
  expect_error(xl_cond_scale("A2:A4", colors = c("red", "green"),
                             rule_types = "number"), "one entry per colour")
  expect_error(xl_cond_scale(), "must name the cells")
})

# ── Data bars ─────────────────────────────────────────────────────────────────

test_that("a data bar is written with its colour", {
  x <- cf(xl_cond_bar("A2:A4", color = "steelblue"))
  expect_match(x, 'type="dataBar"', fixed = TRUE)
  expect_match(x, "FF4682B4", ignore.case = TRUE)
})

test_that("bar options reach the file", {
  expect_match(cf(xl_cond_bar("A2:A4", color = "green", bar_only = TRUE)),
               "dataBar", fixed = TRUE)
  # the 2010 extensions carry the solid/direction/axis settings
  x <- cf(xl_cond_bar("A2:A4", color = "green", solid = TRUE,
                      direction = "right to left", axis = "midpoint"))
  expect_match(x, "x14:", fixed = TRUE)
})

test_that("bar ends accept values and rule types", {
  x <- cf(xl_cond_bar("A2:A4", color = "green", values = c(0, 100),
                      rule_types = c("number", "number")))
  expect_match(x, 'type="num"', fixed = TRUE)
})

test_that("bar arguments are validated", {
  expect_error(xl_cond_bar("A2:A4", values = 1), "two ends")
  expect_error(xl_cond_bar("A2:A4", rule_types = "number"), "two ends")
  expect_error(xl_cond_bar("A2:A4", direction = "sideways"), "direction")
  expect_error(xl_cond_bar("A2:A4", axis = "diagonal"), "axis")
  expect_error(xl_cond_bar(), "must name the cells")
})

# ── Icon sets ─────────────────────────────────────────────────────────────────

test_that("an icon set is written", {
  expect_match(cf(xl_cond_icons("A2:A4", style = "3_traffic_lights")),
               'type="iconSet"', fixed = TRUE)
})

test_that("every icon style maps to its Excel name", {
  # the style is an index into libxlsxwriter's enum, and three of the
  # conditional type constants lack the TYPE_ prefix -- getting the offset
  # wrong would silently write a colour scale instead, so enumerate them all
  expected <- c(
    "3_arrows" = "3Arrows", "3_arrows_gray" = "3ArrowsGray",
    "3_flags" = "3Flags", "3_traffic_lights" = NA,   # the default: no attribute
    "3_traffic_lights_rimmed" = "3TrafficLights2", "3_signs" = "3Signs",
    "3_symbols_circled" = "3Symbols", "3_symbols" = "3Symbols2",
    "4_arrows" = "4Arrows", "4_arrows_gray" = "4ArrowsGray",
    "4_red_to_black" = "4RedToBlack", "4_ratings" = "4Rating",
    "4_traffic_lights" = "4TrafficLights", "5_arrows" = "5Arrows",
    "5_arrows_gray" = "5ArrowsGray", "5_ratings" = "5Rating",
    "5_quarters" = "5Quarters")
  for (st in names(expected)) {
    x <- cf(xl_cond_icons("A2:A4", style = st))
    if (is.na(expected[[st]]))
      expect_match(x, "<iconSet>", fixed = TRUE, label = st)
    else
      expect_match(x, sprintf('iconSet="%s"', expected[[st]]), fixed = TRUE,
                   label = st)
  }
})

test_that("all 17 icon styles are covered by the map", {
  # a newly added style in a libxlsxwriter update should fail here rather than
  # be silently unreachable from R
  expect_equal(length(.LXW_COND_ICONS), 17L)
})

test_that("icon options are written", {
  expect_match(cf(xl_cond_icons("A2:A4", icons_only = TRUE)),
               'showValue="0"', fixed = TRUE)
  expect_match(cf(xl_cond_icons("A2:A4", reverse = TRUE)),
               'reverse="1"', fixed = TRUE)
})

test_that("icon arguments are validated", {
  expect_error(xl_cond_icons("A2:A4", style = "6_smileys"), "style")
  expect_error(xl_cond_icons(), "must name the cells")
})

test_that("icon sets embed no image parts", {
  # Excel draws these itself, so nothing goes into xl/media
  p <- write_tmp(list(D = xl_sheet(data.frame(v = c(10, 50, 90)),
                                   conditional = xl_cond_icons("A2:A4"))))
  d <- tempfile(); dir.create(d); utils::unzip(p, exdir = d)
  expect_false(dir.exists(file.path(d, "xl", "media")))
})

# ── The four clusters together ────────────────────────────────────────────────

test_that("all four kinds coexist on one sheet", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = c(10, 50, 90), b = c(10, 50, 90),
                   cc = c(10, 50, 90), d = c(10, 50, 90))
  p <- write_tmp(list(D = xl_sheet(df, conditional = list(
    xl_cond_cell("A2:A4", criteria = ">", value = 20,
                 format = xl_fill(background = "red")),
    xl_cond_scale("B2:B4"),
    xl_cond_bar("C2:C4", color = "steelblue"),
    xl_cond_icons("D2:D4")))))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  for (ty in c("cellIs", "colorScale", "dataBar", "iconSet"))
    expect_match(x, sprintf('type="%s"', ty), fixed = TRUE, label = ty)
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
})

# ── The branches nothing else reached ────────────────────────────────────────

test_that("a rule type is inferred from the criteria when none is given", {
  expect_equal(.cond_infer_type(NULL, NULL), "cell")
  expect_equal(.cond_infer_type(NULL, "contains"), "text")
  expect_equal(.cond_infer_type(NULL, "yesterday"), "time_period")
  expect_equal(.cond_infer_type(NULL, "above"), "average")
  expect_equal(.cond_infer_type(NULL, ">"), "cell")
  # an explicit type always wins
  expect_equal(.cond_infer_type("duplicate", "contains"), "duplicate")
  # a type with no criteria table of its own is left alone
  expect_null(.check_conditional_criteria("duplicate", NULL))
})

test_that("a colour scale takes string thresholds as well as numbers", {
  p <- write_tmp(list(D = xl_sheet(
    data.frame(a = 1:5),
    conditional = xl_cond_scale("A2:A6", colors = c("red", "green"),
                                rule_types = c("formula", "formula"),
                                values = c("=MIN($A$2:$A$6)",
                                           "=MAX($A$2:$A$6)")))))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(x, "MIN($A$2:$A$6)", fixed = TRUE)
  expect_match(x, "colorScale", fixed = TRUE)
})

test_that("an icon set defaults to three traffic lights", {
  plain <- write_tmp(list(D = xl_sheet(data.frame(a = 1:5),
                                       conditional = xl_cond_icons("A2:A6"))))
  named <- write_tmp(list(D = xl_sheet(
    data.frame(a = 1:5),
    conditional = xl_cond_icons("A2:A6", style = "3_traffic_lights"))))
  expect_equal(xlsx_part(plain, "xl/worksheets/sheet1.xml", raw = TRUE),
               xlsx_part(named, "xl/worksheets/sheet1.xml", raw = TRUE))
})

test_that("conditional formatting names what was wrong", {
  expect_error(xl_cond_icons(NULL), "must name the cells to format")
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                           conditional = list("nope")))),
               "must be an xl_conditional object")
})

test_that("a rule type with no criteria of its own accepts none", {
  # blanks, duplicates, errors and the rest take no criteria at all
  for (ty in c("duplicate", "unique", "blanks", "no_blanks", "errors",
               "no_errors"))
    expect_null(.check_conditional_criteria(ty, NULL))
  # and a type outside the table is left alone entirely
  expect_null(.check_conditional_criteria("2_color_scale", "anything"))
})

test_that("an icon style of NA falls back to three traffic lights", {
  a <- write_tmp(list(D = xl_sheet(data.frame(a = 1:5),
                                   conditional = xl_cond_icons("A2:A6",
                                                               style = NA))))
  b <- write_tmp(list(D = xl_sheet(data.frame(a = 1:5),
                                   conditional = xl_cond_icons("A2:A6"))))
  expect_equal(xlsx_part(a, "xl/worksheets/sheet1.xml", raw = TRUE),
               xlsx_part(b, "xl/worksheets/sheet1.xml", raw = TRUE))
})

test_that("every conditional constructor insists on a range", {
  expect_error(xl_cond_cell(NULL, type = "duplicate"),
               "must name the cells to format")
  expect_error(xl_cond_scale(NULL), "must name the cells to format")
  expect_error(xl_cond_bar(NULL), "must name the cells to format")
})

test_that("every data bar colour reaches the file", {
  p <- write_tmp(list(D = xl_sheet(data.frame(a = c(-5, 3, 8)),
    conditional = xl_cond_bar("A2:A4", color = "#0000FF",
                              negative_color = "#FF0000",
                              border_color = "#00FF00",
                              negative_border_color = "#FFFF00",
                              axis_color = "#FF00FF"))))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  for (col in c("0000FF", "FF0000", "00FF00", "FFFF00", "FF00FF"))
    expect_match(x, col, fixed = TRUE)
})

# =============================================================================
# Outline (grouping) display, and the error indicators Excel shows in cells
# =============================================================================
#
# Two small worksheet-level settings that share nothing but their size.
#
# Outline settings control how the grouping controls that xl_col_spec(level =)
# and xl_row_spec(level =) create are *drawn* -- not which rows are grouped.
#
# Ignoring errors turns off the green triangle Excel puts in a cell it thinks
# is wrong.  The common case is a number deliberately written as text: Excel
# flags it, and there is no other way to silence that.
# -----------------------------------------------------------------------------

#' Control how outline (grouping) symbols are drawn
#'
#' @description
#' Grouping itself comes from `level` in [xl_col_spec()] / [xl_row_spec()].
#' `xl_outline()` only changes how the controls are *displayed*, which is
#' rarely needed --- the defaults match Excel's own.
#'
#' @param visible Show the outline symbols at all.  `FALSE` keeps the grouping
#'   (and so the collapsing) but hides the +/- controls.
#' @param symbols_below Put the summary row *below* the detail rows, which is
#'   Excel's default.  `FALSE` puts it above.
#' @param symbols_right Put the summary column to the *right* of the detail
#'   columns, Excel's default.  `FALSE` puts it to the left.
#' @param auto_style Apply Excel's automatic outline styling to the grouped
#'   rows and columns.
#' @return An `xl_outline` object, for [xl_sheet()]'s `outline` argument.
#' @family worksheet layout
#' @seealso [xl_sheet], [xl_colrow_spec]
#' @export
#' @examples
#' xl_outline(symbols_below = FALSE)
#' xl_outline(visible = FALSE)
xl_outline <- function(visible = TRUE, symbols_below = TRUE,
                       symbols_right = TRUE, auto_style = FALSE) {
  flag <- function(x, arg) {
    if (!is.logical(x) || length(x) != 1L || is.na(x))
      stop(sprintf("`%s` must be TRUE or FALSE", arg), call. = FALSE)
    as.integer(x)
  }
  structure(list(visible = flag(visible, "visible"),
                 symbols_below = flag(symbols_below, "symbols_below"),
                 symbols_right = flag(symbols_right, "symbols_right"),
                 auto_style = flag(auto_style, "auto_style")),
            class = "xl_outline")
}

#' @export
print.xl_outline <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_outline: visible=%s below=%s right=%s auto_style=%s>\n",
              p$visible == 1L, p$symbols_below == 1L, p$symbols_right == 1L,
              p$auto_style == 1L))
  invisible(x)
}

.resolve_outline <- function(el) {
  if (!inherits(el, "xl_sheet") || is.null(el$outline)) return(NULL)
  if (!inherits(el$outline, "xl_outline"))
    stop("`outline` must be an xl_outline object", call. = FALSE)
  unclass(el$outline)
}

# The error indicators Excel can be told to stop showing.  The names mirror
# libxlsxwriter's LXW_IGNORE_* constants, lowercased.
.LXW_IGNORE_TYPE <- c(
  number_stored_as_text = 1L, eval_error = 2L, formula_differs = 3L,
  formula_range = 4L, formula_unlocked = 5L, empty_cell_reference = 6L,
  list_data_validation = 7L, calculated_column = 8L, two_digit_text_year = 9L
)

# Resolve xl_sheet(ignore_errors =) to the C payloads.  Each element is named
# for the error type and holds a range, in any form the shared resolver takes.
.resolve_ignore_errors <- function(el, df, header_offset) {
  if (!inherits(el, "xl_sheet") || is.null(el$ignore_errors)) return(list())
  ie <- el$ignore_errors
  if (!is.list(ie)) ie <- as.list(ie)
  nms <- names(ie)
  if (is.null(nms) || any(!nzchar(nms)))
    stop("`ignore_errors` must be a named list, each name an error type ",
         "and each value a range", call. = FALSE)
  bad <- setdiff(nms, names(.LXW_IGNORE_TYPE))
  if (length(bad))
    stop(sprintf("`ignore_errors`: unknown error type(s): %s.\n  Available: %s",
                 paste(bad, collapse = ", "),
                 paste(names(.LXW_IGNORE_TYPE), collapse = ", ")),
         call. = FALSE)
  if (anyDuplicated(nms))
    stop("`ignore_errors`: each error type may be given only once; combine ",
         "the ranges into one entry", call. = FALSE)
  lapply(seq_along(ie), function(i) {
    rng <- .xl_resolve_range(ie[[i]], arg = sprintf("ignore_errors$%s", nms[i]),
                             df = df, header_offset = header_offset,
                             allow_cell = TRUE)
    list(kind = "ignore_errors",
         type = unname(.LXW_IGNORE_TYPE[[nms[i]]]),
         range = as.integer(rng))
  })
}

# =============================================================================
# xl_format: the shared cell/row/column/workbook formatting engine
# =============================================================================
#
# An `xl_format` is a simple S3 object: a list with a class.  Each of its slots
# is either absent (unset) or a validated named list describing one group of
# formatting properties:
#
#   font        - font_name/size/color/bold/italic/underline/...
#   fill        - background/foreground/pattern
#   border      - per-side styles and colors, diagonal
#   align       - horizontal/vertical/wrap/rotation/indent/shrink/reading_order
#   num_format  - number-format string or built-in index
#   protection  - locked/hidden
#
# plus the scalar slots `quote_prefix` and `hyperlink`.
#
# Human-readable values (enum strings, R colors) are stored in the object so it
# prints legibly; conversion to the C-ready integer payload libxlsxwriter needs
# happens once, at write time, in `.xl_format_payload()`.
# -----------------------------------------------------------------------------

# Enum string -> libxlsxwriter integer maps.  These MUST match the values in
# src/libxlsxwriter/include/xlsxwriter/format.h.
.LXW <- list(
  underline = c("none" = 0L, "single" = 1L, "double" = 2L,
                "single-accounting" = 3L, "double-accounting" = 4L),
  script    = c("super" = 1L, "sub" = 2L),
  align_h   = c("none" = 0L, "left" = 1L, "center" = 2L, "right" = 3L,
                "fill" = 4L, "justify" = 5L, "center-across" = 6L,
                "distributed" = 7L),
  align_v   = c("top" = 8L, "bottom" = 9L, "center" = 10L, "justify" = 11L,
                "distributed" = 12L),
  pattern   = c("none" = 0L, "solid" = 1L, "medium-gray" = 2L, "dark-gray" = 3L,
                "light-gray" = 4L, "dark-horizontal" = 5L, "dark-vertical" = 6L,
                "dark-down" = 7L, "dark-up" = 8L, "dark-grid" = 9L,
                "dark-trellis" = 10L, "light-horizontal" = 11L,
                "light-vertical" = 12L, "light-down" = 13L, "light-up" = 14L,
                "light-grid" = 15L, "light-trellis" = 16L, "gray-125" = 17L,
                "gray-0625" = 18L),
  border    = c("none" = 0L, "thin" = 1L, "medium" = 2L, "dashed" = 3L,
                "dotted" = 4L, "thick" = 5L, "double" = 6L, "hair" = 7L,
                "medium-dashed" = 8L, "dash-dot" = 9L, "medium-dash-dot" = 10L,
                "dash-dot-dot" = 11L, "medium-dash-dot-dot" = 12L,
                "slant-dash-dot" = 13L),
  diagonal  = c("up" = 1L, "down" = 2L, "up-down" = 3L),
  reading_order = c("default" = 0L, "ltr" = 1L, "rtl" = 2L)
)

# --- validation helpers ------------------------------------------------------
# Each returns the coerced value, or NULL when the property is unset (NA).

.is_unset <- function(x) is.null(x) || (length(x) == 1L && is.na(x))

.val_flag <- function(x, arg) {
  if (.is_unset(x)) return(NULL)
  if (!is.logical(x) || length(x) != 1L)
    stop(sprintf("`%s` must be a single logical (TRUE/FALSE) or NA", arg), call. = FALSE)
  x
}

.val_enum <- function(x, choices, arg) {
  if (.is_unset(x)) return(NULL)
  if (!is.character(x) || length(x) != 1L || !x %in% choices)
    stop(sprintf("`%s` must be one of: %s (or NA)",
                 arg, paste(choices, collapse = ", ")), call. = FALSE)
  x
}

.val_str <- function(x, arg) {
  if (.is_unset(x)) return(NULL)
  if (!is.character(x) || length(x) != 1L)
    stop(sprintf("`%s` must be a single string or NA", arg), call. = FALSE)
  x
}

# What to write where a value has none.  Any single atomic value is kept with
# its own type, so `na = 0` stays a number and `na = "-"` a string -- the cell
# writer that handles ordinary values handles this one, which is why no type
# table is needed here.  NA returns NULL, meaning "nothing set at this level":
# a workbook then leaves the cell blank, and a column or a cell falls through
# to the level above it.
.val_na <- function(x, arg = "na") {
  if (is.null(x)) return(NULL)
  if (is.factor(x)) x <- as.character(x)
  if (!is.atomic(x) || length(x) != 1L)
    stop(sprintf(paste0("`%s` must be a single value of any type, or NA to ",
                        "leave the cell blank"), arg), call. = FALSE)
  # NaN is as unrepresentable as NA, so it cannot stand in for one either
  if (is.na(x)) return(NULL)
  x
}

.val_int <- function(x, arg, min = NA, max = NA) {
  if (.is_unset(x)) return(NULL)
  if (!is.numeric(x) || length(x) != 1L)
    stop(sprintf("`%s` must be a single number or NA", arg), call. = FALSE)
  x <- as.integer(round(x))
  if (!is.na(min) && x < min || !is.na(max) && x > max)
    stop(sprintf("`%s` must be between %s and %s", arg, min, max), call. = FALSE)
  x
}

.val_num <- function(x, arg, min = NA, max = NA) {
  if (.is_unset(x)) return(NULL)
  if (!is.numeric(x) || length(x) != 1L)
    stop(sprintf("`%s` must be a single number or NA", arg), call. = FALSE)
  x <- as.numeric(x)
  if (!is.na(min) && x < min || !is.na(max) && x > max)
    stop(sprintf("`%s` must be between %s and %s", arg, min, max), call. = FALSE)
  x
}

.val_color <- function(x, arg) {
  if (.is_unset(x)) return(NULL)
  tryCatch(xl_color(x),
           error = function(e)
             stop(sprintf("`%s`: %s", arg, conditionMessage(e)), call. = FALSE))
}

# Build a named list, dropping NULL entries.
.drop_null <- function(x) x[!vapply(x, is.null, logical(1))]

#' Normalize a color to a libxlsxwriter RGB integer
#'
#' Converts an R color name (e.g. `"navy"`), a hex string (`"#FF0000"` or
#' `"FF0000"`), or an integer in the range `0x000000`..`0xFFFFFF` into the
#' single `0xRRGGBB` integer that libxlsxwriter expects.  Used internally by
#' every `color`/`background`/`foreground` argument of the formatting
#' constructors, and exported so it can be used directly.
#'
#' @param x A single color: an R color name, a hex string, or an integer.
#'   `NA` returns `NA_integer_` (an unset color).
#' @return A single integer in `0x000000`..`0xFFFFFF`, or `NA_integer_`.
#' @family cell formatting
#' @export
#' @examples
#' xl_color("red")
#' xl_color("#0000FF")
#' xl_color(255L)
xl_color <- function(x) {
  if (length(x) != 1L)
    stop("xl_color() expects a single color, got length ", length(x), call. = FALSE)
  if (is.na(x)) return(NA_integer_)
  if (is.numeric(x)) {
    v <- as.integer(round(x))
    if (v < 0L || v > 0xFFFFFFL)
      stop("Numeric color must be in the range 0x000000..0xFFFFFF", call. = FALSE)
    return(v)
  }
  x <- as.character(x)
  if (grepl("^[0-9A-Fa-f]{6}$", x)) x <- paste0("#", x)
  rgb <- tryCatch(grDevices::col2rgb(x),
                  error = function(e) stop("Unknown color: '", x, "'", call. = FALSE))
  as.integer(rgb[1L] * 65536L + rgb[2L] * 256L + rgb[3L])
}

# --- internal constructor ----------------------------------------------------

new_xl_format <- function(font = NULL, fill = NULL, border = NULL, align = NULL,
                          num_format = NULL, protection = NULL,
                          quote_prefix = NULL, hyperlink = NULL,
                          class = character(0)) {
  slots <- .drop_null(list(
    font = font, fill = fill, border = border, align = align,
    num_format = num_format, protection = protection,
    quote_prefix = quote_prefix, hyperlink = hyperlink
  ))
  structure(slots, class = c(class, "xl_format"))
}

#' Test whether an object is an `xl_format`
#' @param x An object.
#' @return `TRUE` if `x` inherits from `"xl_format"`.
#' @family cell formatting
#' @export
is_xl_format <- function(x) inherits(x, "xl_format")

# =============================================================================
# Group constructors.  Each returns a complete `xl_format` with only its own
# group populated, so any one can be used directly as a cell/row/column format.
# Every property defaults to NA (= leave unset).
# =============================================================================

#' Cell formatting groups
#'
#' @description
#' These constructors build [xl_format] objects, each populating one group of
#' Excel cell-formatting properties.  Every property defaults to `NA`, meaning
#' "leave unset"; unset properties are simply not written.  Enum-like arguments
#' take lowercase strings and are validated against a fixed set of choices.
#'
#' Each constructor returns a full `xl_format`, so a single group can be used on
#' its own (`xl_cell_general(1, format = xl_font(bold = TRUE))`), and groups can
#' be combined with `+` (see [xl_format]).
#'
#' @param name Font name, e.g. `"Calibri"`.
#' @param size Font size in points (1--409).
#' @param color,background,foreground,left_color,right_color,top_color,bottom_color,diagonal_color
#'   A color: an R color name, a hex string, or an integer (see [xl_color]).
#' @param bold,italic,strikeout,outline,shadow,condense,extend,font_only Logical
#'   font flags.
#' @param underline One of `"none"`, `"single"`, `"double"`,
#'   `"single-accounting"`, `"double-accounting"`.
#' @param script One of `"super"`, `"sub"`.
#' @param family,charset,theme,color_indexed Advanced integer font properties
#'   (rarely needed).
#' @param scheme Font scheme string (advanced).
#' @return An [xl_format] object.
#' @name xl_format_groups
#' @family cell formatting
#' @seealso [xl_format], [xl_color]
#' @examples
#' xl_font(bold = TRUE, color = "navy", size = 12)
#' xl_fill(background = "#FFF2CC")
#' xl_border(all = "thin", color = "gray")
#' xl_align(horizontal = "center", vertical = "top", wrap = TRUE)
#' xl_num_format("#,##0.00")
#' xl_protection(locked = FALSE)
NULL

#' @rdname xl_format_groups
#' @export
xl_font <- function(bold = NA, italic = NA, color = NA, size = NA, name = NA,
                    underline = NA, strikeout = NA, script = NA, family = NA,
                    charset = NA, outline = NA, shadow = NA, condense = NA,
                    extend = NA, scheme = NA, theme = NA, color_indexed = NA,
                    font_only = NA) {
  font <- .drop_null(list(
    name          = .val_str(name, "name"),
    size          = .val_num(size, "size", min = 1, max = 409),
    color         = .val_color(color, "color"),
    bold          = .val_flag(bold, "bold"),
    italic        = .val_flag(italic, "italic"),
    underline     = .val_enum(underline, names(.LXW$underline), "underline"),
    strikeout     = .val_flag(strikeout, "strikeout"),
    script        = .val_enum(script, names(.LXW$script), "script"),
    family        = .val_int(family, "family"),
    charset       = .val_int(charset, "charset"),
    outline       = .val_flag(outline, "outline"),
    shadow        = .val_flag(shadow, "shadow"),
    condense      = .val_flag(condense, "condense"),
    extend        = .val_flag(extend, "extend"),
    scheme        = .val_str(scheme, "scheme"),
    theme         = .val_int(theme, "theme"),
    color_indexed = .val_int(color_indexed, "color_indexed"),
    font_only     = .val_flag(font_only, "font_only")
  ))
  new_xl_format(font = if (length(font)) font else NULL)
}

#' @rdname xl_format_groups
#' @param pattern Fill pattern, e.g. `"solid"`, `"light-gray"` (18 choices).
#'   When `background` is supplied and `pattern` is unset, a `"solid"` pattern
#'   is assumed.
#' @export
xl_fill <- function(background = NA, foreground = NA, pattern = NA,
                    transparency = NA) {
  bg  <- .val_color(background, "background")
  fg  <- .val_color(foreground, "foreground")
  pat <- .val_enum(pattern, names(.LXW$pattern), "pattern")
  if (is.null(pat) && !is.null(bg) && is.null(fg)) pat <- "solid"
  fill <- .drop_null(list(background = bg, foreground = fg, pattern = pat,
                          transparency = .val_int(transparency, "transparency",
                                                  min = 0, max = 100)))
  new_xl_format(fill = if (length(fill)) fill else NULL)
}

#' @rdname xl_format_groups
#' @param transparency Percentage transparency, 0--100.  **Charts only**: Excel
#'   has no transparency for a cell's fill or border, so this is ignored
#'   everywhere except a chart's line and fill (see [xl_chart()]).
#' @param all Border style applied to all four sides at once (one of `"none"`,
#'   `"thin"`, `"medium"`, `"dashed"`, `"dotted"`, `"thick"`, `"double"`,
#'   `"hair"`, `"medium-dashed"`, `"dash-dot"`, `"medium-dash-dot"`,
#'   `"dash-dot-dot"`, `"medium-dash-dot-dot"`, `"slant-dash-dot"`).
#' @param left,right,top,bottom Per-side border styles (override `all`).
#' @param diagonal Diagonal border direction: `"up"`, `"down"`, `"up-down"`.
#' @param diagonal_style Diagonal border style (same choices as `all`).
#' @export
xl_border <- function(all = NA, left = NA, right = NA, top = NA, bottom = NA,
                      color = NA, left_color = NA, right_color = NA,
                      top_color = NA, bottom_color = NA, diagonal = NA,
                      diagonal_style = NA, diagonal_color = NA,
                      transparency = NA) {
  bchoice <- names(.LXW$border)
  all_s   <- .val_enum(all, bchoice, "all")
  all_c   <- .val_color(color, "color")
  border <- .drop_null(list(
    left           = .val_enum(left, bchoice, "left")   %||% all_s,
    right          = .val_enum(right, bchoice, "right")  %||% all_s,
    top            = .val_enum(top, bchoice, "top")      %||% all_s,
    bottom         = .val_enum(bottom, bchoice, "bottom") %||% all_s,
    left_color     = .val_color(left_color, "left_color")     %||% all_c,
    right_color    = .val_color(right_color, "right_color")   %||% all_c,
    top_color      = .val_color(top_color, "top_color")       %||% all_c,
    bottom_color   = .val_color(bottom_color, "bottom_color") %||% all_c,
    diagonal       = .val_enum(diagonal, names(.LXW$diagonal), "diagonal"),
    diagonal_style = .val_enum(diagonal_style, bchoice, "diagonal_style"),
    diagonal_color = .val_color(diagonal_color, "diagonal_color"),
    transparency   = .val_int(transparency, "transparency", min = 0, max = 100)
  ))
  new_xl_format(border = if (length(border)) border else NULL)
}

#' @rdname xl_format_groups
#' @param horizontal Horizontal alignment: `"left"`, `"center"`, `"right"`,
#'   `"fill"`, `"justify"`, `"center-across"`, `"distributed"`.
#' @param vertical Vertical alignment: `"top"`, `"bottom"`, `"center"`,
#'   `"justify"`, `"distributed"`.
#' @param wrap Logical; wrap text in the cell.
#' @param rotation Text rotation in degrees (-90..90, or 270).
#' @param indent Integer indentation level.
#' @param shrink Logical; shrink text to fit.
#' @param reading_order One of `"default"`, `"ltr"`, `"rtl"`.
#' @export
xl_align <- function(horizontal = NA, vertical = NA, wrap = NA, rotation = NA,
                     indent = NA, shrink = NA, reading_order = NA) {
  align <- .drop_null(list(
    horizontal    = .val_enum(horizontal, names(.LXW$align_h), "horizontal"),
    vertical      = .val_enum(vertical, names(.LXW$align_v), "vertical"),
    wrap          = .val_flag(wrap, "wrap"),
    rotation      = .val_int(rotation, "rotation", min = -90, max = 270),
    indent        = .val_int(indent, "indent", min = 0),
    shrink        = .val_flag(shrink, "shrink"),
    reading_order = .val_enum(reading_order, names(.LXW$reading_order), "reading_order")
  ))
  if (!is.null(align$rotation) && align$rotation > 90 && align$rotation != 270)
    stop("`rotation` must be between -90 and 90, or exactly 270", call. = FALSE)
  new_xl_format(align = if (length(align)) align else NULL)
}

#' @rdname xl_format_groups
#' @param format A number-format string, e.g. `"#,##0.00"`, `"0%"`,
#'   `"yyyy-mm-dd"`.
#' @param index An Excel built-in number-format index (alternative to
#'   `format`).
#' @export
xl_num_format <- function(format = NA, index = NA) {
  num <- .drop_null(list(
    format = .val_str(format, "format"),
    index  = .val_int(index, "index", min = 0, max = 255)
  ))
  new_xl_format(num_format = if (length(num)) num else NULL)
}

#' @rdname xl_format_groups
#' @param locked Logical; whether the cell is locked (Excel's default is
#'   `TRUE`).  Only takes effect when the worksheet is protected.  `FALSE`
#'   unlocks the cell.
#' @param hidden Logical; hide the cell's formula.
#' @export
xl_protection <- function(locked = NA, hidden = NA) {
  prot <- .drop_null(list(
    locked = .val_flag(locked, "locked"),
    hidden = .val_flag(hidden, "hidden")
  ))
  new_xl_format(protection = if (length(prot)) prot else NULL)
}

`%||%` <- function(a, b) if (is.null(a)) b else a

# =============================================================================
# xl_format(): merge any number of group objects into one format.
# =============================================================================

#' Combine cell-formatting groups into a single format
#'
#' @description
#' `xl_format()` merges any number of [xl_format] objects (typically the
#' single-group objects returned by [xl_font], [xl_fill], [xl_border],
#' [xl_align], [xl_num_format] and [xl_protection]) into one combined format.
#' The `+` operator does the same for two formats.
#'
#' Merging is right-biased and works property-by-property: where two formats set
#' the *same* property the later one wins, but properties set by only one side
#' are all preserved.  So partial groups accumulate rather than overwrite.
#'
#' @param ... [xl_format] objects (and/or `NULL`s, which are ignored).  May also
#'   include the scalar flags `quote_prefix` and `hyperlink`.
#' @param quote_prefix Logical; treat the cell contents as literal text (as if
#'   prefixed with a single quote in Excel).
#' @param hyperlink Logical; apply the internal hyperlink style flag (advanced).
#' @return A combined [xl_format] object.
#' @family cell formatting
#' @seealso [xl_font], [xl_fill], [xl_border], [xl_align], [xl_num_format],
#'   [xl_protection], [xl_color]
#' @export
#' @examples
#' xl_format(xl_font(bold = TRUE), xl_border(bottom = "thin"),
#'           xl_num_format("#,##0.00"))
#'
#' # equivalent, using +
#' xl_font(bold = TRUE) + xl_border(bottom = "thin") + xl_num_format("#,##0.00")
xl_format <- function(..., quote_prefix = NA, hyperlink = NA) {
  parts <- list(...)
  parts <- parts[!vapply(parts, is.null, logical(1))]
  for (p in parts)
    if (!is_xl_format(p))
      stop("all `...` arguments to xl_format() must be xl_format objects ",
           "(e.g. from xl_font(), xl_fill(), ...)", call. = FALSE)
  out <- if (length(parts)) Reduce(merge_xl_format, parts) else new_xl_format()
  qp <- .val_flag(quote_prefix, "quote_prefix")
  hl <- .val_flag(hyperlink, "hyperlink")
  if (!is.null(qp)) out <- merge_xl_format(out, new_xl_format(quote_prefix = qp))
  if (!is.null(hl)) out <- merge_xl_format(out, new_xl_format(hyperlink = hl))
  out
}

# Merge b onto a: group-by-group then property-by-property, b wins on conflict.
# Subclass "extras" carried in the `xl_geometry`/`xl_target` attributes are
# merged too, and the result takes the more-specific class of the two.
merge_xl_format <- function(a, b) {
  if (is.null(a)) return(b)
  if (is.null(b)) return(a)
  out <- unclass(a)
  lb  <- unclass(b)
  for (slot in names(lb)) {
    # groups (named lists) merge property-by-property; scalar flags replace
    out[[slot]] <- if (is.list(lb[[slot]]) && is.list(out[[slot]]))
      utils::modifyList(out[[slot]], lb[[slot]]) else lb[[slot]]
  }
  # subclass extras (geometry/targeting) carried as attributes
  for (at in c("xl_geometry", "xl_target")) {
    av <- attr(a, at, exact = TRUE)
    bv <- attr(b, at, exact = TRUE)
    merged <- if (is.null(bv)) av else if (is.null(av)) bv else utils::modifyList(av, bv)
    if (!is.null(merged)) attr(out, at) <- merged
  }
  class(out) <- .result_format_class(a, b)
  out
}

# The more-specific of two format classes (a subclass such as xl_col_spec wins
# over a bare xl_format; if both are subclasses, the right operand wins).
.result_format_class <- function(a, b) {
  ca <- class(a); cb <- class(b)
  spec_b <- setdiff(cb, "xl_format")
  spec_a <- setdiff(ca, "xl_format")
  if (length(spec_b)) cb else if (length(spec_a)) ca else "xl_format"
}

#' @export
#' @rdname xl_format
#' @param e1,e2 [xl_format] objects to combine.
`+.xl_format` <- function(e1, e2) {
  if (is.null(e1)) return(e2)
  if (is.null(e2)) return(e1)
  if (!is_xl_format(e1) || !is_xl_format(e2))
    stop("both operands of `+` must be xl_format objects", call. = FALSE)
  merge_xl_format(e1, e2)
}

# =============================================================================
# Printing
# =============================================================================

#' @export
print.xl_format <- function(x, ...) {
  cls <- setdiff(class(x), "xl_format")
  label <- if (length(cls)) cls[1L] else "xl_format"
  cat(sprintf("<%s>\n", label))
  f <- unclass(x)
  tgt <- attr(x, "xl_target", exact = TRUE)
  geo <- attr(x, "xl_geometry", exact = TRUE)
  if (!is.null(tgt)) cat("  target:", .fmt_kv(tgt), "\n")
  if (!is.null(geo)) cat("  geometry:", .fmt_kv(geo), "\n")
  shown <- FALSE
  for (slot in names(f)) {
    v <- f[[slot]]
    if (is.list(v)) {
      cat(sprintf("  %s: %s\n", slot, .fmt_kv(v))); shown <- TRUE
    } else if (isTRUE(v)) {
      cat(sprintf("  %s: TRUE\n", slot)); shown <- TRUE
    }
  }
  if (!shown && is.null(tgt) && is.null(geo))
    cat("  (empty)\n")
  invisible(x)
}

.fmt_kv <- function(lst) {
  paste(vapply(names(lst), function(nm) {
    v <- lst[[nm]]
    if (is.null(v)) return(NA_character_)
    paste0(nm, "=", paste(format(v), collapse = ","))
  }, character(1)), collapse = ", ")
}

# =============================================================================
# Payload conversion + global dedup registry (used at write time)
# =============================================================================

# Flatten an xl_format to a named list of C-ready scalars.  Keys correspond
# 1:1 to format_set_* calls in src/write_xlsx.c.  Enum strings become the
# libxlsxwriter integer; colors are already integers; flags are TRUE.
.xl_format_payload <- function(fmt) {
  f <- unclass(fmt)
  p <- list()
  fo <- f$font
  if (!is.null(fo)) {
    if (!is.null(fo$name))          p$font_name     <- as.character(fo$name)
    if (!is.null(fo$size))          p$font_size     <- as.numeric(fo$size)
    if (!is.null(fo$color))         p$font_color    <- as.integer(fo$color)
    if (isTRUE(fo$bold))            p$bold          <- TRUE
    if (isTRUE(fo$italic))          p$italic        <- TRUE
    if (!is.null(fo$underline))     p$underline     <- .LXW$underline[[fo$underline]]
    if (isTRUE(fo$strikeout))       p$strikeout     <- TRUE
    if (!is.null(fo$script))        p$script        <- .LXW$script[[fo$script]]
    if (!is.null(fo$family))        p$font_family   <- as.integer(fo$family)
    if (!is.null(fo$charset))       p$font_charset  <- as.integer(fo$charset)
    if (isTRUE(fo$outline))         p$font_outline  <- TRUE
    if (isTRUE(fo$shadow))          p$font_shadow   <- TRUE
    if (isTRUE(fo$condense))        p$font_condense <- TRUE
    if (isTRUE(fo$extend))          p$font_extend   <- TRUE
    if (!is.null(fo$scheme))        p$font_scheme   <- as.character(fo$scheme)
    if (!is.null(fo$theme))         p$theme         <- as.integer(fo$theme)
    if (!is.null(fo$color_indexed)) p$color_indexed <- as.integer(fo$color_indexed)
    if (isTRUE(fo$font_only))       p$font_only     <- TRUE
  }
  fi <- f$fill
  if (!is.null(fi)) {
    if (!is.null(fi$background)) p$bg_color <- as.integer(fi$background)
    if (!is.null(fi$foreground)) p$fg_color <- as.integer(fi$foreground)
    if (!is.null(fi$pattern))    p$pattern  <- .LXW$pattern[[fi$pattern]]
  }
  bd <- f$border
  if (!is.null(bd)) {
    if (!is.null(bd$left))           p$border_left   <- .LXW$border[[bd$left]]
    if (!is.null(bd$right))          p$border_right  <- .LXW$border[[bd$right]]
    if (!is.null(bd$top))            p$border_top    <- .LXW$border[[bd$top]]
    if (!is.null(bd$bottom))         p$border_bottom <- .LXW$border[[bd$bottom]]
    if (!is.null(bd$left_color))     p$left_color    <- as.integer(bd$left_color)
    if (!is.null(bd$right_color))    p$right_color   <- as.integer(bd$right_color)
    if (!is.null(bd$top_color))      p$top_color     <- as.integer(bd$top_color)
    if (!is.null(bd$bottom_color))   p$bottom_color  <- as.integer(bd$bottom_color)
    if (!is.null(bd$diagonal))       p$diag_type     <- .LXW$diagonal[[bd$diagonal]]
    if (!is.null(bd$diagonal_style)) p$diag_border   <- .LXW$border[[bd$diagonal_style]]
    if (!is.null(bd$diagonal_color)) p$diag_color    <- as.integer(bd$diagonal_color)
  }
  al <- f$align
  if (!is.null(al)) {
    if (!is.null(al$horizontal))    p$align_h       <- .LXW$align_h[[al$horizontal]]
    if (!is.null(al$vertical))      p$align_v       <- .LXW$align_v[[al$vertical]]
    if (isTRUE(al$wrap))            p$text_wrap     <- TRUE
    if (!is.null(al$rotation))      p$rotation      <- as.integer(al$rotation)
    if (!is.null(al$indent))        p$indent        <- as.integer(al$indent)
    if (isTRUE(al$shrink))          p$shrink        <- TRUE
    if (!is.null(al$reading_order)) p$reading_order <- .LXW$reading_order[[al$reading_order]]
  }
  nf <- f$num_format
  if (!is.null(nf)) {
    if (!is.null(nf$format)) p$num_format       <- as.character(nf$format)
    if (!is.null(nf$index))  p$num_format_index <- as.integer(nf$index)
  }
  pr <- f$protection
  if (!is.null(pr)) {
    if (isFALSE(pr$locked)) p$unlocked <- TRUE
    if (isTRUE(pr$hidden))  p$hidden   <- TRUE
  }
  if (isTRUE(f$quote_prefix)) p$quote_prefix  <- TRUE
  if (isTRUE(f$hyperlink))    p$set_hyperlink <- TRUE
  p
}

# A stable string key for a payload, used to deduplicate identical formats.
.payload_key <- function(p) {
  if (length(p) == 0L) return("")
  p <- p[order(names(p))]
  paste0(names(p), ":",
         vapply(p, function(v) paste0(typeof(v), "=", as.character(v)), character(1)),
         collapse = "|")
}

# A registry accumulates unique format payloads across a whole workbook and
# hands back a 1-based integer id for each (0 means "no format").
.new_format_registry <- function() {
  reg <- new.env(parent = emptyenv())
  reg$keys  <- character(0)
  reg$table <- list()
  reg
}

.register_format <- function(reg, fmt) {
  if (is.null(fmt)) return(0L)
  p <- .xl_format_payload(fmt)
  if (length(p) == 0L) return(0L)
  k <- .payload_key(p)
  idx <- match(k, reg$keys)
  if (is.na(idx)) {
    reg$table[[length(reg$table) + 1L]] <- p
    reg$keys <- c(reg$keys, k)
    idx <- length(reg$table)
  }
  idx
}

# Integration tests for cell-level formatting via xl_cell_general(format=).
# Formatting is verified by inspecting the written xl/styles.xml.

test_that("xl_cell_general accepts a single format, a list, or NULL", {
  x <- xl_cell_general(value = 1:3, format = xl_font(bold = TRUE))
  expect_true(is_xl_format(x[[1]]$format))
  expect_true(is_xl_format(x[[3]]$format))

  # per-cell list, recycled
  y <- xl_cell_general(value = 1:2,
                       format = list(xl_font(bold = TRUE), xl_font(italic = TRUE)))
  expect_true(unclass(y[[1]]$format)$font$bold)
  expect_true(unclass(y[[2]]$format)$font$italic)

  # NULL -> no format
  z <- xl_cell_general(value = 1:2)
  expect_null(z[[1]]$format)

  expect_error(xl_cell_general(value = 1, format = list("bad")),
               "must be an xl_format")
  expect_error(xl_cell_general(value = 1, format = 42), "must be an xl_format")
})

test_that("print shows a format marker", {
  expect_output(print(xl_cell_general(value = 1, format = xl_font(bold = TRUE))),
                "format=<set>")
})

test_that("font formatting reaches styles.xml", {
  df <- data.frame(a = 1L)
  df$b <- xl_cell_general(value = "hi",
                          format = xl_font(bold = TRUE, italic = TRUE,
                                           color = "red", size = 16))
  s <- styles_string(df)
  expect_match(s, "<b/>")
  expect_match(s, "<i/>")
  expect_match(s, "FFFF0000", ignore.case = TRUE)  # red font color, ARGB
  expect_match(s, 'val="16"')
})

test_that("fill formatting reaches styles.xml", {
  df <- data.frame(a = 1L)
  df$b <- xl_cell_general(value = 1, format = xl_fill(background = "yellow"))
  s <- styles_string(df)
  expect_match(s, "<patternFill")
  expect_match(s, "FFFFFF00", ignore.case = TRUE)   # yellow
})

test_that("border formatting reaches styles.xml", {
  df <- data.frame(a = 1L)
  df$b <- xl_cell_general(value = 1, format = xl_border(all = "thin", color = "black"))
  s <- styles_string(df)
  expect_match(s, "<border")
  expect_match(s, "thin")
})

test_that("alignment formatting reaches styles.xml", {
  df <- data.frame(a = 1L)
  df$b <- xl_cell_general(value = "x",
                          format = xl_align(horizontal = "center", vertical = "top",
                                            wrap = TRUE))
  s <- styles_string(df)
  expect_match(s, 'horizontal="center"')
  expect_match(s, 'vertical="top"')
  expect_match(s, 'wrapText="1"')
})

test_that("number format reaches styles.xml", {
  df <- data.frame(a = 1L)
  df$b <- xl_cell_general(value = 1234.5, format = xl_num_format("#,##0.00"))
  s <- styles_string(df)
  expect_match(s, "#,##0.00", fixed = TRUE)
})

test_that("protection formatting reaches styles.xml", {
  df <- data.frame(a = 1L)
  df$b <- xl_cell_general(value = 1, format = xl_protection(locked = FALSE, hidden = TRUE))
  # unprotected sheet warns that the cell locking has no effect (tested
  # separately); silence it here.
  s <- suppressWarnings(styles_string(df))
  expect_match(s, 'locked="0"')
  expect_match(s, 'hidden="1"')
})

test_that("date/time general cells get the default number format merged", {
  df <- data.frame(a = 1L)
  df$d <- xl_cell_general(value = as.Date("2020-03-01"), format = xl_font(bold = TRUE))
  df$t <- xl_cell_general(value = as.POSIXct("2020-03-01 12:00", tz = "UTC"))
  s <- styles_string(df)
  expect_match(s, "yyyy", fixed = TRUE)          # date number format present
  expect_match(s, "<b/>")                        # bold preserved alongside it
  # explicit num_format on a date cell wins over the default
  df2 <- data.frame(a = 1L)
  df2$d <- xl_cell_general(value = as.Date("2020-03-01"),
                           format = xl_num_format("dd/mm/yyyy"))
  expect_match(styles_string(df2), "dd/mm/yyyy", fixed = TRUE)
})

test_that("values still round-trip with formatting applied", {
  df <- data.frame(x = 1:3)
  df$n <- xl_cell_general(value = c(1.5, 2.5, 3.5),
                          format = xl_num_format("0.0"))
  tmp <- write_tmp(df)
  rd <- readxl::read_xlsx(tmp)
  expect_equal(rd$n, c(1.5, 2.5, 3.5))
})

test_that("format= flows through xl_formula and xl_hyperlink_cell", {
  df <- data.frame(a = 1L)
  df$f <- xl_formula("=1+2", format = xl_font(bold = TRUE))
  expect_match(styles_string(df), "<b/>")

  df2 <- data.frame(a = 1L)
  df2$h <- xl_hyperlink_cell("https://example.com", value = "link",
                             format = xl_font(color = "green"))
  s2 <- styles_string(df2)
  expect_match(s2, "FF00FF00", ignore.case = TRUE)  # R "green" == #00FF00

  df3 <- data.frame(a = 1L)
  df3$h <- xl_hyperlink("https://example.com", "link", format = xl_font(italic = TRUE))
  expect_match(styles_string(df3), "<i/>")
})

test_that("wrappers coerce factor input while applying a format", {
  df <- data.frame(a = 1L)
  df$f <- xl_formula(factor("=1+1"), format = xl_font(bold = TRUE))
  expect_match(styles_string(df), "<b/>")

  df2 <- data.frame(a = 1L)
  df2$h <- xl_hyperlink(factor("http://example.com"), format = xl_font(italic = TRUE))
  expect_match(styles_string(df2), "<i/>")

  df3 <- data.frame(a = 1L)
  df3$h <- xl_hyperlink_cell(factor("http://example.com"),
                             format = xl_font(strikeout = TRUE))
  expect_match(styles_string(df3), "strike")
})

test_that("identical formats deduplicate to one style definition", {
  df <- data.frame(a = 1:50)
  df$b <- xl_cell_general(value = 1:50, format = xl_num_format("0.000"))
  s <- styles_string(df)
  hits <- gregexpr("0.000", s, fixed = TRUE)[[1]]
  expect_equal(sum(hits > 0), 1L)   # the custom number format is defined once
})

# Tests for xl_cell_general() constructor, S3 methods, and write_xlsx integration

# ── Construction: value types ─────────────────────────────────────────────────

test_that("xl_cell_general: numeric value", {
  x <- xl_cell_general(value = 42.5)
  expect_s3_class(x, "xl_cell_general")
  expect_s3_class(x, "xl_cell")
  expect_equal(length(x), 1L)
  expect_equal(x[[1L]][["value"]], 42.5)
  expect_true(is.na(x[[1L]][["formula"]]))
  expect_identical(x[[1L]][["hyperlink"]], NA)
})

test_that("xl_cell_general: integer value", {
  x <- xl_cell_general(value = 7L)
  expect_equal(x[[1L]][["value"]], 7L)
})

test_that("xl_cell_general: logical value", {
  x <- xl_cell_general(value = TRUE)
  expect_equal(x[[1L]][["value"]], TRUE)
})

test_that("xl_cell_general: character value", {
  x <- xl_cell_general(value = "hello")
  expect_equal(x[[1L]][["value"]], "hello")
})

test_that("xl_cell_general: Date value", {
  d <- as.Date("2024-01-15")
  x <- xl_cell_general(value = d)
  expect_equal(x[[1L]][["value"]], d)
})

test_that("xl_cell_general: POSIXct value", {
  dt <- as.POSIXct("2024-01-15 12:00:00", tz = "UTC")
  x <- xl_cell_general(value = dt)
  expect_equal(x[[1L]][["value"]], dt)
})

test_that("xl_cell_general: mixed-type list value", {
  x <- xl_cell_general(value = list(1.5, "text", TRUE))
  expect_equal(length(x), 3L)
  expect_equal(x[[1L]][["value"]], 1.5)
  expect_equal(x[[2L]][["value"]], "text")
  expect_equal(x[[3L]][["value"]], TRUE)
})

test_that("xl_cell_general: NA value produces blank cell", {
  x <- xl_cell_general(value = NA)
  expect_equal(length(x), 1L)
  expect_true(is.na(x[[1L]][["value"]]))
})

test_that("xl_cell_general: no arguments errors with informative message", {
  expect_error(xl_cell_general(), "At least one")
})

test_that("xl_cell_general: value = NA produces explicit empty cell", {
  x <- xl_cell_general(value = NA)
  expect_equal(length(x), 1L)
  expect_true(is.na(x[[1L]][["value"]]))
})

# ── Construction: formula ──────────────────────────────────────────────────────

test_that("xl_cell_general: formula only", {
  x <- xl_cell_general(formula = "=SUM(A1:A10)")
  expect_equal(x[[1L]][["formula"]], "=SUM(A1:A10)")
  expect_identical(x[[1L]][["hyperlink"]], NA)
})

test_that("xl_cell_general: formula with numeric pre-calculated value", {
  x <- xl_cell_general(value = 55.0, formula = "=SUM(A1:A10)")
  expect_equal(x[[1L]][["formula"]], "=SUM(A1:A10)")
  expect_equal(x[[1L]][["value"]], 55.0)
})

test_that("xl_cell_general: formula with character pre-calculated value", {
  x <- xl_cell_general(value = "result", formula = "=TEXT(A1,\"0\")")
  expect_equal(x[[1L]][["formula"]], "=TEXT(A1,\"0\")")
  expect_equal(x[[1L]][["value"]], "result")
})

test_that("xl_cell_general: formula with NA value -> plain formula", {
  x <- xl_cell_general(value = NA, formula = "=A1+1")
  expect_equal(x[[1L]][["formula"]], "=A1+1")
  expect_true(is.na(x[[1L]][["value"]]))
})

test_that("xl_cell_general: formula vector of length > 1", {
  x <- xl_cell_general(formula = c("=A1", "=A2", "=A3"))
  expect_equal(length(x), 3L)
  expect_equal(x[[2L]][["formula"]], "=A2")
})

test_that("xl_cell_general: NA formula -> no formula written", {
  x <- xl_cell_general(formula = NA_character_)
  expect_true(is.na(x[[1L]][["formula"]]))
})

test_that("xl_cell_general: formula not starting with '=' errors", {
  expect_error(xl_cell_general(formula = "SUM(A1:A10)"),
               "must start with '='")
})

# ── Construction: hyperlink ────────────────────────────────────────────────────

test_that("xl_cell_general: hyperlink as character URL", {
  x <- xl_cell_general(hyperlink = "https://example.com")
  expect_equal(x[[1L]][["hyperlink"]], "https://example.com")
})

test_that("xl_cell_general: hyperlink as named list with url only", {
  x <- xl_cell_general(hyperlink = list(url = "https://example.com"))
  expect_equal(x[[1L]][["hyperlink"]][["url"]], "https://example.com")
})

test_that("xl_cell_general: hyperlink as named list with url and tooltip", {
  x <- xl_cell_general(hyperlink = list(url = "https://example.com",
                                         tooltip = "Go to example.com"))
  h <- x[[1L]][["hyperlink"]]
  expect_equal(h[["url"]],     "https://example.com")
  expect_equal(h[["tooltip"]], "Go to example.com")
})

test_that("xl_cell_general: value provides display text for hyperlink cell", {
  x <- xl_cell_general(value    = "Visit",
                        hyperlink = list(url     = "https://example.com",
                                          tooltip = "Go to example.com"))
  expect_equal(x[[1L]][["value"]], "Visit")
  expect_equal(x[[1L]][["hyperlink"]][["url"]], "https://example.com")
})

test_that("xl_cell_general: NA hyperlink -> no hyperlink written", {
  x <- xl_cell_general(hyperlink = NA)
  expect_identical(x[[1L]][["hyperlink"]], NA)
})

test_that("xl_cell_general: hyperlink vector of length > 1", {
  x <- xl_cell_general(hyperlink = c("https://a.com", "https://b.com"))
  expect_equal(length(x), 2L)
  expect_equal(x[[2L]][["hyperlink"]], "https://b.com")
})

test_that("xl_cell_general: invalid hyperlink format errors", {
  expect_error(
    xl_cell_general(hyperlink = list(list(noturl = "x"))),
    "hyperlink"
  )
})

test_that("xl_cell_general: hyperlink list missing url errors", {
  expect_error(
    xl_cell_general(hyperlink = list(list(tooltip = "hover"))),
    "hyperlink"
  )
})

# ── Vectorization ─────────────────────────────────────────────────────────────

test_that("xl_cell_general: length determined from longest input", {
  x <- xl_cell_general(value = 1:3, formula = "=A1")
  expect_equal(length(x), 3L)
  # formula recycled to length 3
  expect_equal(x[[2L]][["formula"]], "=A1")
  expect_equal(x[[3L]][["formula"]], "=A1")
})

test_that("xl_cell_general: all NULL inputs errors", {
  expect_error(xl_cell_general(), "At least one")
})

test_that("xl_cell_general: length() returns number of cells", {
  x <- xl_cell_general(value = 1:5)
  expect_equal(length(x), 5L)
})

# ── S3 methods ────────────────────────────────────────────────────────────────

test_that("[.xl_cell_general preserves class and picks correct cells", {
  x <- xl_cell_general(value = 1:4)
  sub <- x[2:3]
  expect_s3_class(sub, "xl_cell_general")
  expect_equal(length(sub), 2L)
  expect_equal(sub[[1L]][["value"]], 2L)
  expect_equal(sub[[2L]][["value"]], 3L)
})

test_that("c.xl_cell_general concatenates and preserves class", {
  a <- xl_cell_general(value = 1)
  b <- xl_cell_general(value = 2)
  ab <- c(a, b)
  expect_s3_class(ab, "xl_cell_general")
  expect_equal(length(ab), 2L)
  expect_equal(ab[[1L]][["value"]], 1)
  expect_equal(ab[[2L]][["value"]], 2)
})

test_that("rep.xl_cell_general repeats correctly", {
  x <- xl_cell_general(value = 42)
  r <- rep(x, times = 3L)
  expect_s3_class(r, "xl_cell_general")
  expect_equal(length(r), 3L)
  expect_equal(r[[3L]][["value"]], 42)
})

test_that("as.data.frame.xl_cell_general: data.frame() uses the argument name", {
  x <- xl_cell_general(value = 1:3)
  # names(as.data.frame(x)) is NULL so data.frame() keeps the argument name
  expect_null(names(as.data.frame(x)))
  df <- data.frame(a = 1:3, my_col = x)
  expect_equal(names(df), c("a", "my_col"))
  # not a list, and that is the point: `[<-.data.frame` reads a list value as a
  # list of columns, so a list-backed cell vector could not be assigned with
  # `df[, j] <-` at all
  expect_s3_class(df$my_col, "xl_cell_general")
  expect_false(is.list(df$my_col))
})

test_that("rep.xl_cell_general with length.out", {
  x <- xl_cell_general(value = c(1, 2))
  r <- rep(x, length.out = 5L)
  expect_equal(length(r), 5L)
})


test_that("print.xl_cell_general runs without error for length-1 cell", {
  x <- xl_cell_general(value = 99)
  expect_output(print(x), "xl_cell_general")
})

test_that("print.xl_cell_general shows formula and hyperlink", {
  x <- xl_cell_general(formula = "=A1+1")
  expect_output(print(x), "formula")

  y <- xl_cell_general(hyperlink = "https://x.com")
  expect_output(print(y), "hyperlink")
})

test_that("print.xl_cell_general truncates at max and shows count", {
  x <- xl_cell_general(value = seq_len(20))
  expect_output(print(x, max = 5L), "more")
})

test_that("print.xl_cell_general shows <empty> for explicit NA cell", {
  x <- xl_cell_general(value = NA)
  expect_output(print(x), "empty")
})

# ── Data frame integration ────────────────────────────────────────────────────

test_that("length-1 xl_formula recycles in data frame", {
  df <- data.frame(x = 1:3)
  df$f <- xl_formula("=A1*2")
  expect_true(file.exists(write_xlsx(df)))
})

test_that("length-1 xl_hyperlink recycles in data frame", {
  df <- data.frame(x = 1:3)
  df$h <- xl_hyperlink("https://example.com")
  expect_true(file.exists(write_xlsx(df)))
})

test_that("length-1 xl_cell_general(value) recycles in data frame", {
  df <- data.frame(x = 1:3)
  df$v <- xl_cell_general(value = 99L)
  expect_true(file.exists(write_xlsx(df)))
})

test_that("full-length xl_cell_general column writes without error", {
  df <- data.frame(x = 1:3)
  df$c <- xl_cell_general(value = c(10, 20, 30))
  expect_true(file.exists(write_xlsx(df)))
})

test_that("mixed-type xl_cell_general column (numeric + string + formula) writes", {
  mixed <- c(
    xl_cell_general(value = 1.5),
    xl_cell_general(value = "note"),
    xl_cell_general(formula = "=A1+A2")
  )
  df <- data.frame(x = 1:3)
  df$m <- mixed
  expect_true(file.exists(write_xlsx(df)))
})

test_that("formula with numeric pre-calc value writes successfully", {
  df <- data.frame(x = 1:2)
  df$f <- c(
    xl_cell_general(value = 55.0, formula = "=SUM(A1:A2)"),
    xl_cell_general(value = 10.0, formula = "=A1*2")
  )
  expect_true(file.exists(write_xlsx(df)))
})

test_that("formula with character pre-calc value writes successfully", {
  df <- data.frame(x = 1:2)
  df$f <- c(
    xl_cell_general(value = "total", formula = "=TEXT(A1,\"0\")"),
    xl_cell_general(formula = "=A1+1")
  )
  expect_true(file.exists(write_xlsx(df)))
})

test_that("hyperlink with value display text and tooltip writes successfully", {
  df <- data.frame(x = 1:2)
  df$h <- c(
    xl_cell_general(value    = "Visit",
                    hyperlink = list(url     = "https://example.com",
                                      tooltip = "tooltip text")),
    xl_cell_general(hyperlink = "https://other.com")
  )
  expect_true(file.exists(write_xlsx(df)))
})

test_that("NA hyperlink in xl_cell_general column writes blank cell", {
  df <- data.frame(x = 1:2)
  df$h <- c(
    xl_cell_general(hyperlink = "https://example.com"),
    xl_cell_general(hyperlink = NA)
  )
  expect_true(file.exists(write_xlsx(df)))
})

# ── Backward compatibility ─────────────────────────────────────────────────────

test_that("xl_formula() returns xl_cell_general with correct class", {
  x <- xl_formula("=A1+1")
  expect_s3_class(x, "xl_cell_general")
  expect_s3_class(x, "xl_cell")
})

test_that("xl_formula() validates formula starts with '='", {
  expect_error(xl_formula("A1+1"), "start with")
})

test_that("xl_formula() accepts NA values", {
  x <- xl_formula(c("=A1", NA))
  expect_equal(length(x), 2L)
  expect_true(is.na(x[[2L]][["formula"]]))
})

test_that("xl_formula() writes valid xlsx", {
  df <- data.frame(x = 1:3)
  df$f <- xl_formula("=A1*2")
  expect_true(file.exists(write_xlsx(df)))
})

test_that("xl_hyperlink() returns xl_cell_general with correct class", {
  x <- xl_hyperlink("https://example.com")
  expect_s3_class(x, "xl_cell_general")
})

test_that("xl_hyperlink() with no name stores HYPERLINK formula", {
  x <- xl_hyperlink(c("https://a.com", "https://b.com"))
  expect_equal(length(x), 2L)
  expect_equal(x[[1L]][["formula"]], '=HYPERLINK("https://a.com")')
  expect_equal(x[[2L]][["formula"]], '=HYPERLINK("https://b.com")')
})

test_that("xl_hyperlink() with name stores HYPERLINK formula with display text", {
  x <- xl_hyperlink(c("https://a.com", "https://b.com"), c("Site A", "Site B"))
  expect_equal(length(x), 2L)
  expect_equal(x[[1L]][["formula"]], '=HYPERLINK("https://a.com","Site A")')
  expect_equal(x[[2L]][["formula"]], '=HYPERLINK("https://b.com","Site B")')
})

test_that("xl_hyperlink() with NA url produces NA formula (blank cell)", {
  x <- xl_hyperlink(c("https://a.com", NA))
  expect_equal(length(x), 2L)
  expect_true(is.na(x[[2L]][["formula"]]))
})

test_that("xl_hyperlink() escapes internal double quotes in url and value", {
  x <- xl_hyperlink('http://example.com/q?a="1"', value = 'Say "hello"')
  expect_equal(x[[1L]][["formula"]],
               '=HYPERLINK("http://example.com/q?a=""1""","Say ""hello""")')
})

test_that("xl_hyperlink() writes valid xlsx", {
  df <- data.frame(name = c("UCLA", "Berkeley"))
  df$website <- xl_hyperlink(c("http://www.ucla.edu", "http://www.berkeley.edu"),
                               "homepage")
  expect_true(file.exists(write_xlsx(df)))
})

test_that("xl_hyperlink() with NA url writes valid xlsx", {
  df <- data.frame(x = c("a", "b"))
  df$link <- xl_hyperlink(c("http://a.com", NA))
  expect_true(file.exists(write_xlsx(df)))
})

# ── xl_hyperlink_cell() ───────────────────────────────────────────────────────

test_that("xl_hyperlink_cell() returns xl_cell_general with correct class", {
  x <- xl_hyperlink_cell("https://example.com")
  expect_s3_class(x, "xl_cell_general")
})

test_that("xl_hyperlink_cell() with no value stores URL as hyperlink field", {
  x <- xl_hyperlink_cell(c("https://a.com", "https://b.com"))
  expect_equal(length(x), 2L)
  expect_equal(x[[1L]][["hyperlink"]], "https://a.com")
  expect_equal(x[[2L]][["hyperlink"]], "https://b.com")
})

test_that("xl_hyperlink_cell() with value stores display text and hyperlink", {
  x <- xl_hyperlink_cell(c("https://a.com", "https://b.com"), value = c("A", "B"))
  expect_equal(x[[1L]][["value"]],     "A")
  expect_equal(x[[1L]][["hyperlink"]], "https://a.com")
  expect_equal(x[[2L]][["value"]],     "B")
  expect_equal(x[[2L]][["hyperlink"]], "https://b.com")
})

test_that("xl_hyperlink_cell() recycles value to url length", {
  x <- xl_hyperlink_cell(c("https://a.com", "https://b.com"), value = "Link")
  expect_equal(x[[1L]][["value"]], "Link")
  expect_equal(x[[2L]][["value"]], "Link")
})

test_that("xl_hyperlink_cell() with NA url produces blank cell", {
  x <- xl_hyperlink_cell(c("https://a.com", NA))
  expect_equal(length(x), 2L)
  expect_identical(x[[2L]][["hyperlink"]], NA)
})

test_that("xl_hyperlink_cell() with NA url and value: NA url clears value", {
  x <- xl_hyperlink_cell(c("https://a.com", NA), value = "Link")
  expect_equal(x[[1L]][["value"]], "Link")
  expect_true(is.na(x[[2L]][["value"]]))
})

test_that("xl_hyperlink_cell() writes valid xlsx", {
  df <- data.frame(name = c("UCLA", "Berkeley"))
  df$website <- xl_hyperlink_cell(
    c("http://www.ucla.edu", "http://www.berkeley.edu"),
    value = "homepage"
  )
  expect_true(file.exists(write_xlsx(df)))
})

test_that("xl_hyperlink_cell() with NA url writes valid xlsx", {
  df <- data.frame(x = c("a", "b"))
  df$link <- xl_hyperlink_cell(c("http://a.com", NA))
  expect_true(file.exists(write_xlsx(df)))
})

# ── normalize_df regression ───────────────────────────────────────────────────

test_that("normalize_df passes xl_cell_general columns through unchanged", {
  df <- data.frame(x = 1:3)
  df$c <- xl_cell_general(value = c(1, 2, 3))
  norm <- normalize_df(df)
  expect_s3_class(norm$c, "xl_cell_general")
})

test_that("xl_cell_general rejects value types writexl cannot write", {
  # the same predicate guards columns and cell objects, so an unsupported type
  # cannot slip in by being wrapped in a cell
  expect_error(xl_cell_general(value = complex(real = 1, imaginary = 1)),
               "cannot write these value type")
  expect_error(xl_cell_general(value = as.raw(1)), "raw")
  # inside a mixed-type list the offending element is identified by position
  expect_error(xl_cell_general(value = list(1, complex(real = 1, imaginary = 1))),
               "value[[2]]", fixed = TRUE)

  # supported values, including the blank-cell sentinels, are unaffected
  expect_s3_class(xl_cell_general(value = list(1.5, "text", TRUE)), "xl_cell_general")
  expect_s3_class(xl_cell_general(value = list(as.Date("2020-01-01"),
                                               as.POSIXct("2020-01-01", tz = "UTC"))),
                  "xl_cell_general")
  expect_s3_class(xl_cell_general(value = NA), "xl_cell_general")
  expect_s3_class(xl_cell_general(value = list(1, NULL)), "xl_cell_general")
  expect_s3_class(xl_cell_general(formula = "=1+1"), "xl_cell_general")
})

# ── Formatted blank cells ─────────────────────────────────────────────────────
#
# A cell with a format but no writable value must still exist, or the format is
# silently lost.  Excel ignores blank cells that carry no format, so the
# unformatted case must continue to write no cell at all.

# The style index of one cell, or NA_character_ if the cell is absent.
cell_style <- function(path, ref) {
  w <- xlsx_part(path, "xl/worksheets/sheet1.xml", raw = TRUE)
  m <- regmatches(w, regexec(sprintf('<c r="%s"([^>]*)>?', ref), w))[[1]]
  if (!length(m)) return(NA_character_)
  s <- regmatches(m[2L], regexec('s="([0-9]+)"', m[2L]))[[1]]
  if (length(s) == 2L) s[2L] else ""
}

test_that("an NA value with a format writes a formatted blank cell", {
  df <- data.frame(x = 1:2)
  df$b <- xl_cell_general(value = c(NA, NA),
                          format = xl_fill(background = "yellow"))
  p <- write_tmp(df)

  # the cell exists and carries a real (non-default) style
  expect_false(is.na(cell_style(p, "B2")))
  expect_true(nzchar(cell_style(p, "B2")))
  expect_true(as.integer(cell_style(p, "B2")) > 0L)

  # ... and it is genuinely blank: no value and no inline string
  w <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  blank <- regmatches(w, regexec('<c r="B2"[^>]*/>', w))[[1]]
  expect_length(blank, 1L)

  # the fill really reached styles.xml
  expect_match(xlsx_part(p, "xl/styles.xml", raw = TRUE), "FFFFFF00",
               ignore.case = TRUE)
})

test_that("an NA value with no format writes no cell at all", {
  df <- data.frame(x = 1:2)
  df$b <- xl_cell_general(value = c(NA, NA))
  expect_true(is.na(cell_style(write_tmp(df), "B2")))
})

test_that("every NA flavour with a format writes a formatted blank", {
  fmt <- xl_fill(background = "yellow")
  for (v in list(NA, NA_character_, NA_real_, NA_integer_, NaN,
                 as.Date(NA), as.POSIXct(NA_real_, origin = "1970-01-01",
                                         tz = "UTC"))) {
    df <- data.frame(x = 1L)
    df$b <- xl_cell_general(value = list(v), format = fmt)
    lbl <- paste(class(v), collapse = "/")
    expect_true(as.integer(cell_style(write_tmp(df), "B2")) > 0L, label = lbl)
  }
})

test_that("a formatted blank survives alongside a comment on the same cell", {
  df <- data.frame(x = 1L)
  df$b <- xl_cell_general(value = NA, comment = "note",
                          format = xl_fill(background = "yellow"))
  p <- write_tmp(df)
  expect_true(as.integer(cell_style(p, "B2")) > 0L)
  expect_match(xlsx_part(p, "xl/comments1.xml", raw = TRUE), "note")
})

test_that("NA in a plain column is left alone even under a column format", {
  # deliberate: Excel already applies the column format to empty cells, so
  # plain columns keep writing no cell for an NA
  sheet <- xl_sheet(data.frame(x = 1L, b = NA_character_),
                    cols = xl_col_spec("b", format = xl_fill(background = "yellow")))
  expect_true(is.na(cell_style(write_tmp(list(S = sheet)), "B2")))
})

# ── Array and dynamic formulas ────────────────────────────────────────────────

# The <sheetData> section of a written sheet, as a single string.
sheet_data <- function(path) {
  w <- xlsx_part(path, "xl/worksheets/sheet1.xml", raw = TRUE)
  substring(w, regexpr("<sheetData>", w, fixed = TRUE))
}

# A one-formula sheet: column A holds data, column B the formula cell.
array_sheet <- function(..., .cm = NA) {
  df <- data.frame(x = 1:2)
  df$f <- xl_cell_general(formula = c("=SUM(A1:A2)", NA), ...)
  write_tmp(df, constant_memory = .cm)
}

test_that("a single-cell array formula is stored as an array over its own cell", {
  s <- sheet_data(array_sheet(array = TRUE))
  expect_match(s, '<f t="array" ref="B2">SUM(A1:A2)</f>', fixed = TRUE)
})

test_that("a single-cell dynamic formula is marked as one and adds metadata", {
  p <- write_tmp(local({
    df <- data.frame(x = 1:2)
    df$f <- xl_cell_general(formula = c("=UNIQUE(A1:A2)", NA), dynamic = TRUE)
    df
  }))
  # cm="1" on the cell is what marks a dynamic array for Excel ...
  expect_match(sheet_data(p), '<c r="B2" cm="1">', fixed = TRUE)
  # ... and it requires the metadata part, which a plain array formula omits
  expect_false(is.null(xlsx_part(p, "xl/metadata.xml")))
  expect_true(is.null(xlsx_part(array_sheet(array = TRUE), "xl/metadata.xml")))
})

test_that("a pre-calculated result is stored for double and integer alike", {
  # the integer case used to fall through to the no-result writer, silently
  # dropping the value and leaving Excel's placeholder zero behind
  for (v in list(3, 3L)) {
    s <- sheet_data(array_sheet(array = TRUE, value = list(v, NA)))
    expect_match(s, '<f t="array" ref="B2">SUM(A1:A2)</f><v>3</v>',
                 fixed = TRUE, label = typeof(v))
  }
  # ... and the same fix applies to a plain (non-array) formula
  df <- data.frame(x = 1L)
  df$f <- xl_cell_general(formula = "=A1", value = 7L)
  expect_match(sheet_data(write_tmp(df)), "<v>7</v>", fixed = TRUE)
})

test_that("a declared multi-cell range is written and padded", {
  df <- data.frame(x = 1:2)
  df$f <- xl_cell_general(formula = c("=TRANSPOSE(A1:A2)", NA), array = TRUE,
                          array_range = list("B2:D2", NA))
  p <- write_tmp(df)
  s <- sheet_data(p)
  expect_match(s, 'ref="B2:D2"', fixed = TRUE)
  # the cells beyond the anchor exist -- libxlsxwriter pads them
  expect_match(s, '<c r="C2">', fixed = TRUE)
  expect_match(s, '<c r="D2">', fixed = TRUE)
})

test_that("a multi-cell range turns row streaming off, and says why", {
  df <- data.frame(x = 1:2)
  df$f <- xl_cell_general(formula = c("=TRANSPOSE(A1:A2)", NA), array = TRUE,
                          array_range = list("B2:D2", NA))
  d <- .resolve_sheet_formats(df, .new_format_registry(), xl_properties(), 1L)
  cm <- .resolve_constant_memory(list(d), xl_properties())
  expect_equal(cm$on, 0L)
  expect_length(cm$reasons, 1L)
  expect_match(cm$reasons, "multi-cell array formula range")

  # observable end to end: with streaming off, strings move to the shared table
  expect_false(is.null(xlsx_part(write_tmp(df), "xl/sharedStrings.xml")))
  # a single-cell array formula leaves streaming on
  # asked for explicitly: a frame this small would not stream on size alone
  expect_true(is.null(xlsx_part(array_sheet(array = TRUE,
                                            .cm = TRUE),
                                "xl/sharedStrings.xml")))
})

test_that("array flags are inert on a cell with no formula", {
  # recycling one TRUE across a column of mixed formula and value cells is
  # normal, so a flag on a value-only cell must simply do nothing
  df <- data.frame(x = 1:2)
  df$f <- xl_cell_general(formula = c("=SUM(A1:A2)", NA),
                          value = list(NA, 5), array = TRUE)
  s <- sheet_data(write_tmp(df))
  expect_match(s, 'ref="B2"', fixed = TRUE)
  expect_match(s, '<c r="B3"><v>5</v></c>', fixed = TRUE)
})

test_that("array arguments are validated", {
  # a flag that can never apply to anything
  expect_error(xl_cell_general(value = 1, array = TRUE), "no `formula`")
  expect_error(xl_cell_general(value = 1, dynamic = TRUE), "no `formula`")
  # a range with nothing to apply to
  expect_error(xl_cell_general(formula = "=A1", array_range = "A1:B2"),
               "applies only to an `array` or `dynamic`")
  # Excel stores no cached string result for an array formula
  expect_error(xl_cell_general(formula = "=A1", value = "txt", array = TRUE),
               "character `value`")
  expect_error(xl_cell_general(formula = "=A1", value = "txt", dynamic = TRUE),
               "character `value`")
  # a character value is still fine for a plain formula
  expect_s3_class(xl_cell_general(formula = "=A1", value = "txt"),
                  "xl_cell_general")
  # the flags are decisions, so NA is not a valid setting
  expect_error(xl_cell_general(formula = "=A1", array = NA), "TRUE or FALSE")
  expect_error(xl_cell_general(formula = "=A1", dynamic = NA), "TRUE or FALSE")
  # an unusable range spelling
  expect_error(xl_cell_general(formula = "=A1", array = TRUE,
                               array_range = list(1L)),
               "must be NA, an Excel range string")
})

test_that("a declared range must start at the cell holding the formula", {
  df <- data.frame(x = 1:2)
  df$f <- xl_cell_general(formula = c("=A1", NA), array = TRUE,
                          array_range = list("C5:D6", NA))
  expect_error(write_xlsx(df), "must start at the cell that holds the formula")
})

test_that("a declared range may not overlap cells the sheet writes itself", {
  df <- data.frame(x = 1:3)
  # B2:B4 is column B for all three data rows -- cells this sheet writes
  df$f <- xl_cell_general(formula = c("=A1", NA, NA), array = TRUE,
                          array_range = list("B2:B4", NA, NA))
  expect_error(write_xlsx(df), "that the sheet writes itself")
  expect_error(write_xlsx(df), "covers 3 cell", fixed = TRUE)
  expect_error(write_xlsx(df), "spills automatically")
})

test_that("array_range accepts a data-frame-relative spec", {
  df <- data.frame(x = 1:2)
  # cols 2:3 does not exist in a 2-column frame, so name the range in A1 terms
  # and check the list spelling on a frame wide enough to hold it
  df$f <- xl_cell_general(formula = c("=A1", NA), array = TRUE)
  expect_silent(write_xlsx(df))

  wide <- data.frame(a = 1:2, b = 1:2, cc = 1:2)
  wide$f <- xl_cell_general(formula = c("=A1", NA), array = TRUE,
                            array_range = list(list(rows = 1, cols = 4), NA))
  # a single-cell spec resolves to the anchor and needs no memory-mode change
  expect_silent(write_xlsx(wide))
})

# ── xl_hyperlink(): name deprecated in favour of value ────────────────────────

test_that("xl_hyperlink(value =) is the supported spelling", {
  x <- xl_hyperlink("https://a.com", value = "Site A")
  expect_match(x[[1L]][["formula"]], 'HYPERLINK("https://a.com","Site A")',
               fixed = TRUE)
})

test_that("xl_hyperlink(name =) still works but warns", {
  expect_warning(x <- xl_hyperlink("https://a.com", name = "Site A"),
                 "deprecated")
  expect_warning(xl_hyperlink("https://a.com", name = "Site A"), "use `value`")
  # and it produces exactly what value= produces
  expect_equal(unclass(x),
               unclass(xl_hyperlink("https://a.com", value = "Site A")))
})

test_that("giving both value and name is an error", {
  expect_error(xl_hyperlink("https://a.com", value = "A", name = "B"),
               "not both")
})

test_that("positional display text is unaffected by the rename", {
  # `value` took the position `name` used to occupy, so old positional code
  # keeps working and must not warn
  expect_silent(x <- xl_hyperlink("https://a.com", "Site A"))
  expect_match(x[[1L]][["formula"]], 'HYPERLINK("https://a.com","Site A")',
               fixed = TRUE)
  # ... and a positional format argument still lands on format
  y <- xl_hyperlink("https://a.com", "Site A", xl_font(bold = TRUE))
  expect_true(is_xl_format(y[[1L]][["format"]]))
})

# ── The guards nothing else reached ──────────────────────────────────────────

test_that("a single array_range spec is one range, not a list of them", {
  # list(rows = , cols = ) has to be wrapped, or its two elements would be
  # taken as two cells' worth of ranges
  cells <- xl_cell_general(formula = "=A1:B2", array = TRUE,
                           array_range = list(rows = 1:2, cols = 1:2))
  expect_length(cells, 1L)
  expect_equal(cells[[1L]]$array_range, list(rows = 1:2, cols = 1:2))
})

test_that("a NULL hyperlink in a list is a cell with no hyperlink", {
  cells <- xl_cell_general(value = c("a", "b"),
                           hyperlink = list(NULL, "https://example.com"))
  expect_true(is.na(cells[[1L]]$hyperlink) || is.null(cells[[1L]]$hyperlink))
  expect_equal(cells[[2L]]$hyperlink, "https://example.com")
})

test_that("an unset array flag is FALSE for every cell", {
  expect_equal(.cell_flag_vec(NULL, "array", 3L), rep(FALSE, 3L))
  expect_error(.cell_flag_vec(NA, "array", 1L), "must be TRUE or FALSE")
})

test_that("an array range reaching past the sheet's data is accepted", {
  # the overlap check only has something to say where the two rectangles meet;
  # the range starts at its own cell (B2) and spills into empty columns
  df <- data.frame(a = 1:2)
  df$f <- xl_cell_general(formula = c("=TRANSPOSE(A2:A3)", NA), array = TRUE,
                          array_range = list("B2:E2", NA))
  expect_silent(write_tmp(df))
})

test_that("a frame with no rows has no data rectangle to overlap", {
  empty <- data.frame(a = numeric(0))
  empty$f <- xl_cell_general(formula = character(0), array = TRUE)
  expect_silent(write_tmp(empty))
})

test_that("a dynamic formula carries its cached result, spilling or not", {
  read_back <- function(cells) {
    df <- data.frame(a = c(1, 2, 3))
    df$f <- cells
    p <- write_tmp(df)
    list(sheet = xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE),
         got = as.data.frame(readxl::read_xlsx(p)))
  }
  # single cell, with the value Excel would compute
  r <- read_back(xl_cell_general(value = c(6, NA, NA),
                                 formula = c("=SUM(A2:A4)", NA, NA),
                                 dynamic = TRUE))
  expect_equal(r$got$f[1L], 6)

  # a declared range spills into columns the sheet does not write, again with
  # a cached value for the anchor cell
  r <- read_back(xl_cell_general(value = c(1, NA, NA),
                                 formula = c("=TRANSPOSE(A2:A4)", NA, NA),
                                 dynamic = TRUE,
                                 array_range = list("B2:D2", NA, NA)))
  expect_match(r$sheet, 'cm="1"', fixed = TRUE)
  expect_equal(r$got$f[1L], 1)

  # and without a cached value, on both paths
  r <- read_back(xl_cell_general(formula = c("=SUM(A2:A4)", NA, NA),
                                 dynamic = TRUE))
  expect_match(r$sheet, "SUM(A2:A4)", fixed = TRUE)
  r <- read_back(xl_cell_general(formula = c("=TRANSPOSE(A2:A4)", NA, NA),
                                 dynamic = TRUE,
                                 array_range = list("B2:D2", NA, NA)))
  expect_match(r$sheet, "TRANSPOSE", fixed = TRUE)
})

# Tests for the shared format engine (R/xl_format.R)

test_that("xl_color accepts names, hex, and integers", {
  expect_equal(xl_color("red"), 0xFF0000)
  expect_equal(xl_color("navy"), xl_color("#000080"))
  expect_equal(xl_color("#0000FF"), 0x0000FF)
  expect_equal(xl_color("00FF00"), 0x00FF00)   # bare 6-hex
  expect_equal(xl_color(255L), 255L)
  expect_equal(xl_color(0xFFFFFF), 0xFFFFFF)
  expect_identical(xl_color(NA), NA_integer_)
})

test_that("xl_color rejects bad input", {
  expect_error(xl_color("notacolor"), "Unknown color")
  expect_error(xl_color(c("red", "blue")), "single color")
  expect_error(xl_color(-1L), "range")
  expect_error(xl_color(0xFFFFFF + 1L), "range")
})

test_that("xl_font sets and validates properties", {
  f <- xl_font(name = "Arial", size = 14, color = "red", bold = TRUE,
               italic = TRUE, underline = "double", strikeout = TRUE,
               script = "super")
  fo <- unclass(f)$font
  expect_equal(fo$name, "Arial")
  expect_equal(fo$size, 14)
  expect_equal(fo$color, xl_color("red"))
  expect_true(fo$bold)
  expect_equal(fo$underline, "double")
  expect_equal(fo$script, "super")
  expect_true(is_xl_format(f))
})

test_that("xl_font handles rarely-used properties", {
  f <- xl_font(family = 2, charset = 1, outline = TRUE, shadow = TRUE,
               condense = TRUE, extend = TRUE, scheme = "minor", theme = 1,
               color_indexed = 8, font_only = TRUE)
  p <- .xl_format_payload(f)
  expect_equal(p$font_family, 2L)
  expect_equal(p$font_charset, 1L)
  expect_true(p$font_outline)
  expect_true(p$font_shadow)
  expect_true(p$font_condense)
  expect_true(p$font_extend)
  expect_equal(p$font_scheme, "minor")
  expect_equal(p$theme, 1L)
  expect_equal(p$color_indexed, 8L)
  expect_true(p$font_only)
})

test_that("xl_font validation errors are typed and specific", {
  expect_error(xl_font(underline = "wiggly"), "underline")
  expect_error(xl_font(size = 0), "between 1 and 409")
  expect_error(xl_font(size = 500), "between 1 and 409")
  expect_error(xl_font(bold = "yes"), "bold")
  expect_error(xl_font(script = "middle"), "script")
  expect_error(xl_font(name = 1), "name")
  expect_error(xl_font(color = "notacolor"), "color")
})

test_that("empty xl_font is an empty format", {
  f <- xl_font()
  expect_true(is_xl_format(f))
  expect_null(unclass(f)$font)
})

test_that("xl_fill maps background/foreground/pattern and auto-solids", {
  f <- xl_fill(background = "yellow")
  expect_equal(unclass(f)$fill$pattern, "solid")   # auto solid
  f2 <- xl_fill(background = "yellow", foreground = "black", pattern = "light-gray")
  fi <- unclass(f2)$fill
  expect_equal(fi$pattern, "light-gray")            # explicit pattern preserved
  p <- .xl_format_payload(f2)
  expect_equal(p$bg_color, xl_color("yellow"))
  expect_equal(p$fg_color, xl_color("black"))
  expect_equal(p$pattern, 4L)
  expect_error(xl_fill(pattern = "sparkle"), "pattern")
})

test_that("xl_border 'all' and 'color' fan out to four sides", {
  b <- xl_border(all = "thin", color = "gray")
  u <- unclass(b)$border
  expect_equal(u$left, "thin"); expect_equal(u$right, "thin")
  expect_equal(u$top, "thin");  expect_equal(u$bottom, "thin")
  expect_equal(u$left_color, xl_color("gray"))
  # per-side overrides 'all'
  b2 <- xl_border(all = "thin", top = "thick")
  expect_equal(unclass(b2)$border$top, "thick")
  expect_equal(unclass(b2)$border$bottom, "thin")
})

test_that("xl_border diagonal and payload", {
  b <- xl_border(diagonal = "up-down", diagonal_style = "hair",
                 diagonal_color = "red", bottom = "double", bottom_color = "blue")
  p <- .xl_format_payload(b)
  expect_equal(p$diag_type, 3L)
  expect_equal(p$diag_border, 7L)
  expect_equal(p$diag_color, xl_color("red"))
  expect_equal(p$border_bottom, 6L)
  expect_equal(p$bottom_color, xl_color("blue"))
  expect_error(xl_border(all = "wavy"), "all")
  expect_error(xl_border(diagonal = "sideways"), "diagonal")
})

test_that("xl_align validates and maps alignment", {
  a <- xl_align(horizontal = "center", vertical = "top", wrap = TRUE,
                rotation = 45, indent = 2, shrink = TRUE, reading_order = "rtl")
  p <- .xl_format_payload(a)
  expect_equal(p$align_h, 2L)
  expect_equal(p$align_v, 8L)
  expect_true(p$text_wrap)
  expect_equal(p$rotation, 45L)
  expect_equal(p$indent, 2L)
  expect_true(p$shrink)
  expect_equal(p$reading_order, 2L)
  expect_equal(.xl_format_payload(xl_align(rotation = 270))$rotation, 270L)
  expect_error(xl_align(horizontal = "middle"), "horizontal")
  expect_error(xl_align(vertical = "sideways"), "vertical")
  expect_error(xl_align(rotation = 120), "270")
})

test_that("xl_num_format takes a string or an index", {
  expect_equal(.xl_format_payload(xl_num_format("#,##0.00"))$num_format, "#,##0.00")
  expect_equal(.xl_format_payload(xl_num_format(index = 3))$num_format_index, 3L)
  expect_error(xl_num_format(format = 1), "format")
  expect_error(xl_num_format(index = 999), "between 0 and 255")
})

test_that("xl_protection inverts locked and sets hidden", {
  p <- .xl_format_payload(xl_protection(locked = FALSE, hidden = TRUE))
  expect_true(p$unlocked)
  expect_true(p$hidden)
  # locked = TRUE is Excel's default -> no unlocked call
  expect_null(.xl_format_payload(xl_protection(locked = TRUE))$unlocked)
  expect_error(xl_protection(locked = "no"), "locked")
})

test_that("xl_format merges groups property-by-property, right-biased", {
  f <- xl_format(xl_font(bold = TRUE), xl_border(bottom = "thin"),
                 xl_num_format("#,##0.00"))
  u <- unclass(f)
  expect_true(u$font$bold)
  expect_equal(u$border$bottom, "thin")
  expect_equal(u$num_format$format, "#,##0.00")
  # accumulate within a group
  f2 <- xl_font(bold = TRUE) + xl_font(italic = TRUE)
  expect_true(unclass(f2)$font$bold)
  expect_true(unclass(f2)$font$italic)
  # right operand wins on conflict
  f3 <- xl_font(color = "red") + xl_font(color = "blue")
  expect_equal(unclass(f3)$font$color, xl_color("blue"))
})

test_that("xl_format scalar flags and NULL handling", {
  f <- xl_format(xl_font(bold = TRUE), quote_prefix = TRUE, hyperlink = TRUE)
  p <- .xl_format_payload(f)
  expect_true(p$quote_prefix)
  expect_true(p$set_hyperlink)
  expect_true(is_xl_format(xl_format()))              # empty
  expect_true(is_xl_format(xl_format(NULL, xl_font(bold = TRUE), NULL)))
  expect_error(xl_format("not a format"), "must be xl_format")
})

test_that("+ operator handles NULL and rejects non-formats", {
  f <- xl_font(bold = TRUE)
  expect_identical(f + NULL, f)
  expect_identical(NULL + f, f)
  expect_error(f + 1, "must be xl_format")
})

test_that("print methods run for coverage", {
  expect_output(print(xl_font(bold = TRUE, color = "red")), "font")
  expect_output(print(xl_format()), "empty")
  expect_output(print(xl_format(xl_font(bold = TRUE), quote_prefix = TRUE)), "quote_prefix")
})

test_that("format registry deduplicates and reserves 0 for none", {
  reg <- .new_format_registry()
  i1 <- .register_format(reg, xl_font(bold = TRUE))
  i2 <- .register_format(reg, xl_font(bold = TRUE))
  i3 <- .register_format(reg, xl_font(italic = TRUE))
  i0 <- .register_format(reg, NULL)
  iE <- .register_format(reg, xl_format())    # empty -> 0
  expect_equal(i1, 1L)
  expect_equal(i2, 1L)
  expect_equal(i3, 2L)
  expect_equal(i0, 0L)
  expect_equal(iE, 0L)
  expect_length(reg$table, 2L)
})

test_that("full font payload round-trips every property", {
  f <- xl_font(name = "Arial", size = 12, color = "red", bold = TRUE,
               italic = TRUE, underline = "single", strikeout = TRUE,
               script = "sub")
  p <- .xl_format_payload(f)
  expect_equal(p$font_name, "Arial")
  expect_equal(p$font_size, 12)
  expect_equal(p$font_color, xl_color("red"))
  expect_true(p$bold); expect_true(p$italic); expect_true(p$strikeout)
  expect_equal(p$underline, 1L)
  expect_equal(p$script, 2L)
})

test_that("full border payload round-trips every side and color", {
  b <- xl_border(left = "thin", right = "medium", top = "thick", bottom = "double",
                 left_color = "red", right_color = "green", top_color = "blue",
                 bottom_color = "black")
  p <- .xl_format_payload(b)
  expect_equal(p$border_left, 1L)
  expect_equal(p$border_right, 2L)
  expect_equal(p$border_top, 5L)
  expect_equal(p$border_bottom, 6L)
  expect_equal(p$left_color, xl_color("red"))
  expect_equal(p$right_color, xl_color("green"))
  expect_equal(p$top_color, xl_color("blue"))
  expect_equal(p$bottom_color, xl_color("black"))
})

test_that("numeric validators reject non-numeric and non-scalar input", {
  expect_error(xl_font(size = "big"), "single number")
  expect_error(xl_font(size = c(1, 2)), "single number")
  expect_error(xl_font(family = "two"), "single number")
  expect_error(xl_align(indent = -1), "between 0")
  expect_error(xl_align(rotation = -200), "270")
})

test_that("merge_xl_format tolerates NULL operands directly", {
  f <- xl_font(bold = TRUE)
  expect_identical(merge_xl_format(NULL, f), f)
  expect_identical(merge_xl_format(f, NULL), f)
})

test_that(".fmt_kv tolerates NULL values", {
  expect_type(.fmt_kv(list(a = NULL)), "character")
})

test_that("payload key is order-independent", {
  a <- .payload_key(list(bold = TRUE, italic = TRUE))
  b <- .payload_key(list(italic = TRUE, bold = TRUE))
  expect_equal(a, b)
  expect_equal(.payload_key(list()), "")
})

# Merged cells.  worksheet_merge_range() writes the top-left text itself and
# blanks the rest, so a merge carries its own content and is applied after the
# sheet's rows.

merged <- function(..., df = data.frame(a = 1:3, b = 4:6)) {
  x <- xlsx_part(write_tmp(list(D = xl_sheet(df, merge = list(...)))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  m <- regmatches(x, regexpr("<mergeCells.*?</mergeCells>", x))
  if (!length(m)) NA_character_ else m
}

# ── Construction ──────────────────────────────────────────────────────────────

test_that("xl_merge validates its arguments", {
  expect_error(xl_merge(), "must name the cells")
  expect_error(xl_merge("A1:B1", value = 1), "single non-NA string")
  expect_error(xl_merge("A1:B1", value = c("a", "b")), "single non-NA string")
  expect_error(xl_merge("A1:B1", format = "bold"), "must be an xl_format")
  expect_s3_class(xl_merge("A1:B1"), "xl_merge")
  expect_s3_class(xl_merge("A1:B1", "T", xl_font(bold = TRUE)), "xl_merge")
})

test_that("the print method runs", {
  expect_output(print(xl_merge("A1:C1", "Title")), "xl_merge")
  expect_output(print(xl_merge("A1:C1", "Title")), "A1:C1", fixed = TRUE)
})

# ── Writing ───────────────────────────────────────────────────────────────────

test_that("a merge is recorded in mergeCells", {
  expect_match(merged(xl_merge("A5:B5", "Total")), '<mergeCell ref="A5:B5"/>',
               fixed = TRUE)
})

test_that("several merges are all recorded", {
  x <- merged(xl_merge("A5:B5", "Total"), xl_merge("A6:B6", "Note"))
  expect_match(x, 'count="2"', fixed = TRUE)
  expect_match(x, 'ref="A5:B5"', fixed = TRUE)
  expect_match(x, 'ref="A6:B6"', fixed = TRUE)
})

test_that("a single xl_merge need not be wrapped in a list", {
  x <- xlsx_part(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                             merge = xl_merge("A5:B5", "T")))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(x, 'ref="A5:B5"', fixed = TRUE)
})

test_that("the merged text is written and readable", {
  skip_if_not_installed("readxl")
  p <- write_tmp(list(D = xl_sheet(data.frame(a = 1:3, b = 4:6),
                                   merge = xl_merge("A5:B5", "Total"))))
  rd <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(rd$a[4], "Total")
})

test_that("a merge over written cells keeps only the merged value", {
  # this is Excel's own behaviour: merging discards everything but the
  # top-left value, and it is why merges are applied after the rows
  skip_if_not_installed("readxl")
  p <- write_tmp(list(D = xl_sheet(data.frame(a = 1:3, b = 4:6),
                                   merge = xl_merge("A2:B2", "Merged"))))
  rd <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(rd$a[1], "Merged")   # replaced the 1 that was there
  expect_equal(rd$a[2], "2")        # later rows untouched
})

test_that("a merge format reaches styles.xml", {
  p <- write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                   merge = xl_merge("A5:B5", "Total",
                                     format = xl_fill(background = "yellow")))))
  expect_match(xlsx_part(p, "xl/styles.xml", raw = TRUE), "FFFFFF00",
               ignore.case = TRUE)
})

test_that("a merge with no text writes an empty merged cell", {
  expect_match(merged(xl_merge("A5:B5")), 'ref="A5:B5"', fixed = TRUE)
})

test_that("a merge accepts a data-frame-relative spec", {
  df <- data.frame(a = 1:3, b = 4:6)
  x <- merged(xl_merge(list(rows = 1, cols = c("a", "b")), "Merged"), df = df)
  # data row 1 is sheet row 2 once the header is counted
  expect_match(x, 'ref="A2:B2"', fixed = TRUE)
})

# ── Validation ────────────────────────────────────────────────────────────────

test_that("a single-cell merge is refused", {
  # libxlsxwriter returns a bare parameter-validation error for this, so the
  # message is raised in R where it can name the merge
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                           merge = xl_merge("A5:A5", "x")))),
               "single cell")
  # a range string that is not a rectangle is rejected by the range parser
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                           merge = xl_merge("A5", "x")))),
               "merge")
})

test_that("non-merge objects in the list are rejected", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1),
                                           merge = list("A1:B1")))),
               "must be an xl_merge object")
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1), merge = "A1:B1"))),
               "must be an xl_merge object")
})

test_that("an inverted merge range is rejected by the shared parser", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
                                           merge = xl_merge("B5:A5", "x")))),
               "inverted")
})

# ── Memory mode ───────────────────────────────────────────────────────────────

test_that("a merge turns row streaming off, and says why", {
  df <- data.frame(a = 1:3, b = 4:6)
  sheets <- list(D = xl_sheet(df, merge = xl_merge("A5:B5", "Total")))
  plan <- Map(function(el, d) .resolve_sheet_plan(el, d, .new_format_registry(),
                                                  1L, xl_properties()),
              sheets, list(df))
  cm <- .resolve_constant_memory(list(df), xl_properties(), plan)
  expect_equal(cm$on, 0L)
  expect_match(cm$reasons, "merged range")

  # observable end to end: with streaming off, strings go to the shared table
  p <- write_tmp(sheets)
  expect_false(is.null(xlsx_part(p, "xl/sharedStrings.xml")))
  # ... and a sheet with no merge keeps streaming on
  expect_true(is.null(xlsx_part(write_tmp(list(D = xl_sheet(df)),
                                          constant_memory = TRUE),
                                "xl/sharedStrings.xml")))
})

# ── Coercing the merged value ─────────────────────────────────────────────────

test_that("a merge takes its value from anything that can render itself", {
  # a cell built for a sheet, reused where only its displayed value matters
  expect_equal(xl_merge("A1:B1", xl_cell_general(value = "Total"))$value,
               "Total")
  expect_equal(xl_merge("A1:B1", factor("Total"))$value, "Total")
  # a rich string is refused rather than silently flattened: worksheet_merge_
  # range() writes one run, so the per-run fonts would be dropped
  rs <- xl_rich_string("a", xl_rich_run("b", xl_font(bold = TRUE)))
  expect_error(xl_merge("A1:B1", rs), "cannot be a rich string")
  expect_error(xl_merge("A1:B1", xl_cell_general(value = rs)),
               "cannot be a rich string")
  expect_error(xl_merge("A1:B1", xl_cell_general(value = c("a", "b"))),
               "single non-NA string")
})

# Translating an ordinary xl_format into the line / fill / pattern / font a
# chart understands.  The translation is lossy in known ways, and every loss is
# an error -- these tests exist mostly to pin that.

cf <- function(f) .chart_format_payload(f, "format")

# ── What translates ───────────────────────────────────────────────────────────

test_that("a font maps to the chart font's five properties", {
  # lxw_chart_font carries size, bold, italic, underline and colour -- and
  # nothing else, which is why the typeface is refused below
  ent <- cf(xl_font(bold = TRUE, italic = TRUE, size = 12, color = "navy",
                    underline = "single"))$font
  expect_equal(ent$size, 12)
  expect_equal(ent$bold, 1L)
  expect_equal(ent$italic, 1L)
  expect_equal(ent$underline, 1L)
  expect_equal(ent$color, xl_color("navy"))
})

test_that("Excel's five underline styles collapse to a flag", {
  # the chart font has a boolean, not a choice, so anything but "none" is on
  for (u in c("single", "double", "single-accounting", "double-accounting"))
    expect_equal(cf(xl_font(underline = u))$font$underline, 1L, label = u)
  expect_null(cf(xl_font(underline = "none", bold = TRUE))$font$underline)
})

test_that("a plain fill maps to the chart fill, with transparency", {
  ent <- cf(xl_fill(background = "red", transparency = 40))$fill
  expect_equal(ent$color, xl_color("red"))
  expect_equal(ent$transparency, 40L)
  # and "none" means no fill rather than a colour
  expect_equal(cf(xl_fill(pattern = "none"))$fill$none, 1L)
})

test_that("a real pattern maps to the chart pattern, not the fill", {
  p <- cf(xl_fill(pattern = "dark-horizontal", foreground = "red",
                  background = "white"))
  expect_null(p$fill)
  expect_equal(p$pattern$fg_color, xl_color("red"))
  expect_equal(p$pattern$bg_color, xl_color("white"))
  # the number is LXW_CHART_PATTERN_DARK_HORIZONTAL, not the cell enum's
  # LXW_PATTERN_DARK_HORIZONTAL -- two different vocabularies at the same names
  expect_equal(p$pattern$type, 24L)
  expect_false(identical(p$pattern$type, unname(.LXW$pattern[["dark-horizontal"]])))
})

test_that("every chart pattern name maps to its own place in the chart enum", {
  # the cell and chart pattern enums overlap only in these names, and a
  # wrong index here draws a silently different pattern
  expect_equal(vapply(names(.CHART_PATTERN), function(nm)
    cf(xl_fill(pattern = nm, foreground = "red", background = "white"))$pattern$type,
    integer(1), USE.NAMES = FALSE), unname(.CHART_PATTERN))
  expect_false(any(duplicated(.CHART_PATTERN)))
})

test_that("a cell pattern with no chart counterpart is refused, not guessed", {
  for (nm in setdiff(names(.LXW$pattern), c(names(.CHART_PATTERN), "none", "solid")))
    expect_error(cf(xl_fill(pattern = nm, foreground = "red",
                            background = "white")),
                 "cannot be drawn on a chart")
  expect_error(cf(xl_fill(pattern = "gray-125", foreground = "red",
                          background = "white")),
               "Chart patterns accept")
})

test_that("a border maps to the chart line, with transparency", {
  ent <- cf(xl_border(all = "dashed", color = "gray", transparency = 25))$line
  expect_equal(ent$dash_type, 3L)
  expect_equal(ent$color, xl_color("gray"))
  expect_equal(ent$transparency, 25L)
  expect_equal(cf(xl_border(all = "none"))$line$none, 1L)
})

test_that("the four supported border styles map to their exact dash types", {
  # only exact equivalents are mapped; the table is short on purpose, since the
  # alternative is approximating a style the user asked for
  expect_equal(cf(xl_border(all = "thin"))$line$dash_type, 0L)      # solid
  expect_equal(cf(xl_border(all = "dashed"))$line$dash_type, 3L)
  expect_equal(cf(xl_border(all = "dash-dot"))$line$dash_type, 4L)
  expect_equal(cf(xl_border(all = "dotted"))$line$dash_type, 8L)
  expect_equal(length(.CHART_DASH), 4L)
})

test_that("groups combine into one payload", {
  p <- cf(xl_font(bold = TRUE) + xl_fill(background = "yellow") +
            xl_border(all = "thin", color = "black"))
  expect_named(p, c("font", "fill", "line"), ignore.order = TRUE)
})

# ── What is refused ───────────────────────────────────────────────────────────

test_that("every border style a chart cannot draw is refused, with a reason", {
  # enumerated: each of the 14 cell styles is either mapped or refused, so a
  # style added to writexl cannot silently fall through
  mapped <- names(.CHART_DASH)
  refused <- names(.CHART_DASH_UNSUPPORTED)
  all_styles <- setdiff(names(writexl:::.LXW$border), "none")
  expect_setequal(c(mapped, refused), all_styles)

  for (st in refused)
    expect_error(cf(xl_border(all = st)), "cannot be drawn on a chart",
                 label = st)
  # the width cases say so, and the pattern cases say something different
  expect_error(cf(xl_border(all = "thick")), "no width")
  expect_error(cf(xl_border(all = "slant-dash-dot")),
               "no chart dash pattern matching it")
  # and the message lists what would have worked
  expect_error(cf(xl_border(all = "medium")), "Chart lines accept")
})

test_that("the format groups a chart shape has no concept of are refused", {
  expect_error(cf(xl_num_format("0.0")), "set it on the axis or the data labels")
  expect_error(cf(xl_align(horizontal = "center")),
               "a chart shape places its own text")
  expect_error(cf(xl_align(vertical = "top", rotation = 45)),
               "takes only `rotation`", fixed = TRUE)
  expect_error(cf(xl_protection(locked = FALSE)), "cannot be locked or hidden")
})

test_that("a font name and a rotation reach the chart font", {
  expect_equal(cf(xl_font(name = "Arial"))$font$name, "Arial")
  expect_equal(cf(xl_align(rotation = -45))$font$rotation, -45L)
  expect_equal(cf(xl_align(rotation = 270))$font$rotation, 270L)
  # lxw_chart_font reads a rotation of 0 as unset, so an explicit upright
  # travels as libxlsxwriter's 360
  expect_equal(cf(xl_align(rotation = 0))$font$rotation, 360L)
  # and a rotation alone is a font, even with no xl_font() beside it
  expect_equal(names(cf(xl_align(rotation = 30))), "font")
})

test_that("a chart has one line, so differing sides are refused", {
  expect_error(cf(xl_border(left = "thin", right = "dashed")),
               "a chart has one line")
  expect_error(cf(xl_border(all = "thin", left_color = "red",
                            right_color = "blue")),
               "a chart has one line")
  # all four the same is fine
  expect_silent(cf(xl_border(all = "thin", color = "red")))
})

test_that("a pattern without both colours is refused", {
  expect_error(cf(xl_fill(pattern = "gray-125")), "give both")
  expect_error(cf(xl_fill(pattern = "gray-125", foreground = "red")),
               "give both")
})

test_that("an empty or non-format argument is refused", {
  expect_error(cf(xl_format()), "is an empty format")
  expect_error(cf("red"), "must be an xl_format object")
  expect_null(.chart_format_payload(NULL, "format"))
})

# ── transparency reaches the format objects ───────────────────────────────────

test_that("transparency is carried by xl_fill and xl_border", {
  expect_equal(unclass(xl_fill(background = "red",
                               transparency = 40))$fill$transparency, 40L)
  expect_equal(unclass(xl_border(all = "thin",
                                 transparency = 25))$border$transparency, 25L)
  expect_error(xl_fill(background = "red", transparency = 101),
               "`transparency` must be between 0 and 100")
  expect_error(xl_border(all = "thin", transparency = -1),
               "`transparency` must be between 0 and 100")
  # NA means unset, as everywhere else
  expect_null(unclass(xl_fill(background = "red"))$fill$transparency)
})

test_that("transparency does not disturb a cell format", {
  # it is inert for cells -- Excel has no transparency there -- so it must not
  # change the styles a normal write produces
  df <- data.frame(a = 1:2)
  plain <- styles_string(list(D = xl_sheet(df,
    cols = xl_col_spec("a", format = xl_fill(background = "red")))))
  with_t <- styles_string(list(D = xl_sheet(df,
    cols = xl_col_spec("a", format = xl_fill(background = "red",
                                             transparency = 40)))))
  expect_equal(plain, with_t)
})

# ── What each chart part will accept ──────────────────────────────────────────

test_that("a series takes no font, and says where the font belongs", {
  # libxlsxwriter's series has a line, a fill and a pattern but no font: a
  # series is a shape, not text.  Dropping the font silently would leave the
  # caller with an unstyled chart and no reason why.
  expect_error(.chart_format_payload(xl_font(bold = TRUE), "format",
                                     accept = c("line", "fill", "pattern"),
                                     part = "a chart series"),
               "a shape and has no text")
  expect_error(.chart_format_payload(xl_font(bold = TRUE), "format",
                                     accept = c("line", "fill", "pattern"),
                                     part = "a chart series"),
               "xl_border() or xl_fill()", fixed = TRUE)
  # the rest of the same format still translates
  p <- .chart_format_payload(xl_fill(background = "red"), "format",
                             accept = c("line", "fill", "pattern"),
                             part = "a chart series")
  expect_equal(p$fill$color, xl_color("red"))
})

test_that("a title takes only a font, and says so for each other group", {
  ttl <- function(f) .chart_format_payload(f, "title_format", accept = "font",
                                           part = "a chart title")
  expect_equal(ttl(xl_font(size = 14))$font$size, 14)
  expect_error(ttl(xl_border(all = "thin")), "takes no line")
  expect_error(ttl(xl_fill(background = "red")), "takes no fill")
  expect_error(ttl(xl_fill(pattern = "dark-up", foreground = "red",
                           background = "white")), "takes no fill")
  expect_error(ttl(xl_border(all = "thin")), "Use xl_font()", fixed = TRUE)
})

test_that("an empty format names the groups the part could use", {
  expect_error(.chart_format_payload(xl_format(), "title_format",
                                     accept = "font", part = "a chart title"),
               "give xl_font()", fixed = TRUE)
})

test_that("black reaches a chart, rather than reading as no colour at all", {
  # every chart colour is truth-tested in libxlsxwriter -- `if (line->color)` --
  # so plain black, which xl_color() gives as 0, would be taken for "unset" and
  # the part drawn with Excel's default.  LXW_COLOR_BLACK is 0x1000000 for this
  # reason, and LXW_COLOR_MASK strips the bit again on the way out.
  expect_equal(.chart_color(xl_color("black")), 0x1000000L)
  expect_equal(.chart_color(xl_color("navy")), xl_color("navy"))
  expect_null(.chart_color(NULL))
  for (f in list(xl_border(all = "thin", color = "black"),
                 xl_fill(background = "black"),
                 xl_font(color = "black")))
    expect_true(any(vapply(cf(f), function(e) identical(e$color, 0x1000000L),
                           logical(1))))
  # a pattern's two colours go the same way
  p <- cf(xl_fill(pattern = "dark-up", foreground = "black",
                  background = "white"))$pattern
  expect_equal(p$fg_color, 0x1000000L)
})

test_that("a black chart line is written as a black line", {
  t <- tempfile(fileext = ".xlsx")
  df <- data.frame(a = c("x", "y"), b = c(1, 2), stringsAsFactors = FALSE)
  write_xlsx(list(Data = xl_sheet(df, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "b")),
    plot_area_format = xl_border(all = "thin", color = "black")))), t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  x <- paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                                  full.names = TRUE)[1L], warn = FALSE),
             collapse = "")
  plot_area <- substr(x, regexpr("<c:plotArea>", x), regexpr("</c:plotArea>", x))
  expect_match(plot_area, '<a:srgbClr val="000000"/>', fixed = TRUE)
})

test_that("the pattern and font that reach the chart XML are the ones asked for", {
  # the payload tests above check the number; this checks the number means what
  # it should once libxlsxwriter has written it
  chart_xml <- function(fmt, ...) {
    df <- data.frame(a = c("x", "y"), b = c(1, 2), stringsAsFactors = FALSE)
    t <- tempfile(fileext = ".xlsx")
    write_xlsx(list(Data = xl_sheet(df, chart = xl_chart("column",
      xl_chart_series(values = list(cols = "b"), format = fmt), ...))), t)
    xlsx_part(t, "xl/charts/chart1.xml", raw = TRUE)
  }
  x <- chart_xml(xl_fill(pattern = "dark-horizontal", foreground = "red",
                         background = "white"))
  expect_match(x, '<a:pattFill prst="dkHorz">', fixed = TRUE)
  expect_false(grepl('prst="pct30"', x, fixed = TRUE))

  x <- chart_xml(NULL, title = "Revenue",
                 title_format = xl_font(name = "Courier New", size = 14) +
                                xl_align(rotation = -45))
  expect_match(x, 'typeface="Courier New"', fixed = TRUE)
  # 60,000ths of a degree, as _chart_convert_font_args() converts it
  expect_match(x, 'rot="-2700000"', fixed = TRUE)
})

test_that("an absent or explicitly absent border is no chart dash", {
  expect_null(.chart_dash(NULL, "format"))
  expect_null(.chart_dash("none", "format"))
})

# Tests for cell comments: the xl_comment object, its format verification, and
# the write path (comments live in xl/comments1.xml and the VML drawing).

comments_xml <- function(x) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(x, tmp)
  xlsx_part(tmp, "xl/comments1.xml", raw = TRUE)
}
vml_xml <- function(x) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(x, tmp)
  xlsx_part(tmp, "xl/drawings/vmlDrawing1.vml", raw = TRUE)
}

# --- the xl_comment object ---------------------------------------------------

test_that("xl_comment builds a payload from text and options", {
  cm <- xl_comment("hello", author = "QA", width_pixels = 200, height_pixels = 90,
                   x_scale = 1.5, y_scale = 2, start_row = 3, start_col = 4,
                   x_offset = 5, y_offset = 6)
  p <- unclass(cm)
  expect_true(is_xl_comment(cm))
  expect_equal(p$value, "hello")
  expect_equal(p$author, "QA")
  expect_equal(p$width, 200L)
  expect_equal(p$height, 90L)
  expect_equal(p$x_scale, 1.5)
  expect_equal(p$y_scale, 2)
  expect_equal(p$start_row, 3L)
  expect_equal(p$start_col, 4L)
  expect_equal(p$x_offset, 5L)
  expect_equal(p$y_offset, 6L)
})

test_that("visible maps NA/TRUE/FALSE to the libxlsxwriter enum", {
  expect_false("visible" %in% names(unclass(xl_comment("x"))))       # NA -> unset
  expect_equal(unclass(xl_comment("x", visible = TRUE))$visible, 2L)  # VISIBLE
  expect_equal(unclass(xl_comment("x", visible = FALSE))$visible, 1L) # HIDDEN
})

test_that("xl_comment validates its inputs", {
  expect_error(xl_comment(c("a", "b")), "single non-NA string")
  expect_error(xl_comment(NA), "single non-NA string")
  expect_error(xl_comment(1), "single non-NA string")
  expect_error(xl_comment("x", format = "nope"), "must be an xl_format")
  expect_error(xl_comment("x", width_pixels = -1), "between 0")
  expect_error(xl_comment("x", visible = "yes"), "logical")
})

test_that("print.xl_comment runs", {
  expect_output(print(xl_comment("hi", author = "me")), "xl_comment")
  expect_output(print(xl_comment("hi", author = "me")), "author")
  expect_output(print(xl_comment("hi", format = xl_font(name = "Arial"))), "Arial")
})

# --- styling: the xl_format is kept, then flattened for C --------------------

test_that("the comment keeps its xl_format and stays editable", {
  cm <- xl_comment("note", format = xl_font(name = "Arial", size = 10))
  expect_true(is_xl_format(unclass(cm)$format))
  # the whole point of keeping it: a format can be adjusted after construction
  cm$format <- cm$format + xl_fill(background = "lightyellow")
  p <- .comment_c_payload(unclass(cm))
  expect_equal(p$font_name, "Arial")
  expect_equal(p$color, xl_color("lightyellow"))
})

test_that(".comment_c_payload flattens the format to the supported C fields", {
  cm <- xl_comment("note",
                   format = xl_font(name = "Arial", size = 10, family = 2) +
                            xl_fill(background = "lightyellow"))
  p <- .comment_c_payload(unclass(cm))
  expect_null(p$format)                     # replaced by flat fields
  expect_equal(p$value, "note")
  expect_equal(p$color, xl_color("lightyellow"))
  expect_equal(p$font_name, "Arial")
  expect_equal(p$font_size, 10)
  expect_equal(p$font_family, 2L)
  # no format -> no styling fields, and NULL passes through
  expect_null(.comment_c_payload(unclass(xl_comment("x")))$color)
  expect_null(.comment_c_payload(NULL))
})

test_that(".check_comment_format warns on unsupported properties", {
  expect_warning(xl_comment("x", format = xl_font(bold = TRUE)), "font\\$bold")
  expect_warning(xl_comment("x", format = xl_font(color = "red")), "font\\$color")
  expect_warning(xl_comment("x", format = xl_fill(foreground = "red")), "fill\\$foreground")
  expect_warning(xl_comment("x", format = xl_border(all = "thin")), "border")
  expect_warning(xl_comment("x", format = xl_align(horizontal = "center")), "align")
  expect_warning(xl_comment("x", format = xl_num_format("0.0")), "num_format")
  expect_warning(xl_comment("x", format = xl_protection(locked = FALSE)), "protection")
  expect_warning(xl_comment("x", format = xl_format(quote_prefix = TRUE)), "quote_prefix")
  expect_warning(xl_comment("x", format = xl_format(hyperlink = TRUE)), "hyperlink")
})

test_that("solid pattern is silent, non-solid pattern warns", {
  expect_no_warning(xl_comment("x", format = xl_fill(background = "yellow")))  # auto solid
  expect_warning(xl_comment("x", format = xl_fill(background = "yellow", pattern = "light-gray")),
                 "fill\\$pattern")
})

test_that("supported-only format does not warn", {
  expect_no_warning(
    xl_comment("x", format = xl_font(name = "Calibri", size = 9, family = 2) +
                            xl_fill(background = "white")))
})

test_that(".comment_payload normalizes strings, xl_comment, and NA", {
  expect_equal(.comment_payload("hi")$value, "hi")
  expect_equal(.comment_payload(xl_comment("hey"))$value, "hey")
  expect_null(.comment_payload(NA))
  expect_null(.comment_payload(NA_character_))
  expect_null(.comment_payload(NULL))
  expect_error(.comment_payload(42), "NA, a single string, or an xl_comment")
})

# --- attaching comments to cells and writing them ---------------------------

test_that("a character vector attaches per-cell comment text", {
  df <- data.frame(x = c("a", "b", "c"))
  df$x <- xl_cell_general(value = c("a", "b", "c"),
                          comment = c("first note", NA, "third note"))
  cm <- comments_xml(df)
  expect_match(cm, "first note")
  expect_match(cm, "third note")
  expect_false(grepl("second", cm))     # NA -> no comment
})

test_that("a single xl_comment recycles to every cell", {
  df <- data.frame(x = 1:2)
  df$x <- xl_cell_general(value = 1:2, comment = xl_comment("same", author = "QA"))
  cm <- comments_xml(df)
  expect_equal(lengths(regmatches(cm, gregexpr("same", cm)))[1], 2L)
  expect_match(cm, "QA")
})

test_that("comment styling reaches comments1.xml and the VML", {
  df <- data.frame(x = 1L)
  df$x <- xl_cell_general(value = 1,
    comment = xl_comment("styled", author = "Rev", visible = TRUE, width_pixels = 200,
                         format = xl_font(name = "Arial", size = 12) +
                                  xl_fill(background = "red")))
  cm <- comments_xml(df)
  vml <- vml_xml(df)
  expect_match(cm, "Rev")
  expect_match(cm, "Arial")            # rFont
  expect_match(cm, 'val="12"')         # font size
  expect_match(vml, "FF0000", ignore.case = TRUE)  # box fill color
  expect_match(vml, "visible")         # forced visible
})

test_that("a per-cell list mixes strings, xl_comment, and NA", {
  df <- data.frame(x = 1:3)
  df$x <- xl_cell_general(
    value = 1:3,
    comment = list("plain", xl_comment("rich", author = "me"), NA))
  cm <- comments_xml(df)
  expect_match(cm, "plain")
  expect_match(cm, "rich")
  expect_match(cm, "me")
})

test_that("a comment can stand alone on an otherwise-blank cell", {
  df <- data.frame(a = 1L)
  df$note <- xl_cell_general(comment = "standalone")
  expect_match(comments_xml(df), "standalone")
})

test_that("print shows a comment marker on the cell", {
  expect_output(print(xl_cell_general(value = 1, comment = "hi")), "comment=<set>")
})

test_that("values round-trip when comments are attached", {
  df <- data.frame(x = 1:3)
  df$x <- xl_cell_general(value = c(10, 20, 30), comment = c("a", "b", "c"))
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(df, tmp)
  expect_equal(readxl::read_xlsx(tmp)$x, c(10, 20, 30))
})

test_that("xl_cell_general validates the comment argument", {
  expect_error(xl_cell_general(value = 1, comment = 42),
               "character vector, an xl_comment, or a list")
  expect_error(xl_cell_general(value = 1, comment = list(1, 2)),
               "NA, a single string, or an xl_comment")
})

# --- sheet / workbook comment defaults --------------------------------------

test_that("sheet comment_author and show_comments apply", {
  df <- data.frame(x = 1L); df$x <- xl_cell_general(value = 1, comment = "a note")
  s <- xl_sheet(df, comment_author = "SheetAuthor", show_comments = TRUE)
  tmp <- tempfile(fileext = ".xlsx"); write_xlsx(list(D = s), tmp)
  expect_match(xlsx_part(tmp, "xl/comments1.xml", raw = TRUE), "SheetAuthor")
  expect_match(xlsx_part(tmp, "xl/drawings/vmlDrawing1.vml", raw = TRUE), "visible")
})

test_that("a per-comment author overrides the sheet default", {
  df <- data.frame(x = 1L)
  df$x <- xl_cell_general(value = 1, comment = xl_comment("n", author = "CellWins"))
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(list(D = xl_sheet(df, comment_author = "SheetAuthor")), tmp)
  expect_match(xlsx_part(tmp, "xl/comments1.xml", raw = TRUE), "CellWins")
})

test_that("comments work on a sheet with no comment defaults set", {
  df <- data.frame(x = 1L); df$x <- xl_cell_general(value = 1, comment = "n")
  cm <- comments_xml(list(D = xl_sheet(df)))
  expect_match(cm, "n")
  # comments are hidden by default
  expect_false(grepl("visible", vml_xml(list(D = xl_sheet(df)))))
})

test_that("show_comments is validated", {
  expect_error(xl_sheet(data.frame(a = 1), show_comments = "x"), "TRUE or FALSE")
  expect_error(xl_sheet(data.frame(a = 1), show_comments = NA), "TRUE or FALSE")
})

test_that("a comment takes its value from anything that can render itself", {
  expect_equal(unclass(xl_comment(xl_cell_general(value = "note")))$value,
               "note")
  expect_equal(unclass(xl_comment(factor("note")))$value, "note")
  # Excel comment boxes draw one font, so a rich string cannot be one
  expect_error(xl_comment(xl_rich_string("a", "b")), "cannot be a rich string")
  expect_error(xl_comment(NULL), "single non-NA string")
})

# Tests for the workbook layer: xl_properties(), xl_workbook(), and the
# de-hardcoded default formats.

test_that("xl_properties validates its arguments", {
  expect_error(xl_properties(default_format = "x"), "must be an xl_format")
  expect_error(xl_properties(header_format = 1), "must be an xl_format")
  expect_error(xl_properties(custom = list(1, 2)), "must be a named list")
  expect_error(xl_properties(window_size = 100), "length-2")
  expect_error(xl_properties(names = list("=1")), "named list")
  p <- xl_properties(title = "T", author = "A")
  expect_s3_class(p, "xl_properties")
})

test_that("xl_workbook validates and stores its parts", {
  expect_error(xl_workbook(1:3), "must be a data frame")
  expect_error(xl_workbook(data.frame(a = 1), properties = list()),
               "must be an xl_properties")
  wb <- xl_workbook(data.frame(a = 1), col_names = FALSE)
  expect_s3_class(wb, "xl_workbook")
  expect_false(wb$col_names)
  expect_length(wb$sheets, 1L)
})

test_that("print methods run", {
  expect_output(print(xl_properties(title = "T", author = "A", read_only = TRUE)),
                "xl_properties")
  expect_output(print(xl_workbook(data.frame(a = 1))), "xl_workbook")
})

test_that("document metadata is written to docProps/core.xml", {
  wb <- xl_workbook(data.frame(a = 1),
                    properties = xl_properties(title = "My Report",
                                               author = "Bill",
                                               company = "HP",
                                               subject = "Q1"))
  tmp <- write_tmp(wb)
  core <- xlsx_part(tmp, "docProps/core.xml", raw = TRUE)
  expect_match(core, "My Report")
  expect_match(core, "Bill")
  expect_match(core, "Q1")
})

test_that("custom properties of every scalar type are written", {
  wb <- xl_workbook(data.frame(a = 1),
                    properties = xl_properties(custom = list(
                      Project  = "Alpha",     # string
                      Revision = 3L,          # integer
                      Score    = 9.5,         # number
                      Final    = TRUE         # boolean
                    )))
  tmp <- write_tmp(wb)
  cust <- xlsx_part(tmp, "docProps/custom.xml", raw = TRUE)
  expect_match(cust, "Project");  expect_match(cust, "Alpha")
  expect_match(cust, "Revision"); expect_match(cust, "<vt:i4>3")
  expect_match(cust, "Score");    expect_match(cust, "9.5")
  expect_match(cust, "Final");    expect_match(cust, "<vt:bool>true")
})

test_that("read_only, window size, and defined names reach workbook.xml", {
  wb <- xl_workbook(data.frame(a = 1),
                    properties = xl_properties(read_only = TRUE,
                                               window_size = c(1000L, 600L),
                                               names = list(tax = "=0.2")))
  tmp <- write_tmp(wb)
  wx <- xlsx_part(tmp, "xl/workbook.xml", raw = TRUE)
  expect_match(wx, "readOnlyRecommended")
  expect_match(wx, 'windowWidth="15000"')   # set_size converts to twips (x15)
  expect_match(wx, 'windowHeight="9000"')
  expect_match(wx, 'name="tax"')
})

test_that("default_format cascades under every cell and header", {
  wb <- xl_workbook(data.frame(a = 1),
                    properties = xl_properties(default_format = xl_font(name = "Arial")))
  s <- styles_string(wb)
  expect_match(s, "Arial")
})

test_that("header_format and hyperlink_format can be overridden", {
  wb <- xl_workbook(
    data.frame(a = 1),
    properties = xl_properties(
      header_format = xl_font(bold = TRUE, color = "white") + xl_fill(background = "navy")
    )
  )
  s <- styles_string(wb)
  expect_match(s, "FF000080", ignore.case = TRUE)  # navy header fill

  df <- data.frame(a = 1L)
  df$h <- xl_hyperlink_cell("http://x.com")
  wb2 <- xl_workbook(list(S = df),
                     properties = xl_properties(
                       hyperlink_format = xl_font(color = "red", underline = "double")))
  s2 <- styles_string(wb2)
  expect_match(s2, "FFFF0000", ignore.case = TRUE)  # red hyperlink
})

test_that("date_format can be overridden at the workbook level", {
  wb <- xl_workbook(data.frame(d = as.Date("2020-01-01")),
                    properties = xl_properties(date_format = xl_num_format("dd/mm/yyyy")))
  expect_match(styles_string(wb), "dd/mm/yyyy", fixed = TRUE)
})

test_that("date column width and header row height are configurable", {
  wb <- xl_workbook(data.frame(d = as.Date("2020-01-01")),
                    properties = xl_properties(date_col_width = 30,
                                               header_row_height = 40))
  tmp <- write_tmp(wb)
  w <- xlsx_part(tmp, "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(w, 'width="30')     # date column widened (Excel pads the value)
  expect_match(w, 'ht="40"')       # header row height
  # datetime width too
  wb2 <- xl_workbook(data.frame(t = as.POSIXct("2020-01-01 00:00", tz = "UTC")),
                     properties = xl_properties(datetime_col_width = 28))
  w2 <- xlsx_part(write_tmp(wb2), "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(w2, 'width="28')
})

test_that("layout dimensions are validated", {
  expect_error(xl_properties(date_col_width = -1), "non-negative")
  expect_error(xl_properties(header_row_height = "tall"), "non-negative")
  expect_error(xl_properties(datetime_col_width = c(1, 2)), "single")
})

test_that("workbook col_names / format_headers override write_xlsx args", {
  wb <- xl_workbook(data.frame(a = 1:2), col_names = FALSE)
  tmp <- write_tmp(wb, col_names = TRUE)   # workbook wins -> no header
  rd <- readxl::read_xlsx(tmp, col_names = FALSE)
  expect_equal(as.integer(rd[[1]]), 1:2)
})

test_that("long and duplicate sheet names are still handled", {
  long <- list(data.frame(a = 1))
  names(long) <- paste(rep("x", 40), collapse = "")
  expect_warning(write_tmp(long), "Truncating")
  dup <- list(data.frame(a = 1), data.frame(b = 2))
  names(dup) <- c("S", "S")
  expect_warning(write_tmp(dup), "Deduplicating")
})

test_that("de-hardcoded defaults match the previous behavior (regression)", {
  # plain data frame, no workbook customization
  df <- data.frame(name = c("a", "b"), when = as.Date("2020-01-01") + 0:1)
  df$link <- xl_hyperlink_cell(c("http://x.com", "http://y.com"))
  s <- styles_string(df)
  expect_match(s, "<b/>")                          # bold header
  expect_match(s, 'horizontal="center"')           # centered header
  expect_match(s, "yyyy", fixed = TRUE)            # date number format
  expect_match(s, "FF0000FF", ignore.case = TRUE)  # blue hyperlink font
})

# --- constant memory --------------------------------------------------------

test_that("constant memory is off for small data, on when asked for", {
  # a one-cell frame saves nothing by streaming, so the default is off; the
  # request is what separates "nothing forbids it" from "it is worth it"
  small <- list(data.frame(a = 1))
  expect_equal(.resolve_constant_memory(small, xl_properties())$on, 0L)
  cm <- .resolve_constant_memory(small, xl_properties(), request = TRUE)
  expect_equal(cm$on, 1L)
  expect_equal(cm$reasons, character(0))
  expect_equal(.properties_payload(xl_properties())$constant_memory, 1L)
  expect_equal(.properties_payload(xl_properties(), 0L)$constant_memory, 0L)
})

test_that("workbooks write the same content with constant memory on and off", {
  skip_if_not_installed("readxl")
  df <- data.frame(txt = c("alpha", "beta", "alpha"),
                   num = c(1.5, 2, 3),
                   flag = c(TRUE, FALSE, NA),
                   when = as.Date("2020-01-01") + 0:2,
                   stringsAsFactors = FALSE)
  on_path <- write_tmp(list(D = xl_sheet(df, autofilter = TRUE, freeze = "A2")),
                       constant_memory = TRUE)
  # The off path is driven by mocking the resolver so that this test covers the
  # C side alone, whatever the blacklist happens to contain.  It has to stay
  # exercised: libxlsxwriter behaves differently with row streaming disabled.
  off_path <- local({
    local_mocked_bindings(
      .resolve_constant_memory = function(elems, props, ...)
        list(on = 0L, reasons = "forced off by test")
    )
    write_tmp(list(D = xl_sheet(df, autofilter = TRUE, freeze = "A2")))
  })

  # Content must match; the bytes need not -- with constant memory off, strings
  # go to the shared string table instead of being written inline.
  rd_on  <- as.data.frame(readxl::read_xlsx(on_path))
  rd_off <- as.data.frame(readxl::read_xlsx(off_path))
  expect_equal(rd_off, rd_on)
  expect_equal(rd_off$txt, df$txt)
  expect_equal(as.Date(rd_off$when), df$when)

  # ... and the worksheet features survive either way
  for (p in c(on_path, off_path)) {
    w <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
    expect_match(w, '<autoFilter ref="A1:D4"')
    expect_match(w, "<pane")
  }

  # the storage difference itself, so the off path cannot rot unnoticed
  expect_null(xlsx_part(on_path, "xl/sharedStrings.xml"))
  expect_s3_class(xlsx_part(off_path, "xl/sharedStrings.xml"), "xml_document")
})

# ── Hyperlink styling opt-out ────────────────────────────────────────────────

hyperlink_book <- function(hf) {
  df <- data.frame(x = 1L)
  df$h <- xl_hyperlink_cell("https://example.com")
  write_tmp(xl_workbook(list(S = df),
                        properties = xl_properties(hyperlink_format = hf)))
}

test_that("hyperlink_format = NULL writes an unstyled hyperlink", {
  # libxlsxwriter substitutes its own blue-underline format whenever a cell
  # reaches worksheet_write_url_opt() with none, so clearing that default is
  # the only way to get an unstyled link
  s <- xlsx_part(hyperlink_book(NULL), "xl/styles.xml", raw = TRUE)
  expect_false(grepl("<u/>", s, fixed = TRUE))
})

test_that("an empty xl_format() is not the opt-out", {
  # it is the neutral element of the cascade, so the default still applies
  s <- xlsx_part(hyperlink_book(xl_format()), "xl/styles.xml", raw = TRUE)
  expect_true(grepl("<u/>", s, fixed = TRUE))
})

test_that("the default hyperlink format is still blue and underlined", {
  s <- xlsx_part(hyperlink_book(xl_properties()$hyperlink_format),
                 "xl/styles.xml", raw = TRUE)
  expect_true(grepl("<u/>", s, fixed = TRUE))
  expect_match(s, "FF0000FF", ignore.case = TRUE)
})

test_that("hyperlink_format rejects a non-format that is not NULL", {
  expect_error(xl_properties(hyperlink_format = "blue"),
               "must be an xl_format object or NULL")
  expect_s3_class(xl_properties(hyperlink_format = NULL), "xl_properties")
})

# ── Datetime custom properties ───────────────────────────────────────────────

test_that("Date and POSIXct custom properties are written as datetimes", {
  wb <- xl_workbook(data.frame(a = 1),
                    properties = xl_properties(custom = list(
                      Released = as.Date("2024-03-17"),
                      Built    = as.POSIXct("2024-03-17 14:35:09", tz = "UTC")
                    )))
  cust <- xlsx_part(write_tmp(wb), "docProps/custom.xml", raw = TRUE)
  # vt:filetime, not vt:lpwstr -- these used to be stringified
  expect_match(cust, "<vt:filetime>2024-03-17T00:00:00Z</vt:filetime>",
               fixed = TRUE)
  expect_match(cust, "<vt:filetime>2024-03-17T14:35:09Z</vt:filetime>",
               fixed = TRUE)
  expect_false(grepl("<vt:lpwstr>2024-03-17", cust, fixed = TRUE))
})

test_that("a datetime custom property follows the workbook time zone rule", {
  # one zone across the workbook: the zone is dropped and wall-clock written,
  # exactly as for cells, rather than the property using a second policy
  wb <- xl_workbook(
    data.frame(t = as.POSIXct("2024-03-17 09:00:00", tz = "Australia/Perth")),
    properties = xl_properties(custom = list(
      Built = as.POSIXct("2024-03-17 14:35:09", tz = "Australia/Perth")))
  )
  cust <- xlsx_part(write_tmp(wb), "docProps/custom.xml", raw = TRUE)
  expect_match(cust, "2024-03-17T14:35:09Z", fixed = TRUE)
})

test_that("the other custom property types are unaffected", {
  wb <- xl_workbook(data.frame(a = 1),
                    properties = xl_properties(custom = list(
                      S = "txt", I = 3L, N = 9.5, B = TRUE,
                      D = as.Date("2020-01-02"))))
  cust <- xlsx_part(write_tmp(wb), "docProps/custom.xml", raw = TRUE)
  expect_match(cust, "<vt:lpwstr>txt")
  expect_match(cust, "<vt:i4>3")
  expect_match(cust, "9.5")
  expect_match(cust, "<vt:bool>true")
  expect_match(cust, "<vt:filetime>2020-01-02T00:00:00Z", fixed = TRUE)
})

test_that("the constant_memory decision follows its five rules in order", {
  small <- list(data.frame(a = 1:5))
  # ~1.4M cells, comfortably past the default threshold at 110 B/cell
  big <- list(data.frame(matrix(1, nrow = 130000, ncol = 12)))
  blocked <- list(list(merges = list(1)))
  clear <- list(list(merges = NULL))

  # 1. FALSE wins outright, without consulting anything else
  # FALSE returns before the blacklist is even consulted, so it is silent too
  expect_silent(r <- .resolve_constant_memory(big, list(), blocked, FALSE))
  expect_equal(r$on, 0L)
  expect_match(r$reasons, "constant_memory = FALSE")

  # 2. the blacklist is absolute -- TRUE cannot override it, only be told.
  # (These also emit the size suggestion, which its own test covers.)
  expect_equal(suppressMessages(
    .resolve_constant_memory(big, list(), blocked))$on, 0L)
  expect_warning(r <- suppressMessages(
    .resolve_constant_memory(big, list(), blocked, TRUE)), "cannot be honoured")
  expect_equal(r$on, 0L)

  # 3. TRUE overrides the size estimate
  expect_equal(.resolve_constant_memory(small, list(), clear, TRUE)$on, 1L)

  # 4. size decides when nothing else does
  expect_equal(.resolve_constant_memory(small, list(), clear)$on, 0L)
  expect_match(.resolve_constant_memory(small, list(), clear)$reasons,
               "below the .* threshold")
  expect_equal(.resolve_constant_memory(big, list(), clear)$on, 1L)
  # ... and the threshold is what moves that line
  expect_equal(.resolve_constant_memory(big, list(), clear, NA,
                                        threshold = 4 * 1024^3)$on, 0L)
})

test_that("cells are counted across the whole workbook", {
  # libxlsxwriter holds every sheet's cell table until close, so two sheets of
  # half the size must decide the same way as one of the full size
  half <- data.frame(matrix(1, nrow = 65000, ncol = 12))
  clear <- list(list(merges = NULL))
  expect_equal(.resolve_constant_memory(list(half), list(), clear)$on, 0L)
  expect_equal(.resolve_constant_memory(list(half, half), list(), clear)$on, 1L)
})

test_that("the suggestion appears only when streaming would have mattered", {
  small <- list(data.frame(a = 1:5))
  big <- list(data.frame(matrix(1, nrow = 130000, ncol = 12)))
  blocked <- list(list(merges = list(1)))
  # a small workbook blocked by a feature has nothing to act on, so says nothing
  expect_silent(.resolve_constant_memory(small, list(), blocked))
  # a large one is told what it cost, and why
  expect_message(.resolve_constant_memory(big, list(), blocked), "merged range")
  expect_message(.resolve_constant_memory(big, list(), blocked),
                 "MB more memory")
})

test_that("the decision is silent whenever there is nothing to report", {
  # write_xlsx() is called constantly on small frames; a message on every one
  # would be noise, so only a refusal that cost something speaks
  small <- list(data.frame(a = 1:5))
  clear <- list(list(merges = NULL))
  expect_silent(.resolve_constant_memory(small, list(), clear))
  expect_silent(.resolve_constant_memory(small, list(), clear, TRUE))
  expect_silent(.resolve_constant_memory(small, list(), clear, FALSE))
  big <- list(data.frame(matrix(1, nrow = 130000, ncol = 12)))
  expect_silent(.resolve_constant_memory(big, list(), clear))
})

test_that("an invalid constant_memory request is refused", {
  expect_error(.resolve_constant_memory(list(data.frame(a = 1)), list(), NULL,
                                        "yes"),
               "must be a single logical")
})

test_that("a pinned creation time makes two runs byte-identical", {
  core <- function(p) {
    tmp <- write_tmp(xl_workbook(data.frame(a = 1), properties = p))
    list(core = xlsx_part(tmp, "docProps/core.xml", raw = TRUE),
         bytes = readBin(tmp, "raw", file.size(tmp)))
  }
  when <- as.POSIXct("2020-06-15 12:34:56", tz = "UTC")
  a <- core(xl_properties(created = when))
  expect_match(a$core, "2020-06-15T12:34:56Z", fixed = TRUE)
  # the whole file, not just the timestamp: that is the point of pinning it
  expect_equal(core(xl_properties(created = when))$bytes, a$bytes)

  # a Date is midnight UTC on that day
  expect_match(core(xl_properties(created = as.Date("2020-06-15")))$core,
               "2020-06-15T00:00:00Z", fixed = TRUE)
  # and left unset libxlsxwriter stamps the moment of writing
  expect_false(grepl("2020-06-15", core(xl_properties())$core, fixed = TRUE))
})

test_that("`created` must be a single date or datetime", {
  bad <- function(x) write_tmp(xl_workbook(data.frame(a = 1),
                                           properties = xl_properties(created = x)))
  expect_error(bad("2020-06-15"), "must be a Date or POSIXct")
  expect_error(bad(1592223296), "must be a Date or POSIXct")
  expect_error(bad(as.Date(c("2020-06-15", "2020-06-16"))),
               "must be a single date or datetime")
})

# Rich strings: one cell, several differently formatted runs.

sheet_xml <- function(path) xlsx_part(path, "xl/worksheets/sheet1.xml", raw = TRUE)

# A one-cell sheet whose column B holds the given rich string.
# constant_memory = TRUE throughout: these tests are about the inline-string
# form used while streaming, which a frame this small would not otherwise get
# (row streaming is only chosen when it would save enough memory to matter).
rich_sheet <- function(rs, ...) {
  df <- data.frame(x = 1L)
  df$r <- xl_cell_general(value = rs, ...)
  write_tmp(df, constant_memory = TRUE)
}

# ── xl_rich_run() ─────────────────────────────────────────────────────────────

test_that("xl_rich_run validates its text", {
  expect_error(xl_rich_run(1), "single non-NA string")
  expect_error(xl_rich_run(NA_character_), "single non-NA string")
  expect_error(xl_rich_run(c("a", "b")), "single non-NA string")
  # Excel cannot represent an empty run, and libxlsxwriter rejects one outright
  expect_error(xl_rich_run(""), "must not be empty")
})

test_that("xl_rich_run validates its format", {
  expect_error(xl_rich_run("a", format = "bold"), "must be an xl_format")
  expect_s3_class(xl_rich_run("a", xl_font(bold = TRUE)), "xl_rich_run")
  expect_s3_class(xl_rich_run("a"), "xl_rich_run")
})

test_that("a run warns about format properties it cannot render", {
  # a run is drawn with a font only, so every other group is ignored
  expect_warning(xl_rich_run("a", xl_fill(background = "yellow")), "fill")
  expect_warning(xl_rich_run("a", xl_border(all = "thin")), "border")
  expect_warning(xl_rich_run("a", xl_align(horizontal = "center")), "align")
  expect_warning(xl_rich_run("a", xl_num_format("0.00")), "num_format")
  expect_warning(xl_rich_run("a", xl_protection(locked = FALSE)), "protection")
  expect_warning(xl_rich_run("a", xl_format(quote_prefix = TRUE)), "quote_prefix")
  # the whole font group is supported, so none of it warns
  expect_silent(xl_rich_run("a", xl_font(bold = TRUE, italic = TRUE,
                                         underline = "single", strikeout = TRUE,
                                         script = "sub", color = "red",
                                         name = "Arial", size = 9)))
})

# ── xl_rich_string() ──────────────────────────────────────────────────────────

test_that("a rich string needs at least two runs", {
  expect_error(xl_rich_string("only one"), "at least 2 runs")
  expect_error(xl_rich_string(), "at least 2 runs")
  expect_error(xl_rich_string(xl_rich_run("a", xl_font(bold = TRUE))),
               "ordinary character value")
  expect_s3_class(xl_rich_string("a", "b"), "xl_rich_string")
})

test_that("bare strings become unformatted runs and lists are flattened", {
  rs <- xl_rich_string("a", xl_rich_run("b", xl_font(bold = TRUE)), "c")
  expect_length(unclass(rs), 3L)
  expect_null(unclass(rs)[[1L]]$format)
  expect_true(is_xl_format(unclass(rs)[[2L]]$format))

  # a character vector contributes one run per element
  expect_length(unclass(xl_rich_string(c("a", "b"))), 2L)
  # nested lists splice in, so runs can be built with lapply()
  built <- lapply(c("x", "y"), xl_rich_run)
  expect_length(unclass(xl_rich_string("pre", built)), 3L)
})

test_that("xl_rich_string rejects run types it cannot use", {
  expect_error(xl_rich_string("a", 42), "must be a string or an xl_rich_run")
  expect_error(xl_rich_string("a", character(0)), "at least one string")
})

test_that("rich string helpers behave", {
  rs <- xl_rich_string("a", "b")
  expect_true(is_xl_rich_string(rs))
  expect_false(is_xl_rich_string("a"))
  # format() gives the plain concatenated text, which auto_colwidth measures
  expect_equal(format(rs), "ab")
  expect_output(print(rs), "2 runs")
  expect_output(print(xl_rich_run("a")), "xl_rich_run")
})

# ── Writing ───────────────────────────────────────────────────────────────────

test_that("a rich string writes one cell with per-run fonts", {
  p <- rich_sheet(xl_rich_string("This is ",
                                 xl_rich_run("bold", xl_font(bold = TRUE)),
                                 " text"))
  s <- sheet_xml(p)
  # three runs, the middle one bold, in the inline-string form used while row
  # streaming is on
  expect_match(s, '<c r="B2" t="inlineStr">', fixed = TRUE)
  expect_match(s, "<r><rPr><b/>", fixed = TRUE)
  expect_match(s, "<t>bold</t>", fixed = TRUE)
  # leading/trailing spaces in a run must be preserved
  expect_match(s, 'xml:space="preserve"', fixed = TRUE)
  # and it reads back as the concatenated text
  skip_if_not_installed("readxl")
  expect_equal(as.data.frame(readxl::read_xlsx(p))$r, "This is bold text")
})

test_that("a rich string is stored in the shared table when streaming is off", {
  # the two storage forms are different code paths in libxlsxwriter, so both
  # need covering; nothing writexl writes turns streaming off on its own here
  local_mocked_bindings(
    .resolve_constant_memory = function(dfs, props, ...)
      list(on = 0L, reasons = "forced off by test")
  )
  p <- rich_sheet(xl_rich_string("This is ",
                                 xl_rich_run("bold", xl_font(bold = TRUE)),
                                 " text"))
  ss <- xlsx_part(p, "xl/sharedStrings.xml", raw = TRUE)
  expect_match(ss, "<r><rPr><b/>", fixed = TRUE)
  expect_match(ss, "<t>bold</t>", fixed = TRUE)
  # the cell now references the table rather than carrying the text inline
  expect_match(sheet_xml(p), '<c r="B2" t="s">', fixed = TRUE)
  skip_if_not_installed("readxl")
  expect_equal(as.data.frame(readxl::read_xlsx(p))$r, "This is bold text")
})

test_that("a cell-level format applies alongside the run fonts", {
  p <- rich_sheet(xl_rich_string("a", xl_rich_run("b", xl_font(bold = TRUE))),
                  format = xl_fill(background = "yellow"))
  # the cell carries a style ...
  expect_match(sheet_xml(p), '<c r="B2" s="', fixed = TRUE)
  # ... and the fill reached styles.xml, which a run alone could not do
  expect_match(xlsx_part(p, "xl/styles.xml", raw = TRUE), "FFFFFF00",
               ignore.case = TRUE)
})

test_that("a rich string counts as one cell and recycles", {
  rs <- xl_rich_string("a", "b")
  expect_equal(length(xl_cell_general(value = rs)), 1L)
  # recycled across a longer column, like any other length-1 value
  df <- data.frame(x = 1:3)
  df$r <- xl_cell_general(value = rs)
  expect_equal(length(df$r), 3L)
  expect_true(file.exists(write_tmp(df)))
})

test_that("a rich string cannot share a cell with a formula or hyperlink", {
  rs <- xl_rich_string("a", "b")
  expect_error(xl_cell_general(value = rs, formula = "=A1"),
               "rich string `value` with a `formula`")
  expect_error(xl_cell_general(value = rs, hyperlink = "https://example.com"),
               "rich string `value` with a `hyperlink`")
  # an NA hyperlink is not a hyperlink, so it is allowed
  expect_s3_class(xl_cell_general(value = rs, hyperlink = NA),
                  "xl_cell_general")
})

test_that("a rich string works with a comment and with auto_colwidth", {
  df <- data.frame(x = 1L)
  df$r <- xl_cell_general(value = xl_rich_string("hello ", "world"),
                          comment = "note")
  p <- write_tmp(list(S = xl_sheet(df, auto_colwidth = TRUE)))
  expect_match(xlsx_part(p, "xl/comments1.xml", raw = TRUE), "note")
  # auto_colwidth must measure the concatenated run text, not treat the cell as
  # empty: "hello world" is 11 characters wide.  (The stored number carries
  # libxlsxwriter's own padding, so compare against the text width rather than
  # pinning that constant.)
  w <- regmatches(sheet_xml(p),
                  regexec('<col min="2" max="2" width="([0-9.]+)"',
                          sheet_xml(p)))[[1L]]
  expect_length(w, 2L)
  expect_gt(as.numeric(w[2L]), nchar("hello world"))
})

test_that("a rich string is accepted by the cell value type check", {
  # the unsupported-type guard must not reject it for being a list
  expect_s3_class(xl_cell_general(value = xl_rich_string("a", "b")),
                  "xl_cell_general")
  # mixed with ordinary values in one column
  cells <- c(xl_cell_general(value = xl_rich_string("a", "b")),
             xl_cell_general(value = 1.5))
  expect_equal(length(cells), 2L)
  df <- data.frame(x = 1:2)
  df$r <- cells
  expect_true(file.exists(write_tmp(df)))
})

# ── as.character() ────────────────────────────────────────────────────────────

test_that("as.character() drops the runs' fonts and keeps the text", {
  rs <- xl_rich_string("This is ", xl_rich_run("bold", xl_font(bold = TRUE)),
                       " text")
  expect_equal(as.character(rs), "This is bold text")
  # which is what a data frame column of them shows
  expect_equal(as.character(xl_cell_general(value = rs)), "This is bold text")
})

test_that("as.character() of a cell is what the cell displays", {
  expect_equal(as.character(xl_cell_general(value = c(1.5, 2))), c("1.5", "2"))
  expect_equal(as.character(xl_cell_general(value = "a")), "a")
  # a formula's displayed value comes from Excel, so writexl has none to give
  expect_equal(as.character(xl_cell_general(formula = "=SUM(A1:A2)")),
               NA_character_)
  expect_equal(as.character(xl_cell_general(value = NA)), NA_character_)
})

test_that("a run takes its value from anything that can render itself", {
  expect_equal(xl_rich_run(xl_cell_general(value = "a"))$value, "a")
  expect_equal(xl_rich_run(factor("a"))$value, "a")
  # a run is one font by definition, so a rich string cannot be one
  expect_error(xl_rich_run(xl_rich_string("a", "b")), "cannot be a rich string")
})

# =============================================================================
# Chartsheets: a sheet that is one chart and nothing else
# =============================================================================
#
# A chartsheet holds a single chart full-page and has no cells, so it takes no
# data frame and none of the worksheet options that address cells.
# libxlsxwriter reflects that: lxw_chartsheet wraps a worksheet but exposes
# only 17 functions, against the worksheet's several hundred.
#
# Its chart therefore has to plot another sheet.  Every range in the series must
# name one, since there is no data on the chartsheet for a bare
# list(cols = ) spec to resolve against.
#
# The subsets a chartsheet supports are named here rather than left to Excel:
# xl_page_setup() has 23 options and a chartsheet has setters for five,
# xl_sheet_view() has nine and a chartsheet has four.  The rest are refused by
# name.
# -----------------------------------------------------------------------------

# The xl_page_setup() options a chartsheet can express: chartsheet_set_paper(),
# _set_margins(), _set_landscape()/_set_portrait(), _set_header(), _set_footer()
# and their _opt() variants carrying the margins.
.CHARTSHEET_PAGE <- c("orientation", "paper", "margins", "header", "footer",
                      "header_margin", "footer_margin")

# The xl_sheet_view() options it can express: chartsheet_activate(),
# _select(), _hide() and _set_first_sheet().  Selection is the same argument
# here as on a worksheet -- see apply_chartsheet() in src/write_xlsx.c for the
# one extra flag a chartsheet needs to make it reach the file.
.CHARTSHEET_VIEW <- c("active", "selected", "visible", "first_tab")

#' A sheet holding a single chart
#'
#' @description
#' `xl_chartsheet()` is a worksheet-sized chart: a tab of its own holding one
#' chart and no cells.  Give it to [write_xlsx()] in place of a data frame.
#'
#' Because a chartsheet has no cells, every range in the chart's series must
#' name the sheet it plots --- `list(sheet = "Data", cols = "revenue")` or
#' `"Data!B2:B10"`.  A bare `list(cols = )` has nothing to resolve against and
#' is refused.
#'
#' A chartsheet supports only part of what a worksheet does, and the parts it
#' does not are refused rather than dropped: of [xl_page_setup()] it takes the
#' orientation, paper size, margins and the header and footer; of
#' [xl_sheet_view()] it takes `active`, `selected`, `visible` and
#'   `first_tab`.
#'
#' @param chart The [xl_chart()] to fill the sheet with.
#' @param tab_color The colour of the sheet tab.
#' @param zoom The zoom level as a percentage, 10 to 400.
#' @param protect `TRUE`, a password string, or a named list. A chartsheet
#'   has no cells, so of Excel's protection options it takes only
#'   `no_content` (let the chart be edited) and `no_objects` (let the shapes
#'   on it be edited); the worksheet options are refused by name.
#' @param page An [xl_page_setup()] describing how it prints.
#' @param view An [xl_sheet_view()] setting the tab state.
#' @return An `xl_chartsheet` object.
#' @family images and charts
#' @seealso [xl_chart], [xl_sheet]
#' @export
#' @examples
#' sales <- data.frame(quarter = c("Q1", "Q2"), revenue = c(10, 25))
#' chart <- xl_chart("column",
#'                   xl_chart_series(values = list(sheet = "Data",
#'                                                 cols = "revenue")))
#' write_xlsx(list(Data = sales, Overview = xl_chartsheet(chart)),
#'            tempfile(fileext = ".xlsx"))
xl_chartsheet <- function(chart, tab_color = NULL, zoom = NA, protect = NULL,
                          page = NULL, view = NULL) {
  if (missing(chart) || is.null(chart))
    stop("`chart` must be the xl_chart() to fill the sheet with", call. = FALSE)
  .validate_protect(protect, "chartsheet")
  ch <- .chart_list(chart)
  if (length(ch) != 1L)
    stop(sprintf(paste0("a chartsheet holds exactly one chart, not %d. Put ",
                        "the others on worksheets with xl_sheet(chart = )."),
                 length(ch)), call. = FALSE)

  structure(.drop_null(list(
    chart = ch[[1L]],
    view = .chartsheet_subset(view, "view", "xl_sheet_view",
                              .CHARTSHEET_VIEW),
    page = .chartsheet_subset(page, "page", "xl_page_setup",
                              .CHARTSHEET_PAGE),
    protect = protect,
    tab_color = .val_color(tab_color, "tab_color"),
    zoom = .val_int(zoom, "zoom", min = 10, max = 400)
  )), class = "xl_chartsheet")
}

# Keep the options a chartsheet has a setter for, and name the ones it does not.
.chartsheet_subset <- function(x, arg, cls, keep) {
  if (is.null(x)) return(NULL)
  if (!inherits(x, cls))
    stop(sprintf("`%s` must be an %s() object", arg, cls), call. = FALSE)
  # xl_sheet_view() keeps a slot per option, unset ones NULL, so only what was
  # actually set can be complained about
  p <- Filter(Negate(is.null), unclass(x))
  extra <- setdiff(names(p), keep)
  if (length(extra))
    stop(sprintf(paste0("`%s` sets %s, which a chartsheet has no setter for; ",
                        "Excel would ignore it.\n  A chartsheet takes: %s."),
                 arg, paste(sprintf("`%s`", extra), collapse = ", "),
                 paste(keep, collapse = ", ")), call. = FALSE)
  p
}

#' @export
print.xl_chartsheet <- function(x, ...) {
  p <- unclass(x)
  cat("<xl_chartsheet> ", unclass(p$chart)$type, " chart\n", sep = "")
  set <- setdiff(names(p), "chart")
  if (length(set)) cat("  set:", paste(set, collapse = ", "), "\n")
  invisible(x)
}

# --- Resolution --------------------------------------------------------------

# A chartsheet has no cells, so it stands in the data-frame list as a data frame
# with no columns.  Nothing writes rows for it: the C side branches on the
# plan's `chartsheet` flag before the row loop.
.chartsheet_placeholder <- function() data.frame()

# The plan C reads for one chartsheet.  `charts` carries the single chart in the
# same slot a worksheet uses, so .check_drawing_order() sees it without knowing
# the difference.
.resolve_chartsheet_plan <- function(el, charts) {
  p <- unclass(el)
  ent <- list(chartsheet = 1L, charts = charts)
  vw <- p[["view"]]
  if (!is.null(vw)) {
    if (isTRUE(vw$active))       ent$activate <- 1L
    if (isTRUE(vw$selected))     ent$select <- 1L
    if (identical(vw$visible, FALSE)) ent$hide <- 1L
    if (isTRUE(vw$first_tab))    ent$first_sheet <- 1L
  }
  pg <- p[["page"]]
  if (!is.null(pg)) {
    if (!is.null(pg$orientation))
      ent$landscape <- as.integer(identical(pg$orientation, "landscape"))
    ent$paper <- pg$paper
    if (!is.null(pg$margins)) ent$margins <- unlist(pg$margins, use.names = FALSE)
    ent$header <- pg$header
    ent$footer <- pg$footer
    ent$header_margin <- pg$header_margin
    ent$footer_margin <- pg$footer_margin
  }
  ent$tab_color <- p[["tab_color"]]
  ent$zoom <- p[["zoom"]]
  if (!is.null(p[["protect"]])) {
    pr <- .resolve_protect(p[["protect"]], "chartsheet")
    ent$protect <- pr$flag
    ent$protect_password <- pr$password
    ent$protect_options <- pr$options
  }
  .drop_null(ent)
}

# Every range a chartsheet's chart plots must name its sheet: there is nothing
# on the chartsheet itself for a bare list(cols = ) or an unqualified "B2:B10"
# to resolve against.
.check_chartsheet_ranges <- function(chart, own) {
  # `r` is never NULL: the loop below skips a part that carries no range
  named <- function(r) {
    if (!is.null(r[["sheet"]])) return(TRUE)
    is.character(r[["spec"]]) && !is.null(.split_sheet_ref(r[["spec"]])$sheet)
  }
  p <- unclass(chart)
  parts <- list(title = p[["title"]])
  for (i in seq_along(p[["series"]])) {
    q <- unclass(p[["series"]][[i]])
    for (k in c("values", "categories", "name"))
      parts[[sprintf("series[[%d]]$%s", i, k)]] <- q[[k]]
  }
  for (nm in names(parts)) {
    r <- parts[[nm]]
    # a literal title or series name carries no range at all
    if (is.null(r) || !is.null(r[["text"]]) || isTRUE(r[["off"]])) next
    if (named(r)) next
    stop(sprintf(paste0("`%s` does not name a sheet, and chartsheet \"%s\" ",
                        "has no cells of its own to resolve it against.\n  ",
                        "Give the sheet holding the data: ",
                        "list(sheet = \"Data\", cols = \"revenue\") or ",
                        "\"Data!B2:B10\"."),
                 nm, own), call. = FALSE)
  }
  invisible(NULL)
}

# Chartsheets: a sheet that is one chart and no cells.  A chartsheet supports a
# small subset of what a worksheet does, and the tests below are mostly about
# refusing the rest rather than letting Excel drop it.

cs_sales <- data.frame(quarter = c("Q1", "Q2", "Q3", "Q4"),
                       revenue = c(10, 25, 18, 32),
                       stringsAsFactors = FALSE)

cs_series <- function()
  xl_chart_series(values = list(sheet = "Data", cols = "revenue"),
                  categories = list(sheet = "Data", cols = "quarter"))

# Write a workbook holding one chartsheet and return its parts.
cs_parts <- function(..., name = "Overview") {
  t <- tempfile(fileext = ".xlsx")
  wb <- list(Data = cs_sales)
  wb[[name]] <- xl_chartsheet(...)
  write_xlsx(wb, t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  f <- list.files(file.path(d, "xl/chartsheets"), pattern = "^sheet",
                  full.names = TRUE)
  list(files = list.files(d, recursive = TRUE),
       sheet = if (length(f)) paste(readLines(f[1L], warn = FALSE),
                                    collapse = "") else "",
       workbook = paste(readLines(file.path(d, "xl/workbook.xml"),
                                  warn = FALSE), collapse = ""))
}

# ── What reaches the file ─────────────────────────────────────────────────────

test_that("a chartsheet is its own part, with no worksheet beside it", {
  r <- cs_parts(xl_chart("column", cs_series(), title = "Revenue"))
  expect_true(any(grepl("^xl/chartsheets/sheet1.xml$", r$files)))
  expect_true(any(grepl("^xl/charts/chart1.xml$", r$files)))
  expect_true(any(grepl("^xl/drawings/drawing1.xml$", r$files)))
  # the data sheet is the only worksheet
  expect_length(grep("^xl/worksheets/sheet", r$files), 1L)
  expect_match(r$sheet, "<drawing r:id=", fixed = TRUE)
  expect_match(r$workbook, "Overview", fixed = TRUE)
})

test_that("the chart on a chartsheet carries everything a chart can", {
  r <- cs_parts(xl_chart("line",
                         xl_chart_series(values = list(sheet = "Data",
                                                       cols = "revenue"),
                                         categories = list(sheet = "Data",
                                                           cols = "quarter"),
                                         marker = xl_chart_marker("circle", 7)),
                         title = "T",
                         legend = xl_chart_legend("bottom"),
                         y_axis = xl_chart_axis(title = "Revenue", min = 0)))
  # the chart itself is covered by the chart tests; what matters here is that
  # a fully-featured chart still reaches a chartsheet
  expect_true(any(grepl("^xl/charts/chart1.xml$", r$files)))
  expect_true(any(grepl("^xl/chartsheets/sheet1.xml$", r$files)))
})

test_that("the settings a chartsheet does support reach the file", {
  r <- cs_parts(xl_chart("pie", cs_series()),
                tab_color = "red", zoom = 150,
                view = xl_sheet_view(active = TRUE),
                page = xl_page_setup(orientation = "landscape", paper = "A4",
                                     header = "&LTop"),
                protect = "secret")
  expect_match(r$sheet, "FF0000", fixed = TRUE)
  expect_match(r$sheet, 'zoomScale="150"', fixed = TRUE)
  expect_match(r$sheet, 'orientation="landscape"', fixed = TRUE)
  expect_match(r$sheet, 'paperSize="9"', fixed = TRUE)
  expect_match(r$sheet, "&amp;LTop", fixed = TRUE)
  expect_match(r$sheet, "sheetProtection", fixed = TRUE)
})

test_that("a chartsheet takes part in the workbook-wide tab rules", {
  # the visibility checks span every sheet, and a chartsheet is a sheet
  expect_error(
    write_xlsx(list(Data = cs_sales,
                    Hidden = xl_chartsheet(xl_chart("pie", cs_series()),
                                           view = xl_sheet_view(visible = FALSE,
                                                                active = TRUE))),
               tempfile(fileext = ".xlsx")),
    "hidden")
  # and one can be the active tab
  expect_silent(write_xlsx(list(
    Data = cs_sales,
    Front = xl_chartsheet(xl_chart("pie", cs_series()),
                          view = xl_sheet_view(active = TRUE))),
    tempfile(fileext = ".xlsx")))
})

# ── What a chartsheet refuses ─────────────────────────────────────────────────

test_that("a chartsheet holds exactly one chart", {
  expect_error(xl_chartsheet(list(xl_chart("column",
                                           xl_chart_series(values = "A1:A5")),
                                  xl_chart("pie",
                                           xl_chart_series(values = "A1:A5")))),
               "holds exactly one chart, not 2")
  expect_error(xl_chartsheet(), "must be the xl_chart")
})

test_that("every range must name its sheet, since a chartsheet has none", {
  bad <- function(...) write_xlsx(list(Data = cs_sales,
                                       Bad = xl_chartsheet(xl_chart(...))),
                                  tempfile(fileext = ".xlsx"))
  expect_error(bad("column", xl_chart_series(values = list(cols = "revenue"))),
               "does not name a sheet")
  expect_error(bad("column", xl_chart_series(values = list(cols = "revenue"))),
               'chartsheet "Bad" has no cells of its own')
  # an unqualified A1 range is refused for the same reason
  expect_error(bad("column", xl_chart_series(values = "B2:B5")),
               "does not name a sheet")
  # the categories, the series name and the title are checked too
  expect_error(bad("column",
                   xl_chart_series(values = list(sheet = "Data",
                                                 cols = "revenue"),
                                   categories = list(cols = "quarter"))),
               "series\\[\\[1\\]\\]\\$categories` does not name a sheet")
  expect_error(bad("column", cs_series(), title = list(header = "revenue")),
               "`title` does not name a sheet")
  # a literal title or name carries no range, so neither is checked
  expect_silent(bad("column", xl_chart_series(values = list(sheet = "Data",
                                                            cols = "revenue"),
                                              name = "Revenue"),
                    title = "Revenue"))
})

test_that("the page options a chartsheet has no setter for are refused", {
  ch <- xl_chart("column", cs_series())
  expect_error(xl_chartsheet(ch, page = xl_page_setup(scale = 90)),
               "`page` sets `scale`, which a chartsheet has no setter for",
               fixed = TRUE)
  expect_error(xl_chartsheet(ch, page = xl_page_setup(print_area = "A1:B2")),
               "A chartsheet takes: orientation, paper, margins", fixed = TRUE)
  # and the five it does have are accepted
  expect_s3_class(xl_chartsheet(ch, page = xl_page_setup(
                    orientation = "landscape", paper = "A4",
                    margins = c(1, 1, 1, 1), header = "&Lx", footer = "&Ry")),
                  "xl_chartsheet")
})

test_that("the view options a chartsheet has no setter for are refused", {
  ch <- xl_chart("column", cs_series())
  for (opt in list(list(hide_zero = TRUE), list(right_to_left = TRUE),
                   list(selection = "A1"), list(top_left = "A1")))
    expect_error(xl_chartsheet(ch, view = do.call(xl_sheet_view, opt)),
                 "which a chartsheet has no setter for", label = names(opt))
  for (opt in list(list(active = TRUE), list(selected = TRUE),
                   list(visible = FALSE), list(first_tab = TRUE)))
    expect_true(inherits(xl_chartsheet(ch, view = do.call(xl_sheet_view, opt)),
                         "xl_chartsheet"), info = names(opt))
})

test_that("chartsheet arguments are validated", {
  ch <- xl_chart("column", cs_series())
  expect_error(xl_chartsheet(ch, view = list(active = TRUE)),
               "must be an xl_sheet_view")
  expect_error(xl_chartsheet(ch, page = list(paper = "A4")),
               "must be an xl_page_setup")
  expect_error(xl_chartsheet(ch, zoom = 5), "must be between 10 and 400")
  expect_error(xl_chartsheet(ch, tab_color = "not a colour"), "tab_color")
  expect_error(xl_chartsheet(ch, protect = list(nope = TRUE)), "protect")
})

test_that("a chartsheet takes the two protection options it has, not a sheet's", {
  # the attribute asserts protection, so asking for no_* drops it: Excel
  # writes content="1" for a protected chart and nothing for an editable one
  plain <- cs_parts(xl_chart("column", cs_series()), protect = TRUE)
  expect_match(plain$sheet, 'content="1"', fixed = TRUE)
  expect_match(plain$sheet, 'objects="1"', fixed = TRUE)

  r <- cs_parts(xl_chart("column", cs_series()),
                protect = list(password = "s", no_content = TRUE,
                               no_objects = TRUE))
  expect_match(r$sheet, "sheetProtection", fixed = TRUE)
  expect_false(grepl('content="1"', r$sheet, fixed = TRUE))
  expect_false(grepl('objects="1"', r$sheet, fixed = TRUE))

  # one at a time, so neither flag is standing in for the other
  only_obj <- cs_parts(xl_chart("column", cs_series()),
                       protect = list(no_objects = TRUE))
  expect_match(only_obj$sheet, 'content="1"', fixed = TRUE)
  expect_false(grepl('objects="1"', only_obj$sheet, fixed = TRUE))
  only_con <- cs_parts(xl_chart("column", cs_series()),
                       protect = list(no_content = TRUE))
  expect_match(only_con$sheet, 'objects="1"', fixed = TRUE)
  expect_false(grepl('content="1"', only_con$sheet, fixed = TRUE))
})

test_that("a worksheet's protection options are refused on a chartsheet, and vice versa", {
  ch <- xl_chart("column", cs_series())
  expect_error(xl_chartsheet(ch, protect = list(format_cells = TRUE)),
               "a chartsheet has no cells")
  expect_error(xl_chartsheet(ch, protect = list(format_cells = TRUE)),
               "It accepts: no_content, no_objects", fixed = TRUE)
  expect_error(xl_sheet(cs_sales, protect = list(no_content = TRUE)),
               "belong to a chartsheet")
})

# ── Alongside the rest of the workbook ────────────────────────────────────────

test_that("a chartsheet counts as a drawing in the ordering check", {
  # a chartsheet's chart is a drawing, so it is a victim of the same
  # libxlsxwriter numbering bug a floating image is
  png <- tempfile(fileext = ".png")
  writeBin(as.raw(c(
    0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D,
    0x49, 0x48, 0x44, 0x52, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01,
    0x08, 0x06, 0x00, 0x00, 0x00, 0x1F, 0x15, 0xC4, 0x89, 0x00, 0x00, 0x00,
    0x0A, 0x49, 0x44, 0x41, 0x54, 0x78, 0x9C, 0x63, 0x00, 0x01, 0x00, 0x00,
    0x05, 0x00, 0x01, 0x0D, 0x0A, 0x2D, 0xB4, 0x00, 0x00, 0x00, 0x00, 0x49,
    0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82)), png)
  expect_error(
    write_xlsx(list(
      Data = xl_sheet(cs_sales,
                      page = xl_page_setup(header = "&L&G",
                                           header_image = list(left = png))),
      Chart = xl_chartsheet(xl_chart("column", cs_series()))),
      tempfile(fileext = ".xlsx")),
    "earlier in the workbook")
})

test_that("a chartsheet contributes no cells to the streaming estimate", {
  # it has no data frame, so it cannot tip the constant-memory decision
  res <- .resolve_constant_memory(
    list(Data = cs_sales, Overview = .chartsheet_placeholder()),
    xl_properties(), NULL, NA, 1)
  expect_equal(res$on, 1L)
})

test_that("a chartsheet may be given to xl_workbook() like any other sheet", {
  wb <- xl_workbook(list(Data = cs_sales,
                         Overview = xl_chartsheet(xl_chart("pie",
                                                           cs_series()))),
                    properties = xl_properties(title = "T"))
  expect_s3_class(wb$sheets$Overview, "xl_chartsheet")
  expect_silent(write_xlsx(wb, tempfile(fileext = ".xlsx")))
})

test_that("a chartsheet takes the page and view options it supports", {
  r <- cs_parts(xl_chart("column", cs_series()),
                view = xl_sheet_view(active = TRUE, first_tab = TRUE),
                page = xl_page_setup(orientation = "portrait",
                                     margins = c(1, 1, 1.2, 1.2),
                                     header = "&LLeft", footer = "&RRight",
                                     header_margin = 0.6,
                                     footer_margin = 0.7))
  expect_match(r$sheet, 'orientation="portrait"', fixed = TRUE)
  expect_match(r$sheet, 'left="1"', fixed = TRUE)
  expect_match(r$sheet, 'header="0.6"', fixed = TRUE)
  expect_match(r$sheet, 'footer="0.7"', fixed = TRUE)
  expect_match(r$sheet, "&amp;LLeft", fixed = TRUE)
  expect_match(r$sheet, "&amp;RRight", fixed = TRUE)
  expect_match(r$workbook, "activeTab", fixed = TRUE)
  expect_match(r$workbook, "firstSheet", fixed = TRUE)
})

test_that("a chartsheet's tab is selected by the same argument a worksheet's is", {
  # chartsheet_select() only sets a flag on the chartsheet, while <sheetView>
  # is written from the worksheet it wraps -- so writexl sets that one too, and
  # xl_sheet_view(selected =) means the same thing on either kind of sheet
  view_of <- function(part) sub(".*(<sheetView [^>]*>).*", "\1", part)

  cs <- cs_parts(xl_chart("column", cs_series()),
                 view = xl_sheet_view(selected = TRUE))
  expect_match(cs$sheet, 'tabSelected="1"', fixed = TRUE)

  p <- write_tmp(list(A = cs_sales,
                      B = xl_sheet(cs_sales,
                                   view = xl_sheet_view(selected = TRUE))))
  ws <- xlsx_part(p, "xl/worksheets/sheet2.xml", raw = TRUE)
  expect_match(ws, 'tabSelected="1"', fixed = TRUE)
  # the same element, byte for byte, on a chartsheet and on a worksheet
  expect_equal(view_of(cs$sheet), view_of(ws))
})

test_that("the active sheet is selected too, on either kind of sheet", {
  # Excel cannot show an active tab that is not also selected, and
  # libxlsxwriter's *_activate() set both
  cs <- cs_parts(xl_chart("column", cs_series()),
                 view = xl_sheet_view(active = TRUE))
  expect_match(cs$sheet, 'tabSelected="1"', fixed = TRUE)
  expect_match(cs$workbook, "activeTab", fixed = TRUE)

  p <- write_tmp(list(A = cs_sales,
                      B = xl_sheet(cs_sales,
                                   view = xl_sheet_view(active = TRUE))))
  expect_match(xlsx_part(p, "xl/worksheets/sheet2.xml", raw = TRUE),
               'tabSelected="1"', fixed = TRUE)
})

test_that("a sheet nobody selected carries no tabSelected", {
  cs <- cs_parts(xl_chart("column", cs_series()))
  expect_false(grepl("tabSelected", cs$sheet, fixed = TRUE))
})

test_that("a hidden chartsheet is hidden in the workbook", {
  r <- cs_parts(xl_chart("column", cs_series()),
                view = xl_sheet_view(visible = FALSE), name = "Hidden")
  expect_match(r$workbook, 'state="hidden"', fixed = TRUE)
})

test_that("a chart with no title has no unqualified range to complain about", {
  # named() returns TRUE for an absent range, so a titleless chart passes
  expect_s3_class(xl_chartsheet(xl_chart("column", cs_series())),
                  "xl_chartsheet")
})

#' General cell objects for Excel writing
#'
#' @description
#' `xl_cell_general` creates a vector of cell objects, each optionally
#' containing a **value**, a **formula**, and/or a **hyperlink**. It is the
#' fundamental building block used internally by [xl_formula()] and
#' [xl_hyperlink()], and can be used directly for mixed-type columns or cells
#' that combine multiple features (e.g., a formula with a pre-calculated
#' result, or a URL with separate display text and tooltip).
#'
#' An `xl_cell_general` behaves like a vector: it has a `length()`, supports
#' `[`, `[[`, `c()` and `rep()`, and recycles automatically when assigned to a
#' data frame column of a different length (just like [xl_formula()]).  It is
#' deliberately not a list, so that `df[, j] <- cells` assigns one column
#' rather than being read as a list of them.
#'
#' `df[i, j] <- x` sets that cell's `value`, whatever `x`'s type; a formula
#' needs [xl_formula()], since a cell column carries no column-wide notion of
#' "these are all formulas".
#'
#' @param value An atomic vector or a list of scalars, one per cell. Use `NA`
#'   for a cell that is empty unless `na` says otherwise (see `na` below, and
#'   note that a workbook-wide `na` reaches these values too). A list enables
#'   mixed types across cells
#'   in the same column (e.g., `list(1.5, "text", TRUE)`). Date and POSIXct
#'   scalars are supported and formatted as in [write_xlsx()].  When
#'   `hyperlink` is also set for the same cell, a **character** `value` is
#'   used as the display text shown in the cell instead of the raw URL; all
#'   other types are ignored for hyperlink cells.
#' @param formula A character vector of Excel formulas (each must start with
#'   `"="`), or `NA` for cells with no formula.  When both `value` and
#'   `formula` are supplied for the same cell, `value` is used as a
#'   pre-calculated result stored alongside the formula via
#'   `worksheet_write_formula_num()` (numeric value) or
#'   `worksheet_write_formula_str()` (character value).  This allows static
#'   xlsx exports that display formula text in the formula bar but do not
#'   require Excel to recalculate on open.
#' @param hyperlink A character vector of URLs, or a list where each element
#'   is `NA`, a single character URL, or a named list with elements:
#'   \describe{
#'     \item{`url`}{(required) The target URL.}
#'     \item{`tooltip`}{(optional) Tooltip text shown on hover.}
#'   }
#'   Supply a character `value` alongside `hyperlink` to show custom display
#'   text in the cell instead of the raw URL.  The hyperlink is written via
#'   `worksheet_write_url_opt()`.
#'
#' @param format An [xl_format] object (applied to every cell), or a list of
#'   `xl_format` objects (one per cell, recycled), or `NULL` for no formatting.
#'   Build formats with [xl_font()], [xl_fill()], [xl_border()], [xl_align()],
#'   [xl_num_format()] and [xl_protection()], combined with [xl_format()] or
#'   `+`.  When a cell's value is a date/time and its format sets no number
#'   format, the default date/time number format is applied automatically.
#' @param comment Cell comments (notes): a character vector of comment text
#'   (one per cell, recycled; `NA` for no comment), a single [xl_comment()]
#'   (recycled to every cell), or a list mixing strings / `xl_comment` / `NA`
#'   per cell.  `NULL` for no comments.
#' @param array,dynamic Logical (one per cell, recycled): how the cell's
#'   `formula` is stored.  `array = TRUE` writes a legacy *array* (Ctrl-Shift-
#'   Enter) formula; `dynamic = TRUE` writes a modern *dynamic array* formula,
#'   which Excel spills over as many cells as the result needs.  Neither can
#'   carry a character `value`, because Excel stores no cached string result for
#'   an array formula.  On a cell that has no `formula` the flags are inert, so a
#'   single `array = TRUE` can be recycled across a column that mixes formula
#'   and value cells.
#' @param na What to write in a cell that has no value, one per cell and
#'   recycled.  `NA` (the default) inherits the column's
#'   [xl_col_spec()]`(na = )`, and failing that the workbook's
#'   [xl_properties()]`(na = )`; a cell that sets its own overrides both.  This
#'   is also how a cell asks for a blank where the column or workbook would
#'   otherwise substitute something --- give it the value you want, `""` for an
#'   empty string.
#' @param array_range The range a legacy array formula covers, for the rare case
#'   where it must be declared: an Excel range string (`"C2:C11"`) or a
#'   `list(rows = , cols = )` spec, one per cell (`NA` for none). It must start
#'   at the cell holding the formula, and must extend into cells the sheet does
#'   not otherwise write --- the range is padded on write, so an overlap would
#'   have the padding and the sheet's own values overwrite each other.
#'
#'   Leave it unset for almost everything: a single-cell `dynamic` formula
#'   spills automatically in Excel 365 / 2021, and a single-cell `array` formula
#'   is the right spelling for a `SUMPRODUCT`-style aggregate. Supplying it
#'   forces the workbook out of the memory-efficient row-streaming mode, since
#'   libxlsxwriter cannot pad an array range while streaming.
#' @return An object of class `c("xl_cell_general", "xl_cell")`, which is a
#'   list of length `n` where each element is a named list with fields
#'   `value`, `formula`, `hyperlink`, `format`, `comment`, `array`, `dynamic`
#'   and `array_range`.
#'
#' @family cell content
#' @seealso [xl_formula()], [xl_hyperlink()], [write_xlsx()]
#' @export
#' @examples
#' # Value-only cell
#' xl_cell_general(value = 42)
#'
#' # Formula with a pre-calculated numeric result (static export)
#' xl_cell_general(value = 42.0, formula = "=SUM(A1:A10)")
#'
#' # Hyperlink with display text (value) and tooltip
#' xl_cell_general(
#'   value    = "Visit",
#'   hyperlink = list(url = "https://example.com", tooltip = "Go to example.com")
#' )
#'
#' # Vector of cells: value and formula cells in one column
#' cells <- c(
#'   xl_cell_general(value = 1.5),
#'   xl_cell_general(value = "note"),
#'   xl_cell_general(formula = "=A1+A2")
#' )
#'
#' # Used in a data frame (length-1 recycles to fill all rows, as with
#' # xl_formula())
#' df <- data.frame(x = 1:3)
#' df$formula_col <- xl_formula("=A1*2")   # backward-compatible shorthand
#' df$cell_col    <- xl_cell_general(value = 99L)  # all rows get 99
xl_cell_general <- function(value = NULL, formula = NULL, hyperlink = NULL,
                            format = NULL, comment = NULL, array = FALSE,
                            dynamic = FALSE, array_range = NULL, na = NA) {

  # Require at least one content argument -------------------------------------
  if (is.null(value) && is.null(formula) && is.null(hyperlink) &&
      is.null(comment))
    stop("At least one of 'value', 'formula', 'hyperlink', or 'comment' must ",
         "be provided. Use value = NA for a cell with nothing in it.",
         call. = FALSE)

  # Reject values writexl cannot write ----------------------------------------
  # Uses the same predicate as the column-level check in normalize_df(), so an
  # unsupported type cannot slip in by being wrapped in a cell object.  NULL
  # and NA are allowed: both mean "blank cell".
  if (!is.null(value)) {
    vals <- if (is_xl_rich_string(value)) list(value)
            else if (is.list(value)) value else list(value)
    keep <- !vapply(vals, is.null, logical(1))
    ok <- vapply(vals, function(v)
      is_xl_rich_string(v) || .is_supported_column(v), logical(1))
    bad <- which(keep & !ok)
    if (length(bad)) {
      detail <- vapply(bad, function(k) {
        v <- vals[[k]]
        cls <- paste(class(v), collapse = "/")
        ty <- typeof(v)
        lab <- if (identical(cls, ty)) cls else sprintf("%s (%s)", cls, ty)
        if (is.list(value)) sprintf("value[[%d]]: %s", k, lab)
        else sprintf("value: %s", lab)
      }, character(1))
      stop("xl_cell_general() cannot write these value type(s):\n  ",
           paste(detail, collapse = "\n  "), call. = FALSE)
    }
  }

  # Pre-normalise the hyperlink argument --------------------------------------
  # A named list with a 'url' element is a single hyperlink spec, not a list of
  # multiple hyperlinks.  Wrap it so length() = 1.
  if (is.list(hyperlink) && !is.null(hyperlink[["url"]])) {
    hyperlink <- list(hyperlink)
  }

  # Pre-normalise the array_range argument ------------------------------------
  # Same wrapping problem as `hyperlink`: a list(rows = , cols = ) spec is one
  # range, not a list of ranges.
  if (is.list(array_range) &&
      (!is.null(array_range[["rows"]]) || !is.null(array_range[["cols"]]))) {
    array_range <- list(array_range)
  }

  # Determine the output length n ---------------------------------------------
  # A single xl_comment counts as one comment (not its number of fields).
  comment_n <- if (is.null(comment)) 0L
               else if (is_xl_comment(comment)) 1L
               else length(comment)
  n <- max(c(
    if (!is.null(value))     .value_length(value) else 0L,
    if (!is.null(formula))   length(formula)   else 0L,
    if (!is.null(hyperlink)) length(hyperlink) else 0L,
    if (!is.null(array_range)) length(array_range) else 0L,
    comment_n
  ))

  # Normalise value to a list of length n -------------------------------------
  # An xl_rich_string is itself a list (of runs), so it has to be recognised
  # before the is.list() branch or its runs would be spread across cells.
  value_list <- if (is.null(value)) {
    rep(list(NA), n)
  } else if (is_xl_rich_string(value)) {
    rep(list(value), n)
  } else {
    rep_len(if (is.list(value)) value else as.list(value), n)
  }

  # Normalise formula to a character vector of length n -----------------------
  formula_vec <- if (is.null(formula)) {
    rep(NA_character_, n)
  } else {
    formula <- as.character(formula)
    bad <- !is.na(formula) & !startsWith(formula, "=")
    if (any(bad))
      stop("All non-NA formulas must start with '='", call. = FALSE)
    rep_len(formula, n)
  }

  # Normalise hyperlink to a list of length n ---------------------------------
  hyperlink_list <- if (is.null(hyperlink)) {
    rep(list(NA), n)
  } else {
    h <- if (is.list(hyperlink)) hyperlink else as.list(hyperlink)
    for (i in seq_along(h)) {
      el <- h[[i]]
      if (is.null(el))                                              next
      if (identical(el, NA))                                        next
      # character NA -> normalise to logical NA sentinel
      if (is.character(el) && length(el) == 1L && is.na(el)) {
        h[[i]] <- NA
        next
      }
      if (is.character(el) && length(el) == 1L)                    next
      if (is.list(el) && !is.null(el[["url"]]) &&
          is.character(el[["url"]]) && length(el[["url"]]) == 1L)  next
      stop(sprintf(
        paste0("hyperlink[[%d]] must be NA, a single character URL, or a ",
               "named list with a character 'url' element"),
        i), call. = FALSE)
    }
    rep_len(h, n)
  }

  # Normalise format to a list of length n ------------------------------------
  format_list <- if (is.null(format)) {
    vector("list", n)
  } else if (is_xl_format(format)) {
    rep(list(format), n)
  } else if (is.list(format)) {
    for (el in format)
      if (!is.null(el) && !is_xl_format(el))
        stop("each element of `format` must be an xl_format object or NULL",
             call. = FALSE)
    rep_len(format, n)
  } else {
    stop("`format` must be an xl_format object or a list of xl_format objects",
         call. = FALSE)
  }

  # Normalise comment to a list of length n -----------------------------------
  # Each element is NULL (no comment) or a C-ready comment payload.  A
  # character vector gives per-cell text with default options; a single
  # xl_comment recycles to every cell; a list mixes strings/xl_comment/NA.
  comment_list <- if (is.null(comment)) {
    vector("list", n)
  } else if (is_xl_comment(comment)) {
    rep(list(unclass(comment)), n)
  } else if (is.character(comment) || is.list(comment)) {
    rep_len(lapply(comment, .comment_payload), n)
  } else {
    stop("`comment` must be a character vector, an xl_comment, or a list of ",
         "strings/xl_comment objects", call. = FALSE)
  }

  # Normalise the array / dynamic flags and their range -----------------------
  array_vec   <- .cell_flag_vec(array, "array", n)
  dynamic_vec <- .cell_flag_vec(dynamic, "dynamic", n)
  range_list  <- .cell_range_list(array_range, n)
  .check_array_args(value_list, formula_vec, array_vec, dynamic_vec, range_list,
                    formula_given = !is.null(formula))
  .check_rich_value(value_list, formula_vec, hyperlink_list)

  # Normalise `na` to one entry per cell -------------------------------------
  # NULL where the cell inherits, so C can tell "this cell says blank is fine"
  # from "this cell has nothing to say".
  na_src <- if (is.list(na)) na else as.list(na)
  if (length(na_src) != 1L && length(na_src) != n)
    stop(sprintf("`na` must be length 1 or %d (one per cell), not %d", n,
                 length(na_src)), call. = FALSE)
  na_list <- rep_len(lapply(na_src, .val_na, arg = "na"), n)

  # Build the per-cell records ------------------------------------------------
  cells <- lapply(seq_len(n), function(i) {
    list(
      value       = value_list[[i]],
      formula     = formula_vec[[i]],
      hyperlink   = hyperlink_list[[i]],
      format      = format_list[[i]],
      comment     = comment_list[[i]],
      array       = array_vec[[i]],
      dynamic     = dynamic_vec[[i]],
      array_range = range_list[[i]],
      na          = na_list[[i]]
    )
  })

  .new_cell_vector(cells)
}

# --- The object ---------------------------------------------------------------
#
# An xl_cell_general is an integer vector carrying its per-cell records in an
# attribute, the shape factor uses.  The integers are bookkeeping: they are
# always seq_along(records), and nothing reads them.  What matters is that the
# object is *not* a list, because `[<-.data.frame` treats a list value as a
# list of columns -- so `df[, j] <- cells` assigned one cell record per column
# and lost the class, which is what broke a reverse dependency on 2.0.0.
#
# The carrier is an integer rather than the displayed text on purpose.  If the
# class is ever stripped -- recycling into a longer frame does it, in base R
# and in writexl 1.5.4 alike -- an integer leaves an obviously wrong 1, 2, 3
# rather than plausible-looking text that would write as silently wrong cells.

# The carrier is the authority on length, not the records.  R extends a data
# frame's rows through xpdrows.data.frame(), which does `class(y) <- NULL;
# length(y) <- n; attributes(y) <- saved` -- it strips the class, pads the
# carrier, and puts the old attributes back verbatim, so no method of ours is
# consulted and the records come back short.  Padding them here on the way out
# makes the object heal itself: the new cells read as empty, which is what
# extending a data frame means.
.cell_records <- function(x) {
  r <- attr(x, "records", exact = TRUE)
  n <- length(unclass(x))
  if (length(r) != n) {
    length(r) <- n
    r <- .fill_empty_cells(r)
  }
  r
}

.new_cell_vector <- function(records) {
  structure(seq_along(records), records = records,
            class = c("xl_cell_general", "xl_cell"))
}

#' @description
#' `as.character()` returns what each cell displays, so a cell built for a sheet
#' can be reused wherever a plain string is wanted --- [xl_merge()]'s or
#' [xl_comment()]'s `value`, for instance.  A cell that carries only a formula
#' shows `NA`: its displayed value comes from Excel, not from writexl.
#' @param x An `xl_cell_general`.
#' @param ... Ignored.
#' @rdname xl_cell_general
#' @export
as.character.xl_cell_general <- function(x, ...) {
  vapply(.cell_records(x), function(el) {
    v <- el[["value"]]
    if (is_xl_rich_string(v)) as.character(v)
    else if (is.null(v) || length(v) != 1L) NA_character_
    else as.character(v)
  }, character(1))
}

# A rich string is the cell's text, so it cannot share the cell with a formula
# (whose result Excel supplies) or a hyperlink (whose display text is a plain
# string).  Both would silently discard the runs, so refuse the combination.
.check_rich_value <- function(values, formulas, hyperlinks) {
  for (i in seq_along(values)) {
    if (!is_xl_rich_string(values[[i]])) next
    if (!is.na(formulas[[i]]))
      stop(sprintf(paste0("cell %d combines a rich string `value` with a ",
                          "`formula`; a formula's result comes from Excel, so ",
                          "the runs would be discarded"), i), call. = FALSE)
    hl <- hyperlinks[[i]]
    hl_set <- !(is.null(hl) || identical(hl, NA) ||
                (is.character(hl) && length(hl) == 1L && is.na(hl)))
    if (hl_set)
      stop(sprintf(paste0("cell %d combines a rich string `value` with a ",
                          "`hyperlink`; a hyperlink's display text is a plain ",
                          "string, so the runs would be discarded"), i),
           call. = FALSE)
  }
  invisible(NULL)
}

# The number of cells a `value` argument describes.  An xl_rich_string is one
# cell's worth of text, not one cell per run.
.value_length <- function(value) {
  if (is_xl_rich_string(value)) 1L else length(value)
}

# --- array / dynamic formula helpers -----------------------------------------

# Recycle a logical flag argument to length n, refusing NA (a flag is a
# decision, and NA would leave it ambiguous).
.cell_flag_vec <- function(x, arg, n) {
  if (is.null(x)) return(rep(FALSE, n))
  if (!is.logical(x) || !length(x) || anyNA(x))
    stop(sprintf("`%s` must be TRUE or FALSE (no NA)", arg), call. = FALSE)
  rep_len(x, n)
}

# Normalise array_range to a list of length n whose elements are NULL (no
# declared range) or a range spec to be resolved at write time.
.cell_range_list <- function(x, n) {
  if (is.null(x)) return(vector("list", n))
  r <- if (is.list(x)) x else as.list(x)
  r <- lapply(r, function(el) {
    if (is.null(el) || (length(el) == 1L && !is.list(el) && is.na(el)))
      return(NULL)
    if (is.character(el) && length(el) == 1L) return(el)
    if (is.list(el)) return(el)
    stop("each `array_range` must be NA, an Excel range string, or a ",
         "list(rows = , cols = ) spec", call. = FALSE)
  })
  rep_len(r, n)
}

# Validate the array/dynamic arguments.
#
# `array` / `dynamic` describe how a *formula* is stored, so on a cell with no
# formula they are simply inert -- which is what recycling a single TRUE across
# a column of mixed formula and value cells has to mean.  What is worth
# refusing is a flag that can never apply to anything, and a combination Excel
# cannot represent.
.check_array_args <- function(values, formulas, arrays, dynamics, ranges,
                              formula_given) {
  any_flag  <- any(arrays) || any(dynamics)
  any_range <- any(!vapply(ranges, is.null, logical(1)))
  if (any_flag && !formula_given)
    stop("`array`/`dynamic` describe how a formula is stored, but no `formula` ",
         "was supplied", call. = FALSE)
  if (any_range && !any_flag)
    stop("`array_range` applies only to an `array` or `dynamic` formula, and ",
         "neither is set", call. = FALSE)
  for (i in seq_along(formulas)) {
    if (is.na(formulas[[i]])) next          # flags are inert without a formula
    if (!(arrays[[i]] || dynamics[[i]])) next
    # libxlsxwriter has write_array_formula_num() but no _str() counterpart:
    # Excel stores no cached string result for an array formula.  Erroring beats
    # silently downgrading to a plain formula and losing the array semantics.
    v <- values[[i]]
    if (is.character(v) && length(v) == 1L && !is.na(v))
      stop(sprintf(paste0("cell %d combines a character `value` with ",
                          "`array`/`dynamic`; only a numeric pre-calculated ",
                          "result can be stored for an array formula"), i),
           call. = FALSE)
  }
  invisible(NULL)
}

# --- S3 methods --------------------------------------------------------------

#' @export
length.xl_cell_general <- function(x) length(unclass(x))

#' @export
`[.xl_cell_general` <- function(x, i, ...) {
  .new_cell_vector(.cell_records(x)[i])
}

# `x[[i]]` is one cell's record, the named list of its fields -- the carrier is
# bookkeeping and never the answer to a question about a cell.
#' @export
`[[.xl_cell_general` <- function(x, i, ...) .cell_records(x)[[i]]

#' @export
`[[<-.xl_cell_general` <- function(x, i, value) {
  recs <- .cell_records(x)
  recs[[i]] <- value
  .new_cell_vector(recs)
}

# Assigning into a cell column has to reach the records.  Without this method
# `df[i, j] <- value` writes into the integer carrier instead -- silently, and
# leaving the records one short -- because that is where `[<-.data.frame`
# sends it.  A bare value becomes the cell's `value`; an xl_cell_general on
# the right replaces those cells outright.
#' @export
`[<-.xl_cell_general` <- function(x, i, value) {
  recs <- .cell_records(x)
  .warn_formula_as_text(value)
  repl <- if (inherits(value, "xl_cell_general")) .cell_records(value)
          else .cell_records(xl_cell_general(value = value))
  # `i` may point past the end -- `d[3, 2] <-` on a one-row frame does exactly
  # that -- so index the list directly and let it grow.  seq_along(recs)[i]
  # would give NA there and drop the assignment on the floor.
  k <- if (is.logical(i)) sum(i, na.rm = TRUE) else length(i)
  recs[i] <- rep_len(repl, k)
  .new_cell_vector(.fill_empty_cells(recs))
}

# A string beginning with "=" assigned into a cell column is written as those
# characters, not as a formula: a cell column has no column-wide "these are all
# formulas".  That is the documented rule, but it is a quiet way to lose every
# formula on a sheet -- a reverse dependency shipped a workbook with twenty
# inert HYPERLINK strings and nothing said a word -- so say something.
#
# Only on assignment, where the ambiguity is.  xl_cell_general(value = "=x")
# writing text is deliberate and stays silent.
.warn_formula_as_text <- function(value) {
  if (inherits(value, "xl_cell_general") || !is.character(value)) return(invisible(NULL))
  v <- value[!is.na(value)]
  if (!length(v) || !any(startsWith(v, "="))) return(invisible(NULL))
  warning("assigning a string beginning with \"=\" into a cell column writes ",
          "those characters, not a formula. Use xl_formula() for a formula, ",
          "either per cell or over the finished column:
",
          "  df[[j]] <- xl_formula(df[[j]])", call. = FALSE)
  invisible(NULL)
}

# Extending a data frame pads its other columns, and `[<-.data.frame` does that
# through `length<-`.  Without this method the cell column stays short and the
# frame is refused with "replacement has N rows, data has N+1".
#' @export
`length<-.xl_cell_general` <- function(x, value) {
  recs <- .cell_records(x)
  length(recs) <- value
  .new_cell_vector(.fill_empty_cells(recs))
}

# `[<-` and `length<-` can both leave NULL holes past the old end.  A hole is a
# cell with nothing in it, which is what an unwritten cell already means.
.fill_empty_cells <- function(recs) {
  holes <- vapply(recs, is.null, logical(1))
  if (any(holes))
    recs[holes] <- list(.cell_records(xl_cell_general(value = NA))[[1L]])
  recs
}

#' @export
c.xl_cell_general <- function(x, ...) {
  parts <- lapply(list(x, ...), function(el)
    if (inherits(el, "xl_cell_general")) .cell_records(el)
    else .cell_records(xl_cell_general(value = el)))
  .new_cell_vector(do.call(c, parts))
}

#' @export
rep.xl_cell_general <- function(x, ...) {
  .new_cell_vector(rep(.cell_records(x), ...))
}

#' @export
as.data.frame.xl_cell_general <- function(x, row.names = NULL, optional = FALSE, ...) {
  # Omit names so that data.frame() uses the caller's argument name rather
  # than an arbitrary inner name.  This mirrors how I(list(...)) behaves:
  # names(xi) == NULL causes data.frame() to leave vnames[[i]] as the
  # argument name instead of overriding it.
  structure(
    list(x),
    class     = "data.frame",
    row.names = if (is.null(row.names)) seq_along(x) else row.names
  )
}

#' @export
print.xl_cell_general <- function(x, max = 10L, ...) {
  n <- length(x)
  cat(sprintf("[xl_cell_general: %d cell%s]\n", n, if (n == 1L) "" else "s"))
  show <- seq_len(min(n, max))
  for (i in show) {
    cell  <- .cell_records(x)[[i]]
    parts <- character(0L)

    val <- cell[["value"]]
    has_val <- !is.null(val) && length(val) > 0L &&
               !all(is.na(val))
    if (has_val)
      parts <- c(parts, paste0("value=", format(val)))

    fml <- cell[["formula"]]
    if (!is.null(fml) && !is.na(fml))
      parts <- c(parts, paste0("formula=", fml))

    hlk <- cell[["hyperlink"]]
    hlk_set <- !is.null(hlk) && !identical(hlk, NA) &&
               !(is.character(hlk) && length(hlk) == 1L && is.na(hlk))
    if (hlk_set)
      parts <- c(parts, "hyperlink=<set>")

    if (is_xl_format(cell[["format"]]))
      parts <- c(parts, "format=<set>")

    if (!is.null(cell[["comment"]]))
      parts <- c(parts, "comment=<set>")

    cat(sprintf("  [%d] %s\n", i,
                if (length(parts)) paste(parts, collapse = ", ")
                else "<empty>"))
  }
  if (n > max)
    cat(sprintf("  ... (%d more)\n", n - max))
  invisible(x)
}

#' Version
#'
#' Shows version of bundled libxlsxwriter.
#'
#' @export
#' @rdname writexl
#' @useDynLib writexl C_lxw_version
lxw_version <- function(){
  version <- .Call(C_lxw_version)
  as.numeric_version(version)
}

#' @useDynLib writexl C_set_tempdir
.onLoad <- function(lib, pkg){
  # runs before any test does, so covr never records it
  .Call(C_set_tempdir, tempdir())  # nocov
}

# =============================================================================
# Charts
# =============================================================================
#
# A chart is a drawing placed on a worksheet, so its placement arguments are
# xl_image()'s: lxw_chart_options is lxw_image_options minus url, tip and
# cell_format, with the same field names.  Keeping the R names identical is not
# cosmetic -- the API-consistency gate enforces it.
#
# A series names two ranges: the values, and optionally the categories to plot
# them against.  chart_series_set_values() and _set_categories() take a sheet
# name plus a 0-based quad rather than an A1 string, so the shared range
# resolver's output plugs straight in.  A series may point at a different sheet
# from the one the chart sits on, which is why ranges carry an optional `sheet`
# and are resolved once the whole workbook is known.
#
# Charts themselves produce a drawing part, so they do not desync
# libxlsxwriter's drawing-id counter -- but they are *victims* of it in the
# same way a floating image is, which .check_drawing_order() accounts for.
# -----------------------------------------------------------------------------

# The chart types, named as Excel names them rather than as the enum spells it.
.LXW_CHART_TYPE <- c(
  area = 1L, area_stacked = 2L, area_stacked_percent = 3L,
  bar = 4L, bar_stacked = 5L, bar_stacked_percent = 6L,
  column = 7L, column_stacked = 8L, column_stacked_percent = 9L,
  doughnut = 10L,
  line = 11L, line_stacked = 12L, line_stacked_percent = 13L,
  pie = 14L,
  scatter = 15L, scatter_straight = 16L, scatter_straight_markers = 17L,
  scatter_smooth = 18L, scatter_smooth_markers = 19L,
  radar = 20L, radar_markers = 21L, radar_filled = 22L
)

# In the enum's own order, so it can be read against the header.  NONE is 0
# rather than one past the end, and chart_legend_set_position() rejects
# anything above OVERLAY_TOP_RIGHT.
.LXW_CHART_LEGEND <- c(
  none = 0L, right = 1L, left = 2L, top = 3L, bottom = 4L, top_right = 5L,
  overlay_right = 6L, overlay_left = 7L, overlay_top_right = 8L
)

# --- Which features a chart type supports ------------------------------------
#
# libxlsxwriter documents these restrictions in prose and then ignores them:
# chart_set_hole_size() on a pie chart, or up-down bars on a column chart, are
# accepted and silently dropped by Excel.  The matrix is here so the refusal
# happens in R, where it can name the type.
.CHART_FAMILY <- function(type) {
  if (type %in% c("pie", "doughnut")) return("pie")
  if (startsWith(type, "scatter")) return("scatter")
  if (startsWith(type, "radar")) return("radar")
  if (startsWith(type, "bar") || startsWith(type, "column")) return("bar")
  if (startsWith(type, "line")) return("line")
  if (startsWith(type, "area")) return("area")
  # every name in .LXW_CHART_TYPE matches one of the tests above, and every
  # caller passes a type .val_enum() has already checked; a test pins that
  "other"  # nocov
}

# feature -> the families that support it
.CHART_FEATURE_FAMILIES <- list(
  axes            = c("scatter", "radar", "bar", "line", "area", "other"),
  hole_size       = "pie",     # doughnut only, narrowed below
  rotation        = "pie",
  up_down_bars    = "line",
  high_low_lines  = "line",
  drop_lines      = c("line", "area"),
  series_gap      = "bar",
  series_overlap  = "bar",
  smooth          = c("line", "scatter")
)

# Is `feature` legal for `type`?  Two features are narrower than their family.
.chart_supports <- function(type, feature) {
  if (feature == "hole_size") return(identical(type, "doughnut"))
  fam <- .CHART_FAMILY(type)
  fam %in% .CHART_FEATURE_FAMILIES[[feature]]
}

.check_chart_feature <- function(type, feature, arg) {
  if (.chart_supports(type, feature)) return(invisible(NULL))
  ok <- Filter(function(t) .chart_supports(t, feature), names(.LXW_CHART_TYPE))
  stop(sprintf(paste0("`%s` does not apply to a \"%s\" chart; Excel drops it ",
                      "silently.\n  It applies to: %s."),
               arg, type, paste(ok, collapse = ", ")), call. = FALSE)
}

# --- Ranges a series points at ------------------------------------------------
#
# A range may name a different sheet, so it cannot be resolved until every
# sheet is known.  Until then it is carried as given.
.chart_range <- function(x, arg) {
  if (is.null(x)) return(NULL)
  if (is.character(x) && length(x) == 1L && !is.na(x)) return(list(spec = x))
  if (is.list(x)) {
    nms <- names(x)
    if (is.null(nms) || any(!nzchar(nms)))
      stop(sprintf(paste0("`%s` list must be named, e.g. ",
                          "list(sheet = \"Data\", cols = \"revenue\")"), arg),
           call. = FALSE)
    unknown <- setdiff(nms, c("sheet", "rows", "cols", "header"))
    if (length(unknown))
      stop(sprintf("unknown `%s` element(s): %s", arg,
                   paste(unknown, collapse = ", ")), call. = FALSE)
    # `header` names one column's header cell.  Its own element rather than
    # `rows = 0`, because `rows` counts data rows everywhere else in writexl.
    if (!is.null(x[["header"]])) {
      if (!is.null(x[["rows"]]) || !is.null(x[["cols"]]))
        stop(sprintf(paste0("`%s` gives `header` as well as rows/cols; ",
                            "`header` is the header cell of one column, so it ",
                            "stands alone."), arg), call. = FALSE)
      if (length(x[["header"]]) != 1L || anyNA(x[["header"]]))
        stop(sprintf("`%s$header` must name a single column", arg),
             call. = FALSE)
    }
    sheet <- x[["sheet"]]
    if (!is.null(sheet) &&
        (!is.character(sheet) || length(sheet) != 1L || is.na(sheet)))
      stop(sprintf("`%s$sheet` must be a single sheet name", arg),
           call. = FALSE)
    return(list(spec = x[setdiff(nms, "sheet")], sheet = sheet))
  }
  stop(sprintf(paste0("`%s` must be a range string or a list(sheet = , rows = ",
                      ", cols = ) spec"), arg), call. = FALSE)
}

#' A data series within a chart
#'
#' @description
#' `xl_chart_series()` names the values a chart plots, and optionally the
#' categories to plot them against and a name for the legend.  A series that
#' plots a column is named after that column's header unless told otherwise.
#'
#' Each range may live on a different sheet from the chart, so it takes an
#' optional `sheet`:
#'
#' * `"Data!B2:B10"` --- an A1 range, sheet-qualified;
#' * `list(cols = "revenue")` --- resolved against the chart's own sheet;
#' * `list(sheet = "Data", cols = "revenue")` --- against another sheet;
#' * `list(header = "revenue")` --- that column's header cell, which is where
#'   a series name usually lives.
#'
#' A range that selects no data is an error rather than an empty chart.
#'
#' @param values The range holding the numbers to plot.
#' @param categories The range holding the labels to plot them against.  Omit
#'   for a chart that numbers its points.
#' @param name The series name, shown in the legend.  Left unset, a series that
#'   plots a column takes its name from that column's header cell, which is
#'   what Excel does when you chart a column along with its header; `FALSE`
#'   leaves it unnamed.  A string is always taken literally --- a series may
#'   legitimately be called `"Q1!"` --- so to take the name from another cell,
#'   give a range spec: `name = list(header = "cost")` for a different column's
#'   header, or `name = list(rows = 1, cols = 1)` for a data cell.
#' @param format An [xl_format] styling the series --- its line and fill.  See
#'   [xl_chart()] for which format properties a chart can express.
#' @param marker An [xl_chart_marker()] drawn at each point.
#' @param labels An [xl_chart_labels()] printing the numbers beside the points.
#' @param trendline An [xl_chart_trendline()] fitted through the series.
#' @param x_error_bars,y_error_bars An [xl_chart_error_bars()] on each point.
#' @param points An [xl_format()] per point, as a list, styling individual
#'   points --- one slice of a pie, one bar of a column chart.  `NULL` in the
#'   list leaves that point as it is.
#' @param smooth Draw the line smoothed.  Line and scatter charts only.
#' @param invert_if_negative Fill negative values with the inverse colour.
#' @return An `xl_chart_series` object.
#' @family images and charts
#' @seealso [xl_chart]
#' @export
#' @examples
#' xl_chart_series(values = list(cols = "revenue"))
#' xl_chart_series(values = "Data!B2:B10", categories = "Data!A2:A10",
#'                 name = "2024")
xl_chart_series <- function(values, categories = NULL, name = NULL,
                            format = NULL, marker = NULL, labels = NULL,
                            trendline = NULL, x_error_bars = NULL,
                            y_error_bars = NULL, points = NULL, smooth = NA,
                            invert_if_negative = NA) {
  if (missing(values) || is.null(values))
    stop("`values` must name the range holding the numbers to plot",
         call. = FALSE)
  if (!is.null(format) && !is_xl_format(format))
    stop("`format` must be an xl_format object", call. = FALSE)
  # Translated again at write time, when the payload is needed; here too so
  # that a format a series cannot use is refused where it was written.
  .chart_format_payload(format, "format",
                        accept = c("line", "fill", "pattern"),
                        part = "a chart series")
  parts <- list(marker = marker, labels = labels, trendline = trendline,
                x_error_bars = x_error_bars, y_error_bars = y_error_bars)
  for (k in names(parts))
    if (!is.null(parts[[k]]) && !inherits(parts[[k]], .PART_CLASS[[k]]))
      stop(sprintf("`%s` must be an %s() object", k,
                   sub("^xl", "xl", .PART_CLASS[[k]])), call. = FALSE)
  nm <- if (is.null(name)) NULL
        else if (identical(name, FALSE)) list(off = TRUE)
        else if (is.character(name) && length(name) == 1L && !is.na(name))
          list(text = name)
        else .chart_range(name, "name")
  structure(.drop_null(list(
    values = .chart_range(values, "values"),
    categories = .chart_range(categories, "categories"),
    name = nm, format = format,
    marker = marker, labels = labels, trendline = trendline,
    x_error_bars = x_error_bars, y_error_bars = y_error_bars,
    points = .points_payload(points, "points"),
    smooth = .val_flag(smooth, "smooth"),
    invert_if_negative = .val_flag(invert_if_negative, "invert_if_negative")
  )), class = "xl_chart_series")
}

#' @export
print.xl_chart_series <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_chart_series%s>\n",
              if (!is.null(p[["name"]][["text"]]))
                paste0(": ", p[["name"]][["text"]]) else ""))
  invisible(x)
}

#' Add a chart to a worksheet
#'
#' @description
#' `xl_chart()` builds a chart from one or more [xl_chart_series()] and places
#' it on a sheet, anchored to a cell.  Pass one or a list of them as
#' `xl_sheet(chart = )`.
#'
#' Placement works exactly as it does for [xl_image()] --- `at`, `scale`,
#' `offset`, `position`, `description` and `decorative` mean the same things,
#' because libxlsxwriter describes both with the same fields.
#'
#' @section What a chart type supports:
#' Excel silently drops options a chart type cannot use, so writexl refuses them
#' instead, naming the types that would work.  Pie and doughnut charts have no
#' axes; only a doughnut has a hole; only pie and doughnut rotate; up-down bars
#' and high-low lines are line-only; the series gap and overlap are bar and
#' column only.
#'
#' @param type The chart type: `"column"`, `"bar"`, `"line"`, `"pie"`,
#'   `"doughnut"`, `"area"`, `"scatter"`, `"radar"`, and the stacked,
#'   percent-stacked, smoothed and marker variants.
#' @param series One [xl_chart_series()], or a list of them.  Every series of a
#'   scatter chart must have `categories`, which are its x axis.
#' @param title The chart title.  A string is always taken literally, so to
#'   take the title from a cell give a range spec ---
#'   `list(header = "revenue")` for a column's header cell, or
#'   `list(rows = 1, cols = 1)` for a data cell.  `FALSE` removes the title
#'   Excel would otherwise generate.
#' @param title_format An [xl_format()] styling the title text.  A title is
#'   text, so only the [xl_font()] group applies.
#' @param title_layout Where to put the title by hand, as `c(x, y)` fractions
#'   of the chart.  Excel places it for you otherwise.
#' @param title_overlay Let the title sit over the plot rather than above it.
#' @param legend An [xl_chart_legend()] moving, styling or removing the legend.
#' @param data_table An [xl_chart_table()] printing the plotted numbers in a
#'   grid beneath the chart.
#' @param plot_area_format,chart_area_format An [xl_format()] styling the plot
#'   area --- the panel the data is drawn in --- and the chart area around it:
#'   [xl_border()] for the line, [xl_fill()] for the fill or pattern.
#' @param plot_area_layout Where to put the plot area by hand, as `c(x, y)`
#'   or `c(x, y, width, height)` fractions of the chart.
#' @param drop_lines Drop lines from each point to the category axis: `TRUE`,
#'   or an [xl_format()] giving the line to draw them with.  Line and area
#'   charts.
#' @param high_low_lines A line joining the highest and lowest series at each
#'   category, the same way.  Line charts.
#' @param up_down_bars Bars between the first and last series at each category:
#'   `TRUE`, or `list(up = , down = )` with an [xl_format()] for either bar.
#'   Line charts.
#' @param hole_size The size of a doughnut's hole, 10 to 90 percent.
#' @param rotation Where a pie or doughnut starts, 0 to 360 degrees clockwise
#'   from the top.
#' @param series_gap The gap between category groups on a bar or column chart,
#'   0 to 500 percent of a bar's width.
#' @param series_overlap How far bars of one category overlap, -100 to 100
#'   percent.  100 stacks them, -100 pushes them apart.
#' @param show_blanks What an empty cell does to the plot: leave a `"gap"`,
#'   plot it as `"zero"`, or join across it with `"connected"`.
#' @param show_hidden_data Plot data from rows and columns that are hidden.
#'   Excel leaves them out otherwise.
#' @param x_axis,y_axis An [xl_chart_axis()] describing that axis.  Pie and
#'   doughnut charts have none, and several axis options apply to a value or a
#'   category axis only --- see [xl_chart_axis()].
#' @param at The cell the chart's top-left corner is anchored to.
#' @param scale Scale factor: one number for both axes, or `c(x, y)`.
#' @param offset Offset from the anchor cell's corner in pixels, as `c(x, y)`.
#' @param position How the chart behaves when rows and columns change size; see
#'   [xl_image()].
#' @param description Alt text, for screen readers.
#' @param decorative Mark the chart as decorative, so screen readers skip it.
#' @param style Excel's built-in chart style, 1--48.
#' @return An `xl_chart` object.
#' @family images and charts
#' @seealso [xl_chart_series], [xl_sheet]
#' @export
#' @examples
#' xl_chart("column", xl_chart_series(values = list(cols = "revenue")))
#' xl_chart("pie", xl_chart_series(values = "Data!B2:B5"), title = "Share")
xl_chart <- function(type, series, title = NULL, title_format = NULL,
                     title_layout = NULL, title_overlay = NA,
                     x_axis = NULL, y_axis = NULL,
                     legend = NULL, data_table = NULL,
                     plot_area_format = NULL, plot_area_layout = NULL,
                     chart_area_format = NULL,
                     drop_lines = NA, high_low_lines = NA, up_down_bars = NA,
                     hole_size = NA, rotation = NA,
                     series_gap = NA, series_overlap = NA,
                     show_blanks = NULL, show_hidden_data = NA,
                     at = "A1", scale = 1,
                     offset = NULL, position = "move_and_size",
                     description = NULL, decorative = FALSE, style = NA) {
  ty <- .val_enum(type, names(.LXW_CHART_TYPE), "type")
  if (is.null(ty))
    stop("`type` must name a chart type, e.g. \"column\"", call. = FALSE)
  if (missing(series))
    stop("`series` must give at least one xl_chart_series() to plot",
         call. = FALSE)
  ss <- .chart_series_list(series)
  if (!length(ss))
    stop("a chart needs at least one series", call. = FALSE)
  .check_axis(x_axis, "x", ty, "x_axis")
  .check_axis(y_axis, "y", ty, "y_axis")
  for (nm in c("legend", "data_table"))
    if (!is.null(get(nm)) && !inherits(get(nm), paste0("xl_chart_",
                                                       sub("data_", "", nm))))
      stop(sprintf("`%s` must be an xl_chart_%s() object", nm,
                   sub("data_", "", nm)), call. = FALSE)

  # the six that belong to one family, checked against the matrix below
  drop_l <- .chart_line_switch(drop_lines, "drop_lines")
  high_l <- .chart_line_switch(high_low_lines, "high_low_lines")
  updown <- .chart_up_down(up_down_bars, "up_down_bars")
  hole <- .val_int(hole_size, "hole_size", min = 10, max = 90)
  rot <- .val_int(rotation, "rotation", min = 0, max = 360)
  gap <- .val_int(series_gap, "series_gap", min = 0, max = 500)
  ovl <- .val_int(series_overlap, "series_overlap", min = -100, max = 100)
  for (f in list(list(drop_l, "drop_lines"), list(high_l, "high_low_lines"),
                 list(updown, "up_down_bars"), list(hole, "hole_size"),
                 list(rot, "rotation"), list(gap, "series_gap"),
                 list(ovl, "series_overlap")))
    if (!is.null(f[[1L]])) .check_chart_feature(ty, f[[2L]], f[[2L]])
  for (i in seq_along(ss))
    .check_series_parts(ss[[i]], ty, sprintf("series[[%d]]", i))

  for (i in seq_along(ss)) {
    p <- unclass(ss[[i]])
    if (isTRUE(p[["smooth"]]))
      .check_chart_feature(ty, "smooth", sprintf("series[[%d]]$smooth", i))
    # A scatter plots x against y, so a series with no categories has no x
    # values.  libxlsxwriter requires them too, but only checks when the range
    # is passed to chart_add_series() as a string; set through
    # chart_series_set_values() it reaches _chart_write_x_val(), which unlike
    # _chart_write_cat() has no NULL guard, and strpbrk() is handed the NULL
    # formula.
    if (identical(.CHART_FAMILY(ty), "scatter") && is.null(p[["categories"]]))
      stop(sprintf(paste0("`series[[%d]]` has no `categories`, which a \"%s\" ",
                          "chart needs: a scatter plots values against ",
                          "categories, so the categories are its x axis."),
                   i, ty), call. = FALSE)
  }

  pair <- function(x, arg, what) {
    if (is.null(x)) return(NULL)
    if (!is.numeric(x) || !length(x) %in% c(1L, 2L) || anyNA(x))
      stop(sprintf("`%s` must be one number or two, %s", arg, what),
           call. = FALSE)
    if (length(x) == 1L) x <- c(x, x)
    as.numeric(x)
  }
  sc <- pair(scale, "scale", "as c(x, y)")
  if (!is.null(sc) && any(sc <= 0))
    stop("`scale` must be positive", call. = FALSE)

  ttl <- if (is.null(title)) NULL
         else if (identical(title, FALSE)) list(off = TRUE)
         else if (is.character(title) && length(title) == 1L && !is.na(title))
           list(text = title)
         else .chart_range(title, "title")
  # A title's only styling is its font, so a fill or a border here has
  # nowhere to go, and a title switched off has nothing to style.
  if (!is.null(title_format)) {
    if (isTRUE(ttl[["off"]]))
      stop("`title_format` styles the title, but `title = FALSE` removes it",
           call. = FALSE)
    # The automatic title of a single-series chart is Excel's own: with no
    # `title` there is no <c:title> element for the font to attach to.
    if (is.null(ttl))
      stop(paste0("`title_format` styles the title, so give a `title` too. ",
                  "The title Excel generates for a single-series chart is not ",
                  "in the file, so there is nothing to style."), call. = FALSE)
    ttl <- c(if (is.null(ttl)) list() else ttl,
             list(format = .chart_format_payload(title_format, "title_format",
                                                 accept = "font",
                                                 part = "a chart title")))
  }

  ttl <- .chart_title_extras(ttl, title_layout, title_overlay)

  structure(.drop_null(list(
    type = ty, series = ss, title = ttl, x_axis = x_axis, y_axis = y_axis,
    legend = legend, data_table = data_table,
    plot_area_format = .chart_format_payload(
      plot_area_format, "plot_area_format",
      accept = c("line", "fill", "pattern"), part = "a plot area"),
    plot_area_layout = .chart_layout(plot_area_layout,
                                     "plot_area_layout"),
    chart_area_format = .chart_format_payload(
      chart_area_format, "chart_area_format",
      accept = c("line", "fill", "pattern"), part = "a chart area"),
    drop_lines = drop_l, high_low_lines = high_l, up_down_bars = updown,
    hole_size = hole, rotation = rot, series_gap = gap, series_overlap = ovl,
    show_blanks = .val_enum(show_blanks, names(.LXW_CHART_BLANKS),
                            "show_blanks"),
    show_hidden_data = .val_flag(show_hidden_data, "show_hidden_data"),
    at = at, scale = sc, offset = pair(offset, "offset", "of pixels as c(x, y)"),
    position = .val_enum(position, names(.LXW_OBJECT_POSITION), "position"),
    description = if (is.null(description)) NULL else {
      if (!is.character(description) || length(description) != 1L ||
          is.na(description))
        stop("`description` must be a single non-NA string", call. = FALSE)
      description
    },
    decorative = .val_flag(decorative, "decorative"),
    style = .val_int(style, "style", min = 1, max = 48)
  )), class = "xl_chart")
}

#' @export
print.xl_chart <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_chart: %s, %d series%s>\n", p[["type"]],
              length(p[["series"]]),
              if (!is.null(p[["title"]][["text"]]))
                paste0(", \"", p[["title"]][["text"]], "\"") else ""))
  invisible(x)
}

# --- Resolving a series range against the workbook ---------------------------
#
# A series may plot data from a sheet other than the one the chart sits on, so
# a range cannot be resolved until every sheet is known.  `sheets` maps sheet
# name to the resolved data frame; `own` is the chart's own sheet, used when the
# range names none.
# The header cell of the single column a values spec names, or NULL when there
# is no one column to take it from.
.series_header_name <- function(values, header_offset) {
  if (header_offset < 1L || is.null(values)) return(NULL)
  spec <- values[["spec"]]
  if (!is.list(spec)) return(NULL)             # an A1 range, not a column spec
  cols <- spec[["cols"]]
  if (is.null(cols) || length(cols) != 1L || anyNA(cols)) return(NULL)
  list(spec = list(header = cols), sheet = values[["sheet"]])
}

.resolve_chart_range <- function(rng, arg, sheets, own, header_offset) {
  if (is.null(rng)) return(NULL)
  spec <- rng[["spec"]]
  sheet <- rng[["sheet"]]
  # a string may carry its own qualifier -- "Data!B2:B10" -- which the shared
  # resolver parses; the list form carries it as an element
  if (is.character(spec) && is.null(sheet)) {
    sp <- .split_sheet_ref(spec)
    sheet <- sp$sheet
    spec <- sp$rest
  }
  target <- if (is.null(sheet)) own else sheet
  if (!target %in% names(sheets))
    stop(sprintf(paste0("`%s` names sheet \"%s\", which is not in the ",
                        "workbook.\n  Sheets are: %s."),
                 arg, target, paste(sprintf("\"%s\"", names(sheets)),
                                    collapse = ", ")), call. = FALSE)
  df <- sheets[[target]]
  if (is.list(spec) && !is.null(spec[["header"]])) {
    if (header_offset < 1L)
      stop(sprintf(paste0("`%s` names a header cell, but this workbook is ",
                          "written without headers (col_names = FALSE), so ",
                          "there is no header row."), arg), call. = FALSE)
    col <- .resolve_col_index(spec[["header"]], names(df), arg)
    # 0-based, and the header is the row above data row 1
    return(list(sheet = target,
                range = as.integer(c(0L, col - 1L, 0L, col - 1L))))
  }
  q <- .xl_resolve_range(spec, arg = arg, df = df,
                         header_offset = header_offset, allow_cell = TRUE)
  # A range outside the data plots nothing, and Excel shows an empty chart with
  # no complaint -- so the emptiness is caught here rather than delivered.
  last_row <- (nrow(df) - 1L) + header_offset
  if (q[1L] > last_row || q[2L] > length(df) - 1L)
    stop(sprintf(paste0("`%s` selects no data: sheet \"%s\" has %d row(s) and ",
                        "%d column(s), and the range starts outside them."),
                 arg, target, nrow(df), length(df)), call. = FALSE)
  if (nrow(df) < 1L)
    stop(sprintf("`%s` names sheet \"%s\", which has no data rows", arg,
                 target), call. = FALSE)
  list(sheet = target, range = as.integer(q))
}

# Resolve a sheet's charts to the C payloads.  Called once the whole workbook is
# known, since a series may point at another sheet.
.resolve_charts <- function(el, df, reg, header_offset, props, sheets, own) {
  if (inherits(el, "xl_chartsheet")) {
    # a chartsheet has no cells, so `own` names a sheet with no columns and a
    # range that does not name its sheet has nothing to resolve against
    .check_chartsheet_ranges(unclass(el)$chart, own)
    cs <- list(unclass(el)$chart)
  } else if (!inherits(el, "xl_sheet")) {
    return(list())
  } else {
    cs <- .chart_list(el$chart)
  }
  lapply(seq_along(cs), function(i) {
    p <- unclass(cs[[i]])
    at <- .xl_resolve_range(p[["at"]], arg = sprintf("chart[[%d]] at", i),
                            df = df, header_offset = header_offset,
                            allow_cell = TRUE)
    if (at[1L] != at[3L] || at[2L] != at[4L])
      stop(sprintf(paste0("`chart[[%d]] at` must name a single cell: a chart ",
                          "is anchored to one cell"), i), call. = FALSE)
    ent <- list(type = unname(.LXW_CHART_TYPE[[p[["type"]]]]),
                row = at[1L], col = at[2L])
    if (!is.null(p[["scale"]])) {
      ent$x_scale <- p[["scale"]][1L]
      ent$y_scale <- p[["scale"]][2L]
    }
    if (!is.null(p[["offset"]])) {
      ent$x_offset <- as.integer(p[["offset"]][1L])
      ent$y_offset <- as.integer(p[["offset"]][2L])
    }
    if (!is.null(p[["position"]]))
      ent$object_position <- unname(.LXW_OBJECT_POSITION[[p[["position"]]]])
    if (!is.null(p[["description"]])) ent$description <- p[["description"]]
    if (isTRUE(p[["decorative"]]))    ent$decorative <- 1L
    if (!is.null(p[["style"]]))       ent$style <- as.integer(p[["style"]])

    ttl <- p[["title"]]
    if (!is.null(ttl)) {
      ent$title_format <- ttl[["format"]]
      if (!is.null(ttl[["layout"]])) ent$title_layout <- ttl[["layout"]]
      if (!is.null(ttl[["overlay"]])) ent$title_overlay <- ttl[["overlay"]]
      if (isTRUE(ttl[["off"]]))        ent$title_off <- 1L
      else if (!is.null(ttl[["text"]])) ent$title <- ttl[["text"]]
      # `title_format` alone leaves Excel's automatic title in place and only
      # styles it, so there is no range to resolve
      else if (!is.null(ttl[["spec"]])) {
        r <- .resolve_chart_range(ttl, sprintf("chart[[%d]] title", i),
                                  sheets, own, header_offset)
        ent$title_sheet <- r$sheet
        ent$title_range <- r$range
      }
    }

    ent$legend     <- .legend_payload(p[["legend"]])
    ent$data_table <- .chart_table_payload(p[["data_table"]])
    for (k in c("plot_area_format", "chart_area_format")) {
      f <- p[[k]]
      if (is.null(f)) next
      pre <- sub("_format$", "", k)
      ent[[paste0(pre, "_line")]]    <- f[["line"]]
      ent[[paste0(pre, "_fill")]]    <- f[["fill"]]
      ent[[paste0(pre, "_pattern")]] <- f[["pattern"]]
    }
    ent$plot_area_layout <- p[["plot_area_layout"]]
    ent$drop_lines     <- p[["drop_lines"]]
    ent$high_low_lines <- p[["high_low_lines"]]
    ent$up_down_bars   <- p[["up_down_bars"]]
    ent$hole_size      <- p[["hole_size"]]
    ent$rotation       <- p[["rotation"]]
    ent$series_gap     <- p[["series_gap"]]
    ent$series_overlap <- p[["series_overlap"]]
    if (!is.null(p[["show_blanks"]]))
      ent$show_blanks <- unname(.LXW_CHART_BLANKS[[p[["show_blanks"]]]])
    if (isTRUE(p[["show_hidden_data"]])) ent$show_hidden_data <- 1L

    ent$x_axis <- .axis_payload(p[["x_axis"]], sprintf("chart[[%d]] x_axis", i),
                                sheets, own, header_offset)
    ent$y_axis <- .axis_payload(p[["y_axis"]], sprintf("chart[[%d]] y_axis", i),
                                sheets, own, header_offset)

    ent$series <- lapply(seq_along(p[["series"]]), function(k) {
      q <- unclass(p[["series"]][[k]])
      where <- function(what) sprintf("chart[[%d]] series[[%d]] %s", i, k, what)
      v <- .resolve_chart_range(q[["values"]], where("values"), sheets, own,
                                header_offset)
      se <- list(values_sheet = v$sheet, values_range = v$range)
      cat_ <- .resolve_chart_range(q[["categories"]], where("categories"),
                                   sheets, own, header_offset)
      if (!is.null(cat_)) {
        se$categories_sheet <- cat_$sheet
        se$categories_range <- cat_$range
      }
      # Unnamed, a series is "Series 1" in the legend while the column it
      # plots already has a label in its header -- the name Excel itself uses.
      # Column specs only: an A1 range says nothing about whether the row
      # above it is a header.
      nm <- q[["name"]]
      if (is.null(nm)) nm <- .series_header_name(q[["values"]], header_offset)
      if (!is.null(nm) && !isTRUE(nm[["off"]])) {
        if (!is.null(nm[["text"]])) se$name <- nm[["text"]]
        else {
          r <- .resolve_chart_range(nm, where("name"), sheets, own,
                                    header_offset)
          se$name_sheet <- r$sheet
          se$name_range <- r$range
        }
      }
      se$marker       <- .marker_payload(q[["marker"]])
      se$labels       <- .labels_payload(q[["labels"]])
      se$trendline    <- .trendline_payload(q[["trendline"]])
      se$x_error_bars <- .error_bars_payload(q[["x_error_bars"]])
      se$y_error_bars <- .error_bars_payload(q[["y_error_bars"]])
      se$points       <- q[["points"]]
      if (isTRUE(q[["smooth"]]))             se$smooth <- 1L
      if (isTRUE(q[["invert_if_negative"]])) se$invert_if_negative <- 1L
      # a series is styled by its line and fill; see .chart_format_payload()
      if (!is.null(q[["format"]]))
        se$format <- .chart_format_payload(
          q[["format"]], where("format"),
          accept = c("line", "fill", "pattern"), part = "a chart series")
      se
    })
    ent
  })
}

# Normalise one series or a list of them.
.chart_series_list <- function(series, arg = "series") {
  if (is.null(series)) return(list())
  ss <- if (inherits(series, "xl_chart_series")) list(series) else series
  if (!is.list(ss))
    stop(sprintf("`%s` must be an xl_chart_series object or a list of them",
                 arg), call. = FALSE)
  for (i in seq_along(ss))
    if (!inherits(ss[[i]], "xl_chart_series"))
      stop(sprintf("`%s[[%d]]` must be an xl_chart_series object", arg, i),
           call. = FALSE)
  ss
}

# Normalise one chart or a list of them.
.chart_list <- function(chart, arg = "chart") {
  if (is.null(chart)) return(list())
  cs <- if (inherits(chart, "xl_chart")) list(chart) else chart
  if (!is.list(cs))
    stop(sprintf("`%s` must be an xl_chart object or a list of them", arg),
         call. = FALSE)
  for (i in seq_along(cs))
    if (!inherits(cs[[i]], "xl_chart"))
      stop(sprintf("`%s[[%d]]` must be an xl_chart object", arg, i),
           call. = FALSE)
  cs
}

# Charts: the constructors, the chart-type feature matrix, range specs, and
# what reaches the chart XML.

# ── Chart types ───────────────────────────────────────────────────────────────

test_that("every chart type libxlsxwriter offers is reachable", {
  # the enum runs 1..22 with no gaps, so a type added upstream shows up here as
  # a count mismatch rather than being quietly unreachable
  expect_equal(length(.LXW_CHART_TYPE), 22L)
  expect_equal(sort(unname(.LXW_CHART_TYPE)), 1:22)
  for (ty in names(.LXW_CHART_TYPE)) {
    # scatter needs categories; see the crash test below
    se <- if (identical(.CHART_FAMILY(ty), "scatter"))
      xl_chart_series(values = "B1:B5", categories = "A1:A5")
    else xl_chart_series(values = "A1:A5")
    expect_true(inherits(xl_chart(ty, se), "xl_chart"), info = ty)
  }
})

test_that("an unknown chart type is refused, listing the real ones", {
  expect_error(xl_chart("bubble", xl_chart_series(values = "A1:A5")),
               "`type` must be one of")
  expect_error(xl_chart("bubble", xl_chart_series(values = "A1:A5")),
               "doughnut")
})

test_that("every type maps to a family", {
  fams <- vapply(names(.LXW_CHART_TYPE), .CHART_FAMILY, character(1))
  expect_false(any(fams == "other"))
  expect_setequal(unique(fams),
                  c("area", "bar", "pie", "line", "scatter", "radar"))
})

# ── The feature matrix ────────────────────────────────────────────────────────

test_that("type-restricted features match what libxlsxwriter documents", {
  # these are documented in prose and not enforced: Excel drops them silently,
  # so the refusal has to be ours
  expect_true(.chart_supports("doughnut", "hole_size"))
  expect_false(.chart_supports("pie", "hole_size"))   # doughnut only, not both

  expect_true(.chart_supports("pie", "rotation"))
  expect_true(.chart_supports("doughnut", "rotation"))
  expect_false(.chart_supports("column", "rotation"))

  for (f in c("up_down_bars", "high_low_lines")) {
    expect_true(.chart_supports("line", f), label = f)
    expect_false(.chart_supports("column", f), label = f)
    expect_false(.chart_supports("pie", f), label = f)
  }

  for (f in c("series_gap", "series_overlap")) {
    expect_true(.chart_supports("bar", f), label = f)
    expect_true(.chart_supports("column_stacked", f), label = f)
    expect_false(.chart_supports("line", f), label = f)
  }

  expect_true(.chart_supports("line", "smooth"))
  expect_true(.chart_supports("scatter", "smooth"))
  expect_false(.chart_supports("column", "smooth"))

  # pie and doughnut have no axes at all
  expect_false(.chart_supports("pie", "axes"))
  expect_false(.chart_supports("doughnut", "axes"))
  expect_true(.chart_supports("column", "axes"))
  expect_true(.chart_supports("scatter", "axes"))
})

test_that("every feature is supported by at least one type, and denied by one", {
  # a feature nothing supports would be dead code; one everything supports
  # would not need to be in the matrix
  for (f in names(.CHART_FEATURE_FAMILIES)) {
    ok <- vapply(names(.LXW_CHART_TYPE), .chart_supports, logical(1), feature = f)
    expect_true(any(ok), label = f)
    expect_true(any(!ok), label = f)
  }
})

test_that("an unsupported feature is refused, naming the types that work", {
  expect_error(xl_chart("column", xl_chart_series(values = "A1", smooth = TRUE)),
               'does not apply to a "column" chart')
  expect_error(xl_chart("column", xl_chart_series(values = "A1", smooth = TRUE)),
               "It applies to: line")
  # and is accepted where it belongs
  expect_s3_class(xl_chart("line", xl_chart_series(values = "A1", smooth = TRUE)),
                  "xl_chart")
  expect_s3_class(xl_chart("scatter_smooth",
                           xl_chart_series(values = "B1:B5", smooth = TRUE,
                                           categories = "A1:A5")),
                  "xl_chart")
})

# ── Series ────────────────────────────────────────────────────────────────────

test_that("a series carries its ranges", {
  s <- unclass(xl_chart_series(values = "Data!B2:B10",
                               categories = "Data!A2:A10", name = "2024"))
  expect_equal(s$values$spec, "Data!B2:B10")
  expect_equal(s$categories$spec, "Data!A2:A10")
  expect_equal(s$name$text, "2024")
})

test_that("a range may name another sheet", {
  s <- unclass(xl_chart_series(values = list(sheet = "Data", cols = "revenue")))
  expect_equal(s$values$sheet, "Data")
  expect_equal(s$values$spec, list(cols = "revenue"))
  # without a sheet it resolves against the chart's own
  s2 <- unclass(xl_chart_series(values = list(cols = "revenue")))
  expect_null(s2$values$sheet)
})

test_that("a series name is literal text, and a range needs a spec", {
  # a string is never guessed at: a series may legitimately be called "Q1!",
  # so taking the name from a cell has to be asked for explicitly
  expect_equal(unclass(xl_chart_series(values = "A1", name = "2024"))$name$text,
               "2024")
  expect_equal(unclass(xl_chart_series(values = "A1",
                                       name = "Data!A1"))$name$text,
               "Data!A1")
  from_cell <- unclass(xl_chart_series(
    values = "A1", name = list(sheet = "Data", rows = 1, cols = 1)))
  expect_equal(from_cell$name$sheet, "Data")
  expect_null(from_cell$name$text)
})

test_that("series arguments are validated", {
  expect_error(xl_chart_series(), "must name the range")
  expect_error(xl_chart_series(values = 42), "must be a range string")
  expect_error(xl_chart_series(values = list(nope = "a")),
               "unknown `values` element\\(s\\): nope")
  expect_error(xl_chart_series(values = list(sheet = 1, cols = "a")),
               "must be a single sheet name")
  expect_error(xl_chart_series(values = list("a")), "list must be named")
  expect_error(xl_chart_series(values = "A1", format = "bold"),
               "must be an xl_format")
  expect_error(xl_chart_series(values = "A1", smooth = "yes"),
               "`smooth` must be")
})

# ── Charts ────────────────────────────────────────────────────────────────────

test_that("a chart carries its type, series and title", {
  ch <- unclass(xl_chart("column",
                         list(xl_chart_series(values = "A1:A5"),
                              xl_chart_series(values = "B1:B5")),
                         title = "Revenue"))
  expect_equal(ch$type, "column")
  expect_length(ch$series, 2L)
  expect_equal(ch$title$text, "Revenue")
})

test_that("title = FALSE removes the title Excel would generate", {
  # a single-series chart gets an automatic title, so switching it off has to
  # be expressible -- NULL means "leave Excel to it", which is different
  expect_true(isTRUE(unclass(xl_chart("pie", xl_chart_series(values = "A1"),
                                      title = FALSE))$title$off))
  expect_null(unclass(xl_chart("pie", xl_chart_series(values = "A1")))$title)
})

test_that("placement uses the same vocabulary as an image", {
  # lxw_chart_options is lxw_image_options minus url/tip/cell_format, so the
  # R names must not drift apart
  shared <- c("at", "scale", "offset", "position", "description", "decorative")
  expect_true(all(shared %in% names(formals(xl_chart))))
  expect_true(all(shared %in% names(formals(xl_image))))
  # and mean the same things
  ch <- unclass(xl_chart("column", xl_chart_series(values = "A1"),
                         at = "E2", scale = c(2, 3), offset = c(4, 5),
                         position = "dont_move_dont_size"))
  expect_equal(ch$at, "E2")
  expect_equal(ch$scale, c(2, 3))
  expect_equal(ch$offset, c(4, 5))
  expect_equal(ch$position, "dont_move_dont_size")
})

test_that("chart arguments are validated", {
  expect_error(xl_chart("column"), "`series` must give at least one")
  expect_error(xl_chart("column", list()), "needs at least one series")
  expect_error(xl_chart("column", "series"), "must be an xl_chart_series")
  expect_error(xl_chart("column", list(xl_chart_series(values = "A1"), "x")),
               "`series\\[\\[2\\]\\]` must be an xl_chart_series")
  expect_error(xl_chart("column", xl_chart_series(values = "A1"), style = 0),
               "`style` must be between 1 and 48")
  expect_error(xl_chart("column", xl_chart_series(values = "A1"), style = 49),
               "`style` must be between 1 and 48")
  expect_error(xl_chart("column", xl_chart_series(values = "A1"), scale = -1),
               "`scale` must be positive")
  expect_error(xl_chart("column", xl_chart_series(values = "A1"),
                        description = 1), "single non-NA string")
})

test_that("charts normalise from one or a list", {
  expect_length(.chart_list(xl_chart("pie", xl_chart_series(values = "A1"))), 1L)
  expect_length(.chart_list(list(xl_chart("pie", xl_chart_series(values = "A1")),
                                 xl_chart("bar", xl_chart_series(values = "A1")))),
                2L)
  expect_length(.chart_list(NULL), 0L)
  expect_error(.chart_list("chart"), "must be an xl_chart object")
})

test_that("the print methods run", {
  expect_output(print(xl_chart("column", xl_chart_series(values = "A1"))),
                "xl_chart")
  expect_output(print(xl_chart("column", xl_chart_series(values = "A1"),
                               title = "T")), "\"T\"")
  expect_output(print(xl_chart_series(values = "A1", name = "S")), "S")
})

# ── Range resolution across the workbook ──────────────────────────────────────

crange_sales <- data.frame(fruit = c("a", "b", "c"), qty = c(5, 150, 300),
                           stringsAsFactors = FALSE)
crange_other <- data.frame(x = 1:4, y = c(2, 4, 6, 8))

# Resolve a workbook's charts the way write_xlsx() does.
chart_plan <- function(wb, sheet = "Data") {
  dfs <- lapply(wb, function(e) if (inherits(e, "xl_sheet")) e$data else e)
  dfs <- lapply(dfs, writexl:::normalize_df)
  .resolve_charts(wb[[sheet]], dfs[[sheet]], .new_format_registry(), 1L,
                  xl_properties(), dfs, sheet)
}

test_that("a series resolves against the chart's own sheet by default", {
  p <- chart_plan(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"),
                    categories = list(cols = "fruit"))))))
  s <- p[[1L]]$series[[1L]]
  expect_equal(s$values_sheet, "Data")
  # column qty is 0-based col 1, data rows 1..3 are sheet rows 1..3
  expect_equal(s$values_range, c(1L, 1L, 3L, 1L))
  expect_equal(s$categories_range, c(1L, 0L, 3L, 0L))
})

test_that("a series may plot another sheet, in either spelling", {
  wb <- function(v) list(Data = xl_sheet(crange_sales,
                                         chart = xl_chart("line",
                                           xl_chart_series(values = v))),
                         Other = crange_other)
  a <- chart_plan(wb("Other!B2:B5"))[[1L]]$series[[1L]]
  b <- chart_plan(wb(list(sheet = "Other", cols = "y")))[[1L]]$series[[1L]]
  expect_equal(a$values_sheet, "Other")
  expect_equal(b$values_sheet, "Other")
  # the two spellings must resolve to the same block
  expect_equal(a$values_range, b$values_range)
})

test_that("an unknown sheet is refused, listing the real ones", {
  expect_error(chart_plan(list(Data = xl_sheet(crange_sales,
    chart = xl_chart("column",
      xl_chart_series(values = list(sheet = "Nope", cols = "qty")))))),
    'names sheet "Nope", which is not in the workbook')
  expect_error(chart_plan(list(Data = xl_sheet(crange_sales,
    chart = xl_chart("column",
      xl_chart_series(values = list(sheet = "Nope", cols = "qty")))))),
    '"Data"')
})

test_that("a range holding no data is refused, not plotted empty", {
  # Excel shows an empty chart with no complaint, so the emptiness has to be
  # caught here or it is delivered as a silently blank chart
  expect_error(chart_plan(list(Data = xl_sheet(crange_sales,
    chart = xl_chart("column", xl_chart_series(values = "A99:A200"))))),
    "selects no data")
  expect_error(chart_plan(list(Data = xl_sheet(crange_sales[0, ],
    chart = xl_chart("column", xl_chart_series(values = "A2:A3"))))),
    "selects no data")
})

test_that("a title or series name may come from a cell", {
  p <- chart_plan(list(Data = xl_sheet(crange_sales, chart = xl_chart("pie",
    xl_chart_series(values = list(cols = "qty"),
                    name = list(rows = 1, cols = "fruit")),
    title = list(rows = 1, cols = "fruit")))))
  expect_equal(p[[1L]]$title_range, c(1L, 0L, 1L, 0L))
  expect_equal(p[[1L]]$series[[1L]]$name_range, c(1L, 0L, 1L, 0L))
  # a string title stays a string
  q <- chart_plan(list(Data = xl_sheet(crange_sales, chart = xl_chart("pie",
    xl_chart_series(values = list(cols = "qty")), title = "Share"))))
  expect_equal(q[[1L]]$title, "Share")
  expect_null(q[[1L]]$title_range)
})

test_that("a chart is anchored to one cell", {
  expect_error(chart_plan(list(Data = xl_sheet(crange_sales,
    chart = xl_chart("column", xl_chart_series(values = list(cols = "qty")),
                     at = "A1:B2")))),
    "must name a single cell")
})

test_that("a chart is a victim of the drawing-id desync, like an image", {
  # charts produce a drawing so they never cause the desync, but they are
  # numbered from the same counter and so are subject to it.  Established with
  # a pure-C reprex before charts existed here; nothing else would notice.
  logo <- tempfile(fileext = ".png")
  writeBin(as.raw(c(
    0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D,
    0x49, 0x48, 0x44, 0x52, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01,
    0x08, 0x06, 0x00, 0x00, 0x00, 0x1F, 0x15, 0xC4, 0x89, 0x00, 0x00, 0x00,
    0x0A, 0x49, 0x44, 0x41, 0x54, 0x78, 0x9C, 0x63, 0x00, 0x01, 0x00, 0x00,
    0x05, 0x00, 0x01, 0x0D, 0x0A, 0x2D, 0xB4, 0x00, 0x00, 0x00, 0x00, 0x49,
    0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82)), logo)
  hdr <- xl_sheet(crange_sales,
                  page = xl_page_setup(header = "&L&G",
                                       header_image = list(left = logo)))
  charted <- xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"))))
  expect_error(write_tmp(list(H = hdr, C = charted)),
               'sheet "C" has a chart')
  # the other order is fine, and is what the message recommends
  expect_silent(write_tmp(list(C = charted, H = hdr)))
})

# ── What reaches the file ─────────────────────────────────────────────────────

cfile <- function(..., df = crange_sales) {
  p <- write_tmp(list(Data = xl_sheet(df, ...)))
  d <- tempfile(); dir.create(d); utils::unzip(p, exdir = d)
  ch <- list.files(file.path(d, "xl/charts"), pattern = "^chart",
                   full.names = TRUE)
  list(files = list.files(d, recursive = TRUE),
       chart = if (length(ch)) paste(readLines(ch[1L], warn = FALSE),
                                     collapse = "") else "",
       sheet = xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE))
}

test_that("a chart reaches the file with its drawing and reference", {
  r <- cfile(chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"),
                    categories = list(cols = "fruit"), name = "Qty"),
    title = "Fruit", at = "D2"))
  expect_true(any(grepl("^xl/charts/chart", r$files)))
  expect_true(any(grepl("^xl/drawings/drawing", r$files)))
  expect_match(r$sheet, "<drawing r:id=", fixed = TRUE)
  expect_match(r$chart, "<c:barChart>", fixed = TRUE)
  # the series points at the resolved range, absolute and sheet-qualified
  expect_match(r$chart, "<c:f>Data!$B$2:$B$4</c:f>", fixed = TRUE)
  expect_match(r$chart, "<c:f>Data!$A$2:$A$4</c:f>", fixed = TRUE)
  expect_match(r$chart, "Fruit", fixed = TRUE)
  expect_match(r$chart, "Qty", fixed = TRUE)
})

test_that("each chart type writes its own plot element", {
  kinds <- c(column = "barChart", bar = "barChart", line = "lineChart",
             pie = "pieChart", doughnut = "doughnutChart", area = "areaChart",
             scatter = "scatterChart", radar = "radarChart")
  for (ty in names(kinds)) {
    se <- if (identical(.CHART_FAMILY(ty), "scatter"))
      xl_chart_series(values = list(cols = "qty"),
                      categories = list(cols = "fruit"))
    else xl_chart_series(values = list(cols = "qty"))
    expect_match(cfile(chart = xl_chart(ty, se))$chart,
                 sprintf("<c:%s>", kinds[[ty]]), fixed = TRUE, label = ty)
  }
})

test_that("a series format reaches the chart's shape properties", {
  r <- cfile(chart = xl_chart("line", xl_chart_series(
    values = list(cols = "qty"),
    format = xl_border(all = "dashed", color = "red") +
             xl_fill(background = "yellow", transparency = 30))))
  # charts use 6-digit RGB, not the 8-digit ARGB of cell styles
  expect_match(r$chart, "<a:srgbClr val=\"FF0000\"/>", fixed = TRUE)
  expect_match(r$chart, "<a:srgbClr val=\"FFFF00\">", fixed = TRUE)
  expect_match(r$chart, "<a:prstDash val=\"dash\"/>", fixed = TRUE)
  # transparency 30 becomes 70% opacity
  expect_match(r$chart, "<a:alpha val=\"70000\"/>", fixed = TRUE)
})

test_that("a cross-sheet series names the other sheet", {
  p <- write_tmp(list(
    Chart = xl_sheet(data.frame(z = 1), chart = xl_chart("pie",
      xl_chart_series(values = list(sheet = "Data", cols = "qty")))),
    Data = crange_sales))
  d <- tempfile(); dir.create(d); utils::unzip(p, exdir = d)
  x <- paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                                  full.names = TRUE)[1L], warn = FALSE),
             collapse = "")
  expect_match(x, "Data!", fixed = TRUE)
})

test_that("several charts on one sheet share its drawing", {
  r <- cfile(chart = list(
    xl_chart("column", xl_chart_series(values = list(cols = "qty")), at = "D2"),
    xl_chart("pie", xl_chart_series(values = list(cols = "qty")), at = "D20")))
  expect_equal(sum(grepl("^xl/charts/chart", r$files)), 2L)
  # one drawing per sheet, two charts anchored in it
  expect_equal(sum(grepl("^xl/drawings/drawing", r$files)), 1L)
})

test_that("placement and style options are written", {
  r <- cfile(chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty")),
    at = "A1", offset = c(20, 10), style = 12))
  expect_match(r$chart, "<c:style val=\"12\"/>", fixed = TRUE)
  d <- tempfile(); dir.create(d)
  p <- write_tmp(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty")), at = "A1",
    offset = c(20, 10)))))
  utils::unzip(p, exdir = d)
  dr <- paste(readLines(file.path(d, "xl/drawings/drawing1.xml"), warn = FALSE),
              collapse = "")
  expect_match(dr, sprintf("<xdr:colOff>%d</xdr:colOff>", 20 * 9525),
               fixed = TRUE)
})

test_that("title = FALSE writes autoTitleDeleted", {
  expect_match(cfile(chart = xl_chart("pie",
    xl_chart_series(values = list(cols = "qty")), title = FALSE))$chart,
    "<c:autoTitleDeleted val=\"1\"/>", fixed = TRUE)
})

test_that("every relationship a chart adds resolves", {
  p <- write_tmp(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"))))))
  d <- tempfile(); dir.create(d); utils::unzip(p, exdir = d)
  bad <- character(0)
  for (rel in list.files(d, recursive = TRUE, pattern = "_rels/")) {
    x <- paste(readLines(file.path(d, rel), warn = FALSE), collapse = "")
    base <- dirname(dirname(rel))
    for (tg in regmatches(x, gregexpr('Target="[^"]+"[^>]*', x))[[1L]]) {
      if (grepl('TargetMode="External"', tg)) next
      target <- sub('Target="([^"]+)".*', "\\1", tg)
      q <- if (startsWith(target, "/")) sub("^/", "", target)
           else if (base == ".") target else file.path(base, target)
      while (grepl("/[^/]+/\\.\\./", q)) q <- sub("/[^/]+/\\.\\./", "/", q)
      if (!file.exists(file.path(d, q))) bad <- c(bad, target)
    }
  }
  expect_equal(bad, character(0))
})

test_that("a scatter series must have categories", {
  # not a style preference: a scatter series without categories segfaults in
  # libxlsxwriter's _chart_write_x_val(), which has no NULL guard on the
  # formula.
  for (ty in c("scatter", "scatter_straight", "scatter_straight_markers",
               "scatter_smooth", "scatter_smooth_markers"))
    expect_error(xl_chart(ty, xl_chart_series(values = "B1:B5")),
                 "no `categories`", label = ty)
  # with them, every scatter type writes
  for (ty in c("scatter", "scatter_straight", "scatter_smooth"))
    expect_true(inherits(xl_chart(ty, xl_chart_series(values = "B1:B5",
                                                      categories = "A1:A5")),
                         "xl_chart"), info = ty)
  # and no other family is affected
  expect_s3_class(xl_chart("column", xl_chart_series(values = "A1:A5")),
                  "xl_chart")
})

test_that("every chart type writes a file without crashing", {
  # at the C layer a bad assumption is a segfault rather than a failed
  # expectation, so every type is exercised end to end
  for (ty in names(.LXW_CHART_TYPE)) {
    se <- if (identical(.CHART_FAMILY(ty), "scatter"))
      xl_chart_series(values = list(cols = "qty"),
                      categories = list(cols = "fruit"))
    else xl_chart_series(values = list(cols = "qty"))
    expect_silent(write_tmp(list(Data = xl_sheet(crange_sales,
                                                 chart = xl_chart(ty, se)))))
  }
})

# ── The chart title's font ────────────────────────────────────────────────────

test_that("a title format styles the title, and reaches the file", {
  r <- cfile(chart = xl_chart("column", xl_chart_series(values = list(cols = "qty")),
                              title = "Sales",
                              title_format = xl_font(size = 14, bold = TRUE,
                                                     color = "red")))
  expect_match(r$chart, "<a:defRPr", fixed = TRUE)
  # lxw_chart_font.size is in points and written in hundredths
  expect_match(r$chart, 'sz="1400"', fixed = TRUE)
  expect_match(r$chart, 'b="1"', fixed = TRUE)
  expect_match(r$chart, "FF0000", fixed = TRUE)
})

test_that("a title format needs a title to attach to", {
  # with no title libxlsxwriter writes no <c:title> element at all, so
  # chart_title_set_name_font() has nothing to style.  The automatic title of
  # a single-series chart is Excel's own and never reaches the file.
  expect_error(xl_chart("column", xl_chart_series(values = "A1:A5"),
                        title_format = xl_font(italic = TRUE)),
               "give a `title` too", fixed = TRUE)
  r <- cfile(chart = xl_chart("column",
                              xl_chart_series(values = list(cols = "qty")),
                              title_format = xl_font(italic = TRUE),
                              title = "Sales"))
  expect_match(r$chart, "<c:title>", fixed = TRUE)
  expect_match(r$chart, 'i="1"', fixed = TRUE)
})

test_that("styling a title that has been removed is refused", {
  expect_error(xl_chart("column", xl_chart_series(values = "A1:A5"),
                        title = FALSE, title_format = xl_font(bold = TRUE)),
               "`title = FALSE` removes it", fixed = TRUE)
})

test_that("a title format may not set a fill or a border", {
  expect_error(xl_chart("column", xl_chart_series(values = "A1:A5"),
                        title = "T", title_format = xl_fill(background = "red")),
               "takes no fill")
  expect_error(xl_chart("column", xl_chart_series(values = "A1:A5"),
                        title = "T", title_format = xl_border(all = "thin")),
               "takes no line")
})

test_that("a series format may not set a font", {
  # refused by the constructor, not at write time
  expect_error(xl_chart_series(values = list(cols = "qty"),
                               format = xl_font(bold = TRUE)),
               "a shape and has no text")
})

# ── The header cell ───────────────────────────────────────────────────────────

test_that("a name or title may come from a column's header cell", {
  # the header is where a series name usually lives, and `rows` cannot reach it
  # -- rows count data rows, so row 1 is the first row *under* the header
  r <- cfile(chart = xl_chart("column",
                              xl_chart_series(values = list(cols = "qty"),
                                              name = list(header = "qty")),
                              title = list(header = "fruit")))
  expect_match(r$chart, "<c:f>Data!$B$1</c:f>", fixed = TRUE)
  expect_match(r$chart, "<c:f>Data!$A$1</c:f>", fixed = TRUE)
})

test_that("a header cell may name its column by position, and another sheet", {
  p <- chart_plan(list(
    Chart = xl_sheet(data.frame(z = 1),
                     chart = xl_chart("pie",
                       xl_chart_series(values = list(sheet = "Data",
                                                     cols = "qty"),
                                       name = list(sheet = "Data",
                                                   header = 2)))),
    Data = crange_sales), sheet = "Chart")
  nm <- p[[1L]]$series[[1L]]
  expect_equal(nm$name_sheet, "Data")
  # 0-based: the header row is 0, and column 2 is index 1
  expect_equal(nm$name_range, c(0L, 1L, 0L, 1L))
})

test_that("a header spec stands alone, and needs a header row to exist", {
  expect_error(xl_chart_series(values = "A1:A5",
                               name = list(header = "qty", rows = 1)),
               "stands alone")
  expect_error(xl_chart_series(values = "A1:A5",
                               name = list(header = c("a", "b"))),
               "must name a single column")
  # written without headers there is no such cell, and saying so beats
  # silently plotting the first data row as the name
  expect_error(
    write_tmp(list(Data = xl_sheet(crange_sales,
                                   chart = xl_chart("column",
                                     xl_chart_series(values = list(cols = "qty"),
                                                     name = list(header = "qty"))))),
              col_names = FALSE),
    "written without headers")
})

# ── The default series name ───────────────────────────────────────────────────

test_that("a series plotting a column is named after that column's header", {
  # Excel names a series this way when you chart a column with its header, and
  # the alternative is a legend reading "Series 1"
  p <- chart_plan(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"))))))
  s <- p[[1L]]$series[[1L]]
  expect_equal(s$name_sheet, "Data")
  expect_equal(s$name_range, c(0L, 1L, 0L, 1L))   # header row, column qty
  expect_null(s$name)
  # and it reaches the file as a reference, not as copied text
  expect_match(cfile(chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"))))$chart,
    "<c:f>Data!$B$1</c:f>", fixed = TRUE)
})

test_that("an explicit name still wins, in any spelling", {
  nm <- function(...) {
    p <- chart_plan(list(Data = xl_sheet(crange_sales,
                                         chart = xl_chart("column",
                                           xl_chart_series(...)))))
    p[[1L]]$series[[1L]]
  }
  expect_equal(nm(values = list(cols = "qty"), name = "Quantity")$name,
               "Quantity")
  # another column's header
  expect_equal(nm(values = list(cols = "qty"),
                  name = list(header = "fruit"))$name_range,
               c(0L, 0L, 0L, 0L))
  # a data cell
  expect_equal(nm(values = list(cols = "qty"),
                  name = list(rows = 1, cols = "fruit"))$name_range,
               c(1L, 0L, 1L, 0L))
})

test_that("name = FALSE leaves the series unnamed", {
  p <- chart_plan(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = "qty"), name = FALSE)))))
  s <- p[[1L]]$series[[1L]]
  expect_null(s$name)
  expect_null(s$name_range)
})

test_that("the default only applies where a header is known to exist", {
  no_name <- function(wb, ...) {
    p <- chart_plan(wb, ...)
    s <- p[[1L]]$series[[1L]]
    expect_null(s$name)
    expect_null(s$name_range)
  }
  # an A1 range says nothing about whether the row above it is a header
  no_name(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = "Data!B2:B4")))))
  # neither does a span of several columns
  no_name(list(Data = xl_sheet(crange_sales, chart = xl_chart("column",
    xl_chart_series(values = list(cols = c("fruit", "qty")))))))
  # and a workbook written without headers has no header cell at all
  s <- .resolve_charts(
    xl_sheet(crange_sales,
             chart = xl_chart("column",
                              xl_chart_series(values = list(cols = "qty")))),
    writexl:::normalize_df(crange_sales), .new_format_registry(), 0L,
    xl_properties(), list(Data = writexl:::normalize_df(crange_sales)), "Data")
  expect_null(s[[1L]]$series[[1L]]$name)
  expect_null(s[[1L]]$series[[1L]]$name_range)
})

test_that("the default follows the sheet the values come from", {
  p <- chart_plan(list(
    Chart = xl_sheet(data.frame(z = 1),
                     chart = xl_chart("pie",
                       xl_chart_series(values = list(sheet = "Data",
                                                     cols = "qty")))),
    Data = crange_sales), sheet = "Chart")
  expect_equal(p[[1L]]$series[[1L]]$name_sheet, "Data")
})

# ── The options that had no test of their own ────────────────────────────────

chart_file <- function(...) {
  df <- data.frame(quarter = c("Q1", "Q2", "Q3"), revenue = c(10, 25, 18),
                   stringsAsFactors = FALSE)
  t <- tempfile(fileext = ".xlsx")
  write_xlsx(list(Data = xl_sheet(df, chart = xl_chart(...))), t)
  xlsx_part(t, "xl/charts/chart1.xml", raw = TRUE)
}

test_that("alt text and a smoothed, inverted series reach the file", {
  s <- xl_chart_series(values = list(cols = "revenue"), smooth = TRUE,
                       invert_if_negative = TRUE)
  x <- chart_file("line", s, description = "Quarterly revenue",
                  decorative = TRUE)
  expect_match(x, "<c:smooth val=\"1\"/>", fixed = TRUE)
  expect_match(x, "invertIfNegative", fixed = TRUE)
})

test_that("a chart's alt text reaches the drawing that holds it", {
  drawing <- function(...) {
    df <- data.frame(quarter = c("Q1", "Q2"), revenue = c(1, 2),
                     stringsAsFactors = FALSE)
    t <- tempfile(fileext = ".xlsx")
    write_xlsx(list(Data = xl_sheet(df, chart = xl_chart("line",
      xl_chart_series(values = list(cols = "revenue")), ...))), t)
    # alt text belongs to the drawing anchor, not the chart part
    xlsx_part(t, "xl/drawings/drawing1.xml", raw = TRUE)
  }
  expect_match(drawing(description = "Quarterly revenue"),
               'descr="Quarterly revenue"', fixed = TRUE)
  # marking a chart decorative replaces the description rather than joining it:
  # Excel's model is that a decorative object has nothing to announce
  d <- drawing(description = "Quarterly revenue", decorative = TRUE)
  expect_match(d, "adec:decorative", fixed = TRUE)
  expect_false(grepl("Quarterly revenue", d, fixed = TRUE))
})

test_that("a numeric axis crossing and styled gridlines reach the file", {
  x <- chart_file("column", xl_chart_series(values = list(cols = "revenue")),
                  y_axis = xl_chart_axis(
                    crossing = 5, major_gridlines = TRUE,
                    major_gridlines_format = xl_border(all = "dashed",
                                                       color = "red")))
  expect_match(x, "<c:crossesAt val=\"5\"/>", fixed = TRUE)
  expect_match(x, "majorGridlines", fixed = TRUE)
  expect_match(x, "FF0000", fixed = TRUE)
  # and the two named crossings take the other branches
  expect_match(chart_file("column", xl_chart_series(values = list(cols = "revenue")),
                          y_axis = xl_chart_axis(crossing = "min")),
               "<c:crosses val=\"min\"/>", fixed = TRUE)
  expect_match(chart_file("column", xl_chart_series(values = list(cols = "revenue")),
                          y_axis = xl_chart_axis(crossing = "max")),
               "<c:crosses val=\"max\"/>", fixed = TRUE)
})

test_that("a chart area fill and a plot area fill are told apart", {
  x <- chart_file("column", xl_chart_series(values = list(cols = "revenue")),
                  chart_area_format = xl_fill(background = "yellow"),
                  plot_area_format = xl_fill(background = "#C0C0C0"))
  expect_match(x, "FFFF00", fixed = TRUE)
  expect_match(x, "C0C0C0", fixed = TRUE)
})

test_that("chart arguments name what was wrong", {
  s <- xl_chart_series(values = list(cols = "revenue"))
  expect_error(xl_chart(NA, s), "must name a chart type")
  expect_error(xl_chart("bubble", s), "`type` must be one of")
  expect_error(xl_chart("column", s, at = "A1", scale = c(1, 2, 3)),
               "must be one number or two")
  expect_error(xl_chart("column", s, description = 1),
               "must be a single non-NA string")
  expect_error(xl_chart("column", list(s, "not a series")),
               "must be an xl_chart_series")
  expect_equal(.chart_list(NULL), list())
  expect_error(.chart_list(list(xl_chart("column", s), "x")),
               "must be an xl_chart object")
})

test_that("a series may not plot a sheet that has no data rows", {
  empty <- data.frame(a = numeric(0))
  plot_empty <- function(values)
    write_tmp(list(Empty = empty,
                   D = xl_sheet(data.frame(revenue = c(1, 2)),
                                chart = xl_chart("column",
                                  xl_chart_series(values = values)))))
  # by column: the spec resolver has no rows to turn into a range
  expect_error(plot_empty(list(sheet = "Empty", cols = "a")), "covers no rows")
  # by A1 range: the range itself is fine, and the sheet is what is empty
  expect_error(plot_empty("Empty!A1:A1"), "which has no data rows")
  # and a range starting past the data says so instead
  expect_error(plot_empty("Empty!A2:A3"), "selects no data")
})

test_that("every chart type has a family, so the fallback is unreachable", {
  # the feature matrix is keyed by family, so a type falling through to
  # "other" would silently be allowed everything and refused nothing
  fams <- vapply(names(.LXW_CHART_TYPE), .CHART_FAMILY, character(1))
  expect_setequal(unique(fams),
                  c("pie", "bar", "line", "area", "scatter", "radar"))
  expect_false("other" %in% fams)
})

test_that("a chart with no series to plot resolves to nothing", {
  expect_equal(.chart_series_list(NULL), list())
  expect_error(.chart_series_list("nope"),
               "must be an xl_chart_series object or a list of them")
})

# =============================================================================
# The parts of a chart series: markers, data labels, trendlines, error bars
# and individual points
# =============================================================================
#
# Five constructors over the chart_series_*() functions.  Each takes a
# `format` translated by .chart_format_payload() into whichever of line / fill
# / pattern / font that part can use: a trendline is a line and takes nothing
# else, a data label is a shape with text and takes all four.
#
# Four restrictions are documented in chart.h and enforced nowhere:
#
#   * the positions a data label may take depend on the chart type, and the
#     header carries the whole table (Line/Scatter, Bar/Column, Pie/Doughnut,
#     Area/Radar);
#   * a moving-average trendline has no forecast, equation or R-squared;
#   * an intercept applies only to exponential, linear and polynomial
#     trendlines;
#   * an automatic marker cannot also be given a size, a line or a fill.
#
# Excel drops all of them without a word, so they are refused here.
# -----------------------------------------------------------------------------

.LXW_MARKER_TYPE <- c(
  automatic = 0L, none = 1L, square = 2L, diamond = 3L, triangle = 4L,
  x = 5L, star = 6L, short_dash = 7L, long_dash = 8L, circle = 9L, plus = 10L
)

.LXW_LABEL_POSITION <- c(
  default = 0L, center = 1L, right = 2L, left = 3L, above = 4L, below = 5L,
  inside_base = 6L, inside_end = 7L, outside_end = 8L, best_fit = 9L
)

# Which families accept each position, from the table in chart.h.  "default"
# means "leave it to Excel" and is legal everywhere.
.LABEL_POSITION_FAMILIES <- list(
  default     = c("line", "scatter", "bar", "pie", "area", "radar", "other"),
  center      = c("line", "scatter", "bar", "pie", "area", "radar", "other"),
  right       = c("line", "scatter"),
  left        = c("line", "scatter"),
  above       = c("line", "scatter"),
  below       = c("line", "scatter"),
  inside_base = c("bar", "other"),
  inside_end  = c("bar", "pie", "other"),
  outside_end = c("bar", "pie", "other"),
  best_fit    = "pie"
)

.LXW_LABEL_SEPARATOR <- c(comma = 0L, semicolon = 1L, period = 2L,
                          newline = 3L, space = 4L)

.LXW_TRENDLINE_TYPE <- c(linear = 0L, log = 1L, poly = 2L, power = 3L,
                         exp = 4L, average = 5L)

.LXW_ERROR_BAR_TYPE <- c(std_error = 0L, fixed = 1L, percentage = 2L,
                         std_dev = 3L)

.LXW_ERROR_BAR_DIRECTION <- c(both = 0L, plus = 1L, minus = 2L)

# --- Markers ------------------------------------------------------------------

#' A marker on a chart series
#'
#' @description
#' `xl_chart_marker()` draws a symbol at each point of a series.  Line, scatter
#' and radar charts are where they show; on other types Excel ignores them.
#'
#' `type = "automatic"` asks Excel for the default marker of that series, and
#' is the one type that cannot be given a size or a format --- libxlsxwriter
#' documents that, and Excel drops them, so both are refused here.
#'
#' @param type The symbol: `"automatic"`, `"none"`, `"square"`, `"diamond"`,
#'   `"triangle"`, `"x"`, `"star"`, `"short_dash"`, `"long_dash"`, `"circle"`
#'   or `"plus"`.
#' @param size The symbol's size in points, 2 to 72.
#' @param format An [xl_format()] styling the symbol: [xl_border()] for its
#'   outline, [xl_fill()] for its fill or pattern.
#' @return An `xl_chart_marker` object.
#' @family images and charts
#' @seealso [xl_chart_series]
#' @export
#' @examples
#' xl_chart_marker(type = "circle", size = 8)
#' xl_chart_marker(type = "none")
xl_chart_marker <- function(type = NULL, size = NA, format = NULL) {
  ty <- .val_enum(type, names(.LXW_MARKER_TYPE), "type")
  sz <- .val_int(size, "size", min = 2, max = 72)
  fmt <- .chart_format_payload(format, "format",
                               accept = c("line", "fill", "pattern"),
                               part = "a chart marker")
  if (identical(ty, "automatic") && (!is.null(sz) || !is.null(fmt)))
    stop(paste0("`type = \"automatic\"` asks Excel for the series' own default ",
                "marker, which cannot then be given a size or a format. Name ",
                "the symbol you want instead."), call. = FALSE)
  structure(.drop_null(list(type = ty, size = sz, format = fmt)),
            class = "xl_chart_marker")
}

# --- Data labels --------------------------------------------------------------

#' Data labels on a chart series
#'
#' @description
#' `xl_chart_labels()` prints the numbers next to the points that carry them.
#' With no arguments it shows the value of each point, which is Excel's own
#' default.
#'
#' `xl_chart_label()` describes one label, for `custom`: a label can be given
#' its own text or hidden, one point at a time.
#'
#' Excel allows different label positions for different chart types, and drops
#' one that does not apply.  The table is in libxlsxwriter's header and is
#' enforced here: `"center"` is allowed everywhere, `"right"`, `"left"`,
#' `"above"` and `"below"` on line and scatter charts, `"inside_base"` on bar
#' and column, `"inside_end"` and `"outside_end"` on bar, column, pie and
#' doughnut, and `"best_fit"` on pie and doughnut.
#'
#' @param show_value,show_name,show_category,show_percentage What each label
#'   holds: the point's value, the series name, the category, and the value as
#'   a percentage of the series.  Naming any of them means the label holds
#'   exactly those, so `show_percentage = TRUE` alone gives a percentage and
#'   nothing else; naming none of them leaves Excel's default, the value.
#'   Percentages are meaningful on pie and doughnut charts.
#' @param show_legend_key Print the series' legend swatch in each label.
#' @param num_format A number format for the labels, as an Excel format string
#'   or an [xl_num_format()].
#' @param position Where the label sits relative to its point; see the
#'   description for which chart types allow which.
#' @param separator What joins the parts of a label when it holds more than
#'   one: `"comma"`, `"semicolon"`, `"period"`, `"newline"` or `"space"`.
#' @param format An [xl_format()] styling the labels.  A label is a shape with
#'   text in it, so all of [xl_font()], [xl_border()] and [xl_fill()] apply.
#' @param leader_lines Draw a line from a label back to its point.  Excel only
#'   shows one once the label has been dragged away from the point.
#' @param custom A list of [xl_chart_label()]s, one per point in order, giving
#'   individual labels their own text, styling, or `hide = TRUE`.  `NULL` in
#'   the list leaves that point's label alone.
#' @return An `xl_chart_labels` object.
#' @family images and charts
#' @seealso [xl_chart_series]
#' @export
#' @examples
#' xl_chart_labels()
#' xl_chart_labels(show_category = TRUE, show_percentage = TRUE,
#'                 separator = "newline", position = "outside_end")
xl_chart_labels <- function(show_value = NA, show_name = NA,
                            show_category = NA, show_percentage = NA,
                            show_legend_key = NA, num_format = NULL,
                            position = NULL, separator = NULL,
                            format = NULL, leader_lines = NA, custom = NULL) {
  cu <- custom
  if (!is.null(cu)) {
    if (inherits(cu, "xl_chart_label")) cu <- list(cu)
    if (!is.list(cu))
      stop("`custom` must be a list of xl_chart_label() objects", call. = FALSE)
    for (i in seq_along(cu))
      if (!is.null(cu[[i]]) && !inherits(cu[[i]], "xl_chart_label"))
        stop(sprintf("`custom[[%d]]` must be an xl_chart_label() or NULL", i),
             call. = FALSE)
  }
  structure(.drop_null(list(
    show_value = .val_flag(show_value, "show_value"),
    show_name = .val_flag(show_name, "show_name"),
    show_category = .val_flag(show_category, "show_category"),
    show_percentage = .val_flag(show_percentage, "show_percentage"),
    show_legend_key = .val_flag(show_legend_key, "show_legend_key"),
    num_format = .chart_num_format(num_format, "num_format"),
    position = .val_enum(position, names(.LXW_LABEL_POSITION), "position"),
    separator = .val_enum(separator, names(.LXW_LABEL_SEPARATOR), "separator"),
    format = .chart_format_payload(format, "format", part = "a data label"),
    leader_lines = .val_flag(leader_lines, "leader_lines"),
    custom = cu
  )), class = "xl_chart_labels")
}

#' @param value The label's text.  A string beginning with `"="` is a formula,
#'   so `"=Sheet1!$A$1"` takes the text from a cell.
#' @param hide Remove this point's label, leaving the others.
#' @rdname xl_chart_labels
#' @export
xl_chart_label <- function(value = NULL, hide = NA, format = NULL) {
  hd <- .val_flag(hide, "hide")
  val <- .val_str(value, "value")
  if (isTRUE(hd) && (!is.null(val) || !is.null(format)))
    stop(paste0("`hide = TRUE` removes the label, so it cannot also carry a ",
                "value or a format"), call. = FALSE)
  structure(.drop_null(list(
    value = val, hide = if (isTRUE(hd)) 1L else NULL,
    format = .chart_format_payload(format, "format", part = "a data label")
  )), class = "xl_chart_label")
}

# --- Trendlines ---------------------------------------------------------------

#' A trendline on a chart series
#'
#' @description
#' `xl_chart_trendline()` fits a line through a series.
#'
#' Two of Excel's own restrictions are enforced, because it discards these
#' rather than complain: a **moving average** has no forecast, no equation and
#' no R-squared, and an **intercept** applies only to exponential, linear and
#' polynomial fits.
#'
#' @param type `"linear"`, `"log"`, `"poly"`, `"power"`, `"exp"` or
#'   `"average"` for a moving average.
#' @param order The order of a polynomial fit, 2 or more.  `"poly"` only.
#' @param period The number of points a moving average covers, 2 or more.
#'   `"average"` only.
#' @param forward,backward How far to project the line beyond the data, in
#'   categories.
#' @param intercept Force the line through this value on the y axis.
#'   Exponential, linear and polynomial fits only.
#' @param equation Print the fitted equation on the chart.
#' @param r_squared Print the R-squared value on the chart.
#' @param name The trendline's name in the legend.  Excel generates one
#'   otherwise.
#' @param format An [xl_format()] styling the line --- [xl_border()] only,
#'   since a trendline is a line.
#' @return An `xl_chart_trendline` object.
#' @family images and charts
#' @seealso [xl_chart_series]
#' @export
#' @examples
#' xl_chart_trendline("linear", equation = TRUE, r_squared = TRUE)
#' xl_chart_trendline("poly", order = 3)
#' xl_chart_trendline("average", period = 2)
xl_chart_trendline <- function(type, order = NA, period = NA, forward = NA,
                               backward = NA, intercept = NA, equation = NA,
                               r_squared = NA, name = NULL, format = NULL) {
  if (missing(type))
    stop("`type` must name the fit, e.g. \"linear\"", call. = FALSE)
  ty <- .val_enum(type, names(.LXW_TRENDLINE_TYPE), "type")
  if (is.null(ty))
    stop("`type` must name the fit, e.g. \"linear\"", call. = FALSE)
  or <- .val_int(order, "order", min = 2)
  pe <- .val_int(period, "period", min = 2)
  if (!is.null(or) && !identical(ty, "poly"))
    stop(sprintf("`order` is the degree of a polynomial fit, not of a \"%s\" one",
                 ty), call. = FALSE)
  if (!is.null(pe) && !identical(ty, "average"))
    stop(sprintf(paste0("`period` is how many points a moving average covers; ",
                        "a \"%s\" fit has none"), ty), call. = FALSE)
  if (identical(ty, "poly") && is.null(or))
    stop("a \"poly\" trendline needs an `order`", call. = FALSE)
  if (identical(ty, "average") && is.null(pe))
    stop("an \"average\" trendline needs a `period`", call. = FALSE)

  fw <- .val_num(forward, "forward", min = 0)
  bw <- .val_num(backward, "backward", min = 0)
  ic <- .val_num(intercept, "intercept")
  eq <- .val_flag(equation, "equation")
  rs <- .val_flag(r_squared, "r_squared")
  if (identical(ty, "average")) {
    given <- c("forward", "backward", "equation", "r_squared")[
      c(!is.null(fw), !is.null(bw), isTRUE(eq), isTRUE(rs))]
    if (length(given))
      stop(sprintf(paste0("`%s` does not apply to a moving average; Excel ",
                          "offers a forecast, an equation and an R-squared ",
                          "for the fitted types only."), given[[1L]]),
           call. = FALSE)
  }
  if (!is.null(ic) && !ty %in% c("exp", "linear", "poly"))
    stop(sprintf(paste0("`intercept` applies to an exponential, linear or ",
                        "polynomial fit, not to a \"%s\" one."), ty),
         call. = FALSE)

  structure(.drop_null(list(
    type = ty, order = or, period = pe, forward = fw, backward = bw,
    intercept = ic, equation = if (isTRUE(eq)) 1L else NULL,
    r_squared = if (isTRUE(rs)) 1L else NULL,
    name = .val_str(name, "name"),
    format = .chart_format_payload(format, "format", accept = "line",
                                   part = "a trendline")
  )), class = "xl_chart_trendline")
}

# --- Error bars ---------------------------------------------------------------

#' Error bars on a chart series
#'
#' @description
#' `xl_chart_error_bars()` draws an error bar at each point, and is given to
#' [xl_chart_series()] as `x_error_bars` or `y_error_bars`.
#'
#' @param type How the size of each bar is worked out: `"std_error"` for the
#'   standard error, `"fixed"` for a constant, `"percentage"` of the point's
#'   own value, or `"std_dev"` for that many standard deviations.
#' @param value The constant, the percentage, or the number of standard
#'   deviations.  `"std_error"` needs none.
#' @param direction `"both"`, `"plus"` or `"minus"`.
#' @param endcap Draw the cap at the end of each bar.  On by default.
#' @param format An [xl_format()] styling the bars --- [xl_border()] only,
#'   since an error bar is a line.
#' @return An `xl_chart_error_bars` object.
#' @family images and charts
#' @seealso [xl_chart_series]
#' @export
#' @examples
#' xl_chart_error_bars("percentage", 5)
#' xl_chart_error_bars("std_dev", 1, direction = "plus", endcap = FALSE)
xl_chart_error_bars <- function(type, value = NA, direction = NULL,
                                endcap = NA, format = NULL) {
  if (missing(type))
    stop("`type` must say how the bars are sized, e.g. \"percentage\"",
         call. = FALSE)
  ty <- .val_enum(type, names(.LXW_ERROR_BAR_TYPE), "type")
  if (is.null(ty))
    stop("`type` must say how the bars are sized, e.g. \"percentage\"",
         call. = FALSE)
  val <- .val_num(value, "value", min = 0)
  if (is.null(val) && !identical(ty, "std_error"))
    stop(sprintf(paste0("a \"%s\" error bar needs a `value` -- the %s it is ",
                        "measured in."), ty,
                 switch(ty, fixed = "constant amount",
                        percentage = "percentage",
                        std_dev = "number of standard deviations")),
         call. = FALSE)
  structure(.drop_null(list(
    type = ty, value = val,
    direction = .val_enum(direction, names(.LXW_ERROR_BAR_DIRECTION),
                          "direction"),
    endcap = .val_flag(endcap, "endcap"),
    format = .chart_format_payload(format, "format", accept = "line",
                                   part = "an error bar")
  )), class = "xl_chart_error_bars")
}

# --- Shared helpers -----------------------------------------------------------

# A number format may be written as the Excel string or as xl_num_format().
.chart_num_format <- function(x, arg) {
  if (is.null(x)) return(NULL)
  if (is_xl_format(x)) {
    f <- unclass(x)$num_format$format
    if (is.null(f))
      stop(sprintf("`%s` is a format with no number format in it", arg),
           call. = FALSE)
    return(f)
  }
  if (!is.character(x) || length(x) != 1L || is.na(x))
    stop(sprintf("`%s` must be an Excel format string, e.g. \"#,##0.0\"", arg),
         call. = FALSE)
  x
}

# --- Checks that need the chart's type ----------------------------------------

.check_label_position <- function(pos, type, arg) {
  if (is.null(pos)) return(invisible(NULL))
  fams <- .LABEL_POSITION_FAMILIES[[pos]]
  if (.CHART_FAMILY(type) %in% fams) return(invisible(NULL))
  ok <- Filter(function(t) .CHART_FAMILY(t) %in% fams, names(.LXW_CHART_TYPE))
  allowed <- Filter(function(p) .CHART_FAMILY(type) %in%
                      .LABEL_POSITION_FAMILIES[[p]],
                    names(.LXW_LABEL_POSITION))
  stop(sprintf(paste0("`%s$position = \"%s\"` does not apply to a \"%s\" ",
                      "chart; Excel drops it.\n  On this chart the positions ",
                      "are: %s.\n  \"%s\" applies to: %s."),
               arg, pos, type, paste(allowed, collapse = ", "), pos,
               paste(ok, collapse = ", ")), call. = FALSE)
}

# Which constructor each xl_chart_series() part argument expects.
.PART_CLASS <- c(marker = "xl_chart_marker", labels = "xl_chart_labels",
                 trendline = "xl_chart_trendline",
                 x_error_bars = "xl_chart_error_bars",
                 y_error_bars = "xl_chart_error_bars")

# What needs the chart's type, and so cannot be checked until the series is
# put into a chart.  The parts validate themselves otherwise.
.check_series_parts <- function(se, type, arg) {
  p <- unclass(se)
  if (!is.null(p[["labels"]]))
    .check_label_position(unclass(p[["labels"]])[["position"]], type,
                          sprintf("%s$labels", arg))
  invisible(NULL)
}

# --- Payloads -----------------------------------------------------------------

.marker_payload <- function(m) {
  if (is.null(m)) return(NULL)
  p <- unclass(m)
  ent <- list()
  if (!is.null(p[["type"]]))
    ent$type <- unname(.LXW_MARKER_TYPE[[p[["type"]]]])
  ent$size <- p[["size"]]
  ent$line <- p[["format"]][["line"]]
  ent$fill <- p[["format"]][["fill"]]
  ent$pattern <- p[["format"]][["pattern"]]
  .drop_null(ent)
}

.label_payload <- function(l) {
  p <- unclass(l)
  .drop_null(list(value = p[["value"]], hide = p[["hide"]],
                  font = p[["format"]][["font"]],
                  line = p[["format"]][["line"]],
                  fill = p[["format"]][["fill"]],
                  pattern = p[["format"]][["pattern"]]))
}

.labels_payload <- function(l) {
  if (is.null(l)) return(NULL)
  p <- unclass(l)
  ent <- list(on = 1L)
  # Excel puts the value in a label unless told otherwise, so asking for the
  # percentage alone has to switch the value off -- a label would otherwise
  # read "10, 11.8%".  Naming any part means the label holds exactly those.
  shown <- c("show_name", "show_category", "show_value")
  asked <- c(shown, "show_percentage")
  if (any(vapply(asked, function(k) !is.null(p[[k]]), logical(1))))
    ent$options <- as.integer(vapply(shown, function(k) isTRUE(p[[k]]),
                                     logical(1)))
  if (isTRUE(p[["show_percentage"]])) ent$percentage <- 1L
  if (isTRUE(p[["show_legend_key"]])) ent$legend <- 1L
  if (isTRUE(p[["leader_lines"]]))    ent$leader_lines <- 1L
  if (!is.null(p[["position"]]))
    ent$position <- unname(.LXW_LABEL_POSITION[[p[["position"]]]])
  if (!is.null(p[["separator"]]))
    ent$separator <- unname(.LXW_LABEL_SEPARATOR[[p[["separator"]]]])
  ent$num_format <- p[["num_format"]]
  ent$font    <- p[["format"]][["font"]]
  ent$line    <- p[["format"]][["line"]]
  ent$fill    <- p[["format"]][["fill"]]
  ent$pattern <- p[["format"]][["pattern"]]
  if (!is.null(p[["custom"]]))
    # a NULL entry leaves that point alone, and reaches C as an empty label
    ent$custom <- lapply(p[["custom"]], function(x)
      if (is.null(x)) list() else .label_payload(x))
  .drop_null(ent)
}

.trendline_payload <- function(t) {
  if (is.null(t)) return(NULL)
  p <- unclass(t)
  # chart_series_set_trendline() takes one `value`: the polynomial order or
  # the moving-average period, depending on the type
  val <- if (!is.null(p[["order"]])) p[["order"]] else p[["period"]]
  .drop_null(list(
    type = unname(.LXW_TRENDLINE_TYPE[[p[["type"]]]]),
    value = if (is.null(val)) NULL else as.integer(val),
    forward = p[["forward"]], backward = p[["backward"]],
    intercept = p[["intercept"]], equation = p[["equation"]],
    r_squared = p[["r_squared"]], name = p[["name"]],
    line = p[["format"]][["line"]]))
}

.error_bars_payload <- function(e) {
  if (is.null(e)) return(NULL)
  p <- unclass(e)
  .drop_null(list(
    type = unname(.LXW_ERROR_BAR_TYPE[[p[["type"]]]]),
    value = if (is.null(p[["value"]])) 0 else p[["value"]],
    direction = if (is.null(p[["direction"]])) NULL else
      unname(.LXW_ERROR_BAR_DIRECTION[[p[["direction"]]]]),
    endcap = if (is.null(p[["endcap"]])) NULL else
      as.integer(p[["endcap"]]),
    line = p[["format"]][["line"]]))
}

.points_payload <- function(pts, arg) {
  if (is.null(pts)) return(NULL)
  if (is_xl_format(pts)) pts <- list(pts)
  if (!is.list(pts))
    stop(sprintf(paste0("`%s` must be a list of xl_format objects, one per ",
                        "point, with NULL for a point to leave alone"), arg),
         call. = FALSE)
  out <- lapply(seq_along(pts), function(i) {
    x <- pts[[i]]
    if (is.null(x)) return(list())
    f <- .chart_format_payload(x, sprintf("%s[[%d]]", arg, i),
                               accept = c("line", "fill", "pattern"),
                               part = "a chart point")
    .drop_null(list(line = f[["line"]], fill = f[["fill"]],
                    pattern = f[["pattern"]]))
  })
  if (!length(out)) return(NULL)
  out
}

#' @export
print.xl_chart_marker <- function(x, ...) .print_part(x, "xl_chart_marker")

#' @export
print.xl_chart_labels <- function(x, ...) .print_part(x, "xl_chart_labels")

#' @export
print.xl_chart_label <- function(x, ...) .print_part(x, "xl_chart_label")

#' @export
print.xl_chart_trendline <- function(x, ...) .print_part(x, "xl_chart_trendline")

#' @export
print.xl_chart_error_bars <- function(x, ...)
  .print_part(x, "xl_chart_error_bars")

.print_part <- function(x, cls) {
  p <- unclass(x)
  cat("<", cls, ">\n", sep = "")
  if (length(p)) cat("  set:", paste(names(p), collapse = ", "), "\n")
  invisible(x)
}

# Every object writexl hands back prints something.  A print method is the one
# piece of a package nothing else exercises -- no other code path calls it, so
# a typo in one sits there until a user types the object's name at the prompt.
#
# Each object is built twice where it matters: once bare and once with the
# optional parts filled in, since most of these methods only describe what is
# set.

expect_prints <- function(x, ...) {
  out <- utils::capture.output(print(x))
  expect_true(length(out) >= 1L)
  for (pattern in c(...)) expect_true(any(grepl(pattern, out, fixed = TRUE)),
                                      info = pattern)
  # print() returns its argument invisibly, as R expects of a print method
  vis <- NULL
  utils::capture.output(vis <- withVisible(print(x))$visible)
  expect_identical(vis, FALSE)
  invisible(out)
}

df <- data.frame(quarter = c("Q1", "Q2"), revenue = c(10, 25),
                 stringsAsFactors = FALSE)

test_that("the cell content objects print", {
  expect_prints(xl_cell_general(value = 1:2), "xl_cell_general")
  expect_prints(xl_comment("note", author = "QA"), "value:", "note")
  expect_prints(xl_merge("A1:B1", "Total"), "xl_merge", "Total")
  expect_prints(xl_merge(list(rows = 1, cols = 1:2)), "<spec>")
  expect_prints(xl_rich_run("bold", xl_font(bold = TRUE)), "formatted")
  expect_prints(xl_rich_run("plain"), "xl_rich_run")
  expect_prints(xl_rich_string("a ", xl_rich_run("b", xl_font(bold = TRUE))),
                "2 runs", "formatted")
})

test_that("the format objects print", {
  expect_prints(xl_format(), "xl_format")
  expect_prints(xl_font(bold = TRUE) + xl_fill(background = "yellow"), "font")
  expect_prints(xl_col_spec("revenue", width = 12), "<xl_col_spec>", "width=12")
  expect_prints(xl_row_spec(1, height = 20), "<xl_row_spec>", "height=20")
})

test_that("the worksheet and workbook objects print", {
  expect_prints(xl_sheet(df), "2 rows x 2 cols")
  expect_prints(xl_sheet(df, cols = xl_col_spec("revenue", width = 12),
                         rows = xl_row_spec(1, height = 20)),
                "column specs:", "row specs:")
  expect_prints(xl_workbook(list(D = df)), "xl_workbook")
  expect_prints(xl_properties(), "xl_properties")
  expect_prints(xl_properties(title = "T", author = "A", company = "C",
                              subject = "S", read_only = TRUE),
                "title:", "read_only: TRUE")
  expect_prints(xl_page_setup(orientation = "landscape"), "xl_page_setup")
  expect_prints(xl_sheet_view(), "xl_sheet_view")
  expect_prints(xl_sheet_view(active = TRUE, hide_zero = TRUE),
                "active", "hide_zero")
  expect_prints(xl_outline(symbols_below = FALSE), "xl_outline")
})

test_that("the worksheet feature objects print", {
  expect_prints(xl_validation("A1:A5", type = "integer", criteria = ">",
                              value = 0), "xl_validation")
  expect_prints(xl_cond_cell("A1:A5", type = "cell", criteria = ">",
                             value = 5, format = xl_font(bold = TRUE)),
                "xl_conditional")
  expect_prints(xl_cond_scale("A1:A5"), "xl_conditional")
  expect_prints(xl_filter(col = "quarter", criteria = "==", value = "Q1"),
                "xl_filter")
  expect_prints(xl_table(name = "Sales"), "xl_table")
  expect_prints(xl_table_column("revenue", total = "sum"), "revenue", "total=")
  expect_prints(xl_table_column("revenue", formula = "=1"), "formula")
  expect_prints(xl_image(png_file(), "A1"), "xl_image")
})

test_that("the chart objects print", {
  s <- xl_chart_series(values = list(cols = "revenue"))
  expect_prints(s, "xl_chart_series")
  expect_prints(xl_chart("column", s), "xl_chart", "column")
  expect_prints(xl_chart_axis(title = "Quarter", min = 0), "xl_chart_axis")
  expect_prints(xl_chart_axis(), "xl_chart_axis")
  expect_prints(xl_chart_marker(type = "circle", size = 8), "xl_chart_marker")
  expect_prints(xl_chart_labels(show_value = TRUE), "xl_chart_labels")
  expect_prints(xl_chart_label(value = "peak"), "xl_chart_label")
  expect_prints(xl_chart_trendline(type = "linear"), "xl_chart_trendline")
  expect_prints(xl_chart_error_bars(type = "percentage", value = 5),
                "xl_chart_error_bars")
  expect_prints(xl_chart_legend(position = "bottom"), "xl_chart_legend")
  expect_prints(xl_chart_table(show_keys = TRUE), "xl_chart_table")
  expect_prints(xl_chartsheet(xl_chart("pie", s)), "xl_chartsheet", "pie")
  expect_prints(xl_chartsheet(xl_chart("pie", s), zoom = 150), "set:", "zoom")
})

# Markers, data labels, trendlines, error bars and individual points.  As with
# the axes, the point of most of this is that libxlsxwriter documents a
# restriction and then lets Excel drop the setting in silence.

part_sales <- data.frame(quarter = c("Q1", "Q2", "Q3", "Q4"),
                         revenue = c(10, 20, 30, 40),
                         stringsAsFactors = FALSE)

part_xml <- function(..., type = "line", df = part_sales) {
  t <- tempfile(fileext = ".xlsx")
  se <- xl_chart_series(values = list(cols = "revenue"),
                        categories = list(cols = "quarter"), ...)
  write_xlsx(list(Data = xl_sheet(df, chart = xl_chart(type, se))), t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                             full.names = TRUE)[1L], warn = FALSE),
        collapse = "")
}

# ── Markers ───────────────────────────────────────────────────────────────────

test_that("the marker types are libxlsxwriter's, in its order", {
  expect_equal(unname(.LXW_MARKER_TYPE), 0:10)
  expect_equal(names(.LXW_MARKER_TYPE)[1:2], c("automatic", "none"))
})

test_that("a marker reaches the file with its size and styling", {
  x <- part_xml(marker = xl_chart_marker(type = "diamond", size = 9,
                                         format = xl_fill(background = "red")))
  expect_match(x, '<c:symbol val="diamond"/>', fixed = TRUE)
  expect_match(x, '<c:size val="9"/>', fixed = TRUE)
  expect_match(x, "FF0000", fixed = TRUE)
  expect_match(part_xml(marker = xl_chart_marker(type = "none")),
               '<c:symbol val="none"/>', fixed = TRUE)
})

test_that("an automatic marker cannot also be sized or styled", {
  # documented in chart.h: "If automatic is on then other marker properties
  # such as size, line or fill cannot be set"
  expect_error(xl_chart_marker(type = "automatic", size = 8),
               "cannot then be given a size or a format")
  expect_error(xl_chart_marker(type = "automatic",
                               format = xl_fill(background = "red")),
               "cannot then be given a size or a format")
  expect_s3_class(xl_chart_marker(type = "automatic"), "xl_chart_marker")
  # and the restriction is on "automatic" alone
  expect_s3_class(xl_chart_marker(type = "circle", size = 8),
                  "xl_chart_marker")
})

test_that("marker arguments are validated", {
  expect_error(xl_chart_marker(type = "blob"), "`type` must be one of")
  expect_error(xl_chart_marker(size = 1), "must be between 2 and 72")
  expect_error(xl_chart_marker(format = xl_font(bold = TRUE)),
               "a shape and has no text")
})

# ── Data labels ───────────────────────────────────────────────────────────────

test_that("labels default to showing the value, as Excel does", {
  x <- part_xml(labels = xl_chart_labels())
  expect_match(x, "<c:dLbls>", fixed = TRUE)
  expect_match(x, '<c:showVal val="1"/>', fixed = TRUE)
})

test_that("every label option reaches the file", {
  x <- part_xml(labels = xl_chart_labels(
    show_value = TRUE, show_category = TRUE, show_name = FALSE,
    show_legend_key = TRUE, separator = "newline", position = "above",
    num_format = "0.0", leader_lines = TRUE,
    format = xl_font(size = 9, bold = TRUE) +
             xl_fill(background = "#FFF2CC") +
             xl_border(all = "thin", color = "gray")))
  for (w in c('<c:showVal val="1"/>', '<c:showCatName val="1"/>',
              '<c:showLegendKey val="1"/>',
              "<c:separator>", '<c:dLblPos val="t"/>', 'formatCode="0.0"',
              '<c:showLeaderLines val="1"/>', 'sz="900"', 'b="1"',
              "FFF2CC"))
    expect_true(grepl(w, x, fixed = TRUE), info = w)
  # what is switched off is left out rather than written as 0, which is how
  # Excel writes it too -- these all default to off
  expect_false(grepl("showSerName", x, fixed = TRUE))
})

test_that("a label holds exactly the parts named, and the value otherwise", {
  # Excel puts the value in every label unless told otherwise, so the
  # percentage alone would otherwise read "10, 11.8%"
  only_pc <- part_xml(labels = xl_chart_labels(show_percentage = TRUE),
                      type = "pie")
  expect_match(only_pc, '<c:showPercent val="1"/>', fixed = TRUE)
  expect_false(grepl("showVal", only_pc, fixed = TRUE))

  # naming nothing leaves Excel's default
  expect_match(part_xml(labels = xl_chart_labels(), type = "pie"),
               '<c:showVal val="1"/>', fixed = TRUE)

  # and both together, when both are asked for
  both <- part_xml(labels = xl_chart_labels(show_value = TRUE,
                                            show_percentage = TRUE),
                   type = "pie")
  expect_match(both, '<c:showVal val="1"/>', fixed = TRUE)
  expect_match(both, '<c:showPercent val="1"/>', fixed = TRUE)

  # naming the category alone drops the value too
  cat_only <- part_xml(labels = xl_chart_labels(show_category = TRUE),
                       type = "pie")
  expect_match(cat_only, '<c:showCatName val="1"/>', fixed = TRUE)
  expect_false(grepl("showVal", cat_only, fixed = TRUE))
})

test_that("a label position is refused where the chart type has no such place", {
  # the table lives in chart.h and Excel drops what does not apply
  expect_error(xl_chart("column", xl_chart_series(values = "A1:A5",
                          labels = xl_chart_labels(position = "above"))),
               'does not apply to a "column" chart')
  expect_error(xl_chart("column", xl_chart_series(values = "A1:A5",
                          labels = xl_chart_labels(position = "above"))),
               "the positions are: default, center, inside_base", fixed = TRUE)
  expect_error(xl_chart("line", xl_chart_series(values = "A1:A5",
                          labels = xl_chart_labels(position = "best_fit"))),
               "applies to: doughnut, pie", fixed = TRUE)
  # and each is accepted where the table says it belongs
  expect_s3_class(xl_chart("line", xl_chart_series(values = "A1:A5",
                            labels = xl_chart_labels(position = "above"))),
                  "xl_chart")
  expect_s3_class(xl_chart("column", xl_chart_series(values = "A1:A5",
                            labels = xl_chart_labels(position = "inside_end"))),
                  "xl_chart")
  expect_s3_class(xl_chart("pie", xl_chart_series(values = "A1:A5",
                            labels = xl_chart_labels(position = "best_fit"))),
                  "xl_chart")
  # "center" is the one position every type allows
  for (ty in names(.LXW_CHART_TYPE)) {
    se <- if (identical(.CHART_FAMILY(ty), "scatter"))
      xl_chart_series(values = "B1:B5", categories = "A1:A5",
                      labels = xl_chart_labels(position = "center"))
    else xl_chart_series(values = "A1:A5",
                         labels = xl_chart_labels(position = "center"))
    expect_true(inherits(xl_chart(ty, se), "xl_chart"), info = ty)
  }
})

test_that("custom labels give one point its own text, styling or nothing", {
  x <- part_xml(type = "pie", labels = xl_chart_labels(
    custom = list(xl_chart_label(value = "biggest"), NULL,
                  xl_chart_label(hide = TRUE),
                  xl_chart_label(format = xl_font(bold = TRUE)))))
  expect_match(x, "biggest", fixed = TRUE)
  expect_match(x, '<c:delete val="1"/>', fixed = TRUE)
})

test_that("a hidden custom label carries nothing else", {
  expect_error(xl_chart_label(hide = TRUE, value = "x"),
               "cannot also carry a value or a format")
  expect_error(xl_chart_label(hide = TRUE, format = xl_font(bold = TRUE)),
               "cannot also carry a value or a format")
  expect_error(xl_chart_labels(custom = list("text")),
               "must be an xl_chart_label\\(\\) or NULL")
  # one label need not be wrapped in a list
  expect_length(unclass(xl_chart_labels(custom = xl_chart_label(value = "a")))$custom,
                1L)
})

test_that("label arguments are validated", {
  expect_error(xl_chart_labels(position = "middle"), "`position` must be one of")
  expect_error(xl_chart_labels(separator = "tab"), "`separator` must be one of")
  expect_error(xl_chart_labels(num_format = 42), "must be an Excel format string")
  expect_error(xl_chart_labels(num_format = xl_font(bold = TRUE)),
               "no number format in it")
  expect_equal(unclass(xl_chart_labels(num_format = xl_num_format("0%")))$num_format,
               "0%")
})

# ── Trendlines ────────────────────────────────────────────────────────────────

test_that("a trendline reaches the file with all of its options", {
  x <- part_xml(trendline = xl_chart_trendline(
    "linear", forward = 1, backward = 0.5, equation = TRUE, r_squared = TRUE,
    intercept = 0, name = "fit",
    format = xl_border(all = "dashed", color = "green")))
  for (w in c("<c:trendline>", '<c:trendlineType val="linear"/>',
              '<c:forward val="1"/>', '<c:backward val="0.5"/>',
              '<c:dispEq val="1"/>', '<c:dispRSqr val="1"/>',
              '<c:intercept val="0"/>', "fit", '<a:prstDash val="dash"/>'))
    expect_true(grepl(w, x, fixed = TRUE), info = w)
})

test_that("a polynomial needs an order and a moving average a period", {
  expect_error(xl_chart_trendline("poly"), "needs an `order`")
  expect_error(xl_chart_trendline("average"), "needs a `period`")
  expect_error(xl_chart_trendline("linear", order = 2),
               "not of a \"linear\" one")
  expect_error(xl_chart_trendline("linear", period = 2),
               "a \"linear\" fit has none")
  x <- part_xml(trendline = xl_chart_trendline("poly", order = 3))
  expect_match(x, '<c:order val="3"/>', fixed = TRUE)
  y <- part_xml(trendline = xl_chart_trendline("average", period = 2))
  expect_match(y, '<c:trendlineType val="movingAvg"/>', fixed = TRUE)
  expect_match(y, '<c:period val="2"/>', fixed = TRUE)
})

test_that("a moving average has no forecast, equation or R-squared", {
  # chart.h: "This feature isn't available for Moving Average in Excel"
  for (a in list(list(forward = 1), list(backward = 1), list(equation = TRUE),
                 list(r_squared = TRUE)))
    expect_error(do.call(xl_chart_trendline,
                         c(list("average", period = 2), a)),
                 "does not apply to a moving average", label = names(a))
  # the same options are fine on a fitted type
  expect_s3_class(xl_chart_trendline("exp", forward = 1, equation = TRUE),
                  "xl_chart_trendline")
})

test_that("an intercept applies to three of the six fits", {
  # chart.h: "only available in Excel for Exponential, Linear and Polynomial"
  for (ty in c("exp", "linear"))
    expect_s3_class(xl_chart_trendline(ty, intercept = 0),
                    "xl_chart_trendline")
  expect_s3_class(xl_chart_trendline("poly", order = 2, intercept = 0),
                  "xl_chart_trendline")
  for (ty in c("log", "power"))
    expect_error(xl_chart_trendline(ty, intercept = 0),
                 "exponential, linear or polynomial", label = ty)
  expect_error(xl_chart_trendline("average", period = 2, intercept = 0),
               "exponential, linear or polynomial")
})

test_that("trendline arguments are validated", {
  expect_error(xl_chart_trendline(), "must name the fit")
  expect_error(xl_chart_trendline("cubic"), "`type` must be one of")
  expect_error(xl_chart_trendline("poly", order = 1), "must be between 2 and NA")
  expect_error(xl_chart_trendline("linear", format = xl_fill(background = "red")),
               "takes no fill")
})

# ── Error bars ────────────────────────────────────────────────────────────────

test_that("error bars reach both axes of a series", {
  t <- tempfile(fileext = ".xlsx")
  write_xlsx(list(Data = xl_sheet(part_sales, chart = xl_chart("scatter",
    xl_chart_series(values = list(cols = "revenue"),
                    categories = list(cols = "revenue"),
                    x_error_bars = xl_chart_error_bars("fixed", 2,
                                                       direction = "plus"),
                    y_error_bars = xl_chart_error_bars("std_dev", 1))))), t)
  d <- tempfile(); dir.create(d); utils::unzip(t, exdir = d)
  x <- paste(readLines(list.files(file.path(d, "xl/charts"), pattern = "^chart",
                                  full.names = TRUE)[1L], warn = FALSE),
             collapse = "")
  expect_length(regmatches(x, gregexpr("<c:errBars>", x))[[1L]], 2L)
  expect_match(x, '<c:errBarType val="plus"/>', fixed = TRUE)
  expect_match(x, '<c:errValType val="stdDev"/>', fixed = TRUE)
})

test_that("an error bar's endcap and line reach the file", {
  x <- part_xml(y_error_bars = xl_chart_error_bars(
    "percentage", 10, endcap = FALSE,
    format = xl_border(all = "dotted", color = "purple")))
  expect_match(x, '<c:noEndCap val="1"/>', fixed = TRUE)
  expect_match(x, '<c:errValType val="percentage"/>', fixed = TRUE)
  expect_match(x, '<c:val val="10"/>', fixed = TRUE)
  expect_match(x, '<a:prstDash val="dot"/>', fixed = TRUE)
})

test_that("every error-bar type but the standard error needs a value", {
  for (ty in c("fixed", "percentage", "std_dev"))
    expect_error(xl_chart_error_bars(ty), "needs a `value`", label = ty)
  expect_s3_class(xl_chart_error_bars("std_error"), "xl_chart_error_bars")
  expect_error(xl_chart_error_bars(), "must say how the bars are sized")
  expect_error(xl_chart_error_bars("guess", 1), "`type` must be one of")
  expect_error(xl_chart_error_bars("fixed", 1, direction = "up"),
               "`direction` must be one of")
})

# ── Individual points ─────────────────────────────────────────────────────────

test_that("points are styled one at a time, and NULL leaves one alone", {
  x <- part_xml(type = "pie",
                points = list(xl_fill(background = "red"),
                              xl_fill(background = "green"), NULL,
                              xl_fill(background = "blue")))
  # three styled, one left as it was
  expect_length(regmatches(x, gregexpr("<c:dPt>", x))[[1L]], 3L)
  expect_match(x, "00FF00", fixed = TRUE)
})

test_that("a point takes no font, and one format need not be a list", {
  expect_error(xl_chart_series(values = "A1:A5",
                               points = list(xl_font(bold = TRUE))),
               "a shape and has no text")
  expect_length(unclass(xl_chart_series(values = "A1:A5",
                                        points = xl_fill(background = "red")))$points,
                1L)
})

# ── The series arguments themselves ───────────────────────────────────────────

test_that("each part argument insists on its own constructor", {
  for (nm in names(.PART_CLASS)) {
    args <- list(values = "A1:A5")
    args[[nm]] <- "yes please"
    expect_error(do.call(xl_chart_series, args),
                 sprintf("`%s` must be an %s", nm, .PART_CLASS[[nm]]),
                 fixed = TRUE, label = nm)
  }
})

# ── Coverage of the C API ─────────────────────────────────────────────────────

test_that("every chart_series_*() function libxlsxwriter offers is called", {
  hdr <- NULL
  for (p in c("../../src/libxlsxwriter/include/xlsxwriter/chart.h",
              "../../../src/libxlsxwriter/include/xlsxwriter/chart.h"))
    if (file.exists(p)) hdr <- p
  skip_if(is.null(hdr), "libxlsxwriter sources not alongside the tests")
  src <- NULL
  for (p in c("../../src/write_xlsx.c", "../../../src/write_xlsx.c"))
    if (file.exists(p)) src <- p
  skip_if(is.null(src), "write_xlsx.c not alongside the tests")

  # declarations only: the header's prose mentions chart_series_set_marker_xxx()
  decl <- grep("^(void|lxw_error|lxw_series_error_bars \\*|lxw_chart_series \\*)",
               readLines(hdr), value = TRUE)
  fns <- unique(regmatches(decl, regexpr("chart_series_[a-z_]+", decl)))
  fns <- fns[nzchar(fns)]
  expect_equal(length(fns), 40L)

  called <- paste(readLines(src), collapse = "\n")
  for (f in fns)
    expect_true(grepl(paste0(f, "("), called, fixed = TRUE), info = f)
})

test_that("the part constructors name what was wrong", {
  expect_error(xl_chart_labels(custom = "not a label"),
               "must be a list of xl_chart_label")
  expect_error(xl_chart_trendline(type = NA), "must name the fit")
  expect_error(xl_chart_error_bars(type = NA, value = 1),
               "must say how the bars are sized")
  expect_error(.points_payload("not a format", "points"),
               "must be a list of xl_format objects")
  # an empty list of points is nothing to apply, not an error
  expect_null(.points_payload(list(), "points"))
  # a single format is taken as one point's worth
  expect_length(.points_payload(xl_fill(background = "red"), "points"), 1L)
})

#' Export to xlsx
#'
#' Writes a data frame to an xlsx file. To create an xlsx with (multiple) named
#' sheets, simply set \code{x} to a named list of data frames.
#'
#' Supports strings, numbers, booleans and dates automatically. For cell
#' formatting (fonts, fills, borders, number formats, ...), worksheet layout
#' (column widths, frozen panes, ...), and workbook metadata, wrap columns with
#' \code{\link{xl_cell_general}}, sheets with \code{\link{xl_sheet}}, and the
#' whole workbook with \code{\link{xl_workbook}}. See the "Formatting and
#' workbook properties" vignette and \code{\link{xl_format}}.
#'
#' \if{html}{
#' \out{
#' <link rel="stylesheet" type="text/css" href="https://jeroen.github.io/clippy/clippy.min.css" media="all">
#' <script src="https://jeroen.github.io/clippy/bundle.js"></script>
#' }}
#'
#' @export
#' @aliases writexl
#' @useDynLib writexl C_write_data_frame_list
#' @param x a data frame, an [xl_sheet], an [xl_workbook], or a (named) list of
#'   data frames / `xl_sheet`s that become the sheets in the xlsx
#' @param path a file name to write to
#' @param col_names write column names as the header row at the top of the sheet?
#' @param format_headers apply the workbook's header format to that header row?
#'   The default header format is bold and centered; change it with
#'   \code{\link{xl_properties}(header_format = )}.
#' @param use_zip64 use \href{https://en.wikipedia.org/wiki/Zip_(file_format)#ZIP64}{zip64}
#' to enable support for 4GB+ xlsx files. Not all platforms can read this.
#' @param na what to write where a value has none. `NA` (the default) leaves
#'   the cell blank, as writexl has always done; anything else is written in
#'   its place, keeping its own type. Shorthand for
#'   \code{\link{xl_properties}(na = )}, so give it there instead when `x` is
#'   already an [xl_workbook]. A column or a single cell can override it: see
#'   [xl_col_spec()] and [xl_cell_general()].
#' @param constant_memory stream rows to disk instead of building the whole
#'   workbook in memory. `NA` (the default) decides per workbook: on for large
#'   data, off for small, and always off when a feature needs it off. `TRUE`
#'   forces it on for a workbook that would otherwise be judged too small;
#'   `FALSE` forces it off. Features that cannot be written while streaming ---
#'   merged ranges, tables, embedded images and multi-cell array formulas ---
#'   turn it off regardless, with a warning if `TRUE` was asked for, because
#'   the alternative is a file that opens cleanly and is missing cells.
#' @param constant_memory_threshold how much extra memory not streaming would
#'   have to cost, in bytes, before streaming is worth it. Default 128 MiB.
#'   The cost is *estimated* from the number of cells in the workbook, using a
#'   fixed per-cell figure calibrated against a range of data; the true cost
#'   varies with the data, and is lowest for text that repeats. Streaming saves
#'   memory but produces slightly larger files, so it is not used for workbooks
#'   small enough that the saving would not be noticed.
#' @family workbook settings
#' @examples # Roundtrip example with single excel sheet named 'mysheet'
#' tmp <- write_xlsx(list(mysheet = iris))
#' readxl::read_xlsx(tmp)
write_xlsx <- function(x, path = tempfile(fileext = ".xlsx"), col_names = TRUE,
                       format_headers = TRUE, na = NA, use_zip64 = FALSE,
                       constant_memory = NA,
                       constant_memory_threshold = 128 * 1024^2){
  # Resolve the input to an xl_workbook.  A bare data frame / xl_sheet / list
  # is wrapped in a workbook with default properties; an explicit xl_workbook
  # overrides col_names/format_headers.
  if(inherits(x, "xl_workbook")){
    # a workbook carries its own properties, so the shorthand has nowhere to go
    # rather than somewhere surprising
    if(!is.null(.val_na(na)))
      stop("`na` is a shorthand for xl_properties(na = ); `x` is already an ",
           "xl_workbook, so set it there", call. = FALSE)
    wb <- x
    col_names <- wb$col_names
    format_headers <- wb$format_headers
  } else {
    wb <- xl_workbook(x, properties = xl_properties(na = na),
                      col_names = col_names, format_headers = format_headers)
  }
  props <- wb$properties
  elems <- wb$sheets
  # a chartsheet has no cells, so it stands in the data-frame list as a data
  # frame with no columns; C branches on the plan before the row loop
  is_cs <- vapply(elems, inherits, logical(1), "xl_chartsheet")
  dfs <- lapply(seq_along(elems), function(i)
    if (is_cs[[i]]) .chartsheet_placeholder()
    else if (inherits(elems[[i]], "xl_sheet")) elems[[i]]$data else elems[[i]])
  names(dfs) <- names(elems)
  dfs <- lapply(dfs, normalize_df)
  names(dfs) <- .resolve_sheet_names(names(elems), length(dfs))
  # tab visibility is the one worksheet setting whose rules span the whole
  # workbook, so it is checked here rather than per sheet
  .resolve_sheet_visibility(elems, names(dfs))
  stopifnot(is.character(path) && length(path))
  path <- normalizePath(path, mustWork = FALSE)
  # Excel has no concept of a time zone, so decide once, for the whole workbook,
  # how POSIXct values are written (see .resolve_timezones()).
  tz_res <- .resolve_timezones(dfs, props)
  dfs <- tz_res$dfs
  props <- tz_res$props
  # Resolve everything into one deduplicated workbook format table: the header
  # format, each general cell's effective format, and each sheet's column/row
  # plan -- all cascaded over the workbook default_format.
  header_offset <- if(isTRUE(as.logical(col_names))) 1L else 0L
  reg <- .new_format_registry()
  header_id <- .register_format(reg, merge_xl_format(props$default_format,
                                                     props$header_format))
  dfs <- lapply(dfs, .resolve_sheet_formats, reg = reg, props = props,
                header_offset = header_offset)
  # Table names are unique across the workbook, not the sheet, so they are
  # resolved here alongside the sheet names rather than inside xl_table().
  table_names <- .resolve_table_names(elems, names(dfs))
  # Charts resolve after the sheets are known: a series may plot data from any
  # sheet, so its range cannot be resolved against one sheet in isolation.
  chart_plans <- Map(function(el, nm)
    .resolve_charts(el, dfs[[nm]], reg, header_offset, props, dfs, nm),
    elems, names(dfs))
  sheets <- Map(function(el, df, tn, ch)
                  if (inherits(el, "xl_chartsheet"))
                    .resolve_chartsheet_plan(el, ch)
                  else
                    .resolve_sheet_plan(el, df, reg, header_offset, props, tn,
                                        ch),
                elems, dfs, table_names, chart_plans)
  # Ordering matters to libxlsxwriter's drawing numbering, so this spans the
  # whole workbook rather than any one sheet
  .check_drawing_order(sheets, names(dfs))
  cm <- .resolve_constant_memory(dfs, props, sheets, constant_memory,
                                 constant_memory_threshold)
  ret <- .Call(C_write_data_frame_list, dfs, path, col_names, format_headers,
               use_zip64, reg$table, sheets, header_id,
               .properties_payload(props, cm$on))
  invisible(ret)
}

# =============================================================================
# Time zones
# =============================================================================
#
# Excel stores a datetime as a naive serial number: it has no time zone.  Rather
# than silently converting every POSIXct to UTC, writexl decides per workbook:
#
#   * all POSIXct values share one time zone -> drop the zone and write local
#     wall-clock time (2025-12-01 01:00 in Perth is written as 01:00), and drop
#     the " UTC" suffix from the default datetime format so nothing is mislabelled;
#   * the time zones differ -> convert everything to UTC and warn, since no
#     single wall clock can represent them all.
#
# The survey covers plain POSIXct columns and POSIXct values inside
# xl_cell_general cells.

# The time zone of a POSIXct, with NULL/"" (meaning local time) resolved to the
# session zone so that an unset zone and an explicit local zone compare equal.
.tzone_of <- function(x){
  tz <- attr(x, "tzone", exact = TRUE)
  if(is.null(tz) || !length(tz) || !nzchar(tz[1])){
    loc <- Sys.timezone()
    # Sys.timezone() reports NA only where the OS has no zone configured
    if(is.na(loc)) "" else loc  # nocov
  } else {
    tz[1]
  }
}

# Every distinct time zone used by a POSIXct anywhere in the workbook.  Custom
# document properties count: they are datetimes in the same workbook, and
# writing them under a different rule than the cells would be incoherent.
.collect_tzones <- function(dfs, props = NULL){
  out <- character(0)
  if(!is.null(props$custom))
    for(v in props$custom)
      if(inherits(v, "POSIXct"))
        out <- c(out, .tzone_of(v))
  for(df in dfs){
    for(col in df){
      if(inherits(col, "POSIXct")){
        out <- c(out, .tzone_of(col))
      } else if(inherits(col, "xl_cell_general")){
        for(rec in .cell_records(col))
          if(inherits(rec$value, "POSIXct"))
            out <- c(out, .tzone_of(rec$value))
      }
    }
  }
  unique(out)
}

# Re-express a POSIXct so that its wall-clock reading is what reaches the cell.
# The offset is taken per instant, so daylight saving is handled correctly.
.drop_tzone <- function(x){
  off <- as.POSIXlt(x, tz = .tzone_of(x))$gmtoff
  # Not every platform reports gmtoff: older Windows builds of R return NULL or
  # a zero-length vector.  Fall back to leaving the value alone (i.e. UTC)
  # rather than writing something wrong -- note the length check, without which
  # a zero-length offset would recycle to nothing and silently empty the column.
  if(is.null(off) || length(off) != length(x) ||
     any(is.na(off) & !is.na(as.numeric(x))))
    return(x)  # nocov  (only on a platform whose POSIXlt has no gmtoff)
  off[is.na(off)] <- 0
  structure(as.numeric(x) + off, class = c("POSIXct", "POSIXt"), tzone = "UTC")
}

# Apply .drop_tzone() to every POSIXct in a sheet, including inside cell objects.
.drop_tzones_sheet <- function(df){
  for(j in seq_along(df)){
    col <- df[[j]]
    if(inherits(col, "POSIXct")){
      df[[j]] <- .drop_tzone(col)
    } else if(inherits(col, "xl_cell_general")){
      recs <- .cell_records(col)
      touched <- FALSE
      for(k in seq_along(recs)){
        if(inherits(recs[[k]]$value, "POSIXct")){
          recs[[k]]$value <- .drop_tzone(recs[[k]]$value)
          touched <- TRUE
        }
      }
      if(touched)
        df[[j]] <- .new_cell_vector(recs)
    }
  }
  df
}

.resolve_timezones <- function(dfs, props){
  tzs <- .collect_tzones(dfs, props)
  if(length(tzs) > 1L){
    warning("The workbook contains datetimes in ", length(tzs),
            " different time zones (", paste(tzs, collapse = ", "),
            "); all of them were converted to UTC because Excel does not ",
            "support time zones. Convert them yourself beforehand if you want ",
            "different behaviour.", call. = FALSE)
  } else if(length(tzs) == 1L){
    dfs <- lapply(dfs, .drop_tzones_sheet)
    # custom document properties follow the same rule as the cells
    if(!is.null(props$custom))
      for(k in seq_along(props$custom))
        if(inherits(props$custom[[k]], "POSIXct"))
          props$custom[[k]] <- .drop_tzone(props$custom[[k]])
    # The default format labels datetimes "UTC"; that is no longer true once the
    # zone has been dropped.  Only replace it if the caller kept the default.
    if(identical(unclass(props$datetime_format),
                 unclass(.default_datetime_format())))
      props$datetime_format <- xl_num_format("yyyy-mm-dd HH:mm:ss")
  }
  list(dfs = dfs, props = props)
}

# Walk a sheet's xl_cell_general columns, register each cell's effective format
# in the workbook registry, attach the resulting integer id vector, flatten each
# comment's xl_format into the flat fields C applies, and resolve each array
# formula's range to the 0-based quad C writes it with.
#
# Multi-cell array ranges are collected on the returned data frame (attribute
# "writexl_array_multicell") because they decide the workbook's constant-memory
# flag -- see .resolve_constant_memory().
.resolve_sheet_formats <- function(df, reg, props, header_offset = 1L) {
  multicell <- list()
  for (j in seq_along(df)) {
    col <- df[[j]]
    if (!inherits(col, "xl_cell_general")) next
    recs <- .cell_records(col)
    ids <- vapply(recs, .resolve_cell_format_id, integer(1), reg = reg, props = props)
    for (k in seq_along(recs)) {
      rec <- recs[[k]]
      if (!is.null(rec$comment)) rec$comment <- .comment_c_payload(rec$comment)
      fm_set <- !is.null(rec$formula) && length(rec$formula) == 1L &&
                !is.na(rec$formula)
      if ((isTRUE(rec$array) || isTRUE(rec$dynamic)) && fm_set) {
        anchor <- c((k - 1L) + as.integer(header_offset), j - 1L)
        q <- .resolve_array_quad(rec$array_range, anchor, df, header_offset,
                                 names(df)[j], k)
        rec$array_quad <- q
        if (q[3L] > q[1L] || q[4L] > q[2L]) {
          .check_array_no_overlap(q, anchor, df, header_offset, names(df)[j], k)
          multicell <- c(multicell, list(q))
        }
      }
      # a rich string's runs each carry their own font, registered like any
      # other format; C receives the parallel text / id vectors
      if (is_xl_rich_string(rec$value)) {
        rich <- .rich_string_c_payload(rec$value, reg, props)
        rec$rich_text <- rich$rich_text
        rec$rich_format_id <- rich$rich_format_id
        rec$value <- NULL
      }
      # the unresolved user spec has served its purpose; keep the C payload tight
      rec$array_range <- NULL
      recs[[k]] <- rec
    }
    col <- .new_cell_vector(recs)
    attr(col, "writexl_format_ids") <- ids
    df[[j]] <- col
  }
  attr(df, "writexl_array_multicell") <- multicell
  df
}

# Resolve one array formula's range.  With no declared range the formula covers
# only its own cell, which is the modern dynamic-array spelling: Excel spills
# the result itself.  A declared range must start at the cell, because
# libxlsxwriter writes the formula into the range's top-left corner -- were the
# two allowed to differ the formula would land in a different cell than the one
# the caller wrote it to.
.resolve_array_quad <- function(spec, anchor, df, header_offset, colname, k) {
  if (is.null(spec))
    return(as.integer(c(anchor[1L], anchor[2L], anchor[1L], anchor[2L])))
  arg <- sprintf('array_range (column "%s", cell %d)', colname, k)
  q <- .xl_resolve_range(spec, arg = arg, df = df,
                         header_offset = header_offset, allow_cell = TRUE)
  if (q[1L] != anchor[1L] || q[2L] != anchor[2L])
    stop(sprintf(paste0("`%s` must start at the cell that holds the formula ",
                        "(row %d, column %d of the sheet, 1-based)"),
                 arg, anchor[1L] + 1L, anchor[2L] + 1L), call. = FALSE)
  q
}

# Refuse a multi-cell array range that overlaps cells the sheet writes itself.
#
# libxlsxwriter pads a multi-cell array range with formatted zeroes, which the
# row loop would then overwrite (or which would overwrite the sheet's own
# values, depending on ordering).  Rather than emit a file whose array range is
# half data, require the range to spill into cells the sheet does not occupy.
.check_array_no_overlap <- function(q, anchor, df, header_offset, colname, k) {
  nrows <- nrow(df)
  ncols <- length(df)
  # unreachable from a written sheet: the check runs per cell, and a frame with
  # no rows or no columns has no cells
  if (nrows < 1L || ncols < 1L) return(invisible(NULL))  # nocov
  # the sheet's own data rectangle, 0-based
  first_row <- as.integer(header_offset)
  last_row  <- first_row + nrows - 1L
  r1 <- max(q[1L], first_row); r2 <- min(q[3L], last_row)
  c1 <- max(q[2L], 0L);        c2 <- min(q[4L], ncols - 1L)
  # likewise: the range is required to start at the formula's own cell, which
  # is inside the data, so the two rectangles always overlap
  if (r2 < r1 || c2 < c1) return(invisible(NULL))  # nocov
  ncell <- (as.numeric(r2 - r1) + 1) * (as.numeric(c2 - c1) + 1)
  if (ncell == 1 && r1 == anchor[1L] && c1 == anchor[2L])
    return(invisible(NULL))
  stop(sprintf(paste0(
    'the `array_range` on column "%s", cell %d covers %.0f cell(s) that the ',
    "sheet writes itself. A multi-cell array range must extend into cells the ",
    "sheet does not occupy, because the range is padded on write and the two ",
    "would overwrite each other. Either drop `array_range` (a single-cell ",
    "dynamic formula spills automatically in modern Excel) or move the range ",
    "beyond the data."), colname, k, ncell), call. = FALSE)
}

# Resolve one cell record's effective format to a registry id (0 = none),
# cascading workbook default_format, then the hyperlink/date default, then the
# cell's own format.
.resolve_cell_format_id <- function(rec, reg, props) {
  fmt <- rec$format
  val <- rec$value
  fm  <- rec$formula
  hl  <- rec$hyperlink
  formula_set   <- !(is.null(fm) || (length(fm) == 1L && is.na(fm)))
  hyperlink_set <- !(is.null(hl) || identical(hl, NA) ||
                     (is.character(hl) && length(hl) == 1L && is.na(hl)))
  base <- props$default_format
  if (hyperlink_set) {
    base <- merge_xl_format(base, props$hyperlink_format)
  } else if (!formula_set && !is.null(val) &&
             (inherits(val, "Date") || inherits(val, "POSIXct"))) {
    has_num <- is_xl_format(fmt) && !is.null(unclass(fmt)$num_format)
    if (!has_num)
      base <- merge_xl_format(base, if (inherits(val, "POSIXct"))
                                props$datetime_format else props$date_format)
  }
  .register_format(reg, merge_xl_format(base, fmt))
}

normalize_df <- function(df){
  if(nrow(df) > 1024^2){
    stop("the xlsx format does not support tables with 1M+ rows")
  }
  # Types to coerce to strings
  for(i in which(vapply(df, inherits, logical(1), c("factor", "hms")))){
    df[[i]] <- as.character(df[[i]])
  }
  for(i in which(vapply(df, function(x){is.integer(x) && inherits(x, "POSIXct")}, logical(1)))){
    # as.double() drops the class, so this dispatches to as.POSIXct.numeric(),
    # whose `origin` had no default before R 4.3 -- there it is an error,
    # "'origin' must be supplied".  1970-01-01 is what the later default uses,
    # so naming it gives the same instant on every version rather than a
    # value that depends on which R is running.
    df[[i]] <- as.POSIXct(as.double(df[[i]]), origin = "1970-01-01")
  }
  for(i in which(vapply(df, inherits, logical(1), "POSIXlt"))){
    df[[i]] <- as.POSIXct(df[[i]])
  }
  for(i in which(vapply(df, inherits, logical(1), "integer64"))){
    warning(sprintf("Coercing column %s from int64 to double", names(df)[i]), call. = FALSE)
    getNamespace("bit64")
    df[[i]] <- as.double(df[[i]])
  }
  .check_supported_columns(df)
  df
}

# Whether a column can be written at all.  This mirrors get_type() in
# src/write_xlsx.c: anything backed by a character, integer, double or logical
# vector is writable (a classed column such as Date, POSIXct or difftime is
# written from its underlying values), and xl_cell_general carries its own
# writer.  Everything else -- complex, raw, a plain list column -- has no
# representation in xlsx.
.is_supported_column <- function(x){
  if(inherits(x, "xl_cell_general"))
    return(TRUE)
  is.character(x) || is.integer(x) || is.double(x) || is.logical(x)
}

# Fail loudly on column types writexl cannot represent, rather than silently
# writing the underlying values (or nothing at all).
.check_supported_columns <- function(df){
  bad <- which(!vapply(df, .is_supported_column, logical(1)))
  if(!length(bad))
    return(invisible(NULL))
  detail <- vapply(bad, function(i){
    cls <- paste(class(df[[i]]), collapse = "/")
    ty <- typeof(df[[i]])
    sprintf("'%s' (column %d): %s", names(df)[i], i,
            if(identical(cls, ty)) cls else sprintf("%s (%s)", cls, ty))
  }, character(1))
  stop("writexl cannot write column(s) of these types:\n  ",
       paste(detail, collapse = "\n  "), call. = FALSE)
}

test_that("Performance is OK", {
  skip_if_not_installed("nycflights13")
  flights <- nycflights13::flights
  tmp <- writexl::write_xlsx(flights)
  out <- readxl::read_xlsx(tmp)
  unlink(tmp)
  # Excel has no time zones. Every datetime here shares one zone, so writexl
  # writes local wall-clock time rather than the UTC instant (see the time-zone
  # tests in test-types.R). Shift back to recover the instant; the offset is
  # per value because this column spans both EST and EDT.
  out$time_hour <- out$time_hour - as.POSIXlt(flights$time_hour)$gmtoff
  attr(out$time_hour, "tzone") <- attr(flights$time_hour, "tzone")
  # max_diffs caps the report: testthat uses max_diffs = Inf on CI, so a
  # mismatch in this 336k-row frame would otherwise print one line per row.
  expect_equal(out, flights, max_diffs = 10)
})

# Coverage of the bundled libxlsxwriter's option structs.
#
# A caller reaches most of libxlsxwriter through a handful of option structs:
# fill one in, hand it over, and every field left at zero is a feature writexl
# does not offer.  Nothing fails when a field is missed -- the file is written,
# Excel opens it, and the option simply is not there -- so the gap is invisible
# until someone reads the header.
#
# The chart gates elsewhere read the same header for *function* coverage.  This
# one reads it for *field* coverage, which is where the misses actually were: a
# chart font's typeface and rotation, a chartsheet's two protection options, an
# outline group's collapsed state, a table column's cached total.  Each was a
# documented field nothing in writexl ever assigned.
#
# A field is "reached" when src/write_xlsx.c assigns it.  UNUSED lists the ones
# that are deliberately not, each with the reason, so adding a field upstream
# fails here rather than passing unnoticed.
#
# This is a floor, not a proof: the check is textual, so a field whose name
# another struct also carries -- `name`, `color`, `size` -- reads as reached
# once any of them is set.  Telling them apart needs the type of the variable
# being assigned, which needs a C parser.  Removing each of the five fields
# above and re-running catches four; the fifth, lxw_chart_font.name, is pinned
# instead by test-xl_chart_format.R asserting typeface= in the chart XML.

lxw_header <- function(name) {
  for (p in c(file.path("../..", "src/libxlsxwriter/include/xlsxwriter", name),
              file.path("../../..", "src/libxlsxwriter/include/xlsxwriter", name)))
    if (file.exists(p)) return(paste(readLines(p, warn = FALSE), collapse = "\n"))
  NULL
}

# The documented fields of one option struct.  Undocumented members (no /** */)
# are libxlsxwriter's own bookkeeping and are not part of the API.
struct_fields <- function(header, struct) {
  txt <- lxw_header(header)
  if (is.null(txt)) return(NULL)
  m <- regmatches(txt, regexpr(sprintf("typedef struct %s \\{.*?\\n\\}", struct),
                               txt))
  if (!length(m)) return(NA_character_)
  documented <- regmatches(m, gregexpr("/\\*\\*.*?\\*/[^;]*;", m))[[1]]
  decl <- sub(".*\\*/\\s*", "", documented)
  trimws(gsub("^.*?([A-Za-z_][A-Za-z0-9_]*)\\s*(\\[[^]]*\\])?\\s*;$", "\\1",
              gsub("\n", " ", decl)))
}

# Fields writexl does not set, and why.  Anything not listed here must be
# assigned in src/write_xlsx.c.
UNUSED <- list(
  lxw_doc_properties = character(0),
  # writing to a buffer instead of a path is not write_xlsx()'s contract
  lxw_workbook_options = c("output_buffer", "output_buffer_size"),
  lxw_protection = character(0),
  lxw_image_options = character(0),
  lxw_chart_options = character(0),
  lxw_row_col_options = character(0),
  lxw_header_footer_options = character(0),
  lxw_data_validation = character(0),
  # the Excel 2010 data bar extension: solid bars and the negative-value axis
  lxw_conditional_format = c("data_bar_2010", "bar_negative_color_same",
                             "bar_negative_border_color_same"),
  lxw_table_options = character(0),
  lxw_table_column = character(0),
  lxw_comment_options = character(0),
  lxw_filter_rule = character(0),
  lxw_rich_string_tuple = character(0),
  # "rarely required, set to 0" in chart.h: a pitch family and character set
  # are Windows font-matching hints, and baseline is super/subscript, which a
  # chart title has no use for
  lxw_chart_font = c("pitch_family", "charset", "baseline"),
  lxw_chart_line = character(0),
  lxw_chart_fill = character(0),
  lxw_chart_pattern = character(0),
  lxw_chart_point = character(0),
  lxw_chart_layout = character(0)
)

STRUCT_HEADER <- c(
  lxw_doc_properties = "workbook.h", lxw_workbook_options = "workbook.h",
  lxw_protection = "worksheet.h", lxw_image_options = "worksheet.h",
  lxw_chart_options = "worksheet.h", lxw_row_col_options = "worksheet.h",
  lxw_header_footer_options = "worksheet.h",
  lxw_data_validation = "worksheet.h", lxw_conditional_format = "worksheet.h",
  lxw_table_options = "worksheet.h", lxw_table_column = "worksheet.h",
  lxw_comment_options = "worksheet.h", lxw_filter_rule = "worksheet.h",
  lxw_rich_string_tuple = "worksheet.h", lxw_chart_font = "chart.h",
  lxw_chart_line = "chart.h", lxw_chart_fill = "chart.h",
  lxw_chart_pattern = "chart.h", lxw_chart_point = "chart.h",
  lxw_chart_layout = "chart.h"
)

writexl_c <- function() {
  for (p in c("../../src/write_xlsx.c", "../../../src/write_xlsx.c"))
    if (file.exists(p)) return(paste(readLines(p, warn = FALSE), collapse = "\n"))
  NULL
}

test_that("every documented option-struct field is either set or listed unused", {
  skip_if(is.null(writexl_c()), "source tree not available")
  src <- writexl_c()
  missed <- character(0)
  for (struct in names(UNUSED)) {
    fields <- struct_fields(STRUCT_HEADER[[struct]], struct)
    expect_false(identical(fields, NA_character_),
                 info = paste("struct not found in the header:", struct))
    if (is.null(fields)) next
    for (f in setdiff(fields, UNUSED[[struct]])) {
      # reached by assignment (`x.f =`, `x->f =`) or by address (`&x.f`, which
      # is how the datetime fields and the metadata strings are filled).  Not
      # by the field's name appearing as a payload key: `"rotation"` is a
      # string writexl reads from R, and says nothing about where it is put.
      if (!grepl(sprintf("[.>]%s\\b", f), src))
        missed <- c(missed, paste0(struct, "$", f))
    }
  }
  expect_equal(missed, character(0))
})

test_that("nothing is listed unused that writexl has since started setting", {
  # keeps UNUSED honest: a stale entry would hide a field going missing again
  skip_if(is.null(writexl_c()), "source tree not available")
  for (struct in names(UNUSED)) {
    fields <- struct_fields(STRUCT_HEADER[[struct]], struct)
    if (is.null(fields) || identical(fields, NA_character_)) next
    expect_equal(as.character(setdiff(UNUSED[[struct]], fields)),
                 character(0),
                 info = paste("UNUSED names a field", struct,
                              "no longer has"))
  }
})

test_that("the pattern enums writexl bridges are the sizes the headers declare", {
  # .CHART_PATTERN maps cell pattern names onto chart pattern numbers, two
  # enums that share names but not values.  A member added to either shifts
  # nothing by itself, but it does mean the bridge needs another look.
  fmt <- lxw_header("format.h")
  cht <- lxw_header("chart.h")
  skip_if(is.null(fmt) || is.null(cht), "source tree not available")
  # one is a plain `enum name {`, the other a `typedef enum name {`
  count <- function(txt, enum, prefix) {
    m <- regmatches(txt, regexpr(sprintf("enum %s \\{.*?\\n\\}", enum), txt))
    if (!length(m)) return(NA_integer_)
    length(regmatches(m, gregexpr(prefix, m))[[1]])
  }
  expect_equal(count(fmt, "lxw_format_patterns", "LXW_PATTERN_"), 19L)
  expect_equal(count(cht, "lxw_chart_pattern_type", "LXW_CHART_PATTERN_"), 49L)
  # every bridged number is inside the chart enum, and none is the cell one
  expect_true(all(.CHART_PATTERN < 49L))
  expect_equal(length(.CHART_PATTERN), 11L)
})

test_that("C_set_tempdir rejects paths that are too long", {
  long_path <- strrep("a", 2048)
  expect_error(
    .Call(C_set_tempdir, long_path),
    "tempdir path too long"
  )
})

test_that("write_xlsx errors informatively for too many columns", {
  df_wide <- as.data.frame(matrix(1L, nrow = 1, ncol = 16385))
  expect_error(
    write_xlsx(df_wide),
    "too many columns"
  )
})

test_that("write_xlsx errors informatively for too many rows with header", {
  # 1048576 data rows + 1 header row = 1048577 total, exceeds xlsx limit of 1048576
  df_tall <- data.frame(x = integer(1048576))
  expect_error(
    write_xlsx(df_tall, col_names = TRUE),
    "too many rows"
  )
})

test_that("write_xlsx allows exactly LXW_ROW_MAX rows when col_names=FALSE", {
  skip_on_cran()
  df_max <- data.frame(x = integer(1048576))
  expect_true(file.exists(write_xlsx(df_max, col_names = FALSE)))
})

# =============================================================================
# Page setup: how a worksheet prints
# =============================================================================
#
# xl_page_setup() collects everything Excel's "Page Layout" ribbon controls:
# orientation, paper size, margins, scaling, centring, the print options, and
# the header/footer.  None of it affects the cells -- it only affects printing
# and the page-break preview -- so it is a plain bag of scalars that flattens
# into the per-sheet plan and is applied before any row is written.
# -----------------------------------------------------------------------------

# Excel's limit on page breaks per worksheet (LXW_BREAKS_MAX in worksheet.h).
.BREAKS_MAX <- 1023L

# Resolve `repeat_rows` / `repeat_cols` to the 0-based c(first, last) pair
# libxlsxwriter wants.
#
# These are *sheet* rows and columns, counted from 1 with the header row
# included -- unlike xl_row_spec(), which indexes data rows.  That is
# deliberate: repeating the header row on every printed page is the whole point
# of the feature, and data-row numbering could not name it.
.resolve_repeat <- function(x, arg, kind) {
  if (.is_unset(x)) return(NULL)
  if (is.numeric(x) && length(x) == 1L) {
    n <- .val_int(x, arg, min = 1)
    return(c(0L, as.integer(n) - 1L))
  }
  if (is.character(x) && length(x) == 1L) {
    ok <- if (kind == "row") grepl("^[$]?[0-9]+:[$]?[0-9]+$", x)
          else               grepl("^[$]?[A-Za-z]+:[$]?[A-Za-z]+$", x)
    if (!ok)
      stop(sprintf('`%s` must be a count, or a %s range like "%s"', arg,
                   if (kind == "row") "row" else "column",
                   if (kind == "row") "1:2" else "A:B"), call. = FALSE)
    q <- .xl_resolve_range(x, arg = arg, df = NULL, allow_cell = FALSE)
    return(if (kind == "row") c(q[1L], q[3L]) else c(q[2L], q[4L]))
  }
  stop(sprintf("`%s` must be a single count or range string", arg),
       call. = FALSE)
}

# Resolve a page-break vector to the 0-based positions libxlsxwriter wants.
#
# A break "at" row N means the new page starts at row N, which is 0-based N-1.
# Row/column 1 is therefore rejected: 0 is the array terminator in
# libxlsxwriter, so a break there would silently truncate the whole list rather
# than doing nothing visible.
.resolve_breaks <- function(x, arg, kind) {
  if (is.null(x) || (length(x) == 1L && !is.list(x) && is.na(x))) return(NULL)
  if (is.character(x) && kind == "col") x <- .col_letters_to_index(x) + 1L
  if (!is.numeric(x) || !length(x) || anyNA(x))
    stop(sprintf("`%s` must be a numeric vector of 1-based %s positions", arg,
                 if (kind == "row") "row" else "column"), call. = FALSE)
  v <- sort(unique(as.integer(x)))
  if (any(v < 2L))
    stop(sprintf(paste0("`%s` positions must be 2 or greater: a break at %s 1 ",
                        "would come before any content, and libxlsxwriter uses ",
                        "0 to terminate the break list"), arg,
                 if (kind == "row") "row" else "column"), call. = FALSE)
  if (length(v) > .BREAKS_MAX)
    stop(sprintf("`%s` may have at most %d breaks (got %d)", arg,
                 .BREAKS_MAX, length(v)), call. = FALSE)
  v - 1L
}

# Excel's paper types, from the table in worksheet.h.  Only the sizes worth
# naming are here; the ~40 envelope and specialist types are reachable by
# passing the index directly.
.LXW_PAPER <- c(
  "default"   = 0L,
  "letter"    = 1L,
  "tabloid"   = 3L,
  "ledger"    = 4L,
  "legal"     = 5L,
  "statement" = 6L,
  "executive" = 7L,
  "A3"        = 8L,
  "A4"        = 9L,
  "A5"        = 11L,
  "B4"        = 12L,
  "B5"        = 13L,
  "folio"     = 14L,
  "quarto"    = 15L
)

# libxlsxwriter caps a header/footer string at 255 characters
# (LXW_HEADER_FOOTER_MAX in worksheet.h).
.HEADER_FOOTER_MAX <- 255L

# Resolve a paper size to Excel's numeric type.  A name is looked up
# case-insensitively; an integer passes through so the unnamed sizes stay
# reachable.
.resolve_paper <- function(x) {
  if (.is_unset(x)) return(NULL)
  if (is.character(x) && length(x) == 1L) {
    i <- match(tolower(x), tolower(names(.LXW_PAPER)))
    if (is.na(i))
      stop("`paper` must be one of ", paste(names(.LXW_PAPER), collapse = ", "),
           ", or an Excel paper-type integer", call. = FALSE)
    return(unname(.LXW_PAPER[[i]]))
  }
  # libxlsxwriter's own valid range; outside it worksheet_set_paper() warns and
  # ignores the value, which would be a silent no-op from R
  .val_int(x, "paper", min = 0, max = 118)
}

# Validate a header or footer string.  Excel's &-codes are passed through
# untouched; the length and the image placeholders are our business.
.check_header_footer <- function(x, arg) {
  if (.is_unset(x)) return(NULL)
  s <- .val_str(x, arg)   # non-NULL: .is_unset() covered the NULL case above
  # libxlsxwriter measures this in UTF-8 characters, as nchar(type = "chars")
  # does, and errors above the limit
  if (nchar(s, type = "chars") > .HEADER_FOOTER_MAX)
    stop(sprintf("`%s` must be at most %d characters (got %d)", arg,
                 .HEADER_FOOTER_MAX, nchar(s, type = "chars")), call. = FALSE)
  s
}

# How many image placeholders a header/footer string carries.  Excel spells the
# same thing two ways, and libxlsxwriter counts both.
.hf_placeholders <- function(s) {
  if (is.null(s)) return(0L)
  n <- function(pat) length(gregexpr(pat, s, fixed = TRUE)[[1L]][
    gregexpr(pat, s, fixed = TRUE)[[1L]] > 0L])
  n("&G") + n("&[Picture]")
}

# The three positions a header/footer image can occupy.
.HF_POSITIONS <- c("left", "center", "right")

# Resolve a header/footer image argument to file paths.
#
# libxlsxwriter takes only filenames here -- lxw_header_footer_options has no
# buffer form (upstream issue 384) -- so an in-memory image is written to a
# temporary file, which lives as long as the session's tempdir.
.resolve_hf_images <- function(x, arg) {
  if (is.null(x) || (length(x) == 1L && !is.list(x) && is.na(x)))
    return(NULL)
  if (!is.list(x)) x <- as.list(x)
  nms <- names(x)
  if (is.null(nms) || any(!nzchar(nms)))
    stop(sprintf(paste0("`%s` must be named by position, e.g. ",
                        "list(left = \"logo.png\")"), arg), call. = FALSE)
  bad <- setdiff(nms, .HF_POSITIONS)
  if (length(bad))
    stop(sprintf("`%s`: unknown position(s): %s. Use %s.", arg,
                 paste(bad, collapse = ", "),
                 paste(.HF_POSITIONS, collapse = ", ")), call. = FALSE)
  if (anyDuplicated(nms))
    stop(sprintf("`%s`: each position may be given only once", arg),
         call. = FALSE)

  out <- list()
  for (pos in nms) {
    img <- x[[pos]]
    where <- sprintf("%s$%s", arg, pos)
    if (.is_raster_like(img)) img <- .raster_to_png(img, where)
    .check_image(img, where)
    out[[pos]] <- if (is.raw(img)) {
      p <- tempfile(fileext = ".img")
      writeBin(img, p)
      p
    } else {
      img
    }
  }
  out
}

# libxlsxwriter requires the placeholder count to equal the image count, and
# rejects a mismatch with a parameter-validation error that says nothing about
# which side is short.  Checking here says which.
.check_hf_images <- function(text, images, text_arg, image_arg) {
  n_ph <- .hf_placeholders(text)
  n_im <- length(images)
  if (n_ph == n_im) return(invisible(NULL))
  if (n_im > n_ph)
    stop(sprintf(paste0("`%s` gives %d image(s) but `%s` has %d image ",
                        "placeholder(s); add &G where each image should sit, ",
                        "e.g. \"&L&G\" for a left image"),
                 image_arg, n_im, text_arg, n_ph), call. = FALSE)
  stop(sprintf(paste0("`%s` has %d image placeholder(s) (&G or &[Picture]) but ",
                      "`%s` gives %d image(s)"),
               text_arg, n_ph, image_arg, n_im), call. = FALSE)
}

# Validate a header/footer margin.  libxlsxwriter applies it only when it is
# greater than zero (`if (options->margin > 0.0)`), so a zero would be silently
# ignored rather than doing what it says.
.check_hf_margin <- function(x, arg) {
  v <- .val_num(x, arg, min = 0)
  if (!is.null(v) && v <= 0)
    stop(sprintf("`%s` must be greater than 0 (Excel's default is 0.3)", arg),
         call. = FALSE)
  v
}

# Normalise a margins argument to the four named sides.
.resolve_margins <- function(x) {
  if (is.null(x) || (length(x) == 1L && is.na(x))) return(NULL)
  sides <- c("left", "right", "top", "bottom")
  if (!is.numeric(x))
    stop("`margins` must be a numeric vector of inches", call. = FALSE)
  if (is.null(names(x))) {
    if (length(x) == 1L) x <- stats::setNames(rep(x, 4L), sides)
    else if (length(x) == 4L) x <- stats::setNames(x, sides)
    else stop("`margins` must be one value, four values, or a named vector ",
              "with any of: ", paste(sides, collapse = ", "), call. = FALSE)
  }
  bad <- setdiff(names(x), sides)
  if (length(bad))
    stop("unknown `margins` element(s): ", paste(bad, collapse = ", "),
         call. = FALSE)
  if (anyNA(x) || any(x < 0))
    stop("`margins` must be non-negative", call. = FALSE)
  as.list(x)
}

# Normalise fit_to to c(width, height); 0 means "unconstrained in that
# direction", which is how Excel spells "fit to 1 page wide, any number tall".
.resolve_fit_to <- function(x) {
  if (is.null(x) || (length(x) == 1L && is.na(x))) return(NULL)
  if (!is.numeric(x) || !length(x) || length(x) > 2L || anyNA(x) || any(x < 0))
    stop("`fit_to` must be one or two non-negative numbers, c(width, height)",
         call. = FALSE)
  if (!is.null(names(x))) {
    bad <- setdiff(names(x), c("width", "height"))
    if (length(bad))
      stop("unknown `fit_to` element(s): ", paste(bad, collapse = ", "),
           call. = FALSE)
    w <- if (is.na(x["width"])) 1L else as.integer(x[["width"]])
    h <- if (is.na(x["height"])) 1L else as.integer(x[["height"]])
    return(c(width = w, height = h))
  }
  if (length(x) == 1L) x <- c(x, 1)
  c(width = as.integer(x[1L]), height = as.integer(x[2L]))
}

#' How a worksheet prints
#'
#' @description
#' `xl_page_setup()` collects Excel's page-layout settings for one worksheet:
#' orientation, paper size, margins, scaling, centring, the print options, and
#' the header and footer.  Pass it as `xl_sheet(page = )`.
#'
#' None of these affect the cell data --- they change only how the sheet prints
#' and how it looks in Excel's page-break preview.
#'
#' @param orientation `"portrait"` or `"landscape"`.
#' @param paper Paper size: a name (`"A4"`, `"letter"`, `"legal"`, `"tabloid"`,
#'   `"ledger"`, `"statement"`, `"executive"`, `"A3"`, `"A5"`, `"B4"`, `"B5"`,
#'   `"folio"`, `"quarto"`, `"default"`), or an Excel paper-type integer for the
#'   envelope and specialist sizes that have no name here.
#' @param margins Page margins in inches: a single number for all four sides,
#'   four numbers in the order left, right, top, bottom, or a named vector using
#'   any of `left`, `right`, `top`, `bottom`.
#' @param scale Print scaling as a percentage (10--400).  Ignored by Excel when
#'   `fit_to` is set.
#' @param fit_to Fit the printout to `c(width, height)` pages.  A `0` means
#'   "as many pages as needed in that direction", so `c(width = 1, height = 0)`
#'   is Excel's "fit all columns on one page".
#' @param center_horizontally,center_vertically Logical; centre the printed
#'   output on the page.
#' @param header,footer Header and footer text, at most 255 characters, using
#'   Excel's own codes: `&L`, `&C`, `&R` start the left, centre and right
#'   sections, `&P` is the page number, `&N` the page count, `&D` the date, `&A`
#'   the sheet name, and `&&` a literal ampersand.  For example
#'   `"&LQ1 report&RPage &P of &N"`.  Image placeholders (`&G` /
#'   `&[Picture]`) are not supported yet and are rejected.
#' @param header_image,footer_image Images to place in the header or footer,
#'   named by position: `list(left = , center = , right = )`.  Each may be a
#'   file path, a raw vector, or an in-memory image, exactly as [xl_image()]
#'   accepts.  Each image needs a matching `&G` placeholder in the
#'   corresponding section of `header`/`footer` --- `"&L&G"` puts one on the
#'   left --- and the counts must agree.
#' @param header_margin,footer_margin Header/footer margin in inches (Excel's
#'   default is 0.3).  Must be greater than 0.
#' @param page_view Logical; open the sheet in Excel's page-layout view rather
#'   than normal view.
#' @param first_page The page number to start numbering from.
#' @param across Logical; print pages left-to-right before top-to-bottom
#'   (Excel's "over, then down").
#' @param black_and_white Logical; print without colour.
#' @param row_col_headers Logical; print the row numbers and column letters.
#' @param print_area The range to print: an Excel range string such as
#'   `"A1:F50"`, or a `list(rows = , cols = )` spec naming data rows and
#'   columns, as elsewhere in writexl.
#' @param repeat_rows,repeat_cols Rows/columns to repeat at the top or left of
#'   every printed page.  Either a count (`repeat_rows = 1` repeats the first
#'   sheet row, which is the header when `col_names = TRUE`) or a range string
#'   (`"1:2"`, `"A:B"`).
#'
#'   Note these count *sheet* rows from 1 with the header included, unlike
#'   [xl_row_spec()], which indexes data rows --- repeating the header row is
#'   the usual reason to use this, and data-row numbering could not name it.
#' @param h_breaks,v_breaks Manual page breaks: 1-based sheet positions at which
#'   a new page starts, so `h_breaks = 21` breaks between rows 20 and 21.
#'   `v_breaks` also accepts column letters.  Positions must be 2 or greater,
#'   and at most 1023 breaks are allowed.  Excel ignores manual breaks when
#'   `fit_to` is set, which warns.
#' @return An `xl_page_setup` object.
#' @family worksheet layout
#' @seealso [xl_sheet], [write_xlsx]
#' @export
#' @examples
#' xl_page_setup(orientation = "landscape", paper = "A4",
#'               fit_to = c(width = 1, height = 0))
#'
#' # margins in inches, and a header with page numbers
#' xl_page_setup(margins = c(left = 1, right = 1),
#'               header = "&LQuarterly report&RPage &P of &N")
#'
#' df <- data.frame(x = 1:3)
#' tmp <- write_xlsx(list(Data = xl_sheet(df, page = xl_page_setup(
#'   orientation = "landscape", header = "&CDraft"
#' ))))
xl_page_setup <- function(orientation = NA, paper = NA, margins = NULL,
                          scale = NA, fit_to = NULL,
                          center_horizontally = NA, center_vertically = NA,
                          header = NA, footer = NA,
                          header_margin = NA, footer_margin = NA,
                          header_image = NULL, footer_image = NULL,
                          page_view = NA, first_page = NA, across = NA,
                          black_and_white = NA, row_col_headers = NA,
                          print_area = NULL, repeat_rows = NA,
                          repeat_cols = NA, h_breaks = NULL, v_breaks = NULL) {
  hdr_images <- .resolve_hf_images(header_image, "header_image")
  ftr_images <- .resolve_hf_images(footer_image, "footer_image")
  .check_hf_images(.check_header_footer(header, "header"), hdr_images,
                   "header", "header_image")
  .check_hf_images(.check_header_footer(footer, "footer"), ftr_images,
                   "footer", "footer_image")
  if (!is.null(h_breaks) || !is.null(v_breaks)) {
    fit <- .resolve_fit_to(fit_to)
    if (!is.null(fit))
      warning("Excel ignores manual page breaks when `fit_to` is set",
              call. = FALSE)
  }
  out <- .drop_null(list(
    orientation         = .val_enum(orientation, c("portrait", "landscape"),
                                    "orientation"),
    paper               = .resolve_paper(paper),
    margins             = .resolve_margins(margins),
    scale               = .val_int(scale, "scale", min = 10, max = 400),
    fit_to              = .resolve_fit_to(fit_to),
    center_horizontally = .val_flag(center_horizontally, "center_horizontally"),
    center_vertically   = .val_flag(center_vertically, "center_vertically"),
    header              = .check_header_footer(header, "header"),
    footer              = .check_header_footer(footer, "footer"),
    header_margin       = .check_hf_margin(header_margin, "header_margin"),
    footer_margin       = .check_hf_margin(footer_margin, "footer_margin"),
    header_image        = hdr_images,
    footer_image        = ftr_images,
    page_view           = .val_flag(page_view, "page_view"),
    first_page          = .val_int(first_page, "first_page", min = 0),
    across              = .val_flag(across, "across"),
    black_and_white     = .val_flag(black_and_white, "black_and_white"),
    row_col_headers     = .val_flag(row_col_headers, "row_col_headers"),
    # the print area is resolved later, when the sheet's data frame is known
    print_area          = print_area,
    repeat_rows         = .resolve_repeat(repeat_rows, "repeat_rows", "row"),
    repeat_cols         = .resolve_repeat(repeat_cols, "repeat_cols", "col"),
    h_breaks            = .resolve_breaks(h_breaks, "h_breaks", "row"),
    v_breaks            = .resolve_breaks(v_breaks, "v_breaks", "col")
  ))
  structure(out, class = "xl_page_setup")
}

#' @export
print.xl_page_setup <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_page_setup: %d setting%s>\n", length(p),
              if (length(p) == 1L) "" else "s"))
  for (k in names(p))
    cat(sprintf("  %s: %s\n", k, paste(format(unlist(p[[k]])), collapse = ", ")))
  invisible(x)
}

# Flatten a page setup into the flat scalars C applies.  Margins and fit_to are
# spread into individual keys so the C side only ever reads scalars.
.page_setup_payload <- function(page, df = NULL, header_offset = 1L) {
  if (is.null(page)) return(NULL)
  if (!inherits(page, "xl_page_setup"))
    stop("`page` must be an xl_page_setup object", call. = FALSE)
  p <- unclass(page)
  out <- list()
  # The print area is the one setting that needs the sheet, so it is resolved
  # here rather than in the constructor.
  if (!is.null(p$print_area))
    out$print_area <- .xl_resolve_range(p$print_area, arg = "print_area",
                                        df = df, header_offset = header_offset,
                                        allow_cell = FALSE)
  for (k in c("repeat_rows", "repeat_cols", "h_breaks", "v_breaks"))
    if (!is.null(p[[k]])) out[[k]] <- as.integer(p[[k]])
  if (!is.null(p$orientation))
    out$landscape <- as.integer(identical(p$orientation, "landscape"))
  if (!is.null(p$paper))  out$paper <- as.integer(p$paper)
  if (!is.null(p$scale))  out$scale <- as.integer(p$scale)
  if (!is.null(p$margins))
    for (side in names(p$margins))
      out[[paste0("margin_", side)]] <- as.numeric(p$margins[[side]])
  if (!is.null(p$fit_to)) {
    out$fit_width  <- as.integer(p$fit_to[["width"]])
    out$fit_height <- as.integer(p$fit_to[["height"]])
  }
  for (k in c("center_horizontally", "center_vertically", "page_view",
              "across", "black_and_white", "row_col_headers"))
    if (!is.null(p[[k]])) out[[k]] <- as.integer(isTRUE(p[[k]]))
  if (!is.null(p$first_page)) out$first_page <- as.integer(p$first_page)
  if (!is.null(p$header)) out$header <- p$header
  if (!is.null(p$footer)) out$footer <- p$footer
  if (!is.null(p$header_margin)) out$header_margin <- as.numeric(p$header_margin)
  if (!is.null(p$footer_margin)) out$footer_margin <- as.numeric(p$footer_margin)
  for (side in c("header", "footer"))
    for (pos in .HF_POSITIONS) {
      f <- p[[paste0(side, "_image")]][[pos]]
      if (!is.null(f))
        out[[sprintf("%s_image_%s", side, pos)]] <- f
    }
  if (length(out)) out else NULL
}

# =============================================================================
# Split panes: converting a cell reference into libxlsxwriter's units
# =============================================================================
#
# worksheet_split_panes() does not take a row and a column.  Its two arguments
# are distances, measured in the units Excel uses for row height and column
# width, and the two units differ from each other:
#
#     worksheet_split_panes(worksheet1, 15, 0);    // First row.
#     worksheet_split_panes(worksheet2, 0,  8.43); // First column.
#
# So `vertical` is a number of row-heights and positions the split *between
# rows*, while `horizontal` is a number of column-widths and positions it
# between columns.  Passing 1 for "one row" would put the split a fifteenth of
# the way down the first row, silently.
#
# writexl therefore takes a cell reference, as `freeze` does, and converts:
# the distance above the split is the summed height of the rows before it, and
# the distance to its left is the summed width of the columns before it.  Those
# use the sheet's real geometry -- header row height, xl_row_spec() heights,
# xl_col_spec() widths and auto_colwidth results -- rather than assuming the
# defaults, so a split lands where it was asked for on a sheet whose rows or
# columns have been resized.
#
# A raw list(vertical =, horizontal =) is still accepted for callers who want
# to specify the units directly.
# -----------------------------------------------------------------------------

# Excel's defaults, and the values libxlsxwriter's own examples use.
.DEFAULT_ROW_HEIGHT <- 15
.DEFAULT_COL_WIDTH  <- 8.43

# The height of one 0-based sheet row: an explicit xl_row_spec() height, else
# the sheet default, else Excel's.
.row_height_at <- function(i, header_offset, props, default_row_height,
                           row_row, row_height) {
  k <- which(row_row == i)
  if (length(k) && !is.na(row_height[k[1L]])) return(row_height[k[1L]])
  if (header_offset > 0L && i == 0L && !is.null(props$header_row_height))
    return(as.numeric(props$header_row_height))
  if (!is.na(default_row_height)) return(as.numeric(default_row_height))
  .DEFAULT_ROW_HEIGHT
}

# The width of one 1-based column.
.col_width_at <- function(j, col_width) {
  if (j <= length(col_width) && !is.na(col_width[j])) col_width[j]
  else .DEFAULT_COL_WIDTH
}

.resolve_split <- function(spec, df, header_offset, props, default_row_height,
                           row_row, row_height, col_width) {
  # raw units, for callers who want to place the split exactly
  if (is.list(spec) &&
      (!is.null(spec[["vertical"]]) || !is.null(spec[["horizontal"]]))) {
    bad <- setdiff(names(spec), c("vertical", "horizontal"))
    if (length(bad))
      stop("unknown `split` element(s): ", paste(bad, collapse = ", "),
           call. = FALSE)
    v <- if (is.null(spec$vertical)) 0 else as.numeric(spec$vertical)
    h <- if (is.null(spec$horizontal)) 0 else as.numeric(spec$horizontal)
    if (anyNA(c(v, h)) || any(c(v, h) < 0))
      stop("`split` units must be non-negative numbers", call. = FALSE)
    return(c(vertical = v, horizontal = h))
  }

  q <- .xl_resolve_range(spec, arg = "split", df = df,
                         header_offset = header_offset, allow_cell = TRUE)
  n_rows <- q[1L]   # 0-based row index == how many rows sit above the split
  n_cols <- q[2L]

  vertical <- 0
  if (n_rows > 0L)
    vertical <- sum(vapply(seq_len(n_rows) - 1L, .row_height_at, numeric(1),
                           header_offset = header_offset, props = props,
                           default_row_height = default_row_height,
                           row_row = row_row, row_height = row_height))
  horizontal <- 0
  if (n_cols > 0L)
    horizontal <- sum(vapply(seq_len(n_cols), .col_width_at, numeric(1),
                             col_width = col_width))

  c(vertical = vertical, horizontal = horizontal)
}

# =============================================================================
# Worksheet tables
# =============================================================================
#
# A table is a named, styled range that Excel treats as a unit: banded rows, a
# filter dropdown in the header, structured references in formulas, and an
# optional total row.
#
# The thing to know before reading further is that worksheet_add_table() writes
# cells of its own.  _write_table_column_data() in worksheet.c writes each
# column's header string into the header row, and _set_default_table_columns()
# defaults those strings to "Column 1", "Column 2", ...  A table added over a
# writexl data frame would therefore replace the header row with generic
# captions.
#
# It is worse than losing the captions: the table part records the column names
# independently of the cells, and Excel treats a mismatch between the two as a
# corrupt file.  So xl_table() always sends the data frame's column names
# through as the headers -- the generic captions are not reachable by accident.
#
# That also settles the ordering.  Tables are applied *after* the sheet's rows,
# beside merges, so the table's header and total-row writes land last; applying
# them first would let the row loop overwrite them and lose header_format.
# Writing over already-written rows is impossible while libxlsxwriter streams,
# and worksheet_add_table() refuses outright in that mode, so any table turns
# constant memory off -- see .resolve_constant_memory().
# -----------------------------------------------------------------------------

# Style types, and the style numbers each one allows.  libxlsxwriter only warns
# on a bad number and silently substitutes Medium 9, so the checking is ours.
.LXW_TABLE_STYLE_TYPE <- c(none = 0L, light = 1L, medium = 2L, dark = 3L)
.TABLE_STYLE_RANGE <- list(light = c(0L, 21L), medium = c(1L, 28L),
                           dark = c(1L, 11L))

# The total-row functions.  The enum is sparse (101-105, 107, 109, 110), which
# is a good reason not to make callers know it.
.LXW_TABLE_FUNCTION <- c(
  average = 101L, count_nums = 102L, count = 103L, max = 104L, min = 105L,
  std_dev = 107L, sum = 109L, var = 110L
)

# Parse "medium 9" / "light 21" / "none" into the type and number C wants.
.parse_table_style <- function(style) {
  if (is.null(style)) return(list(type = 2L, number = 9L))
  if (!is.character(style) || length(style) != 1L || is.na(style))
    stop('`style` must be a single string like "medium 9" or "none"',
         call. = FALSE)
  s <- tolower(trimws(style))
  # "none" is Light 0 in libxlsxwriter's numbering
  if (identical(s, "none")) return(list(type = 1L, number = 0L))
  parts <- strsplit(s, "[[:space:]]+")[[1L]]
  if (length(parts) != 2L)
    stop(sprintf(paste0('`style` must be "none" or a type and number like ',
                        '"medium 9" (got "%s")'), style), call. = FALSE)
  type <- parts[1L]
  if (!type %in% names(.TABLE_STYLE_RANGE))
    stop(sprintf('`style` type must be "light", "medium" or "dark" (got "%s")',
                 parts[1L]), call. = FALSE)
  num <- suppressWarnings(as.integer(parts[2L]))
  rng <- .TABLE_STYLE_RANGE[[type]]
  if (is.na(num) || num < rng[1L] || num > rng[2L])
    stop(sprintf("`style`: %s styles are numbered %d to %d (got \"%s\")",
                 type, rng[1L], rng[2L], parts[2L]), call. = FALSE)
  list(type = unname(.LXW_TABLE_STYLE_TYPE[[type]]), number = num)
}

#' Describe a column of a worksheet table
#'
#' @description
#' `xl_table_column()` overrides what [xl_table()] does with one column: its
#' header caption, a formula filling the column, what its total-row cell shows,
#' and the formats applied to the header and the data cells.
#'
#' Only the columns you name need an entry; the rest take the data frame's
#' column name as their header and are left otherwise alone.
#'
#' @param col The column: a name or a 1-based position within the table's range.
#' @param header The header caption.  Defaults to the data frame's column name
#'   --- which is also what makes a table safe to add, since Excel rejects a
#'   file whose table definition and header cells disagree.
#' @param formula A formula filling every data cell of the column, usually with
#'   a structured reference such as `"=SUM(Sales[@[Q1]:[Q4]])"`.  Because those
#'   name the table, see the `name` argument of [xl_table()].
#' @param total The function shown in this column's total-row cell: one of
#'   `"sum"`, `"average"`, `"count"`, `"count_nums"`, `"max"`, `"min"`,
#'   `"std_dev"` or `"var"`.  Needs `total_row = TRUE` on the table.
#' @param total_label A string for the total-row cell instead of a function ---
#'   typically `"Total"` under the first column.
#' @param total_value The number the total-row cell already holds. Excel
#'   recalculates `total` on open and ignores this, but a reader that does not
#'   evaluate formulas --- [readxl::read_xlsx()] among them --- shows what is
#'   cached, which without this is nothing.
#' @param format An [xl_format] applied to the column's data cells.
#' @param header_format An [xl_format] applied to the column's header cell.
#' @return An `xl_table_column` object.
#' @family worksheet features
#' @seealso [xl_table], [xl_sheet]
#' @export
#' @examples
#' xl_table_column("qty", total = "sum")
#' xl_table_column("fruit", total_label = "Total")
#' xl_table_column("margin", formula = "=Sales[@revenue] * 0.3")
xl_table_column <- function(col, header = NULL, formula = NULL, total = NULL,
                            total_label = NULL, total_value = NULL,
                            format = NULL, header_format = NULL) {
  if (missing(col) || is.null(col))
    stop("`col` must name or index the table column", call. = FALSE)
  str1 <- function(x, arg) {
    if (is.null(x)) return(NULL)
    if (!is.character(x) || length(x) != 1L || is.na(x))
      stop(sprintf("`%s` must be a single non-NA string", arg), call. = FALSE)
    x
  }
  fmt <- function(x, arg) {
    if (is.null(x)) return(NULL)
    if (!is_xl_format(x))
      stop(sprintf("`%s` must be an xl_format object", arg), call. = FALSE)
    x
  }
  formula <- str1(formula, "formula")
  if (!is.null(formula) && !startsWith(formula, "="))
    stop("`formula` must start with '='", call. = FALSE)
  tot <- .val_enum(total, names(.LXW_TABLE_FUNCTION), "total")
  if (!is.null(tot) && !is.null(total_label))
    stop("give either `total` or `total_label`, not both", call. = FALSE)
  if (!is.null(total_value)) {
    if (!is.numeric(total_value) || length(total_value) != 1L ||
        is.na(total_value))
      stop("`total_value` must be a single non-NA number", call. = FALSE)
    if (is.null(tot))
      stop("`total_value` caches the result of `total`, so give `total` too",
           call. = FALSE)
    total_value <- as.numeric(total_value)
  }
  structure(.drop_null(list(
    col = col, header = str1(header, "header"), formula = formula,
    total = tot, total_label = str1(total_label, "total_label"),
    total_value = total_value,
    format = fmt(format, "format"),
    header_format = fmt(header_format, "header_format")
  )), class = "xl_table_column")
}

#' @export
print.xl_table_column <- function(x, ...) {
  p <- unclass(x)
  # `[[` not `$`: see .table_columns_payload() -- `$total` would partial-match
  # `total_label` on a column that only carries a label
  cat(sprintf("<xl_table_column: %s%s%s>\n", p[["col"]],
              if (!is.null(p[["formula"]])) " formula" else "",
              if (!is.null(p[["total"]])) paste0(" total=", p[["total"]]) else ""))
  invisible(x)
}

#' Add a worksheet table
#'
#' @description
#' `xl_table()` turns a range into an Excel table: a named, styled block with
#' banded rows, a filter dropdown in its header, and an optional total row.
#' Pass one or a list of them as `xl_sheet(table = )`.
#'
#' Column headers default to the data frame's column names.  That is not just a
#' convenience: Excel records a table's column names separately from the header
#' cells and rejects a file where the two disagree, so the default is what makes
#' a table safe to add at all.
#'
#' @section Table names:
#' Excel requires table names to be unique across the workbook, and formulas
#' refer to a table by name.  `name = NULL` (the default) has writexl generate a
#' unique name from the sheet name and write it explicitly, so it cannot shift
#' when another table is added elsewhere.  Give a string to choose your own; it
#' is validated and checked for collisions.  Give `NA` to write no name at all
#' and let Excel assign `Table1`, `Table2`, ... by insertion order --- which
#' makes any formula naming the table fragile, so that combination warns.
#'
#' @section The total row:
#' A total row is written by libxlsxwriter below the data, and the sheet's row
#' plan does not know about it: `auto_colwidth` does not measure it, and an
#' [xl_row_spec()] aimed at that row will fight it.  It has no cells of its own
#' until a column gives a `total` or `total_label`.
#'
#' @param range The cells the table covers, as an Excel range string or a
#'   `list(rows = , cols = )` spec.  Defaults to the sheet's used range,
#'   including the header row and, with `total_row = TRUE`, one row below.
#' @param name The table's name.  See "Table names".
#' @param style The table style: `"none"`, or a type and number such as
#'   `"medium 9"` (the default), `"light 21"` or `"dark 11"`.  Light styles are
#'   numbered 0--21, medium 1--28 and dark 1--11.
#' @param header_row Show the header row.  Turning it off also removes the
#'   filter dropdown, as it does in Excel.
#' @param autofilter Show the filter dropdown in the header row.
#' @param banded_rows,banded_columns Alternating row / column shading.
#' @param first_column,last_column Highlight the first / last column.
#' @param total_row Add a total row below the data.  See "The total row".
#' @param columns One [xl_table_column()], or a list of them, overriding
#'   individual columns.
#' @return An `xl_table` object.
#' @family worksheet features
#' @seealso [xl_table_column], [xl_sheet]
#' @export
#' @examples
#' xl_table(style = "light 9")
#' xl_table(name = "Sales", total_row = TRUE,
#'          columns = list(xl_table_column("fruit", total_label = "Total"),
#'                         xl_table_column("qty", total = "sum")))
xl_table <- function(range = NULL, name = NULL, style = "medium 9",
                     header_row = TRUE, autofilter = TRUE, banded_rows = TRUE,
                     banded_columns = FALSE, first_column = FALSE,
                     last_column = FALSE, total_row = FALSE, columns = NULL) {
  flag <- function(x, arg) {
    if (!is.logical(x) || length(x) != 1L || is.na(x))
      stop(sprintf("`%s` must be TRUE or FALSE", arg), call. = FALSE)
    isTRUE(x)
  }
  # NA means "write no name"; NULL means "generate one"
  if (!is.null(name)) {
    if (length(name) != 1L)
      stop("`name` must be a single string, NULL or NA", call. = FALSE)
    if (!is.na(name)) .check_table_name(name)
  }
  sty <- .parse_table_style(style)
  cols <- .table_column_list(columns)

  header <- flag(header_row, "header_row")
  af <- flag(autofilter, "autofilter")
  # libxlsxwriter forces this itself; mirroring it keeps the object honest
  if (!header) af <- FALSE
  tot <- flag(total_row, "total_row")

  if (!tot) {
    # `[[` not `$`: `$total` partial-matches `total_label` once the NULLs are
    # dropped, so the two conditions would not be independent
    bad <- vapply(cols, function(cc)
      !is.null(cc[["total"]]) || !is.null(cc[["total_label"]]), logical(1))
    if (any(bad))
      stop(sprintf(paste0("column %s gives a total, but the table has no total ",
                          "row; set `total_row = TRUE`"),
                   paste(vapply(cols[bad], function(cc) sprintf('"%s"', cc$col),
                                character(1)), collapse = ", ")),
           call. = FALSE)
  }
  if (!is.null(name) && length(name) == 1L && is.na(name) &&
      any(vapply(cols, function(cc) !is.null(cc$formula), logical(1))))
    warning("`name = NA` leaves Excel to name the table (Table1, Table2, ...), ",
            "but a column formula refers to the table by name, so it will ",
            "break if the numbering shifts. Give `name` a string instead.",
            call. = FALSE)

  structure(.drop_null(list(
    range = range, name = name,
    style_type = sty$type, style_number = sty$number,
    header_row = header, autofilter = af, banded_rows = flag(banded_rows, "banded_rows"),
    banded_columns = flag(banded_columns, "banded_columns"),
    first_column = flag(first_column, "first_column"),
    last_column = flag(last_column, "last_column"),
    total_row = tot,
    columns = if (length(cols)) cols else NULL
  )), class = "xl_table")
}

#' @export
print.xl_table <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_table: %s%s%s>\n",
              if (is.null(p$name)) "unnamed" else
                if (is.na(p$name)) "named by Excel" else p$name,
              if (isTRUE(p$total_row)) ", total row" else "",
              sprintf(", %d column override(s)", length(p$columns))))
  invisible(x)
}

# Normalise one table or a list of them.
.table_list <- function(table, arg = "table") {
  if (is.null(table)) return(list())
  ts <- if (inherits(table, "xl_table")) list(table) else table
  if (!is.list(ts))
    stop(sprintf("`%s` must be an xl_table object or a list of them", arg),
         call. = FALSE)
  for (i in seq_along(ts))
    if (!inherits(ts[[i]], "xl_table"))
      stop(sprintf("`%s[[%d]]` must be an xl_table object", arg, i),
           call. = FALSE)
  ts
}

# --- Table names, resolved across the whole workbook -------------------------
#
# Excel requires table names to be unique workbook-wide, and libxlsxwriter never
# checks that -- two tables sharing a name reach the file and Excel offers to
# repair it.  Uniqueness therefore cannot be settled inside xl_table(), which
# sees one table; it is settled here, where the sheets are assembled, next to
# the sheet-name dedupe that has the same shape.

# Turn any string into something Excel will accept as a table name.
.sanitize_table_name <- function(x) {
  chars <- strsplit(x, "", fixed = TRUE)[[1L]]
  chars[chars %in% .TABLE_NAME_BAD_CHARS] <- "_"
  out <- paste(chars, collapse = "")
  if (!nzchar(out)) out <- "Table"
  if (grepl("^[0-9]", out)) out <- paste0("T_", out)
  # a bare C/R or a cell-reference lookalike is reserved; suffixing clears both
  if (toupper(out) %in% c("C", "R") || grepl("^[A-Za-z]{1,3}[0-9]+$", out))
    out <- paste0(out, "_tbl")
  if (nchar(out) > 250L) out <- substr(out, 1L, 250L)
  out
}

# Append _2, _3, ... until the name is free.  Excel compares table names
# case-insensitively, so "Sales" collides with "sales" and the comparison has
# to be folded on both sides -- folding only the stored set silently lets every
# generated name through.
.unique_table_name <- function(base, taken) {
  taken <- tolower(taken)
  if (!tolower(base) %in% taken) return(base)
  i <- 2L
  repeat {
    cand <- paste0(base, "_", i)
    if (!tolower(cand) %in% taken) return(cand)
    i <- i + 1L
  }
}

# Resolve every table's name across the workbook.  Returns one character vector
# per sheet, NA meaning "write no name and let Excel number it".
.resolve_table_names <- function(elems, sheet_names) {
  tabs <- lapply(elems, function(el)
    if (inherits(el, "xl_sheet")) .table_list(el$table) else list())
  if (!length(unlist(tabs, recursive = FALSE))) return(rep(list(character(0)), length(elems)))

  # explicit names first: they are fixed points, so a collision between two of
  # them cannot be resolved by renaming either one
  explicit <- list()
  for (i in seq_along(tabs)) {
    for (k in seq_along(tabs[[i]])) {
      nm <- unclass(tabs[[i]][[k]])$name
      if (is.null(nm) || is.na(nm)) next
      prev <- explicit[[tolower(nm)]]
      if (!is.null(prev))
        stop(sprintf(paste0("two tables are both named \"%s\" (on sheet \"%s\" ",
                            "and sheet \"%s\"); Excel requires table names to ",
                            "be unique across the workbook"),
                     nm, prev, sheet_names[i]), call. = FALSE)
      explicit[[tolower(nm)]] <- sheet_names[i]
    }
  }

  taken <- names(explicit)
  out <- vector("list", length(elems))
  for (i in seq_along(tabs)) {
    nms <- character(length(tabs[[i]]))
    base <- .sanitize_table_name(sheet_names[i])
    for (k in seq_along(tabs[[i]])) {
      nm <- unclass(tabs[[i]][[k]])$name
      if (is.null(nm)) {
        gen <- .unique_table_name(base, taken)
        taken <- c(taken, tolower(gen))
        nms[k] <- gen
      } else if (is.na(nm)) {
        nms[k] <- NA_character_
      } else {
        nms[k] <- nm
      }
    }
    out[[i]] <- nms
  }
  out
}

# The per-column formats a sheet's tables set, indexed by data frame column.
#
# These have to reach the sheet's own column plan.  libxlsxwriter applies a
# table column's `format` only to cells it writes itself -- the total cell in
# _write_column_function() and a calculated column in _write_column_formula() --
# and never to data already written by the row loop.  Left to it, setting a
# format on an ordinary column would do nothing at all, silently.
#
# This is by design, not a bug: the other ports take the table's data as a
# parameter, so the library writes the cells and can format them, while
# libxlsxwriter's caller writes them first (jmcnamara/libxlsxwriter#520).  The
# recommended fix is exactly this -- put the format on the column.
#
# BOTH paths are needed and neither is redundant.  The column plan formats the
# data cells; the format_id passed through .table_columns_payload() becomes the
# table part's dataDxfId, which upstream says is required "for strict
# correctness with Excel".  A test asserts both survive, because after the
# column plan lands the dataDxfId looks like the removable one.
.table_column_formats <- function(el, df) {
  out <- vector("list", length(df))
  if (!inherits(el, "xl_sheet")) return(out)
  for (tb in .table_list(el$table))
    for (cc in unclass(tb)[["columns"]]) {
      q <- unclass(cc)
      if (is.null(q[["format"]])) next
      for (idx in .resolve_col_index(q[["col"]], names(df), "table column"))
        out[[idx]] <- q[["format"]]
    }
  out
}

# --- Conflicts -----------------------------------------------------------
#
# Excel drops or repairs each of these rather than reporting them, and
# libxlsxwriter passes them all through, so they are refused here.  Each
# message names the table's range so the offending pair can be found.

# Do two 0-based c(first_row, first_col, last_row, last_col) quads overlap?
.ranges_overlap <- function(a, b) {
  a[1L] <= b[3L] && b[1L] <= a[3L] && a[2L] <= b[4L] && b[2L] <= a[4L]
}

.range_a1 <- function(r) {
  cell <- function(row, col) {
    letters_of <- function(i) {
      out <- ""
      repeat {
        out <- paste0(LETTERS[(i %% 26L) + 1L], out)
        i <- i %/% 26L - 1L
        if (i < 0L) break
      }
      out
    }
    paste0(letters_of(col), row + 1L)
  }
  paste0(cell(r[1L], r[2L]), ":", cell(r[3L], r[4L]))
}

# `tables` is the resolved payload list; the rest come off the sheet element.
.check_table_conflicts <- function(el, df, header_offset, tables) {
  if (!length(tables)) return(invisible(NULL))
  ranges <- lapply(tables, function(t) t$range)

  refuse <- function(rng, what)
    stop(sprintf("the table at %s %s", .range_a1(rng), what), call. = FALSE)

  for (rng in ranges) {
    # a table brings its own autofilter; a second one over the same cells is
    # what Excel repairs the file over
    af <- el$autofilter
    if (isTRUE(af) || is.character(af)) {
      afr <- if (is.character(af))
        .parse_range(af, "autofilter", df, header_offset)
      else c(0L, 0L, (nrow(df) - 1L) + header_offset, length(df) - 1L)
      if (.ranges_overlap(rng, afr))
        # the remedy is the sheet's autofilter, not the table's: turning the
        # table's off leaves the sheet-level one over the same cells
        refuse(rng, paste0("overlaps `xl_sheet(autofilter =)`; a table brings ",
                           "its own filter dropdown, and Excel allows only one ",
                           "over a range. Drop the sheet's `autofilter`, or ",
                           "point it at cells outside the table."))
    }
    # libxlsxwriter: "Filter conditions within the table are not supported"
    for (f in .filter_list(el$filter)) {
      idx <- .resolve_col_index(unclass(f)$col, names(df), "filter col") - 1L
      if (idx >= rng[2L] && idx <= rng[4L])
        refuse(rng, sprintf(paste0("covers column \"%s\", which `filter` also ",
                                   "filters; filter criteria inside a table ",
                                   "are not supported"),
                            names(df)[idx + 1L]))
    }
    # Excel forbids merged cells inside a table
    for (m in .merge_list(el$merge)) {
      mr <- .xl_resolve_range(unclass(m)$range, arg = "merge range", df = df,
                              header_offset = header_offset, allow_cell = FALSE)
      if (.ranges_overlap(rng, mr))
        refuse(rng, sprintf(paste0("overlaps the merge at %s; Excel does not ",
                                   "allow merged cells inside a table"),
                            .range_a1(mr)))
    }
  }
  # two tables may not overlap each other either
  if (length(ranges) > 1L)
    for (i in seq_len(length(ranges) - 1L))
      for (j in seq(i + 1L, length(ranges)))
        if (.ranges_overlap(ranges[[i]], ranges[[j]]))
          stop(sprintf("the tables at %s and %s overlap",
                       .range_a1(ranges[[i]]), .range_a1(ranges[[j]])),
               call. = FALSE)
  invisible(NULL)
}

# --- The C payload -----------------------------------------------------------
#
# Every column of the table's range gets an entry, whether or not the caller
# named it, because _write_table_column_data() writes a header string for each
# one and defaults it to "Column 1", "Column 2", ...  Sending the data frame's
# own names through is what keeps the table definition and the header cells in
# agreement, which is the difference between a file Excel opens and one it
# offers to repair.
.resolve_tables <- function(el, df, reg, header_offset, props, table_names) {
  if (!inherits(el, "xl_sheet")) return(list())
  ts <- .table_list(el$table)
  if (!length(ts)) return(list())
  if (nrow(df) < 1L)
    stop("a table needs at least one data row", call. = FALSE)

  out <- lapply(seq_along(ts), function(i) {
    p <- unclass(ts[[i]])
    rng <- if (is.null(p$range)) {
      last_row <- (nrow(df) - 1L) + header_offset + as.integer(p$total_row)
      c(0L, 0L, as.integer(last_row), length(df) - 1L)
    } else {
      .xl_resolve_range(p$range, arg = "table range", df = df,
                        header_offset = header_offset, allow_cell = FALSE)
    }
    # libxlsxwriter counts the header row but not the total row as data
    non_header <- rng[3L] - rng[1L] - as.integer(p$total_row) +
      as.integer(!p$header_row)
    if (non_header < 1L)
      stop("a table needs at least one row that is not the header or the ",
           "total row", call. = FALSE)

    # A column past the data has no name to take, and the generic "Column N"
    # caption is exactly what this design exists to make unreachable.
    if (rng[4L] > length(df) - 1L)
      stop(sprintf(paste0("the table at %s reaches column %d, but the data ",
                          "frame has %d; writexl has no header name for the ",
                          "extra column(s)"),
                   .range_a1(rng), rng[4L] + 1L, length(df)), call. = FALSE)
    # worksheet_add_table() writes its header captions into the range's first
    # row.  If that is not the sheet's header row it lands on data, silently
    # replacing values with column names.
    if (p$header_row && header_offset > 0L && rng[1L] != 0L)
      stop(sprintf(paste0("the table at %s has a header row, but its first row ",
                          "is not the sheet's header row; the table's captions ",
                          "would overwrite data. Start the table at row 1, or ",
                          "set `header_row = FALSE`."),
                   .range_a1(rng)), call. = FALSE)

    cols <- .table_columns_payload(p, df, rng, reg, props)
    ent <- list(range = as.integer(rng),
                style_type = p$style_type, style_number = p$style_number,
                header_row = as.integer(p$header_row),
                autofilter = as.integer(p$autofilter),
                banded_rows = as.integer(p$banded_rows),
                banded_columns = as.integer(p$banded_columns),
                first_column = as.integer(p$first_column),
                last_column = as.integer(p$last_column),
                total_row = as.integer(p$total_row),
                columns = cols)
    nm <- if (i <= length(table_names)) table_names[[i]] else NA_character_
    if (!is.na(nm)) ent$name <- nm
    ent
  })
  .check_table_conflicts(el, df, header_offset, out)
  out
}

# One entry per column of the range, with the caller's overrides merged over
# the data frame's column names.
.table_columns_payload <- function(p, df, rng, reg, props) {
  first <- rng[2L] + 1L                    # 1-based data frame column
  n <- rng[4L] - rng[2L] + 1L
  specs <- vector("list", n)
  for (cc in p$columns) {
    q <- unclass(cc)
    idx <- .resolve_col_index(q$col, names(df), "table column")
    k <- idx - first + 1L
    if (length(k) != 1L || k < 1L || k > n)
      stop(sprintf("table column \"%s\" is outside the table's range",
                   as.character(q$col)), call. = FALSE)
    specs[[k]] <- q
  }
  # `[[` throughout, never `$`: these lists have had their NULLs dropped, and
  # `$` partial-matches, so on a column carrying only `total_label` a `$total`
  # would quietly return the label and a `$format` would return the header
  # format.  Both are silent wrong answers rather than errors.
  lapply(seq_len(n), function(k) {
    q <- specs[[k]]
    dfcol <- first + k - 1L
    default_header <- if (dfcol <= length(df)) names(df)[dfcol]
                      else sprintf("Column %d", k)
    hdr <- q[["header"]]
    ent <- list(header = if (!is.null(hdr)) hdr else default_header)
    if (!is.null(q[["formula"]]))     ent$formula <- q[["formula"]]
    if (!is.null(q[["total"]]))
      ent$total_function <- unname(.LXW_TABLE_FUNCTION[[q[["total"]]]])
    if (!is.null(q[["total_label"]])) ent$total_string <- q[["total_label"]]
    if (!is.null(q[["total_value"]])) ent$total_value <- q[["total_value"]]
    if (!is.null(q[["format"]]))
      ent$format_id <- .register_format(reg,
        merge_xl_format(props$default_format, q[["format"]]))
    if (!is.null(q[["header_format"]]))
      # cascade over the workbook's header format rather than replacing it, so
      # an italic header stays bold and centred like every other header
      ent$header_format_id <- .register_format(reg,
        merge_xl_format(merge_xl_format(props$default_format,
                                        props$header_format),
                        q[["header_format"]]))
    ent
  })
}

# Normalise one column spec or a list of them.
.table_column_list <- function(columns, arg = "columns") {
  if (is.null(columns)) return(list())
  cs <- if (inherits(columns, "xl_table_column")) list(columns) else columns
  if (!is.list(cs))
    stop(sprintf("`%s` must be an xl_table_column object or a list of them",
                 arg), call. = FALSE)
  for (i in seq_along(cs))
    if (!inherits(cs[[i]], "xl_table_column"))
      stop(sprintf("`%s[[%d]]` must be an xl_table_column object", arg, i),
           call. = FALSE)
  cs
}

# Excel's rules for a table name.  libxlsxwriter checks length, a bare C or R,
# and punctuation; it does not reject a cell reference, which Excel does.
#
# The punctuation set is the one _check_table_name() passes to strpbrk() in
# worksheet.c.  Testing membership rather than matching a regex mirrors what
# strpbrk does and keeps the set readable -- as a bracket expression it needs
# escaping that is easy to get wrong ("," "-" "/" reads as a character range).
.TABLE_NAME_BAD_CHARS <- strsplit(
  " !\"#$%&'()*+,-/:;<=>?@[\\]^`{|}~", "", fixed = TRUE)[[1L]]

.check_table_name <- function(name, arg = "name") {
  if (!is.character(name) || length(name) != 1L || !nzchar(name))
    stop(sprintf("`%s` must be a single non-empty string, NULL or NA", arg),
         call. = FALSE)
  if (nchar(name) > 255L)
    stop(sprintf("`%s` exceeds Excel's limit of 255 characters", arg),
         call. = FALSE)
  if (grepl("^[0-9]", name))
    stop(sprintf("`%s` may not start with a digit (got \"%s\")", arg, name),
         call. = FALSE)
  if (toupper(name) %in% c("C", "R"))
    stop(sprintf("`%s` may not be \"C\" or \"R\": Excel reserves them", arg),
         call. = FALSE)
  if (any(strsplit(name, "", fixed = TRUE)[[1L]] %in% .TABLE_NAME_BAD_CHARS))
    stop(sprintf(paste0("`%s` may not contain spaces or punctuation (got ",
                        "\"%s\"); use letters, digits, \".\" or \"_\""),
                 arg, name), call. = FALSE)
  # a name that reads as a cell reference is rejected by Excel but not by
  # libxlsxwriter, so it would reach the file and prompt a repair
  if (grepl("^[A-Za-z]{1,3}[0-9]+$", name))
    stop(sprintf(paste0("`%s` looks like a cell reference (\"%s\"), which ",
                        "Excel will not accept as a table name"), arg, name),
         call. = FALSE)
  invisible(name)
}

# Worksheet tables: the constructors, their validation, and what reaches the
# file.

# ── Style parsing ─────────────────────────────────────────────────────────────

test_that("a style string parses to the type and number C wants", {
  expect_equal(.parse_table_style("medium 9"), list(type = 2L, number = 9L))
  expect_equal(.parse_table_style("dark 11"), list(type = 3L, number = 11L))
  expect_equal(.parse_table_style("light 21"), list(type = 1L, number = 21L))
  # "none" is Light 0 in libxlsxwriter's numbering, not a fourth type
  expect_equal(.parse_table_style("none"), list(type = 1L, number = 0L))
  expect_equal(.parse_table_style("light 0"), .parse_table_style("none"))
  # case and spacing are forgiving
  expect_equal(.parse_table_style("  MEDIUM   9 "), list(type = 2L, number = 9L))
  # the default matches Excel's own default table style
  expect_equal(.parse_table_style(NULL), list(type = 2L, number = 9L))
})

test_that("each style type enforces its own number range", {
  # libxlsxwriter only warns and silently substitutes Medium 9, so writexl has
  # to be the one that refuses -- and the ranges genuinely differ per type
  expect_error(.parse_table_style("light 22"), "light styles are numbered 0 to 21")
  expect_error(.parse_table_style("medium 0"), "medium styles are numbered 1 to 28")
  expect_error(.parse_table_style("medium 29"), "medium styles are numbered 1 to 28")
  expect_error(.parse_table_style("dark 0"), "dark styles are numbered 1 to 11")
  expect_error(.parse_table_style("dark 12"), "dark styles are numbered 1 to 11")
  # the boundaries themselves are accepted
  expect_silent(.parse_table_style("light 0"))
  expect_silent(.parse_table_style("light 21"))
  expect_silent(.parse_table_style("medium 1"))
  expect_silent(.parse_table_style("medium 28"))
  expect_silent(.parse_table_style("dark 1"))
  expect_silent(.parse_table_style("dark 11"))
})

test_that("a malformed style is rejected", {
  expect_error(.parse_table_style("rainbow 3"), 'must be "light", "medium" or "dark"')
  expect_error(.parse_table_style("medium"), 'a type and number like "medium 9"')
  expect_error(.parse_table_style("medium 9 extra"),
               "a type and number like")
  expect_error(.parse_table_style("medium nine"), "numbered 1 to 28")
  expect_error(.parse_table_style(9), "must be a single string")
  expect_error(.parse_table_style(NA_character_), "must be a single string")
})

# ── Table names ───────────────────────────────────────────────────────────────

test_that("Excel's table-name rules are enforced", {
  expect_error(.check_table_name("My Table"), "spaces or punctuation")
  expect_error(.check_table_name("a-b"), "spaces or punctuation")
  expect_error(.check_table_name("2024sales"), "may not start with a digit")
  expect_error(.check_table_name("C"), 'may not be "C" or "R"')
  expect_error(.check_table_name("r"), 'may not be "C" or "R"')
  expect_error(.check_table_name(""), "non-empty string")
  expect_error(.check_table_name(strrep("a", 256L)), "limit of 255")
  # libxlsxwriter does not reject this one, but Excel does
  expect_error(.check_table_name("A1"), "looks like a cell reference")
  expect_error(.check_table_name("XFD1048576"), "looks like a cell reference")
})

test_that("valid table names pass", {
  for (nm in c("Sales", "Sales_2024", "T.1", "a", "Table1x", strrep("a", 255L)))
    expect_silent(.check_table_name(nm))
  # a letter-and-digit name that is not a cell reference is fine
  expect_silent(.check_table_name("Q1data"))
})

test_that("the punctuation set matches libxlsxwriter's strpbrk set", {
  # enumerated rather than sampled: every character worksheet.c rejects must be
  # rejected here, or a name would reach the file and prompt a repair
  bad <- strsplit(" !\"#$%&'()*+,-/:;<=>?@[\\]^`{|}~", "", fixed = TRUE)[[1L]]
  for (ch in bad)
    expect_error(.check_table_name(paste0("a", ch, "b")),
                 "spaces or punctuation", label = ch)
})

# ── Constructors ──────────────────────────────────────────────────────────────

test_that("xl_table carries its options", {
  t <- xl_table(name = "Sales", style = "light 9", total_row = TRUE,
                banded_columns = TRUE, first_column = TRUE)
  p <- unclass(t)
  expect_s3_class(t, "xl_table")
  expect_equal(p$name, "Sales")
  expect_equal(p$style_type, 1L)
  expect_equal(p$style_number, 9L)
  expect_true(p$total_row)
  expect_true(p$banded_columns)
  expect_true(p$first_column)
  expect_false(p$last_column)
})

test_that("turning off the header row also turns off the autofilter", {
  # libxlsxwriter forces this itself; the object must not claim otherwise
  expect_false(unclass(xl_table(header_row = FALSE))$autofilter)
  expect_false(unclass(xl_table(header_row = FALSE, autofilter = TRUE))$autofilter)
  expect_true(unclass(xl_table())$autofilter)
})

test_that("a total without a total row is refused, naming the column", {
  expect_error(xl_table(columns = xl_table_column("qty", total = "sum")),
               'column "qty" gives a total')
  expect_error(xl_table(columns = xl_table_column("qty", total_label = "Total")),
               "no total row")
  expect_silent(xl_table(total_row = TRUE,
                         columns = xl_table_column("qty", total = "sum")))
})

test_that("name = NA with a column formula warns", {
  # structured references name the table, so leaving Excel to number it is a
  # real hazard -- but an allowed one, so it warns rather than errors
  expect_warning(xl_table(name = NA,
                          columns = xl_table_column("q", formula = "=1")),
                 "refers to the table by name")
  # without a formula there is nothing to break
  expect_silent(xl_table(name = NA))
  # and a real name is never a hazard
  expect_silent(xl_table(name = "Sales",
                         columns = xl_table_column("q", formula = "=1")))
})

test_that("xl_table_column carries its options and validates them", {
  cc <- unclass(xl_table_column("qty", header = "Quantity", total = "sum",
                                format = xl_font(bold = TRUE)))
  expect_equal(cc$col, "qty")
  expect_equal(cc$header, "Quantity")
  expect_equal(cc$total, "sum")
  expect_true(is_xl_format(cc$format))

  expect_error(xl_table_column(), "must name or index")
  expect_error(xl_table_column("q", formula = "SUM(1)"), "must start with '='")
  expect_error(xl_table_column("q", total = "median"), "`total` must be one of")
  expect_error(xl_table_column("q", total = "sum", total_label = "T"),
               "either `total` or `total_label`, not both")
  expect_error(xl_table_column("q", header = 1), "single non-NA string")
  expect_error(xl_table_column("q", format = "bold"), "must be an xl_format")
  expect_error(xl_table_column("q", header_format = "bold"),
               "`header_format` must be an xl_format")
})

test_that("every documented total function is accepted", {
  # the docs quantify over the list, so enumerate it
  for (f in names(.LXW_TABLE_FUNCTION))
    expect_silent(xl_table_column("q", total = f))
  expect_equal(length(.LXW_TABLE_FUNCTION), 8L)
  # the enum is sparse, which is the reason callers get names instead
  expect_equal(unname(.LXW_TABLE_FUNCTION[["std_dev"]]), 107L)
  expect_equal(unname(.LXW_TABLE_FUNCTION[["sum"]]), 109L)
})

test_that("columns normalise from one or a list", {
  one <- xl_table(columns = xl_table_column("a"))
  many <- xl_table(columns = list(xl_table_column("a"), xl_table_column("b")))
  expect_length(unclass(one)$columns, 1L)
  expect_length(unclass(many)$columns, 2L)
  expect_null(unclass(xl_table())$columns)
  expect_error(xl_table(columns = "a"), "must be an xl_table_column object")
  expect_error(xl_table(columns = list(xl_table_column("a"), "b")),
               "`columns\\[\\[2\\]\\]` must be an xl_table_column object")
})

test_that("flag arguments are validated", {
  for (arg in c("header_row", "autofilter", "banded_rows", "banded_columns",
                "first_column", "last_column", "total_row")) {
    args <- stats::setNames(list(NA), arg)
    expect_error(do.call(xl_table, args), sprintf("`%s` must be TRUE or FALSE", arg),
                 label = arg)
  }
  expect_error(xl_table(name = c("a", "b")), "single string, NULL or NA")
})

# ── What reaches the file ─────────────────────────────────────────────────────

sdf <- data.frame(fruit = c("apple", "banana", "cherry"), qty = c(5, 150, 300),
                  stringsAsFactors = FALSE)

# The table part for a sheet built with `...` passed to xl_sheet().
table_xml <- function(..., df = sdf) {
  p <- write_tmp(list(Sales = xl_sheet(df, ...)))
  xlsx_part(p, "xl/tables/table1.xml", raw = TRUE)
}

# The text a header cell actually shows, resolved through the shared strings.
header_text <- function(..., df = sdf) {
  p <- write_tmp(list(Sales = xl_sheet(df, ...)))
  w <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  ss <- xlsx_part(p, "xl/sharedStrings.xml", raw = TRUE)
  strings <- regmatches(ss, gregexpr("<t>[^<]*</t>", ss))[[1L]]
  strings <- sub("</t>$", "", sub("^<t>", "", strings))
  row1 <- regmatches(w, regexpr('<row r="1".*?</row>', w))
  idx <- as.integer(sub(".*<v>([0-9]+)</v>.*", "\\1",
                        regmatches(row1, gregexpr("<v>[0-9]+</v>", row1))[[1L]]))
  strings[idx + 1L]
}

test_that("a table's columns take the data frame's column names", {
  # the whole point of the commit: _write_table_column_data() would otherwise
  # default these to "Column 1", "Column 2", and Excel rejects a file whose
  # table definition and header cells disagree
  x <- table_xml(table = xl_table())
  expect_match(x, '<tableColumn id="1" name="fruit"/>', fixed = TRUE)
  expect_match(x, '<tableColumn id="2" name="qty"/>', fixed = TRUE)
  expect_no_match(x, "Column 1", fixed = TRUE)
  # and the header cells say the same thing, which is what Excel checks
  expect_equal(header_text(table = xl_table()), c("fruit", "qty"))
})

test_that("an overridden header reaches both the definition and the cell", {
  x <- table_xml(table = xl_table(columns = xl_table_column("qty",
                                                            header = "Quantity")))
  expect_match(x, 'name="Quantity"', fixed = TRUE)
  expect_equal(header_text(table = xl_table(
    columns = xl_table_column("qty", header = "Quantity"))),
    c("fruit", "Quantity"))
})

test_that("the range defaults to the sheet's used range", {
  expect_match(table_xml(table = xl_table()), 'ref="A1:B4"', fixed = TRUE)
  # a total row extends it by one, and the autofilter stays on the data only
  x <- table_xml(table = xl_table(total_row = TRUE))
  expect_match(x, 'ref="A1:B5"', fixed = TRUE)
  expect_match(x, '<autoFilter ref="A1:B4"/>', fixed = TRUE)
  expect_match(x, 'totalsRowCount="1"', fixed = TRUE)
})

test_that("total functions and labels are written", {
  x <- table_xml(table = xl_table(total_row = TRUE, columns = list(
    xl_table_column("fruit", total_label = "Total"),
    xl_table_column("qty", total = "sum"))))
  expect_match(x, 'totalsRowLabel="Total"', fixed = TRUE)
  expect_match(x, 'totalsRowFunction="sum"', fixed = TRUE)
})

test_that("a column carrying only a total label does not confuse the total", {
  # regression: `$total` partial-matches `total_label` once NULLs are dropped,
  # so a label-only column looked up the label in the function table and errored
  x <- table_xml(table = xl_table(total_row = TRUE,
                                  columns = xl_table_column("fruit",
                                                            total_label = "Total")))
  expect_match(x, 'totalsRowLabel="Total"', fixed = TRUE)
  expect_no_match(x, "totalsRowFunction", fixed = TRUE)
})

test_that("style, banding and highlight flags are written", {
  expect_match(table_xml(table = xl_table(style = "dark 3")),
               'name="TableStyleDark3"', fixed = TRUE)
  # "none" (Light 0) is written as a tableStyleInfo with no name at all, not
  # as a style called TableStyleLight0
  x0 <- table_xml(table = xl_table(style = "none"))
  expect_match(x0, "<tableStyleInfo showFirstColumn", fixed = TRUE)
  expect_no_match(x0, "TableStyle", fixed = TRUE)
  x <- table_xml(table = xl_table(first_column = TRUE, banded_columns = TRUE,
                                  banded_rows = FALSE))
  expect_match(x, 'showFirstColumn="1"', fixed = TRUE)
  expect_match(x, 'showColumnStripes="1"', fixed = TRUE)
  expect_match(x, 'showRowStripes="0"', fixed = TRUE)
})

test_that("turning off the header row removes the autofilter too", {
  x <- table_xml(table = xl_table(header_row = FALSE))
  expect_match(x, 'headerRowCount="0"', fixed = TRUE)
  expect_no_match(x, "<autoFilter", fixed = TRUE)
  # and the autofilter can be dropped on its own
  expect_no_match(table_xml(table = xl_table(autofilter = FALSE)),
                  "<autoFilter", fixed = TRUE)
})

test_that("a column formula is written as a calculated column", {
  d <- sdf
  d$double <- NA_real_
  x <- table_xml(df = d, table = xl_table(
    name = "Sales",
    columns = xl_table_column("double",
                              formula = "=Sales[[#This Row],[qty]]*2")))
  expect_match(x, "<calculatedColumnFormula>", fixed = TRUE)
  expect_match(x, "Sales[[#This Row],[qty]]*2", fixed = TRUE)
})

test_that("a column format reaches the data cells", {
  # libxlsxwriter applies a table column's format only to cells it writes
  # itself -- the total cell and a calculated column -- never to data already
  # written, so left to it this argument would silently do nothing
  p <- write_tmp(list(Sales = xl_sheet(sdf, table = xl_table(
    columns = xl_table_column("qty", format = xl_num_format("$#,##0.00"))))))
  s <- xlsx_part(p, "xl/styles.xml", raw = TRUE)
  w <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(s, "$#,##0.00", fixed = TRUE)
  # the column carries the style, and so does the data cell
  expect_match(w, '<col min="2" max="2"', fixed = TRUE)
  expect_match(regmatches(w, regexpr('<row r="2".*?</row>', w)),
               '<c r="B2" s=', fixed = TRUE)
})

test_that("a column format reaches both the column plan and the table part", {
  # Upstream (jmcnamara/libxlsxwriter#520) confirms the column plan is the
  # supported way to format a table's data cells, and adds that the table
  # column's own format is still needed "for strict correctness with Excel",
  # where it becomes dataDxfId.  Neither half is redundant; this pins both, so
  # a later tidy-up cannot drop the one that looks superfluous.
  p <- write_tmp(list(Sales = xl_sheet(sdf, table = xl_table(
    columns = xl_table_column("qty", format = xl_num_format("$#,##0.00"))))))
  w <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  tbl <- xlsx_part(p, "xl/tables/table1.xml", raw = TRUE)
  expect_match(w, '<col min="2" max="2"', fixed = TRUE)      # the column plan
  expect_match(tbl, "dataDxfId=", fixed = TRUE)              # strict correctness
})

test_that("a column format merges over an xl_col_spec format", {
  # the table column is the more specific of the two, so it wins on conflict
  # while leaving the rest of the column's format in place
  p <- write_tmp(list(Sales = xl_sheet(sdf,
    cols = xl_col_spec("qty", format = xl_font(bold = TRUE)),
    table = xl_table(columns = xl_table_column("qty",
                                               format = xl_num_format("0.000"))))))
  s <- xlsx_part(p, "xl/styles.xml", raw = TRUE)
  expect_match(s, "0.000", fixed = TRUE)
  expect_match(s, "<b/>", fixed = TRUE)
})

test_that("a header format cascades over the workbook header format", {
  # an italic header should stay bold and centred like every other header,
  # rather than replacing the default outright
  p <- write_tmp(list(Sales = xl_sheet(sdf, table = xl_table(
    columns = xl_table_column("qty", header_format = xl_font(italic = TRUE))))))
  s <- xlsx_part(p, "xl/styles.xml", raw = TRUE)
  fonts <- regmatches(s, regexpr("<fonts.*?</fonts>", s))
  # some font carries both, which only happens if the two were merged
  expect_match(fonts, "<b/><i/>|<i/><b/>")
})

test_that("a table turns off constant memory", {
  # worksheet_add_table() returns LXW_ERROR_FEATURE_NOT_SUPPORTED in optimize
  # mode, so this is a hard requirement rather than a preference
  cm <- .resolve_constant_memory(list(D = sdf), list(),
                                 list(list(tables = list(1))))
  expect_equal(cm$on, 0L)
  expect_match(cm$reasons, "does not support tables while row streaming")
  # request = TRUE isolates the feature's effect from the size rule
  expect_equal(.resolve_constant_memory(list(D = sdf), list(),
                                        list(list(tables = NULL)),
                                        request = TRUE)$on, 1L)
})

test_that("a table needs a data row", {
  expect_error(write_tmp(list(D = xl_sheet(sdf[0, ], table = xl_table()))),
               "at least one data row")
  # one data row plus a total row still leaves one non-total row
  expect_silent(write_tmp(list(D = xl_sheet(sdf[1, , drop = FALSE],
                                            table = xl_table(total_row = TRUE)))))
})

test_that("two tables on one sheet both reach the file", {
  p <- write_tmp(list(Sales = xl_sheet(sdf, table = list(
    xl_table(range = "A1:A4"), xl_table(range = "B1:B4")))))
  expect_match(xlsx_part(p, "xl/tables/table1.xml", raw = TRUE),
               'ref="A1:A4"', fixed = TRUE)
  expect_match(xlsx_part(p, "xl/tables/table2.xml", raw = TRUE),
               'ref="B1:B4"', fixed = TRUE)
})

test_that("a range reaching past the data columns is refused", {
  # there is no header name for the extra column, and "Column 3" is exactly the
  # generic caption this design exists to make unreachable
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = xl_table(range = "A1:D4")))),
               "writexl has no header name for the extra column")
  # the full width is fine
  expect_silent(write_tmp(list(D = xl_sheet(sdf, table = xl_table(range = "A1:B4")))))
})

test_that("a header row that would land on data is refused", {
  # worksheet_add_table() writes its captions into the range's first row, so a
  # table starting below row 1 would silently replace values with column names
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = xl_table(range = "A2:B4")))),
               "would overwrite data")
  # ... unless the table has no header row, which is the legitimate way to
  # cover the data rows only
  expect_silent(write_tmp(list(D = xl_sheet(sdf, table = xl_table(
    range = "A2:B4", header_row = FALSE)))))
  # and with no header written at all, any first row is fine
  expect_silent(write_tmp(list(D = xl_sheet(sdf, table = xl_table(range = "A2:B4"))),
                          col_names = FALSE))
})

test_that("a column outside the table's range is refused", {
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = xl_table(
    range = "A1:A4", columns = xl_table_column("qty"))))),
    "outside the table's range")
})

# ── Conflicts ─────────────────────────────────────────────────────────────────

test_that("a range renders back to A1 notation", {
  expect_equal(.range_a1(c(0L, 0L, 3L, 1L)), "A1:B4")
  expect_equal(.range_a1(c(1L, 2L, 5L, 27L)), "C2:AB6")   # past Z
  expect_equal(.range_a1(c(0L, 0L, 0L, 0L)), "A1:A1")
})

test_that("range overlap is inclusive on both edges", {
  expect_true(.ranges_overlap(c(0, 0, 3, 1), c(0, 0, 3, 1)))
  expect_true(.ranges_overlap(c(0, 0, 3, 1), c(3, 1, 5, 2)))   # touching corner
  expect_false(.ranges_overlap(c(0, 0, 3, 0), c(0, 1, 3, 1)))  # side by side
  expect_false(.ranges_overlap(c(0, 0, 1, 1), c(2, 0, 3, 1)))  # stacked
})

test_that("a table and a sheet autofilter over the same cells is refused", {
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = xl_table(),
                                           autofilter = TRUE))),
               "brings its own filter dropdown")
  # the message must name the remedy that works: turning the table's own
  # autofilter off leaves the sheet-level one over the same cells
  expect_error(write_tmp(list(D = xl_sheet(sdf,
                                           table = xl_table(autofilter = FALSE),
                                           autofilter = TRUE))),
               "Drop the sheet's `autofilter`")
  # an autofilter outside the table is fine
  wide <- cbind(sdf, note = c("a", "b", "c"))
  expect_silent(write_tmp(list(D = xl_sheet(wide,
                                            table = xl_table(range = "A1:B4"),
                                            autofilter = "C1:C4"))))
})

test_that("a filter on a table column is refused", {
  # libxlsxwriter: "Filter conditions within the table are not supported"
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = xl_table(),
                                           filter = xl_filter("qty", ">", 100)))),
               'covers column "qty", which `filter` also filters')
})

test_that("a merge inside a table is refused", {
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = xl_table(),
                                           merge = xl_merge("A1:B1", "x")))),
               "does not allow merged cells inside a table")
  # a merge below the table is fine
  expect_silent(write_tmp(list(D = xl_sheet(sdf,
                                            table = xl_table(range = "A1:B4"),
                                            merge = xl_merge("A6:B6", "Note")))))
})

test_that("two overlapping tables are refused", {
  expect_error(write_tmp(list(D = xl_sheet(sdf, table = list(
    xl_table(range = "A1:B4"), xl_table(range = "B1:B4"))))),
    "tables at A1:B4 and B1:B4 overlap")
  # side by side is fine
  expect_silent(write_tmp(list(D = xl_sheet(sdf, table = list(
    xl_table(range = "A1:A4"), xl_table(range = "B1:B4"))))))
})

# ── Name resolution across the workbook ───────────────────────────────────────

tdf <- data.frame(a = 1:2)
tsheet <- function(x) xl_sheet(tdf, table = x)
resolved <- function(elems) .resolve_table_names(elems, names(elems))

test_that("a sheet name is sanitized into a usable table name", {
  expect_equal(.sanitize_table_name("Sales"), "Sales")
  expect_equal(.sanitize_table_name("My Sheet"), "My_Sheet")   # space is legal
  expect_equal(.sanitize_table_name("a-b/c"), "a_b_c")         # in a sheet name
  expect_equal(.sanitize_table_name("Q1.data"), "Q1.data")     # "." is allowed
  expect_equal(.sanitize_table_name(""), "Table")
  # the three reserved shapes get cleared rather than passed through
  expect_equal(.sanitize_table_name("2024"), "T_2024")
  expect_equal(.sanitize_table_name("A1"), "A1_tbl")
  expect_equal(.sanitize_table_name("C"), "C_tbl")
  # whatever comes out must itself be a legal table name
  for (s in c("Sales", "My Sheet", "a-b/c", "", "2024", "A1", "C", "R"))
    expect_silent(.check_table_name(.sanitize_table_name(s)))
})

test_that("an unnamed table takes its name from the sheet", {
  expect_equal(resolved(list(Sales = tsheet(xl_table()))), list("Sales"))
  expect_equal(resolved(list(`My Sheet` = tsheet(xl_table()))),
               list("My_Sheet"))
})

test_that("generated names do not collide", {
  # regression: comparing an original-case name against a folded set let every
  # generated name through, so two tables on one sheet both came out "Sales"
  expect_equal(resolved(list(Sales = tsheet(list(xl_table(), xl_table())))),
               list(c("Sales", "Sales_2")))
  # ... including against an explicit name, which is a fixed point
  expect_equal(
    resolved(list(Sales = tsheet(list(xl_table(name = "Sales"), xl_table())))),
    list(c("Sales", "Sales_2")))
  # and case must not hide a collision, since Excel folds table names
  expect_equal(
    resolved(list(Sales = tsheet(list(xl_table(name = "sales"), xl_table())))),
    list(c("sales", "Sales_2")))
})

test_that("name = NA is left for Excel to number", {
  expect_equal(resolved(list(Sales = tsheet(xl_table(name = NA)))),
               list(NA_character_))
  # and it does not take part in uniqueness, having no name to collide
  expect_equal(
    resolved(list(S = tsheet(list(xl_table(name = NA), xl_table())))),
    list(c(NA, "S")))
})

test_that("a duplicate explicit name is refused, naming both sheets", {
  # libxlsxwriter never checks this, and Excel repairs the file
  expect_error(
    resolved(list(A = tsheet(xl_table(name = "Totals")),
                  B = tsheet(xl_table(name = "Totals")))),
    'both named "Totals" \\(on sheet "A" and sheet "B"\\)')
  # Excel compares case-insensitively, so this is the same collision
  expect_error(
    resolved(list(A = tsheet(xl_table(name = "Totals")),
                  B = tsheet(xl_table(name = "totals")))),
    "unique across the workbook")
  # the same name twice on one sheet is caught too
  expect_error(
    resolved(list(A = tsheet(list(xl_table(name = "T"), xl_table(name = "T"))))),
    "unique across the workbook")
})

test_that("a workbook with no tables resolves to nothing", {
  expect_equal(resolved(list(A = xl_sheet(tdf), B = xl_sheet(tdf))),
               list(character(0), character(0)))
})

test_that("duplicate table names are caught when writing", {
  expect_error(
    write_tmp(list(A = tsheet(xl_table(name = "Totals")),
                   B = tsheet(xl_table(name = "Totals")))),
    "unique across the workbook")
})

test_that("tables normalise from one or a list", {
  expect_error(write_tmp(list(A = xl_sheet(tdf, table = "t"))),
               "must be an xl_table object")
  expect_error(write_tmp(list(A = xl_sheet(tdf, table = list(xl_table(), "t")))),
               "`table\\[\\[2\\]\\]` must be an xl_table object")
})

test_that("the print methods run", {
  expect_output(print(xl_table(name = "Sales")), "Sales")
  expect_output(print(xl_table()), "unnamed")
  expect_output(print(xl_table(name = NA)), "named by Excel")
  expect_output(print(xl_table(total_row = TRUE)), "total row")
  expect_output(print(xl_table_column("qty", total = "sum")), "total=sum")
  expect_output(print(xl_table_column("qty", formula = "=1")), "formula")
})

test_that("a cached total is written for readers that do not evaluate formulas", {
  df <- data.frame(fruit = c("apple", "pear"), qty = c(3, 4),
                   stringsAsFactors = FALSE)
  sheet <- xl_sheet(df, table = xl_table(total_row = TRUE, columns = list(
    xl_table_column("fruit", total_label = "Total"),
    xl_table_column("qty", total = "sum", total_value = 7))))
  p <- write_tmp(list(D = sheet))
  # readxl reads the cached value, so without one the total row reads blank
  got <- as.data.frame(readxl::read_xlsx(p))
  expect_equal(got$qty[3L], 7)
  expect_equal(got$fruit[3L], "Total")
  # the formula is still what Excel recalculates on open
  expect_match(xlsx_part(p, "xl/tables/table1.xml", raw = TRUE),
               'totalsRowFunction="sum"', fixed = TRUE)
})

test_that("a cached total needs the function it caches", {
  expect_error(xl_table_column("qty", total_value = 7),
               "caches the result of `total`", fixed = TRUE)
  expect_error(xl_table_column("qty", total = "sum", total_value = "7"),
               "must be a single non-NA number")
  expect_error(xl_table_column("qty", total = "sum", total_value = c(1, 2)),
               "must be a single non-NA number")
  expect_error(xl_table_column("qty", total = "sum", total_value = NA_real_),
               "must be a single non-NA number")
  expect_equal(unclass(xl_table_column("q", total = "sum",
                                       total_value = 7))$total_value, 7)
})

test_that("a generated table name is trimmed and made unique", {
  # Excel caps a table name at 255 and compares them case-insensitively
  long <- paste(rep("a", 300), collapse = "")
  expect_equal(nchar(.sanitize_table_name(long)), 250L)
  expect_equal(.unique_table_name("Sales", c("sales", "sales_2")), "Sales_3")
  expect_equal(.unique_table_name("Sales", character(0)), "Sales")
})

test_that("a table needs a row that is neither header nor total", {
  df <- data.frame(a = 1, b = 2)
  expect_error(write_tmp(list(D = xl_sheet(df, table = xl_table(
    range = "A1:B2", total_row = TRUE)))),
    "at least one row that is not the header")
})

test_that("a sheet with no table has no column formats to collect", {
  expect_equal(.table_column_formats(data.frame(a = 1), data.frame(a = 1)),
               list(NULL))
  expect_null(.check_table_conflicts(xl_sheet(data.frame(a = 1)),
                                     data.frame(a = 1), 1L, list()))
})

# =============================================================================
# Worksheet-level formatting: xl_sheet(), xl_col_spec(), xl_row_spec()
# =============================================================================
#
# xl_col_spec and xl_row_spec are *subclasses* of xl_format: they carry the
# same formatting groups plus targeting (which columns/rows) and geometry
# (width/height/hidden/level), stored in the `xl_target` / `xl_geometry`
# attributes.  Because they are xl_format objects they combine with `+` just
# like any other format group.
# -----------------------------------------------------------------------------

# Shared constructor for column and row specs.
.new_colrow_spec <- function(kind, target, geometry, format) {
  base <- if (is.null(format)) new_xl_format() else format
  if (!is_xl_format(base))
    stop("`format` must be an xl_format object", call. = FALSE)
  out <- unclass(base)
  attr(out, "xl_geometry") <- geometry
  attr(out, "xl_target")   <- target
  class(out) <- c(paste0("xl_", kind, "_spec"), "xl_colrow_spec", "xl_format")
  out
}

#' Column and row specifications for a worksheet
#'
#' @description
#' `xl_col_spec()` and `xl_row_spec()` describe formatting and geometry for a
#' set of columns or rows within a sheet built by [xl_sheet()].  They are
#' *subclasses* of [xl_format]: they carry the usual formatting groups (so they
#' combine with `+` and the group constructors) plus a target (which
#' columns/rows) and geometry (width/height, hidden, outline level).
#'
#' @param cols Columns to target: a character vector of column names or a
#'   numeric vector of 1-based positions.
#' @param rows Rows to target: a numeric vector of 1-based data-row indices
#'   (row 1 is the first data row, ignoring the header).
#' @param width Column width (in Excel character units).
#' @param height Row height (in points).
#' @param width_pixels,height_pixels The same geometry given in pixels instead.
#'   Give one or the other, not both.  Excel stores character units and points,
#'   so the pixel value is converted on the way in --- read back, a width set
#'   as 100 pixels is 13.57 character units.
#' @param hidden Logical; hide the column/row.
#' @param level Integer outline (grouping) level, 0--7.
#' @param na What to write in this column where a value has none, overriding
#'   the workbook's [xl_properties()]`(na = )`. `NA` (the default) inherits it.
#'   Columns are where this usually belongs: a substitute that suits a numeric
#'   column rarely suits a date one.
#' @param collapsed Logical; draw this column/row as the collapsed summary of
#'   the group beside it.  Excel does not derive this --- the rows or columns
#'   of the group itself need `hidden = TRUE` as well, exactly as clicking the
#'   grouping symbol would leave them.
#' @param format An optional [xl_format] applied to the column/row as its
#'   default cell format.  Combine groups with `+` (e.g.
#'   `xl_font(bold = TRUE) + xl_fill(background = "yellow")`).
#' @return An `xl_col_spec` / `xl_row_spec` object (also an [xl_format]).
#' @family worksheet layout
#' @seealso [xl_sheet], [xl_format]
#' @examples
#' xl_col_spec("revenue", width = 14, format = xl_num_format("#,##0.00"))
#' xl_col_spec(c(1, 2), width = 10) + xl_font(bold = TRUE)
#' xl_col_spec("logo", width_pixels = 100)
#' xl_row_spec(1, height = 24, format = xl_font(bold = TRUE))
#' xl_row_spec(1, height_pixels = 40)
#' @name xl_colrow_spec
NULL

# Pixel geometry, converted exactly as _pixels_to_width() / _pixels_to_height()
# in libxlsxwriter's worksheet.c do.  Converting here rather than calling
# worksheet_set_*_pixels() keeps one geometry path through to C, which matters
# because auto_colwidth and the split-pane resolver both read these widths.
.pixels_to_width <- function(px) {
  if (px == 64) return(8.43)          # LXW_DEF_COL_WIDTH_PIXELS -> default
  if (px <= 12) px / 12 else (px - 5) / 7
}

.pixels_to_height <- function(px) {
  if (px == 20) 15 else px * 0.75     # LXW_DEF_ROW_HEIGHT_PIXELS -> default
}

# Resolve a units pair to the single value the geometry carries.
.one_of_units <- function(value, pixels, arg, px_arg, convert) {
  v  <- .val_num(value, arg, min = 0)
  pv <- .val_num(pixels, px_arg, min = 0)
  if (!is.null(v) && !is.null(pv))
    stop(sprintf("give either `%s` or `%s`, not both", arg, px_arg),
         call. = FALSE)
  if (!is.null(pv)) convert(pv) else v
}

#' @rdname xl_colrow_spec
#' @export
xl_col_spec <- function(cols, width = NA, hidden = NA, level = NA,
                        format = NULL, width_pixels = NA, collapsed = NA,
                        na = NA) {
  if (missing(cols) || length(cols) < 1L)
    stop("`cols` must name or index at least one column", call. = FALSE)
  if (!is.character(cols) && !is.numeric(cols))
    stop("`cols` must be a character (names) or numeric (positions) vector",
         call. = FALSE)
  geometry <- .drop_null(list(
    width  = .one_of_units(width, width_pixels, "width", "width_pixels",
                           .pixels_to_width),
    hidden = .val_flag(hidden, "hidden"),
    level  = .val_int(level, "level", min = 0, max = 7),
    collapsed = .val_flag(collapsed, "collapsed"),
    na = .val_na(na)
  ))
  .new_colrow_spec("col", list(kind = "col", index = cols), geometry, format)
}

#' @rdname xl_colrow_spec
#' @export
xl_row_spec <- function(rows, height = NA, hidden = NA, level = NA,
                        format = NULL, height_pixels = NA, collapsed = NA) {
  if (missing(rows) || length(rows) < 1L)
    stop("`rows` must index at least one row", call. = FALSE)
  if (!is.numeric(rows))
    stop("`rows` must be a numeric vector of 1-based data-row indices",
         call. = FALSE)
  if (any(rows < 1))
    stop("`rows` must be positive (1-based data-row indices)", call. = FALSE)
  geometry <- .drop_null(list(
    height = .one_of_units(height, height_pixels, "height", "height_pixels",
                           .pixels_to_height),
    hidden = .val_flag(hidden, "hidden"),
    level  = .val_int(level, "level", min = 0, max = 7),
    collapsed = .val_flag(collapsed, "collapsed")
  ))
  .new_colrow_spec("row", list(kind = "row", index = rows), geometry, format)
}

# Normalise a single spec or a list of specs to a list of specs.
.as_spec_list <- function(x, cls, arg) {
  if (is.null(x)) return(list())
  if (inherits(x, cls)) return(list(x))
  if (is.list(x)) {
    for (el in x)
      if (!inherits(el, cls))
        stop(sprintf("each element of `%s` must be an %s object", arg, cls),
             call. = FALSE)
    return(x)
  }
  stop(sprintf("`%s` must be an %s object or a list of them", arg, cls),
       call. = FALSE)
}

#' A worksheet with formatting and layout options
#'
#' @description
#' `xl_sheet()` wraps a data frame together with worksheet-level options
#' (column/row formatting and geometry, frozen panes, gridlines, tab color,
#' zoom).  Pass it anywhere [write_xlsx()] accepts a data frame; a plain data
#' frame continues to behave exactly as before.
#'
#' @param data A data frame (the sheet contents).
#' @param cols An [xl_col_spec()], or a list of them.
#' @param rows An [xl_row_spec()], or a list of them.
#' @param freeze Frozen panes: an Excel cell reference such as `"A2"` (freeze
#'   the rows above and columns left of that cell), or `list(row =, col =)`
#'   giving the number of rows/columns to freeze.
#' @param gridlines Logical; show (`TRUE`) or hide (`FALSE`) screen gridlines.
#'   `NA` leaves Excel's default.
#' @param tab_color Sheet tab color (see [xl_color]).
#' @param zoom Zoom level as a percentage (10--400).
#' @param default_row_height Default height (in points) for rows in the sheet.
#' @param auto_colwidth If `TRUE`, size each column to fit its contents (a
#'   character-count heuristic, since the xlsx format has no true "AutoFit").
#'   Columns given an explicit width via [xl_col_spec()] are left untouched.
#' @param autofilter Add an autofilter (filter dropdowns). `TRUE` covers the
#'   whole used range (header plus data); an Excel range string such as
#'   `"A1:D51"` restricts it; `FALSE` (default) adds none.
#' @param protect Protect the worksheet: `FALSE` (default) leaves it
#'   unprotected, `TRUE` applies the standard protection, and a string sets a
#'   password. A named list gives fine-grained control, e.g.
#'   `list(password = "secret", format_cells = TRUE)`; an editing option set to
#'   `TRUE` *allows* that action on the protected sheet. Available option names:
#'   `format_cells`, `format_columns`, `format_rows`, `insert_columns`,
#'   `insert_rows`, `insert_hyperlinks`, `delete_columns`, `delete_rows`,
#'   `sort`, `autofilter`, `pivot_tables`, `scenarios`, `objects`,
#'   `no_select_locked_cells`, `no_select_unlocked_cells`. Cell locking via
#'   [xl_protection()] only has an effect on a protected sheet.
#' @param comment_author Default author for this sheet's cell comments (a
#'   per-comment `author` overrides it).
#' @param show_comments If `TRUE`, all comments on the sheet are initially
#'   shown (individual comments can still be forced via `xl_comment(visible=)`).
#' @param view An [xl_sheet_view()] describing the sheet's tab state and
#'   opening view (active/selected/hidden tab, selection, scroll position,
#'   zero display, direction, split panes).
#' @param page An [xl_page_setup()] describing how the sheet prints
#'   (orientation, paper size, margins, scaling, header and footer). Affects
#'   printing only, never the cell data.
#' @param merge One [xl_merge()], or a list of them, merging rectangles of cells
#'   into single cells.  Merges are applied after the sheet's rows are written,
#'   so a merge over cells the data frame filled keeps only the merged text ---
#'   as merging in Excel does.  Any merge turns off the memory-efficient
#'   row-streaming mode, since it writes back over rows already emitted.
#' @param validation One [xl_validation()], or a list of them, restricting what
#'   may be typed into a range --- a dropdown, a numeric or date bound, a text
#'   length limit or a custom formula.
#' @param conditional One conditional format ([xl_cond_cell()],
#'   [xl_cond_scale()], [xl_cond_bar()], [xl_cond_icons()]), or a list of them,
#'   formatting cells according to their contents.
#' @param filter One [xl_filter()], or a list of them, setting autofilter
#'   criteria.  Each also hides the rows it excludes, because Excel does not
#'   apply a filter when a file is opened --- criteria on their own produce a
#'   sheet that looks filtered but shows every row.  Implies `autofilter = TRUE`
#'   over the used range when `autofilter` is not set separately.
#' @param outline An [xl_outline()] controlling how the grouping symbols
#'   created by `level` in [xl_col_spec()] / [xl_row_spec()] are drawn.  It
#'   changes their display only, never which rows are grouped.
#' @param table One [xl_table()], or a list of them, turning a range into an
#'   Excel table --- a named, styled block with banded rows, a filter dropdown
#'   and an optional total row.  Any table turns off the memory-efficient
#'   row-streaming mode, which libxlsxwriter refuses to combine with tables.
#' @param image One [xl_image()], or a list of them, placing images on the
#'   sheet --- floating over the cells, or inside a cell with `embed = TRUE`.
#' @param chart One [xl_chart()], or a list of them, placed on the sheet and
#'   anchored to a cell.  A chart's series may plot data from any sheet in the
#'   workbook, not only this one.
#' @param background_image An image tiled behind the sheet's cells, in any shape
#'   [xl_image()] accepts.  It is a screen backdrop only --- Excel never prints
#'   it.
#' @param ignore_errors A named list turning off the green error triangle Excel
#'   shows in cells it believes are wrong.  Each name is an error type and each
#'   value a range, e.g. `list(number_stored_as_text = "A2:A99")`.  Types:
#'   `number_stored_as_text`, `eval_error`, `formula_differs`, `formula_range`,
#'   `formula_unlocked`, `empty_cell_reference`, `list_data_validation`,
#'   `calculated_column`, `two_digit_text_year`.
#' @return An `xl_sheet` object.
#' @family worksheet layout
#' @seealso [xl_col_spec], [xl_row_spec], [write_xlsx]
#' @export
#' @examples
#' df <- data.frame(name = c("a", "b"), revenue = c(1000.5, 2000.25))
#' sheet <- xl_sheet(
#'   df,
#'   cols   = xl_col_spec("revenue", width = 14, format = xl_num_format("#,##0.00")),
#'   freeze = "A2",
#'   tab_color = "steelblue"
#' )
#' tmp <- write_xlsx(list(Data = sheet))
xl_sheet <- function(data, cols = NULL, rows = NULL, freeze = NULL,
                     gridlines = NA, tab_color = NA, zoom = NA,
                     default_row_height = NA, auto_colwidth = FALSE,
                     autofilter = FALSE, protect = FALSE,
                     comment_author = NA, show_comments = FALSE,
                     page = NULL, view = NULL, merge = NULL,
                     validation = NULL, conditional = NULL,
                     filter = NULL, outline = NULL, ignore_errors = NULL,
                     table = NULL, image = NULL, background_image = NULL,
                     chart = NULL) {
  if (!is.data.frame(data))
    stop("`data` must be a data frame", call. = FALSE)
  if (!is.logical(auto_colwidth) || length(auto_colwidth) != 1L || is.na(auto_colwidth))
    stop("`auto_colwidth` must be TRUE or FALSE", call. = FALSE)
  if (!(is.logical(autofilter) || is.character(autofilter)) ||
      length(autofilter) != 1L)
    stop('`autofilter` must be TRUE/FALSE or an Excel range string like "A1:D51"',
         call. = FALSE)
  .validate_protect(protect)
  if (!is.logical(show_comments) || length(show_comments) != 1L || is.na(show_comments))
    stop("`show_comments` must be TRUE or FALSE", call. = FALSE)
  structure(
    list(
      data      = data,
      cols      = .as_spec_list(cols, "xl_col_spec", "cols"),
      rows      = .as_spec_list(rows, "xl_row_spec", "rows"),
      freeze    = freeze,
      gridlines = gridlines,
      tab_color = tab_color,
      zoom      = zoom,
      default_row_height = default_row_height,
      auto_colwidth = auto_colwidth,
      autofilter = autofilter,
      protect    = protect,
      comment_author = comment_author,
      show_comments  = show_comments,
      page           = page,
      view           = view,
      merge          = merge,
      validation     = validation,
      conditional    = conditional,
      filter         = filter,
      outline        = outline,
      ignore_errors  = ignore_errors,
      table          = table,
      image          = image,
      background_image = background_image,
      chart          = chart
    ),
    class = "xl_sheet"
  )
}

# lxw_protection, split by the sheet kind each field applies to.  The last two
# are what Excel offers on a chartsheet, where there are no cells to lock.
.LXW_PROTECT_OPTS <- c(
  "no_select_locked_cells", "no_select_unlocked_cells", "format_cells",
  "format_columns", "format_rows", "insert_columns", "insert_rows",
  "insert_hyperlinks", "delete_columns", "delete_rows", "sort", "autofilter",
  "pivot_tables", "scenarios", "objects"
)
.LXW_CHARTSHEET_PROTECT_OPTS <- c("no_content", "no_objects")

.protect_opts <- function(kind) {
  if (identical(kind, "chartsheet")) .LXW_CHARTSHEET_PROTECT_OPTS
  else .LXW_PROTECT_OPTS
}

.validate_protect <- function(protect, kind = "sheet") {
  if (is.null(protect)) return(invisible())
  if (is.logical(protect) && length(protect) == 1L) return(invisible())
  if (is.character(protect) && length(protect) == 1L) return(invisible())
  if (is.list(protect)) {
    nms <- names(protect)
    if (is.null(nms) || any(nms == ""))
      stop("`protect` list must be fully named", call. = FALSE)
    ok <- .protect_opts(kind)
    bad <- setdiff(nms, c("password", ok))
    if (length(bad)) {
      other <- setdiff(c(.LXW_PROTECT_OPTS, .LXW_CHARTSHEET_PROTECT_OPTS), ok)
      why <- if (all(bad %in% other))
        sprintf(" -- %s, so %s.", paste(sprintf("`%s`", bad), collapse = ", "),
                if (identical(kind, "chartsheet"))
                  "belong to a worksheet, and a chartsheet has no cells"
                else "belong to a chartsheet")
      else "."
      stop("unknown `protect` option(s) for a ", kind, ": ",
           paste(bad, collapse = ", "), why, "
  It accepts: ",
           paste(ok, collapse = ", "), call. = FALSE)
    }
    return(invisible())
  }
  stop("`protect` must be TRUE/FALSE, a password string, or a named list",
       call. = FALSE)
}

#' @export
print.xl_sheet <- function(x, ...) {
  cat(sprintf("<xl_sheet: %d rows x %d cols>\n", nrow(x$data), ncol(x$data)))
  if (length(x$cols)) cat("  column specs:", length(x$cols), "\n")
  if (length(x$rows)) cat("  row specs:", length(x$rows), "\n")
  if (!is.null(x$freeze)) cat("  freeze:", format(x$freeze), "\n")
  invisible(x)
}

# --- resolution: turn a sheet into the plan C consumes --------------------

# Build the (flag, password, options) triple C uses for worksheet protection.
.resolve_protect <- function(protect, kind = "sheet") {
  out <- list(flag = 0L, password = NA_character_, options = NULL)
  if (isTRUE(protect)) {
    out$flag <- 1L
  } else if (is.character(protect) && length(protect) == 1L) {
    out$flag <- 1L; out$password <- protect
  } else if (is.list(protect)) {
    out$flag <- 1L
    if (!is.null(protect$password)) out$password <- as.character(protect$password)
    optnames <- setdiff(names(protect), "password")
    if (length(optnames)) {
      # a named list (not a bare vector) so the C side can look up by name
      ok <- .protect_opts(kind)
      out$options <- stats::setNames(
        lapply(ok, function(nm) as.integer(isTRUE(protect[[nm]]))), ok)
    }
  }
  out
}

# Estimate a column width (in Excel character units) from its rendered content.
# There is no native "AutoFit" in the file format, so this is a heuristic based
# on character counts of the displayed values and the header.
.auto_col_width <- function(col, header) {
  vals_w <- .content_nchar(col)
  hdr_w  <- if (nzchar(header)) nchar(header, type = "width") else 0L
  w <- max(c(hdr_w, vals_w, 0L), na.rm = TRUE)
  min(w + 1L, 255L)   # small padding; Excel's maximum column width is 255
}

# Per-cell display width for a column (0 for blank/NA cells).
.content_nchar <- function(col) {
  if (inherits(col, "xl_cell_general")) {
    vapply(.cell_records(col), function(rec) {
      s <- .cell_display_string(rec)
      if (is.na(s)) 0L else nchar(s, type = "width")
    }, integer(1))
  } else {
    x <- format(col, trim = TRUE)
    w <- nchar(x, type = "width")
    w[is.na(col)] <- 0L
    w
  }
}

# The string a general cell displays (hyperlink display / value / formula).
.cell_display_string <- function(rec) {
  # A rich string has been flattened to its runs before widths are measured, so
  # the displayed text is the runs joined -- rec$value is gone by now.
  if (!is.null(rec$rich_text))
    return(paste(rec$rich_text, collapse = ""))
  v <- rec$value
  if (!is.null(v) && length(v) && !all(is.na(v)))
    return(format(v, trim = TRUE)[1L])
  fm <- rec$formula
  if (!is.null(fm) && length(fm) && !is.na(fm)) return(as.character(fm))
  hl <- rec$hyperlink
  if (is.character(hl) && length(hl) && !is.na(hl)) return(hl)
  if (is.list(hl) && !is.null(hl$url)) return(as.character(hl$url))
  NA_character_
}

# Resolve targeted columns (names or positions) to 1-based indices.  `arg` names
# the calling argument in the error messages.
.resolve_col_index <- function(index, colnames, arg = "cols") {
  if (is.character(index)) {
    idx <- match(index, colnames)
    if (anyNA(idx))
      stop(sprintf("`%s`: unknown column(s): ", arg),
           paste(index[is.na(idx)], collapse = ", "), call. = FALSE)
    idx
  } else {
    idx <- as.integer(index)
    if (any(idx < 1L | idx > length(colnames)))
      stop(sprintf("`%s`: column index out of range", arg), call. = FALSE)
    idx
  }
}

# --- sheet overlays -------------------------------------------------------
#
# A sheet overlay is a range-scoped worksheet feature applied in one pass after
# the per-sheet scalar options and *before* the row loop -- required under
# libxlsxwriter's constant-memory mode, where each row is flushed as it is
# written.  Each payload is a named list carrying a `kind` string that C
# dispatches on, so a new feature adds a payload kind rather than a new
# argument to C_write_data_frame_list().
#
# The autofilter is the first (and currently only) kind.  `xl_sheet()` has no
# public `overlay` argument yet; an `overlay` element set on the sheet object
# is picked up here so the mechanism is reachable for testing and for the
# features that will populate it.

# The autofilter payload for an already-resolved 0-based range.
.overlay_autofilter <- function(range) {
  list(kind = "autofilter", range = as.integer(range))
}

# Normalise a single payload or a list of payloads to a list of payloads.
.as_overlay_list <- function(x) {
  if (is.null(x)) return(list())
  if (!is.list(x))
    stop("`overlay` must be an overlay payload or a list of them", call. = FALSE)
  if (!is.null(x$kind)) list(x) else x
}

# Build the per-sheet plan (column/row/scalar options) that C applies.
# `table_names` carries this sheet's tables' workbook-resolved names, since
# uniqueness is settled across sheets rather than within one.
.resolve_sheet_plan <- function(el, df, reg, header_offset, props,
                                table_names = character(0),
                                charts = list()) {
  ncols  <- length(df)
  cnames <- names(df)

  # per-column base format: workbook default_format, plus the date/time number
  # format for Date/POSIXct columns
  col_fmt  <- vector("list", ncols)
  col_width  <- rep(NA_real_, ncols)
  col_hidden <- rep(NA_integer_, ncols)
  col_level  <- rep(NA_integer_, ncols)
  col_collapsed <- rep(NA_integer_, ncols)
  # a list, not a vector: each column's `na` keeps its own type
  col_na <- vector("list", ncols)
  explicit_width <- rep(FALSE, ncols)   # columns whose width the user set
  for (i in seq_len(ncols)) {
    base <- props$default_format
    if (inherits(df[[i]], "POSIXct")) {
      base <- merge_xl_format(base, props$datetime_format)
      col_width[i] <- props$datetime_col_width
    } else if (inherits(df[[i]], "Date")) {
      base <- merge_xl_format(base, props$date_format)
      col_width[i] <- props$date_col_width
    }
    col_fmt[[i]] <- base
  }

  row_row <- integer(0); row_height <- numeric(0)
  row_fmt_id <- integer(0); row_hidden <- integer(0); row_level <- integer(0)
  row_collapsed <- integer(0)
  freeze <- c(-1L, -1L)
  gridlines <- -1L; tab_color <- -1L; zoom <- 0L; default_row_height <- NA_real_
  overlay <- list()
  protect <- list(flag = 0L, password = NA_character_, options = NULL)
  # comment defaults are worksheet-scoped (as in libxlsxwriter); a comment's own
  # author overrides the sheet default
  comment_author <- NA_character_
  show_comments <- FALSE
  page_payload <- NULL
  view <- list()

  if (inherits(el, "xl_sheet")) {
    # column specs
    for (spec in el$cols) {
      idx <- .resolve_col_index(attr(spec, "xl_target")$index, cnames)
      geo <- attr(spec, "xl_geometry")
      for (i in idx) {
        col_fmt[[i]] <- merge_xl_format(col_fmt[[i]], spec)
        if (!is.null(geo$width)) { col_width[i] <- geo$width; explicit_width[i] <- TRUE }
        if (!is.null(geo$hidden)) col_hidden[i] <- as.integer(isTRUE(geo$hidden))
        if (!is.null(geo$level))  col_level[i]  <- as.integer(geo$level)
        if (!is.null(geo$collapsed))
          col_collapsed[i] <- as.integer(isTRUE(geo$collapsed))
        if (!is.null(geo$na)) col_na[[i]] <- geo$na
      }
    }
    # row specs
    for (spec in el$rows) {
      geo <- attr(spec, "xl_geometry")
      fid <- .register_format(reg, spec)
      for (r in as.integer(attr(spec, "xl_target")$index)) {
        row_row    <- c(row_row, (r - 1L) + header_offset)
        row_height <- c(row_height, if (!is.null(geo$height)) geo$height else NA_real_)
        row_fmt_id <- c(row_fmt_id, fid)
        row_hidden <- c(row_hidden,
          if (!is.null(geo$hidden)) as.integer(isTRUE(geo$hidden))
          else NA_integer_)
        row_level  <- c(row_level, if (!is.null(geo$level)) as.integer(geo$level) else NA_integer_)
        row_collapsed <- c(row_collapsed,
          if (!is.null(geo$collapsed)) as.integer(isTRUE(geo$collapsed)) else NA_integer_)
      }
    }
    # A table column's format reaches the data cells only through the column
    # plan; libxlsxwriter never applies it to rows already written.  It merges
    # over any xl_col_spec() format, being the more specific of the two.
    tfmt <- .table_column_formats(el, df)
    for (i in seq_along(tfmt))
      if (!is.null(tfmt[[i]]))
        col_fmt[[i]] <- merge_xl_format(col_fmt[[i]], tfmt[[i]])

    freeze <- .parse_freeze(el$freeze)
    gridlines <- if (is.na(el$gridlines)) -1L else if (isTRUE(el$gridlines)) 3L else 0L
    tab_color <- if (is.na(el$tab_color)) -1L else xl_color(el$tab_color)
    zoom <- if (is.na(el$zoom)) 0L else as.integer(el$zoom)
    default_row_height <- if (is.na(el$default_row_height)) NA_real_
                          else as.numeric(el$default_row_height)
    # libxlsxwriter requires the autofilter range before any column criteria,
    # so the filter payloads are appended after the autofilter one below
    fr <- .resolve_filters(el, df, header_offset)
    af <- el$autofilter
    # a filter is meaningless without the autofilter it belongs to
    if (length(fr$payloads) && !is.character(af) && !isTRUE(af)) af <- TRUE
    if (is.character(af)) {
      overlay <- c(overlay, list(.overlay_autofilter(
        .parse_range(af, "autofilter", df, header_offset))))
    } else if (isTRUE(af)) {
      last_row <- (nrow(df) - 1L) + header_offset
      if (ncols > 0L && last_row >= 0L)
        overlay <- c(overlay, list(.overlay_autofilter(
          c(0L, 0L, as.integer(last_row), ncols - 1L))))
    }
    overlay <- c(overlay, fr$payloads)
    overlay <- c(overlay, .as_overlay_list(el$overlay))
    protect <- .resolve_protect(el$protect)
    page_payload <- .page_setup_payload(el$page, df, header_offset)
    vw <- .sheet_view_of(el)
    for (k in c("active", "selected", "visible", "first_tab", "hide_zero",
                "right_to_left"))
      if (!is.null(vw[[k]])) view[[k]] <- as.integer(isTRUE(vw[[k]]))
    if (!is.null(vw$selection))
      view$selection <- .xl_resolve_range(vw$selection, arg = "selection",
                                          df = df, header_offset = header_offset,
                                          allow_cell = TRUE)
    if (!is.null(vw$top_left))
      view$top_left <- .xl_resolve_range(vw$top_left, arg = "top_left",
                                         df = df, header_offset = header_offset,
                                         allow_cell = TRUE)[1:2]
    comment_author <- el$comment_author
    show_comments <- isTRUE(el$show_comments)
    # auto column widths (for columns the user did not size explicitly)
    if (isTRUE(el$auto_colwidth)) {
      for (i in seq_len(ncols)) {
        if (explicit_width[i]) next
        header <- if (header_offset > 0L) cnames[i] else ""
        col_width[i] <- .auto_col_width(df[[i]], header)
      }
    }
  }

  # The split is resolved last, because converting a cell reference into
  # libxlsxwriter's units needs the sheet's final row heights and column widths
  # (including any set by auto_colwidth just above).
  vw_split <- .sheet_view_of(el)$split
  if (!is.null(vw_split)) {
    if (!is.null(el$freeze) && !(length(el$freeze) == 1L && is.na(el$freeze)))
      stop("`xl_sheet_view(split =)` and `xl_sheet(freeze =)` cannot both be ",
           "set: Excel supports frozen panes or a split, not both",
           call. = FALSE)
    view$split <- .resolve_split(vw_split, df, header_offset, props,
                                 default_row_height, row_row, row_height,
                                 col_width)
  }

  # rows excluded by a filter are hidden alongside any set by an xl_row_spec()
  if (exists("fr", inherits = FALSE) && length(fr$hide)) {
    for (rr in fr$hide) {
      k <- which(row_row == rr)
      if (length(k)) {
        row_hidden[k[1L]] <- 1L
      } else {
        row_row    <- c(row_row, rr)
        row_height <- c(row_height, NA_real_)
        row_fmt_id <- c(row_fmt_id, 0L)
        row_hidden <- c(row_hidden, 1L)
        row_level  <- c(row_level, NA_integer_)
        row_collapsed <- c(row_collapsed, NA_integer_)
      }
    }
  }

  merges <- .resolve_merges(el, df, reg, header_offset, props)
  tables <- .resolve_tables(el, df, reg, header_offset, props, table_names)
  images <- .resolve_images(el, df, reg, header_offset, props)
  background <- .resolve_background(el)
  overlay <- c(overlay, .resolve_validations(el, df, header_offset),
              .resolve_conditionals(el, df, reg, header_offset, props),
              .resolve_ignore_errors(el, df, header_offset))

  col_format_id <- vapply(col_fmt, function(f) .register_format(reg, f), integer(1))

  list(
    col_width = col_width, col_format_id = col_format_id,
    col_hidden = col_hidden, col_level = col_level,
    col_collapsed = col_collapsed,
    col_na = if (any(lengths(col_na))) col_na else NULL,
    row_row = row_row, row_height = row_height, row_format_id = row_fmt_id,
    row_hidden = row_hidden, row_level = row_level,
    row_collapsed = row_collapsed,
    freeze_row = freeze[1L], freeze_col = freeze[2L],
    gridlines = gridlines, tab_color = tab_color, zoom = zoom,
    default_row_height = default_row_height,
    overlay = overlay, protect = protect$flag,
    protect_password = protect$password, protect_options = protect$options,
    comment_author = as.character(comment_author),
    show_comments = as.integer(show_comments),
    page = page_payload,
    view = if (length(view)) view else NULL,
    merges = if (length(merges)) merges else NULL,
    outline = .resolve_outline(el),
    tables = if (length(tables)) tables else NULL,
    images = if (length(images)) images else NULL,
    background = background,
    charts = if (length(charts)) charts else NULL
  )
}

# Outline symbol display, and the cell error indicators Excel can be told to
# stop drawing.  Both are worksheet-level settings with no cell data of their
# own, so the tests read the worksheet part directly.

ws <- function(...) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(list(D = xl_sheet(data.frame(a = 1:3, b = c("x", "y", "z"),
                                          stringsAsFactors = FALSE), ...)),
             tmp)
  xlsx_part(tmp, "xl/worksheets/sheet1.xml", raw = TRUE)
}

# ── Outline ───────────────────────────────────────────────────────────────────

test_that("outline settings reach sheetPr", {
  w <- ws(cols = xl_col_spec("a", level = 1),
          outline = xl_outline(symbols_below = FALSE, symbols_right = FALSE))
  expect_match(w, 'summaryBelow="0"', fixed = TRUE)
  expect_match(w, 'summaryRight="0"', fixed = TRUE)
})

test_that("the defaults match Excel's, so nothing is written for them", {
  # Excel's own defaults are below/right, so writing them would be noise
  w <- ws(cols = xl_col_spec("a", level = 1), outline = xl_outline())
  expect_no_match(w, 'summaryBelow="0"', fixed = TRUE)
  expect_no_match(w, 'summaryRight="0"', fixed = TRUE)
})

test_that("outline display is independent of the grouping itself", {
  # hiding the symbols must not disturb the outlineLevel that creates the group
  w <- ws(cols = xl_col_spec("a", level = 2), outline = xl_outline(visible = FALSE))
  expect_match(w, 'outlineLevel="2"', fixed = TRUE)
})

test_that("outline arguments are validated", {
  expect_error(xl_outline(visible = NA), "`visible` must be TRUE or FALSE")
  expect_error(xl_outline(symbols_below = "yes"),
               "`symbols_below` must be TRUE or FALSE")
  expect_error(xl_outline(auto_style = c(TRUE, FALSE)),
               "`auto_style` must be TRUE or FALSE")
  expect_error(ws(outline = "yes"), "must be an xl_outline object")
})

test_that("the outline print method runs", {
  expect_output(print(xl_outline()), "xl_outline")
  expect_output(print(xl_outline(visible = FALSE)), "visible=FALSE")
})

# ── Ignoring cell errors ──────────────────────────────────────────────────────

test_that("an ignored error reaches ignoredErrors with its range", {
  w <- ws(ignore_errors = list(number_stored_as_text = "B2:B4"))
  expect_match(w, '<ignoredError sqref="B2:B4" numberStoredAsText="1"/>',
               fixed = TRUE)
})

test_that("several types are written together", {
  w <- ws(ignore_errors = list(number_stored_as_text = "B2:B4",
                               eval_error = "A2:A4"))
  expect_match(w, 'numberStoredAsText="1"', fixed = TRUE)
  expect_match(w, 'evalError="1"', fixed = TRUE)
})

test_that("the range takes any spelling the shared resolver accepts", {
  # a data-frame-relative spec resolves against the sheet's columns, with the
  # header row accounted for -- column b, data rows 1:3, is B2:B4
  w <- ws(ignore_errors = list(number_stored_as_text = list(cols = "b")))
  expect_match(w, 'sqref="B2:B4"', fixed = TRUE)
  expect_match(ws(ignore_errors = list(eval_error = list(cols = "a", rows = 1:2))),
               'sqref="A2:A3"', fixed = TRUE)
})

test_that("every documented error type is accepted", {
  # the docs quantify over the whole list, so enumerate it rather than
  # spot-checking: a newly added type fails here until it is covered
  for (ty in names(.LXW_IGNORE_TYPE)) {
    args <- list(ignore_errors = stats::setNames(list("A2:A4"), ty))
    expect_match(do.call(ws, args), "<ignoredError ", fixed = TRUE, label = ty)
  }
  expect_equal(length(.LXW_IGNORE_TYPE), 9L)
})

test_that("ignore_errors arguments are validated", {
  expect_error(ws(ignore_errors = list("A1:A2")), "must be a named list")
  expect_error(ws(ignore_errors = list(nope = "A1:A2")),
               "unknown error type\\(s\\): nope")
  # the message lists what would have worked
  expect_error(ws(ignore_errors = list(nope = "A1:A2")), "number_stored_as_text")
  expect_error(ws(ignore_errors = list(eval_error = "A1", eval_error = "B1")),
               "only once")
  expect_error(ws(ignore_errors = list(eval_error = "not a range")),
               "range must look like")
})

test_that("no outline or ignored errors means neither element is written", {
  w <- ws()
  expect_no_match(w, "<ignoredErrors>", fixed = TRUE)
  expect_no_match(w, "outlinePr", fixed = TRUE)
})

test_that("a collapsed group writes the summary row and column", {
  # Excel does not derive `collapsed` from the group: the detail is hidden and
  # the summary beside it carries the flag, which is the state clicking the
  # grouping symbol leaves behind
  w <- ws(cols = list(xl_col_spec("a", level = 1, hidden = TRUE),
                      xl_col_spec("b", collapsed = TRUE)),
          rows = list(xl_row_spec(1, level = 1, hidden = TRUE),
                      xl_row_spec(2, collapsed = TRUE)))
  expect_match(w, 'collapsed="1"', fixed = TRUE)
  expect_equal(lengths(regmatches(w, gregexpr('collapsed="1"', w, fixed = TRUE))),
               2L)
  # and it is off unless asked for
  expect_false(grepl("collapsed", ws(cols = xl_col_spec("a", level = 1)),
                     fixed = TRUE))
})

test_that("collapsed alone is enough to reach the file", {
  # no width, format, hidden or level beside it -- the column loop skips a
  # column with nothing set, so `collapsed` has to count as something
  expect_match(ws(cols = xl_col_spec("a", collapsed = TRUE)),
               'collapsed="1"', fixed = TRUE)
  expect_match(ws(rows = xl_row_spec(1, collapsed = TRUE)),
               'collapsed="1"', fixed = TRUE)
  expect_error(xl_col_spec("a", collapsed = "yes"), "`collapsed` must be")
  expect_error(xl_row_spec(1, collapsed = "yes"), "`collapsed` must be")
})

test_that("ignore_errors takes a bare vector as well as a list", {
  w <- ws(ignore_errors = c(number_stored_as_text = "A2:A4"))
  expect_match(w, "ignoredError", fixed = TRUE)
  expect_error(ws(ignore_errors = list("A2:A4")), "must be a named list")
})

# Sheet tab state.  These are the only worksheet settings whose rules span the
# whole workbook, so most of these fixtures need more than one sheet.

one <- function() data.frame(x = 1L)

wb_xml <- function(sheets) {
  xlsx_part(write_tmp(sheets), "xl/workbook.xml", raw = TRUE)
}

# ── What reaches the file ─────────────────────────────────────────────────────

test_that("a hidden sheet is marked hidden in workbook.xml", {
  x <- wb_xml(list(A = xl_sheet(one()), B = xl_sheet(one(), view = xl_sheet_view(visible = FALSE))))
  expect_match(x, '<sheet name="B" sheetId="2" state="hidden"', fixed = TRUE)
  # the visible one is untouched
  expect_match(x, '<sheet name="A" sheetId="1" r:id="rId1"/>', fixed = TRUE)
})

test_that("active and first_tab reach workbookView", {
  x <- wb_xml(list(A = xl_sheet(one()),
                   B = xl_sheet(one(), view = xl_sheet_view(active = TRUE)),
                   C = xl_sheet(one(), view = xl_sheet_view(first_tab = TRUE))))
  expect_match(x, 'activeTab="1"', fixed = TRUE)   # 0-based: sheet B
  expect_match(x, 'firstSheet="2"', fixed = TRUE)  # 0-based: sheet C
})

test_that("selected marks the tab without making it active", {
  x <- xlsx_part(write_tmp(list(A = xl_sheet(one()),
                                B = xl_sheet(one(), view = xl_sheet_view(selected = TRUE)))),
                 "xl/worksheets/sheet2.xml", raw = TRUE)
  expect_match(x, 'tabSelected="1"', fixed = TRUE)
})

test_that("visible = TRUE is not the same as unset, and neither hides", {
  x <- wb_xml(list(A = xl_sheet(one(), view = xl_sheet_view(visible = TRUE)), B = xl_sheet(one())))
  expect_false(grepl("state=", x, fixed = TRUE))
})

test_that("sheet view settings leave the cells alone", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = 1:3, stringsAsFactors = FALSE)
  p <- write_tmp(list(A = xl_sheet(df),
                      B = xl_sheet(df, view = xl_sheet_view(active = TRUE, first_tab = TRUE))))
  expect_equal(as.data.frame(readxl::read_xlsx(p, sheet = "B")), df)
})

# ── The cross-sheet rules ─────────────────────────────────────────────────────

test_that("hiding every sheet is refused", {
  expect_error(write_tmp(list(A = xl_sheet(one(), view = xl_sheet_view(visible = FALSE)),
                              B = xl_sheet(one(), view = xl_sheet_view(visible = FALSE)))),
               "at least one visible sheet")
})

test_that("a hidden sheet cannot also be active or selected", {
  expect_error(write_tmp(list(A = xl_sheet(one()),
                              B = xl_sheet(one(), view = xl_sheet_view(
                                visible = FALSE, active = TRUE)))),
               "hidden but also marked active/selected")
  expect_error(write_tmp(list(A = xl_sheet(one()),
                              B = xl_sheet(one(), view = xl_sheet_view(
                                visible = FALSE, selected = TRUE)))),
               "hidden but also marked active/selected")
})

test_that("only one sheet may be active", {
  expect_error(write_tmp(list(A = xl_sheet(one(), view = xl_sheet_view(active = TRUE)),
                              B = xl_sheet(one(), view = xl_sheet_view(active = TRUE)))),
               "only one sheet may be active")
})

test_that("the first sheet cannot be hidden unless another is activated", {
  # Excel opens on the first sheet, so hiding it with nothing else active
  # produces a workbook with no sheet to show
  expect_error(write_tmp(list(A = xl_sheet(one(), view = xl_sheet_view(visible = FALSE)),
                              B = xl_sheet(one()))),
               "cannot be hidden unless another sheet")
  # ... and it is allowed once another sheet is active
  x <- wb_xml(list(A = xl_sheet(one(), view = xl_sheet_view(visible = FALSE)),
                   B = xl_sheet(one(), view = xl_sheet_view(active = TRUE))))
  expect_match(x, '<sheet name="A" sheetId="1" state="hidden"', fixed = TRUE)
  expect_match(x, 'activeTab="1"', fixed = TRUE)
})

test_that("error messages name the offending sheet", {
  expect_error(write_tmp(list(Data = xl_sheet(one()),
                              Notes = xl_sheet(one(), view = xl_sheet_view(
                                visible = FALSE, active = TRUE)))),
               '"Notes"', fixed = TRUE)
  expect_error(write_tmp(list(First = xl_sheet(one(), view = xl_sheet_view(active = TRUE)),
                              Second = xl_sheet(one(), view = xl_sheet_view(active = TRUE)))),
               '"First" and "Second"', fixed = TRUE)
})

test_that("unnamed sheets are identified by position in errors", {
  expect_error(write_tmp(list(xl_sheet(one(), view = xl_sheet_view(active = TRUE)),
                              xl_sheet(one(), view = xl_sheet_view(active = TRUE)))),
               "sheets 1 and 2")
})

# ── Interaction with plain data frames ────────────────────────────────────────

test_that("plain data frames are treated as unset and never trip the rules", {
  # a bare data frame has no tab settings, so it must not be read as hidden
  expect_silent(write_tmp(list(A = one(), B = one())))
  # mixed with an xl_sheet
  expect_silent(write_tmp(list(
    A = one(), B = xl_sheet(one(), view = xl_sheet_view(active = TRUE)))))
  # a single plain data frame is not "every sheet hidden"
  expect_silent(write_tmp(one()))
})

test_that("the flags are validated as logicals", {
  expect_error(xl_sheet(one(), view = xl_sheet_view(active = "yes")), "active")
  expect_error(xl_sheet(one(), view = xl_sheet_view(visible = 1)), "visible")
  expect_error(xl_sheet(one(), view = xl_sheet_view(first_tab = c(TRUE, TRUE))), "first_tab")
})

# ── Zero display, direction, selection and scroll position ───────────────────

view_xml <- function(...) {
  xlsx_part(write_tmp(list(D = xl_sheet(data.frame(x = 1:5),
                                        view = xl_sheet_view(...)))),
            "xl/worksheets/sheet1.xml", raw = TRUE)
}

test_that("hide_zero and right_to_left reach sheetView", {
  expect_match(view_xml(hide_zero = TRUE), 'showZeros="0"', fixed = TRUE)
  expect_match(view_xml(right_to_left = TRUE), 'rightToLeft="1"', fixed = TRUE)
  # FALSE is not TRUE: nothing is written
  expect_false(grepl("showZeros", view_xml(hide_zero = FALSE), fixed = TRUE))
})

test_that("selection accepts a single cell and a range", {
  expect_match(view_xml(selection = "B2"), 'sqref="B2"', fixed = TRUE)
  x <- view_xml(selection = "B2:C4")
  expect_match(x, 'sqref="B2:C4"', fixed = TRUE)
  # the top-left corner becomes the active cell
  expect_match(x, 'activeCell="B2"', fixed = TRUE)
})

test_that("selection accepts a data-frame-relative spec", {
  df <- data.frame(a = 1:5, b = 1:5)
  x <- xlsx_part(write_tmp(list(D = xl_sheet(df, view = xl_sheet_view(selection = list(rows = 1:2,
                                                                 cols = "a"))))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  # data rows 1:2 are sheet rows 2:3 once the header is counted
  expect_match(x, 'sqref="A2:A3"', fixed = TRUE)
})

test_that("an inverted selection is rejected by the shared range parser", {
  # Excel would read the corner order as naming the active cell; writexl gives
  # that up in exchange for one strict range parser, and says so rather than
  # silently normalising
  expect_error(write_tmp(list(D = xl_sheet(data.frame(x = 1:5),
                 view = xl_sheet_view(selection = "C4:B2")))),
               "inverted")
})

test_that("top_left sets the scrolled-to cell", {
  expect_match(view_xml(top_left = "A3"), 'topLeftCell="A3"', fixed = TRUE)
  expect_match(view_xml(top_left = "C10"), 'topLeftCell="C10"', fixed = TRUE)
})

test_that("selection and top_left are independent", {
  x <- view_xml(selection = "B2", top_left = "A3")
  expect_match(x, 'topLeftCell="A3"', fixed = TRUE)
  expect_match(x, 'sqref="B2"', fixed = TRUE)
})

test_that("the view scalars are validated", {
  expect_error(xl_sheet(data.frame(x = 1), view = xl_sheet_view(hide_zero = "yes")), "hide_zero")
  expect_error(xl_sheet(data.frame(x = 1),
                        view = xl_sheet_view(right_to_left = 1)),
               "right_to_left")
  expect_error(write_tmp(list(D = xl_sheet(data.frame(x = 1),
                 view = xl_sheet_view(selection = "not a range")))),
               "selection")
})

test_that("the view scalars leave the cells alone", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = 1:5, stringsAsFactors = FALSE)
  p <- write_tmp(list(D = xl_sheet(df, view = xl_sheet_view(
    hide_zero = TRUE, right_to_left = TRUE, selection = "A2",
    top_left = "A2"))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
})

# ── Split panes ───────────────────────────────────────────────────────────────
#
# worksheet_split_panes() positions the split by distance in row-height and
# column-width units, not by row and column, so the interesting part is the
# conversion.

# `split` is a view setting; everything else passed here is a sheet setting
# (row heights, column widths) that the split conversion has to account for.
pane <- function(split, ...) {
  sheet <- xl_sheet(data.frame(a = 1:5, b = 1:5, cc = 1:5),
                    view = xl_sheet_view(split = split), ...)
  x <- xlsx_part(write_tmp(list(D = sheet)),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  m <- regmatches(x, regexpr("<pane[^/]*/>", x))
  if (!length(m)) NA_character_ else m
}

pane_num <- function(p, attr) {
  m <- regmatches(p, regexec(sprintf('%s="([0-9.]+)"', attr), p))[[1L]]
  if (length(m) == 2L) as.numeric(m[2L]) else 0
}

test_that("a split is placed from a cell reference", {
  expect_match(pane(split = "B3"), "<pane ", fixed = TRUE)
  # a split at B3 is below one header row plus one data row, and right of one
  # column, so both distances are non-zero
  expect_gt(pane_num(pane(split = "B3"), "ySplit"), 0)
  expect_gt(pane_num(pane(split = "B3"), "xSplit"), 0)
})

test_that("a split on one axis only leaves the other unset", {
  expect_equal(pane_num(pane(split = "A3"), "xSplit"), 0)
  expect_gt(pane_num(pane(split = "A3"), "ySplit"), 0)
  expect_equal(pane_num(pane(split = "B1"), "ySplit"), 0)
  expect_gt(pane_num(pane(split = "B1"), "xSplit"), 0)
})

test_that("the split distance follows the sheet's real row heights", {
  # this is the whole point of converting rather than assuming 15 per row:
  # taller rows above the split must push it further down
  base <- pane_num(pane(split = "A3"), "ySplit")
  taller <- pane_num(pane(split = "A3", default_row_height = 30), "ySplit")
  tallest <- pane_num(pane(split = "A3", rows = xl_row_spec(1, height = 45)),
                      "ySplit")
  expect_gt(taller, base)
  expect_gt(tallest, taller)
})

test_that("the split distance follows the sheet's real column widths", {
  base <- pane_num(pane(split = "B1"), "xSplit")
  wider <- pane_num(pane(split = "B1", cols = xl_col_spec("a", width = 20)),
                    "xSplit")
  expect_gt(wider, base)
})

test_that("raw units can be given directly", {
  p <- pane(split = list(vertical = 15, horizontal = 8.43))
  expect_gt(pane_num(p, "ySplit"), 0)
  expect_gt(pane_num(p, "xSplit"), 0)
  # one axis only
  expect_equal(pane_num(pane(split = list(vertical = 15)), "xSplit"), 0)
})

test_that("raw split units are validated", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(x = 1),
                 view = xl_sheet_view(split = list(vertical = -1))))),
               "non-negative")
  expect_error(write_tmp(list(D = xl_sheet(data.frame(x = 1),
                 view = xl_sheet_view(split = list(sideways = 1))))),
               "unknown `split`")
})

test_that("split and freeze cannot both be set", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(x = 1:5), freeze = "B2",
                 view = xl_sheet_view(split = "B2")))),
               "cannot both be set")
  # either alone is fine
  expect_silent(write_tmp(list(D = xl_sheet(data.frame(x = 1:5),
                  view = xl_sheet_view(split = "B2")))))
  expect_silent(write_tmp(list(D = xl_sheet(data.frame(x = 1:5), freeze = "B2"))))
})

test_that("a split leaves the cells alone", {
  skip_if_not_installed("readxl")
  df <- data.frame(a = 1:5, stringsAsFactors = FALSE)
  p <- write_tmp(list(D = xl_sheet(df, view = xl_sheet_view(split = "B3"))))
  expect_equal(as.data.frame(readxl::read_xlsx(p)), df)
})

test_that("the guards on an empty or malformed view hold", {
  expect_null(.resolve_sheet_visibility(list()))
  expect_null(.sheet_view_of(data.frame(a = 1)))
  bad <- xl_sheet(data.frame(a = 1))
  bad$view <- list(active = TRUE)
  expect_error(.sheet_view_of(bad), "must be an xl_sheet_view object")
  expect_error(write_tmp(list(D = bad)), "must be an xl_sheet_view object")
})

test_that("a raw split spec names its own unknown elements", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(a = 1:3),
    view = xl_sheet_view(split = list(vertical = 10, sideways = 2))))),
    "unknown `split` element(s): sideways", fixed = TRUE)
})

# Tests for the worksheet layer: xl_sheet(), xl_col_spec(), xl_row_spec().

sheet_xml <- function(x, ...) {
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(x, tmp, ...)
  xlsx_part(tmp, "xl/worksheets/sheet1.xml", raw = TRUE)
}

test_that("xl_col_spec / xl_row_spec are xl_format subclasses", {
  cs <- xl_col_spec("a", width = 12, format = xl_num_format("0.0"))
  expect_s3_class(cs, "xl_col_spec")
  expect_s3_class(cs, "xl_colrow_spec")
  expect_s3_class(cs, "xl_format")
  expect_equal(attr(cs, "xl_target")$index, "a")
  expect_equal(attr(cs, "xl_geometry")$width, 12)

  rs <- xl_row_spec(1, height = 20)
  expect_s3_class(rs, "xl_row_spec")
  expect_equal(attr(rs, "xl_geometry")$height, 20)
})

test_that("specs combine with + preserving target, geometry, and class", {
  cs <- xl_col_spec("revenue", width = 14) + xl_font(bold = TRUE)
  expect_s3_class(cs, "xl_col_spec")
  expect_equal(attr(cs, "xl_target")$index, "revenue")
  expect_equal(attr(cs, "xl_geometry")$width, 14)
  expect_true(unclass(cs)$font$bold)
  # a combined format (via +) can be passed as `format`
  cs2 <- xl_col_spec("x", width = 10,
                     format = xl_font(italic = TRUE) + xl_fill(background = "yellow"))
  expect_true(unclass(cs2)$font$italic)
  expect_equal(unclass(cs2)$fill$background, xl_color("yellow"))
})

test_that("print methods run", {
  expect_output(print(xl_col_spec("a", width = 12)), "target")
  expect_output(print(xl_col_spec("a", width = 12)), "geometry")
  expect_output(print(xl_sheet(data.frame(a = 1), cols = xl_col_spec("a", width = 9),
                               rows = xl_row_spec(1, height = 9), freeze = "A2")),
                "xl_sheet")
})

test_that("spec constructors validate their inputs", {
  expect_error(xl_col_spec(), "at least one column")
  expect_error(xl_col_spec(TRUE), "character .* or numeric")
  expect_error(xl_row_spec(), "at least one row")
  expect_error(xl_row_spec("a"), "numeric")
  expect_error(xl_row_spec(0), "positive")
  expect_error(xl_col_spec("a", width = -1), "between 0")
  expect_error(xl_col_spec("a", level = 9), "between 0 and 7")
  expect_error(xl_col_spec("a", format = "x"), "must be an xl_format")
})

test_that("xl_sheet validates data and spec lists", {
  expect_error(xl_sheet(1:3), "must be a data frame")
  expect_error(xl_sheet(data.frame(a = 1), cols = "x"),
               "must be an xl_col_spec")
  expect_error(xl_sheet(data.frame(a = 1), rows = list(xl_col_spec("a"))),
               "must be an xl_row_spec")
  s <- xl_sheet(data.frame(a = 1), cols = list(xl_col_spec("a", width = 5)))
  expect_length(s$cols, 1L)
})

test_that("freeze parsing handles cell refs, lists, and errors", {
  expect_equal(.parse_freeze(NULL), c(-1L, -1L))
  expect_equal(.parse_freeze(NA), c(-1L, -1L))
  expect_equal(.parse_freeze("A2"), c(1L, 0L))
  expect_equal(.parse_freeze("B3"), c(2L, 1L))
  expect_equal(.parse_freeze("AA1"), c(0L, 26L))
  expect_equal(.parse_freeze(list(row = 2, col = 1)), c(2L, 1L))
  expect_error(.parse_freeze("nonsense"), "cell reference")
  expect_error(.parse_freeze(42), "cell reference")
})

test_that("column targeting resolves names and positions", {
  expect_equal(.resolve_col_index(c("b", "a"), c("a", "b")), c(2L, 1L))
  expect_equal(.resolve_col_index(c(1, 3), c("a", "b", "c")), c(1L, 3L))
  expect_error(.resolve_col_index("z", c("a", "b")), "unknown column")
  expect_error(.resolve_col_index(9, c("a", "b")), "out of range")
})

test_that("column width and format reach the worksheet/styles xml", {
  s <- xl_sheet(data.frame(name = c("a", "b"), revenue = c(1.5, 2.5)),
                cols = xl_col_spec("revenue", width = 16,
                                   format = xl_num_format("#,##0.00")))
  w <- sheet_xml(list(D = s))
  expect_match(w, 'customWidth="1"')
  expect_match(w, 'width="16')            # Excel pads the exact value
  expect_match(w, '<col min="2" max="2"')  # targeted the 2nd column
})

test_that("column targeting by position works", {
  s <- xl_sheet(data.frame(a = 1:2, b = 3:4),
                cols = xl_col_spec(1, width = 25))
  w <- sheet_xml(list(D = s))
  expect_match(w, '<col min="1" max="1"')
  expect_match(w, 'width="25')
})

test_that("row height, freeze, gridlines, tab color, and zoom are written", {
  s <- xl_sheet(data.frame(a = 1:3),
                rows = xl_row_spec(1, height = 30),
                freeze = "A2", gridlines = FALSE,
                tab_color = "red", zoom = 120)
  w <- sheet_xml(list(D = s))
  expect_match(w, 'ht="30"')
  expect_match(w, "<pane")
  expect_match(w, 'showGridLines="0"')
  expect_match(w, "tabColor")
  expect_match(w, 'zoomScale="120"')
})

test_that("hidden columns and outline levels are written", {
  s <- xl_sheet(data.frame(a = 1:2, b = 3:4),
                cols = xl_col_spec("b", hidden = TRUE, level = 1))
  w <- sheet_xml(list(D = s))
  expect_match(w, 'hidden="1"')
  expect_match(w, 'outlineLevel="1"')
})

test_that("geometry can be given in pixels instead", {
  # converted with libxlsxwriter's own formulas, so the width Excel stores
  # round-trips back to the pixel count asked for: 100 px -> 13.571 chars,
  # stored as 13.571 + 5/7 = 14.2857, which renders at 100 px again
  expect_equal(.pixels_to_width(100), (100 - 5) / 7)
  expect_equal(.pixels_to_height(40), 30)
  # the two documented special cases: the pixel defaults map to the unit
  # defaults exactly, rather than through the general formula
  expect_equal(.pixels_to_width(64), 8.43)
  expect_equal(.pixels_to_height(20), 15)
  expect_equal(.pixels_to_width(12), 1)      # the <= 12 branch

  s <- xl_sheet(data.frame(a = 1:2, b = 3:4),
                cols = xl_col_spec("a", width_pixels = 100),
                rows = xl_row_spec(1, height_pixels = 40))
  w <- sheet_xml(list(D = s))
  expect_match(w, 'width="14.28515625"', fixed = TRUE)
  expect_match(w, 'ht="30"', fixed = TRUE)
})

test_that("a geometry may not be given in both units", {
  expect_error(xl_col_spec("a", width = 10, width_pixels = 100),
               "either `width` or `width_pixels`, not both")
  expect_error(xl_row_spec(1, height = 10, height_pixels = 40),
               "either `height` or `height_pixels`, not both")
  # each on its own is fine
  expect_s3_class(xl_col_spec("a", width_pixels = 100), "xl_col_spec")
  expect_s3_class(xl_row_spec(1, height_pixels = 40), "xl_row_spec")
})

test_that("plain data frames still work and Date columns stay formatted", {
  # regression: date formatting moved from C to the R sheet plan
  df <- data.frame(d = as.Date("2020-01-01") + 0:2, x = 1:3)
  tmp <- tempfile(fileext = ".xlsx")
  write_xlsx(df, tmp)
  s <- xlsx_part(tmp, "xl/styles.xml", raw = TRUE)
  expect_match(s, "yyyy", fixed = TRUE)
  expect_equal(as.Date(readxl::read_xlsx(tmp)$d), df$d)
})

test_that("a sheet's default row height is applied", {
  s <- xl_sheet(data.frame(a = 1:3), default_row_height = 22)
  w <- sheet_xml(list(D = s))
  expect_match(w, 'defaultRowHeight="22"')
})

test_that("auto_colwidth sizes columns to their content", {
  df <- data.frame(x = c("a", "bb"), header_is_long = c(1L, 2L))
  w <- sheet_xml(list(D = xl_sheet(df, auto_colwidth = TRUE)))
  expect_match(w, '<col min="1" max="1" width="3')    # header "x", max value "bb"
  expect_match(w, '<col min="2" max="2" width="15')   # 14-char header
})

test_that("explicit width overrides auto_colwidth", {
  df <- data.frame(text = c("a very long value here indeed", "x"))
  w <- sheet_xml(list(D = xl_sheet(df, auto_colwidth = TRUE,
                                   cols = xl_col_spec("text", width = 5))))
  expect_match(w, '<col min="1" max="1" width="5')
})

test_that("auto_colwidth handles Date and general columns", {
  d <- data.frame(when = as.Date("2020-01-01") + 0:1)     # "2020-01-01" = 10
  expect_match(sheet_xml(list(D = xl_sheet(d, auto_colwidth = TRUE))),
               '<col min="1" max="1" width="11')
  g <- data.frame(a = 1:2)
  g$b <- xl_cell_general(value = c(1234.5, 6.7))           # "1234.5" = 6
  expect_match(sheet_xml(list(D = xl_sheet(g, auto_colwidth = TRUE))),
               '<col min="2" max="2" width="7')
})

test_that("auto_colwidth is validated", {
  expect_error(xl_sheet(data.frame(a = 1), auto_colwidth = "yes"), "TRUE or FALSE")
  expect_error(xl_sheet(data.frame(a = 1), auto_colwidth = NA), "TRUE or FALSE")
})

test_that("cell display width covers value/formula/hyperlink/blank", {
  gen <- c(
    xl_cell_general(value = 12345),
    xl_cell_general(formula = "=SUM(A1:A9)"),
    xl_cell_general(hyperlink = "http://example.com"),
    xl_cell_general(hyperlink = list(url = "http://foo.com")),
    xl_cell_general(value = NA)
  )
  w <- .content_nchar(gen)
  expect_equal(w[1], 5L)
  expect_equal(w[2], nchar("=SUM(A1:A9)"))
  expect_equal(w[3], nchar("http://example.com"))
  expect_equal(w[4], nchar("http://foo.com"))
  expect_equal(w[5], 0L)
})

test_that("xl_sheet mixes with plain data frames in one workbook", {
  wb <- list(
    Styled = xl_sheet(data.frame(a = 1.5), cols = xl_col_spec("a", width = 12)),
    Plain  = data.frame(b = 1:2)
  )
  tmp <- tempfile(fileext = ".xlsx")
  expect_silent(write_xlsx(wb, tmp))
  expect_equal(readxl::read_xlsx(tmp, sheet = "Plain")$b, 1:2)
})

# --- sheet overlays ---------------------------------------------------------
#
# Range-scoped features are applied through one payload list dispatched on
# `kind` in C.  The autofilter is the first kind (its own behavior is covered
# in test-format-protect.R); these tests cover the mechanism itself.

test_that("the autofilter reaches C as an overlay payload", {
  plan <- .resolve_sheet_plan(xl_sheet(data.frame(a = 1:3), autofilter = TRUE),
                              data.frame(a = 1:3), .new_format_registry(), 1L,
                              xl_properties())
  expect_equal(plan$overlay,
               list(list(kind = "autofilter", range = c(0L, 0L, 3L, 0L))))
  expect_null(plan$autofilter)   # the old scalar slot is gone
})

test_that("a sheet with no range features carries no overlays", {
  plan <- .resolve_sheet_plan(xl_sheet(data.frame(a = 1:3)), data.frame(a = 1:3),
                              .new_format_registry(), 1L, xl_properties())
  expect_equal(plan$overlay, list())
})

test_that("an overlay set on the sheet is applied", {
  s <- xl_sheet(data.frame(a = 1:3))
  s$overlay <- list(kind = "autofilter", range = c(0L, 0L, 3L, 0L))
  expect_match(sheet_xml(list(D = s)), '<autoFilter ref="A1:A4"')
})

test_that("an unknown overlay kind is an error, not a silent no-op", {
  s <- xl_sheet(data.frame(a = 1:3))
  s$overlay <- list(kind = "bogus")
  expect_error(write_tmp(list(D = s)), "unknown sheet overlay kind 'bogus'")
})

test_that("an overlay payload without a kind is an error", {
  s <- xl_sheet(data.frame(a = 1:3))
  s$overlay <- list(list(range = c(0L, 0L, 1L, 0L)))
  expect_error(write_tmp(list(D = s)), "missing a 'kind'")
})

test_that("an autofilter overlay without a usable range is an error", {
  s <- xl_sheet(data.frame(a = 1:3))
  s$overlay <- list(kind = "autofilter", range = 1:2)
  expect_error(write_tmp(list(D = s)), "length-4 range")
})

test_that("overlay payload lists are normalised", {
  one <- list(kind = "autofilter", range = 1:4)
  expect_equal(.as_overlay_list(NULL), list())
  expect_equal(.as_overlay_list(one), list(one))
  expect_equal(.as_overlay_list(list(one, one)), list(one, one))
  expect_error(.as_overlay_list("x"), "overlay payload")
})

test_that("a filter hides a row that already carries a row spec", {
  df <- data.frame(a = c("x", "y", "z"), stringsAsFactors = FALSE)
  p <- write_tmp(list(D = xl_sheet(df,
    rows = xl_row_spec(2, height = 30),
    filter = xl_filter(col = "a", list = "x"))))
  x <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  # the spec's own row keeps its height and gains the filter's hidden flag
  expect_match(x, 'ht="30"', fixed = TRUE)
  expect_match(x, 'hidden="1"', fixed = TRUE)
})

# =============================================================================
# Workbook layer: xl_properties() and xl_workbook()
# =============================================================================

# The default number format for POSIXct columns.  Shared with
# .resolve_timezones(), which swaps it out when the time zone is dropped, so
# the two definitions cannot drift apart.
.default_datetime_format <- function() xl_num_format("yyyy-mm-dd HH:mm:ss UTC")

#' Workbook properties, defaults, and metadata
#'
#' @description
#' `xl_properties()` collects everything that applies at the *workbook* level:
#' document metadata, a few native workbook settings, and the formatting
#' defaults that used to be hard-coded.  The formatting defaults are ordinary
#' [xl_format] objects, so you can override them (e.g. change the header style
#' or the default date format) simply by passing a different `xl_format`.
#'
#' The `default_format` is cascaded *under* every cell (an emulated
#' workbook-wide default: libxlsxwriter has no native "Normal style" setter, so
#' it is merged beneath each cell/column format).  `header_format` styles the
#' header row and `hyperlink_format` styles cell hyperlinks; both are also
#' cascaded over `default_format`.
#'
#' @param title,subject,author,manager,company,category,keywords,comments,status,hyperlink_base
#'   Document metadata strings (Excel's "Properties" dialog).
#' @param created The workbook's creation timestamp, a `Date` or `POSIXct`.
#'   Left `NA` libxlsxwriter stamps the moment the file is written, which is
#'   what makes two runs over the same data differ; pinning it makes them
#'   byte-identical.
#' @param na What to write where a value has none --- an `NA` or `NaN` in the
#'   data, or a cell with nothing in it at all. `NA` (the default) leaves the
#'   cell blank, which is what writexl has always done. Anything else is
#'   written in its place, keeping its own type: `na = "Not available"` writes
#'   a string, `na = 0` a number. A column or an individual cell can override
#'   it --- see [xl_col_spec()] and [xl_cell_general()]. Note that a non-blank
#'   `na` in a numeric column makes that column mixed, so a reader such as
#'   [readxl::read_xlsx()] returns the whole column as character.
#' @param custom A named list of custom document properties.  Values may be
#'   character, integer, numeric, logical, `Date` or `POSIXct`.  A `Date` or
#'   `POSIXct` is written as a real datetime property (not as text) and follows
#'   the same workbook-wide time zone rule as datetime cells, described below.
#' @param read_only Logical; mark the workbook read-only recommended.
#' @param window_size Optional integer vector `c(width, height)` for the
#'   workbook window size.
#' @param names A named list of workbook-scoped defined names, each a formula
#'   string (e.g. `list(tax = "=0.2")`).
#' @param default_format An [xl_format] cascaded under every cell (default:
#'   none).
#' @param header_format An [xl_format] for the header row (default: bold,
#'   centered).
#' @param hyperlink_format An [xl_format] for cell hyperlinks (default: blue,
#'   underlined), or `NULL` for no hyperlink styling at all.  `NULL` is the only
#'   way to write an unstyled hyperlink: an empty `xl_format()` leaves the cell
#'   with no format, and Excel files written that way fall back to
#'   libxlsxwriter's own blue-underlined default.
#' @param date_format,datetime_format [xl_format] number formats applied to
#'   `Date` / `POSIXct` values.
#' @param date_col_width,datetime_col_width Default column width for `Date` /
#'   `POSIXct` columns.
#' @param header_row_height Height (in points) of the header row.
#' @return An `xl_properties` object.
#' @section Time zones:
#' Excel has no concept of a time zone.  When every `POSIXct` in the workbook
#' shares one time zone, writexl drops the zone and writes local wall-clock
#' time, and the default `datetime_format` loses its `" UTC"` suffix so that
#' nothing is mislabelled.  When the time zones differ, all datetimes are
#' converted to UTC with a warning.  Supplying your own `datetime_format`
#' overrides the label in either case.
#' @family workbook settings
#' @seealso [xl_workbook], [write_xlsx]
#' @export
#' @examples
#' xl_properties(title = "Quarterly report", author = "Finance",
#'               header_format = xl_font(bold = TRUE, color = "white") +
#'                               xl_fill(background = "navy"))
xl_properties <- function(title = NA, subject = NA, author = NA, manager = NA,
                          company = NA, category = NA, keywords = NA,
                          comments = NA, status = NA, hyperlink_base = NA,
                          created = NA, na = NA, custom = NULL,
                          read_only = FALSE, window_size = NULL,
                          names = NULL,
                          default_format   = xl_format(),
                          header_format    = xl_font(bold = TRUE) +
                                             xl_align(horizontal = "center"),
                          hyperlink_format = xl_font(color = "blue",
                                                     underline = "single"),
                          date_format      = xl_num_format("yyyy-mm-dd"),
                          datetime_format  = .default_datetime_format(),
                          date_col_width = 20, datetime_col_width = 20,
                          header_row_height = 15) {
  fmts <- list(default_format = default_format, header_format = header_format,
               date_format = date_format, datetime_format = datetime_format)
  for (nm in names(fmts))
    if (!is_xl_format(fmts[[nm]]))
      stop(sprintf("`%s` must be an xl_format object", nm), call. = FALSE)
  # NULL is meaningful for hyperlink_format alone: it opts out of hyperlink
  # styling entirely.  An empty xl_format() cannot express that -- it is the
  # neutral element of the cascade, and libxlsxwriter substitutes its own
  # blue-underline format whenever no format reaches worksheet_write_url_opt().
  if (!is.null(hyperlink_format) && !is_xl_format(hyperlink_format))
    stop("`hyperlink_format` must be an xl_format object or NULL", call. = FALSE)
  for (nm in c("date_col_width", "datetime_col_width", "header_row_height")) {
    v <- get(nm)
    if (!is.numeric(v) || length(v) != 1L || is.na(v) || v < 0)
      stop(sprintf("`%s` must be a single non-negative number", nm), call. = FALSE)
  }
  if (!is.null(custom) && (is.null(base::names(custom)) || any(base::names(custom) == "")))
    stop("`custom` must be a named list", call. = FALSE)
  if (!is.null(window_size) && length(window_size) != 2L)
    stop("`window_size` must be a length-2 vector c(width, height)", call. = FALSE)
  if (!is.null(names) && (is.null(base::names(names)) || any(base::names(names) == "")))
    stop("`names` must be a named list of defined-name formulas", call. = FALSE)
  structure(
    list(title = title, subject = subject, author = author, manager = manager,
         company = company, category = category, keywords = keywords,
         comments = comments, status = status, hyperlink_base = hyperlink_base,
         created = created, na = .val_na(na), custom = custom,
         read_only = read_only, window_size = window_size,
         names = names,
         default_format = default_format, header_format = header_format,
         hyperlink_format = hyperlink_format, date_format = date_format,
         datetime_format = datetime_format, date_col_width = date_col_width,
         datetime_col_width = datetime_col_width,
         header_row_height = header_row_height),
    class = "xl_properties"
  )
}

#' @export
print.xl_properties <- function(x, ...) {
  cat("<xl_properties>\n")
  meta <- c("title", "author", "company", "subject")
  set <- meta[vapply(meta, function(k) !is.null(x[[k]]) && !is.na(x[[k]]), logical(1))]
  if (length(set))
    for (k in set) cat(sprintf("  %s: %s\n", k, x[[k]]))
  if (isTRUE(x$read_only)) cat("  read_only: TRUE\n")
  invisible(x)
}

#' A workbook: sheets plus workbook-level properties
#'
#' @description
#' `xl_workbook()` binds one or more sheets (data frames or [xl_sheet]s) to a
#' set of [xl_properties].  It is the single place to attach workbook-level
#' formatting defaults and metadata.  When passed to [write_xlsx()], the
#' workbook's `col_names` and `format_headers` settings take precedence over
#' `write_xlsx()`'s own arguments.
#'
#' @param sheets A data frame, an [xl_sheet], or a (named) list of them.
#' @param properties An [xl_properties] object.
#' @inheritParams write_xlsx
#' @return An `xl_workbook` object.
#' @family workbook settings
#' @seealso [xl_properties], [xl_sheet], [write_xlsx]
#' @export
#' @examples
#' wb <- xl_workbook(
#'   list(Data = data.frame(x = 1:3)),
#'   properties = xl_properties(title = "Demo", author = "me")
#' )
#' tmp <- write_xlsx(wb)
xl_workbook <- function(sheets, properties = xl_properties(),
                        col_names = TRUE, format_headers = TRUE) {
  if (is.data.frame(sheets) || inherits(sheets, "xl_sheet") ||
      inherits(sheets, "xl_chartsheet"))
    sheets <- list(sheets)
  ok <- is.list(sheets) &&
    all(vapply(sheets, function(el) is.data.frame(el) ||
                 inherits(el, "xl_sheet") || inherits(el, "xl_chartsheet"),
               logical(1)))
  if (!ok)
    stop(paste0("`sheets` must be a data frame, an xl_sheet, an ",
                "xl_chartsheet, or a list of them"),
         call. = FALSE)
  if (!inherits(properties, "xl_properties"))
    stop("`properties` must be an xl_properties object", call. = FALSE)
  structure(
    list(sheets = sheets, properties = properties,
         col_names = col_names, format_headers = format_headers),
    class = "xl_workbook"
  )
}

#' @export
print.xl_workbook <- function(x, ...) {
  cat(sprintf("<xl_workbook: %d sheet%s>\n", length(x$sheets),
              if (length(x$sheets) == 1L) "" else "s"))
  invisible(x)
}

# --- constant memory ---------------------------------------------------------
#
# libxlsxwriter can write a workbook in "constant memory" mode, streaming each
# row to disk as soon as it is written.  That keeps memory flat for large
# exports, which is why writexl uses it, but it constrains what may still be
# applied once a row has been flushed: `worksheet_add_table()` is rejected
# outright, and `worksheet_merge_range()` only works within the current row.
# Data validation, conditional formats, images and charts are unaffected.
#
# A feature that cannot work under row streaming turns the mode off *and says
# why*: append to `reasons` and the flag follows.

# Resolve the constant-memory flag for a workbook, from the sheets it is about
# to write.  Returns the C-side integer flag plus the reasons (if any) the mode
# had to be turned off.
# The extra memory libxlsxwriter uses per cell when it is NOT streaming, in
# bytes.  Measured against nycflights13 at several sizes and column mixes: 37
# B/cell for highly repetitive text (the shared-string table deduplicates it)
# rising to 112 for mixed columns at ~10M cells.  The high end is used, so the
# estimate errs towards keeping streaming on -- which at scale is also the
# faster path, not just the leaner one.
.CM_BYTES_PER_CELL <- 110

# Decide whether to stream rows to disk.
#
#   1. constant_memory = FALSE  -> off, without even looking at the rest.
#   2. a blacklisted feature    -> off.  Nothing overrides this: each of these
#      either fails outright under streaming or, for a multi-cell array range,
#      writes a file that opens cleanly and is quietly missing cells.
#   3. constant_memory = TRUE   -> on, overriding the size estimate.
#   4. estimated extra memory below `threshold` -> off.  Streaming only pays
#      for itself on large data, and not streaming also yields a smaller file,
#      since repeated strings are shared rather than written inline.
#   5. otherwise -> on.
.resolve_constant_memory <- function(dfs, props, sheets = NULL, request = NA,
                                     threshold = 128 * 1024^2) {
  request <- .val_flag(request, "constant_memory")
  if (identical(request, FALSE))
    return(list(on = 0L, reasons = "constant_memory = FALSE"))
  reasons <- character(0)
  # A multi-cell array formula range is padded by libxlsxwriter, and it skips
  # that padding entirely when row streaming is on -- silently, returning
  # success -- so the range would be left half-written.  Collected per sheet by
  # .resolve_sheet_formats().
  n_array <- sum(vapply(dfs, function(df)
    length(attr(df, "writexl_array_multicell")), integer(1)))
  n_merge <- if (is.null(sheets)) 0L else
    sum(vapply(sheets, function(s) length(s$merges), integer(1)))
  if (n_array > 0L)
    reasons <- c(reasons, sprintf(
      paste0("%d multi-cell array formula range(s): libxlsxwriter only pads an ",
             "array range when row streaming is off"), n_array))
  # Merges are applied after the sheet's rows so that Excel's own semantics
  # hold, which means writing back over rows already flushed to disk.
  if (n_merge > 0L)
    reasons <- c(reasons, sprintf(
      paste0("%d merged range(s): merges are applied after the rows are ",
             "written, which row streaming does not allow"), n_merge))
  # worksheet_add_table() refuses in optimize mode outright, returning
  # LXW_ERROR_FEATURE_NOT_SUPPORTED rather than quietly doing nothing
  n_table <- if (is.null(sheets)) 0L else
    sum(vapply(sheets, function(s) length(s$tables), integer(1)))
  if (n_table > 0L)
    reasons <- c(reasons, sprintf(
      paste0("%d worksheet table(s): libxlsxwriter does not support tables ",
             "while row streaming"), n_table))
  # An embedded image writes a cell, and worksheet_embed_image() checks the
  # dimensions with the optimize guard active, so anchoring one above the
  # streaming cursor fails.  A *floating* image is unaffected --
  # worksheet_insert_image_opt() checks no dimensions at all -- so only the
  # embedded ones cost streaming.
  n_embed <- if (is.null(sheets)) 0L else
    sum(vapply(sheets, function(s)
      sum(vapply(s$images, function(i) as.integer(i$embed), integer(1)),
          na.rm = TRUE), integer(1)))
  if (n_embed > 0L)
    reasons <- c(reasons, sprintf(
      paste0("%d embedded image(s): an embedded image writes a cell, which ",
             "row streaming does not allow above the current row"), n_embed))
  # Cells are counted across the whole workbook: libxlsxwriter holds every
  # sheet's cell table until the file is closed, not one sheet at a time.
  cells <- sum(vapply(dfs, function(df)
    as.numeric(nrow(df)) * as.numeric(ncol(df)), numeric(1)))
  est <- cells * .CM_BYTES_PER_CELL

  if (length(reasons)) {
    if (isTRUE(request))
      warning("`constant_memory = TRUE` cannot be honoured: ", reasons[1L],
              ". Row streaming is off.", call. = FALSE)
    # The suggestion is emitted here rather than handed back: this is the only
    # place that knows both why streaming was refused and what it cost, and a
    # caller could only re-derive that.  It is worth saying only when streaming
    # would have saved enough to notice -- below that there is nothing for the
    # reader to act on.
    if (est >= threshold)
      message(sprintf(
        paste0("Row streaming is off because of %s. This workbook is ",
               "estimated to use about %.0f MB more memory as a result."),
        reasons[1L], est / 1024^2))
    return(list(on = 0L, reasons = reasons))
  }
  if (isTRUE(request))
    return(list(on = 1L, reasons = character(0)))
  if (est < threshold)
    return(list(on = 0L, reasons = sprintf(
      paste0("an estimated %.0f MB of extra memory, below the %.0f MB ",
             "threshold at which row streaming starts to pay"),
      est / 1024^2, threshold / 1024^2)))
  list(on = 1L, reasons = character(0))
}

# Break a Date/POSIXct out into the fields lxw_datetime carries.  The value has
# already been through .resolve_timezones() by this point, so a POSIXct holds
# the wall-clock reading that is meant to reach the file and UTC is simply how
# that reading is tagged.  A Date has no time of day and maps to midnight.
.datetime_fields <- function(v) {
  lt <- as.POSIXlt(v[1L], tz = "UTC")
  is_date <- inherits(v, "Date")
  list(year  = as.integer(lt$year + 1900L),
       month = as.integer(lt$mon + 1L),
       day   = as.integer(lt$mday),
       hour  = if (is_date) 0L else as.integer(lt$hour),
       min   = if (is_date) 0L else as.integer(lt$min),
       sec   = if (is_date) 0    else as.numeric(lt$sec))
}

# Build the C-side document-properties payload from an xl_properties object.
# `constant_memory` rides along here rather than as another .Call() argument so
# the C entry point's signature stays put as features are added.
.properties_payload <- function(props, constant_memory = 1L) {
  meta_keys <- c("title", "subject", "author", "manager", "company",
                 "category", "keywords", "comments", "status", "hyperlink_base")
  out <- list()
  for (k in meta_keys) {
    v <- props[[k]]
    if (!is.null(v) && !(length(v) == 1L && is.na(v))) out[[k]] <- as.character(v)
  }
  if (!is.null(props$na)) out$na <- props$na
  cr <- props$created
  if (!is.null(cr) && !(length(cr) == 1L && is.na(cr))) {
    if (!inherits(cr, "Date") && !inherits(cr, "POSIXct"))
      stop("`created` must be a Date or POSIXct", call. = FALSE)
    if (length(cr) != 1L)
      stop("`created` must be a single date or datetime", call. = FALSE)
    # time_t on the C side, and a Date is midnight UTC on that day
    out$created <- as.numeric(as.POSIXct(cr, tz = "UTC"))
  }
  if (!is.null(props$custom) && length(props$custom)) {
    out$custom <- lapply(base::names(props$custom), function(nm) {
      v <- props$custom[[nm]]
      # Date/POSIXct first: neither satisfies is.numeric(), so without this they
      # fall through to as.character() and are written as text.
      type <- if (inherits(v, "Date") || inherits(v, "POSIXct")) "datetime"
              else if (is.logical(v)) "boolean"
              else if (is.integer(v)) "integer"
              else if (is.numeric(v)) "number"
              else "string"
      list(name = nm, type = type,
           value = switch(type,
                          datetime = .datetime_fields(v),
                          string   = as.character(v),
                          v))
    })
  }
  out$read_only <- as.integer(isTRUE(props$read_only))
  # hyperlink_format = NULL opts out of hyperlink styling; without this
  # libxlsxwriter substitutes its own blue-underline default whenever a cell
  # reaches worksheet_write_url_opt() with no format.
  out$unset_url_format <- as.integer(is.null(props$hyperlink_format))
  out$header_row_height <- as.numeric(props$header_row_height)
  out$constant_memory <- as.integer(constant_memory)
  if (!is.null(props$window_size))
    out$window_size <- as.integer(props$window_size)
  if (!is.null(props$names) && length(props$names))
    out$names <- lapply(base::names(props$names),
                        function(nm) list(name = nm,
                                          formula = as.character(props$names[[nm]])))
  out
}

# Coerce a character-ish argument, accepting a factor, and fail with a message
# that names the argument rather than stopifnot()'s expression dump.  Shared by
# the three cell shorthands below, which all take a character vector first.
.as_character_arg <- function(x, arg) {
  if (is.factor(x)) x <- as.character(x)
  if (!is.character(x))
    stop(sprintf("`%s` must be a character vector (got %s)", arg,
                 paste(class(x), collapse = "/")), call. = FALSE)
  x
}

# The one string an object will show in the workbook.  Anything classed is put
# through as.character(), so a cell built for a sheet -- xl_cell_general(value
# = "Total") -- can be reused wherever only its displayed value is wanted.  A
# rich string is refused rather than flattened: these callers write a single
# run, so its per-run fonts would disappear without a word.
.as_display_string <- function(x, arg, null_ok = TRUE) {
  if (is.null(x)) {
    if (null_ok) return(NULL)
    stop(sprintf("`%s` must be a single non-NA string", arg), call. = FALSE)
  }
  if (is_xl_rich_string(x) || .cell_holds_rich_string(x))
    stop(sprintf(paste0("`%s` cannot be a rich string: only one font is ",
                        "written here, so the per-run fonts would be lost"),
                 arg), call. = FALSE)
  if (!is.character(x) && !is.null(attr(x, "class"))) x <- as.character(x)
  if (!is.character(x) || length(x) != 1L || is.na(x))
    stop(sprintf(paste0("`%s` must be a single non-NA string, or an object ",
                        "with an as.character() method such as an ",
                        "xl_cell_general()"), arg), call. = FALSE)
  x
}

.cell_holds_rich_string <- function(x) {
  inherits(x, "xl_cell_general") &&
    any(vapply(.cell_records(x), function(el) is_xl_rich_string(el[["value"]]),
               logical(1)))
}

#' Excel Types
#'
#' Create special column types to write to a spreadsheet.
#'
#' @description
#' * `xl_formula(x)` — wraps a character vector of Excel formulas (each must
#'   start with `"="`). The formulas are written to the xlsx file as-is and
#'   are recalculated by Excel on open.
#'
#' * `xl_hyperlink(url, name)` — convenience wrapper that builds an Excel
#'   `=HYPERLINK(url, name)` **formula** for each element.  Because the
#'   hyperlink is stored as a formula, it is readable by
#'   [readxl::read_xlsx()], which returns the formula text.  Display text is
#'   controlled by the `name` argument.
#'
#' * `xl_hyperlink_cell(url, value)` — creates a **native cell-level
#'   hyperlink** using `worksheet_write_url_opt()` from libxlsxwriter.  The
#'   URL is stored as metadata attached to the cell, not in the formula bar.
#'   An optional `value` argument provides the display text shown in the cell.
#'   A tooltip and further options can be set by passing a named list to
#'   `xl_cell_general()` directly.
#'   **Note:** [readxl::read_xlsx()] cannot read cell-level hyperlinks and
#'   returns `NA` for those cells.  Use `xl_hyperlink()` instead when
#'   round-tripping through readxl is required.
#'
#' @family cell content
#' @param x character vector to be interpreted as formula
#' @param format An optional [xl_format] (or list of `xl_format`, one per
#'   element) applied to the cells.  See [xl_format].
#' @export
#' @rdname xl_formula
#' @examples
#' df <- data.frame(
#'   name = c("UCLA", "Berkeley", "Jeroen"),
#'   founded = c(1919, 1868, 2030),
#'   website = xl_hyperlink(c("http://www.ucla.edu", "http://www.berkeley.edu", NA), "homepage")
#' )
#' df$age <- xl_formula('=(YEAR(TODAY()) - INDIRECT("B" & ROW()))')
#' write_xlsx(df, 'universities.xlsx')
#'
#' # xl_hyperlink_cell() stores the URL as native cell metadata.
#' # readxl cannot read these cells, but they display cleanly in Excel.
#' df2 <- data.frame(
#'   name = c("UCLA", "Berkeley"),
#'   website = xl_hyperlink_cell(c("http://www.ucla.edu", "http://www.berkeley.edu"),
#'                                value = "homepage")
#' )
#' write_xlsx(df2, 'universities2.xlsx')
#'
#' # cleanup
#' unlink(c('universities.xlsx', 'universities2.xlsx'))
xl_formula <- function(x, format = NULL){
  x <- .as_character_arg(x, "x")
  if(!all(grepl("^=", x) | is.na(x)))
    stop("Formulas must start with '='")
  xl_cell_general(formula = x, format = format)
}

#' @rdname xl_formula
#' @export
#' @param url character vector of URLs.  Use `NA` to produce a blank cell.
#' @param name **Deprecated.** The former spelling of `value`, kept for
#'   backward compatibility.  Supplying it warns and points at `value`;
#'   supplying both is an error, since they mean the same thing.  `value` has
#'   taken the argument position `name` used to occupy, so code that passed the
#'   display text positionally keeps working unchanged.
xl_hyperlink <- function(url, value = NULL, format = NULL, name = NULL){
  if(!is.null(name)){
    if(!is.null(value))
      stop("Give either `value` or the deprecated `name`, not both: they mean ",
           "the same thing", call. = FALSE)
    warning("The `name` argument of xl_hyperlink() is deprecated; use `value` ",
            "instead, which is what xl_hyperlink_cell() and xl_cell_general() ",
            "already call it.", call. = FALSE)
    value <- name
  }
  url <- .as_character_arg(url, "url")
  fmlas <- if(!is.null(value)){
    paste0("=HYPERLINK(", dubquote(url), ",", dubquote(value), ")")
  } else {
    paste0("=HYPERLINK(", dubquote(url), ")")
  }
  fmlas[is.na(url)] <- NA_character_
  xl_formula(fmlas, format = format)
}

#' @rdname xl_formula
#' @export
#' @param value character vector (or `NULL`) of display text shown in the cell
#'   instead of the URL.  When `NULL` the URL itself is shown.  Recycled to the
#'   length of `url`, and automatically `NA` for cells whose URL is `NA`.  The
#'   same argument name is used by [xl_hyperlink_cell()] and
#'   [xl_cell_general()].
xl_hyperlink_cell <- function(url, value = NULL, format = NULL){
  url <- .as_character_arg(url, "url")
  if(is.null(value)){
    xl_cell_general(hyperlink = url, format = format)
  } else {
    value_arg <- rep_len(as.character(value), length(url))
    value_arg[is.na(url)] <- NA_character_
    xl_cell_general(value = value_arg, hyperlink = url, format = format)
  }
}

# Wrap x in Excel double-quotes, doubling any internal double-quote characters.
dubquote <- function(x){
  paste0('"', gsub('"', '""', x, fixed = TRUE), '"')
}

# Integration test: single xlsx file exercising every package feature.
#
# When this file is opened in Excel / LibreOffice the "expected" column on
# each sheet describes exactly what should appear in each data column next
# to it, making the file self-documenting for manual spot-checks.

test_that("comprehensive xlsx output exercises all package functionality", {
  skip_if_not_installed("readxl")

  # ── Sheet 1: basic_types ─────────────────────────────────────────────────
  # Three rows: a normal value, a blank (NA), and a second value / edge case.
  # Each type gets its own column so readxl can round-trip them correctly.
  #
  # Col A: expected (description)
  # Col B: logical_col    — TRUE / blank / FALSE
  # Col C: integer_col    — 1 / blank / 2
  # Col D: double_col     — 1.5 / blank / "Inf" (written as text)
  # Col E: char_col       — "hello" / blank / "world"
  # Col F: date_col       — 2024-01-15 / blank / 2024-12-31  (yyyy-mm-dd)
  # Col G: datetime_col   — 2024-06-01 00:00:00 / blank / 2024-12-31 23:59:59 UTC
  basic <- data.frame(
    expected     = c(
      "row 1: normal values in every column",
      "row 2: all columns blank (NA)",
      "row 3: FALSE / 2 / Inf written as text 'Inf' / world / 2024-12-31 / second datetime"
    ),
    logical_col  = c(TRUE,                NA,             FALSE),
    integer_col  = c(1L,                  NA_integer_,    2L),
    double_col   = c(1.5,                 NA_real_,       Inf),
    char_col     = c("hello",             NA_character_,  "world"),
    date_col     = as.Date(c("2024-01-15", NA,            "2024-12-31")),
    datetime_col = as.POSIXct(
      c("2024-06-01 00:00:00", NA, "2024-12-31 23:59:59"), tz = "UTC"
    ),
    stringsAsFactors = FALSE
  )

  # ── Sheet 2: formulas ────────────────────────────────────────────────────
  # Col A: x        — numeric values used by the formulas in col C
  # Col B: expected — description of what col C should show
  # Col C: result   — formula cells; NA formula produces a blank cell
  #
  # Formulas reference col A (the x values).
  formulas <- data.frame(
    x        = c(10, 20, 30),
    expected = c(
      "formula =A2*2  ->  Excel shows 20 after recalculation",
      "formula =A3*2  ->  Excel shows 40 after recalculation",
      "blank (NA formula)"
    ),
    stringsAsFactors = FALSE
  )
  formulas$result <- xl_formula(c("=A2*2", "=A3*2", NA))

  # ── Sheet 3: hyperlinks ───────────────────────────────────────────────────
  # Col A: expected        — description
  # Col B: formula_link    — =HYPERLINK() formula; readable by readxl; NA = blank
  # Col C: named_link      — =HYPERLINK() formula with display text; NA = blank
  # Col D: cell_link       — native cell-level hyperlink via xl_hyperlink_cell();
  #                          NOT readable by readxl (reads as NA)
  links <- data.frame(
    expected = c(
      "formula hyperlink  ->  =HYPERLINK() formula, URL shown, clickable",
      "named formula hyperlink  ->  =HYPERLINK() formula, display text 'CRAN' shown",
      "blank (NA URL)",
      "cell-level hyperlink  ->  native URL metadata, display text 'R Project'"
    ),
    stringsAsFactors = FALSE
  )
  links$formula_link <- xl_hyperlink(
    c("https://www.r-project.org", "https://cran.r-project.org", NA,
      "https://www.r-project.org")
  )
  links$named_link <- xl_hyperlink(
    c("https://www.r-project.org", "https://cran.r-project.org", NA,
      "https://www.r-project.org"),
    value = c("R Project", "CRAN", "none", "R Project")
  )
  links$cell_link <- xl_hyperlink_cell(
    c(NA, NA, NA, "https://www.r-project.org"),
    value = c(NA_character_, NA_character_, NA_character_, "R Project")
  )

  # ── Sheet 4: cell_general ─────────────────────────────────────────────────
  # Col A: expected — description of what col B should show
  # Col B: result   — one xl_cell_general per row demonstrating each feature
  #
  # Formulas in col B reference other cells in col B.
  # B2 = 42  (integer value, used as formula input)
  # B3 = 3.14 (double value, used as formula input)
  # B8 = =B2+1     -> 43
  # B9 = =SUM(B2:B3) with cached result 45.14  -> shows 45.14 immediately
  # B10 = =TEXT(B2,"0") with cached result "42" -> shows 42 immediately
  general <- data.frame(
    expected = c(
      # value-only cells (rows 2–7)
      "42  (integer value)",
      "3.14  (double value)",
      "hello  (character value)",
      "TRUE  (logical value)",
      "2024-03-01 as a serial number (Date value; no column-level date format here)",
      "blank  (value = NA, explicit empty cell)",
      # formula cells (rows 8–10)
      "formula =B2+1, no cached value  ->  Excel shows 43 after recalculation",
      "formula =SUM(B2:B3) with cached result 45.14  ->  shows 45.14 immediately",
      "formula =TEXT(B2,\"0\") with cached result '42'  ->  shows 42 immediately",
      # hyperlink cells (rows 11–14)
      "hyperlink https://example.com  ->  raw URL shown, clickable",
      "hyperlink https://example.com  ->  display text 'Visit' shown, clickable",
      "hyperlink https://example.com  ->  display text 'Hover', tooltip on mouseover",
      "hyperlink https://example.com  ->  display text 'Named', tooltip on mouseover"
    ),
    stringsAsFactors = FALSE
  )
  general$result <- c(
    # value-only
    xl_cell_general(value = 42L),
    xl_cell_general(value = 3.14),
    xl_cell_general(value = "hello"),
    xl_cell_general(value = TRUE),
    xl_cell_general(value = as.Date("2024-03-01")),
    xl_cell_general(value = NA),
    # formula
    xl_cell_general(formula = "=B2+1"),
    xl_cell_general(value = 45.14, formula = "=SUM(B2:B3)"),
    xl_cell_general(value = "42",  formula = "=TEXT(B2,\"0\")"),
    # hyperlink
    xl_cell_general(hyperlink = "https://example.com"),
    xl_cell_general(value = "Visit", hyperlink = "https://example.com"),
    xl_cell_general(value = "Hover", hyperlink = list(
      url     = "https://example.com",
      tooltip = "Go to example.com"
    )),
    xl_cell_general(value = "Named", hyperlink = list(
      url     = "https://example.com",
      tooltip = "Tooltip on named hyperlink"
    ))
  )

  # ── Sheet 5: mixed_types ─────────────────────────────────────────────────
  # Col A: expected — description
  # Col B: result   — mixed-type column built with c()
  #
  # B2 = 1.5 (double)
  # B3 = "note" (character — cannot be used in arithmetic)
  # B4 = =B2*2  ->  3 (references the numeric value in B2)
  # B5 = hyperlink
  mixed <- data.frame(
    expected = c(
      "1.5  (double cell)",
      "note  (character cell)",
      "formula =B2*2  ->  Excel shows 3 after recalculation",
      "hyperlink https://www.r-project.org  ->  display text 'R', clickable"
    ),
    stringsAsFactors = FALSE
  )
  mixed$result <- c(
    xl_cell_general(value = 1.5),
    xl_cell_general(value = "note"),
    xl_cell_general(formula = "=B2*2"),
    xl_cell_general(value = "R", hyperlink = "https://www.r-project.org")
  )

  # ── Write all sheets with bold/centred headers ────────────────────────────
  path <- write_xlsx(
    list(
      basic_types  = basic,
      formulas     = formulas,
      hyperlinks   = links,
      cell_general = general,
      mixed_types  = mixed
    ),
    format_headers = TRUE
  )

  expect_true(file.exists(path))
  expect_gt(file.size(path), 0L)

  # ── col_names = FALSE ─────────────────────────────────────────────────────
  # Data starts in row 1; no header row written.
  path_no_header <- write_xlsx(basic, col_names = FALSE)
  expect_true(file.exists(path_no_header))
  rt_no_header <- readxl::read_xlsx(path_no_header, col_names = FALSE)
  expect_equal(nrow(rt_no_header), nrow(basic))

  # ── format_headers = FALSE ────────────────────────────────────────────────
  # Headers written as plain (unformatted) text.
  path_plain <- write_xlsx(basic, format_headers = FALSE)
  expect_true(file.exists(path_plain))

  # ── Round-trip spot-checks on basic_types ─────────────────────────────────
  rt <- readxl::read_xlsx(path, sheet = "basic_types")
  expect_equal(nrow(rt), nrow(basic))
  expect_equal(ncol(rt), ncol(basic))

  # Verify column names are correct
  expect_equal(names(rt), names(basic))

  # Verify column names on formulas and hyperlinks sheets
  rt_f <- readxl::read_xlsx(path, sheet = "formulas")
  expect_equal(names(rt_f), c("x", "expected", "result"))

  rt_l <- readxl::read_xlsx(path, sheet = "hyperlinks")
  expect_equal(names(rt_l), c("expected", "formula_link", "named_link", "cell_link"))

  expect_true(rt$logical_col[1L])             # readxl returns logical TRUE/FALSE
  expect_false(rt$logical_col[3L])
  expect_true(is.na(rt$logical_col[2L]))

  expect_equal(rt$integer_col[1L], 1)         # readxl returns all numbers as double
  expect_true(is.na(rt$integer_col[2L]))

  # double_col: Inf is written as the string "Inf", causing readxl to read
  # the whole column as character — numeric round-trip is not expected.
  expect_equal(rt$char_col[1L], "hello")
  expect_true(is.na(rt$char_col[2L]))

  # ── lxw_version ───────────────────────────────────────────────────────────
  # Returns the bundled libxlsxwriter version as a numeric_version object.
  ver <- lxw_version()
  expect_s3_class(ver, "numeric_version")
  expect_gte(ver, numeric_version("1.0.0"))
})

# =============================================================================
# Autofilter criteria, and the row hiding they imply
# =============================================================================
#
# Setting a filter criteria is not enough.  Excel stores the criteria and the
# row-hidden flags independently and does *not* apply the filter when the file
# is opened, so a file with criteria alone shows the filter dropdown set and
# every row still visible.  writexl therefore evaluates the filter against the
# data frame and hides the rows that do not match.
#
# That makes writexl responsible for reproducing Excel's matching rules
# exactly.  They are measured rather than inferred: write a workbook with
# criteria and no hidden rows, open it, and press Data > Reapply so that Excel
# computes the match itself.
#
# The rules are not a property of the criteria alone.  They are a property of
# the XML form libxlsxwriter chooses.
# _set_custom_filter() in worksheet.c picks one of two forms, and Excel matches
# them by entirely different rules:
#
#   <filters>        "==" and "blanks", unless the value holds a * or ?
#   <customFilters>  everything else, including "==" with a wildcard
#
#   <filters>        matches the cell's DISPLAYED TEXT, case-insensitively.
#                    The R type of `value` does not enter into it: `== 10`,
#                    `== "10"` and `list = "10"` all write val="10", and all
#                    three match the number 10 and the string "10" alike.
#   <customFilters>  matches by type.  If the written value parses as a number
#                    the comparison is numeric, and a text cell never satisfies
#                    it -- except "!=", which a text cell satisfies vacuously.
#                    Otherwise the comparison is textual, case-insensitive,
#                    with * and ? as wildcards, and a number cell never
#                    satisfies it.
#
# So `== "10"` keeps the number 10 (it is a value list) while `== "1*"` keeps
# nothing (it is a custom filter, and no number is text).  Only the string
# written to val matters, which is why the two disagree.
#
# Blank cells satisfy "blanks" and nothing else; "non-blanks" is its exact
# complement, so the two partition the column.
# -----------------------------------------------------------------------------

.LXW_FILTER_CRITERIA <- c(
  none = 0L, "==" = 1L, "!=" = 2L, ">" = 3L, "<" = 4L, ">=" = 5L, "<=" = 6L,
  blanks = 7L, "non-blanks" = 8L
)

.LXW_FILTER_AND_OR <- c(and = 0L, or = 1L)

# The criteria that compare magnitudes, and so need a numeric column.
.FILTER_NUMERIC_CRITERIA <- c(">", "<", ">=", "<=")

# Excel matches text case-insensitively and honours * and ? as wildcards.
.filter_text_match <- function(x, value) {
  pat <- utils::glob2rx(as.character(value))
  grepl(pat, as.character(x), ignore.case = TRUE)
}

# --- How a column looks to Excel ---------------------------------------------
# Every column is reduced to three parallel vectors: the numeric value of the
# cells Excel stores as numbers, the text of the cells it stores as strings,
# and which cells are blank.  A cell contributes to exactly one of the first
# two, which is what lets a mixed xl_cell_general column be matched per cell.
.filter_cells <- function(x, nm) {
  if (inherits(x, "xl_cell_general")) return(.filter_cells_general(x, nm))
  n <- length(x)
  none <- rep(NA_character_, n)
  if (inherits(x, "Date") || inherits(x, "POSIXct"))
    return(list(num = as.numeric(x), txt = none, blank = is.na(x),
                dated = TRUE))
  if (is.numeric(x))
    return(list(num = as.numeric(x), txt = none, blank = is.na(x),
                dated = FALSE))
  # Excel stores a logical as a boolean, displayed as TRUE/FALSE, and does not
  # compare it numerically -- so it behaves as text here.
  if (is.logical(x))
    return(list(num = rep(NA_real_, n),
                txt = ifelse(is.na(x), NA_character_, as.character(x)),
                blank = is.na(x), dated = FALSE))
  ch <- as.character(x)
  list(num = rep(NA_real_, n), txt = ch, blank = is.na(ch) | !nzchar(ch),
       dated = FALSE)
}

.filter_cells_general <- function(x, nm) {
  recs <- .cell_records(x)
  n <- length(recs)
  num <- rep(NA_real_, n)
  txt <- rep(NA_character_, n)
  dated <- FALSE
  for (i in seq_len(n)) {
    r <- recs[[i]]
    if (!is.null(r$formula) && length(r$formula) && !is.na(r$formula))
      stop(sprintf(paste0("cannot filter column \"%s\": row %d holds a ",
                          "formula, so writexl cannot know the value Excel ",
                          "would match. Filter on a plain column instead."),
                   nm, i), call. = FALSE)
    v <- r$value
    if (is_xl_rich_string(v)) {
      txt[i] <- format(v)
    } else if (is.null(v) || (length(v) == 1L && !is_xl_rich_string(v) &&
                              is.na(v))) {
      # no value: a hyperlink cell displays its URL
      hl <- r$hyperlink
      if (is.character(hl) && length(hl) == 1L && !is.na(hl)) txt[i] <- hl
      else if (is.list(hl) && !is.null(hl$url)) txt[i] <- as.character(hl$url)
    } else if (inherits(v, "Date") || inherits(v, "POSIXct")) {
      num[i] <- as.numeric(v)
      dated <- TRUE
    } else if (is.numeric(v)) {
      num[i] <- as.numeric(v)
    } else {
      txt[i] <- as.character(v)
    }
  }
  list(num = num, txt = txt,
       blank = is.na(num) & (is.na(txt) | !nzchar(txt)), dated = dated)
}

# The text Excel shows for a cell, which is what a <filters> list matches.
.filter_display <- function(cells) {
  out <- cells$txt
  ok <- is.na(out) & !is.na(cells$num)
  # format() per element: a shared format would pad 10 to "10.0" beside 20.5
  out[ok] <- vapply(cells$num[ok], function(z)
    format(z, trim = TRUE, scientific = FALSE, digits = 15), character(1))
  out
}

# --- Which XML form a rule takes ---------------------------------------------
# Mirrors _set_custom_filter() in worksheet.c.  It has to be mirrored rather
# than guessed, because the form decides the matching rules.
.filter_wildcard <- function(v) is.character(v) && any(grepl("[*?]", v))

.filter_is_custom <- function(p) {
  custom <- !(identical(p$criteria, "==") || identical(p$criteria, "blanks"))
  if (!is.null(p$criteria2) && identical(p$and_or, "and")) custom <- TRUE
  if (.filter_wildcard(p$value) || .filter_wildcard(p$value2)) custom <- TRUE
  custom
}

# Does the string libxlsxwriter writes to val parse as a number?  That, not the
# R type, is what decides whether Excel compares numerically.
.filter_numeric_value <- function(v) {
  if (is.null(v)) return(FALSE)
  if (.filter_wildcard(v)) return(FALSE)
  if (is.numeric(v) || inherits(v, "Date") || inherits(v, "POSIXct"))
    return(TRUE)
  !is.na(suppressWarnings(as.numeric(as.character(v))))
}

# --- Evaluation ---------------------------------------------------------------
# A value list: the displayed text, matched case-insensitively.
.filter_display_match <- function(cells, vals) {
  d <- tolower(.filter_display(cells))
  !is.na(d) & d %in% tolower(as.character(vals)) & !cells$blank
}

# One custom-filter rule.
.filter_custom_keep <- function(cells, criteria, value) {
  if (criteria == "blanks")     return(cells$blank)
  if (criteria == "non-blanks") return(!cells$blank)
  n <- cells$num
  if (.filter_numeric_value(value)) {
    v <- as.numeric(value)
    keep <- switch(criteria,
      # a text cell is not a number, so it is never equal to one -- which makes
      # "!=" true for it, and every other comparison false
      "==" = !is.na(n) & n == v, "!=" = is.na(n) | n != v,
      ">"  = !is.na(n) & n >  v, "<"  = !is.na(n) & n <  v,
      ">=" = !is.na(n) & n >= v, "<=" = !is.na(n) & n <= v)
  } else {
    if (criteria %in% .FILTER_NUMERIC_CRITERIA)
      stop(sprintf(paste0("`criteria = \"%s\"` with the text value \"%s\": ",
                          "writexl cannot predict how Excel orders text, so it ",
                          "will not guess which rows to hide. Compare against ",
                          "a number instead."), criteria, value), call. = FALSE)
    m <- !is.na(cells$txt) & .filter_text_match(cells$txt, value)
    keep <- if (criteria == "==") m else !m
  }
  # A blank is not equal to anything, so "!=" keeps it -- measured; every other
  # comparison excludes it.  Excel's own "non-blanks" is written as != " " and
  # would be caught by this, but it never reaches here: it is handled above.
  if (criteria == "!=") keep else keep & !cells$blank
}

.filter_keep_rules <- function(cells, p) {
  if (!.filter_is_custom(p)) {
    # a value list, possibly of two values, plus the blanks flag
    vals <- c(if (!identical(p$criteria, "blanks")) p$value,
              if (!is.null(p$criteria2) && !identical(p$criteria2, "blanks"))
                p$value2)
    keep <- if (length(vals)) .filter_display_match(cells, vals)
            else rep(FALSE, length(cells$blank))
    if (identical(p$criteria, "blanks") || identical(p$criteria2, "blanks"))
      keep <- keep | cells$blank
    return(keep)
  }
  keep <- .filter_custom_keep(cells, p$criteria, p$value)
  if (!is.null(p$criteria2)) {
    k2 <- .filter_custom_keep(cells, p$criteria2, p$value2)
    keep <- if (identical(p$and_or, "or")) keep | k2 else keep & k2
  }
  keep
}

# A value list matches displayed text, and writexl cannot know how Excel will
# format a date, so that combination is refused rather than guessed at.
.filter_check_dated <- function(cells, nm) {
  if (!cells$dated) return(invisible(NULL))
  stop(sprintf(paste0("cannot filter column \"%s\" by an exact value or a ",
                      "`list`: it holds dates, and that form of filter matches ",
                      "the text Excel displays, which depends on the number ",
                      "format. Use a comparison such as `\">=\"` instead."),
               nm), call. = FALSE)
}

#' Filter an autofilter column, hiding the rows that do not match
#'
#' @description
#' `xl_filter()` sets the criteria on one autofilter column *and* hides the rows
#' that do not match.  Both halves are necessary: Excel stores the criteria and
#' the hidden rows separately and does not apply a filter when a file is opened,
#' so criteria alone produce a sheet that looks filtered but shows every row.
#'
#' Because writexl decides which rows to hide, it reproduces Excel's own
#' matching rules, which were measured in Excel rather than assumed.  Those
#' rules depend on which of two forms the filter takes:
#'
#' * `"=="` (with no wildcard) and `"blanks"`, and any `list`, are written as a
#'   **value list**.  Excel matches these against the text a cell *displays*,
#'   case-insensitively --- so `"=="` with `10`, with `"10"`, or a `list` of
#'   `"10"` all match both the number `10` and the string `"10"`.
#' * every other criteria, including `"=="` with a `*` or `?` in it, is written
#'   as a **typed comparison**.  If the value is a number the comparison is
#'   numeric and a text cell never satisfies it (except `"!="`, which a text
#'   cell satisfies because it is not that number).  If the value is text the
#'   comparison is textual, case-insensitive, with `*` and `?` as wildcards,
#'   and a number cell never satisfies it.
#'
#' The consequence worth knowing is that `"=="` with `"10"` keeps the number
#' `10`, while `"=="` with `"1*"` keeps nothing on a numeric column: the
#' wildcard changes the form, and so the rule.
#'
#' Blank means an empty cell or an empty string; `"non-blanks"` is its exact
#' complement.  Blanks are excluded by every comparison except `"!="`, which
#' keeps them --- a blank is not equal to anything.  A mixed-type column built
#' with [xl_cell_general()] is matched cell by cell, each by its own type.
#'
#' @section Limitations:
#' A value list matches displayed text, which writexl can only predict for the
#' General format.  Filtering a `Date` or `POSIXct` column by `"=="` or `list`
#' is therefore refused --- use a comparison such as `">="`.  For the same
#' reason a numeric column carrying a custom number format (say two decimal
#' places, or a currency symbol) may display differently from what writexl
#' compares, so prefer a comparison there too.  Columns whose cells hold
#' formulas cannot be filtered at all, since writexl does not know what Excel
#' would compute.
#'
#' @param col The column to filter: a name or a 1-based position.
#' @param criteria One of `"=="`, `"!="`, `">"`, `"<"`, `">="`, `"<="`,
#'   `"blanks"` or `"non-blanks"`.  The four magnitude comparisons need a
#'   numeric value; writexl will not guess how Excel orders text.
#' @param value The value to compare against.  In a text comparison `*` matches
#'   any run of characters and `?` any single one.
#' @param criteria2,value2 An optional second rule for the same column.
#' @param and_or How the two rules combine: `"and"` (default) or `"or"`.
#' @param list Instead of a criteria, keep only rows whose value is in this
#'   character vector.  Matched case-insensitively, as Excel does.
#' @return An `xl_filter` object.
#' @family worksheet features
#' @seealso [xl_sheet], [xl_filter_keep]
#' @export
#' @examples
#' xl_filter("qty", ">", 100)
#' xl_filter("qty", ">", 100, "<", 200)          # between, via two rules
#' xl_filter("fruit", "==", "ap*")               # wildcard: typed comparison
#' xl_filter("fruit", list = c("apple", "banana"))
#' xl_filter("qty", "==", "10")                  # value list: matches the
#'                                               # number 10 and the text "10"
xl_filter <- function(col, criteria = NA, value = NULL, criteria2 = NA,
                      value2 = NULL, and_or = "and", list = NULL) {
  if (missing(col) || is.null(col))
    stop("`col` must name or index the column to filter", call. = FALSE)
  crit  <- .val_enum(criteria, names(.LXW_FILTER_CRITERIA)[-1L], "criteria")
  crit2 <- .val_enum(criteria2, names(.LXW_FILTER_CRITERIA)[-1L], "criteria2")
  ao    <- .val_enum(and_or, names(.LXW_FILTER_AND_OR), "and_or")

  if (!is.null(list)) {
    if (!is.null(crit))
      stop("give either `list` or `criteria`, not both", call. = FALSE)
    if (!is.character(list) || !length(list) || anyNA(list))
      stop("`list` must be a character vector of values to keep", call. = FALSE)
  } else {
    if (is.null(crit))
      stop("`xl_filter()` needs a `criteria` or a `list`", call. = FALSE)
    if (!crit %in% c("blanks", "non-blanks") && is.null(value))
      stop(sprintf("`criteria = \"%s\"` needs a `value`", crit), call. = FALSE)
    if (!is.null(crit2) && !crit2 %in% c("blanks", "non-blanks") &&
        is.null(value2))
      stop(sprintf("`criteria2 = \"%s\"` needs a `value2`", crit2),
           call. = FALSE)
  }
  structure(.drop_null(base::list(col = col, criteria = crit, value = value,
                                  criteria2 = crit2, value2 = value2,
                                  and_or = ao %||% "and", list = list)),
            class = "xl_filter")
}

#' @export
print.xl_filter <- function(x, ...) {
  p <- unclass(x)
  cat(sprintf("<xl_filter: %s %s>\n", p$col,
              if (!is.null(p$list)) sprintf("in %d value(s)", length(p$list))
              else paste(p$criteria, format(p$value))))
  invisible(x)
}

# Normalise one filter or a list of them to a checked list.
.filter_list <- function(filter, arg = "filter") {
  if (is.null(filter)) return(list())
  fs <- if (inherits(filter, "xl_filter")) list(filter) else filter
  if (!is.list(fs))
    stop(sprintf("`%s` must be an xl_filter object or a list of them", arg),
         call. = FALSE)
  for (i in seq_along(fs))
    if (!inherits(fs[[i]], "xl_filter"))
      stop(sprintf("`%s[[%d]]` must be an xl_filter object", arg, i),
           call. = FALSE)
  fs
}

#' Which rows an Excel autofilter would leave visible
#'
#' @description
#' `xl_filter_keep()` answers, for a data frame and a set of [xl_filter()]
#' rules, the question `xl_sheet(filter =)` has to answer internally: which
#' rows does Excel leave visible?  It writes nothing --- it is the matching
#' rule on its own, exported because reproducing Excel's filter semantics is
#' hard to get right and useful outside writing a file.
#'
#' The rules are described in detail under [xl_filter()], and were established
#' by measurement rather than from documentation: a workbook was written with
#' criteria set and no rows hidden, opened in Excel, and Data > Reapply pressed
#' so that Excel computed each match itself.
#'
#' @section Accuracy:
#' This function tracks Excel's *observed* behaviour, so a case found to
#' disagree with Excel is treated as a bug and fixed, which may change the rows
#' it returns.  One limitation is known: a filter written as a value list
#' (`"=="` without a wildcard, or `list`) matches the text a cell **displays**,
#' which writexl can only predict for the General format.  A numeric column
#' carrying a custom number format may therefore display differently from what
#' is compared here --- prefer a comparison such as `">="` on such a column.
#' Dates are refused outright for the same reason.  See [xl_filter()].
#'
#' @param data A data frame whose columns the filters name.
#' @param filter One [xl_filter()], or a list of them.  Filters on different
#'   columns combine with AND, as they do in Excel.
#' @return A logical vector with one element per row of `data`, `TRUE` where
#'   the row stays visible.
#' @family worksheet features
#' @seealso [xl_filter], [xl_sheet]
#' @export
#' @examples
#' sales <- data.frame(fruit = c("apple", "banana", "cherry"),
#'                     qty = c(5, 150, 300))
#' xl_filter_keep(sales, xl_filter("qty", ">", 100))
#' sales[xl_filter_keep(sales, xl_filter("fruit", "==", "*a*")), ]
#'
#' # filters on different columns combine with AND
#' xl_filter_keep(sales, list(xl_filter("qty", ">", 100),
#'                            xl_filter("fruit", "==", "b*")))
xl_filter_keep <- function(data, filter) {
  if (!is.data.frame(data))
    stop("`data` must be a data frame", call. = FALSE)
  fs <- .filter_list(filter)
  keep <- rep(TRUE, nrow(data))
  for (f in fs) keep <- keep & .filter_keep(f, data)
  keep
}

# Which data rows one filter keeps.
.filter_keep <- function(f, df) {
  p <- unclass(f)
  idx <- .resolve_col_index(p$col, names(df), "filter col")
  nm <- names(df)[idx]
  cells <- .filter_cells(df[[idx]], nm)
  if (!is.null(p$list)) {
    .filter_check_dated(cells, nm)
    return(.filter_display_match(cells, p$list))
  }
  if (!.filter_is_custom(p)) .filter_check_dated(cells, nm)
  .filter_keep_rules(cells, p)
}

# Resolve a sheet's filters: the C payloads, plus the 0-based sheet rows to
# hide because they do not match.
.resolve_filters <- function(el, df, header_offset) {
  if (!inherits(el, "xl_sheet") || is.null(el$filter))
    return(list(payloads = list(), hide = integer(0)))
  fs <- .filter_list(el$filter)
  # the rows to hide come from the exported predicate, so the file and
  # xl_filter_keep() can never disagree about what Excel would show
  keep <- xl_filter_keep(df, fs)
  payloads <- lapply(seq_along(fs), function(i) {
    p <- unclass(fs[[i]])
    out <- list(kind = "filter",
                col = .resolve_col_index(p$col, names(df), "filter col") - 1L)
    if (!is.null(p$list)) {
      out$value_list <- as.character(p$list)
    } else {
      out$criteria <- unname(.LXW_FILTER_CRITERIA[[p$criteria]])
      out <- c(out, .filter_value_payload(p$value, ""))
      if (!is.null(p$criteria2)) {
        out$criteria2 <- unname(.LXW_FILTER_CRITERIA[[p$criteria2]])
        out <- c(out, .filter_value_payload(p$value2, "2"))
        out$and_or <- unname(.LXW_FILTER_AND_OR[[p$and_or]])
      }
    }
    out
  })
  list(payloads = payloads,
       hide = as.integer(which(!keep) - 1L + header_offset))
}

# A rule's value, as the number or string the struct carries.
.filter_value_payload <- function(v, suffix) {
  if (is.null(v)) return(list())
  if (is.character(v))
    return(stats::setNames(list(as.character(v)),
                           paste0("value_string", suffix)))
  stats::setNames(list(as.numeric(v)), paste0("value", suffix))
}

# Autofilter criteria, and the rows they hide.
#
# The expectations below are measured, not documented: each records what Excel
# itself computes for those criteria, via Data > Reapply on a workbook written
# with no rows hidden.  A predicate that drifts away from Excel fails here.

# The sheet rows a filter hides (1-based, as they appear in the XML).
filter_hidden <- function(df, ...) {
  p <- write_tmp(list(D = xl_sheet(df, filter = list(...))))
  w <- xlsx_part(p, "xl/worksheets/sheet1.xml", raw = TRUE)
  rows <- regmatches(w, gregexpr('<row r="[0-9]+"[^>]*>', w))[[1]]
  n <- as.integer(sub('.*r="([0-9]+)".*', "\\1", rows))
  sort(n[grepl("hidden", rows)])
}

# Data row indices kept, derived from the predicate directly.
kept <- function(df, f) which(.filter_keep(f, df))

# ── Semantics measured in Excel ───────────────────────────────────────────────

test_that("text comparison is case-insensitive", {
  df <- data.frame(f = c("apple", "Apple", "APPLE", "banana"),
                   stringsAsFactors = FALSE)
  # Excel: apple, Apple and APPLE all stay visible
  expect_equal(kept(df, xl_filter("f", "==", "Apple")), 1:3)
  expect_equal(filter_hidden(df, xl_filter("f", "==", "Apple")), 5L)
})

test_that("* and ? are wildcards in a text value", {
  df <- data.frame(f = c("apple", "apricot", "banana", "ap"),
                   stringsAsFactors = FALSE)
  # Excel: apple, apricot and ap stay visible; banana goes
  expect_equal(kept(df, xl_filter("f", "==", "ap*")), c(1L, 2L, 4L))
  expect_equal(kept(df, xl_filter("f", "==", "ap?le")), 1L)
})

test_that("blank covers an empty cell and an empty string alike", {
  df <- data.frame(v = c("a", NA, "", "b"), stringsAsFactors = FALSE)
  # Excel: both middle rows stay visible under Blanks ...
  expect_equal(kept(df, xl_filter("v", "blanks")), 2:3)
  # ... and both disappear under Non-blanks, so the two partition the column
  expect_equal(kept(df, xl_filter("v", "non-blanks")), c(1L, 4L))
  expect_equal(filter_hidden(df, xl_filter("v", "blanks")), c(2L, 5L))
  expect_equal(filter_hidden(df, xl_filter("v", "non-blanks")), c(3L, 4L))
})

test_that("a value list matches displayed text, whatever the cell's type", {
  # "==" without a wildcard writes <filters>, which Excel matches against the
  # text a cell displays -- so a text criteria does match a number cell.  The R
  # type of `value` is irrelevant: only the string written to val matters.
  num <- data.frame(q = c(10, 20, 100, 15))
  expect_equal(kept(num, xl_filter("q", "==", "10")), 1L)
  expect_equal(kept(num, xl_filter("q", "==", 10)), 1L)
  expect_equal(kept(num, xl_filter("q", list = c("10", "100"))), c(1L, 3L))
  # logicals display as TRUE/FALSE and match the same way
  expect_equal(kept(data.frame(b = c(TRUE, FALSE, TRUE)),
                    xl_filter("b", list = "TRUE")), c(1L, 3L))
})

test_that("a wildcard turns the same criteria into a typed comparison", {
  # a * or ? forces <customFilters>, which compares by type -- and no number is
  # text, so the wildcard matches nothing at all on a numeric column
  num <- data.frame(q = c(10, 20, 100, 15))
  expect_equal(kept(num, xl_filter("q", "==", "1*")), integer(0))
  # the pair that shows the two forms really do differ: same column, same
  # criteria, and the wildcard is the only change
  expect_equal(kept(num, xl_filter("q", "==", "10")), 1L)
})

test_that("a text comparison leaves number cells alone", {
  # the mirror of the numeric case: a number is not text, so it never equals a
  # text value -- which makes "!=" true for it
  num <- data.frame(q = c(10, 20, 100))
  expect_equal(kept(num, xl_filter("q", "!=", "a*")), c(1L, 2L, 3L))
  txt <- data.frame(f = c("apple", "banana", "cherry"), stringsAsFactors = FALSE)
  expect_equal(kept(txt, xl_filter("f", "!=", "a*")), c(2L, 3L))
})

test_that("a blank survives != but no other comparison", {
  # a blank is not equal to anything, so "!=" keeps it.  This is the one place
  # blanks are not silently excluded, and it is easy to get backwards.
  df <- data.frame(v = c("a", NA, "b"), stringsAsFactors = FALSE)
  expect_equal(kept(df, xl_filter("v", "!=", "a")), c(2L, 3L))
  # every other comparison drops it
  n <- data.frame(q = c(5, NA, 300))
  expect_equal(kept(n, xl_filter("q", ">", 1)), c(1L, 3L))
  expect_equal(kept(n, xl_filter("q", "==", 5)), 1L)
  # "non-blanks" is written as != " " but must still exclude blanks
  expect_equal(kept(df, xl_filter("v", "non-blanks")), c(1L, 3L))
})

test_that("== or == collapses into a two-value list", {
  # libxlsxwriter writes this as <filters> with both values, not a custom
  # filter, so it matches displayed text
  df <- data.frame(f = c("apple", "banana", "cherry"), stringsAsFactors = FALSE)
  f <- xl_filter("f", "==", "apple", "==", "cherry", and_or = "or")
  expect_equal(kept(df, f), c(1L, 3L))
  w <- xlsx_part(write_tmp(list(D = xl_sheet(df, filter = f))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(w, '<filter val="apple"/><filter val="cherry"/>', fixed = TRUE)
})

test_that("a value list matches a text cell that looks numeric", {
  # the mirror of the numeric column case: val="10" matches the string "10"
  # just as it matches the number 10
  df <- data.frame(t = c("10", "20", "abc"), stringsAsFactors = FALSE)
  expect_equal(kept(df, xl_filter("t", "==", 10)), 1L)
  expect_equal(kept(df, xl_filter("t", "==", "10")), 1L)
})

test_that("a custom filter compares numerically when its value is a number", {
  # val="10" is parsed as a number even when it was given as a string
  num <- data.frame(q = c(10, 20, 100, 15))
  expect_equal(kept(num, xl_filter("q", "!=", "10")), c(2L, 3L, 4L))
  # against a text column the same rule keeps everything: a text cell is not a
  # number, so it is never equal to one
  txt <- data.frame(t = c("10", "20", "abc"), stringsAsFactors = FALSE)
  expect_equal(kept(txt, xl_filter("t", "!=", 10)), c(1L, 2L, 3L))
  # and a magnitude comparison excludes every text cell
  expect_equal(kept(txt, xl_filter("t", ">", 15)), integer(0))
})

test_that("numeric comparisons behave as expected", {
  df <- data.frame(q = c(5, 150, 20, 300, 75))
  expect_equal(kept(df, xl_filter("q", ">", 100)), c(2L, 4L))
  expect_equal(kept(df, xl_filter("q", ">=", 75)), c(2L, 4L, 5L))
  expect_equal(kept(df, xl_filter("q", "<", 75)), c(1L, 3L))
  expect_equal(kept(df, xl_filter("q", "<=", 75)), c(1L, 3L, 5L))
  expect_equal(kept(df, xl_filter("q", "==", 150)), 2L)
  expect_equal(kept(df, xl_filter("q", "!=", 150)), c(1L, 3L, 4L, 5L))
  expect_equal(filter_hidden(df, xl_filter("q", ">", 100)), c(2L, 4L, 6L))
})

test_that("a mixed-type column is matched cell by cell", {
  # rows: 1 = number 10, 2 = text "20", 3 = text "abc", 4 = number 30,
  #       5 = number 20, 6 = empty string, 7 = NA
  mixed <- data.frame(row = 1:7)
  mixed$v <- xl_cell_general(value = list(10, "20", "abc", 30, 20, "", NA))

  # a magnitude comparison reaches the number cells only -- the text "20" in
  # row 2 is not a number, so it does not match even though it looks like one
  expect_equal(kept(mixed, xl_filter("v", ">", 15)), c(4L, 5L))
  # a text comparison reaches the text cells only
  expect_equal(kept(mixed, xl_filter("v", "==", "a*")), 3L)
  # both kinds of empty are blank
  expect_equal(kept(mixed, xl_filter("v", "blanks")), c(6L, 7L))
  expect_equal(kept(mixed, xl_filter("v", "non-blanks")), c(1L, 2L, 3L, 4L, 5L))
  # a value list matches displayed text, so val="20" catches the text "20" in
  # row 2 AND the number 20 in row 5
  expect_equal(kept(mixed, xl_filter("v", list = c("20", "abc"))),
               c(2L, 3L, 5L))
  expect_equal(filter_hidden(mixed, xl_filter("v", ">", 15)),
               c(2L, 3L, 4L, 7L, 8L))
})

test_that("a mixed column's rich strings and hyperlinks filter on their text", {
  df <- data.frame(row = 1:3)
  df$v <- xl_cell_general(value = list(
    xl_rich_string(xl_rich_run("ur", xl_font(bold = TRUE)), xl_rich_run("gent")),
    "calm", NA))
  expect_equal(kept(df, xl_filter("v", "==", "urgent")), 1L)

  h <- data.frame(row = 1:2)
  h$v <- xl_cell_general(hyperlink = c("https://a.example", "https://b.example"))
  expect_equal(kept(h, xl_filter("v", "==", "https://a.example")), 1L)
})

test_that("the R type of `value` does not change the file, so nor the match", {
  # == 10 and == "10" write byte-identical criteria, which is why they must
  # match identically; this pins the equivalence mechanically
  crit <- function(f) {
    w <- xlsx_part(write_tmp(list(D = xl_sheet(data.frame(q = c(10, 20)),
                                               filter = f))),
                   "xl/worksheets/sheet1.xml", raw = TRUE)
    regmatches(w, regexpr("<filterColumn.*</filterColumn>", w))
  }
  expect_identical(crit(xl_filter("q", "==", 10)),
                   crit(xl_filter("q", "==", "10")))
  expect_identical(crit(xl_filter("q", "!=", 10)),
                   crit(xl_filter("q", "!=", "10")))
})

test_that("list membership is case-insensitive too", {
  df <- data.frame(f = c("apple", "Apple", "banana", "BANANA"),
                   stringsAsFactors = FALSE)
  # Excel: all four stay visible for the list apple, BANANA
  expect_equal(kept(df, xl_filter("f", list = c("apple", "BANANA"))), 1:4)
  expect_equal(kept(df, xl_filter("f", list = "apple")), 1:2)
  expect_equal(filter_hidden(df, xl_filter("f", list = "apple")), c(4L, 5L))
})

test_that("two rules combine with and / or", {
  df <- data.frame(q = c(50, 120, 180, 250))
  expect_equal(kept(df, xl_filter("q", ">", 100, "<", 200)), 2:3)
  expect_equal(kept(df, xl_filter("q", "<", 100, ">", 200, and_or = "or")),
               c(1L, 4L))
})

test_that("dates filter on their underlying value", {
  df <- data.frame(when = as.Date("2024-01-01") + c(0, 91, 182, 274))
  expect_equal(kept(df, xl_filter("when", ">=", as.Date("2024-04-01"))), 2:4)
})

# ── The criteria reach the file ───────────────────────────────────────────────

test_that("criteria are written alongside the autofilter", {
  df <- data.frame(q = c(5, 150, 300))
  w <- xlsx_part(write_tmp(list(D = xl_sheet(df, filter = xl_filter("q", ">", 100)))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(w, "<autoFilter ", fixed = TRUE)
  expect_match(w, 'operator="greaterThan"', fixed = TRUE)
  expect_match(w, 'val="100"', fixed = TRUE)
})

test_that("a filter implies the autofilter", {
  # criteria without an autofilter range is rejected by libxlsxwriter, and is
  # meaningless anyway
  df <- data.frame(q = c(5, 150))
  w <- xlsx_part(write_tmp(list(D = xl_sheet(df, filter = xl_filter("q", ">", 100)))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(w, '<autoFilter ref="A1:A3">', fixed = TRUE)
})

test_that("a string filter is stored as a filter list", {
  # libxlsxwriter converts a string equality into <filters>, not <customFilter>
  df <- data.frame(f = c("apple", "banana"), stringsAsFactors = FALSE)
  w <- xlsx_part(write_tmp(list(D = xl_sheet(df, filter = xl_filter("f", "==", "apple")))),
                 "xl/worksheets/sheet1.xml", raw = TRUE)
  expect_match(w, '<filter val="apple"/>', fixed = TRUE)
})

test_that("filters on two columns combine with AND", {
  df <- data.frame(q = c(5, 150, 300, 200),
                   f = c("a", "b", "a", "a"), stringsAsFactors = FALSE)
  h <- filter_hidden(df, xl_filter("q", ">", 100), xl_filter("f", "==", "a"))
  # rows 3 (300, "a") and 4 (200, "a") satisfy both; row 1 fails on q and
  # row 2 on f, so sheet rows 2 and 3 go
  expect_equal(h, c(2L, 3L))
  expect_equal(kept(df, xl_filter("q", ">", 100)), c(2L, 3L, 4L))
})

test_that("row hiding survives every special column type", {
  # the filtered column is plain, but the sheet's other columns go through the
  # xl_cell_general machinery; hiding is driven by the row plan, so it must not
  # matter what else the row contains
  base <- data.frame(q = c(5, 150, 300))
  f <- xl_filter("q", ">", 100)
  expect_equal(filter_hidden(base, f), 2L)

  with_col <- function(v) {
    d <- base
    d$extra <- v
    d
  }
  expect_equal(filter_hidden(with_col(xl_cell_general(
    value = c("a", "b", "c"), format = xl_font(bold = TRUE))), f), 2L)
  expect_equal(filter_hidden(with_col(
    xl_formula(c("=A2*2", "=A3*2", "=A4*2"))), f), 2L)
  expect_equal(filter_hidden(with_col(
    xl_cell_general(value = "x", comment = "hi")), f), 2L)
  expect_equal(filter_hidden(with_col(
    xl_hyperlink("https://example.com")), f), 2L)
  expect_equal(filter_hidden(with_col(xl_cell_general(value = list(
    xl_rich_string(xl_rich_run("bo", xl_font(bold = TRUE)), xl_rich_run("ld")),
    xl_rich_string(xl_rich_run("it", xl_font(italic = TRUE)), xl_rich_run("al")),
    xl_rich_string(xl_rich_run("pl"), xl_rich_run("ain"))))), f), 2L)
})

# ── The exported predicate ────────────────────────────────────────────────────

test_that("xl_filter_keep returns one logical per row", {
  sales <- data.frame(fruit = c("apple", "banana", "cherry"),
                      qty = c(5, 150, 300), stringsAsFactors = FALSE)
  expect_identical(xl_filter_keep(sales, xl_filter("qty", ">", 100)),
                   c(FALSE, TRUE, TRUE))
  expect_length(xl_filter_keep(sales, xl_filter("qty", ">", 100)), nrow(sales))
})

test_that("xl_filter_keep combines a list with AND", {
  sales <- data.frame(fruit = c("apple", "banana", "cherry"),
                      qty = c(5, 150, 300), stringsAsFactors = FALSE)
  expect_identical(
    xl_filter_keep(sales, list(xl_filter("qty", ">", 100),
                               xl_filter("fruit", "==", "b*"))),
    c(FALSE, TRUE, FALSE))
  # no filters at all keeps everything
  expect_identical(xl_filter_keep(sales, list()), rep(TRUE, 3))
  expect_identical(xl_filter_keep(sales, NULL), rep(TRUE, 3))
})

test_that("xl_filter_keep agrees with the rows the file hides", {
  # the written file and the exported predicate share one implementation, and
  # this pins that they cannot drift apart
  df <- data.frame(q = c(5, 150, 20, 300, 75))
  f <- xl_filter("q", ">", 100)
  keep <- xl_filter_keep(df, f)
  # data row i is sheet row i + 1, since the header takes row 1
  expect_equal(filter_hidden(df, f), which(!keep) + 1L)
})

test_that("xl_filter_keep validates its arguments", {
  df <- data.frame(q = 1:3)
  expect_error(xl_filter_keep(1:3, xl_filter("q", ">", 1)),
               "`data` must be a data frame")
  expect_error(xl_filter_keep(df, "q>1"), "must be an xl_filter object")
  expect_error(xl_filter_keep(df, list(xl_filter("q", ">", 1), "nope")),
               "`filter\\[\\[2\\]\\]` must be an xl_filter object")
  expect_error(xl_filter_keep(df, xl_filter("nope", ">", 1)), "unknown column")
})

test_that("xl_filter_keep handles a zero-row data frame", {
  df <- data.frame(q = numeric(0))
  expect_identical(xl_filter_keep(df, xl_filter("q", ">", 1)), logical(0))
})

# ── Validation ────────────────────────────────────────────────────────────────

test_that("a filter on a formula column is refused, naming the row", {
  # writexl cannot know what Excel would compute, so it must not guess
  df <- data.frame(x = 1:2)
  df$f <- xl_cell_general(formula = c("=A1", "=A2"))
  expect_error(write_tmp(list(D = xl_sheet(df, filter = xl_filter("f", "==", 1)))),
               "row 1 holds a formula")
  # one formula among plain values is still enough to refuse
  d2 <- data.frame(x = 1:3)
  d2$f <- xl_cell_general(value = list(1, 2, NA),
                          formula = c(NA, NA, "=A2+A3"))
  expect_error(write_tmp(list(D = xl_sheet(d2, filter = xl_filter("f", ">", 1)))),
               "row 3 holds a formula")
})

test_that("a date cannot be filtered by exact value or list", {
  # <filters> matches displayed text, and the displayed text of a date depends
  # on the number format, so writexl refuses rather than guessing
  df <- data.frame(when = as.Date("2024-01-01") + 0:2)
  expect_error(write_tmp(list(D = xl_sheet(df,
    filter = xl_filter("when", list = "2024-01-01")))), "holds dates")
  expect_error(write_tmp(list(D = xl_sheet(df,
    filter = xl_filter("when", "==", as.Date("2024-01-01"))))), "holds dates")
  # a comparison is fine: it is a typed numeric comparison
  expect_equal(kept(df, xl_filter("when", ">=", as.Date("2024-01-02"))), 2:3)
})

test_that("an unorderable text comparison is refused, not guessed", {
  df <- data.frame(f = c("apple", "banana"), stringsAsFactors = FALSE)
  expect_error(kept(df, xl_filter("f", ">", "apple")),
               "cannot predict how Excel orders text")
})

test_that("filter arguments are validated", {
  expect_error(xl_filter(), "must name or index")
  expect_error(xl_filter("q"), "needs a `criteria` or a `list`")
  expect_error(xl_filter("q", ">"), "needs a `value`")
  expect_error(xl_filter("q", "~", 1), "criteria")
  expect_error(xl_filter("q", "==", 1, list = "a"), "not both")
  expect_error(xl_filter("q", list = character(0)), "character vector")
  expect_error(xl_filter("q", list = c("a", NA)), "character vector")
  expect_error(xl_filter("q", ">", 1, and_or = "xor"), "and_or")
})

test_that("an unknown column is reported", {
  df <- data.frame(q = 1:2)
  expect_error(write_tmp(list(D = xl_sheet(df, filter = xl_filter("nope", ">", 1)))),
               "unknown column")
})

test_that("a magnitude comparison needs a numeric column", {
  df <- data.frame(f = c("a", "b"), stringsAsFactors = FALSE)
  # no coercion, so this simply matches nothing rather than erroring
  expect_equal(kept(df, xl_filter("f", ">", 1)), integer(0))
})

test_that("non-filter objects are rejected", {
  expect_error(write_tmp(list(D = xl_sheet(data.frame(q = 1), filter = "q>1"))),
               "must be an xl_filter object")
})

test_that("the print method runs", {
  expect_output(print(xl_filter("q", ">", 100)), "xl_filter")
  expect_output(print(xl_filter("f", list = c("a", "b"))), "2 value")
})

# ── The branches nothing else reached ────────────────────────────────────────

test_that("a hyperlink cell filters on the text Excel shows", {
  df <- data.frame(site = c("a", "b"), stringsAsFactors = FALSE)
  # a hyperlink with no value of its own displays its URL, so that is what a
  # filter has to match
  df$site <- c(xl_cell_general(hyperlink = "https://example.com"),
               xl_cell_general(hyperlink = list(url = "https://r-project.org")))
  keep <- xl_filter_keep(df, xl_filter(col = "site",
                                       list = "https://example.com"))
  expect_equal(keep, c(TRUE, FALSE))
  keep <- xl_filter_keep(df, xl_filter(col = "site",
                                       list = "https://r-project.org"))
  expect_equal(keep, c(FALSE, TRUE))
})

test_that("a date column compares as a number, not as its text", {
  df <- data.frame(when = as.Date(c("2024-01-01", "2024-06-01", "2024-12-01")))
  expect_equal(xl_filter_keep(df, xl_filter(col = "when", criteria = ">",
                                            value = as.Date("2024-05-01"))),
               c(FALSE, TRUE, TRUE))
  expect_true(.filter_numeric_value(as.Date("2024-01-01")))
  expect_true(.filter_numeric_value(as.POSIXct("2024-01-01", tz = "UTC")))
  expect_true(.filter_numeric_value("42"))
  expect_false(.filter_numeric_value(NULL))
  expect_false(.filter_numeric_value("a*"))
  expect_false(.filter_numeric_value("apple"))
})

test_that("blanks and non-blanks are their own criteria", {
  df <- data.frame(a = c("x", NA, ""), stringsAsFactors = FALSE)
  expect_equal(xl_filter_keep(df, xl_filter(col = "a", criteria = "blanks")),
               c(FALSE, TRUE, TRUE))
  expect_equal(xl_filter_keep(df, xl_filter(col = "a", criteria = "non-blanks")),
               c(TRUE, FALSE, FALSE))
})

test_that("a second criteria needs its own value", {
  expect_error(xl_filter(col = "a", criteria = ">", value = 1,
                         criteria2 = "<"),
               "needs a `value2`")
})

test_that("blanks reaches the custom rule when it is the second criteria", {
  # a lone "blanks" is a standard filter; paired with a comparison under "and"
  # it becomes a custom rule, which is a different evaluator
  df <- data.frame(a = c("x", NA, "y"), stringsAsFactors = FALSE)
  expect_equal(xl_filter_keep(df, xl_filter(col = "a", criteria = "!=",
                                            value = "x", criteria2 = "blanks",
                                            and_or = "or")),
               c(FALSE, TRUE, TRUE))
  # "blanks" AND "non-blanks" is the empty set, and each half is evaluated
  expect_equal(xl_filter_keep(df, xl_filter(col = "a", criteria = "blanks",
                                            criteria2 = "non-blanks",
                                            and_or = "and")),
               c(FALSE, FALSE, FALSE))
})

test_that("a date inside a mixed cell column compares as a number", {
  df <- data.frame(a = 1:3)
  df$when <- xl_cell_general(value = list(as.Date("2024-01-01"),
                                          as.Date("2024-06-01"),
                                          as.Date("2024-12-01")))
  expect_equal(xl_filter_keep(df, xl_filter(col = "when", criteria = ">",
                                            value = as.Date("2024-05-01"))),
               c(FALSE, TRUE, TRUE))
  # and a value list against a date column is refused, not guessed
  expect_error(xl_filter_keep(df, xl_filter(col = "when",
                                            list = "2024-01-01")),
               "date")
})

---
title: "Charts and images"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Charts and images}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
sales <- data.frame(quarter = c("Q1", "Q2", "Q3", "Q4"),
                    revenue = c(120, 145, 133, 180),
                    cost    = c(90, 110, 105, 130))
```

## A chart

`xl_chart()` puts a chart on a sheet; `xl_chart_series()` says what it plots.

```{r}
sheet <- xl_sheet(sales,
                  chart = xl_chart("column",
                                   xl_chart_series(values = list(cols = "revenue")),
                                   title = "Revenue by quarter", at = "E2"))
path <- write_xlsx(list(Sales = sheet), tempfile(fileext = ".xlsx"))
```

All 22 types Excel offers are available: `"column"`, `"bar"`, `"line"`,
`"pie"`, `"doughnut"`, `"area"`, `"scatter"`, `"radar"`, and the stacked,
percent-stacked, smoothed and marker variants.

### Naming a range

A series range is written either **by column**, `list(cols = "revenue")`, or as
**A1 text**, `"Data!B2:B5"`. Prefer the column form: it is checked against the
data frame, moves with the column rather than with a letter, and accounts for
the header row itself. `rows` narrows it, and `sheet` points at another sheet:

```{r}
xl_chart_series(values = list(sheet = "Sales", cols = "revenue", rows = 1:2))
```

A range reaching past the data is an error rather than a chart with blanks on
the end --- a chart of the wrong numbers looks plausible, which is the worst
way to be wrong.

A series that plots a column is **named after that column's header**, which is
what Excel does when you chart a column along with its header. The header is
referenced, not copied, so renaming it in Excel renames the series.
`name = FALSE` leaves it unnamed, and a string is always taken literally --- a
series may legitimately be called `"Q1!"` --- so to take a name from a cell,
give a spec: `list(header = "cost")`, or `list(rows = 1, cols = 1)`.

### Styling

Charts reuse the format objects from [Formatting cells](b-formatting.html).
`xl_border()` becomes the line, `xl_fill()` the fill or pattern, `xl_font()`
the text:

```{r}
xl_chart("line",
         xl_chart_series(values = list(cols = "revenue"),
                         format = xl_border(all = "dashed", color = "#4472C4")),
         title = "Revenue", title_format = xl_font(size = 14, bold = TRUE))
```

Each part takes the groups it can use and **refuses the rest by name**. A
series is a shape, so it takes a line and a fill but no font; a title is text,
so it takes a font and nothing else. `xl_fill()` and `xl_border()` also take
`transparency`, which charts honour and cells ignore, and chart lines have
fewer dash styles than cell borders because a chart line has no width. None of
this is dropped quietly.

`style` picks one of Excel's 48 built-in chart styles.

## Axes

`xl_chart_axis()` covers everything an axis has --- the title and its font,
tick labels with their number format, bounds, a log scale, tick marks and their
spacing, gridlines, and the axis line itself:

```{r}
xl_chart("column", xl_chart_series(values = list(cols = "revenue")),
         x_axis = xl_chart_axis(title = "Quarter", major_tick = "none"),
         y_axis = xl_chart_axis(title = "Revenue", min = 0,
                                num_format = "$#,##0",
                                major_gridlines_format =
                                  xl_border(all = "dotted", color = "gray")))
```

There are two kinds of axis and several options belong to one of them. A
**scatter** chart plots numbers against numbers, so both its axes are value
axes; every other type has a **category** x axis and a value y axis --- bar
charts included, since Excel draws their categories up the side but does not
rename the axes.

| value axes | category axes |
|---|---|
| `min`, `max`, `log_base`, `major_unit`, `minor_unit`, `display_units`, `display_units_visible` | `position`, `label_align`, `interval_unit`, `interval_tick` |

Setting a minimum on a category axis is accepted by Excel and then ignored, so
writexl refuses it and names the axis that would work. Pie and doughnut charts
have no axes at all.

## The parts of a series

A series can carry a marker at each point, labels beside them, a trendline
through them and error bars on them, and individual points can be styled:

```{r}
xl_chart_series(
  values       = list(cols = "revenue"),
  marker       = xl_chart_marker(type = "circle", size = 7),
  labels       = xl_chart_labels(num_format = "$#,##0", position = "above"),
  trendline    = xl_chart_trendline("linear", equation = TRUE),
  y_error_bars = xl_chart_error_bars("percentage", 5))
```

`xl_chart_labels()` with no arguments prints each point's value, which is
Excel's default. Naming any part means the label holds **exactly** those, so
`show_percentage = TRUE` alone gives a percentage with no value beside it.
`custom =` takes a list of `xl_chart_label()`s, one per point, so a single
label can be renamed or removed with `hide = TRUE`. `points =` takes a list of
formats the same way --- one red slice, the rest untouched:

```{r}
xl_chart("pie", xl_chart_series(values = list(cols = "revenue"),
                                points = list(xl_fill(background = "red"),
                                              NULL, NULL, NULL)))
```

Four more of Excel's silent rules are checked here: which **label positions**
each chart type allows, that a **moving average** has no forecast, equation or
R-squared, that an **intercept** applies only to exponential, linear and
polynomial fits, and that an **automatic marker** takes neither a size nor a
format.

## The chart as a whole

```{r}
xl_chart("column", xl_chart_series(values = list(cols = "revenue")),
         legend     = xl_chart_legend(position = "bottom"),
         data_table = xl_chart_table(show_keys = TRUE),
         plot_area_format  = xl_fill(background = "#F8F8F8"),
         chart_area_format = xl_border(all = "thin", color = "gray"))
```

`xl_chart_legend(position = "none")` removes the legend, and `delete_series`
leaves one series out of the key while still plotting it. Anything placed by
hand takes a `layout` rather than an `at` --- `title_layout`,
`plot_area_layout`, `xl_chart_axis(title_layout =)` and
`xl_chart_legend(layout =)` are fractions of the chart, where `at` means a cell
everywhere else. `show_blanks` decides what an empty cell does, and
`show_hidden_data` plots rows Excel would leave out.

Six options belong to one family of chart: the doughnut `hole_size`, pie
`rotation`, `drop_lines`, `high_low_lines`, `up_down_bars`, and the bar
`series_gap` and `series_overlap`. Each takes `TRUE` for Excel's own drawing or
a format to style it, and each is refused on a type that has no such feature.

A scatter series must have `categories` --- they are its x axis, and
libxlsxwriter crashes without them.

## A chart on a sheet of its own

`xl_chartsheet()` is a tab holding one chart and no cells:

```{r}
chart <- xl_chart("column",
                  xl_chart_series(values = list(sheet = "Data",
                                                cols = "revenue"),
                                  categories = list(sheet = "Data",
                                                    cols = "quarter")),
                  title = "Revenue")
path <- write_xlsx(list(Data = sales, Overview = xl_chartsheet(chart)),
                   tempfile(fileext = ".xlsx"))
```

Every range must name its sheet, since a chartsheet has no cells for
`list(cols = )` to resolve against. It supports a fraction of what a worksheet
does --- of `xl_page_setup()` the orientation, paper, margins, header and
footer; of `xl_sheet_view()` the four tab-state options --- plus `tab_color`,
`zoom` and `protect`. The rest is refused rather than dropped.

## Images

`xl_image()` places a picture, anchored to a cell:

```{r eval = isTRUE(capabilities("png"))}
card <- as.raster(matrix(c("red", "blue", "green", "gold"), nrow = 2))
xl_sheet(sales, image = xl_image(card, at = "G2", scale = 20))
```

The image may be a **file path**, a **raw vector** of encoded bytes, or an
**in-memory picture** --- a `raster`, a colour matrix, an RGB array or a
`nativeRaster`. Anything `rasterImage()` can draw can be written, so a plot
never has to touch the disk; that path uses R's PNG device, which is why
`grDevices` is a suggested package.

PNG, JPEG, GIF and BMP are supported, and the format is read from the file's
own bytes rather than its extension, so a mislabelled `.png` is caught rather
than misfiled. `scale`, `offset` and `position` control size and placement,
`description` and `decorative` set the alt text, and `url` makes the image a
hyperlink.

An image can also go in the printed header or footer, matched to a `&G`
placeholder, or be tiled behind the cells as a screen watermark:

```{r eval = FALSE}
xl_page_setup(header = "&L&G", header_image = list(left = logo))
xl_sheet(sales, background_image = logo)
```

`embed = TRUE` puts the image *inside* a cell, the Excel 365 "place in cell"
feature; older Excel shows `#VALUE!` in its place.

Two combinations are refused, because libxlsxwriter miscounts them and Excel
repairs the file: an embedded image alongside any other image, and a
header/footer or background image on a sheet *before* one with a floating image
or a chart. Both errors name the sheets and the arrangement that works.

---
title: "Formatting cells"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Formatting cells}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
```

One kind of object describes appearance everywhere in writexl: an `xl_format`.
The same object styles a cell, a whole column, a chart series or a conditional
rule, so there is one vocabulary to learn rather than five.

## Building a format

Six group constructors, each covering one family of Excel properties. Every
argument defaults to unset, so you name only what you want to change:

```{r}
xl_font(bold = TRUE, color = "navy", size = 12)
xl_fill(background = "#FFF2CC")
xl_border(all = "thin", color = "gray")
xl_align(horizontal = "center", vertical = "top", wrap = TRUE)
xl_num_format("#,##0.00")
xl_protection(locked = FALSE)
```

Colours accept R colour names, hex strings or integers, via `xl_color()`.

Each constructor already returns a complete format, so one can be used alone.
Combine them with `+`, which merges **property by property** --- later values
win, but nothing you did not touch is lost:

```{r}
money <- xl_num_format("$#,##0.00") + xl_font(color = "darkgreen")
money + xl_font(bold = TRUE)     # still green, now bold
```

## Applying it to cells

`xl_cell_general()` attaches a format to values. One format covers the column;
a list gives a format per cell:

```{r}
df <- data.frame(item = c("Widget", "Gadget", "Gizmo"))
df$price <- xl_cell_general(value = c(1234.5, 67.25, 890), format = money)
df$flag  <- xl_cell_general(
  value  = c(10, -5, 3),
  format = list(xl_fill(background = "lightgreen"),
                xl_fill(background = "salmon"),
                xl_fill(background = "lightgreen")))
path <- write_xlsx(df, tempfile(fileext = ".xlsx"))
```

A cell with a format and no value is a formatted blank --- useful for ruling a
sheet out to a fixed size:

```{r}
xl_cell_general(value = NA, format = xl_fill(background = "#EEEEEE"))
```

Dates and times take the workbook's date format automatically when their own
format sets no number format.

## Applying it to columns and rows

More often you want a whole column. `xl_col_spec()` and `xl_row_spec()` *are*
`xl_format` objects that additionally carry a target and some geometry, so they
combine with `+` like any other format:

```{r}
sheet <- xl_sheet(
  data.frame(date = as.Date("2024-01-01") + 0:2,
             revenue = c(1000.5, 2000.25, 1500.75)),
  cols = list(xl_col_spec("date", width = 12),
              xl_col_spec("revenue", width = 14, format = money)),
  rows = xl_row_spec(1, height = 20))
path <- write_xlsx(list(Sales = sheet), tempfile(fileext = ".xlsx"))
```

Columns are named or indexed; rows are counted from 1 as *data* rows, ignoring
the header.

## Where a format comes from

Formats cascade. What Excel finally applies to a cell is the workbook's
`default_format`, merged with the column or row format, merged with the cell's
own. Identical results are written once, so styling thousands of cells stays
compact. See [Worksheets and workbooks](c-worksheets-workbooks.html) for the workbook end of
that.

## Conditional formatting

A conditional format is a rule plus a format, applied to a range and evaluated
by Excel rather than by you:

```{r}
scores <- data.frame(name = c("a", "b", "c", "d"), score = c(35, 78, 55, 92))
sheet <- xl_sheet(scores, conditional = list(
  xl_cond_cell(list(cols = "score"), "cell", ">=", 80,
               format = xl_fill(background = "lightgreen")),
  xl_cond_cell(list(cols = "score"), "cell", "<", 50,
               format = xl_fill(background = "salmon"))))
path <- write_xlsx(list(Scores = sheet), tempfile(fileext = ".xlsx"))
```

Rules are tried in order, and `stop_if_true` stops at the first match.

The format on a rule is a **differential** format: it says what to change, not
what the cell looks like. A rule setting only a fill leaves the font alone,
which is why these are stored apart from ordinary cell styles.

`type` picks what is being tested --- a cell's value, whether it is in the top
or bottom *n*, above or below average, a duplicate, blank, an error, text
containing something, a date in a relative window, or a formula of your own.

### Scales, bars and icons

Three constructors cover Excel's graphical rules, none of which needs a format:

```{r}
xl_cond_scale(list(cols = "score"), colors = c("salmon", "lightgreen"))
xl_cond_bar(list(cols = "score"), color = "steelblue")
xl_cond_icons(list(cols = "score"), style = "3_traffic_lights")
```

A two-colour scale takes two colours, a three-colour scale three. Data bars
draw behind the value in the cell. Icon sets take one of Excel's built-in
`style`s, optionally reversed or shown without their values.

Every one of these takes its range the same way as anything else in writexl:
`list(cols = )` against the data frame, or A1 text.

## What is elsewhere

* The same format objects style **charts**, translated to what a chart can
  draw: [Charts and images](d-charts-images.html).
* **Number formats** are ordinary Excel format strings --- anything the Custom
  box in Excel accepts.
* `xl_protection()` only takes effect once the sheet itself is protected: see
  [Worksheets and workbooks](c-worksheets-workbooks.html).

---
title: "Worksheets and workbooks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Worksheets and workbooks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
```

`xl_sheet()` wraps a data frame with everything that is true of the sheet
rather than of a cell. `xl_workbook()` does the same one level up.

## The sheet

```{r}
sheet <- xl_sheet(
  data.frame(id = 1:3, note = c("a", "b", "c")),
  cols      = xl_col_spec("note", width = 30),
  freeze    = "A2",          # keep the header row visible
  tab_color = "steelblue",
  zoom      = 120,
  gridlines = FALSE)
path <- write_xlsx(list(Notes = sheet), tempfile(fileext = ".xlsx"))
```

`auto_colwidth = TRUE` sizes every column to its contents. The xlsx format has
no true AutoFit, so this is a character-count estimate; columns given an
explicit width are left alone.

Widths and heights can be given in **pixels** instead of Excel's character
units and points, which is easier to reason about when sizing a column around
an image:

```{r}
xl_col_spec("thumbnail", width_pixels = 100)
xl_row_spec(1, height_pixels = 40)
```

Excel stores the character units, so the value is converted on the way in --- a
column set to 100 pixels reads back as 13.57 units and renders at 100 pixels
again.

### Grouping and outlines

`level` puts columns or rows into an outline group, which Excel draws with the
+/- controls that collapse it. `xl_outline()` changes how those controls are
*drawn* --- never which rows are grouped --- and its defaults already match
Excel's:

```{r}
xl_sheet(data.frame(q1 = 1:2, q2 = 3:4, total = 4:5),
         cols    = xl_col_spec(c("q1", "q2"), level = 1),
         outline = xl_outline(symbols_right = FALSE))
```

### Protection

```{r}
xl_sheet(data.frame(id = 1:3),
         protect = list(password = "secret", sort = TRUE))
```

`protect = TRUE` locks everything; a list allows named actions through. Cell
locking set with `xl_protection()` only takes effect once the sheet is
protected, and `locked = FALSE` is how a cell is left editable on a protected
sheet.

### Silencing Excel's error indicators

Excel puts a green triangle in cells it thinks are wrong, most often a number
deliberately stored as text:

```{r}
xl_sheet(data.frame(zip = c("01234", "02138")),
         ignore_errors = list(number_stored_as_text = "A2:A3"))
```

Nine error types can be silenced, each taking a range in any spelling the other
range arguments accept.

## The tab strip and the opening view

`xl_sheet_view()` collects the rarely-used settings that would otherwise widen
`xl_sheet()` past reading:

```{r}
xl_sheet(data.frame(x = 1),
         view = xl_sheet_view(active = TRUE, selection = "B2",
                              hide_zero = TRUE))
```

Which tab is active, which are selected, which are hidden, which is leftmost,
where the sheet is scrolled to, whether zeros show, right-to-left column order,
and split panes all live here. `freeze` stays on `xl_sheet()` because keeping
the header row visible is the one worksheet option almost everyone wants.

Excel's rules about tabs span the whole workbook, so they are checked before
anything is written: at most one active sheet, no hidden-and-active sheet, and
at least one sheet visible. The error names the sheet at fault.

## Printing

`xl_page_setup()` collects everything Excel's Page Layout ribbon controls. None
of it affects the cells:

```{r}
xl_sheet(data.frame(x = 1:3),
         page = xl_page_setup(
           orientation = "landscape", paper = "A4",
           fit_to = c(1, 0),                 # one page wide, any number tall
           center_horizontally = TRUE,
           header = "&LQuarterly report&RPage &P of &N",
           repeat_rows = 1,                  # the header row, on every page
           print_area = "A1:C4"))
```

Headers and footers use Excel's own `&`-codes: `&L`/`&C`/`&R` for the three
sections, `&P` and `&N` for the page number and count, `&D` for the date, `&G`
for a picture. Manual page breaks go in with `h_breaks` and `v_breaks`.

## The workbook

`xl_workbook()` binds sheets to `xl_properties()`, which holds the document
metadata and the formatting defaults that would otherwise be built in:

```{r}
wb <- xl_workbook(
  list(Report = data.frame(quarter = c("Q1", "Q2"), sales = c(12000, 15000))),
  properties = xl_properties(
    title   = "Quarterly Report",
    author  = "Finance Team",
    custom  = list(Project = "Alpha", Reviewed = TRUE),
    header_format = xl_font(bold = TRUE, color = "white") +
                    xl_fill(background = "navy")))
path <- write_xlsx(wb, tempfile(fileext = ".xlsx"))
```

Because the defaults are ordinary `xl_format` objects you can change any of
them: the header row's style, the date and datetime formats and their column
widths, the hyperlink style, and a workbook-wide `default_format` beneath every
cell. Custom properties may be text, numbers, dates or logicals.

A workbook-wide `default_format` is *emulated* --- it is merged beneath every
cell --- because libxlsxwriter has no native "Normal style" hook. The effect is
the same; the file is slightly larger.

## Memory

For a large workbook writexl can stream each row to disk as it is written
rather than holding the whole sheet in memory. It decides per workbook:

```{r eval = FALSE}
write_xlsx(big, path, constant_memory = TRUE)
```

Left at `NA` it turns streaming on only when the workbook is large enough for
the saving to matter --- estimated from the cell count, with the threshold
configurable through `constant_memory_threshold`. Some features cannot be
written while streaming, because they touch cells above the row being written:
worksheet tables, multi-cell array formulas, merged ranges and embedded images.
Those turn it off, and asking for it anyway warns rather than producing a
half-written file.

## Elsewhere

* Styling the cells themselves: [Formatting cells](b-formatting.html)
* Charts, chartsheets and pictures: [Charts and images](d-charts-images.html)
* Validation, filters, tables and merges: [Formulas, tables and the
  rest](e-formulas-and-more.html)

---
title: "Everything at once"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Everything at once}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
```

This vignette is one runnable script. It exercises every part of the package
into **two** workbooks, so that reviewing what writexl can do means opening two
files rather than reading five vignettes.

Two files rather than one because of a single constraint: an embedded image
cannot share a workbook with any other image, so it gets its own. Everything
else fits together.

Each section says which vignette explains it.

```{r}
sales <- data.frame(
  quarter = c("Q1", "Q2", "Q3", "Q4"),
  revenue = c(120000, 145000, 133000, 180000),
  cost    = c(90000, 110000, 105000, 130000),
  region  = c("North", "South", "North", "South"),
  stringsAsFactors = FALSE)
```

## Formatting --- [vignette](b-formatting.html)

```{r}
money <- xl_num_format("$#,##0") + xl_font(color = "darkgreen")

styled <- xl_sheet(
  sales,
  cols = list(xl_col_spec("quarter", width = 10),
              xl_col_spec(c("revenue", "cost"), width = 14, format = money)),
  rows = xl_row_spec(1, height = 18),
  conditional = list(
    xl_cond_cell(list(cols = "revenue"), "cell", ">=", 1.4e5,
                 format = xl_fill(background = "lightgreen")),
    xl_cond_bar(list(cols = "cost"), color = "steelblue")),
  freeze = "A2", tab_color = "steelblue")
```

## Cell content --- [vignette](e-formulas-and-more.html)

```{r}
cells <- data.frame(item = c("Widget", "Gadget", "Gizmo"),
                    stringsAsFactors = FALSE)
cells$price   <- xl_cell_general(value = c(1234.5, 67.25, 890), format = money)
cells$margin  <- xl_formula(sprintf("=B%d*0.3", 2:4))
cells$link    <- xl_hyperlink(rep("https://ropensci.org", 3), "rOpenSci")
cells$note    <- xl_cell_general(
  value   = c("ok", NA, "check"),
  comment = list(xl_comment("Reviewed", author = "QA"), NULL, NULL),
  format  = list(NULL, xl_fill(background = "#EEEEEE"), NULL))
# a rich string is one cell's worth of text, so it goes in a list alongside
# the other rows' values
cells$rich    <- xl_cell_general(value = list(
  xl_rich_string("Plain ", xl_rich_run("bold", xl_font(bold = TRUE)),
                 " and ", xl_rich_run("red", xl_font(color = "red"))),
  "plain text", NA))

content <- xl_sheet(
  cells,
  merge = xl_merge("A5:C5", "A merged heading",
                   format = xl_align(horizontal = "center") +
                            xl_font(bold = TRUE)),
  validation = xl_validation(list(cols = "item"), type = "list",
                             list = c("Widget", "Gadget", "Gizmo"),
                             input_title = "Pick one"),
  ignore_errors = list(number_stored_as_text = "A2:A4"))
```

## Worksheet features --- [vignette](e-formulas-and-more.html)

```{r}
filtered <- xl_sheet(sales, filter = xl_filter("region", "==", "North"))

tabled <- xl_sheet(sales[, c("quarter", "revenue")],
                   table = xl_table(name = "Revenue", style = "medium 9",
                                    total_row = TRUE,
                                    columns = xl_table_column("revenue",
                                                              total = "sum")))
```

## Printing and the sheet view --- [vignette](c-worksheets-workbooks.html)

```{r}
printed <- xl_sheet(
  sales,
  page = xl_page_setup(orientation = "landscape", paper = "A4",
                       fit_to = c(1, 0), center_horizontally = TRUE,
                       header = "&LwritexlShowcase&RPage &P of &N",
                       repeat_rows = 1),
  view = xl_sheet_view(hide_zero = TRUE),
  outline = xl_outline(symbols_below = FALSE),
  protect = list(password = "secret", sort = TRUE))
```

## Charts --- [vignette](d-charts-images.html)

```{r}
charted <- xl_sheet(
  sales,
  chart = list(
    xl_chart("column",
             list(xl_chart_series(values = list(cols = "revenue"),
                                  categories = list(cols = "quarter"),
                                  labels = xl_chart_labels(num_format = "$#,##0",
                                                           position = "outside_end")),
                  xl_chart_series(values = list(cols = "cost"),
                                  categories = list(cols = "quarter"))),
             title = "Revenue and cost",
             title_format = xl_font(size = 13, bold = TRUE),
             x_axis = xl_chart_axis(title = "Quarter"),
             y_axis = xl_chart_axis(title = "Dollars", min = 0,
                                    num_format = "$#,##0"),
             legend = xl_chart_legend(position = "bottom"),
             series_gap = 60, at = "G2"),
    xl_chart("pie",
             xl_chart_series(values = list(cols = "revenue"),
                             categories = list(cols = "quarter"),
                             labels = xl_chart_labels(show_percentage = TRUE),
                             points = list(xl_fill(background = "#C00000"),
                                           NULL, NULL, NULL)),
             title = "Share of revenue", at = "G20"),
    xl_chart("scatter_straight_markers",
             xl_chart_series(values = list(cols = "revenue"),
                             categories = list(cols = "cost"),
                             marker = xl_chart_marker("circle", size = 7),
                             trendline = xl_chart_trendline("linear",
                                                            equation = TRUE),
                             y_error_bars = xl_chart_error_bars("percentage", 5)),
             title = "Revenue against cost", at = "P2")))

overview <- xl_chartsheet(
  xl_chart("line",
           xl_chart_series(values = list(sheet = "Charts", cols = "revenue"),
                           categories = list(sheet = "Charts",
                                             cols = "quarter")),
           title = "Revenue, full page",
           y_axis = xl_chart_axis(title = "Revenue", min = 0)),
  tab_color = "red")
```

## An image --- [vignette](d-charts-images.html)

```{r}
have_png <- isTRUE(capabilities("png")) &&
  requireNamespace("grDevices", quietly = TRUE)
```

```{r eval = have_png}
card <- as.raster(matrix(c("#4472C4", "#ED7D31", "#A5A5A5", "#FFC000"),
                         nrow = 2))
pictured <- xl_sheet(sales[, 1:2], image = xl_image(card, at = "D2",
                                                    scale = 30,
                                                    description = "A test card"))
```

```{r eval = !have_png, echo = FALSE}
pictured <- xl_sheet(sales[, 1:2])
```

## The workbook --- [vignette](c-worksheets-workbooks.html)

```{r}
wb <- xl_workbook(
  list(Formatted = styled,
       Content   = content,
       Filtered  = filtered,
       Table     = tabled,
       Printed   = printed,
       Charts    = charted,
       Picture   = pictured,
       Overview  = overview),
  properties = xl_properties(
    title    = "writexl showcase",
    author   = "writexl",
    subject  = "Every feature in one workbook",
    custom   = list(Generated = Sys.Date(), Reviewed = FALSE),
    header_format = xl_font(bold = TRUE, color = "white") +
                    xl_fill(background = "#1F3864")))

showcase <- write_xlsx(wb, tempfile(fileext = ".xlsx"))
```

The sheet order matters: any sheet with a header, footer or background image
must come *after* every sheet carrying a floating image or a chart, because
libxlsxwriter numbers the drawing relationships from a counter that skips the
first kind. writexl checks the order and refuses rather than writing a file
Excel repairs.

## The second workbook: an embedded image

An embedded image sits *inside* a cell and sizes with it. libxlsxwriter
numbers its metadata from a counter shared with every other image, so mixing
the two writes a cell reference past the end of the metadata and Excel repairs
the sheet. writexl refuses that combination, which is why this is a second
file:

```{r eval = have_png}
embedded <- write_xlsx(
  list(Embedded = xl_sheet(sales[, 1:2],
                           image = xl_image(card, at = "D2", embed = TRUE))),
  tempfile(fileext = ".xlsx"))
```

## What was written

```{r}
sheets <- c("Formatted", "Content", "Filtered", "Table", "Printed",
            "Charts", "Picture", "Overview")
data.frame(sheet = sheets,
           shows = c("formats, conditional rules, frozen pane",
                     "formulas, links, comments, rich text, merge, validation",
                     "an autofilter with its rows hidden",
                     "a worksheet table with a total row",
                     "page setup, outline symbols, protection",
                     "a column, a pie and a scatter chart",
                     "a picture anchored to a cell",
                     "a chartsheet: one chart, no cells"))
```

```{r}
file.size(showcase)
```

Open it and every feature above is visible in one place. For why any of it
behaves as it does, follow the links back to
[formatting](b-formatting.html),
[worksheets and workbooks](c-worksheets-workbooks.html),
[charts and images](d-charts-images.html), or
[formulas, tables and the rest](e-formulas-and-more.html).

---
title: "Getting started with writexl"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with writexl}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
```

writexl writes data frames to xlsx. It needs no Java and no Excel, and for the
common case it needs nothing from you either.

## One data frame

```{r}
path <- write_xlsx(iris, tempfile(fileext = ".xlsx"))
```

`write_xlsx()` returns the path it wrote, so it composes with anything that
takes a file name.

## Several sheets

A named list becomes one sheet per element, and the names become the tabs:

```{r}
path <- write_xlsx(list(Flowers = iris, Cars = mtcars[1:5, 1:4]),
                   tempfile(fileext = ".xlsx"))
```

Sheet names are repaired rather than rejected: Excel forbids `[ ] : * ? / \`,
limits a name to 31 characters, and refuses duplicates. writexl fixes each of
those and warns, naming the original and the replacement, because a sheet name
is often taken from data and losing the whole export over one stray `/` would
be unkind.

## What is written

Column types map to Excel's own:

| R | Excel |
|---|---|
| `character` | text |
| `numeric`, `integer` | number |
| `logical` | TRUE/FALSE |
| `Date`, `POSIXct` | a formatted date or datetime |
| `factor` | its labels, as text |
| `difftime` | a number of seconds |

`NA` becomes an empty cell. A column of a type xlsx has no representation for
--- `complex`, `raw`, a plain list column --- is an error rather than a silent
approximation.

`na` writes something else in place of an empty cell, for `NaN` as well as
`NA`. It keeps its own type, so a number stays a number:

```{r}
tmp <- write_xlsx(data.frame(x = c(1.5, NA)), na = "not measured")
```

Set it for the whole workbook with `xl_properties(na = )`, for one column with
`xl_col_spec(na = )`, or for one cell with `xl_cell_general(na = )`; the
innermost one that is set wins. Substituting a string into a numeric column
makes that column mixed, so reading it back gives character.

Excel has no concept of a time zone, so `POSIXct` needs a decision. writexl
makes it once for the whole workbook: if every value shares one zone the local
wall-clock reading is written, and if they differ everything is converted to
UTC and you are warned. Either way nothing is mislabelled.

The header row is written by default and styled bold; `col_names = FALSE` and
`format_headers = FALSE` turn each off.

## Everything else

The rest of the package is optional. Nothing below is needed to write a
workbook, and each has a vignette of its own:

* **[Formatting cells](b-formatting.html)** --- fonts, fills, borders, number
  formats and alignment, built from group constructors that combine with `+`,
  plus conditional formatting.

* **[Worksheets and workbooks](c-worksheets-workbooks.html)** --- column widths and row
  heights, frozen panes, tab colours, printing and page setup, and the document
  properties.

* **[Charts and images](d-charts-images.html)** --- all 22 chart types Excel offers,
  their axes, series parts and legends; chartsheets; and pictures placed on or
  in a sheet.

* **[Formulas, tables and the rest](e-formulas-and-more.html)** --- formulas and array
  formulas, hyperlinks, comments, rich strings, data validation, autofilters,
  worksheet tables and merged cells.

* **[Everything at once](f-everything-at-once.html)** --- one runnable script that
  exercises the whole package into a couple of workbooks, for reviewing what it
  can do without reading five vignettes first.

Two ideas run through all of them and are worth knowing early.

**A format is an object.** `xl_font()`, `xl_fill()` and the rest each return a
complete `xl_format`, they combine with `+`, and the same object styles a cell,
a column, a chart series or a conditional rule.

```{r}
money <- xl_num_format("$#,##0.00") + xl_font(bold = TRUE)
```

**A range is written one of two ways.** Either as A1 text, or against the data
frame by column name --- and the second is worth preferring, because it moves
with the column and accounts for the header row on its own:

```{r}
xl_sheet(iris, autofilter = "A1:E151")                   # A1 text
xl_sheet(iris, filter = xl_filter("Species", "==", "setosa"))  # by column
```

## Reading it back

writexl only writes. To read, use readxl:

```{r eval = requireNamespace("readxl", quietly = TRUE)}
readxl::read_xlsx(write_xlsx(head(iris, 3), tempfile(fileext = ".xlsx")))
```

---
title: "Formulas, tables and the rest"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Formulas, tables and the rest}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(writexl)
```

Everything a cell can hold besides a plain value, and the worksheet features
that act on ranges.

## What a cell can hold

`xl_cell_general()` is the general cell: any combination of a value, a formula,
a hyperlink, a format and a comment. The narrower constructors --- `xl_formula()`,
`xl_hyperlink()`, `xl_comment()` --- return the same kind of object.

### Formulas

```{r}
df <- data.frame(x = 1:3, y = c(10, 20, 30))
df$total <- xl_formula(sprintf("=A%d*B%d", 2:4, 2:4))
path <- write_xlsx(df, tempfile(fileext = ".xlsx"))
```

writexl writes the formula, not its result: Excel computes the value when the
file is opened. Give `result =` if a reader that does not calculate needs
something to show.

**Array formulas** cover a range and are entered once:

```{r}
xl_cell_general(formula = "=SUM(A2:A4*B2:B4)", array = TRUE)
xl_cell_general(formula = "=UNIQUE(A2:A10)", dynamic = TRUE)
```

`array = TRUE` writes the legacy CSE form and `dynamic = TRUE` the modern
spilling form. A multi-cell array range is declared with `array_range =` and
must start at the cell holding the formula, and must not overlap cells the
sheet writes itself.

### Hyperlinks

```{r}
xl_hyperlink("https://example.com", "Example")
```

The second argument is the text the cell shows; without it the URL is
shown. External URLs, `mailto:` addresses, other files and internal
`#Sheet1!A1` references all work. Styling comes from the workbook's
`hyperlink_format`, and `xl_properties(hyperlink_format = NULL)` writes them
unstyled.

### Comments

```{r}
xl_cell_general(value = 42, comment = xl_comment("Checked", author = "QA"))
```

A comment carries its own box: size, position, colour, and whether it is
visible when the file opens.

### Rich strings

One cell whose text is split into differently formatted runs:

```{r}
xl_rich_string("Plain ", xl_rich_run("bold", xl_font(bold = TRUE)),
               " and ", xl_rich_run("red", xl_font(color = "red")))
```

A run carries a font and nothing else, because that is all Excel renders on
one. A rich string cannot share a cell with a formula or a hyperlink, both of
which would discard the runs.

### Mixed columns

A column may hold different kinds of cell at different rows --- a number, a
formula, a blank, a hyperlink --- by giving `xl_cell_general()` lists:

```{r}
xl_cell_general(value = list(1, NA, "text"),
                format = list(NULL, xl_fill(background = "yellow"), NULL))
```

## Data validation

`xl_validation()` restricts what may be typed into a range, and carries the
messages Excel shows:

```{r}
xl_sheet(data.frame(size = c("S", "M", "L")),
         validation = xl_validation(list(cols = "size"), type = "list",
                                    list = c("S", "M", "L"),
                                    input_title = "Pick a size"))
```

Dropdown lists, numeric, date, time and text-length bounds, and custom formulas
are all available, with the criteria spelled the same way as elsewhere
(`"=="`, `"between"`, ...).

## Autofilters

```{r}
xl_sheet(data.frame(fruit = c("apple", "pear"), qty = c(5, 12)),
         filter = xl_filter("fruit", "==", "apple"))
```

Excel stores filter criteria and hidden rows independently and does **not**
apply a filter when a file is opened, so criteria alone would give a sheet that
looks filtered while showing every row. writexl therefore evaluates the filter
and hides the rows it excludes.

That makes it responsible for reproducing Excel's matching rules, which were
measured rather than inferred. An exact value or a `list` matches the text a
cell *displays*, so it matches the number `10` and the string `"10"` alike;
every other criteria compares by type. Text matching is case-insensitive, `*`
and `?` are wildcards, and blank covers an empty cell as well as an empty
string.

`xl_filter_keep()` exposes the rule on its own, without writing anything:

```{r}
df <- data.frame(fruit = c("apple", "pear", "plum"), qty = c(5, 12, 7))
xl_filter_keep(df, xl_filter("qty", ">", 6))
```

## Worksheet tables

A table is a named, styled range Excel treats as a unit: banded rows, a filter
dropdown, structured references, an optional total row.

```{r}
xl_sheet(data.frame(item = c("a", "b"), qty = c(3, 4)),
         table = xl_table(name = "Stock", total_row = TRUE,
                          columns = xl_table_column("qty", total = "sum")))
```

Column headers always come from the data frame's names. `worksheet_add_table()`
would otherwise write its own generic captions over them, and Excel treats a
mismatch between the table part and the header cells as a corrupt file.

Table names are made unique across the workbook. A table turns off the
row-streaming mode described in [Worksheets and workbooks](c-worksheets-workbooks.html), and
cannot sit on the same range as the sheet's own autofilter.

## Merged cells

```{r}
xl_sheet(data.frame(a = 1:2, b = 3:4),
         merge = xl_merge("A1:B1", "Heading",
                          format = xl_align(horizontal = "center")))
```

A merged range holds one value, so `xl_merge()` carries its own text. Merging
over cells the data frame filled keeps only the merged text, exactly as merging
does in Excel.

## Elsewhere

* Styling any of this: [Formatting cells](b-formatting.html)
* Sheet layout, printing, workbook properties: [Worksheets and
  workbooks](c-worksheets-workbooks.html)
* Charts, chartsheets, pictures: [Charts and images](d-charts-images.html)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_image.R
\name{xl_image}
\alias{xl_image}
\title{Insert an image into a worksheet}
\usage{
xl_image(
  image,
  at = "A1",
  scale = 1,
  offset = NULL,
  position = "move_and_size",
  description = NULL,
  decorative = FALSE,
  url = NULL,
  tip = NULL,
  embed = FALSE,
  format = NULL
)
}
\arguments{
\item{image}{The image, in any of four shapes: a path to a PNG, JPEG, GIF or
BMP; a raw vector holding one of those encoded; a `raster` (or anything
[grDevices::as.raster()] accepts, such as a colour matrix or an RGB/RGBA
array); or a `nativeRaster`.  The last two are what
[graphics::rasterImage()] draws, so anything you can plot can be written.}

\item{at}{The cell the image is anchored to, such as `"B2"`, or a
`list(rows = , cols = )` spec selecting a single cell.}

\item{scale}{Scale factor: one number for both axes, or `c(x, y)`.}

\item{offset}{Offset from the anchor cell's top-left corner in pixels, as
`c(x, y)`.}

\item{position}{How the image behaves when rows and columns change size:
`"move_and_size"` (the default), `"move_dont_size"`,
`"dont_move_dont_size"`, `"move_and_size_after"`, or `"default"` for
Excel's own default.  Ignored when `embed = TRUE`.}

\item{description}{Alt text, for screen readers.  Excel defaults it to the
file name; `""` writes none.}

\item{decorative}{Mark the image as decorative, so screen readers skip it.
Excel does not write a description for a decorative image.}

\item{url}{An optional hyperlink the image links to.}

\item{tip}{An optional mouseover tip for `url`.}

\item{embed}{Place the image inside the cell rather than floating above it.
Requires a version of Excel that supports images in cells.}

\item{format}{An [xl_format] for the cell holding an embedded image.  Only
meaningful with `embed = TRUE`.}
}
\value{
An `xl_image` object.
}
\description{
`xl_image()` places an image on a sheet, either floating over the cells and
anchored to one of them (the default) or, with `embed = TRUE`, inside a cell.
Pass one or a list of them as `xl_sheet(image = )`.

The image may be a file path or a raw vector, which is convenient when a plot
has just been written by a graphics device and never touched the disk.  PNG,
JPEG, GIF and BMP are supported --- the formats Excel reads --- and the
format is detected from the file's own bytes rather than its extension.
}
\section{Floating versus embedded}{

An inserted image floats above the grid: it has a position but occupies no
cell, and `position` decides whether it moves and resizes as rows and columns
change.  An embedded image (`embed = TRUE`) lives *in* a cell and sizes with
it, which is the Excel 365 "place in cell" behaviour.  Excel versions without
that feature show `#VALUE!` in place of an embedded image, so it is worth
choosing deliberately.
}

\section{Caveats inherited from Excel}{

An image's scaling can shift if it crosses a row whose height changed --- for
a taller font or wrapped text --- so set the height explicitly with
[xl_row_spec()] for rows an image spans.  BMP is supported only for backward
compatibility and must be 24-bit true colour; prefer PNG.  SVG is refused,
because Excel stores it converted to PNG anyway.
}

\examples{
logo <- system.file("help", "figures", "logo.png", package = "writexl")
if (nzchar(logo)) {
  xl_image(logo, at = "C2", scale = 0.5)
  xl_image(logo, at = "C2", url = "https://example.com", tip = "Home")
}
}
\seealso{
[xl_sheet]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_sheet.R
\name{xl_sheet}
\alias{xl_sheet}
\title{A worksheet with formatting and layout options}
\usage{
xl_sheet(
  data,
  cols = NULL,
  rows = NULL,
  freeze = NULL,
  gridlines = NA,
  tab_color = NA,
  zoom = NA,
  default_row_height = NA,
  auto_colwidth = FALSE,
  autofilter = FALSE,
  protect = FALSE,
  comment_author = NA,
  show_comments = FALSE,
  page = NULL,
  view = NULL,
  merge = NULL,
  validation = NULL,
  conditional = NULL,
  filter = NULL,
  outline = NULL,
  ignore_errors = NULL,
  table = NULL,
  image = NULL,
  background_image = NULL,
  chart = NULL
)
}
\arguments{
\item{data}{A data frame (the sheet contents).}

\item{cols}{An [xl_col_spec()], or a list of them.}

\item{rows}{An [xl_row_spec()], or a list of them.}

\item{freeze}{Frozen panes: an Excel cell reference such as `"A2"` (freeze
the rows above and columns left of that cell), or `list(row =, col =)`
giving the number of rows/columns to freeze.}

\item{gridlines}{Logical; show (`TRUE`) or hide (`FALSE`) screen gridlines.
`NA` leaves Excel's default.}

\item{tab_color}{Sheet tab color (see [xl_color]).}

\item{zoom}{Zoom level as a percentage (10--400).}

\item{default_row_height}{Default height (in points) for rows in the sheet.}

\item{auto_colwidth}{If `TRUE`, size each column to fit its contents (a
character-count heuristic, since the xlsx format has no true "AutoFit").
Columns given an explicit width via [xl_col_spec()] are left untouched.}

\item{autofilter}{Add an autofilter (filter dropdowns). `TRUE` covers the
whole used range (header plus data); an Excel range string such as
`"A1:D51"` restricts it; `FALSE` (default) adds none.}

\item{protect}{Protect the worksheet: `FALSE` (default) leaves it
unprotected, `TRUE` applies the standard protection, and a string sets a
password. A named list gives fine-grained control, e.g.
`list(password = "secret", format_cells = TRUE)`; an editing option set to
`TRUE` *allows* that action on the protected sheet. Available option names:
`format_cells`, `format_columns`, `format_rows`, `insert_columns`,
`insert_rows`, `insert_hyperlinks`, `delete_columns`, `delete_rows`,
`sort`, `autofilter`, `pivot_tables`, `scenarios`, `objects`,
`no_select_locked_cells`, `no_select_unlocked_cells`. Cell locking via
[xl_protection()] only has an effect on a protected sheet.}

\item{comment_author}{Default author for this sheet's cell comments (a
per-comment `author` overrides it).}

\item{show_comments}{If `TRUE`, all comments on the sheet are initially
shown (individual comments can still be forced via `xl_comment(visible=)`).}

\item{page}{An [xl_page_setup()] describing how the sheet prints
(orientation, paper size, margins, scaling, header and footer). Affects
printing only, never the cell data.}

\item{view}{An [xl_sheet_view()] describing the sheet's tab state and
opening view (active/selected/hidden tab, selection, scroll position,
zero display, direction, split panes).}

\item{merge}{One [xl_merge()], or a list of them, merging rectangles of cells
into single cells.  Merges are applied after the sheet's rows are written,
so a merge over cells the data frame filled keeps only the merged text ---
as merging in Excel does.  Any merge turns off the memory-efficient
row-streaming mode, since it writes back over rows already emitted.}

\item{validation}{One [xl_validation()], or a list of them, restricting what
may be typed into a range --- a dropdown, a numeric or date bound, a text
length limit or a custom formula.}

\item{conditional}{One conditional format ([xl_cond_cell()],
[xl_cond_scale()], [xl_cond_bar()], [xl_cond_icons()]), or a list of them,
formatting cells according to their contents.}

\item{filter}{One [xl_filter()], or a list of them, setting autofilter
criteria.  Each also hides the rows it excludes, because Excel does not
apply a filter when a file is opened --- criteria on their own produce a
sheet that looks filtered but shows every row.  Implies `autofilter = TRUE`
over the used range when `autofilter` is not set separately.}

\item{outline}{An [xl_outline()] controlling how the grouping symbols
created by `level` in [xl_col_spec()] / [xl_row_spec()] are drawn.  It
changes their display only, never which rows are grouped.}

\item{ignore_errors}{A named list turning off the green error triangle Excel
shows in cells it believes are wrong.  Each name is an error type and each
value a range, e.g. `list(number_stored_as_text = "A2:A99")`.  Types:
`number_stored_as_text`, `eval_error`, `formula_differs`, `formula_range`,
`formula_unlocked`, `empty_cell_reference`, `list_data_validation`,
`calculated_column`, `two_digit_text_year`.}

\item{table}{One [xl_table()], or a list of them, turning a range into an
Excel table --- a named, styled block with banded rows, a filter dropdown
and an optional total row.  Any table turns off the memory-efficient
row-streaming mode, which libxlsxwriter refuses to combine with tables.}

\item{image}{One [xl_image()], or a list of them, placing images on the
sheet --- floating over the cells, or inside a cell with `embed = TRUE`.}

\item{background_image}{An image tiled behind the sheet's cells, in any shape
[xl_image()] accepts.  It is a screen backdrop only --- Excel never prints
it.}

\item{chart}{One [xl_chart()], or a list of them, placed on the sheet and
anchored to a cell.  A chart's series may plot data from any sheet in the
workbook, not only this one.}
}
\value{
An `xl_sheet` object.
}
\description{
`xl_sheet()` wraps a data frame together with worksheet-level options
(column/row formatting and geometry, frozen panes, gridlines, tab color,
zoom).  Pass it anywhere [write_xlsx()] accepts a data frame; a plain data
frame continues to behave exactly as before.
}
\examples{
df <- data.frame(name = c("a", "b"), revenue = c(1000.5, 2000.25))
sheet <- xl_sheet(
  df,
  cols   = xl_col_spec("revenue", width = 14, format = xl_num_format("#,##0.00")),
  freeze = "A2",
  tab_color = "steelblue"
)
tmp <- write_xlsx(list(Data = sheet))
}
\seealso{
[xl_col_spec], [xl_row_spec], [write_xlsx]

Other worksheet layout:
\code{\link{xl_colrow_spec}},
\code{\link[=xl_outline]{xl_outline()}},
\code{\link[=xl_page_setup]{xl_page_setup()}},
\code{\link[=xl_sheet_view]{xl_sheet_view()}}
}
\concept{worksheet layout}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_sheet.R
\name{xl_colrow_spec}
\alias{xl_colrow_spec}
\alias{xl_col_spec}
\alias{xl_row_spec}
\title{Column and row specifications for a worksheet}
\usage{
xl_col_spec(
  cols,
  width = NA,
  hidden = NA,
  level = NA,
  format = NULL,
  width_pixels = NA,
  collapsed = NA,
  na = NA
)

xl_row_spec(
  rows,
  height = NA,
  hidden = NA,
  level = NA,
  format = NULL,
  height_pixels = NA,
  collapsed = NA
)
}
\arguments{
\item{cols}{Columns to target: a character vector of column names or a
numeric vector of 1-based positions.}

\item{width}{Column width (in Excel character units).}

\item{hidden}{Logical; hide the column/row.}

\item{level}{Integer outline (grouping) level, 0--7.}

\item{format}{An optional [xl_format] applied to the column/row as its
default cell format.  Combine groups with `+` (e.g.
`xl_font(bold = TRUE) + xl_fill(background = "yellow")`).}

\item{width_pixels, height_pixels}{The same geometry given in pixels instead.
Give one or the other, not both.  Excel stores character units and points,
so the pixel value is converted on the way in --- read back, a width set
as 100 pixels is 13.57 character units.}

\item{collapsed}{Logical; draw this column/row as the collapsed summary of
the group beside it.  Excel does not derive this --- the rows or columns
of the group itself need `hidden = TRUE` as well, exactly as clicking the
grouping symbol would leave them.}

\item{na}{What to write in this column where a value has none, overriding
the workbook's [xl_properties()]`(na = )`. `NA` (the default) inherits it.
Columns are where this usually belongs: a substitute that suits a numeric
column rarely suits a date one.}

\item{rows}{Rows to target: a numeric vector of 1-based data-row indices
(row 1 is the first data row, ignoring the header).}

\item{height}{Row height (in points).}
}
\value{
An `xl_col_spec` / `xl_row_spec` object (also an [xl_format]).
}
\description{
`xl_col_spec()` and `xl_row_spec()` describe formatting and geometry for a
set of columns or rows within a sheet built by [xl_sheet()].  They are
*subclasses* of [xl_format]: they carry the usual formatting groups (so they
combine with `+` and the group constructors) plus a target (which
columns/rows) and geometry (width/height, hidden, outline level).
}
\examples{
xl_col_spec("revenue", width = 14, format = xl_num_format("#,##0.00"))
xl_col_spec(c(1, 2), width = 10) + xl_font(bold = TRUE)
xl_col_spec("logo", width_pixels = 100)
xl_row_spec(1, height = 24, format = xl_font(bold = TRUE))
xl_row_spec(1, height_pixels = 40)
}
\seealso{
[xl_sheet], [xl_format]

Other worksheet layout:
\code{\link[=xl_outline]{xl_outline()}},
\code{\link[=xl_page_setup]{xl_page_setup()}},
\code{\link[=xl_sheet]{xl_sheet()}},
\code{\link[=xl_sheet_view]{xl_sheet_view()}}
}
\concept{worksheet layout}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_xlsx.R
\name{write_xlsx}
\alias{write_xlsx}
\alias{writexl}
\title{Export to xlsx}
\usage{
write_xlsx(
  x,
  path = tempfile(fileext = ".xlsx"),
  col_names = TRUE,
  format_headers = TRUE,
  na = NA,
  use_zip64 = FALSE,
  constant_memory = NA,
  constant_memory_threshold = 128 * 1024^2
)
}
\arguments{
\item{x}{a data frame, an [xl_sheet], an [xl_workbook], or a (named) list of
data frames / `xl_sheet`s that become the sheets in the xlsx}

\item{path}{a file name to write to}

\item{col_names}{write column names as the header row at the top of the sheet?}

\item{format_headers}{apply the workbook's header format to that header row?
The default header format is bold and centered; change it with
\code{\link{xl_properties}(header_format = )}.}

\item{na}{what to write where a value has none. `NA` (the default) leaves
the cell blank, as writexl has always done; anything else is written in
its place, keeping its own type. Shorthand for
\code{\link{xl_properties}(na = )}, so give it there instead when `x` is
already an [xl_workbook]. A column or a single cell can override it: see
[xl_col_spec()] and [xl_cell_general()].}

\item{use_zip64}{use \href{https://en.wikipedia.org/wiki/Zip_(file_format)#ZIP64}{zip64}
to enable support for 4GB+ xlsx files. Not all platforms can read this.}

\item{constant_memory}{stream rows to disk instead of building the whole
workbook in memory. `NA` (the default) decides per workbook: on for large
data, off for small, and always off when a feature needs it off. `TRUE`
forces it on for a workbook that would otherwise be judged too small;
`FALSE` forces it off. Features that cannot be written while streaming ---
merged ranges, tables, embedded images and multi-cell array formulas ---
turn it off regardless, with a warning if `TRUE` was asked for, because
the alternative is a file that opens cleanly and is missing cells.}

\item{constant_memory_threshold}{how much extra memory not streaming would
have to cost, in bytes, before streaming is worth it. Default 128 MiB.
The cost is *estimated* from the number of cells in the workbook, using a
fixed per-cell figure calibrated against a range of data; the true cost
varies with the data, and is lowest for text that repeats. Streaming saves
memory but produces slightly larger files, so it is not used for workbooks
small enough that the saving would not be noticed.}
}
\description{
Writes a data frame to an xlsx file. To create an xlsx with (multiple) named
sheets, simply set \code{x} to a named list of data frames.
}
\details{
Supports strings, numbers, booleans and dates automatically. For cell
formatting (fonts, fills, borders, number formats, ...), worksheet layout
(column widths, frozen panes, ...), and workbook metadata, wrap columns with
\code{\link{xl_cell_general}}, sheets with \code{\link{xl_sheet}}, and the
whole workbook with \code{\link{xl_workbook}}. See the "Formatting and
workbook properties" vignette and \code{\link{xl_format}}.

\if{html}{
\out{
<link rel="stylesheet" type="text/css" href="https://jeroen.github.io/clippy/clippy.min.css" media="all">
<script src="https://jeroen.github.io/clippy/bundle.js"></script>
}}
}
\examples{
# Roundtrip example with single excel sheet named 'mysheet'
tmp <- write_xlsx(list(mysheet = iris))
readxl::read_xlsx(tmp)
}
\seealso{
Other workbook settings:
\code{\link[=xl_properties]{xl_properties()}},
\code{\link[=xl_workbook]{xl_workbook()}}
}
\concept{workbook settings}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_page_setup.R
\name{xl_page_setup}
\alias{xl_page_setup}
\title{How a worksheet prints}
\usage{
xl_page_setup(
  orientation = NA,
  paper = NA,
  margins = NULL,
  scale = NA,
  fit_to = NULL,
  center_horizontally = NA,
  center_vertically = NA,
  header = NA,
  footer = NA,
  header_margin = NA,
  footer_margin = NA,
  header_image = NULL,
  footer_image = NULL,
  page_view = NA,
  first_page = NA,
  across = NA,
  black_and_white = NA,
  row_col_headers = NA,
  print_area = NULL,
  repeat_rows = NA,
  repeat_cols = NA,
  h_breaks = NULL,
  v_breaks = NULL
)
}
\arguments{
\item{orientation}{`"portrait"` or `"landscape"`.}

\item{paper}{Paper size: a name (`"A4"`, `"letter"`, `"legal"`, `"tabloid"`,
`"ledger"`, `"statement"`, `"executive"`, `"A3"`, `"A5"`, `"B4"`, `"B5"`,
`"folio"`, `"quarto"`, `"default"`), or an Excel paper-type integer for the
envelope and specialist sizes that have no name here.}

\item{margins}{Page margins in inches: a single number for all four sides,
four numbers in the order left, right, top, bottom, or a named vector using
any of `left`, `right`, `top`, `bottom`.}

\item{scale}{Print scaling as a percentage (10--400).  Ignored by Excel when
`fit_to` is set.}

\item{fit_to}{Fit the printout to `c(width, height)` pages.  A `0` means
"as many pages as needed in that direction", so `c(width = 1, height = 0)`
is Excel's "fit all columns on one page".}

\item{center_horizontally, center_vertically}{Logical; centre the printed
output on the page.}

\item{header, footer}{Header and footer text, at most 255 characters, using
Excel's own codes: `&L`, `&C`, `&R` start the left, centre and right
sections, `&P` is the page number, `&N` the page count, `&D` the date, `&A`
the sheet name, and `&&` a literal ampersand.  For example
`"&LQ1 report&RPage &P of &N"`.  Image placeholders (`&G` /
`&[Picture]`) are not supported yet and are rejected.}

\item{header_margin, footer_margin}{Header/footer margin in inches (Excel's
default is 0.3).  Must be greater than 0.}

\item{header_image, footer_image}{Images to place in the header or footer,
named by position: `list(left = , center = , right = )`.  Each may be a
file path, a raw vector, or an in-memory image, exactly as [xl_image()]
accepts.  Each image needs a matching `&G` placeholder in the
corresponding section of `header`/`footer` --- `"&L&G"` puts one on the
left --- and the counts must agree.}

\item{page_view}{Logical; open the sheet in Excel's page-layout view rather
than normal view.}

\item{first_page}{The page number to start numbering from.}

\item{across}{Logical; print pages left-to-right before top-to-bottom
(Excel's "over, then down").}

\item{black_and_white}{Logical; print without colour.}

\item{row_col_headers}{Logical; print the row numbers and column letters.}

\item{print_area}{The range to print: an Excel range string such as
`"A1:F50"`, or a `list(rows = , cols = )` spec naming data rows and
columns, as elsewhere in writexl.}

\item{repeat_rows, repeat_cols}{Rows/columns to repeat at the top or left of
  every printed page.  Either a count (`repeat_rows = 1` repeats the first
  sheet row, which is the header when `col_names = TRUE`) or a range string
  (`"1:2"`, `"A:B"`).

  Note these count *sheet* rows from 1 with the header included, unlike
  [xl_row_spec()], which indexes data rows --- repeating the header row is
  the usual reason to use this, and data-row numbering could not name it.}

\item{h_breaks, v_breaks}{Manual page breaks: 1-based sheet positions at which
a new page starts, so `h_breaks = 21` breaks between rows 20 and 21.
`v_breaks` also accepts column letters.  Positions must be 2 or greater,
and at most 1023 breaks are allowed.  Excel ignores manual breaks when
`fit_to` is set, which warns.}
}
\value{
An `xl_page_setup` object.
}
\description{
`xl_page_setup()` collects Excel's page-layout settings for one worksheet:
orientation, paper size, margins, scaling, centring, the print options, and
the header and footer.  Pass it as `xl_sheet(page = )`.

None of these affect the cell data --- they change only how the sheet prints
and how it looks in Excel's page-break preview.
}
\examples{
xl_page_setup(orientation = "landscape", paper = "A4",
              fit_to = c(width = 1, height = 0))

# margins in inches, and a header with page numbers
xl_page_setup(margins = c(left = 1, right = 1),
              header = "&LQuarterly report&RPage &P of &N")

df <- data.frame(x = 1:3)
tmp <- write_xlsx(list(Data = xl_sheet(df, page = xl_page_setup(
  orientation = "landscape", header = "&CDraft"
))))
}
\seealso{
[xl_sheet], [write_xlsx]

Other worksheet layout:
\code{\link{xl_colrow_spec}},
\code{\link[=xl_outline]{xl_outline()}},
\code{\link[=xl_sheet]{xl_sheet()}},
\code{\link[=xl_sheet_view]{xl_sheet_view()}}
}
\concept{worksheet layout}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_comment.R
\name{is_xl_comment}
\alias{is_xl_comment}
\title{Test whether an object is an `xl_comment`}
\usage{
is_xl_comment(x)
}
\arguments{
\item{x}{An object.}
}
\value{
`TRUE` if `x` inherits from `"xl_comment"`.
}
\description{
Test whether an object is an `xl_comment`
}
\seealso{
Other cell content:
\code{\link[=xl_cell_general]{xl_cell_general()}},
\code{\link[=xl_comment]{xl_comment()}},
\code{\link[=xl_formula]{xl_formula()}},
\code{\link[=xl_rich_run]{xl_rich_run()}},
\code{\link[=xl_rich_string]{xl_rich_string()}}
}
\concept{cell content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_merge.R
\name{xl_merge}
\alias{xl_merge}
\title{Merge a range of cells}
\usage{
xl_merge(range, value = NULL, format = NULL)
}
\arguments{
\item{range}{The cells to merge: an Excel range string such as `"A1:C1"`, or
a `list(rows = , cols = )` spec.  It must cover more than one cell ---
Excel has no single-cell merge.}

\item{value}{The value shown in the merged cell: a string, or anything with
an [as.character()] method such as an [xl_rich_string()].  `NULL` leaves
the cell empty.}

\item{format}{An optional [xl_format] applied to the whole merged range.
Merged cells usually want `xl_align(horizontal = "center")`.}
}
\value{
An `xl_merge` object.
}
\description{
`xl_merge()` merges a rectangle of cells into one, as Excel's "Merge and
Centre" does.  Pass one or a list of them as `xl_sheet(merge = )`.

A merged range holds a single value, so `xl_merge()` carries its own `value`
rather than taking it from the data frame.  Merging over cells the data frame
filled keeps only the merged value, exactly as merging in Excel discards
everything but the top-left value.
}
\examples{
xl_merge("A1:C1", "Quarterly results",
         format = xl_align(horizontal = "center") + xl_font(bold = TRUE))

df <- data.frame(a = 1:3, b = 4:6)
sheet <- xl_sheet(df, merge = xl_merge("A5:B5", "Total",
                                       format = xl_font(bold = TRUE)))
tmp <- write_xlsx(list(Data = sheet))
}
\seealso{
[xl_sheet], [xl_format]

Other worksheet features:
\code{\link{xl_conditional}},
\code{\link[=xl_filter]{xl_filter()}},
\code{\link[=xl_filter_keep]{xl_filter_keep()}},
\code{\link[=xl_table]{xl_table()}},
\code{\link[=xl_table_column]{xl_table_column()}},
\code{\link[=xl_validation]{xl_validation()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_validation.R
\name{xl_validation}
\alias{xl_validation}
\title{Restrict what can be typed into a range}
\usage{
xl_validation(
  range,
  type = "any",
  criteria = NA,
  value = NULL,
  min = NULL,
  max = NULL,
  list = NULL,
  input_title = NA,
  input_message = NA,
  error_title = NA,
  error_message = NA,
  error_type = NA,
  ignore_blank = NA,
  show_input = NA,
  show_error = NA,
  dropdown = NA
)
}
\arguments{
\item{range}{The cells to validate: an Excel range string such as
`"B2:B100"`, a single cell, or a `list(rows = , cols = )` spec.}

\item{type}{The kind of value allowed: `"integer"`, `"decimal"`, `"date"`,
`"time"`, `"length"` (of the text entered), `"custom"` (any formula that
must evaluate `TRUE`), or `"any"` (no restriction, useful when you only
want the input message).  Ignored when `list` is given.}

\item{criteria}{How `value` (or `min`/`max`) limits the entry: `"between"`,
`"not between"`, `"=="`, `"!="`, `">"`, `"<"`, `">="` or `"<="`.  Required
for the numeric, date, time and length kinds; must not be given for
`list`, `custom` or `any`, which carry their own meaning.  Supplying `min`
and `max` implies `"between"`.}

\item{value}{The single limit the criteria applies to.  A number, a `Date` or
`POSIXct`, or a `"=..."` formula.}

\item{min, max}{The two limits for `"between"` / `"not between"`.  Supplying
both and omitting `criteria` implies `"between"`.}

\item{list}{A dropdown of allowed values: a character vector of choices, or a
single `"=..."` formula naming a range that holds them.  The choices are
stored joined by commas, and Excel limits that joined string to 255
characters.}

\item{input_title, input_message}{Text shown in a tooltip when the cell is
selected.  Titles are limited to 32 characters and messages to 255.}

\item{error_title, error_message}{Text shown when an invalid entry is made.
Same limits.}

\item{error_type}{What Excel does on an invalid entry: `"stop"` (refuse it),
`"warning"` or `"information"` (both allow it through).}

\item{ignore_blank}{Logical; allow an empty cell.  `TRUE` by default, as in
Excel.}

\item{show_input, show_error}{Logical; whether the input tooltip and the error
alert are shown at all.  Both `TRUE` by default.}

\item{dropdown}{Logical; show the in-cell dropdown arrow for a `list`
validation.  `TRUE` by default.}
}
\value{
An `xl_validation` object.
}
\description{
`xl_validation()` adds Excel data validation to a range: a dropdown list, a
numeric or date bound, a text-length limit, or a custom formula.  Pass one or
a list of them as `xl_sheet(validation = )`.

Excel's 17 internal validation types are collapsed into the five `type`
kinds below.  Whether a limit is a literal, a cell formula or a date is
inferred from what you pass: a string starting with `"="` is a formula, and a
`Date` or `POSIXct` is a date/time bound.
}
\examples{
# a dropdown
xl_validation("C2:C100", list = c("open", "high", "close"))

# a numeric bound, with the message Excel shows on a bad entry
xl_validation("B2:B100", type = "integer", min = 1, max = 10,
              error_message = "Enter a whole number from 1 to 10")

# a date bound
xl_validation("D2:D100", type = "date", criteria = ">=",
              value = as.Date("2024-01-01"))

df <- data.frame(qty = 1:3)
tmp <- write_xlsx(list(Data = xl_sheet(df,
  validation = xl_validation("A2:A4", type = "integer", min = 0, max = 99))))
}
\seealso{
[xl_sheet]

Other worksheet features:
\code{\link{xl_conditional}},
\code{\link[=xl_filter]{xl_filter()}},
\code{\link[=xl_filter_keep]{xl_filter_keep()}},
\code{\link[=xl_merge]{xl_merge()}},
\code{\link[=xl_table]{xl_table()}},
\code{\link[=xl_table_column]{xl_table_column()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel_types.R
\name{xl_formula}
\alias{xl_formula}
\alias{xl_hyperlink}
\alias{xl_hyperlink_cell}
\title{Excel Types}
\usage{
xl_formula(x, format = NULL)

xl_hyperlink(url, value = NULL, format = NULL, name = NULL)

xl_hyperlink_cell(url, value = NULL, format = NULL)
}
\arguments{
\item{x}{character vector to be interpreted as formula}

\item{format}{An optional [xl_format] (or list of `xl_format`, one per
element) applied to the cells.  See [xl_format].}

\item{url}{character vector of URLs.  Use `NA` to produce a blank cell.}

\item{value}{character vector (or `NULL`) of display text shown in the cell
instead of the URL.  When `NULL` the URL itself is shown.  Recycled to the
length of `url`, and automatically `NA` for cells whose URL is `NA`.  The
same argument name is used by [xl_hyperlink_cell()] and
[xl_cell_general()].}

\item{name}{**Deprecated.** The former spelling of `value`, kept for
backward compatibility.  Supplying it warns and points at `value`;
supplying both is an error, since they mean the same thing.  `value` has
taken the argument position `name` used to occupy, so code that passed the
display text positionally keeps working unchanged.}
}
\description{
* `xl_formula(x)` — wraps a character vector of Excel formulas (each must
  start with `"="`). The formulas are written to the xlsx file as-is and
  are recalculated by Excel on open.

* `xl_hyperlink(url, name)` — convenience wrapper that builds an Excel
  `=HYPERLINK(url, name)` **formula** for each element.  Because the
  hyperlink is stored as a formula, it is readable by
  [readxl::read_xlsx()], which returns the formula text.  Display text is
  controlled by the `name` argument.

* `xl_hyperlink_cell(url, value)` — creates a **native cell-level
  hyperlink** using `worksheet_write_url_opt()` from libxlsxwriter.  The
  URL is stored as metadata attached to the cell, not in the formula bar.
  An optional `value` argument provides the display text shown in the cell.
  A tooltip and further options can be set by passing a named list to
  `xl_cell_general()` directly.
  **Note:** [readxl::read_xlsx()] cannot read cell-level hyperlinks and
  returns `NA` for those cells.  Use `xl_hyperlink()` instead when
  round-tripping through readxl is required.
}
\details{
Create special column types to write to a spreadsheet.
}
\examples{
df <- data.frame(
  name = c("UCLA", "Berkeley", "Jeroen"),
  founded = c(1919, 1868, 2030),
  website = xl_hyperlink(c("http://www.ucla.edu", "http://www.berkeley.edu", NA), "homepage")
)
df$age <- xl_formula('=(YEAR(TODAY()) - INDIRECT("B" & ROW()))')
write_xlsx(df, 'universities.xlsx')

# xl_hyperlink_cell() stores the URL as native cell metadata.
# readxl cannot read these cells, but they display cleanly in Excel.
df2 <- data.frame(
  name = c("UCLA", "Berkeley"),
  website = xl_hyperlink_cell(c("http://www.ucla.edu", "http://www.berkeley.edu"),
                               value = "homepage")
)
write_xlsx(df2, 'universities2.xlsx')

# cleanup
unlink(c('universities.xlsx', 'universities2.xlsx'))
}
\seealso{
Other cell content:
\code{\link[=is_xl_comment]{is_xl_comment()}},
\code{\link[=xl_cell_general]{xl_cell_general()}},
\code{\link[=xl_comment]{xl_comment()}},
\code{\link[=xl_rich_run]{xl_rich_run()}},
\code{\link[=xl_rich_string]{xl_rich_string()}}
}
\concept{cell content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_format.R
\name{xl_format_groups}
\alias{xl_format_groups}
\alias{xl_font}
\alias{xl_fill}
\alias{xl_border}
\alias{xl_align}
\alias{xl_num_format}
\alias{xl_protection}
\title{Cell formatting groups}
\usage{
xl_font(
  bold = NA,
  italic = NA,
  color = NA,
  size = NA,
  name = NA,
  underline = NA,
  strikeout = NA,
  script = NA,
  family = NA,
  charset = NA,
  outline = NA,
  shadow = NA,
  condense = NA,
  extend = NA,
  scheme = NA,
  theme = NA,
  color_indexed = NA,
  font_only = NA
)

xl_fill(background = NA, foreground = NA, pattern = NA, transparency = NA)

xl_border(
  all = NA,
  left = NA,
  right = NA,
  top = NA,
  bottom = NA,
  color = NA,
  left_color = NA,
  right_color = NA,
  top_color = NA,
  bottom_color = NA,
  diagonal = NA,
  diagonal_style = NA,
  diagonal_color = NA,
  transparency = NA
)

xl_align(
  horizontal = NA,
  vertical = NA,
  wrap = NA,
  rotation = NA,
  indent = NA,
  shrink = NA,
  reading_order = NA
)

xl_num_format(format = NA, index = NA)

xl_protection(locked = NA, hidden = NA)
}
\arguments{
\item{bold, italic, strikeout, outline, shadow, condense, extend, font_only}{Logical
font flags.}

\item{color, background, foreground, left_color, right_color, top_color, bottom_color, diagonal_color}{A color: an R color name, a hex string, or an integer (see [xl_color]).}

\item{size}{Font size in points (1--409).}

\item{name}{Font name, e.g. `"Calibri"`.}

\item{underline}{One of `"none"`, `"single"`, `"double"`,
`"single-accounting"`, `"double-accounting"`.}

\item{script}{One of `"super"`, `"sub"`.}

\item{family, charset, theme, color_indexed}{Advanced integer font properties
(rarely needed).}

\item{scheme}{Font scheme string (advanced).}

\item{pattern}{Fill pattern, e.g. `"solid"`, `"light-gray"` (18 choices).
When `background` is supplied and `pattern` is unset, a `"solid"` pattern
is assumed.}

\item{transparency}{Percentage transparency, 0--100.  **Charts only**: Excel
has no transparency for a cell's fill or border, so this is ignored
everywhere except a chart's line and fill (see [xl_chart()]).}

\item{all}{Border style applied to all four sides at once (one of `"none"`,
`"thin"`, `"medium"`, `"dashed"`, `"dotted"`, `"thick"`, `"double"`,
`"hair"`, `"medium-dashed"`, `"dash-dot"`, `"medium-dash-dot"`,
`"dash-dot-dot"`, `"medium-dash-dot-dot"`, `"slant-dash-dot"`).}

\item{left, right, top, bottom}{Per-side border styles (override `all`).}

\item{diagonal}{Diagonal border direction: `"up"`, `"down"`, `"up-down"`.}

\item{diagonal_style}{Diagonal border style (same choices as `all`).}

\item{horizontal}{Horizontal alignment: `"left"`, `"center"`, `"right"`,
`"fill"`, `"justify"`, `"center-across"`, `"distributed"`.}

\item{vertical}{Vertical alignment: `"top"`, `"bottom"`, `"center"`,
`"justify"`, `"distributed"`.}

\item{wrap}{Logical; wrap text in the cell.}

\item{rotation}{Text rotation in degrees (-90..90, or 270).}

\item{indent}{Integer indentation level.}

\item{shrink}{Logical; shrink text to fit.}

\item{reading_order}{One of `"default"`, `"ltr"`, `"rtl"`.}

\item{format}{A number-format string, e.g. `"#,##0.00"`, `"0%"`,
`"yyyy-mm-dd"`.}

\item{index}{An Excel built-in number-format index (alternative to
`format`).}

\item{locked}{Logical; whether the cell is locked (Excel's default is
`TRUE`).  Only takes effect when the worksheet is protected.  `FALSE`
unlocks the cell.}

\item{hidden}{Logical; hide the cell's formula.}
}
\value{
An [xl_format] object.
}
\description{
These constructors build [xl_format] objects, each populating one group of
Excel cell-formatting properties.  Every property defaults to `NA`, meaning
"leave unset"; unset properties are simply not written.  Enum-like arguments
take lowercase strings and are validated against a fixed set of choices.

Each constructor returns a full `xl_format`, so a single group can be used on
its own (`xl_cell_general(1, format = xl_font(bold = TRUE))`), and groups can
be combined with `+` (see [xl_format]).
}
\examples{
xl_font(bold = TRUE, color = "navy", size = 12)
xl_fill(background = "#FFF2CC")
xl_border(all = "thin", color = "gray")
xl_align(horizontal = "center", vertical = "top", wrap = TRUE)
xl_num_format("#,##0.00")
xl_protection(locked = FALSE)
}
\seealso{
[xl_format], [xl_color]

Other cell formatting:
\code{\link[=is_xl_format]{is_xl_format()}},
\code{\link[=xl_color]{xl_color()}},
\code{\link[=xl_format]{xl_format()}}
}
\concept{cell formatting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_cell_general.R
\name{xl_cell_general}
\alias{xl_cell_general}
\alias{as.character.xl_cell_general}
\title{General cell objects for Excel writing}
\usage{
xl_cell_general(
  value = NULL,
  formula = NULL,
  hyperlink = NULL,
  format = NULL,
  comment = NULL,
  array = FALSE,
  dynamic = FALSE,
  array_range = NULL,
  na = NA
)

\method{as.character}{xl_cell_general}(x, ...)
}
\arguments{
\item{value}{An atomic vector or a list of scalars, one per cell. Use `NA`
for a cell that is empty unless `na` says otherwise (see `na` below, and
note that a workbook-wide `na` reaches these values too). A list enables
mixed types across cells
in the same column (e.g., `list(1.5, "text", TRUE)`). Date and POSIXct
scalars are supported and formatted as in [write_xlsx()].  When
`hyperlink` is also set for the same cell, a **character** `value` is
used as the display text shown in the cell instead of the raw URL; all
other types are ignored for hyperlink cells.}

\item{formula}{A character vector of Excel formulas (each must start with
`"="`), or `NA` for cells with no formula.  When both `value` and
`formula` are supplied for the same cell, `value` is used as a
pre-calculated result stored alongside the formula via
`worksheet_write_formula_num()` (numeric value) or
`worksheet_write_formula_str()` (character value).  This allows static
xlsx exports that display formula text in the formula bar but do not
require Excel to recalculate on open.}

\item{hyperlink}{A character vector of URLs, or a list where each element
is `NA`, a single character URL, or a named list with elements:
\describe{
  \item{`url`}{(required) The target URL.}
  \item{`tooltip`}{(optional) Tooltip text shown on hover.}
}
Supply a character `value` alongside `hyperlink` to show custom display
text in the cell instead of the raw URL.  The hyperlink is written via
`worksheet_write_url_opt()`.}

\item{format}{An [xl_format] object (applied to every cell), or a list of
`xl_format` objects (one per cell, recycled), or `NULL` for no formatting.
Build formats with [xl_font()], [xl_fill()], [xl_border()], [xl_align()],
[xl_num_format()] and [xl_protection()], combined with [xl_format()] or
`+`.  When a cell's value is a date/time and its format sets no number
format, the default date/time number format is applied automatically.}

\item{comment}{Cell comments (notes): a character vector of comment text
(one per cell, recycled; `NA` for no comment), a single [xl_comment()]
(recycled to every cell), or a list mixing strings / `xl_comment` / `NA`
per cell.  `NULL` for no comments.}

\item{array, dynamic}{Logical (one per cell, recycled): how the cell's
`formula` is stored.  `array = TRUE` writes a legacy *array* (Ctrl-Shift-
Enter) formula; `dynamic = TRUE` writes a modern *dynamic array* formula,
which Excel spills over as many cells as the result needs.  Neither can
carry a character `value`, because Excel stores no cached string result for
an array formula.  On a cell that has no `formula` the flags are inert, so a
single `array = TRUE` can be recycled across a column that mixes formula
and value cells.}

\item{array_range}{The range a legacy array formula covers, for the rare case
  where it must be declared: an Excel range string (`"C2:C11"`) or a
  `list(rows = , cols = )` spec, one per cell (`NA` for none). It must start
  at the cell holding the formula, and must extend into cells the sheet does
  not otherwise write --- the range is padded on write, so an overlap would
  have the padding and the sheet's own values overwrite each other.

  Leave it unset for almost everything: a single-cell `dynamic` formula
  spills automatically in Excel 365 / 2021, and a single-cell `array` formula
  is the right spelling for a `SUMPRODUCT`-style aggregate. Supplying it
  forces the workbook out of the memory-efficient row-streaming mode, since
  libxlsxwriter cannot pad an array range while streaming.}

\item{na}{What to write in a cell that has no value, one per cell and
recycled.  `NA` (the default) inherits the column's
[xl_col_spec()]`(na = )`, and failing that the workbook's
[xl_properties()]`(na = )`; a cell that sets its own overrides both.  This
is also how a cell asks for a blank where the column or workbook would
otherwise substitute something --- give it the value you want, `""` for an
empty string.}

\item{x}{An `xl_cell_general`.}

\item{...}{Ignored.}
}
\value{
An object of class `c("xl_cell_general", "xl_cell")`, which is a
  list of length `n` where each element is a named list with fields
  `value`, `formula`, `hyperlink`, `format`, `comment`, `array`, `dynamic`
  and `array_range`.
}
\description{
`xl_cell_general` creates a vector of cell objects, each optionally
containing a **value**, a **formula**, and/or a **hyperlink**. It is the
fundamental building block used internally by [xl_formula()] and
[xl_hyperlink()], and can be used directly for mixed-type columns or cells
that combine multiple features (e.g., a formula with a pre-calculated
result, or a URL with separate display text and tooltip).

An `xl_cell_general` behaves like a vector: it has a `length()`, supports
`[`, `[[`, `c()` and `rep()`, and recycles automatically when assigned to a
data frame column of a different length (just like [xl_formula()]).  It is
deliberately not a list, so that `df[, j] <- cells` assigns one column
rather than being read as a list of them.

`df[i, j] <- x` sets that cell's `value`, whatever `x`'s type; a formula
needs [xl_formula()], since a cell column carries no column-wide notion of
"these are all formulas".

`as.character()` returns what each cell displays, so a cell built for a sheet
can be reused wherever a plain string is wanted --- [xl_merge()]'s or
[xl_comment()]'s `value`, for instance.  A cell that carries only a formula
shows `NA`: its displayed value comes from Excel, not from writexl.
}
\examples{
# Value-only cell
xl_cell_general(value = 42)

# Formula with a pre-calculated numeric result (static export)
xl_cell_general(value = 42.0, formula = "=SUM(A1:A10)")

# Hyperlink with display text (value) and tooltip
xl_cell_general(
  value    = "Visit",
  hyperlink = list(url = "https://example.com", tooltip = "Go to example.com")
)

# Vector of cells: value and formula cells in one column
cells <- c(
  xl_cell_general(value = 1.5),
  xl_cell_general(value = "note"),
  xl_cell_general(formula = "=A1+A2")
)

# Used in a data frame (length-1 recycles to fill all rows, as with
# xl_formula())
df <- data.frame(x = 1:3)
df$formula_col <- xl_formula("=A1*2")   # backward-compatible shorthand
df$cell_col    <- xl_cell_general(value = 99L)  # all rows get 99
}
\seealso{
[xl_formula()], [xl_hyperlink()], [write_xlsx()]

Other cell content:
\code{\link[=is_xl_comment]{is_xl_comment()}},
\code{\link[=xl_comment]{xl_comment()}},
\code{\link[=xl_formula]{xl_formula()}},
\code{\link[=xl_rich_run]{xl_rich_run()}},
\code{\link[=xl_rich_string]{xl_rich_string()}}
}
\concept{cell content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_comment.R
\name{xl_comment}
\alias{xl_comment}
\title{Create a cell comment}
\usage{
xl_comment(
  value,
  format = NULL,
  author = NA,
  visible = NA,
  width_pixels = NA,
  height_pixels = NA,
  x_scale = NA,
  y_scale = NA,
  start_row = NA,
  start_col = NA,
  x_offset = NA,
  y_offset = NA
)
}
\arguments{
\item{value}{A single string: the comment's text.  Anything with an
[as.character()] method is accepted, so a cell built for a sheet can be
reused here.}

\item{format}{An optional [xl_format]; only its fill background color and
font name/size/family are used (see [xl_fill()], [xl_font()]).  Other
properties are unsupported and warned about.}

\item{author}{Comment author (shown in Excel's status bar).  Defaults to the
sheet/workbook `comment_author` when unset (see [xl_sheet()],
[xl_properties()]).}

\item{visible}{Initial visibility: `NA` follows the sheet default (comments
are hidden unless `show_comments` is set), `TRUE` shows this comment,
`FALSE` hides it.}

\item{width_pixels, height_pixels}{Comment box size in pixels (defaults
128 x 74).}

\item{x_scale, y_scale}{Box scale factors.}

\item{start_row, start_col}{Zero-based anchor cell of the box (by default a
comment sits one row up and one column right of its cell).}

\item{x_offset, y_offset}{Pixel offset of the box from its anchor.}
}
\value{
An `xl_comment` object.
}
\description{
`xl_comment()` builds a comment (note) that can be attached to a cell via
[xl_cell_general()]'s `comment` argument.  The simplest form is just text
(`xl_cell_general(value = x, comment = "see note")`); `xl_comment()` adds
options such as the author, initial visibility, box size and position, and
styling.

Excel comment boxes support only a **background color** and a **font
name/size/family**.  These are supplied by reusing the formatting engine via
the `format` argument (e.g. `xl_font(name = "Arial", size = 10) +
xl_fill(background = "lightyellow")`); any other format property (bold,
borders, number formats, ...) is not supported by comments and triggers a
warning.
}
\examples{
# plain text
xl_comment("Double-check this figure")

# with author and styling reused from the format engine
xl_comment("Estimate", author = "Finance",
           format = xl_font(name = "Arial", size = 10) +
                    xl_fill(background = "lightyellow"))
}
\seealso{
[xl_cell_general], [xl_format]

Other cell content:
\code{\link[=is_xl_comment]{is_xl_comment()}},
\code{\link[=xl_cell_general]{xl_cell_general()}},
\code{\link[=xl_formula]{xl_formula()}},
\code{\link[=xl_rich_run]{xl_rich_run()}},
\code{\link[=xl_rich_string]{xl_rich_string()}}
}
\concept{cell content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_axis.R
\name{xl_chart_axis}
\alias{xl_chart_axis}
\title{An axis of a chart}
\usage{
xl_chart_axis(
  title = NULL,
  title_format = NULL,
  title_layout = NULL,
  label_format = NULL,
  num_format = NULL,
  line_format = NULL,
  visible = NA,
  reverse = NA,
  min = NA,
  max = NA,
  log_base = NA,
  major_unit = NA,
  minor_unit = NA,
  display_units = NULL,
  display_units_visible = NA,
  interval_unit = NA,
  interval_tick = NA,
  position = NULL,
  label_position = NULL,
  label_align = NULL,
  major_tick = NULL,
  minor_tick = NULL,
  crossing = NULL,
  major_gridlines = NA,
  minor_gridlines = NA,
  major_gridlines_format = NULL,
  minor_gridlines_format = NULL
)
}
\arguments{
\item{title}{The axis title: a string, or a range spec holding one --- see
[xl_chart_series()] for the spellings, including
`list(header = "revenue")`.}

\item{title_format}{An [xl_format()] styling the axis title.  A title is
text, so only the [xl_font()] group applies.}

\item{title_layout}{Where to put the axis title by hand, as `c(x, y)`
fractions of the chart, each above 0 and at most 1.  Excel places it for
you otherwise.}

\item{label_format}{An [xl_format()] styling the tick labels --- the
[xl_font()] group only.}

\item{num_format}{A number format for the tick labels, as an Excel format
string (`"#,##0"`) or an [xl_num_format()].}

\item{line_format}{An [xl_format()] styling the axis line itself:
[xl_border()] for the line, [xl_fill()] for the fill behind it.  An axis
has four parts that can be styled, so none of them is just `format`.}

\item{visible}{`FALSE` hides the axis.}

\item{reverse}{Draw the axis in the opposite direction.}

\item{min, max}{The axis bounds.  Value axes only.}

\item{log_base}{Use a logarithmic scale with this base, 2 or more.  Value
axes only.}

\item{major_unit, minor_unit}{The spacing between major and minor tick marks.
Value axes only.}

\item{display_units}{Scale the labels by `"thousands"`, `"millions"`,
`"billions"` and so on; see Details for the full set.  Value axes only.}

\item{display_units_visible}{Whether the caption naming the units --- the
small rotated "Millions" beside the axis --- is drawn.  Setting
`display_units` turns it **on**, as Excel does, so this is really for
`FALSE`: rescaled labels with no caption.  Value axes only.}

\item{interval_unit}{Label one category in every `n`.  Category axes only.}

\item{interval_tick}{Put a tick mark on one category in every `n`.
Category axes only.}

\item{position}{Whether the data sits `"on_tick"` or `"between"` the tick
marks.  Category axes only.}

\item{label_position}{Where the tick labels go: `"next_to"`, `"high"`,
`"low"`, or `"none"` for no labels.}

\item{label_align}{Tick-label alignment: `"center"`, `"left"` or `"right"`.
Category axes only.}

\item{major_tick, minor_tick}{The tick marks: `"default"`, `"none"`,
`"inside"`, `"outside"` or `"crossing"`.}

\item{crossing}{Where the other axis crosses this one: a number, or `"min"`
or `"max"` for either end.}

\item{major_gridlines, minor_gridlines}{Show the gridlines.  A chart's major
y gridlines are on by default and everything else is off.}

\item{major_gridlines_format, minor_gridlines_format}{An [xl_format()]
styling the gridlines --- [xl_border()] only, since a gridline is a line.}
}
\value{
An `xl_chart_axis` object.
}
\description{
`xl_chart_axis()` describes one axis, and is given to [xl_chart()] as
`x_axis` or `y_axis`.

Several options apply to one kind of axis only, and Excel discards the rest
without a word, so writexl refuses them instead.  A scatter chart plots
numbers against numbers, so both of its axes are **value** axes; every other
type has a **category** x axis and a value y axis.  (A bar chart is drawn
with its categories up the side, but the axes keep their names.)  Pie and
doughnut charts have no axes at all.

* value axes only --- `min`, `max`, `log_base`, `major_unit`, `minor_unit`,
  `display_units`, `display_units_visible`;
* category axes only --- `position`, `label_align`, `interval_unit`,
  `interval_tick`.
}
\details{
`display_units` is one of `"none"`, `"hundreds"`, `"thousands"`,
`"ten_thousands"`, `"hundred_thousands"`, `"millions"`, `"ten_millions"`,
`"hundred_millions"`, `"billions"` or `"trillions"`.
}
\examples{
xl_chart_axis(title = "Quarter")
xl_chart_axis(title = "Revenue", min = 0, num_format = "$#,##0",
              major_gridlines = FALSE)
}
\seealso{
[xl_chart], [xl_chart_series]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_parts.R
\name{xl_chart_labels}
\alias{xl_chart_labels}
\alias{xl_chart_label}
\title{Data labels on a chart series}
\usage{
xl_chart_labels(
  show_value = NA,
  show_name = NA,
  show_category = NA,
  show_percentage = NA,
  show_legend_key = NA,
  num_format = NULL,
  position = NULL,
  separator = NULL,
  format = NULL,
  leader_lines = NA,
  custom = NULL
)

xl_chart_label(value = NULL, hide = NA, format = NULL)
}
\arguments{
\item{show_value, show_name, show_category, show_percentage}{What each label
holds: the point's value, the series name, the category, and the value as
a percentage of the series.  Naming any of them means the label holds
exactly those, so `show_percentage = TRUE` alone gives a percentage and
nothing else; naming none of them leaves Excel's default, the value.
Percentages are meaningful on pie and doughnut charts.}

\item{show_legend_key}{Print the series' legend swatch in each label.}

\item{num_format}{A number format for the labels, as an Excel format string
or an [xl_num_format()].}

\item{position}{Where the label sits relative to its point; see the
description for which chart types allow which.}

\item{separator}{What joins the parts of a label when it holds more than
one: `"comma"`, `"semicolon"`, `"period"`, `"newline"` or `"space"`.}

\item{format}{An [xl_format()] styling the labels.  A label is a shape with
text in it, so all of [xl_font()], [xl_border()] and [xl_fill()] apply.}

\item{leader_lines}{Draw a line from a label back to its point.  Excel only
shows one once the label has been dragged away from the point.}

\item{custom}{A list of [xl_chart_label()]s, one per point in order, giving
individual labels their own text, styling, or `hide = TRUE`.  `NULL` in
the list leaves that point's label alone.}

\item{value}{The label's text.  A string beginning with `"="` is a formula,
so `"=Sheet1!$A$1"` takes the text from a cell.}

\item{hide}{Remove this point's label, leaving the others.}
}
\value{
An `xl_chart_labels` object.
}
\description{
`xl_chart_labels()` prints the numbers next to the points that carry them.
With no arguments it shows the value of each point, which is Excel's own
default.

`xl_chart_label()` describes one label, for `custom`: a label can be given
its own text or hidden, one point at a time.

Excel allows different label positions for different chart types, and drops
one that does not apply.  The table is in libxlsxwriter's header and is
enforced here: `"center"` is allowed everywhere, `"right"`, `"left"`,
`"above"` and `"below"` on line and scatter charts, `"inside_base"` on bar
and column, `"inside_end"` and `"outside_end"` on bar, column, pie and
doughnut, and `"best_fit"` on pie and doughnut.
}
\examples{
xl_chart_labels()
xl_chart_labels(show_category = TRUE, show_percentage = TRUE,
                separator = "newline", position = "outside_end")
}
\seealso{
[xl_chart_series]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_format.R
\name{xl_color}
\alias{xl_color}
\title{Normalize a color to a libxlsxwriter RGB integer}
\usage{
xl_color(x)
}
\arguments{
\item{x}{A single color: an R color name, a hex string, or an integer.
`NA` returns `NA_integer_` (an unset color).}
}
\value{
A single integer in `0x000000`..`0xFFFFFF`, or `NA_integer_`.
}
\description{
Converts an R color name (e.g. `"navy"`), a hex string (`"#FF0000"` or
`"FF0000"`), or an integer in the range `0x000000`..`0xFFFFFF` into the
single `0xRRGGBB` integer that libxlsxwriter expects.  Used internally by
every `color`/`background`/`foreground` argument of the formatting
constructors, and exported so it can be used directly.
}
\examples{
xl_color("red")
xl_color("#0000FF")
xl_color(255L)
}
\seealso{
Other cell formatting:
\code{\link[=is_xl_format]{is_xl_format()}},
\code{\link[=xl_format]{xl_format()}},
\code{\link{xl_format_groups}}
}
\concept{cell formatting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chartsheet.R
\name{xl_chartsheet}
\alias{xl_chartsheet}
\title{A sheet holding a single chart}
\usage{
xl_chartsheet(
  chart,
  tab_color = NULL,
  zoom = NA,
  protect = NULL,
  page = NULL,
  view = NULL
)
}
\arguments{
\item{chart}{The [xl_chart()] to fill the sheet with.}

\item{tab_color}{The colour of the sheet tab.}

\item{zoom}{The zoom level as a percentage, 10 to 400.}

\item{protect}{`TRUE`, a password string, or a named list. A chartsheet
has no cells, so of Excel's protection options it takes only
`no_content` (let the chart be edited) and `no_objects` (let the shapes
on it be edited); the worksheet options are refused by name.}

\item{page}{An [xl_page_setup()] describing how it prints.}

\item{view}{An [xl_sheet_view()] setting the tab state.}
}
\value{
An `xl_chartsheet` object.
}
\description{
`xl_chartsheet()` is a worksheet-sized chart: a tab of its own holding one
chart and no cells.  Give it to [write_xlsx()] in place of a data frame.

Because a chartsheet has no cells, every range in the chart's series must
name the sheet it plots --- `list(sheet = "Data", cols = "revenue")` or
`"Data!B2:B10"`.  A bare `list(cols = )` has nothing to resolve against and
is refused.

A chartsheet supports only part of what a worksheet does, and the parts it
does not are refused rather than dropped: of [xl_page_setup()] it takes the
orientation, paper size, margins and the header and footer; of
[xl_sheet_view()] it takes `active`, `selected`, `visible` and
  `first_tab`.
}
\examples{
sales <- data.frame(quarter = c("Q1", "Q2"), revenue = c(10, 25))
chart <- xl_chart("column",
                  xl_chart_series(values = list(sheet = "Data",
                                                cols = "revenue")))
write_xlsx(list(Data = sales, Overview = xl_chartsheet(chart)),
           tempfile(fileext = ".xlsx"))
}
\seealso{
[xl_chart], [xl_sheet]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_conditional.R
\name{xl_conditional}
\alias{xl_conditional}
\alias{xl_cond_cell}
\alias{xl_cond_scale}
\alias{xl_cond_bar}
\alias{xl_cond_icons}
\title{Format cells according to their contents}
\usage{
xl_cond_cell(
  range,
  type = NA,
  criteria = NA,
  value = NULL,
  min = NULL,
  max = NULL,
  format = NULL,
  stop_if_true = NA,
  multi_range = NA
)

xl_cond_scale(
  range,
  colors = c("red", "yellow", "green"),
  values = NULL,
  rule_types = NULL,
  stop_if_true = NA,
  multi_range = NA
)

xl_cond_bar(
  range,
  color = NA,
  values = NULL,
  rule_types = NULL,
  solid = NA,
  negative_color = NA,
  border_color = NA,
  negative_border_color = NA,
  no_border = NA,
  direction = NA,
  axis = NA,
  axis_color = NA,
  bar_only = NA,
  stop_if_true = NA,
  multi_range = NA
)

xl_cond_icons(
  range,
  style = "3_traffic_lights",
  reverse = NA,
  icons_only = NA,
  stop_if_true = NA,
  multi_range = NA
)
}
\arguments{
\item{range}{The cells the rule applies to: an Excel range string such as
`"B2:B100"`, a single cell, or a `list(rows = , cols = )` spec.}

\item{type}{The kind of rule, when it cannot be inferred from `criteria`:
`"cell"`, `"text"`, `"time_period"`, `"average"`, `"top"`, `"bottom"`,
`"duplicate"`, `"unique"`, `"blanks"`, `"no_blanks"`, `"errors"`,
`"no_errors"` or `"formula"`.}

\item{criteria}{How the rule decides, which depends on `type`:
  `"=="`, `"!="`, `">"`, `"<"`, `">="`, `"<="`, `"between"`, `"not between"`
  for `"cell"`; `"contains"`, `"not contains"`, `"begins with"`,
  `"ends with"` for `"text"`; `"yesterday"`, `"today"`, `"tomorrow"`,
  `"last 7 days"`, `"last week"`, `"this week"`, `"next week"`,
  `"last month"`, `"this month"`, `"next month"` for `"time_period"`;
  `"above"`, `"below"`, `"above or equal"`, `"below or equal"` and the
  `"N std dev above"` / `"below"` variants for `"average"`; and `"percent"`
  for `"top"` / `"bottom"`.

  Pairing a criteria with the wrong `type` is an error --- Excel would accept
  the file and silently ignore the rule.}

\item{value}{What the criteria compares against: a number, a string (for the
text criteria), or an `"=..."` formula.  For `"top"` / `"bottom"` it is the
N.  Use `min` and `max` for `"between"` / `"not between"`.}

\item{min, max}{The two bounds for `"between"` / `"not between"`.}

\item{format}{The [xl_format] applied to cells that match.}

\item{stop_if_true}{Logical; if this rule matches, skip the later rules on
the same cells.}

\item{multi_range}{A further set of ranges the rule also covers, as an Excel
multi-range string such as `"B3:K6 B9:K12"`.}

\item{colors}{Two or three colours for the scale, from lowest to highest.
Two gives a two-colour scale, three a three-colour scale.}

\item{values}{Optional values marking where each colour sits, in the same
order as `colors`.  Defaults to the range's minimum, midpoint and maximum.}

\item{rule_types}{How each entry of `values` is interpreted: `"minimum"`,
`"maximum"`, `"number"`, `"percent"`, `"percentile"`, `"formula"`,
`"auto_min"` or `"auto_max"`.}

\item{color}{The bar's fill colour.}

\item{solid}{Logical; a solid bar rather than Excel's default gradient.}

\item{negative_color, border_color, negative_border_color, axis_color}{Colours
for the negative portion, the bar border, the negative border, and the
axis line.}

\item{no_border}{Logical; draw the bar without a border.}

\item{direction}{`"context"` (follow the sheet), `"left to right"` or
`"right to left"`.}

\item{axis}{Where the zero axis sits: `"automatic"`, `"midpoint"` or
`"none"`.}

\item{bar_only}{Logical; show the bar without the cell's value.}

\item{style}{Which built-in icon set: one of `"3_arrows"`, `"3_arrows_gray"`,
`"3_flags"`, `"3_traffic_lights"`, `"3_traffic_lights_rimmed"`,
`"3_signs"`, `"3_symbols_circled"`, `"3_symbols"`, `"4_arrows"`,
`"4_arrows_gray"`, `"4_red_to_black"`, `"4_ratings"`,
`"4_traffic_lights"`, `"5_arrows"`, `"5_arrows_gray"`, `"5_ratings"` or
`"5_quarters"`.  These are Excel's own icons, not images, so nothing is
embedded in the file.}

\item{reverse}{Logical; reverse the order the icons are assigned in.}

\item{icons_only}{Logical; show the icon without the cell's value.}
}
\value{
An `xl_conditional` object.
}
\description{
Excel's conditional formatting, in four flavours:

* `xl_cond_cell()` --- a rule with a format: comparisons, text matches, time
  periods, above/below average, top/bottom N, duplicates, blanks, errors, or
  an arbitrary formula.
* `xl_cond_scale()` --- a two- or three-colour scale across the range.
* `xl_cond_bar()` --- in-cell data bars.
* `xl_cond_icons()` --- one of Excel's built-in icon sets.

Pass one or a list of them as `xl_sheet(conditional = )`.
}
\examples{
xl_cond_cell("B2:B100", criteria = ">", value = 100, format = xl_fill(background = "red"))
xl_cond_cell("C2:C100", type = "text", criteria = "contains", value = "urgent",
             format = xl_font(bold = TRUE))
xl_cond_cell("D2:D100", type = "duplicate",
             format = xl_fill(background = "yellow"))
xl_cond_scale("C2:C100", colors = c("red", "yellow", "green"))
xl_cond_scale("C2:C100", colors = c("white", "steelblue"))
xl_cond_bar("D2:D100", color = "steelblue")
xl_cond_bar("D2:D100", color = "green", solid = TRUE, bar_only = TRUE)
xl_cond_icons("E2:E100", style = "3_traffic_lights")
xl_cond_icons("E2:E100", style = "5_ratings", icons_only = TRUE)
}
\seealso{
[xl_sheet], [xl_format]

Other worksheet features:
\code{\link[=xl_filter]{xl_filter()}},
\code{\link[=xl_filter_keep]{xl_filter_keep()}},
\code{\link[=xl_merge]{xl_merge()}},
\code{\link[=xl_table]{xl_table()}},
\code{\link[=xl_table_column]{xl_table_column()}},
\code{\link[=xl_validation]{xl_validation()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_table.R
\name{xl_table}
\alias{xl_table}
\title{Add a worksheet table}
\usage{
xl_table(
  range = NULL,
  name = NULL,
  style = "medium 9",
  header_row = TRUE,
  autofilter = TRUE,
  banded_rows = TRUE,
  banded_columns = FALSE,
  first_column = FALSE,
  last_column = FALSE,
  total_row = FALSE,
  columns = NULL
)
}
\arguments{
\item{range}{The cells the table covers, as an Excel range string or a
`list(rows = , cols = )` spec.  Defaults to the sheet's used range,
including the header row and, with `total_row = TRUE`, one row below.}

\item{name}{The table's name.  See "Table names".}

\item{style}{The table style: `"none"`, or a type and number such as
`"medium 9"` (the default), `"light 21"` or `"dark 11"`.  Light styles are
numbered 0--21, medium 1--28 and dark 1--11.}

\item{header_row}{Show the header row.  Turning it off also removes the
filter dropdown, as it does in Excel.}

\item{autofilter}{Show the filter dropdown in the header row.}

\item{banded_rows, banded_columns}{Alternating row / column shading.}

\item{first_column, last_column}{Highlight the first / last column.}

\item{total_row}{Add a total row below the data.  See "The total row".}

\item{columns}{One [xl_table_column()], or a list of them, overriding
individual columns.}
}
\value{
An `xl_table` object.
}
\description{
`xl_table()` turns a range into an Excel table: a named, styled block with
banded rows, a filter dropdown in its header, and an optional total row.
Pass one or a list of them as `xl_sheet(table = )`.

Column headers default to the data frame's column names.  That is not just a
convenience: Excel records a table's column names separately from the header
cells and rejects a file where the two disagree, so the default is what makes
a table safe to add at all.
}
\section{Table names}{

Excel requires table names to be unique across the workbook, and formulas
refer to a table by name.  `name = NULL` (the default) has writexl generate a
unique name from the sheet name and write it explicitly, so it cannot shift
when another table is added elsewhere.  Give a string to choose your own; it
is validated and checked for collisions.  Give `NA` to write no name at all
and let Excel assign `Table1`, `Table2`, ... by insertion order --- which
makes any formula naming the table fragile, so that combination warns.
}

\section{The total row}{

A total row is written by libxlsxwriter below the data, and the sheet's row
plan does not know about it: `auto_colwidth` does not measure it, and an
[xl_row_spec()] aimed at that row will fight it.  It has no cells of its own
until a column gives a `total` or `total_label`.
}

\examples{
xl_table(style = "light 9")
xl_table(name = "Sales", total_row = TRUE,
         columns = list(xl_table_column("fruit", total_label = "Total"),
                        xl_table_column("qty", total = "sum")))
}
\seealso{
[xl_table_column], [xl_sheet]

Other worksheet features:
\code{\link{xl_conditional}},
\code{\link[=xl_filter]{xl_filter()}},
\code{\link[=xl_filter_keep]{xl_filter_keep()}},
\code{\link[=xl_merge]{xl_merge()}},
\code{\link[=xl_table_column]{xl_table_column()}},
\code{\link[=xl_validation]{xl_validation()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_format.R
\name{is_xl_format}
\alias{is_xl_format}
\title{Test whether an object is an `xl_format`}
\usage{
is_xl_format(x)
}
\arguments{
\item{x}{An object.}
}
\value{
`TRUE` if `x` inherits from `"xl_format"`.
}
\description{
Test whether an object is an `xl_format`
}
\seealso{
Other cell formatting:
\code{\link[=xl_color]{xl_color()}},
\code{\link[=xl_format]{xl_format()}},
\code{\link{xl_format_groups}}
}
\concept{cell formatting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_filter.R
\name{xl_filter}
\alias{xl_filter}
\title{Filter an autofilter column, hiding the rows that do not match}
\usage{
xl_filter(
  col,
  criteria = NA,
  value = NULL,
  criteria2 = NA,
  value2 = NULL,
  and_or = "and",
  list = NULL
)
}
\arguments{
\item{col}{The column to filter: a name or a 1-based position.}

\item{criteria}{One of `"=="`, `"!="`, `">"`, `"<"`, `">="`, `"<="`,
`"blanks"` or `"non-blanks"`.  The four magnitude comparisons need a
numeric value; writexl will not guess how Excel orders text.}

\item{value}{The value to compare against.  In a text comparison `*` matches
any run of characters and `?` any single one.}

\item{criteria2, value2}{An optional second rule for the same column.}

\item{and_or}{How the two rules combine: `"and"` (default) or `"or"`.}

\item{list}{Instead of a criteria, keep only rows whose value is in this
character vector.  Matched case-insensitively, as Excel does.}
}
\value{
An `xl_filter` object.
}
\description{
`xl_filter()` sets the criteria on one autofilter column *and* hides the rows
that do not match.  Both halves are necessary: Excel stores the criteria and
the hidden rows separately and does not apply a filter when a file is opened,
so criteria alone produce a sheet that looks filtered but shows every row.

Because writexl decides which rows to hide, it reproduces Excel's own
matching rules, which were measured in Excel rather than assumed.  Those
rules depend on which of two forms the filter takes:

* `"=="` (with no wildcard) and `"blanks"`, and any `list`, are written as a
  **value list**.  Excel matches these against the text a cell *displays*,
  case-insensitively --- so `"=="` with `10`, with `"10"`, or a `list` of
  `"10"` all match both the number `10` and the string `"10"`.
* every other criteria, including `"=="` with a `*` or `?` in it, is written
  as a **typed comparison**.  If the value is a number the comparison is
  numeric and a text cell never satisfies it (except `"!="`, which a text
  cell satisfies because it is not that number).  If the value is text the
  comparison is textual, case-insensitive, with `*` and `?` as wildcards,
  and a number cell never satisfies it.

The consequence worth knowing is that `"=="` with `"10"` keeps the number
`10`, while `"=="` with `"1*"` keeps nothing on a numeric column: the
wildcard changes the form, and so the rule.

Blank means an empty cell or an empty string; `"non-blanks"` is its exact
complement.  Blanks are excluded by every comparison except `"!="`, which
keeps them --- a blank is not equal to anything.  A mixed-type column built
with [xl_cell_general()] is matched cell by cell, each by its own type.
}
\section{Limitations}{

A value list matches displayed text, which writexl can only predict for the
General format.  Filtering a `Date` or `POSIXct` column by `"=="` or `list`
is therefore refused --- use a comparison such as `">="`.  For the same
reason a numeric column carrying a custom number format (say two decimal
places, or a currency symbol) may display differently from what writexl
compares, so prefer a comparison there too.  Columns whose cells hold
formulas cannot be filtered at all, since writexl does not know what Excel
would compute.
}

\examples{
xl_filter("qty", ">", 100)
xl_filter("qty", ">", 100, "<", 200)          # between, via two rules
xl_filter("fruit", "==", "ap*")               # wildcard: typed comparison
xl_filter("fruit", list = c("apple", "banana"))
xl_filter("qty", "==", "10")                  # value list: matches the
                                              # number 10 and the text "10"
}
\seealso{
[xl_sheet], [xl_filter_keep]

Other worksheet features:
\code{\link{xl_conditional}},
\code{\link[=xl_filter_keep]{xl_filter_keep()}},
\code{\link[=xl_merge]{xl_merge()}},
\code{\link[=xl_table]{xl_table()}},
\code{\link[=xl_table_column]{xl_table_column()}},
\code{\link[=xl_validation]{xl_validation()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_format.R
\name{xl_format}
\alias{xl_format}
\alias{+.xl_format}
\title{Combine cell-formatting groups into a single format}
\usage{
xl_format(..., quote_prefix = NA, hyperlink = NA)

\method{+}{xl_format}(e1, e2)
}
\arguments{
\item{...}{[xl_format] objects (and/or `NULL`s, which are ignored).  May also
include the scalar flags `quote_prefix` and `hyperlink`.}

\item{quote_prefix}{Logical; treat the cell contents as literal text (as if
prefixed with a single quote in Excel).}

\item{hyperlink}{Logical; apply the internal hyperlink style flag (advanced).}

\item{e1, e2}{[xl_format] objects to combine.}
}
\value{
A combined [xl_format] object.
}
\description{
`xl_format()` merges any number of [xl_format] objects (typically the
single-group objects returned by [xl_font], [xl_fill], [xl_border],
[xl_align], [xl_num_format] and [xl_protection]) into one combined format.
The `+` operator does the same for two formats.

Merging is right-biased and works property-by-property: where two formats set
the *same* property the later one wins, but properties set by only one side
are all preserved.  So partial groups accumulate rather than overwrite.
}
\examples{
xl_format(xl_font(bold = TRUE), xl_border(bottom = "thin"),
          xl_num_format("#,##0.00"))

# equivalent, using +
xl_font(bold = TRUE) + xl_border(bottom = "thin") + xl_num_format("#,##0.00")
}
\seealso{
[xl_font], [xl_fill], [xl_border], [xl_align], [xl_num_format],
  [xl_protection], [xl_color]

Other cell formatting:
\code{\link[=is_xl_format]{is_xl_format()}},
\code{\link[=xl_color]{xl_color()}},
\code{\link{xl_format_groups}}
}
\concept{cell formatting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_parts.R
\name{xl_chart_marker}
\alias{xl_chart_marker}
\title{A marker on a chart series}
\usage{
xl_chart_marker(type = NULL, size = NA, format = NULL)
}
\arguments{
\item{type}{The symbol: `"automatic"`, `"none"`, `"square"`, `"diamond"`,
`"triangle"`, `"x"`, `"star"`, `"short_dash"`, `"long_dash"`, `"circle"`
or `"plus"`.}

\item{size}{The symbol's size in points, 2 to 72.}

\item{format}{An [xl_format()] styling the symbol: [xl_border()] for its
outline, [xl_fill()] for its fill or pattern.}
}
\value{
An `xl_chart_marker` object.
}
\description{
`xl_chart_marker()` draws a symbol at each point of a series.  Line, scatter
and radar charts are where they show; on other types Excel ignores them.

`type = "automatic"` asks Excel for the default marker of that series, and
is the one type that cannot be given a size or a format --- libxlsxwriter
documents that, and Excel drops them, so both are refused here.
}
\examples{
xl_chart_marker(type = "circle", size = 8)
xl_chart_marker(type = "none")
}
\seealso{
[xl_chart_series]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_filter.R
\name{xl_filter_keep}
\alias{xl_filter_keep}
\title{Which rows an Excel autofilter would leave visible}
\usage{
xl_filter_keep(data, filter)
}
\arguments{
\item{data}{A data frame whose columns the filters name.}

\item{filter}{One [xl_filter()], or a list of them.  Filters on different
columns combine with AND, as they do in Excel.}
}
\value{
A logical vector with one element per row of `data`, `TRUE` where
  the row stays visible.
}
\description{
`xl_filter_keep()` answers, for a data frame and a set of [xl_filter()]
rules, the question `xl_sheet(filter =)` has to answer internally: which
rows does Excel leave visible?  It writes nothing --- it is the matching
rule on its own, exported because reproducing Excel's filter semantics is
hard to get right and useful outside writing a file.

The rules are described in detail under [xl_filter()], and were established
by measurement rather than from documentation: a workbook was written with
criteria set and no rows hidden, opened in Excel, and Data > Reapply pressed
so that Excel computed each match itself.
}
\section{Accuracy}{

This function tracks Excel's *observed* behaviour, so a case found to
disagree with Excel is treated as a bug and fixed, which may change the rows
it returns.  One limitation is known: a filter written as a value list
(`"=="` without a wildcard, or `list`) matches the text a cell **displays**,
which writexl can only predict for the General format.  A numeric column
carrying a custom number format may therefore display differently from what
is compared here --- prefer a comparison such as `">="` on such a column.
Dates are refused outright for the same reason.  See [xl_filter()].
}

\examples{
sales <- data.frame(fruit = c("apple", "banana", "cherry"),
                    qty = c(5, 150, 300))
xl_filter_keep(sales, xl_filter("qty", ">", 100))
sales[xl_filter_keep(sales, xl_filter("fruit", "==", "*a*")), ]

# filters on different columns combine with AND
xl_filter_keep(sales, list(xl_filter("qty", ">", 100),
                           xl_filter("fruit", "==", "b*")))
}
\seealso{
[xl_filter], [xl_sheet]

Other worksheet features:
\code{\link{xl_conditional}},
\code{\link[=xl_filter]{xl_filter()}},
\code{\link[=xl_merge]{xl_merge()}},
\code{\link[=xl_table]{xl_table()}},
\code{\link[=xl_table_column]{xl_table_column()}},
\code{\link[=xl_validation]{xl_validation()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_rich_string.R
\name{xl_rich_run}
\alias{xl_rich_run}
\title{One run of a rich (multi-format) string}
\usage{
xl_rich_run(value, format = NULL)
}
\arguments{
\item{value}{A single non-`NA`, non-empty string: the run's text.}

\item{format}{An optional [xl_format].  Only its **font** properties apply
(see [xl_font()]); a run has no fill, border, alignment or number format,
and supplying one warns.  `NULL` draws the run in the cell's own font.}
}
\value{
An `xl_rich_run` object.
}
\description{
`xl_rich_run()` is one fragment of an [xl_rich_string()]: a piece of text
plus the font it is drawn in.
}
\examples{
xl_rich_run("bold", xl_font(bold = TRUE))
xl_rich_run("plain")
}
\seealso{
[xl_rich_string], [xl_format]

Other cell content:
\code{\link[=is_xl_comment]{is_xl_comment()}},
\code{\link[=xl_cell_general]{xl_cell_general()}},
\code{\link[=xl_comment]{xl_comment()}},
\code{\link[=xl_formula]{xl_formula()}},
\code{\link[=xl_rich_string]{xl_rich_string()}}
}
\concept{cell content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version.R
\name{lxw_version}
\alias{lxw_version}
\title{Version}
\usage{
lxw_version()
}
\description{
Shows version of bundled libxlsxwriter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart.R
\name{xl_chart_series}
\alias{xl_chart_series}
\title{A data series within a chart}
\usage{
xl_chart_series(
  values,
  categories = NULL,
  name = NULL,
  format = NULL,
  marker = NULL,
  labels = NULL,
  trendline = NULL,
  x_error_bars = NULL,
  y_error_bars = NULL,
  points = NULL,
  smooth = NA,
  invert_if_negative = NA
)
}
\arguments{
\item{values}{The range holding the numbers to plot.}

\item{categories}{The range holding the labels to plot them against.  Omit
for a chart that numbers its points.}

\item{name}{The series name, shown in the legend.  Left unset, a series that
plots a column takes its name from that column's header cell, which is
what Excel does when you chart a column along with its header; `FALSE`
leaves it unnamed.  A string is always taken literally --- a series may
legitimately be called `"Q1!"` --- so to take the name from another cell,
give a range spec: `name = list(header = "cost")` for a different column's
header, or `name = list(rows = 1, cols = 1)` for a data cell.}

\item{format}{An [xl_format] styling the series --- its line and fill.  See
[xl_chart()] for which format properties a chart can express.}

\item{marker}{An [xl_chart_marker()] drawn at each point.}

\item{labels}{An [xl_chart_labels()] printing the numbers beside the points.}

\item{trendline}{An [xl_chart_trendline()] fitted through the series.}

\item{x_error_bars, y_error_bars}{An [xl_chart_error_bars()] on each point.}

\item{points}{An [xl_format()] per point, as a list, styling individual
points --- one slice of a pie, one bar of a column chart.  `NULL` in the
list leaves that point as it is.}

\item{smooth}{Draw the line smoothed.  Line and scatter charts only.}

\item{invert_if_negative}{Fill negative values with the inverse colour.}
}
\value{
An `xl_chart_series` object.
}
\description{
`xl_chart_series()` names the values a chart plots, and optionally the
categories to plot them against and a name for the legend.  A series that
plots a column is named after that column's header unless told otherwise.

Each range may live on a different sheet from the chart, so it takes an
optional `sheet`:

* `"Data!B2:B10"` --- an A1 range, sheet-qualified;
* `list(cols = "revenue")` --- resolved against the chart's own sheet;
* `list(sheet = "Data", cols = "revenue")` --- against another sheet;
* `list(header = "revenue")` --- that column's header cell, which is where
  a series name usually lives.

A range that selects no data is an error rather than an empty chart.
}
\examples{
xl_chart_series(values = list(cols = "revenue"))
xl_chart_series(values = "Data!B2:B10", categories = "Data!A2:A10",
                name = "2024")
}
\seealso{
[xl_chart]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_table.R
\name{xl_table_column}
\alias{xl_table_column}
\title{Describe a column of a worksheet table}
\usage{
xl_table_column(
  col,
  header = NULL,
  formula = NULL,
  total = NULL,
  total_label = NULL,
  total_value = NULL,
  format = NULL,
  header_format = NULL
)
}
\arguments{
\item{col}{The column: a name or a 1-based position within the table's range.}

\item{header}{The header caption.  Defaults to the data frame's column name
--- which is also what makes a table safe to add, since Excel rejects a
file whose table definition and header cells disagree.}

\item{formula}{A formula filling every data cell of the column, usually with
a structured reference such as `"=SUM(Sales[@[Q1]:[Q4]])"`.  Because those
name the table, see the `name` argument of [xl_table()].}

\item{total}{The function shown in this column's total-row cell: one of
`"sum"`, `"average"`, `"count"`, `"count_nums"`, `"max"`, `"min"`,
`"std_dev"` or `"var"`.  Needs `total_row = TRUE` on the table.}

\item{total_label}{A string for the total-row cell instead of a function ---
typically `"Total"` under the first column.}

\item{total_value}{The number the total-row cell already holds. Excel
recalculates `total` on open and ignores this, but a reader that does not
evaluate formulas --- [readxl::read_xlsx()] among them --- shows what is
cached, which without this is nothing.}

\item{format}{An [xl_format] applied to the column's data cells.}

\item{header_format}{An [xl_format] applied to the column's header cell.}
}
\value{
An `xl_table_column` object.
}
\description{
`xl_table_column()` overrides what [xl_table()] does with one column: its
header caption, a formula filling the column, what its total-row cell shows,
and the formats applied to the header and the data cells.

Only the columns you name need an entry; the rest take the data frame's
column name as their header and are left otherwise alone.
}
\examples{
xl_table_column("qty", total = "sum")
xl_table_column("fruit", total_label = "Total")
xl_table_column("margin", formula = "=Sales[@revenue] * 0.3")
}
\seealso{
[xl_table], [xl_sheet]

Other worksheet features:
\code{\link{xl_conditional}},
\code{\link[=xl_filter]{xl_filter()}},
\code{\link[=xl_filter_keep]{xl_filter_keep()}},
\code{\link[=xl_merge]{xl_merge()}},
\code{\link[=xl_table]{xl_table()}},
\code{\link[=xl_validation]{xl_validation()}}
}
\concept{worksheet features}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_extras.R
\name{xl_chart_table}
\alias{xl_chart_table}
\title{The table of values under a chart}
\usage{
xl_chart_table(
  show_keys = NA,
  horizontal_border = NA,
  vertical_border = NA,
  outline_border = NA,
  format = NULL
)
}
\arguments{
\item{show_keys}{Print each series' legend swatch in the table.}

\item{horizontal_border, vertical_border, outline_border}{Which of the grid's
borders to draw.}

\item{format}{An [xl_format()] styling the table's text --- the [xl_font()]
group only.}
}
\value{
An `xl_chart_table` object.
}
\description{
`xl_chart_table()` prints the plotted numbers in a grid beneath the chart,
which is Excel's "Data Table" chart element.  It is given to [xl_chart()] as
`data_table`.

Naming none of the grid options leaves Excel's own: horizontal, vertical and
outline borders drawn, and no legend keys.
}
\examples{
xl_chart_table()
xl_chart_table(show_keys = TRUE, vertical_border = FALSE)
}
\seealso{
[xl_chart]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_workbook.R
\name{xl_workbook}
\alias{xl_workbook}
\title{A workbook: sheets plus workbook-level properties}
\usage{
xl_workbook(
  sheets,
  properties = xl_properties(),
  col_names = TRUE,
  format_headers = TRUE
)
}
\arguments{
\item{sheets}{A data frame, an [xl_sheet], or a (named) list of them.}

\item{properties}{An [xl_properties] object.}

\item{col_names}{write column names as the header row at the top of the sheet?}

\item{format_headers}{apply the workbook's header format to that header row?
The default header format is bold and centered; change it with
\code{\link{xl_properties}(header_format = )}.}
}
\value{
An `xl_workbook` object.
}
\description{
`xl_workbook()` binds one or more sheets (data frames or [xl_sheet]s) to a
set of [xl_properties].  It is the single place to attach workbook-level
formatting defaults and metadata.  When passed to [write_xlsx()], the
workbook's `col_names` and `format_headers` settings take precedence over
`write_xlsx()`'s own arguments.
}
\examples{
wb <- xl_workbook(
  list(Data = data.frame(x = 1:3)),
  properties = xl_properties(title = "Demo", author = "me")
)
tmp <- write_xlsx(wb)
}
\seealso{
[xl_properties], [xl_sheet], [write_xlsx]

Other workbook settings:
\code{\link[=write_xlsx]{write_xlsx()}},
\code{\link[=xl_properties]{xl_properties()}}
}
\concept{workbook settings}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_parts.R
\name{xl_chart_error_bars}
\alias{xl_chart_error_bars}
\title{Error bars on a chart series}
\usage{
xl_chart_error_bars(
  type,
  value = NA,
  direction = NULL,
  endcap = NA,
  format = NULL
)
}
\arguments{
\item{type}{How the size of each bar is worked out: `"std_error"` for the
standard error, `"fixed"` for a constant, `"percentage"` of the point's
own value, or `"std_dev"` for that many standard deviations.}

\item{value}{The constant, the percentage, or the number of standard
deviations.  `"std_error"` needs none.}

\item{direction}{`"both"`, `"plus"` or `"minus"`.}

\item{endcap}{Draw the cap at the end of each bar.  On by default.}

\item{format}{An [xl_format()] styling the bars --- [xl_border()] only,
since an error bar is a line.}
}
\value{
An `xl_chart_error_bars` object.
}
\description{
`xl_chart_error_bars()` draws an error bar at each point, and is given to
[xl_chart_series()] as `x_error_bars` or `y_error_bars`.
}
\examples{
xl_chart_error_bars("percentage", 5)
xl_chart_error_bars("std_dev", 1, direction = "plus", endcap = FALSE)
}
\seealso{
[xl_chart_series]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_outline.R
\name{xl_outline}
\alias{xl_outline}
\title{Control how outline (grouping) symbols are drawn}
\usage{
xl_outline(
  visible = TRUE,
  symbols_below = TRUE,
  symbols_right = TRUE,
  auto_style = FALSE
)
}
\arguments{
\item{visible}{Show the outline symbols at all.  `FALSE` keeps the grouping
(and so the collapsing) but hides the +/- controls.}

\item{symbols_below}{Put the summary row *below* the detail rows, which is
Excel's default.  `FALSE` puts it above.}

\item{symbols_right}{Put the summary column to the *right* of the detail
columns, Excel's default.  `FALSE` puts it to the left.}

\item{auto_style}{Apply Excel's automatic outline styling to the grouped
rows and columns.}
}
\value{
An `xl_outline` object, for [xl_sheet()]'s `outline` argument.
}
\description{
Grouping itself comes from `level` in [xl_col_spec()] / [xl_row_spec()].
`xl_outline()` only changes how the controls are *displayed*, which is
rarely needed --- the defaults match Excel's own.
}
\examples{
xl_outline(symbols_below = FALSE)
xl_outline(visible = FALSE)
}
\seealso{
[xl_sheet], [xl_colrow_spec]

Other worksheet layout:
\code{\link{xl_colrow_spec}},
\code{\link[=xl_page_setup]{xl_page_setup()}},
\code{\link[=xl_sheet]{xl_sheet()}},
\code{\link[=xl_sheet_view]{xl_sheet_view()}}
}
\concept{worksheet layout}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart.R
\name{xl_chart}
\alias{xl_chart}
\title{Add a chart to a worksheet}
\usage{
xl_chart(
  type,
  series,
  title = NULL,
  title_format = NULL,
  title_layout = NULL,
  title_overlay = NA,
  x_axis = NULL,
  y_axis = NULL,
  legend = NULL,
  data_table = NULL,
  plot_area_format = NULL,
  plot_area_layout = NULL,
  chart_area_format = NULL,
  drop_lines = NA,
  high_low_lines = NA,
  up_down_bars = NA,
  hole_size = NA,
  rotation = NA,
  series_gap = NA,
  series_overlap = NA,
  show_blanks = NULL,
  show_hidden_data = NA,
  at = "A1",
  scale = 1,
  offset = NULL,
  position = "move_and_size",
  description = NULL,
  decorative = FALSE,
  style = NA
)
}
\arguments{
\item{type}{The chart type: `"column"`, `"bar"`, `"line"`, `"pie"`,
`"doughnut"`, `"area"`, `"scatter"`, `"radar"`, and the stacked,
percent-stacked, smoothed and marker variants.}

\item{series}{One [xl_chart_series()], or a list of them.  Every series of a
scatter chart must have `categories`, which are its x axis.}

\item{title}{The chart title.  A string is always taken literally, so to
take the title from a cell give a range spec ---
`list(header = "revenue")` for a column's header cell, or
`list(rows = 1, cols = 1)` for a data cell.  `FALSE` removes the title
Excel would otherwise generate.}

\item{title_format}{An [xl_format()] styling the title text.  A title is
text, so only the [xl_font()] group applies.}

\item{title_layout}{Where to put the title by hand, as `c(x, y)` fractions
of the chart.  Excel places it for you otherwise.}

\item{title_overlay}{Let the title sit over the plot rather than above it.}

\item{x_axis, y_axis}{An [xl_chart_axis()] describing that axis.  Pie and
doughnut charts have none, and several axis options apply to a value or a
category axis only --- see [xl_chart_axis()].}

\item{legend}{An [xl_chart_legend()] moving, styling or removing the legend.}

\item{data_table}{An [xl_chart_table()] printing the plotted numbers in a
grid beneath the chart.}

\item{plot_area_format, chart_area_format}{An [xl_format()] styling the plot
area --- the panel the data is drawn in --- and the chart area around it:
[xl_border()] for the line, [xl_fill()] for the fill or pattern.}

\item{plot_area_layout}{Where to put the plot area by hand, as `c(x, y)`
or `c(x, y, width, height)` fractions of the chart.}

\item{drop_lines}{Drop lines from each point to the category axis: `TRUE`,
or an [xl_format()] giving the line to draw them with.  Line and area
charts.}

\item{high_low_lines}{A line joining the highest and lowest series at each
category, the same way.  Line charts.}

\item{up_down_bars}{Bars between the first and last series at each category:
`TRUE`, or `list(up = , down = )` with an [xl_format()] for either bar.
Line charts.}

\item{hole_size}{The size of a doughnut's hole, 10 to 90 percent.}

\item{rotation}{Where a pie or doughnut starts, 0 to 360 degrees clockwise
from the top.}

\item{series_gap}{The gap between category groups on a bar or column chart,
0 to 500 percent of a bar's width.}

\item{series_overlap}{How far bars of one category overlap, -100 to 100
percent.  100 stacks them, -100 pushes them apart.}

\item{show_blanks}{What an empty cell does to the plot: leave a `"gap"`,
plot it as `"zero"`, or join across it with `"connected"`.}

\item{show_hidden_data}{Plot data from rows and columns that are hidden.
Excel leaves them out otherwise.}

\item{at}{The cell the chart's top-left corner is anchored to.}

\item{scale}{Scale factor: one number for both axes, or `c(x, y)`.}

\item{offset}{Offset from the anchor cell's corner in pixels, as `c(x, y)`.}

\item{position}{How the chart behaves when rows and columns change size; see
[xl_image()].}

\item{description}{Alt text, for screen readers.}

\item{decorative}{Mark the chart as decorative, so screen readers skip it.}

\item{style}{Excel's built-in chart style, 1--48.}
}
\value{
An `xl_chart` object.
}
\description{
`xl_chart()` builds a chart from one or more [xl_chart_series()] and places
it on a sheet, anchored to a cell.  Pass one or a list of them as
`xl_sheet(chart = )`.

Placement works exactly as it does for [xl_image()] --- `at`, `scale`,
`offset`, `position`, `description` and `decorative` mean the same things,
because libxlsxwriter describes both with the same fields.
}
\section{What a chart type supports}{

Excel silently drops options a chart type cannot use, so writexl refuses them
instead, naming the types that would work.  Pie and doughnut charts have no
axes; only a doughnut has a hole; only pie and doughnut rotate; up-down bars
and high-low lines are line-only; the series gap and overlap are bar and
column only.
}

\examples{
xl_chart("column", xl_chart_series(values = list(cols = "revenue")))
xl_chart("pie", xl_chart_series(values = "Data!B2:B5"), title = "Share")
}
\seealso{
[xl_chart_series], [xl_sheet]

Other images and charts:
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_parts.R
\name{xl_chart_trendline}
\alias{xl_chart_trendline}
\title{A trendline on a chart series}
\usage{
xl_chart_trendline(
  type,
  order = NA,
  period = NA,
  forward = NA,
  backward = NA,
  intercept = NA,
  equation = NA,
  r_squared = NA,
  name = NULL,
  format = NULL
)
}
\arguments{
\item{type}{`"linear"`, `"log"`, `"poly"`, `"power"`, `"exp"` or
`"average"` for a moving average.}

\item{order}{The order of a polynomial fit, 2 or more.  `"poly"` only.}

\item{period}{The number of points a moving average covers, 2 or more.
`"average"` only.}

\item{forward, backward}{How far to project the line beyond the data, in
categories.}

\item{intercept}{Force the line through this value on the y axis.
Exponential, linear and polynomial fits only.}

\item{equation}{Print the fitted equation on the chart.}

\item{r_squared}{Print the R-squared value on the chart.}

\item{name}{The trendline's name in the legend.  Excel generates one
otherwise.}

\item{format}{An [xl_format()] styling the line --- [xl_border()] only,
since a trendline is a line.}
}
\value{
An `xl_chart_trendline` object.
}
\description{
`xl_chart_trendline()` fits a line through a series.

Two of Excel's own restrictions are enforced, because it discards these
rather than complain: a **moving average** has no forecast, no equation and
no R-squared, and an **intercept** applies only to exponential, linear and
polynomial fits.
}
\examples{
xl_chart_trendline("linear", equation = TRUE, r_squared = TRUE)
xl_chart_trendline("poly", order = 3)
xl_chart_trendline("average", period = 2)
}
\seealso{
[xl_chart_series]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_legend]{xl_chart_legend()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_rich_string.R
\name{xl_rich_string}
\alias{xl_rich_string}
\alias{is_xl_rich_string}
\alias{as.character.xl_rich_string}
\title{A cell whose text has several formats}
\usage{
xl_rich_string(...)

is_xl_rich_string(x)

\method{as.character}{xl_rich_string}(x, ...)
}
\arguments{
\item{...}{Runs, in order.  A bare string is taken as an unformatted run; an
[xl_rich_run()] carries its own font.  Lists of either are flattened, so
runs can be assembled programmatically.}

\item{x}{An object to test.}
}
\value{
An `xl_rich_string` object: a list of runs.
}
\description{
`xl_rich_string()` builds the value of a single cell out of differently
formatted runs, so that one cell can read "This is **bold** text".  Pass it
as the `value` of [xl_cell_general()].

Excel requires at least two runs: a string with one format is an ordinary
character value, so pass it as one.

`as.character()` returns the cell's text with the per-run fonts dropped.
}
\examples{
xl_rich_string("This is ", xl_rich_run("bold", xl_font(bold = TRUE)), " text")

# in a cell, with a cell-wide format alongside the per-run fonts
xl_cell_general(
  value  = xl_rich_string("2 H", xl_rich_run("2", xl_font(script = "sub")), "O"),
  format = xl_align(horizontal = "center")
)
}
\seealso{
[xl_rich_run], [xl_cell_general], [xl_font]

Other cell content:
\code{\link[=is_xl_comment]{is_xl_comment()}},
\code{\link[=xl_cell_general]{xl_cell_general()}},
\code{\link[=xl_comment]{xl_comment()}},
\code{\link[=xl_formula]{xl_formula()}},
\code{\link[=xl_rich_run]{xl_rich_run()}}
}
\concept{cell content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_chart_extras.R
\name{xl_chart_legend}
\alias{xl_chart_legend}
\title{A chart's legend}
\usage{
xl_chart_legend(
  position = NULL,
  format = NULL,
  layout = NULL,
  delete_series = NULL
)
}
\arguments{
\item{position}{Where the legend sits: `"right"` (Excel's default),
`"left"`, `"top"`, `"bottom"`, `"top_right"`, the `"overlay_*"` variants
that let the legend sit over the plot, or `"none"` to remove it.}

\item{format}{An [xl_format()] styling the legend text --- the [xl_font()]
group only, since libxlsxwriter gives a legend a font and nothing else.}

\item{layout}{Where to put the legend by hand, as `c(x, y)` or
`c(x, y, width, height)` --- fractions of the chart, each above 0 and at
most 1.  Excel places it for you otherwise.  `at` is a cell everywhere
else in writexl, so a chart's own fractions are a `layout`.}

\item{delete_series}{Series to leave out of the legend, by position: `2`
drops the second series' entry while still plotting it.  This is how a
trendline or a helper series is kept out of the key.}
}
\value{
An `xl_chart_legend` object.
}
\description{
`xl_chart_legend()` moves, styles or removes the legend, and can leave
individual series out of it.
}
\examples{
xl_chart_legend(position = "bottom")
xl_chart_legend(position = "none")
xl_chart_legend(delete_series = 2)
}
\seealso{
[xl_chart]

Other images and charts:
\code{\link[=xl_chart]{xl_chart()}},
\code{\link[=xl_chart_axis]{xl_chart_axis()}},
\code{\link[=xl_chart_error_bars]{xl_chart_error_bars()}},
\code{\link[=xl_chart_labels]{xl_chart_labels()}},
\code{\link[=xl_chart_marker]{xl_chart_marker()}},
\code{\link[=xl_chart_series]{xl_chart_series()}},
\code{\link[=xl_chart_table]{xl_chart_table()}},
\code{\link[=xl_chart_trendline]{xl_chart_trendline()}},
\code{\link[=xl_chartsheet]{xl_chartsheet()}},
\code{\link[=xl_image]{xl_image()}}
}
\concept{images and charts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_sheet_view.R
\name{xl_sheet_view}
\alias{xl_sheet_view}
\title{How a worksheet appears when it opens}
\usage{
xl_sheet_view(
  active = NA,
  selected = NA,
  visible = NA,
  first_tab = NA,
  selection = NULL,
  top_left = NULL,
  hide_zero = NA,
  right_to_left = NA,
  split = NULL
)
}
\arguments{
\item{active}{Logical; make this the tab Excel opens on.  At most one sheet
in a workbook may be active.}

\item{selected}{Logical; include this tab in the selected group.  The active
sheet is always selected.}

\item{visible}{Logical; `FALSE` hides the sheet's tab.  A hidden sheet cannot
be active or selected, the first sheet cannot be hidden unless another is
made active, and at least one sheet must stay visible or Excel will not
open the file.  All four rules are checked before writing.}

\item{first_tab}{Logical; make this the leftmost visible tab in the tab
strip.  Independent of which sheet is active.}

\item{selection}{The cell or range selected when the sheet opens, as an Excel
  reference (`"B2"`, `"B2:D10"`) or a `list(rows = , cols = )` spec.

  Excel also uses the order of a selection's corners to mark which cell in it
  is active; writexl does not expose that, because ranges are normalised by
  the shared range parser, which rejects an inverted range.}

\item{top_left}{The cell scrolled to the top-left of the window when the
sheet opens, as an Excel reference such as `"A5"`.}

\item{hide_zero}{Logical; display zero values as blank cells.}

\item{right_to_left}{Logical; order the columns right to left, for a sheet in
a right-to-left language.}

\item{split}{Split the sheet into scrollable panes with a visible, movable
  divider, given as the cell reference the split sits above and to the left
  of --- `"B3"` splits above row 3 and left of column B.  Mutually exclusive
  with `xl_sheet(freeze = )`, which does the same thing without the divider.

  libxlsxwriter positions a split by distance, in row-height and column-width
  units, not by row and column number.  writexl converts the cell reference
  using the sheet's actual row heights and column widths, so the split lands
  where you asked even after resizing.  Pass
  `list(vertical = , horizontal = )` to give those units directly.

  Note that libxlsxwriter derives the pane's scroll anchor back from that
  distance assuming default row heights, so on a sheet with resized rows or
  columns the divider is placed correctly but the anchor cell may be a row or
  two out.}
}
\value{
An `xl_sheet_view` object.
}
\description{
`xl_sheet_view()` collects a worksheet's tab state and opening view: which
tab is active, selected or hidden, where the sheet is scrolled and selected,
and a few display options.  Pass it as `xl_sheet(view = )`.

None of these affect the cell data.
}
\examples{
xl_sheet_view(active = TRUE, selection = "B2")
xl_sheet_view(visible = FALSE)

df <- data.frame(x = 1:3)
tmp <- write_xlsx(list(
  Summary = xl_sheet(df, view = xl_sheet_view(active = TRUE)),
  Working = xl_sheet(df, view = xl_sheet_view(visible = FALSE))
))
}
\seealso{
[xl_sheet], [xl_page_setup]

Other worksheet layout:
\code{\link{xl_colrow_spec}},
\code{\link[=xl_outline]{xl_outline()}},
\code{\link[=xl_page_setup]{xl_page_setup()}},
\code{\link[=xl_sheet]{xl_sheet()}}
}
\concept{worksheet layout}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xl_workbook.R
\name{xl_properties}
\alias{xl_properties}
\title{Workbook properties, defaults, and metadata}
\usage{
xl_properties(
  title = NA,
  subject = NA,
  author = NA,
  manager = NA,
  company = NA,
  category = NA,
  keywords = NA,
  comments = NA,
  status = NA,
  hyperlink_base = NA,
  created = NA,
  na = NA,
  custom = NULL,
  read_only = FALSE,
  window_size = NULL,
  names = NULL,
  default_format = xl_format(),
  header_format = xl_font(bold = TRUE) + xl_align(horizontal = "center"),
  hyperlink_format = xl_font(color = "blue", underline = "single"),
  date_format = xl_num_format("yyyy-mm-dd"),
  datetime_format = .default_datetime_format(),
  date_col_width = 20,
  datetime_col_width = 20,
  header_row_height = 15
)
}
\arguments{
\item{title, subject, author, manager, company, category, keywords, comments, status, hyperlink_base}{Document metadata strings (Excel's "Properties" dialog).}

\item{created}{The workbook's creation timestamp, a `Date` or `POSIXct`.
Left `NA` libxlsxwriter stamps the moment the file is written, which is
what makes two runs over the same data differ; pinning it makes them
byte-identical.}

\item{na}{What to write where a value has none --- an `NA` or `NaN` in the
data, or a cell with nothing in it at all. `NA` (the default) leaves the
cell blank, which is what writexl has always done. Anything else is
written in its place, keeping its own type: `na = "Not available"` writes
a string, `na = 0` a number. A column or an individual cell can override
it --- see [xl_col_spec()] and [xl_cell_general()]. Note that a non-blank
`na` in a numeric column makes that column mixed, so a reader such as
[readxl::read_xlsx()] returns the whole column as character.}

\item{custom}{A named list of custom document properties.  Values may be
character, integer, numeric, logical, `Date` or `POSIXct`.  A `Date` or
`POSIXct` is written as a real datetime property (not as text) and follows
the same workbook-wide time zone rule as datetime cells, described below.}

\item{read_only}{Logical; mark the workbook read-only recommended.}

\item{window_size}{Optional integer vector `c(width, height)` for the
workbook window size.}

\item{names}{A named list of workbook-scoped defined names, each a formula
string (e.g. `list(tax = "=0.2")`).}

\item{default_format}{An [xl_format] cascaded under every cell (default:
none).}

\item{header_format}{An [xl_format] for the header row (default: bold,
centered).}

\item{hyperlink_format}{An [xl_format] for cell hyperlinks (default: blue,
underlined), or `NULL` for no hyperlink styling at all.  `NULL` is the only
way to write an unstyled hyperlink: an empty `xl_format()` leaves the cell
with no format, and Excel files written that way fall back to
libxlsxwriter's own blue-underlined default.}

\item{date_format, datetime_format}{[xl_format] number formats applied to
`Date` / `POSIXct` values.}

\item{date_col_width, datetime_col_width}{Default column width for `Date` /
`POSIXct` columns.}

\item{header_row_height}{Height (in points) of the header row.}
}
\value{
An `xl_properties` object.
}
\description{
`xl_properties()` collects everything that applies at the *workbook* level:
document metadata, a few native workbook settings, and the formatting
defaults that used to be hard-coded.  The formatting defaults are ordinary
[xl_format] objects, so you can override them (e.g. change the header style
or the default date format) simply by passing a different `xl_format`.

The `default_format` is cascaded *under* every cell (an emulated
workbook-wide default: libxlsxwriter has no native "Normal style" setter, so
it is merged beneath each cell/column format).  `header_format` styles the
header row and `hyperlink_format` styles cell hyperlinks; both are also
cascaded over `default_format`.
}
\section{Time zones}{

Excel has no concept of a time zone.  When every `POSIXct` in the workbook
shares one time zone, writexl drops the zone and writes local wall-clock
time, and the default `datetime_format` loses its `" UTC"` suffix so that
nothing is mislabelled.  When the time zones differ, all datetimes are
converted to UTC with a warning.  Supplying your own `datetime_format`
overrides the label in either case.
}

\examples{
xl_properties(title = "Quarterly report", author = "Finance",
              header_format = xl_font(bold = TRUE, color = "white") +
                              xl_fill(background = "navy"))
}
\seealso{
[xl_workbook], [write_xlsx]

Other workbook settings:
\code{\link[=write_xlsx]{write_xlsx()}},
\code{\link[=xl_workbook]{xl_workbook()}}
}
\concept{workbook settings}
